% no linearization yet.
% just linear terms are allowed

:- module(lpbineq,[
	'#='/2,
	'#>='/2,
	'#<='/2,
	'#>'/2,
	'#<'/2,
	pb_max/2,    % just calculates Max
	pb_min/2,    % just calculates Min
	pb_maxass/2, % asserts that Objf = Max
	pb_minass/2, % asserts that Objf = Min
	pb_enum/1, % standard labeling; not recommended
        collect_coefflits/5,
	simply_consistent/2
]).

:- use_module(library(lists)).
:- use_module(interface).

'#='(Left,Right) :-
	'#>='(Left,Right),
	'#>='(Right,Left).
'#>'(Left,Right) :- '#>='(Left,Right+1).
'#<'(Left,Right) :- '#>='(Right,Left+1).
'#<='(Left,Right) :- '#>='(Right,Left).
'#>='(Left,Right) :-
	collect_coefflits(Left-Right,[],CoeffLits,0,Const),
	Rhs is -Const,
	simply_consistent(CoeffLits,Rhs),
%	format("adding ~w~n",[add_lpbis([lpbi(CoeffLits,Rhs)])]),
	pb_caddlpbis([lpbi(CoeffLits,Rhs)]).

simply_consistent(_,Rhs) :-
	Rhs =< 0,!.
simply_consistent([C-_|CoeffLits],Rhs) :-
	(C =< 0 ->
	    simply_consistent(CoeffLits,Rhs)
	;
	    Rhs1 is Rhs - C,
	    simply_consistent(CoeffLits,Rhs1)
	).

pb_enum([]).
pb_enum([V|L]) :-
	'#>='(V,1),
	pb_enum(L).
pb_enum([V|L]) :-
	'#>='('~'(V),1),
	pb_enum(L).

pb_max(Objf,Max) :-
	collect_coefflits(Objf,[],CoeffLits,0,Const),
	pb_maximize(CoeffLits,Max0),
	Max is Max0 + Const.

pb_min(Objf,Min) :-
	collect_coefflits(Objf,[],CoeffLits,0,Const),
	pb_minimize(CoeffLits,Min0),
	Min is Min0 + Const.

pb_maxass(Objf,Max) :-
	pb_max(Objf,Max),
	'#='(Objf,Max).

pb_minass(Objf,Min) :-
	pb_min(Objf,Min),
	'#='(Objf,Min).

collect_coefflits(T,L0,L1,R0,R1) :-
	var(T),!,
	L1 = [1-T|L0],
	R1 = R0.
collect_coefflits(T,L0,L1,R0,R1) :-
	integer(T),!,
	L1 = L0,
	R1 is R0 + T.
collect_coefflits('~'(T),L0,L1,R0,R1) :-
	var(T),!,
	L1 = [1- '~'(T)|L0],
	R1 = R0.
collect_coefflits('~'(T),L0,L1,R0,R1) :-
	integer(T),!,
	(T == 0; T == 1),
	(T == 1 ->
	    R1 is R0
	;
	    R1 is R0 + 1
	),
	L0 = L1.
collect_coefflits(T1*T2,L0,L1,R0,R1) :-
	constant(T1,C1),!,
	(constant(T2,C2) -> 
	    R1 is R0 + C1 + C2,
	    L0 = L1
	;
	    collect_coefflits(T2,[],L3,0,C2),
	    mult_const(L3,C1,L4),
	    R1 is R0 + C1*C2,
	    append(L4,L0,L1)
	).
collect_coefflits(T1*T2,L0,L1,R0,R1) :-
	constant(T2,C2),!,
	collect_coefflits(T1,[],L3,0,C1),
	mult_const(L3,C2,L4),
	R1 is R0 + C2*C1,
	append(L4,L0,L1).
collect_coefflits(_T1*_T2,_L0,_L1,_R0,_R1) :-
	format("nonlinear terms not yet supported~n failing ~n",[]),
	fail.
collect_coefflits(T1 + T2,L0,L2,R0,R2) :-
	collect_coefflits(T1,L0,L1,R0,R1),
	collect_coefflits(T2,L1,L2,R1,R2).
collect_coefflits(T1 - T2,L0,L4,R0,R3) :-
	collect_coefflits(T1,L0,L1,R0,R1),
	collect_coefflits(T2,[],L2,0,R2),
	mult_const(L2,-1,L3),
	append(L3,L1,L4),
	R3 is R1 - R2.


mult_const([],_,[]).
mult_const([C0-L0|R0],M,[C1-L0|R1]) :-
	C1 is C0*M,
	mult_const(R0,M,R1).
	
constant(T,C) :- integer(T),!,T = C.
constant(T,_C) :- var(T),!,fail.
constant('~'(T),_C) :- var(T),!,fail.
constant(T1*T2,C) :-
	constant(T1,C1),
	constant(T2,C2),
	C is C1 * C2.
constant(T1+T2,C) :-
	constant(T1,C1),
	constant(T2,C2),
	C is C1 + C2.
constant(T1-T2,C) :-
	constant(T1,C1),
	constant(T2,C2),
	C is C1 - C2.

/* pretty printing of constraints lists */

:-multifile user:portray/1.  % hg

user:portray(X) :- 
	nonvar(X),
	X = [E|Rest],
	nonvar(E),
	functor(E,lpbi,2),!,
	print_lpbis([E|Rest]).

print_lpbis([]).
print_lpbis([lpbi(CL,R)]) :-
	!,
	print_cl(CL),
	format(" #>= ~w ",[R]).
print_lpbis([lpbi(CL,R)|Rest]) :-
	print_cl(CL),
	format(" #>= ~w~n",[R]),
	print_lpbis(Rest).

print_cl([]).
print_cl([C-L]) :-
	!,
	write(C),write('*'),
	print_lit(L).
print_cl([C-L|R]) :-
	!,
	write(C),write('*'),
	print_lit(L),write(' + '),
	print_cl(R).

print_lit(L) :- write(L).
