% test pighole in clppb
pig(M,N) :-
	make_mat(M,N,Mat),
	mat_each_line_at_least(Mat,1),
	mat_transpose(Mat,TMat),
	mat_each_line_at_most(TMat,1).
	


mat_each_line_at_least([],_).
mat_each_line_at_least([Vec|Mat],N) :-
	vec_at_least(Vec,N),
	mat_each_line_at_least(Mat,N).

vec_at_least([V|Rest],N) :- % at least on var else does not work
	vec_at_least(Rest,V,T),
	write(T #>= N),nl,
	T #>= N.
vec_at_least([],T1,T1).
vec_at_least([V|Rest],T1,T2 + V) :-
	vec_at_least(Rest,T1,T2).
	

mat_each_line_at_most([],_).
mat_each_line_at_most([Vec|Mat],N) :-
	vec_at_most(Vec,N),
	mat_each_line_at_most(Mat,N).

vec_at_most([V|Rest],N) :- % at most on var else does not work
	NN is 1-N,
	vec_at_most(Rest,~V,T,NN,R),
	write(T #>= R),nl,
	T #>= R.
vec_at_most([],T1,T1,R,R).
vec_at_most([V|Rest],T1,T2 + ~V,R0,R2) :-
	R1 is R0 + 1,
	vec_at_most(Rest,T1,T2,R1,R2).
	




mat_transpose([],[]).
mat_transpose([Vec|Mat],TMat) :-
	mat_transpose(Vec,1,Mat,TMat).

mat_transpose([],_,_,[]).
mat_transpose([E|Vec],I,Mat,[[E|Row]|TMat]) :-
	vec_row_i(Mat,I,Row),
	I1 is I + 1,
	mat_transpose(Vec,I1,Mat,TMat).

vec_row_i([],_,[]) :- !.
vec_row_i([Vec|Mat],I,[E|Rest]) :- 
	vec_i(Vec,E,I),
	vec_row_i(Mat,I,Rest).

	
vec_i([E|_],E,1) :- !.	
vec_i([_|V],E,I) :- 
	I1 is I-1,
	vec_i(V,E,I1).
	
make_mat(0,_N,[]) :- !.
make_mat(M,N,[Vec|Mat]) :-
	M > 0,
	make_vec(N,Vec),
	M1 is M-1,
	make_mat(M1,N,Mat).

make_vec(0,[]) :- !.
make_vec(N,[_Var|Vec]) :-
	N > 0,
	N1 is N-1,
	make_vec(N1,Vec).

