% -*- Module: formulas; -*-

:- module(formulas,[
		    complete_eq/1,
		    equation_template/1,
		    formula2term/3,
		    left/2,
		    left/3,
		    inconsistency/1,
		    mk_equation/2,
		    mk_equation/5,
		    mk_equation/8,
		    number/2,
		    number/3,
		    orientation/2,
		    orientation/3,
		    proof/2,
		    proof/3,
		    right/2,
		    right/3,
		    rightr/2,
		    rightr/3,
		    lsize/2,
		    lsize/3,
		    rsize/2,
		    rsize/3,
%		    skolemized/3,
		    type/2,
		    type/3,
		    weight/2,
		    weight/3
		   ]).

:- use_module(library(change_arg),[change_arg/4]).

:- use_module(satsrc(myterms),[
	mapargs/3
   ]).

:- use_module(global_vars, [
        cont/2,
	store/2
   ]).

:- use_module(st_terms).



/* -----------------------------------------------------------------------------

Equation    ::= eq(Left,Right,Number,Weight,Type,Proof,Orientation,LSize,RSize)
Left        ::= Term                             % left side
Right       ::= Term                             % right side
Number      ::= nat
Weight, LSize, RSize   ::= nat
Orientation ::= lr                           % uniformly oriented from l to r
             |  rl                           % uniformly oriented from r to l
	     |  check                        % check orientation for instances


Upon creation of an equation, L,R,Type, and Proof have to be provided.
All other attributes will be computed lazily. Note, however, that
whenever on calls, say,

  number(E,N)

with N instantiated, E will be assigned the number N if it didn't
have a number assigned already. 



----------------------------------------------------------------------------- */


equation_template(eq(_,_,_,_,_,_,_)).

mk_equation(L,R,N,S,T,P,O,eq(L,R,N,S,T,P,O,_,_)).

mk_equation(L,R,T,P,eq(L,R,_N,_S,T,P,_O,_,_)).
				% only minimal information fixed

mk_equation(eq(L,R,T,P),eq(L,R,_N,_S,T,P,_O,_,_)).
mk_equation(eq(L=R,T,P),eq(L,R,_N,_S,T,P,_O,_,_)).



formula2term(E,L=R,Ty):-
	left(E,L),
	right(E,R),
	type(E,Ty).





	
complete_eq(R):-
	number(R,_),
	weight(R,_),
	orientation(R,_).
	

orient(L,'$false',lr):-
	L\=='$false',
	!.
orient('$false',L,rl):-
	L\=='$false',
	!.
orient(T,T,check):-
	!.
orient(L,R,Orientation):-
	cont(ord,ord(OType,_OParams)),
	OType:extend_ordering(L,R),
	cont(ord,ord(OType,OParams)),
	(OType:greater(OParams,L,R) ->
	    Orientation=lr
	;OType:greater(OParams,R,L) ->
	    Orientation=rl
	;
	    Orientation=check
	),
	!.

	
left(E,L):-
	arg(1,E,L).

left(E,E1,L):-
	change_arg(1,E,E1,L).

right(E,R):-
	arg(2,E,R).

right(E,E1,R):-
	change_arg(2,E,E1,R).


number(E,N):-
	arg(3,E,N),
	(var(N) ->
	    cont(eq_ctr,M),
	    N is M+1,
%	    format('~Nnew number:~p',[E]),
	    store(eq_ctr,N)
	;
	    true).

number(E,E1,N):-
	change_arg(3,E,E1,N).

weight(E,S):-
	arg(4,E,S),
	(var(S) ->
	    left(E,L),
	    right(E,R),
	    orient(L,R,O),
	    (O\==check ->
		(O=lr ->
		    L1=L,
		    R1=R
		; %O=rl 
		    L1=R,
		    R1=L
		),
		term_weight(L1,SL),
		term_weight(R1,SR),
		S is SL*(SL+SR)	% Waldmeister weight function
	    ;
		term_weight(L,SL),
		term_weight(R,SR),
		S is max(SL,SR)*(SL+SR)
	    )
	;
	    true
	).


weight(E,E1,N):-
	change_arg(4,E,E1,N).

type(E,N):-
	arg(5,E,N).

type(E,E1,N):-
	change_arg(5,E,E1,N).

proof(E,N):-
	arg(6,E,N).

proof(E,E1,N):-
	change_arg(6,E,E1,N).

orientation(Ru,O):-
	arg(7,Ru,O),
	(var(O) ->
	    left(Ru,L),
	    right(Ru,R),
	    orient(L,R,O)
	;
	    true
	).

orientation(R,R1,O):-
	change_arg(7,R,R1,O).


lsize(E,S):-
	arg(8,E,S),
	(var(S) ->
	    left(E,L),
	    term_size(L,S)
	;
	    true
	).


lsize(E,E1,N):-
	change_arg(8,E,E1,N).

rsize(E,S):-
	arg(9,E,S),
	(var(S) ->
	    right(E,R),
	    term_size(R,S)
	;
	    true
	).


rsize(E,E1,N):-
	change_arg(9,E,E1,N).



/* ----------------------------------------------------------------------------

                Term Size Computation


----------------------------------------------------------------------------- */

term_weight( T, 0.7 ):-
        var_(T), !.
%/*
term_weight(T,0.7):-
        object_var_(T), !.
%*/
term_weight('$neq'(L,R),W):-
	term_weight(L,W1),
	term_weight(R,W2),
	W is max(W1,W2)/2.
term_weight( T, S ):-
	functor_(T,_O,N),
	term_weight_from_to(0,N,T,S1),
	S is S1 + 1.


term_weight_from_to(I,N,T,S):-
	(I=N ->
	     S=0
	;
	     I1 is I+1,
	     arg_(I1,T,T1),
	     term_weight(T1,S1),
	     term_weight_from_to(I1,N,T,S2),
	     S is S1+S2
	).



term_size( T, 1 ):-
        var_(T), !.
term_size(T,1):-
        object_var_(T), !.
term_size( T, S ):-
	functor_(T,_O,N),
	term_size_from_to(0,N,T,S1),
	S is S1 + 1.


term_size_from_to(I,N,T,S):-
	(I=N ->
	     S=0
	;
	     I1 is I+1,
	     arg_(I1,T,T1),
	     term_size(T1,S1),
	     term_size_from_to(I1,N,T,S2),
	     S is S1+S2
	).






/* ----------------------------------------------------------------------------

                Testing

----------------------------------------------------------------------------- */

inconsistency(eq('$true','$false',_,_,_,_,_,_,_)).
inconsistency(eq('$false','$true',_,_,_,_,_,_,_)).


/* ----------------------------------------------------------------------------

                Printing

----------------------------------------------------------------------------- */


:- multifile user:portray/1.


user:portray(eq(Left,Right,Number,_Size,_Type,_Proof,Orientation,_LS,_RS)):-
	(Orientation==rl ->
	    L=Right,
	    R=Left
	;
	    L=Left,
	    R=Right
	),
	(var(Number) ->
	    \+ \+ (   numbervars((L,R),0,_),
%		      format('~t  ~7| ~p = ~p [~p,~p]',[L,R,_LS,_RS]))
		      format('~t  ~7| ~p = ~p',[L,R]))
	;
	    \+ \+ (   numbervars((L,R),0,_),
%		      format('~t~d :~7| ~p = ~p [~p,~p]',[Number,L,R,_LS,_RS]))
		      format('~t~d :~7| ~p = ~p',[Number,L,R]))
	).
