%declarations%
:- module(lpo, [
	greater/3,
	greater_eq/3,
	init_ordering/1
   ]).

:- use_module(st_terms).
:- use_module(satsrc(auxiliary),[member212/2]).
:- use_module(satsrc(mylists),[is_list/1]).
:- use_module(library(maplist),[maplist/2]).
:- use_module(satsrc(thisProlog),[prolog_version/1]).

:- ((prolog_version(sicstus);prolog_version(sicstus3)) ->
	use_module(library(ugraphs))
   ;
	use_module(slibrary(ugraphs))
   ).

:- use_module(global_vars, [
        cont/2,
	store/2
   ]).


:- ensure_loaded(error_handling).

%end%








testcase1:-
	format('~N should succeed~n',[]),
	prolog2term(i(X*Y),IXY),
	prolog2term(i(X)*i(Y),IXIY),
	greater(prec([],[i,'*',e]),IXY,IXIY).

testcase2:-
	format('~N should fail~n',[]),
	prolog2term(i(X*Y),IXY),
	prolog2term(i(X)*i(Y),IXIY),
	greater(prec([],['*',i,e]),IXY,IXIY).

testcase3:-
	format('~N should fail~n',[]),
	prolog2term(i(X*_Y),IXY),
	prolog2term(i(X)*i(_Z),IXIZ),
	greater(prec([],[i,'*',e]),IXY,IXIZ).

testcase4:-
	format('~N should succeed~n',[]),
	prolog2term(i(a*b),IXY),
	prolog2term(i(a)*i(b),IXIY),
	greater(prec([],[i,'*',e]),IXY,IXIY).


testcase5:-
	format('~N should succeed~n',[]),
	prolog2term(p(i(X*_Y)),IXY),
	prolog2term(i(X)*i(_Z),IXIZ),
	greater(prec([p],[p,i,'*',e]),IXY,IXIZ).


init_ordering(_).

%------------------------------------------------------------------------------
% greater(+P,+S,+T):
%	S is greater than T with respect to the lexicographic path ordering
%       given by precedence P
%  The terms S and T may contain function symbols not mentioned
%  in the precedence P. These functions are assumed to be smaller
%  than any function mentioned in P. Function symbols not
%  contained in P are compared in the Prolog term order.
%
%  Hence, greater can be used also for terms with (Skolem) constants
%  among which the precedence is irrelevant.
%  In other words, greater is a total ordering on ground terms with Skolem
%  constants.
%
%  Terms can both be Prolog terms with Prolog variables, or st_terms with
%  object variables. Mixed terms should not appear. Prolog variables
%  and object variables are treated the same ordering-wise.
%------------------------------------------------------------------------------

/*
greater( P, S, T):-
	pre((
	     is_precedence(P),
	     is_term(S),
	     is_term(T)
	    ),
	    greater( P, S, T)
	   ).
*/
greater( _,  S, _ ):-
	var_(S),
	!,
	fail.
greater( _,  S, _ ):-
	object_var_(S),
	!,
	fail.
/*
greater( _P, S, T ):-
	nonvar_(T),
	S=f('$VAR'(N),'$VAR'(M)),
	nonvar(N),nonvar(M),
	T=f('$VAR'(M),'$VAR'(N)),
	!,
	N>M.
*/
greater( P, S, T ):-
	nonvar_(T),
	\+ object_var_(T),
	!,
	functor_(S,F,N),
	functor_(T,G,M),
	greater(P,F,G,N,M,S,T).
greater( P, S, T ):-
	% T is a variable
	functor_(S,F,_),
	(is_top_predicate(P,F) -> 
	     true
	;    contains_var(T,S)
	),
	!.


greater(P,F,F,N,M,S,T):-
	!,
        (is_top_predicate(P,F) ->
             lexicographic1(P,0,S,N,T,M)
        ;
             lexicographic(P,0,S,N,T,M)
	).
greater(P,F,G,N,M,S,T):-
        (greater_op(P,F,G)->
             greater_than_list(P,S,0,T,M)

        ; % greater_op(P,G,F)
             some_subterm_greater_equal(P,0,S,N,T)
        ),
        !.



some_subterm_greater_equal(_P,N,_S,N,_T):-
	!,
	fail.
some_subterm_greater_equal(P,I,S,N,T):-
	I1 is I+1,
	arg_(I1,S,SI1),
	(greater_eq( P, SI1, T) ->
	     true
	;
	     some_subterm_greater_equal(P,I1,S,N,T)
	).

	




lexicographic1(_P,N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic1(P,I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg_(I1,S,SI),   % otherwise: N=<M
	(arg_(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic1(P,I1,S,N,T,M)
	     ;
	          greater(P,SI,TI)
	     )
	;% M<N
	     true
	).
	

lexicographic(_P,N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic(P,I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg_(I1,S,SI),   % otherwise: N=<M
	(arg_(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic(P,I1,S,N,T,M)
	     ;greater(P,SI,TI) ->
		  greater_than_list(P,S,I1,T,M)
	     ;
                  some_subterm_greater_equal(P,I1,S,N,T)
	     )
	;% M<N     
	     true
	).

greater_than_list(_P,_,N,_,N):-
	!.
greater_than_list(P,S,I,T,M):-
	I1 is I+1,
	arg_(I1,T,TI),
	greater(P,S,TI),
	greater_than_list(P,S,I1,T,M).




%------------------------------------------------------------------------------
% greater_eq( P, S, T ):
%	S is greater than T with respect to lpo(P) or S is equal to T
%------------------------------------------------------------------------------

greater_eq(_,T1,T2):- T1 == T2,!.
%greater_eq(_,_T1,T2):- '$min' == T2,!.
%greater_eq(T1,T2):- 	'$true$' == T2,	T1\=='$min'.
greater_eq(P,T1,T2):- greater(P,T1,T2),!.






/* ----------------------------------------------------------------------------

               Operations on the precedence
	       
----------------------------------------------------------------------------- */

is_precedence(P):-
	nonvar(P),
	P=prec(TP,Pr),
	is_list(TP),
	is_list(Cons),
	maplist(is_function_symbol,TP),
	maplist(is_function_symbol,Pr).


%------------------------------------------------------------------------------
% is_top_predicate( +P, +F ):
%  succeeds whenever F occurs in the TopPrecidates part of
%  P=prec(TopPredicates,Pr).
%------------------------------------------------------------------------------

is_top_predicate( prec(TopPreds,_), F ):-
	member212(F,TopPreds).




% greater_op( +P, +F , +G)
%    succeeds whenever F occurs before G in the precedence list Pr of
%    P=prec(TopPredicates,Pr).
%
%    It is assumed that F is different from G.
%    If a function symbol does not occur in P, it is assumed to be smaller
%    than any function symbol occurring in P.
%
%    If neither F nor G appear in P we use the Prolog term ordering.
%
%    Note that the ordering in TopPredicates is irrelevant, and it is assumed
%    that TopPredicates also occur in (the initial segment of) Pr.
%
%    Should we postulate the existence of a minimal constant???
/*
greater_op(Prec,F,G):-
	pre((
	     is_precedence(Prec),
	     is_function_symbol(F),
	     is_function_symbol(G),
	     F\==G
	    ),
	    greater_op(Prec,F,G)
	   ).
*/
greater_op(prec(_TopPreds,Prec),F,G):-
	greater_in_prec(Prec,F,G).




% we use == for membership tests so as to avoid instantiation
% of type variables

greater_in_prec([F1|_Fs],F,_G):-
	F==F1,
	% we found F first, hence its the greater one among F and G
	!.
greater_in_prec([G1|_Fs],_F,G):-
	G==G1,
	% we found G first 
	!,
	fail.
greater_in_prec([_|Fs],F,G):-
	greater_in_prec(Fs,F,G).
greater_in_prec([],F,G):-
	% we found neither, so we assume the Prolog term order
	F @> G.




/* ------------------------------------------------------------------------------

 Ordered structures

------------------------------------------------------------------------------ */

extend_ordering(_,_).

