:- module(rewrite,[
	normalform/4,
	normalform1/4,
        reducible/2
   ]).


:- use_module(library(maplist),[maplist/2,maplist/3]).

:- use_module(st_terms).
:- use_module(formulas).
:- use_module(theory,[match_active/4]).

:- use_module(global_vars, [
        cont/2,
        incr/1,
	store/2,
        timer/2
   ]).



/* ------------------------------------------------------------------------------

normalform1(+S,?T,?Used,+Rules)
succeeds if S is reducible, and if T is the normaform of S
with respect to the rewrite rules in Rules.
If reducible, Used is the list of rules applied to normalize S.


normalform(+S,?T,?Used,+Rules)
always succeeds with T the normalform of S wrt Rules.


Having both variants at the same time allows one to avoid
the copying of irreducible terms. We might eventually change this
to using exceptions.


------------------------------------------------------------------------------ */


normalform(S,T,U1/U2,Rs):-
	normalform(S,S,T,U1/U2,Rs).


normalform1(S,T,Us,Rs):-
	normalform1(S,S,T,Us,Rs).

reducible(S,Rs):-
	on_exception(reducible,
		     (   store(irreducibility_check,true),
			 normalform1(S,S,_T,_Us,Rs)
		     ),
		     true),
	store(irreducibility_check,false),
	!.
reducible(_S,_Rs):-
	store(irreducibility_check,false),
	fail.
	


/* ----------------------------------------------------------------------------
   normalform[1](S,S1,T,Used,Rs) 

is like

   normalform[1](S,T,Used,Rs), 

where S1 marks irreducible positions in S by '$VAR'(_).
This is supposed to take advantage from the fact that
we normalize terms bottom-up.

---------------------------------------------------------------------------- */

normalform(S,S1,T,U1/U2,Rs):-
	(normalform1(S,S1,T,U1/U2,Rs) -> 
	     true
	;    
	     U1=U2,
	     T=S
	).


normalform1(_S,'$VAR'(_),_T,_Us,_Rs):-
	% S is irreducible, as indicated by the ``$VAR(_)'' in the S1-argument
	!,
	fail.
normalform1(S,SR,T,U1/U2,Rs):-
	nonvar_(S),
	!,
	(functor_(S,F,N) ->
	     (normalformArgs1(N,S,SR,S1,F,N,U1/U3,Rs) ->
		  (one_step_at_root(S1,T1,T1R,U3/U4,Rs) ->
		      normalform(T1,T1R,T,U4/U2,Rs)
		  ;
		      T=S1,
		      U3=U2
		  )
	     ; % subterms of S are irreducible

		  (one_step_at_root(S,T1,T1R,U1/U3,Rs) ->
		       normalform(T1,T1R,T,U3/U2,Rs)
		  ; % S is irreducible, hence, fail
		   fail
		   )
	     )
	).

/* ----------------------------------------------------------------------------
   normalformArgs1(+I,+S,+S1,+T,+F,+N,+Rs) ist wahr, wenn die Subterme
   T[j] von T, 1=<j<=I, die Normalformen der Subterme S[j] von T
   sind. Zudem mu\3 mindestens einer der Terme S[j] reduzibel sein,
   und es mu\3 F das Topsymbol von S und T sein.
   Context provides constraint and clause context for the reduction.
   S1 marks irreducible positions in S.

---------------------------------------------------------------------------- */


normalformArgs1(I,S,SR,T,F,N,U1/U2,Rs):-
	I>0,
	arg_(I,S,SI),
	arg_(I,SR,SRI),
	I1 is I-1,
	(normalform1(SI,SRI,TI,U1/U3,Rs) ->     % irgendwo in SI wurde eine
	                                 % Regel angewendet
	     functor_(T,F,N),            % neues F-Oberteil
	     arg_(I,T,TI),               % die Normalform TI von SI einhaengen
	     sameSubterms(I,N,S,T),      % die alteren Brueder von SI in S 
                                         % in T einhaengen
	     normalformArgs(I1,S,SR,T,U3/U2,Rs)
	;    
	     normalformArgs1(I1,S,SR,T,F,N,U1/U2,Rs)).



sameSubterms(N,N,_,_).
sameSubterms(I,N,S,T):-
	I<N,
	I1 is I+1,
	arg_(I1,S,Sub),
	arg_(I1,T,Sub),
	sameSubterms(I1,N,S,T).


normalformArgs(0,_,_,_,U/U,_Rs):- !.
normalformArgs(I,S,SR,T,U1/U3,Rs):-
	I>0,
	I1 is I-1,
	arg_(I,S,SI),
	arg_(I,SR,SRI),
	arg_(I,T,TI),
	normalform(SI,SRI,TI,U1/U2,Rs),
	normalformArgs(I1,S,SR,T,U2/U3,Rs).
	




one_step_at_root(Sub,R,RR,[N|Ns]/Ns,Rs):-
	incr(match_ctr),
	match_active(Rs,Sub,R,Ru),
	incr(red_ctr),
	right(Ru,RR),
	number(Ru,N),
	!.








