%declarations%
:- module(lpoc, [
	greater/3,
	greater_eq/3,
	init_ordering/1,
	newconstant/1
   ]).

:- use_module(st_terms).
:- use_module(satsrc(auxiliary),[member212/2]).
:- use_module(satsrc(mylists),[is_list/1]).
:- use_module(library(maplist),[maplist/2,maplist/3]).
:- use_module(satsrc(thisProlog),[prolog_version/1]).

:- ((prolog_version(sicstus);prolog_version(sicstus3)) ->
	use_module(library(ugraphs))
   ;
	use_module(slibrary(ugraphs))
   ).

:- use_module(global_vars, [
        cont/2,
        incr/1,
	store/2
   ]).


:- ensure_loaded(error_handling).

%end%






init_ordering(prec(TP,Prec,Consts,Conss)):-
	% complete Prec so that it totally orderes all function symbols
	% in TP, Prec, Consts, and Conss. These lists  have to be pw disjoint
	% sets.
	append([TP,Prec,Conss],V),
	
	edges(TP,Prec,E1),
	edges(TP,Conss,E3),
	
	edges(Prec,Conss,E5),

	precedenceAsEdges(TP,E7),
	precedenceAsEdges(Prec,E8),
	precedenceAsEdges(Conss,E10),
	
	append([E1,E3,E5,E7,E8,E10],E),

	vertices_edges_to_ugraph(V,E,Graph),
	top_sort(Graph,Prec1),

	vertices_edges_to_ugraph(Consts,[],G),  % now a ugraph; ordering in Consts
	                                        % irrelevant
	store(ord,ord(lpoc,prec(TP,Prec1,(Consts,G),Conss))).


edges([],_,[]):-
	!.
edges([X|Xs],Ys,E):-
	!,
	edges(X,Ys,E1),
	edges(Xs,Ys,E2),
	user:append(E1,E2,E).
edges(_X,[],[]).
edges(X,[Y|Ys],[X-Y|E]):-
	edges(X,Ys,E).


%------------------------------------------------------------------------------
% greater(+P,+S,+T):
%	S is greater than T with respect to the lexicographic path ordering
%       given by precedence P
%  The terms S and T may contain function symbols not mentioned
%  in the precedence P. These functions are assumed to be smaller
%  than any function mentioned in P. Function symbols not
%  contained in P can at most be constants and are assumed to lie between
%  Prec and Conss in the precedence.
%
%  Hence, greater can be used also for terms with (Skolem) constants
%  among which the precedence is irrelevant.
%  In other words, greater is a total ordering on ground terms with Skolem
%  constants.
%
%  Terms can both be Prolog terms with Prolog variables, or st_terms with
%  object variables. Mixed terms should not appear. Prolog variables
%  and object variables are treated the same ordering-wise.
%------------------------------------------------------------------------------

/*
greater( P, S, T):-
	pre((
	     is_precedence(P),
	     is_term(S),
	     is_term(T)
	    ),
	    greater( P, S, T)
	   ).
*/
greater( _,  S, _ ):-
	var_(S),
	!,
	fail.
greater( _,  S, _ ):-
	object_var_(S),
	!,
	fail.
/*
greater( _P, S, T ):-
	nonvar_(T),
	S=f('$VAR'(N),'$VAR'(M)),
	nonvar(N),nonvar(M),
	T=f('$VAR'(M),'$VAR'(N)),
	!,
	N>M.
*/
greater( P, S, T ):-
	nonvar_(T),
	\+ object_var_(T),
	!,
	functor_(S,F,N),
	functor_(T,G,M),
	greater(P,F,G,N,M,S,T).
greater( P, S, T ):-
	% T is a variable
	functor_(S,F,_),
	(is_top_predicate(P,F) -> 
	     true
	;    contains_var(T,S)
	),
	!.


greater(P,F,F,N,M,S,T):-
	!,
        (is_top_predicate(P,F) ->
             lexicographic1(P,0,S,N,T,M)
        ;
             lexicographic(P,0,S,N,T,M)
	).
greater(P,F,G,N,M,S,T):-
        (greater_op(P,F,G)->
             greater_than_list(P,S,0,T,M)

        ; % greater_op(P,G,F)
             some_subterm_greater_equal(P,0,S,N,T)
        ),
        !.



some_subterm_greater_equal(_P,N,_S,N,_T):-
	!,
	fail.
some_subterm_greater_equal(P,I,S,N,T):-
	I1 is I+1,
	arg_(I1,S,SI1),
	(greater_eq( P, SI1, T) ->
	     true
	;
	     some_subterm_greater_equal(P,I1,S,N,T)
	).

	




lexicographic1(_P,N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic1(P,I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg_(I1,S,SI),   % otherwise: N=<M
	(arg_(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic1(P,I1,S,N,T,M)
	     ;
	          greater(P,SI,TI)
	     )
	;% M<N
	     true
	).
	

lexicographic(_P,N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic(P,I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg_(I1,S,SI),   % otherwise: N=<M
	(arg_(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic(P,I1,S,N,T,M)
	     ;greater(P,SI,TI) ->
		  greater_than_list(P,S,I1,T,M)
	     ;
                  some_subterm_greater_equal(P,I1,S,N,T)
	     )
	;% M<N     
	     true
	).

greater_than_list(_P,_,N,_,N):-
	!.
greater_than_list(P,S,I,T,M):-
	I1 is I+1,
	arg_(I1,T,TI),
	greater(P,S,TI),
	greater_than_list(P,S,I1,T,M).




%------------------------------------------------------------------------------
% greater_eq( P, S, T ):
%	S is greater than T with respect to lpo(P) or S is equal to T
%------------------------------------------------------------------------------

greater_eq(_,T1,T2):- T1 == T2,!.
%greater_eq(_,_T1,T2):- '$min' == T2,!.
%greater_eq(T1,T2):- 	'$true$' == T2,	T1\=='$min'.
greater_eq(P,T1,T2):- greater(P,T1,T2),!.






/* ----------------------------------------------------------------------------

               Operations on the precedence
	       
----------------------------------------------------------------------------- */

is_precedence(P):-
	nonvar(P),
	P=prec(TP,Pr,(Consts,_),Conss),
	is_list(TP),
	is_list(Pr),
	is_list(Consts),
	is_list(Conss),
	maplist(is_function_symbol,TP),
	maplist(is_function_symbol,Consts),
	maplist(is_function_symbol,Conss),
	maplist(is_function_symbol,Pr).


%------------------------------------------------------------------------------
% is_top_predicate( +P, +F ):
%  succeeds whenever F occurs in the TopPrecidates part of
%------------------------------------------------------------------------------

is_top_predicate( prec(TopPreds,_,_,_), F ):-
	member212(F,TopPreds).




% greater_op( +P, +F , +G)
%    succeeds whenever F occurs before G in the precedence list Pr of
%    P=prec(TopPredicates,Pr,(Consts,PrecGraph),Conss).
%
%    It is assumed that F is different from G.
%
%    Note that the orderings in TopPredicates, and Conss are
%    is irrelevant, and it is assumed all these symbols 
%    also occur in Pr.
%
%    Conss is the set of constructors. Constructors should be minimal
%    in precedence. The constants in Consts in particular smaller than any of the 
%    symbols in Prec and larger than any of the symbols in Conss.
%
%    Invariant: we never compare two constants in Consts that are not
%    connected on a path in the precedence graph PrecGraph (a SICStus ugraph).
%    Should that become necessary, the precedence was extended before.
%
%
%    Should we postulate the existence of a minimal constant???
/*
greater_op(Prec,F,G):-
	pre((
	     is_precedence(Prec),
	     is_function_symbol(F),
	     is_function_symbol(G),
	     F\==G
	    ),
	    greater_op(Prec,F,G)
	   ).
*/
greater_op(prec(_TopPreds,_Prec,(Consts,PrecGraph),Conss),F,G):-
	member212(F,Consts),
	!,
	(member212(G,Consts) ->
	    reachable(F,PrecGraph,R),
	    member212(G,R)
	;
	    member212(G,Conss)
	).
greater_op(prec(_TopPreds,Prec,(Consts,_),Conss),F,G):-
	% F is not a constant
	member212(F,Prec),
	!,
	(member212(G,Consts) ->
	    \+ member212(F,Conss)
	;
	    greater_in_prec(Prec,F,G)
	).
greater_op(prec(_TopPreds,_Prec,_,Conss),F,G):-
	% F and G are both in Conss
	greater_in_prec(Conss,F,G).



% we use == for membership tests so as to avoid instantiation
% of type variables

greater_in_prec([F1|_Fs],F,_G):-
	% we found F first in Prec
	F==F1,
	!.
greater_in_prec([G1|_Fs],_F,G):-
	G==G1,
	% we found G first 
	!,
	fail.
greater_in_prec([_|Fs],F,G):-
	greater_in_prec(Fs,F,G).
/*
greater_in_prec([],F,G):-
	% we found neither, that should not happen
	raise_exception(user_error('~N! Precedence non total ! Goal: ~p~n',
				   greater_in_prec([],F,G))).
*/




precedencesAsEdges([],[]).
precedencesAsEdges([P|Ps],Es):-
	precedenceAsEdges(P,Es1),
	precedencesAsEdges(Ps,Es2),
	user:append(Es1,Es2,Es).

precedenceAsEdges([],[]).
precedenceAsEdges([_],[]).
precedenceAsEdges([P,Q|Ps],[P-Q|Es]):-
	precedenceAsEdges([Q|Ps],Es).



/* ------------------------------------------------------------------------------

 Ordered structures

------------------------------------------------------------------------------ */

extend_ordering(L,R):-
	cont(ord,ord(lpoc,prec(TP,Prec,AllConsts,Conss))),
	constructor_equation(AllConsts,Conss,L,R,Consts),
	mk_greater_ops(AllConsts,AllConsts1,R,Consts),
	store(ord,ord(lpoc,prec(TP,Prec,AllConsts1,Conss))).
extend_ordering(R,L):-
	cont(ord,ord(lpoc,prec(TP,Prec,AllConsts,Conss))),
	constructor_equation(AllConsts,Conss,L,R,Consts),
	mk_greater_ops(AllConsts,AllConsts1,R,Consts),
	store(ord,ord(lpoc,prec(TP,Prec,AllConsts1,Conss))).
extend_ordering(R,L):-
	cont(ord,ord(lpoc,prec(TP,Prec,AllConsts,Conss))),
	constant_equation(AllConsts,L,R),
	(   mk_greater_ops(AllConsts,AllConsts1,L,[R])
	;
	    mk_greater_ops(AllConsts,AllConsts1,R,[L])
	),
	store(ord,ord(lpoc,prec(TP,Prec,AllConsts1,Conss))).
extend_ordering(_,_).









mk_greater_ops(Conss,Conss,_R,[]).
mk_greater_ops(Conss,Conss1,R,[C|Consts]):-
	mk_greater_op(Conss,Conss0,R,C),
	mk_greater_ops(Conss0,Conss1,R,Consts).

mk_greater_op((_Cs,Graph),(Cs1,Graph1),F,G):-
	% change Prec into Prec1 to make F>G
	% fail, if not possible
	add_edges(Graph,[F-G],Graph1),
	!,
	top_sort(Graph1,Cs1).

			    


constructor_term(Cs,Conss,T,Consts):-
	nonvar_(T),
	univ_(T,[F|Args]),
	member212(F,Conss),
	maplist(constructor_subterm(Cs,Conss),Args,ConstsList),
	append(ConstsList,Consts).

constructor_subterm(Cs,_Conss,T,[T]):-
	nonvar_(T),
	member212(T,Cs),
	!.
constructor_subterm(Cs,Conss,T,Consts):-
	constructor_term(Cs,Conss,T,Consts).


constructor_equation((Cs,_),Conss,L,R,Consts):-
	member212(R,Cs),
	constructor_term(Cs,Conss,L,Consts).

constant_equation((Cs,_),L,R):-
	member212(L,Cs),
	member212(R,Cs).

newconstant(C):-
	cont(ord,ord(lpoc,prec(TP,Prec,(Cs,G),Conss))),
	cont(const_ctr,N),
	incr(const_ctr),
	name(N,NN),
	name('c$',CDN),
	user:append(CDN,NN,CN),
	name(C,CN),
	store(ord,ord(lpoc,prec(TP,Prec,([C|Cs],G),Conss))).
	





append(-, _) :- !, fail.	% reject partial lists.
append([], []).
append([L|Ls], List0) :-
	user:append(L, List1, List0),
	append(Ls, List1).

