%declarations%
:- module(rpo, [ 
	greater/2,
	greater_ground/2,
	greaterEqual/2,
	greater_op/2
   ]).

:- use_module(st_terms).

:- use_module(library(basics), [member/2]).
:- use_module(satsrc(auxiliary), [delete_one312/3,
			  member212/2]).
:- use_module(myterms, [
	apply_subst/3,
	occurs_in/2
   ]).
:- use_module(satsrc(mylists), [
	permutation/2]).
:- use_module(library(sets), [
	select/3
   ]).





%end%

%------------------------------------------------------------------------------
% greater( S, T, P ):
%	S is greater than T with respect to the recursive path ordering
%       over precedence P
%------------------------------------------------------------------------------


greater( S, _, _):-
	var_(S),
	!,
	fail.
greater( S, T, P):-
	\+var_(T),
	!,
	functor_(S,F,N),
	functor_(T,G,M),
	greater_(F,G,N,M,S,T,P).

greater( S, T, P ):-
	functor_(S,F,_),
	(is_first_predicate(F,P) -> 
	     true
	;    contains_var_(T,S)),
	!.
	

greater(F,F,_N,_,S,T,P):-
	multisetStatus(F,P),
	!,
	args_(S,Ss),
	args_(T,Ts),
	msext(greater(P),Ss,Ts),
	!.
greater(F,F,N,M,S,T,P):-
	(is_first_predicate(F,P) ->
	     lexicographic1(0,S,N,T,M,P)
	;     lexicographic(0,S,N,T,M,P)
	),
	!.
greater(F,G,_N,M,S,T,P):-
	(greater_op(F,G,P)->
	     greater_than_list(S,0,T,M,P)
	; % greater_op(G,F) or F=G 
             genarg_(_,S,S1),
             greaterEqual( S1, T, P)
	), 
	!.
	


msext(O,U,V):-
	difference(U,V,UV),
	difference(V,U,VU),
	msext1(O,UV,VU).


difference(A,[],A).
difference(A,[X|Xs],B):-
	(select(Y,A,A1),equal(X,Y) ->
	     difference(A1,Xs,B)
	;difference(A,Xs,B)).

msext1(_,[],[]):-
	!,
	fail.
msext1(O,U,V):-
	\+ ( member(T,V),
	       \+ (member(S,U),
		      functor(O,Rel,_),
		      arg(1,O,Prec),
		      OP=..[Rel,S,T,Prec],
		      P
		  )
	   ).





lexicographic1(N,_,N,_,_M,_):-
	!,
	fail.
lexicographic1(M,_,_N,_,M,_):-
	!,
	fail.
lexicographic1(I,S,N,T,M,P):-
	I1 is I+1,
	arg_(I1,S,SI),
	arg_(I1,T,TI),
	(SI==TI ->
	     lexicographic1(I1,S,N,T,M,P)
	;greater(SI,TI,P)
	).
	

lexicographic(N,_,N,_,_M,P):-
	!,
	fail.
lexicographic(M,_,_N,_,M,P):-
	!,
	fail.
lexicographic(I,S,N,T,M,P):-
	I1 is I+1,
	arg_(I1,S,SI),
	arg_(I1,T,TI),
	(SI==TI ->
	     lexicographic1(I1,S,N,T,M,_)
	;greater(SI,TI,P) ->
	     greater_than_list(S,I1,T,M,P)
	).

greater_than_list(_,N,_,N,_):-
	!.
greater_than_list(S,I,T,M,_):-
	I1 is I+1,
	arg_(I1,T,TI),
	greater(S,TI,P),
	greater_than_list(S,I1,T,M,P).



/* 
The RPO is a quasi-ordering in which two terms are equivalent
if they are equal up to permutation of arguments of function
symbols with multiset status.

unifiable(+S,+T,+P)
is true when there exists a substitution making
S and T equal up to permutativity for operators with multiset status
in precedence P.

*/

unifiable(S,T,P):-
	\+ \+mgu([S=T],P,_).



mgu( L, P, U ):-
	mgu( L, P, [], U ),
	!.

mgu( [], _, U, U ).				% the end
mgu( [T1=T2|R], P, S, U ):-			% delete
	T1==T2,
	!,
	mgu( R, P, S, U ).
mgu( [T1=T2|R], P, S, U ):-			
	(var(T1) ->
	     (\+ contains_var( T1, T2 ) ->           % occur check
		  apply_subst( T1=T2, R, R1 ), 
		  apply_subst( T1=T2, S, S1 ), 
		  mgu(R1, P, [T1=T2|S1], U )
	     )
	;var(T2) ->
	     (\+ contains_var( T2, T1 ) ->           % occur check
		  apply_subst( T2=T1, R, R1 ),
		  apply_subst( T2=T1, S, S1 ),
		  mgu(R1, P, [T2=T1|S1], U )
	     )
	; % decompose
	     functor_(T1,F,N),
	     functor_(T2,F,N),
	     (multisetStatus(F,P) ->
	         permutedSubterms(T1,T1P),
		 equalArgs( N, T1P, T2, L/R )
	     ;	 equalArgs( N, T1, T2, L/R )),
	     mgu( L, P, S, U )
	).
mgu(T1=T2,P,S,U):-
	mgu([T1=T2],P,S,U).


multisetStatus(F,prec(MSOPS,_)):-
	member212(F,MSOPS).




permutedSubterms(T,TP):-
	T=..[F|Ts],
	permutation(Ts,TPs),
	TP=..[F|TPs].
		  


% equalArgs(+I,+N,+S,+T,?E) is true when E is the list of unification
% problems between corresponding subterms of S and T between 
% indexes N-I+1 and N. E is represented in form of a diffence list
% E1/E2 to allow easy appending with surrounding unification problems

equalArgs(0,_,_,L/L):-
	!.
equalArgs(I,T1,T2,[A1=A2|L]/R):-
	arg(I,T1,A1),
	arg(I,T2,A2),
	I1 is I-1,
	equalArgs(I1,T1,T2,L/R).



var1(X):-
	var(X),
	!.
var1('$VAR'(_)).   % for skolemized variables




%------------------------------------------------------------------------------
% greaterEqual( S, T ):
%  S is greater than T with respect rpo or S is equal to T up to permutativity
%------------------------------------------------------------------------------

greaterEqual(T1,T2):- equal(T1,T2),!.
greaterEqual(_T1,T2):- '$min' == T2,!.
greaterEqual(_T1,T2):- '$true$' == T2,!.
greaterEqual(T1,T2):- greater(T1,T2),!.

equal(T1,T2):-
	\+ \+ 
	(  numbervars((T1,T2),1,_),
	   unifiable(T1,T2) ).


%------------------------------------------------------------------------------
% maximal_term( S, T, M ):
%------------------------------------------------------------------------------

maximal_term( S, T, S ):- greaterEqual( S, T ), !.
maximal_term( S, T, T ):- greater( T, S ), !.
