/*
:- module(acTheory, [
      
        currentTheory/1,
	equal/2,
	extension/4,
	match/2,
	match/4,
        mgu/2,
	unifiable/2,
	unify/2

     ]).
/*

/* this module provides the basic operations
   for solving equational constraints
   over finite trees (modulo AC)
   where finite trees are
   at present identified with Prolog terms.
   Equational constraints are single equations
   or lists (representing conjunctions) of equations.
*/


:- use_module(library(basics),[member/2]).

:- use_module(uniac,[ac_uniAC/2]).

:- use_module(match,[ac_match_pt/2,ac_match/4]).

:- use_module(aclpo,[acEqual/2]).

:- use_module(achelp, [
	ac_del_assoc/3,
	ac_ist_AC/1,
	ac_occurs/2
   ]).

:- use_module(myterms, [
        delete_var/3,
	list_of_vars/2,
	occurs_in/2
   ]).

:- use_module(adt_term, [
	free_vars/2
   ]).

:- use_module(rewrite,[normalformBy/3]).

:- use_module(termOrder,[greater/2]).


% ========================================================================



currentTheory(ac).


equal(A,B):-
	acEqual(A,B).

match(A,B):-
	ac_match_pt(A,B).

match(A,B,C,D):-
	ac_match(A,B,C,D).

%mgu(_A,_B):-
%	raise_exception(mgu_not_implemented_for_AC).

% partial imlementation:
mgu([],[]).
mgu([A=B|E],[]):-
	mgu(A=B,_),
	mgu(E,_).
mgu(A=B,[]):-
	acEqual(A,B).



unifiable(A,B):-
	\+ \+ ac_uniAC(A,B).

unify(A,B):-
	ac_uniAC(A,B).


/* 
-------------------------------------------------------------------------------

   Extended Rules

-------------------------------------------------------------------------------
*/


extension(L,R,L,R).
extension(L,R,L1,R1):-
	nonvar(L),
	functor(L,F,2),
	ac_ist_AC(F),
	
	\+ self_extension(L,R),

	L1=..[F,L,X],
	R1=..[F,R,X].



self_extension(L,R):-
	L =.. [O|Args],
	ac_del_assoc(O,Args,OArgs),
	append(A1,[X|A2],OArgs),
	var(X),
	\+ occurs_in(X,A1),
	\+ occurs_in(X,A2),
	free_vars(L,Vs),
	delete_var(Vs,X,VX),
	copy_term((L,R),(L2,R2)),
	copy_term((R,X,VX),(R1,X1,VX)),
	X=..[O,U,V],   % R = R[U+V/X]
	X1=U,
	RR=..[O,R1,V], % RR= R[U/X]+V
	\+ \+ (   numbervars((R,RR),0,_),
		  (greater(L2,R2) ->
		       normalformBy(RR,RRN,[L2=R2]),
		       normalformBy(R,RN,[L2=R2]),
		       equal(RN,RRN)
		  ;
		       equal(RR,R)
		  )
	      ).



