:- module(adt_atom, [ 
	active_term_in_atom/2,
	atom_admits_variable_elimination/1,
	atom_as_rule/3,
	atom_as_term/2,
	atom_as_substitution/1,
	atom_as_reductive_substitution/1,
	atom_for_demodulation/5,
	atoms_as_inequations/2,
	boolean_atom/1,
	chain/8,
	create_atom/2,
	create_atom/4,
	equal_atoms/2,
	equation_as_atom/2,
	equivalent_relation_symbols/2,
	factorization/7,
	false_atom/1,
	get_atom/4,
	get_atom_members/3,
	get_atom_preorientation/2,
	get_atom_pre_rel_left_right/5,
	get_atom_terms/2,
	get_atom_relation_symbol/2,
	get_atom_vars/2,
	get_binary_atom/4,
	get_binary_atom_pre_rel_left_right/5,
	get_equation_members/3,
	get_list_atoms_vars/2,
	get_list_of_atoms_operators/2,
	greater_literal/4,
	greater_relation_symbol/2,
	induced_equivalence/2,
	inverse_atoms_if_possible/3,
	is_atom/1,
	is_abbreviation/1,
	is_caseSplit/1,
	is_equational_atom/1,
	is_equivalence/1,
	is_standard_atom/1,
	is_negated_standard_atom/2,
	is_term_equation/1,
	is_true_atom/1,
	is_false_atom/1,
	is_marked/1,
	is_proposition/1,
	is_propositional_literal/1,
	is_positive_standard_atom/1,
	is_trivial_atom/1,
	lifted_equation/2,
	lifted_inequation/2,
	mark_antecedent/2,
	match_atom_terms/5,
	match_atoms/5,
	match_atoms/4,
	match_atoms/2,
	neg_atom_as_literal/2,
	negative_chain/7,
	normalize_atom/2,
	orient/6,
	orient_atom/2,
	orient_atom_fresh/2,
	orient_by_polarity/5,
	pos_atom_as_literal/2,
	predicate_in_atom/2,
	print_list_atoms/1,
	put_atom_preorientation/3,
	replacesubterm_ofatom_by_giving_/4,
	residual/4,
	resolution/3,
	selection/10,
	selectable_atom/5,
	strongly_inverse_atom/2,
	weakly_inverse_atom/2,
	subsumes_literals/4,
	symmetric_polarity/3,
	symmetric_variants/4,
	symmetric_version_of_atom/2,
	true_atom/1,
	true_literal/1,
	false_literal/1,
	is_true_literal/1,
	is_false_literal/1,
	var_on_literal/2,
	var_on_predicate/2,
	variants_atoms/2
   ]).

%:- use_module(user,_,[append/3]).


:- use_module(equationalTheory, [
        currentTheory/1,
	equal/2,
	extension/4,
	match/2,
	match/4,
	unify/2
   ]).

:- use_module(global, [
	cont/2
   ]).

:- use_module(myterms, [
	function_symbol_in/2,
	general_function_call/1,
	intersection_vars/3,
	list_of_vars/2,
	occurs_in/2
   ]).


:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(maplist), [
	maplist/2
   ]).

:- use_module(library(sets), [
	select/3
   ]).

:- use_module(proofState, [
	rule/2,
	clause/1
   ]).

:- use_module(constraintSolver, [
	simplify/2
   ]).


:- use_module(adt_precedence, [
        abbreviating_proposition/1,
        generated_proposition/1,
	greater_op/2,
	has_external_sort/1,
	is_predicate/1,
	is_propositional_variable/1,
	is_magic_goal_predicate/1,
	minimal_constant/1,
	predicate/1,
        split_proposition/1
   ]).

:- use_module(classes, [
	succedent/2,
	hasSelectedLiteral/1
   ]).

:- use_module(adt_term, [
	free_vars/2,
	logical_operator/1,
	logical_symbol/1,
	standard_atom/1,
	get_term_operators/2,
	get_term_vars/2,
	path_arg1/3,
	predicate_in_formula/3,
	replacesubterm_eq/4,
	term_size/2,
	variant/2
   ]).

:- use_module(achelp, [
	ac_ist_AC/1
   ]).

:- use_module(termOrder,[greater/2, maximal_term/3]).

:- use_module(orderConstraints,[true/1,and/3]).

:- use_module(sequents,[atom_in_expansion/2]).

:- use_module(auxiliary, [
	member212/2,
	delete32/3,
	delete_all/3
   ]).

:- use_module(inference_rules, [
	constraint_grequal_literal/5
   ]).

:- use_module(props, [
	composition_of_relations/3,
	dense_ordering_wo_end_relation_symbol/1,
	partial_congruence_relation_symbol/1,
	congruence_relation_symbol/1,
	firstOrder/0,
	predicate_with_extension/3,
	reflexive_closure/2,
	reflexive_relation_symbol/1,
	residual/3,
	special_relation/1,
	strict_part/2,
	equality_relation_symbol/1,
	symmetric_relation_symbol/1,
	total_ordering_relation_symbol/1,
	transitive_relation_symbol/1,
	flag_on/2,
	partial_equivalence_relation_symbol/1,
	irreflexive_relation_symbol/1,
	total_relation_symbol/1,
	user_selection/1,
	user_selection/2,
	var_overlaps/1
   ]).

:- use_module(orderConstraints, [
	false/1
   ]).

:- use_module(library(change_arg), [
	change_path_arg/4
   ]).

:- use_module(library(samsort), [
	samsort/3
   ]).




%end%
%							ADT_ATOM


/* ========================================================================

[n/l/r, relation symbol, S, T]

n: non_preoriented; l: S is greater than T; r: vice versa.

relation symbol can be:  =, '==', other declared binary transitive symbol 

if relation symbol is == then T will be $true$

======================================================================== */


create_atom( A<=>B, At):-
	!,
	create_atom(A==B,At).
create_atom( '$min', [l,==,'$true$','$min']):-
	!.
create_atom( '$true$', [n,==,'$true$','$true$']):-
	!.
create_atom(S=='$true$',A):-
	!,
	create_atom(S,A).
create_atom(S==T,[n,==,S,T]):-
	!.
create_atom( Rel, [n, F, S, T] ):- 
		Rel =.. [F,S,T],
		(
		special_relation( F )
		;  symmetric_relation_symbol(F),
		    \+firstOrder
		)
		, 
		!.
create_atom( Pred, [l, '==', Pred,'$true$'] ).

create_atom( S, Rel, T, [n,Rel,S,T] ):- 
		(
		  special_relation( Rel )
		;  symmetric_relation_symbol(Rel),
		    \+firstOrder
		)
		  , 
		!.
create_atom( S, Rel, T, [l, '==', Pred,'$true$'] ):-
	Pred=..[Rel,S,T].







normalize_atom([_, '==', Rel,'$true$'],[n, F, S, T]):-
	Rel =.. [F,S,T],
	(
	  special_relation( F )
	;  symmetric_relation_symbol(F),
	  \+firstOrder
		)
	, 
	!.
normalize_atom(A,A).


put_atom_preorientation( P, [_,RS,S,T], [P,RS,S,T] ).
put_atom_relation_symbol( RS, [P,_,S,T], [P,RS,S,T] ).
put_atom_left_side( S, [P,RS,_,T], [P,RS,S,T] ).
put_atom_right_side( T, [P,RS,S,_], [P,RS,S,T] ).


lifted_inequation([_P,RS,S,'$true$'],[n,RS1,L,R]):-
	(RS= = ; RS= ==),
	nonvar(S),
	S=..[RS1,L,R],
	special_relation(RS1),
	!.
lifted_inequation(A,A).

lifted_equation([_P,==,S,'$true$'],[n,RS1,L,R]):-
	S=..[RS1,L,R],
	congruence_relation_symbol(RS1),
	!.
lifted_equation(A,A).


get_atom_preorientation( [P,_,_,_], P1 ):-
	(nonvar(P) -> P1=P; P1=n).
get_atom_relation_symbol( [_,RS,_,_], RS ).
get_atom_left_side( [_,_,S,_], S ).
get_atom_right_side( [_,_,_,T], T ).
get_atom_members( [_,_,S,T], S, T ).
get_atom( [_,RS,S,T], RS, S, T ).
get_atom_preorientation_and_members( [P,_,S,T], P, S, T ).
get_atom_pre_rel_left_right( [P,R,S,T], P, R,S, T ).


get_binary_atom( [_,==,A,'$true$'], RS, S, T ):-
	A=..[RS,S,T],
	adt_precedence:dual_predicate(RS),
	!.
get_binary_atom(A,RS,S,T):-
	get_atom(A,RS,S,T).

get_binary_atom_pre_rel_left_right( [_,==,A,'$true$'], n, RS, S, T ):-
	A=..[RS,S,T],
	adt_precedence:dual_predicate(RS),
	!.
get_binary_atom_pre_rel_left_right(A,O,RS,S,T):-
	get_atom_pre_rel_left_right(A,O,RS,S,T).

get_atom_terms([_,==,A,'$true$'],Ts):-
	A=..[_|Ts],
	!.
get_atom_terms([_,_,L,R],[L,R]).




active_term_in_atom([l,_,S,_],S).
active_term_in_atom([r,_,_,S],S).
active_term_in_atom([n,_,_,S],S).
active_term_in_atom([n,_,S,_],S).


get_equation_members( A,S,T):-
	atom_as_equation(A,E),
	get_atom_members(E,S,T).


% -------------------------------------------------------------------
% true and false atoms
% -------------------------------------------------------------------


true_atom(T):-
	create_atom('$true$',T).

false_atom(T):-
	(var(T) ->
	     create_atom('$min',T)
	;
	     is_false_atom(T)
	).


true_literal([+|E]):-
	true_atom(E).
true_literal([-|E]):-
	false_atom(E).

false_literal([-|E]):-
	true_atom(E).
false_literal([+|E]):-
	false_atom(E).

is_true_literal([+|E]):-
	is_true_atom(E).
is_true_literal([-|E]):-
	is_false_atom(E).

is_false_literal([-|E]):-
	is_true_atom(E).
is_false_literal([+|E]):-
	is_false_atom(E).

is_true_atom( E ):- 
	is_trivial_atom( E ).

is_trivial_atom( [_,R,S,T] ):- 
	reflexive_relation_symbol(R),
	equal(S,T).

boolean_atom( [_,==,_,'$true$'] ).
boolean_atom( [_,==,_,'$min'] ).

is_negated_standard_atom( [_,==,A,'$min'], A):-
	standard_atom(A).
is_negated_standard_atom( [_,==,'$min',A], A):-
	standard_atom(A).


is_false_atom( [_,==,'$true$','$min'] ).
is_false_atom( [_,==,'$min','$true$'] ).
is_false_atom( [_,R,S,T] ):-
	irreflexive_relation_symbol(R),
	equal(S,T).


is_atom([_,RS,_,_]):-
	atom(RS).

is_proposition([_,==,P,'$true$']):-
	      atom(P).

is_proposition([_,==,P,'$min']):-
	      atom(P).



is_propositional_literal([_|A]):-
	is_proposition(A).


reflexivity_atom([_,_RS,S,T]):-
	var(S), var(T), S==T.

equivalence([_,==,_,R]):-
	R\=='$true$'.

is_equivalence([_,==,_,R]):-
	!,
	R\=='$true$'.

is_equational_atom([_,==,_,R]):-
	!,
	R\=='$true$'.
is_equational_atom([_,RS,_,_]):-
	congruence_relation_symbol(RS).

is_term_equation([_,RS,_,_]):-
	RS\== ==,
	transitive_relation_symbol(RS),
	symmetric_relation_symbol(RS).

is_standard_atom(A):-
	(get_atom_pre_rel_left_right(A,_,==,F,G) ->
	     standard_atom(F),
	     standard_atom(G) 
	;
	     true
	).

is_positive_standard_atom(A):-
	is_standard_atom(A),
	!.
is_positive_standard_atom(A):-
	get_atom_pre_rel_left_right(A,P,==,F,G),
	(P=l, standard_atom(F)
	;P=r, standard_atom(G)),
	!.



atom_as_reductive_substitution([_,R,X,Y]):-
	congruence_relation_symbol(R),
	minimal_constant(Min),
	( var(X); X==Min),
	( var(Y); Y==Min),
%	( var(X); atomic(X)),
%	( var(Y); atomic(Y)),
	X=Y.



atom_as_substitution([_,=,X,T]):-
	var(X),
	unify(X,T), % w/ occurs check
	!.   
atom_as_substitution([_,=,T,X]):-
	var(X),
	unify(X,T), % w/ occurs check
	!.    

atom_is_substitution(A):-
	\+ \+ atom_as_substitution(A).


atom_admits_variable_elimination([_,RS,X,T]):-
	congruence_relation_symbol(RS),
	var(X),
	\+ \+ unify(X,T),
	!.    % occurs check
atom_admits_variable_elimination([_,RS,T,X]):-
	congruence_relation_symbol(RS),
	var(X),
	\+ \+ unify(X,T),
	!.    % occurs check

%------------------------------------------------------------------------------

%           auxiliary predicates for selection strategies

%------------------------------------------------------------------------------

selection(K,Am1,Ap1,Sm1,Sp1,Am,Ap,Sm,Sp,M):-
	(\+select_something(K,Am1,Ap1,Sm1,Sp1),M\==1
	->  
	     Am=Am1,Ap=Ap1,
	     Sm=Sm1,Sp=Sp1,
	     M=0  
	;
	     M=1,
	     append(Am1,Ap1,A),
	     samsort(preferable_selection(Am1,Ap1,Sm1,Sp1),A,Ap),
	     Am=[],
	     append(Sm1,Sp1,Sp),
	     Sm=[]
	),
	!.




% Andrei
select_something(_,[],[E],_Sm,_Sp):-
	is_caseSplit(E),
	!,
	fail.
select_something(_,[E],[],_Sm,_Sp):-
	is_caseSplit(E),
	!,
	fail.
select_something(monotonicity(_,[_],_,1,==),_Am,_Ap,_Sm,_Sp):-
	!.
select_something(defClause(_),_Am,_Ap,_Sm,_Sp):-
	!.
select_something(inclusion(_,_),_Am,_Ap,_Sm,_Sp):-
	!.
select_something(_,Am,Ap,Sm,Sp):-
	user_selected_literal(_E,Am,Ap,Sm,Sp),
	!.
select_something(_,Am,Ap,_Sm,_Sp):-
	(member(E,Am);member(E,Ap)),
	is_marked(E).
select_something(_,Am,Ap,Sm,Sp):-
	maplist(is_proposition,Am),
	maplist(is_proposition,Ap),
	maplist(is_proposition,Sm),
	maplist(is_proposition,Sp),
	!,
	fail.
select_something(_,Am,Ap,Sm,Sp):-
	free_vars((Am,Ap,Sm,Sp),[_|_]),
	(member(E,Am),free_vars(E,[]);member(E,Ap),free_vars(E,[])).
select_something(_,Am,Ap,_Sm,_Sp):-
	(member(E,Am),free_vars(E,[]);member(E,Ap),free_vars(E,[])),
	is_equational_atom(E).
%select_something(_,Am,Ap,_Sm,_Sp):-
%	(member(E,Am);member(E,Ap)),
%	is_abbreviation(E).
select_something(_,Am,Ap,_Sm,_Sp):-
	(member(E,Am);member(E,Ap)),
	is_projection(E).
select_something(_,Am,Ap,_Sm,_Sp):-
	(member(E,Am);member(E,Ap)),
	easily_solvable_atom(E).
%select_something(_,Am,Ap,Sm,Sp):-   % not so good
	% antecedent has type atom
%	type_atom(Am,Ap,Sm,Sp,_E).
select_something(_,[E],_,[],Sp):-
	\+ unshieldedVariable(E,[],Sp,_),
	!,
	fail.
select_something(_,[],_,[_|_],_):-    % [_|_] was [_] ??
	!,
	fail.
select_something(_,_,_,_,_).


type_atom(Am,Ap,Sm,Sp,E):-
	(member(E,Am);member(E,Ap)),
	is_type_atom(Sm,Sp,E).

is_type_atom(Sm,Sp,E):-
	boolean_atom(E),
	get_atom_pre_rel_left_right(E,_,==,F,_G),
	nonvar(F),
	functor(F,R,1),
	\+ (member(E1,Sm),pred_occurs_in_atom(E1,R/1,P)),
	\+ (member(E1,Sp),pred_occurs_in_atom(E1,R/1,P)),
	!.


	

proposition_or_nonstandard(E):-
	is_proposition(E).
proposition_or_nonstandard(E):-
	\+is_standard_atom(E).


is_projection(E):-
	atom_as_term(E,A),
	functor(A,P,_),
	name('P',[PC]),
	name(P,[PC|_]).

	

is_abbreviation(E):-
	atom_as_term(E,A),
	functor(A,P,N),
	abbreviating_proposition(P/N).


is_caseSplit(E):-
	atom_as_term(E,A),
	functor(A,P,N),
	split_proposition(P/N).


user_selected_literal(E,Am,Ap,Sm,Sp):-
	user_selection(on),
	append(Am,Ap,A),
	append(Sm,Sp,S),
	maplist(atom_as_term,A,AT),
	maplist(atom_as_term,S,ST),
	user_selection(At,[AT,ST]),
	create_atom(At,E).



selection_weight(E,Am,Ap,Sm,Sp,10000):-
	user_selected_literal(E,Am,Ap,Sm,Sp),
	!.
selection_weight(E,_,_,_,_,9000):-
	is_marked(E),
	!.
selection_weight(E,_,_,_,_,8100):-
	is_projection(E),
	!.
selection_weight(E,_,_,_,_,7000):-
	atom_as_term(E,A),
	functor(A,P,N),
	(is_magic_goal_predicate(P);
	predicate_with_extension(P/N,_,_)),	    
	!.
selection_weight(E,_,_,_,_,8100):-
% Andrei's suggestion
% also need to have this with small search depths ??
%selection_weight(E,_,_,_,_,100):-
	is_abbreviation(E),
	!.
%selection_weight(E,_Am,_Ap,Sm,Sp,1000):-
%	is_type_atom(Sm,Sp,E),
%	!.
selection_weight(E,_,_,_,_,850):-
	easily_solvable_atom(E),
	!.
selection_weight(E,_,_,_,_,850):-
	unsolvable_atom(E),
	!.
selection_weight([_,RS,T,X],_,_,_,_,W):-
	partial_congruence_relation_symbol(RS),
	RS\== ==,
	((var(X);var(T)) ->
	      (\+ unify(X,T) ->
		   W=700
	      ;
	           W=850
	      )
	;
	      W=600
	),
	!.
selection_weight(E,_,_,Sm,Sp,W):-
	setof(X,unshieldedVariable(E,Sm,Sp,X),AX),
	length(AX,N),
	W is max(100-N,0),
	!.
selection_weight(E,_,_,_,_,600):-
	get_term_vars(E,[]),
	!.
selection_weight(E,_,_,Sm,Sp,W):-
	succedentVars(E,Sm,Sp,Vs),
	length(Vs,M),
	get_atom_vars(E,V),
	length(V,N),
	W is 300+5*M+N.

preferable_selection(Am,Ap,Sm,Sp,E1,E2):-
	selection_weight(E1,Am,Ap,Sm,Sp,W1),
	selection_weight(E2,Am,Ap,Sm,Sp,W2),
	!,
	(W1>W2 ->
	     true
	;W1=W2 ->
	     term_size(E1,U1),
	     term_size(E2,U2),
	     U1>U2
	).






unshieldedVariable([_,_,L,R],_,_,L):-
	var(L),
	\+ occurs_in(L,R).  % L==R not considered unshielded
unshieldedVariable([_,_,R,L],_,_,L):-
	var(L),
	\+occurs_in(L,R).

succedentVars(E,Sm,Sp,SV):-
	get_atom_vars(E,V),
	append(Sm,Sp,S),
	get_list_atoms_vars(S,VV),
	intersection_vars(V,VV,SV).








mark_antecedent([],[]).
mark_antecedent([[_|A]|As],[[s|A]|As]).

is_marked([P|_]):-
	P==s.
is_marked(A):-
	get_binary_atom_pre_rel_left_right(A,_,RS,_,_),
	adt_precedence:dual_predicate_pair(RS,_).


selectable_atom(A1,A,A2,Ant,_):-
	append(A1,[A|A2],Ant),
	get_term_vars(A,[]).
%	\+ is_proposition(A).
selectable_atom(A1,A,A2,Ant,_):-
	append(A1,[A|A2],Ant),
	unsolvable_atom(A).
/*
selectable_atom(A1,A,A2,Ant,_):-
	append(A1,[A|A2],Ant),
	easily_solvable_atom(A).
selectable_atom(A1,A,A2,Ant,_):-
	append(A1,[A|A2],Ant),
	sort_atom(A).
*/
selectable_atom(A1,[O,RS,X,Y],A2,Ant,Succ):-
	user_selection(B,S),
	(nonvar(S) ->
	     S=..[RS1,L1,R1],
	     member([_,RS1,X1,Y1],Succ),
	     \+ \+ match_light_atoms(RS1,[L1,R1],[X1,Y1])
	;    true),
	append(A1,[[O,RS,X,Y]|A2],Ant),
	(nonvar(B) ->
	     B=..[RS,L,R],
	     \+ \+ match_light_atoms(RS,[L,R],[X,Y])
	;    true).
selectable_atom(A1,[O,RS,X,Y],A2,Ant,_):-
	append(A1,[[O,RS,X,Y]|A2],Ant),
	X==Y,
	reflexive_relation_symbol(RS).


sort_atom([_,==,X,'$true$']):-
	X=..[_|Args],
	select(Y,Args,Args1),
	get_term_vars(Args1,[]),
	var(Y).

easily_solvable_atom([_,RS,T,X]):-
	reflexive_relation_symbol(RS),
	\+ (strict_part(RS,R),total_ordering_relation_symbol(R)),
	\+ \+ unify(T,X).



unsolvable_atom([_,==,P,'$true$']):-
	!,
	functor(P,F,N),
	\+occurs_positively(F/N,P).
unsolvable_atom([_,RS,_T,_X]):-
	(strict_part(RS,R) -> true; R=RS),
	\+ occurs_positively(R/2,_).

%	\+ ( generating(Q),unify(P,Q) ).



generating(Q):-
	rule(_,C),
	\+ hasSelectedLiteral(C),
	succedent(C,S),
	member([_,Q,==,'$true$'],S).
generating(Q):-
	clause(C),
	succedent(C,S),
	member([_,Q,==,'$true$'],S).





occurs_positively(R/N,P):-
	(  rule(_,C),
	   (hasSelectedLiteral(C) -> 
		N=0
	   ;
	        true)
	;
	   N=0,
	   clause(C)
	),
	succedent(C,S),
	member(A,S),
	pred_occurs_in_atom(A,R/N,P),
	!.

pred_occurs_in_atom(A,R/N,P):-
	(    get_atom_relation_symbol(A,R), 
	     N=2,
	     (nonvar(P) ->
		  \+ \+ match(A,P)
	     ;
	          true
	     )
	;    get_atom_relation_symbol(A,==),
	     get_atom_members(A,S,T),
	     (   predicate_in_formula(S,R/N,P)
	     ;
	         predicate_in_formula(T,R/N,P)
	     )
	),
        !.



	
predicate_in_atom(A,R/N):-
	get_atom_relation_symbol(A,Rel),
	(Rel= == ->
	     get_atom_members(A,S,T),
	     (   predicate_in_formula(S,R/N,_)
	     ;
	         predicate_in_formula(T,R/N,_)
	     )
	;    
	     R=Rel,
	     N=2
	).

	



%------------------------------------------------------------------------------
% returns the possible atoms corresponding to the atom
%------------------------------------------------------------------------------

atom_as_term( [_,'==',Atom,'$true$'], Atom ):- !.
atom_as_term( [_,'==',Atom,'$min'], Atom ):- !.
atom_as_term( [_,RS,S,T], Atom ):-
	Atom =.. [RS,S,T].
atom_as_term( [_,RS,S,T], Atom ):-
	symmetric_relation_symbol( RS ),
	Atom =.. [RS,T,S], !.



%------------------------------------------------------------------------------
% checks whether two atoms are variants*
%------------------------------------------------------------------------------
variants_atoms( [_,_,S,T], [_,_|E] ):-
	( variant( [S,T], E ) ; variant( [T,S], E ) ), !.


equal_atoms( [_|E1], [_|E2] ):- E1 == E2, !.
equal_atoms( [_,RS,S,T], [_,RS|E] ):-
	symmetric_relation_symbol( RS ),
	[T,S]==E.


subset_atoms([],_).
subset_atoms([A|As],Bs):-
	     member(B,Bs),
	     equal_atoms(A,B),
	     !,
	     subset_atoms(As,Bs).

delete_atom(_,[],[]).
delete_atom(A,[B|Bs],Bs):-
	equal_atoms(A,B),
	!.
delete_atom(A,[B|Bs],[B|Cs]):-
	delete_atom(A,Bs,Cs).

%------------------------------------------------------------------------------
% orients (if possible) an atom
%------------------------------------------------------------------------------

orient_atom( [l|R], [l|R] ):-
	!.
orient_atom( [r|R], [r|R] ):-
	!.
orient_atom( A, O ):- 
	orient_atom_fresh( A, O ).

orient_atom_fresh( [_,==,S,'$true$'], [l,==,S,'$true$'] ):-
	!.
orient_atom_fresh( [_,==,S1,T1], [Pre,==,S2,T2] ):-
	firstOrder,
	!,
	(greater_literal(+,[_,==,S1,'$true$'],-,[_,==,T1,'$true$']) ->
	     S2=S1,
	     T2=T1,
	     Pre=l
	;greater_literal(+,[_,==,T1,'$true$'],-,[_,==,S1,'$true$']) ->
	     S2=T1,
	     T2=S1,
	     Pre=l
	;
	     S2=S1,
	     T2=T1,
	     Pre=n
	).
orient_atom_fresh( [_,RS,S1,T1], [l,RS,S1,T1] ):-
	greater( S1, T1 ), !.
orient_atom_fresh( [_,RS,S1,T1], [Pre,RS,T11,S11] ):-
	greater( T1, S1 ),
	!,
	(symmetric_relation_symbol( RS ) ->
	     T11=T1,
	     S11=S1,
	     Pre=l
	;
	     T11=S1,
	     S11=T1,
	     Pre=r
	).
orient_atom_fresh( [_,RS,S1,T1], [n,RS,S1,T1] ).


%------------------------------------------------------------------------------
% gives (under backtracking) all the possible orientations of the given atom
%------------------------------------------------------------------------------



symmetric_version_of_atom( E, E ).
symmetric_version_of_atom( [l,RS,S,T], [r,RS,T,S] ):-
	symmetric_relation_symbol( RS ).
symmetric_version_of_atom( [r,RS,S,T], [l,RS,T,S] ):-
	symmetric_relation_symbol( RS ).
symmetric_version_of_atom( [n,RS,S,T], [n,RS,T,S] ):-
	symmetric_relation_symbol( RS ).


%------------------------------------------------------------------------------
% gives (under backtracking) the possible rewrite rules L --> R of the atom
%------------------------------------------------------------------------------


% outdated; should be eliminated
atom_for_demodulation( [_,=,L,R]==[_,=,L1,R1], l, ==, L=R, L1=R1 ):- 
	!.
atom_for_demodulation( [P,RS,L,R], P1, RS, L1, R1 ):- 
	symmetric_relation_symbol( RS ),
	!,
	orient(P,L,R,P1,L1,R1).
atom_for_demodulation( [_P,RS,L,R], l, ==, L1, '$true$' ):- 
	L1=..[RS,L,R].
atom_for_demodulation( [_P,RS,L,R], l, ==, L1, '$min' ):- 
	total_relation_symbol(RS),
	reflexive_closure(RS,RR),
	(L1=..[RR,R,L];L1=..[RS,R,L]).
atom_for_demodulation( [_P,RR,L,R], l, ==, L1, '$min' ):- 
	reflexive_closure(RS,RR),
	total_relation_symbol(RS),
	L1=..[RS,R,L].



orient( l,L,R, l, L, R ).
orient( r,L,R, l, R, L ). 
orient( n,L,R, n, L, R ).
orient( n,R,L, n, L, R ):-
	\+variant((L,R),(R,L)).



%------------------------------------------------------------------------------
% 
%------------------------------------------------------------------------------

atom_as_equation( [P,RS,L,R], [P, RS, L, R] ):- 
	symmetric_relation_symbol( RS ),!.
atom_as_equation( [_,RS,S,T], [l,'==', Pred, '$true$'] ):-
	Pred =.. [RS,S,T], !.



%------------------------------------------------------------------------------
% 
%------------------------------------------------------------------------------
equation_as_atom( [_,'==',L,_], [n, RS, S, T] ):- 
	nonvar(L),
	L=..[RS,S,T],
	transitive_relation_symbol(RS),
	!.
equation_as_atom( [P,RS,L,R], [P, RS, L, R] ).



atoms_as_inequations([],[]).
atoms_as_inequations([[_P,RS,L,R]|A],[(1,RS,L,R)|I]):-
	!,
	atoms_as_inequations(A,I).
atoms_as_inequations([[_P,RS,R,L]|A],[(-1,RS,L,R)|I]):-
	atoms_as_inequations(A,I).



inverse_atoms_if_possible([],[],[]).
inverse_atoms_if_possible([A|As],[B|Bs],Cs):-
	inverse_atom(A,B),
	inverse_atoms_if_possible(As,Bs,Cs),
	!.
inverse_atoms_if_possible([A|As],Bs,[A|Cs]):-
	inverse_atoms_if_possible(As,Bs,Cs),
	!.




inverse_atom([_,RS,L,R],[n,RS1,R,L]):-
	strict_part(RS,RSS),
	total_ordering_relation_symbol(RSS),
	reflexive_closure(RSS,RSR),
	(RS=RSS -> RS1=RSR; RS1=RSS),
	!.


inverse_equation([_,=,L,R],[n,RS,R,L]):-
	total_ordering_relation_symbol(RS).
inverse_equation([_,=,L,R],[n,RS,L,R]):-
	total_ordering_relation_symbol(RS).


weakly_inverse_atom(A,AI):-
	strongly_inverse_atom(A,AI).
weakly_inverse_atom(A,AI):-
	inverse_equation(A,AI).


strongly_inverse_atom(A,AI):-
	inverse_atom(A,AI),
	!.
strongly_inverse_atom([_,RS,L,R],[n,RS,R,L]):-
	transitive_relation_symbol(RS),
	irreflexive_relation_symbol(RS).



%--------------------------------------------------------------------------------
% gets all the variables of the given atom
%--------------------------------------------------------------------------------

get_atom_vars( [_,_,S,T], L ):-
	get_term_vars( t(S,T), L ).

%--------------------------------------------------------------------------------
% gets all the variables of the given list of atoms
%--------------------------------------------------------------------------------

get_list_atoms_vars( [], [] ):- !.
get_list_atoms_vars( [E|R], L ):-
	get_atom_vars( E, L1 ),
	get_list_atoms_vars( R, L2 ),
	append( L1, L2, L3 ),
	sort( L3, L ), !.

%--------------------------------------------------------------------------------
% subterm searching plus replacement
% replacesubterm_ofatom_by_giving_( Sub, E1, R, E2 )
%	Sub is a subterm of the atom E1 wich is replaced by R giving
%	the new atom E2
%--------------------------------------------------------------------------------




replacesubterm_ofatom_by_giving_( Sub, [_K,Op,S,T], R, [n,==,'$true$',R] ):-
	Op\== ==,
	Op\== =,
	Sub=..[Op,S,T].
replacesubterm_ofatom_by_giving_( Sub, [K,Op,S,T], R, [K,Op,NS,T] ):- 
	replacesubterm_eq( Sub, S, R, NS ).
replacesubterm_ofatom_by_giving_( Sub, [K,Op,S,T], R, [K,Op,S,NT] ):- 
	replacesubterm_eq( Sub, T, R, NT ).


%----------------------------------------------------------------------------
% atom matching
%----------------------------------------------------------------------------

% because of cut we may loose 
% subsumptions due to symmetry

match_atoms(E1,E2,M1,M2):-
	get_atom_relation_symbol(E1,RS),
	match_atoms(RS,E1,E2,M1,M2).



match_atoms(_RS,[_,RS1|E1],[_,RS2|E2],M1,M2):-
	\+flag_on( rp, ccrw ),
	equivalent_relation_symbols(RS1,RS2),
	match( E1, E2, M1, M2 ),
	!.
match_atoms(_RS,[_,RS1|E1],[_,RS2|E2],M1,M2):-
	flag_on( rp, ccrw ),
	equivalent_relation_symbols(RS1,RS2),
	match( E1, E2, M1, M2 ).
match_atoms(RS,[_,RS,S1,T1],[_,RS1|E2],M1,M2):-
	equivalent_relation_symbols(RS1,RS),
	symmetric_relation_symbol( RS ),
	match( [T1,S1], E2, M1, M2 ).

match_light_atoms(_RS,E1,E2,M1,M2):-
	\+flag_on( rp, ccrw ),
	match( E1, E2, M1, M2 ),
	!.
match_light_atoms(_RS,E1,E2,M1,M2):-
	flag_on( rp, ccrw ),
	match( E1, E2, M1, M2 ).
match_light_atoms(RS,E1,E2,M1,M2):-
	symmetric_relation_symbol( RS ),
	E1=[S1,T1],
	match( [T1,S1], E2, M1, M2 ).






match_atom_terms([_,_RS,L,R],[_,_RS1,L1,R1],O,M1,M2):-
	(
	    match([L,R],[L1,R1],M1,M2),
	    O= +
	;
	    match([R,L],[L1,R1],M1,M2),
	    O= -
	).









% for skolemized terms

match_atoms([_,RS1|A1],[_,RS2|A2]):-
	equivalent_relation_symbols(RS1,RS2),
	match_atoms1([_,RS1|A1],[_,RS1|A2]).

match_atoms1([_,RS,S1,T1],[_,RS,S2,T2]):-
	\+flag_on( rp, ccrw ),
	match((S1,T1),(S2,T2)),   % note that this also cuts off all but
	% one ac matchers
	!.
match_atoms1([_,RS,S1,T1],[_,RS,S2,T2]):-
	flag_on( rp, ccrw ),
	match((S1,T1),(S2,T2)).
match_atoms1([_,RS,S1,T1],[_,RS,S2,T2]):-
   	symmetric_relation_symbol( RS ),
	match((T1,S1),(S2,T2)).


% for light-weight atoms

match_light_atoms(_RS,E1,E2):-
	\+flag_on( rp, ccrw ),
	match( E1, E2 ),
	!.
match_light_atoms(_RS,E1,E2):-
	flag_on( rp, ccrw ),
	match( E1, E2 ).
match_light_atoms( RS, [S1,T1], E ):-
	symmetric_relation_symbol( RS ),
	match( [T1,S1], E ).




/*-----------------------------------------------------------------------------



 We chain two atoms A1 and A2 by basically executing 2 steps:

 Step 1:  ( predicates atom_as_rule[_from]):
    A1 and A2 are oriented (using pre-orientation information) into rules 
    L1 RS1^Pol1 R1 and L2 RS2^Pol2 R2 modulo symmetry of RS1 and RS2
    with maximal terms L1 and L2, respectively.


 Step 2: ( predicate peak_candidate ):

    Now, 4 cases are possible:
    
    (i)    K[R1] RS1^{-1*Pol1} K[L1] == L2    RS2^Pol2 R2
    (ii)   R2    RS2^{-1*Pol2} L2    == K[L1] RS1^Pol1 K[R1]  

    (iii)  R1    RS1^{-1*Pol1} L1    == K[L2] RS2^Pol2 K[R2]
    (iv)   K[R2] RS2^{-1*Pol2} K[L2] == L1    RS1^Pol1 R1 

    Cases (iii) and (iv) are dual to (i) and (ii) and are dealt with
    when the premise order is swapped.
    Also, empty contexts K=[] in cases (iii) and (iv) need not be reconsidered
    (cf. inference/7).
    
    Reconsideration of symmetric cases (modulo symmetry of relations) are
    avoided (cf. symmetric_variants/4).

 

 Step 3: ( predicates chain and negative_chain ):
    Independent of the two cases (i) and (ii), L1 must unify a subterm of L2.
    This may be a proper subterm whenever RS1 is a partial congruence.
    In case of AC-operators, extensions for L1 -> R1 must be considered


-----------------------------------------------------------------------------*/

peak_candidate(A1, A2, [Pol1,P1,RS1,S1,T1],[Pol2,P2,RS2,S2,T2]):-
	A1=[_,RS1,S11,T11],
	A2=[_,RS2,S21,T21],
	S11\==T11,
	S21\==T21,
	(RS1= ==, RS2= == ->
	     T11\==T21
	;
	     true
	),


	atom_as_rule_from(A1,[P1,RS1,S1,T1],Pol1), % rule (S1 Pol1(RS1) T1)
	atom_as_rule(A2,[P2,RS2,S2,T2],Pol2),      % rule (S2 Pol1(RS2) T2)
	\+ symmetric_variants(Pol1,RS1,Pol2,RS2),
	Pol1 is -Pol2.



/*-----------------------------------------------------------------------------

 atom_as_rule(A,[P,RS,L,R], Pol)

 orients an atom 
        A = L' RS R'
 into a "rule" 
        L Pol(RS) R, 
 such that L > R is satisfiable and such that Pol(RS)=RS,
 if L=L' and R=R' and Pol(RS) = RS^{-1}, otherwise.
 That is if the ordering requires to swap the terms we
 change the relation RS to its inverse Pol(RS). Technically,
 Pol is either 1 (Pol(RS)=RS), or -1, otherwise.

 Similarly, chaining left, by again superposing L and L',
 is needed if a rewrite proof for RS' can begin with an RS
 step.

 
 For symmetric relations RS, any rule L RS R also gives rise
 to the symmetric variant L RS^{-1} R.

-----------------------------------------------------------------------------*/

atom_as_rule( [P,RS,L,R], [P, RS, L, R], Pol ):-
	P\== r,
	symmetric_polarity(1,RS,Pol),
	(Pol= -1 -> 
	     % P must have been u   % no, but still ok !?
	     L\==R
	; 
	     true
	).
atom_as_rule( [r,RS,L,R], [l, RS, R, L], Pol ):-
	symmetric_polarity(-1,RS,Pol).
atom_as_rule( [n,RS,L,R], [n, RS, R, L], Pol ):-
	R\==L,
	symmetric_polarity(-1,RS,Pol).

atom_as_rule_from( A, AR, Pol ):-
	atom_as_rule(A,AR,Pol).
atom_as_rule_from( A, AR, Pol ):-
	induced_equivalence(A,AI),
	atom_as_rule(AI,AR,Pol).





induced_equivalence( [P,RS,L,R], [P,==,L1,R1] ):-
	RS\== ==,
	\+ partial_congruence_relation_symbol(RS),
	partial_equivalence_relation_symbol(RS),
	predicate(Pr/N),
	\+ (N=2,special_relation(Pr)),
	length(Args,N),
	append(LA,[_|RA],Args),
	append(LA,[L|RA],ArgsL1),
	append(LA,[R|RA],ArgsR1),
	L1=..[Pr|ArgsL1],
	R1=..[Pr|ArgsR1].
	


symmetric_polarity(Pol,RS,P):-
	\+ symmetric_relation_symbol(RS),
	!,
	P=Pol.
symmetric_polarity(_Pol,_RS,1).
symmetric_polarity(_Pol,_RS,-1).


chain(A1, A2, [Pol1,P1,RS1,S1,T1],[Pol2,P2,RS2,S2,T2],[n,CRS,L,R], [[Sub,S1]], Pos, STree ):-
	peak_candidate(A1,A2,[Pol1,P1,RS1,S11,T11],[Pol2,P2,RS2,S21,T21]),
	(partial_congruence_relation_symbol(RS1) -> 
	     extension(S11,T11,S1,T1),
	     (    S2=S21,
		  T2=T21,
		  path_arg1(Pos,S2,Sub)
	     ;
		  partial_congruence_relation_symbol(RS2),
		  Pos=[],
	          extension(S21,T21,S2,T2),
		  S2\==S21,    % for proper extension; only
                               % additional root overlaps needed
		  Sub=S2
	     ),
	 
	     CRS=RS2
	; 
	     Pos=[],
	     S1=S11,T1=T11,
	     (partial_congruence_relation_symbol(RS2) ->
		  extension(S21,T21,S2,T2)
	     ;
	          S2=S21,
		  T2=T21
	     ),
	     (Pol2=1 ->
		  composition_of_relations( RS1, RS2, CRS )
	     ;
	          composition_of_relations( RS2, RS1, CRS )
	     ),
	     S2=Sub
	),

	non_redundant_overlap(RS1,S1,Pos,Sub),
	(( partial_equivalence_relation_symbol(RS2)
	 ; partial_congruence_relation_symbol(RS2) ) ->
	       nonvar(S1)
	;
	       true
	),


	\+ \+ ( 
	      unify( S1, Sub ),
%	      ground(S1),
	      irreducibleLeftOf(Pos,STree,S2),
	      irreducibleNonACSubterms(STree,Sub)   
	      ), 

	change_path_arg(Pos,S2,NS2,T1),
	orient_by_polarity(Pol2,NS2,T2,L,R).



reduction_candidate(A1, A2, [Pol,P1,RS1,S1,T1],[Pol,P2,RS2,S2,T2]):-
	A1=[_,RS1,S11,T11],
	chainingRelation(RS1),
	S11\==T11,
	atom_as_rule_from(A1,[P1,RS1,S1,T1],Pol), % rule (S1 Pol(RS1) T1)
	atom_as_rule(A2,[P2,RS2,S2,T2],Pol),      % rule (S2 Pol(RS2) T2)
	\+ symmetric_variants(-Pol,RS1,Pol,RS2).





%------------------------------------------------------------------------------
% negative_chain(E1, E2, NE1, NE2, NE, L, Pos ):
%	checks whether E1 can be chained with E2 (on the [left|right]
%       hand side of E2)
%	giving also the unification to be done and the position
%	where the superposition takes place.
%       NE1 and NE2 are the possibly symmetric versions of E1 and E2
%       indicating the used orientation in the case of a symmetric atom
%------------------------------------------------------------------------------

negative_chain( A1, A2, [P1,RS1,S1,T1],[P2,RS2,S2,T2],[n,CRS,L,R], [[Sub,S1]], Pos ):-

	reduction_candidate(A1, A2, [Pol,P1,RS1,S11,T11],[Pol,P2,RS2,S2,T2]),
	(partial_congruence_relation_symbol(RS1) -> 
	     CRS=RS2,
	     path_arg1(Pos,S2,Sub),
	     extension(S11,T11,S1,T1)
	; 
	     Pos=[],
	     residual( Pol, RS1, CRS, RS2 ),
	     S11=S1,
	     T11=T1,
	     S2=Sub
	),

	non_redundant_overlap(RS1,S1,Pos,Sub),
	\+ \+ (unify( S1, Sub )
%	       ,ground(S1)
	      ),
	
	change_path_arg(Pos,S2,NS2,T1),
	orient_by_polarity(Pol,NS2,T2,L,R).




residual(-1,RS1,Res,RS2):-
      residual( Res, RS1, RS2 ).
residual(1,RS1,Res,RS2):-
      residual( RS1, Res, RS2 ).

chainingRelation(RS):-
	special_relation(RS).


non_redundant_overlap(==,_S1,[],_Sub):-
	!.
non_redundant_overlap(RS1,S1,Pos,Sub):-
        \+ var_on_predicate( S1, Sub ),

	( partial_equivalence_relation_symbol(RS1) -> 
	     nonvar(Sub)
	;
	    \+ var_overlap(S1,Sub)
	),
	(partial_congruence_relation_symbol(RS1) ->
	     nonvar(Sub)
	;    
             Pos=[],      % only for symmetric relations we build 
                          % monotonicity into the chaining inference
%	     ( symmetric_relation_symbol(RS1) ->
%		  nonvar(Sub)
%	     ;
	    (strict_part(RS1,RSS),
		dense_ordering_wo_end_relation_symbol(RSS)
	    -> 
		nonvar( Sub ),
		nonvar(S1) 
	    ;
		\+ var_overlap(S1,Sub)
		% format('~nVariable overlap!',[])
	    )
%	     )
	).


var_overlap(S1,Sub):-
	% for general transitive relations depending on a flag
	var_overlaps(off),
	!,
	(
	  var(Sub),
	  \+ground(S1)
	;
	  var(S1),
	  \+ground(Sub)
	).

	


irreducibleNonACSubterms(_,T):-
	var(T),
	!.
irreducibleNonACSubterms(STree,T):-
	functor(T,F,N),
	(N=2,currentTheory(ac),ac_ist_AC(F) ->
	     arg(1,T,T1),
	     irreducibleNonACSubterms(STree,T1),
	     arg(2,T,T2),
	     irreducibleNonACSubterms(STree,T2)
	;
	     N1 is N+1,
	     irreducibleLeftOf(N1,STree,T)
	).

irreducibleLeftOf([],_,_).
irreducibleLeftOf([I|Is],STree,T):-
	irreducibleLeftOf(I,STree,T),
	arg(I,T,TI),
	irreducibleLeftOf(Is,STree,TI).
irreducibleLeftOf(1,_,_):-
	!.
irreducibleLeftOf(I,STree,T):-
	number(I),
	J is I-1,
	arg(J,T,TJ),
	true(Top),
	\+ (numbervars(TJ,0,_), rewrite: reducible(termByUnits,ct([],Top),TJ,STree)),
	irreducibleLeftOf(J,STree,T).


symmetric_variants(P1,RS1,P2,RS2):-
	symmetric_relation_symbol(RS1),
	symmetric_relation_symbol(RS2),
	1 is P1,
	-1 is P2,
	!.


orient_by_polarity(1,S,T,S,T).
orient_by_polarity(-1,S,T,T,S).


redundant_reflexive(R,S,T):-
	S==T,
	(reflexive_relation_symbol(R);irreflexive_relation_symbol(R)),
	!.


resolution( [_,RS,X,Y], [_,RS,S1,S2], [[X,S1],[S1,S2]] ):-
	var(X),
	var(Y),
	X==Y,       %special handling of (ir)reflexivity
	\+ \+ unify(S1,S2),
	!.
resolution( [_,RS,S1,S2], [_,RS,X,Y], [[X,S1],[S1,S2]] ):-
	var(X),
	var(Y),
	X==Y,
	\+ \+ unify(S1,S2),
	!.
resolution( A1, A2, [[S1,S2],[T1,T2]] ):-
	get_atom_relation_symbol(A1,RS),
	get_atom_relation_symbol(A2,RS),
	symmetric_relation_symbol(RS),
	\+ chainingRelation(RS),
	!,
	get_atom_members(A1,S1,T1),
	symmetric_version_of_atom(A2, [_P2,RS,S2,T2]),
	\+ \+ (unify( S1,S2), unify(T1,T2)).
resolution( A1, A2, [[S1,S2],[S1,T1],[S1,T2]] ):-
	get_atom_relation_symbol(A1,RS),
	\+irreflexive_relation_symbol(RS),
	get_atom_members(A1,S1,T1),
	symmetric_version_of_atom(A2, [_P2,RS,S2,T2]),
	\+ \+ (unify( S1,S2), unify(S1,T1), unify(S1,T2)).
resolution( A1, A2, [[S1,S2],[T1,T2]] ):-
	get_atom_relation_symbol(A1,RS),
	total_relation_symbol(RS),  % in all other cases the residual
	    % includes =, so that reflexivity resolution is all we need
	    % for total relations we produce not(R) positively as residual
            % upon negative chaining, so that resolution is needed
	get_atom_members(A1,S1,T1),
	symmetric_version_of_atom(A2, [_P2,RS,S2,T2]),
	\+ (\+ unify( [S1,T1], [S2,T2] ) ).

/*
*/

factorization( E1, E2, NE1, NE2, Ant, Suc, [[S1,S2]|Un] ):-
	get_atom_relation_symbol(E1,RS),
	get_atom_relation_symbol(E2,RS),

%	\+ (strict_part(RS,RSS), total_ordering_relation_symbol(RSS)),

	atom_as_rule( E1, NE1, Pol),
	atom_as_rule( E2, NE2, Pol),
	NE1=[_,RS, S1, T1], 
	NE2=[_,RS, S2, T2],
	\+ symmetric_variants(-Pol,RS,Pol,RS),
	\+ var_on_predicate( S1, S2 ),
	\+ var_on_predicate( S2, S1 ),
        \+ \+ unify( S1, S2 ),
	(\+ \+ (unify(S1,S2), (T1==T2; var(T1),var(T2))) ->
	     % this is an optimization
                Ant=[],
		Un=[[T1,T2]],
		Suc=[E1]
	;       orient_by_polarity(Pol,T1,T2,W1,W2),
		(strict_part(RS,RSS),total_relation_symbol(RSS) ->
		     tr_factors(RS,W1,W2,E1,E2,Suc),
		     Ant=[],
		     Un=[]
		;    
		     residual(Pol,RS,RS1,RS),
		     Ant=[[n,RS1,W1,W2]],
		     Un=[],
		     Suc=[E2]
		 )
	).


tr_factors(RS,W1,W2,_E1,E2,[[n,RSI,W2,W1],E2]):-
	(reflexive_relation_symbol(RS) ->
	     strict_part(RS,RSI)
	; reflexive_closure(RS,RSI)).
tr_factors(RS,W1,W2,E1,_E2,[[n,RS,W1,W2],E1]).


var_on_predicate( S1, Sub ):-
	var( S1 ),
	nonvar( Sub ),
	(  Sub='$true$'
	;  Sub='$min'
	;  functor( Sub, F, N ),
	   (  is_predicate( F/N )
	   ;
%	      logical_symbol(F/N)
	      logical_operator(F)
	   ;
	      generated_proposition(F/N)
	   )
	;  functor( Sub, F, _ ),
	   has_external_sort(F)
	),
	!.

var_on_literal( S1, Sub ):-
	var( S1 ),
	nonvar( Sub ),
	(Sub=[R,_,_];Sub=[_,R,_,_]),
	atom(R),
	special_relation(R),
	!.

subsumed_lr_pair('==',S2,RS1):-
	nonvar(S2),
	functor(S2,RS,2),
	composition_of_relations(RS1, RS, _),
	!.






%------------------------------------------------------------------------------
% 
%------------------------------------------------------------------------------


possible_unification_whole_atom( E1, E2, [Atom1,Atom2] ):-
	atom_as_term( E1, Atom1 ),
	atom_as_term( E2, Atom2 ),
	\+ ( \+ unify( Atom1, Atom2 ) ), !.


%------------------------------------------------------------------------------
% greater_literal( P1, A1, P2, A2 ) 
%  is true when
%  A1 with polarity P1 is a greater atom than A2 with P2.
%------------------------------------------------------------------------------


/*  buggy
greater_literal(P1,E1,P2,E2):-
	firstOrder,
	!,
	true(Top),
	atom_in_expansion([P1|E1],[P|E]),
	atom_as_term(E,ET),
	(general_function_call(ET) ->
	    E3=[l,==,ET,'$true$']
	;
	    E3=E),
	\+ (   atom_in_expansion([P2|E2],[Q|F]),
	       atom_as_term(F,FT),
	       ( general_function_call(FT) ->
		   F3=[l,==,FT,'$true$']
	       ;   
		   F3=F),
	      \+ greater_literal(P,E3,Q,F3,Top)
	   ),
	!.
*/
greater_literal(P1,E1,P2,E2):-
	firstOrder,
	!,
	true(Top),
	atom_in_expansion([P1|E1],[P|E]),
	\+ (   atom_in_expansion([P2|E2],[Q|F]),
	      \+ greater_literal(P,E,Q,F,Top)
	   ),
	!.
greater_literal(P1,E1,P2,E2):-
	true(Top),
	greater_literal(P1,E1,P2,E2,Top).




greater_literal(_P1,E1,_P2,E2,_):-
	get_binary_atom(E1,_R1,S1,T1),
	get_binary_atom(E2,_R2,S2,T2),
	( greater(S1,S2), greater(S1,T2)
	; greater(T1,S2), greater(T1,T2)
	; greater(S1,S2), greater(T1,T2)
	; greater(T1,S2), greater(S1,T2)
	),
	!.
/* no, that's too weak for top predicates
greater_literal(_,E1,_,E2,_):-
	get_atom_vars(E1,V1),
	get_atom_vars(E2,V2),
	member(V,V2),
	\+ member212(V,V1),
	!,
	fail.
*/
greater_literal(P1,E1,P2,E2,Ct):-
	% this is the general case when the previous heuristics fail
	constraint_grequal_literal(E2,P2,E1,P1,CS),
	and(Ct,CS,CtS),
	false(CtS),
	!.

% -----------------------------------------------------------------------
%  all congruences are equivalent
%
% -----------------------------------------------------------------------
equivalent_relation_symbols(RS,RS1):-
	nonvar(RS1),
	RS=RS1,
	!.
equivalent_relation_symbols(RS,RS).
equivalent_relation_symbols(RS1,RS2):-
	congruence_relation_symbol(RS1),
	congruence_relation_symbol(RS2).
	


% -----------------------------------------------------------------------
%  = is not in the operator precedence
%
% -----------------------------------------------------------------------
greater_relation_symbol(RS1,=):-
	!,
	RS1\== '='.
greater_relation_symbol(RS1,'=='):-
	!,
	RS1\== '=',
	RS1\=='=='.
greater_relation_symbol(RS1,RS2):-
	partial_equivalence_relation_symbol(RS2),
	!,
	(partial_equivalence_relation_symbol(RS1) ->
	     RS1\== =,
	     greater_op(RS1,RS2)
	;
	     true
	).
greater_relation_symbol(RS1,RS2):-
	RS1\==RS2,
	reflexive_closure(RS1,RS2),
	!.
greater_relation_symbol(RS1,RS2):-
	strict_part(RS1,RSS1),
	strict_part(RS2,RSS2),
	greater_op(RSS1,RSS2),
	!.
greater_relation_symbol(RS1,RS2):-
	greater_op(RS1,RS2).





%--------------------------------------------------------------------------------
% greater_max_term( [S1,T1], [S2,T2] ):
%	max[S1,T1] is greater than max[S2,T2] for all gnd inst
%--------------------------------------------------------------------------------

greater_max_term( [S1,T1], [S2,T2] ):-
	( greater( S1, S2 ) ; greater( T1, S2 ) ),
	( greater( S1, T2 ) ; greater( T1, T2 ) ),!.


%--------------------------------------------------------------------------------
% equal_max_term( [S1,T1], [S2,T2] ):
%	max[S1,T1] is equal to max[S2,T2] for all gnd inst
%--------------------------------------------------------------------------------

equal_max_term( [S1,T1], [S2,T2] ):-
	maximal_term_lpo( S1, T1, M1 ),
	maximal_term_lpo( S2, T2, M2 ), 
	M1 == M2.
equal_max_term( E1, E2 ):-
	equal_ms_two( E1, E2 ), !.

%--------------------------------------------------------------------------------
% equal_ms_two( [S1,T1], [S2,T2] ):
%	[S1,T1] is the same multiset as [S2,T2]
%--------------------------------------------------------------------------------

equal_ms_two( E1, [S2,T2] ):-
	E1 == [S2,T2], !.
equal_ms_two( E1, [S2,T2] ):-
	E1 == [T2,S2], !.
	
%--------------------------------------------------------------------------------
% greater_ms_two( [S1,T1], [S2,T2] ):
%	[S1,T1] is greater than [S2,T2] w.r.t. the multiset extension of lpo
%--------------------------------------------------------------------------------

greater_ms_two( [S1, T1], [S2, T2] ):-
	S1 == S2, !, greater( T1, T2 ), !.
greater_ms_two( [S1, T1], [S2, T2] ):-
	S1 == T2, !, greater( T1, S2 ), !.
greater_ms_two( [S1, T1], [S2, T2] ):-
	T1 == S2, !, greater( S1, T2 ), !.
greater_ms_two( [S1, T1], [S2, T2] ):-
	T1 == T2, !, greater( S1, S2 ), !.
greater_ms_two( S1, T1, S2, T2 ):-
	greater( S1, S2 ), !,
	( greater( S1, T2 ) ; greater( T1, T2 ) ), !.
greater_ms_two( S1, T1, S2, T2 ):-
	greater( T1, S2 ),
	( greater( S1, T2 ) ; greater( T1, T2 ) ), !.

%--------------------------------------------------------------------------------
% gets the operators of the given atoms 
%--------------------------------------------------------------------------------

get_atom_operators( [_,_,S,T], L ):-
	get_term_operators( S, L1 ),
	get_term_operators( T, L2 ),
	append( L1, L2, L ), !.

%--------------------------------------------------------------------------------
% gets the operators of the given list of atoms 
%--------------------------------------------------------------------------------

get_list_of_atoms_operators( [], [] ):-!.
get_list_of_atoms_operators( [E|A], L ):-
	get_atom_operators( E, L1 ),
	get_list_of_atoms_operators( A, L2 ),
	append( L1, L2, L ), !.

%--------------------------------------------------------------------------------
% *
%--------------------------------------------------------------------------------

get_eq_members_list( [], [] ):- !.
get_eq_members_list( [A|R1], [S=T|R2] ):-
	atom_for_demodulation( A, _, =, S, T ), !,
	get_eq_members_list( R1, R2 ).





subsumes_literals([],_,M,M).
subsumes_literals([[Sign,_,RS|A]|RL],CL,M1,M2):-
	delete32([Sign,_,RS|CA],CL,CL1),
	match_light_atoms(RS,A,CA,M1,M3),
	subsumes_literals(RL,CL1,M3,M2).

	     

inverse_sign(-,+).
inverse_sign(+,-).





neg_atom_as_literal(A,[-|A]).
pos_atom_as_literal(A,[+|A]).


%------------------------------------------------------------------------------
%	atom printing	
%------------------------------------------------------------------------------

print_atom([+,Pre,RS,L,R]):-
	print_atom([Pre,RS,L,R]).
print_atom([-,Pre,RS,L,R]):-
	write('~'),
	print_atom([Pre,RS,L,R]).
print_atom( [_, ==,'$true$','$true$'] ):- write( 'true' ), !.
print_atom( [_, ==,'$true$','$min'] ):- write( false ), !.
print_atom( [_, ==,'$min','$true$'] ):- write( false ), !.
print_atom( [_, ==, S,'$true$'] ):- print( S ), !.
print_atom( [_, ==, '$true$', T] ):- print( T ), !.
print_atom( [_P,RS,S,T] ):-
	A=..[RS,S,T],
	print(A).

print_list_atoms( [] ):- !.
print_list_atoms( L ):-
	ple( L ), !.

ple( [] ):- !.
ple( [E] ):-
	print_atom( E ),!.
ple( [E|R] ):-
	print_atom( E ),
	write( ',' ),
	ple( R ),!.





% eof ------------------------------------------------------------------
