

%declarations%

:- module(adt_declarations, [ 
	all_flag_types/1,
	all_keys/2,
	all_numbers/2,
	all_texts/2,
	binary_inference/1,
	hi/3,
	message/2,
	simplification_inference/1,
	unary_inference/1
   ]).

:- use_module(props, [
	flag_on/2
   ]).

:- use_module(orderConstraints, [
	true/1
   ]).

:- use_module(adt_atom, [
	print_list_atoms/1
   ]).

:- use_module(adt_proof, [
	s_depth/2
   ]).

:- use_module(adt_term, [
	deBruijn/2
   ]).

:- use_module(classes, [
	antecedent/2,
	succedent/2,
	history/2,
	implicants/2,
	number/2,
	constraint/3,
	weight/2,
	weightBound/2
   ]).

:- use_module(output, [
	print_terms/1,
	skolem/3
   ]).



%end%

%						ADT_DECLARATIONS

%------------------------------------------------------------------------------
%	Declarations for menus, messages and proofs
%------------------------------------------------------------------------------

%------------------------------------------------------------------------------
%	ci:	constraint inheritance
%	ir:	inference rules
%       es:     eager simplifiactions
%	ma:	marking strategies
%	ms:	messages
%	rp:	redundancy proofs
%	tp:	timing points
%------------------------------------------------------------------------------




% if the flag structure is modified don't forget to accordingly change the
% defaults in defaults.pl


all_flag_types( [rp,ir,es, ci,ma,ms,tp] ).

all_keys( rp, [taut,clsb,rusb,ccon,prop,cdem,rdem,clca,ruca,ccrw,rcrw,
	       tore,exre,rws0,rws2] ).

all_keys( ir, [skex,rere,orre,chri,sncr,fact,trre] ). 
	% the order is significant; potentially simplifying inferences
        % should come first (some cases of rere and orre).

all_keys( es, [vare] ).

all_keys( ci, [chco,orde,equa] ).

all_keys( ma, [maxs,user] ).

all_keys( ms, [	trcl,comw,nwru,
		chri,sncr,fact,rere,orre,trre,
		taut,clsb,rusb,ccon,prop,cdem,rdem,clca,ruca,ccrw,rcrw,vare,
		tore,exre,
		skex,rws0,rws2
	      ] ).

all_keys( tp, [cinf,csim,rsim,cons,clpp,rewr] ).

all_numbers( rp, [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15] ).

all_numbers( ir, [1,2,3,4,5,6,7] ).

all_numbers( es, [1] ).

all_numbers( ci, [1,2] ).

all_numbers( ma, [1,2,3] ).

all_numbers( ms, [1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,
		  20,21,22,23,24,25,26] ).
all_numbers( tp, [1,2,3,4,5,6] ).

all_texts( tp, 	['inference computation',
		'clause simplification',
		'rule simplification',
		'constraint solving',
		'clause preprocessing',
		'rewriting'
		]		).

%------------------------------------------------------------------------------
%	Declaration of redundancy proofs
%------------------------------------------------------------------------------

rp( 0, n,	['		REDUNDANCY PROOFS'		]).
rp( 0, n,	[' '						]).
rp( 1, taut,	['tautology deletion'			]).
rp( 2, clsb,	['forward subsumption'				]).
rp( 3, rusb,	['backward subsumption'				]).
rp( 4, ccon,	['condensement'				]).
rp( 5, prop,	['propositional reasoning for ground clauses'	]).
rp( 6, cdem,	['forward simplification by reduction'	]).
rp( 7, rdem,	['backward simplification by reduction'		]).
rp( 8, clca,	['redundancy test by case analysis (forward)'	]).
rp( 9, ruca,	['redundancy test by case analysis (backward)' 	]).
rp( 10, ccrw,	['redundancy test by clausal rewriting (forward)'	]).
rp( 11, rcrw,	['(toggle not in use)'	]).
rp( 12, tore,	['simplification by totality resolution'	]).
rp( 13, exre,	['simplification by extensionality resolution'		]).
rp( 14, rws0,	['minimal reduction strength'			]).
rp( 15, rws2,	['maximal reduction strength'	]).


%------------------------------------------------------------------------------
%	Declaration of inference rules
%------------------------------------------------------------------------------

ir( 0,	 n,	['		INFERENCE RULES' 			]).
ir( 0,	 n,	[' ' 							]).
ir( 1,	 rere,	['reflexivity resolution'				]).
ir( 2,	 chri,	['right superposition and chaining'			]).
ir( 3,	 sncr,	['left superposition and negative chaining'             ]).
ir( 4,	 fact,	['factoring' 						]).
ir( 5,	 orre,	['ordered resolution' 				        ]).
ir( 6,	 trre,	['transitivity resolution' 			        ]).
ir( 7,	 skex,	['expansion' 			                ]).
ir( 0,	 n,	[' ' 							]).
ir( 0,	 n,	[' ' 							]).
ir( 0,	 n,	['FOR CONSTRAINT INHERITANCE STRATEGIES: type command saci' ]).
ir( 0,	 n,	['FOR SELECTION STRATEGIES: type command sama' 		]).

%------------------------------------------------------------------------------
%	Declaration of eager simplifications
%------------------------------------------------------------------------------
es( 0,	 n,	['		EAGER SIMPLIFICATIONS' 			]).
es( 0,	 n,	[' ' 							]).
es( 1,   vare,  ['variable elimination for orderings'                   ]).



%------------------------------------------------------------------------------
%	Declaration of constraint inheritance
%------------------------------------------------------------------------------

ci( 0,	 n,	['	CONSTRAINT CHECKING AND INHERITANCE' 		]).
ci( 0,	 n,	[' ' 							]).
ci( 1,	 chco,	['complete and immediate checking of ordering constraints' ]).
ci( 2,	 orde,	['hereditary ordering constraints' 			]).
ci( 3,	 equa,	['hereditary equality constraints: basic strategies' 	]).
ci( 0,   n,     [' ']).
ci( 0,   n,     ['Selecting none of the options means approximative and immediate']).
ci( 0,   n,     ['checking of ordering constraints.']).





%--------------------------------------------------------------------------------
%	Declaration of marking strategies
%--------------------------------------------------------------------------------

ma( 0,	 n,	['		LITERAL SELECTION STRATEGIES' 			]).
ma( 0,	 n,	[' ' 							]).
ma( 0,	 n,	['	SELECTION OF ON NEGATIVE LITERALS' 			]).
ma( 0,	 n,	[' ' 							]).
ma( 1, maxs,	['automatic selection of a maximal literal for non-reductive clauses'		]).
ma( 2, poss,	['positive strategy'                    		]).
ma( 3, user,	['user guided literal selection' 			]).

%--------------------------------------------------------------------------------
%	Declaration of timing points
%--------------------------------------------------------------------------------

tp( 0,  n,     ['		TIMING POINTS'	]).
tp( 0,  n,     [' ' 				]).
tp( 1,  cinf,  ['inference computation'		]).
tp( 2,  csim,  ['clause simplification'		]).
tp( 3,  rsim,  ['rule simplification'		]).
tp( 4,  cons,  ['constraint solving'		]).
tp( 5,  clpp,  ['clause preprocessing'		]).
tp( 6,  rewr,  ['rewriting'		        ]).

%--------------------------------------------------------------------------------
%	Declaration of messages
%--------------------------------------------------------------------------------


ms( 0,  n,     ['		MESSAGES', _ 				]).
ms( 0,  n,     [' ',_ 							]).
ms( 0,  n,     ['	CONTROL:', _ 					]).
ms( 0,  n,     [' ', _ 							]).

ms( 1,  trcl,  ['treating clause', ['treating clause ', clnum] 		]).

ms( 2,  comw,  ['computing inferences with',
        ['computing inferences with rule ',runum,
	 ' (weight=', ruWeight, ' depth=', ruDepth, ') up to weight ', weightBound]				]).

ms( 3,  nwru,  ['keeping clause', ['keeping clause', nl, rule] 			]).

ms( 0,  n,	[' ', _ 						]).
ms( 0,  n,	['		INFERENCE SYSTEM:', _ 			]).
ms( 0,  n,	[' ', _ 						]).
ms( 4, chri,	['chaining',
        ['chaining by ', runum, ' from ', runum,
         ' giving', nl, clause]						]).

ms( 5, sncr,	['negative chaining',
        ['negative chaining by ', runum, ' from ', runum,
         ' giving', nl, clause]						]).

ms( 6, fact,	['factoring',
        ['factoring on ', rule, ' giving', nl, clause]			]).

ms( 7, rere,	['reflexivity resolution',
        ['reflexivity resolution on ', runum, ' giving', nl, clause]	]).

ms( 8, orre,	['ordered resolution',
        ['ordered resolution of ', runum, ' on ', runum,
         ' giving', nl, clause]						]).

ms( 9, trre,	['transitivity resolution',
        ['transitivity resolution of ', runum, ' on ', runum,
         ' giving', nl, clause]						]).

ms( 0,  n,     [' ', _ 							]).

ms( 0,  n,     ['		REDUNDANCY PROOFS', _ 			]).

ms( 0,  n,     [' ', _ 							]).

ms( 0,  n,     ['	FOR CLAUSES AND RULES:', _ 				]).

ms( 0,  n,     [' ', _ 							]).

ms( 10,  taut,  ['tautology', ['clause ', clnum, ' is a tautology']	]).

ms( 11,  clsb,  ['subsumed clause',
        ['clause ', clnum, ' is subsumed by rule ', runum]		]).

ms( 12,  rusb,  ['subsumed rule',
        ['clause ', runum, ' is subsumed by ', numlist]		]).

ms( 13,  ccon,  ['condensed clause',
        ['clause ',clnum, ' is condensed into clause', nl, clause]	]).

ms( 14,  prop,  ['ground clause',
        ['clause ', runum, ' is ground and handled propositionally']	]).

ms( 15,  cdem,  ['simplified clause by reduction',
        ['clause ', clnum, ' is reduced by rules ', numlist,
         ' giving', nl, clause]						]).
ms( 16, rdem, ['simplified rule by reduction',
        ['clause ', runum, ' is reducible by ', numlist]		]).
ms( 17, clca, ['redundant clause by case analysis', 
        ['clause ', clnum, ' is redundant by case analysis']		]).
ms( 18, ruca, ['redundant rule by case analysis',
        ['rule ', runum, ' is redundant by case analysis']		]).
ms( 19, ccrw, ['redundant by clausal rewriting',
        ['clause ', clnum, ' is redundant by clausal rewriting']	]).
ms( 20, rcrw,	['simplified rule by clausal rewriting',
        ['rule ', runum, ' is simplified by clausal rewriting by rule ',
        runum, ' giving', nl, clause]					]).
ms( 21, vare,  ['variable elimination',
        ['from ', clnum, ' by variable elimination', nl, clause]	]).
ms( 22, tore,	['totality resolution',
        ['totality resolution on ', clause, ' giving', nl, clause]		]).

ms( 23, exre,	['extensionality resolution',
        ['extensionality resolution on ', clause, ' giving', nl, clause]		]).
ms( 24, skex,  ['expansion',
        ['from ', clnum, ' by expansion', nl, clause]	]).
ms( 25, rws0,  []).
ms( 26, rws2,  ['tautology', ['clause ', clnum, ' deleted by extensionality resolution']	]).

%------------------------------------------------------------------------------
%	messages
%------------------------------------------------------------------------------

message( Flag, Args ):-
	flag_on( ms, Flag ),
	ms( _, Flag, [_, LText] ),
	nl, treat_message( Args, LText ), nl, !.
message( _, _ ):- !.

treat_message( [C|Args], [clnum|LText] ):-
	number( C, N ),
	implicants(C,I),
	print_terms( [N] ),
	(I=[] -> true; 
	  write(' with implicants '),
	  \+ \+((antecedent(C,A),
		succedent(C,S),
		skolem([A,S,I],1,_),
		print_list_atoms(I)))
	 ),
	treat_message( Args, LText ), !.

treat_message( [Key|Args], [infkeytext|LText] ):-
	ms( _, Key, [Text, _] ),	
	print_terms( [Text] ),
	treat_message( Args, LText ), !.

treat_message( [R|Args], [runum|LText] ):-
	number( R, N ),
	print_terms( [N] ),
	treat_message( Args, LText ), !.

treat_message( [R|Args], [weightBound|LText] ):-
	weightBound( R, N ),
	print_terms( [N] ),
	treat_message( Args, LText ), !.

treat_message( [R|Args], [ruWeight|LText] ):-
	weight( R, N ),
	print_terms( [N] ),
	treat_message( Args, LText ), !.

treat_message( [R|Args], [ruDepth|LText] ):-
	history( R, H ),
	s_depth( H, N ),
	print_terms( [N] ),
	treat_message( Args, LText ), !.

treat_message( [RL|Args], [rulistnums|LText] ):-
	get_list_rule_nums( RL, NL ),
	print_terms( [NL] ),
	treat_message( Args, LText ), !.

treat_message( [C|Args], [clause|LText] ):-
	deBruijn(C1,C),
%	\+ \+ printSimpleClause( C1 ),
	print( C1 ),
	treat_message( Args, LText ), !.
treat_message( [C|Args], [clwconst|LText] ):-
	true( Top ),
	constraint( Top, C, C1 ),
	print( C1 ),
	treat_message( Args, LText ), !.

treat_message( [R|Args], [rule|LText] ):-
	deBruijn(R1,R),
%	\+ \+ printSimpleClause( R1 ),
	print( R1 ),
	treat_message( Args, LText ), !.

treat_message( [NL|Args], [numlist|LText] ):-
	print_terms( [NL] ),
	treat_message( Args, LText ), !.

treat_message( Args, [nl|R] ):-
	nl, treat_message( Args, R ), !.

treat_message( Args, [Text|R] ):-
	print_terms([Text] ),
	treat_message( Args, R ), !.

treat_message( [], [] ):- !.


get_list_rule_nums( [], [] ).
get_list_rule_nums( [R|RL], [N|NL] ):-
	number( R, N ),
	get_list_rule_nums( RL, NL ), !.

%------------------------------------------------------------------------------
%	Declaration of histories
%------------------------------------------------------------------------------

hi( init,[],
	' ~t[input~78|]').
hi( prop,[clause,natlist],
	' ~t[propositional reasoning from ~p and ~p~78|]').
hi( ccon, [clause],
	' ~t[condensement of ~p~78|]').
hi( vare, [clause, clause],
	' ~t[variable elimination from ~p~78|]').
hi( cdem, [clause, natlist],
	' ~t[reduction of ~p by ~p~78|]').
hi( chri, [rule, rule],
    	' ~t[chaining of ~p from ~p~78|]').
hi( sncr, [rule, rule],
    	' ~t[negative chaining of ~p from ~p~78|]').
hi( fact, [rule],
	' ~t[factoring on ~p~78|]').
hi( rere, [rule],
	' ~t[reflexivity resolution from ~p~78|]').
hi( orre, [rule, rule],
	' ~t[ordered resolution of ~p on ~p~78|]').
hi( trre, [rule, rule],
	' ~t[transitivity resolution of ~p on ~p~78|]').
hi( tore, [clause],
	' ~t[totality resolution from ~p~78|]').
hi( exre, [clause],
	' ~t[extensionality resolution from ~p~78|]').
hi( skex,   [clause, clause],
	' ~t[expansion from ~p~78|]').


unary_inference(fact).
unary_inference(rere).
unary_inference(tore).
unary_inference(exre).

binary_inference(chri).
binary_inference(sncr).
binary_inference(orre).
binary_inference(trre).

simplification_inference(ccon).
simplification_inference(vare).
simplification_inference(cdem).




% eof ---------------------------------------------------------------------------





