:- class(atom,object,
	 [preOrientation,
	  relationSymbol,
	  left,
	  right],
	 [createAtomFromTerm/2,
	  createAtom/4,


          % accessing
	  leftSide/2,
	  preOrientation/2,
	  rightSide/2,
	  relationSymbol/2,


	  chain/6]).



createAtomFromTerm(Term,A):-
	(functor(Term,RS,2),
	 arg(1,Term,L),
	 arg(2,Term,R),
	 special_relation_symbol(RS) ->
		createAtom(L,RS,R,A)
	;       createEquivalence(Term,'$true',A)).

	

createAtom(L,RS,R,A):- 
	(RS= = ->
	     createEquality(L,R,A)
	;RS= == ->
	     createEquivalence(L,R,E).
	;partial_congruence_relation_symbol(RS) ->
	     createPartialCongruence(L,RS,R,A)
	;partial_equivalence_relation_symbol(RS) ->
	     createPartialEquivalence(L,RS,R,A)
	;symmetric_relation_symbol(RS) ->
	     createSymmetricAtom(L,RS,R,A)
	;dense_ordering_relation_symbol(RS) ->
	     createDenseOrderInequation(L,RS,R,A)
	;total_ordering_relation_symbol(RS) ->
	     createTotalOrderInequation(L,RS,R,A)
	;partial_ordering_relation_symbol(RS) ->
	     createPartialOrderInequation(L,RS,R,A)
	;A=myself([n,RS,L,R])).



leftSide(O,L):-
	iv(O,left,L).

rightSide(O,R):-
	iv(O,right,R).

relationSymbol(O,RS):-
	iv(O,relationSymbol,RS).


preOrientation(O,P):-
	iv(O,preOrientation,P).


components(self(_,[P,RS,L,R]),[P,RS,L,R]).


/*
basicChain(+A1,+A2,+Pos,+Sub,?A,?Middle,?EqC,?OrdC)

is true when A is result of an inference by ordered chaining
between atoms A1 and A2 at subterm Sub which lies
at position Pos in the left side of A1. 
EqC is the associated equality constraint
and OrdC the ordering constraint. Satisfiability of both constraints
is checked individually but not in conjunction.
Middle is the middle term in the chain, then peak.
Pos is the position in the peak at which chaining was performed.
Only if A1 is a partial congruence, Pos can be different from [].
*/


basicChain(A1,A2,Sub,Pos,Chain,L2,EqC,OrdC):-
	A1=self(_,[P1,RS1,L1,R1]),
	A2=self(_,[P2,RS2,L2,R2]),
	side_conditions_ok_left(A1),
	side_conditions_ok_right(A2,Sub),
	chainOrderingConstraint(P1,L1,R1,P2,L2,R2,OrdC),
	\+ \+ (unify(R,U),
	       satisfiable(C)
	      ),
	composition_of_relations(RS1,RS2,RS),
	change_path_arg(Pos,L2,L1,L),
	createAtom(L,RS,R2,Chain).





premise(self(A,[P,RS,L,R]),chri([2,top]),[[self(A,[l,RS,Hole,R]),Hole],[1|Pos],L],OrdC):-
	Pos=[_|_],
	true(True),
	(P=l ->
	     OrdC=True
	;    and_gt(True,[L,R],OrdC)).
premise(self(A,[P,RS,L,R]),chri([2,inner]),[[self(A,[l,RS,LC,R]),Hole],[1|Pos],Sub],OrdC):-
	Pos=[_|_],
	change_path_arg(Pos,L,Sub,LC,Hole),
	true(True),
	(P=l ->
	     OrdC=True
	;    and_gt(True,[L,R],OrdC)).






side_conditions_ok_left(self(_,_)).

side_conditions_ok_right(self(_,_),_).


	
chain(A1,A2,A,M,Pos,EqC,OrdC):-
	basicChain(A1,A2,A,M,Pos,EqC,OrdC).


chainFromSymmetricAtom(self(O,A1),A2,A,M,Pos,EqC,OrdC):-
	basicChain(A2,self(O,A1),A,M,Pos,EqC,OrdC).
chainFromSymmetricAtom(self(O,A1),A2,A,M,Pos,EqC,OrdC):-
	inverse(A2,A2I),
	basicChain(A2I,self(O,A1),A,M,Pos,EqC,OrdC).




chainOrderingConstraint(P1,L1,R1,P2,L2,R2,C):-
	true(T),
	(P1=r -> C1=T; and_gt(T,[R1,L1],C1)),
	(P2=l -> C2=C1;and_gt(C1,[L2,R1],C)).


