%declarations%

:- module(auxiliary, [ 
	appendAtoms/3,
	concat312/3,
	concat33/3,
	delete32/3,
	delete_all/3,
	delete_one312/3,
	delete_repeats/2,
	disjoint_vars/2,
	get_prolog_data_operators/2,
	interval/3,
	less_or_equal/2,
	list_as_conjunction/2,
	list_as_disjunction/2,
	member212/2,
	member22/2,
	lookup_ml/2,
	partition/3,
	subset_vars/2
   ]).

:- use_module(library(basics), [
	member/2
   ]).


:- use_module(library(sets), [
	select/3
   ]).


%end%


%						AUX


%_________________________________________________________________________
%
%	AUXILIARY PREDICATES
%_________________________________________________________________________

% mostly old stuff to be deleted gradually


% lists(equality):member
member22(X,Y):-
	member(X,Y).

% lists(identity):member

member212(X,[Y|_]):- X == Y,!.
member212(X,[_,Y|_]):- X == Y,!.
member212(X,[_,_,Y|_]):- X == Y,!.
member212(X,[_,_,_,Y|_]):- X == Y,!.
member212(X,[_,_,_,_|L]):- member212(X,L).

% lists(equality)::delete
delete32(X,Y,Z):-
	select(X,Y,Z).

concat312( [],L,L):-!.
concat312( [X|L1],L2,[X|L3]):- concat312(L1,L2,L3),!.

concat33( [],L,L):-!.
concat33( [X|L1],L2,[X|L3]):- concat33(L1,L2,L3),!.

delete_all( _, [], [] ):- !.
delete_all( X, [Y|TL], L ):- X == Y,!, delete_all( X, TL, L ).
delete_all( X, [Y|TL], [Y|L]):- delete_all( X, TL, L ).

delete_one32(X,Y,Z):-
	select(X,Y,Z).

delete_one312( X, [Y|L], L ):- X == Y, !.
delete_one312( X, [Y|L], [Y|L1] ):- delete_one312( X, L, L1 ).

delete_repeats( [], [] ):- !.
delete_repeats( [X|R], R1 ):-
	member( X1, R ),
	X == X1, !,
	delete_repeats( R, R1 ).
delete_repeats( [X|R], [X|R1] ):-
	delete_repeats( R, R1 ), !.





/* lookup in dictionary of Key=Info entries.

   Keys are compared with ==, and hence may be variables.
   The Info items are logical.
   ``ml'' comes from ``meta-logical''
*/

lookup_ml( L=T, [X=T|_] ):- L == X,!.
lookup_ml( _, [[X,T]|R] ):-
	raise_exception('~n*** illegal call type lookup_ml [[~p,~p]|~p]~n'-[X,T,R]).
lookup_ml( X, [_|R] ):- lookup_ml( X,R ).







subset_vars(L1,L2):-
	\+ (member(V,L1), \+member212(V,L2)).

disjoint_vars(L1,L2):-
	\+ (member(V,L1), member212(V,L2)),
	\+ (member(V,L2), member212(V,L1)).

%------------------------------------------------------------------------------
% list_of_const(+T,?L) is true
% when L is the set of atomic constants in the term T, where
% list operators do not count, they are supposed to be meta-notation
% for disjunctions.
%------------------------------------------------------------------------------



list_of_const( T, L ):-
        list_const( T, L1/[] ),
        sort( L1, L ),!.

list_const( X, Xs/Xs ):- var(X), !.
list_const( [], Xs/Xs ).
list_const( X, [X|Xs]/Xs ):- atomic(X), !.
list_const( T, Xs/Ys ):-
	functor(T,_,N),
        list_const( N, T, Xs/Ys ).


list_const( 0,_,Xs/Xs ).
list_const( I,T,Xs/Zs):-
	arg(I,T,TI),
	list_const(TI,Xs/Ys),
	I1 is I-1,
        list_const(I1,T,Ys/Zs).


%------------------------------------------------------------------------------
% get_prolog_data_operators(+T,?L) is true
% when L is the set of operators in the Prolog term T, where
% list operators do not count, they are supposed to be meta-notation
% for disjunctions.
%------------------------------------------------------------------------------

get_prolog_data_operators( T, L2 ):-
	data_ops( T, L1/[] ),
	sort( L1, L2 ), !.

data_ops( T, Xs/Xs ):-
	var( T ), 
	!.
data_ops( [T|Ts], Xs/Ys ):-
	data_ops( T, Xs/Zs),
	data_ops( Ts, Zs/Ys ).
data_ops( [], Xs/Xs ).
data_ops( T, [Op/N|Xs]/Ys ):-
	functor(T,Op,N),
	ops( N,T,Xs/Ys ), 
	!.


ops(0,_,Xs/Xs).
ops(I,T,Xs/Zs):-
	arg(I,T,TI),
	data_ops(TI,Xs/Ys),
	I1 is I-1,
	ops(I1,T,Ys/Zs).




card(P,N):-
	PP=..[P,X],
	(setof(X,PP,L) ->
	     length(L,N)
	;
	     N=0
	),
	!.


less_or_equal(N,N):-
	N>0.
less_or_equal(I,N):-
	N>0,
	N1 is N-1,
	less_or_equal(I,N1).



appendAtoms(As,Sep,A):-
	% don't adapt this to new SICStus standards as otherwise
	% it doesn't work in Quintus 
	name(Sep,SN),
	appendAtomsN(As,SN,AN),
	name(A,AN).



appendAtomsN([X],_,XN):-
	name(X,XN).
appendAtomsN([X,Y|As],SN,N3):-
	appendAtomsN([Y|As],SN,N1),
	name(X,XN),
	user:append(SN,N1,N2),
	user:append(XN,N2,N3).


partition(S,S1,S2):-
	partition1(S,S,S1,S2).

partition1([],S,S1,S2):-
	partition2(S,S1,S2).
partition1([_|A],S,S1,S2):-
	(var(A) -> trace; partition1(A,S,S1,S2)).


partition2([X|S],[X|S1],S2):-
	partition2(S,S1,S2).
partition2([X|S],S1,[X|S2]):-
	partition2(S,S1,S2).
partition2([],[],[]).



	
interval( N, M, _ ):-
	M<N, !, fail.
interval( N, M, [N] ):-
	N == M, !.
interval( N, M, [N|L] ):-
	N1 is N+1,
	interval( N1, M, L ), !.


list_as_conjunction([],true).
list_as_conjunction([X,Y|Xs],(X,Conj)):-
	!,
	list_as_conjunction([Y|Xs],Conj).
list_as_conjunction([X],X).

list_as_disjunction([],false).
list_as_disjunction([X,Y|Xs],(X;Dis)):-
	!,
	list_as_disjunction([Y|Xs],Dis).
list_as_disjunction([X],X).

% eof --------------------------------------------------------------------



