%						CLAUSE_REDUNDANCY_PROVER
%declarations%

:- module(clause_redundancy_prover, [ 
	backward_simplification/3,
	condensement/3,
	delete_redundant_old_rules/0,
	demodulate_clause/4,
	extensionality_resolution/2,
	is_tautology/2,
	is_tautology/3,
	redundant_clause_by_case_analysis/1,
	redundant_clause_by_crw/1,
	simplify_by_crw/2,
	redundant_rule/1,
	remove_repetitions_and_identities/2,
	sequent_calculus_expansion/2,
	simplify_by_var_elim/2,
	tautology/1,
	tautology/3,
	totality_resolution/3,
	unshieldedVariable/3
   ]).

:- use_module(props, [
	anti_symmetric_relation_symbol/1,
	congruence_relation_symbol/1,
	crw_depth/1,
	crw_size_bound/1,
	flag_on/2,
	finite_domain/1,
	firstOrder/0,
	irreflexive_relation_symbol/1,
	noendL_relation_symbol/1,
	noendR_relation_symbol/1,
	horn/0,
        horn/1,
	predicate_with_extension/3,
	reflexive_relation_symbol/1,
	total_relation_symbol/1
   ]).

%:- use_module(user,_,[append/3]).


:- use_module(rewrite,[
        containsReducer/1,
        containsContextualReducer/1,
        containsSubsumer/1,
        normalform/6,
        normalform1/6,
	reducible/4,
	rewriteStrength/1,
	subsumes_literals/2
     ]).

:- use_module(equationalTheory, [
        currentTheory/1,
        mgu/2,
	equal/2,
	unify/2
     ]).

:- use_module(library(sets),[select/3,subtract/3]).

:- use_module(library(myordsets),[ord_add_element/3,ord_nonmember/2]).

:- use_module(mylists,[append/2]).

:- use_module(dynamicCode,[recordPrologClause/2]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(proofState, [
	clause/1,
	rule/2,
	deleteAllRules/0,
	proved_by_case_an/2,
	recordClause/1,
	recordIrreducibility/1,
	recordProofByCaseAnalysis/2,
	recordProofOfClause/1,
	recordSplitCase/3,
	splitCase/3
   ]).


:- use_module(constraintSolver, [
	simplify/2
   ]).

:- use_module(inference_computation, [
	builtin_inferences_with/2
   ]).


:- use_module(achelp, [
	ac_del_assoc/3,
	rreduce/3
   ]).

:- use_module(adt_atom, [
	atom_as_term/2,
	atom_for_demodulation/5,
	atoms_as_inequations/2,
	atom_as_reductive_substitution/1,
	atom_as_rule/3,
	atom_as_substitution/1,
	boolean_atom/1,
	create_atom/2,
	create_atom/4,
	equal_atoms/2,
	false_atom/1,
	false_literal/1,
	get_atom/4,
	get_atom_members/3,
	get_atom_relation_symbol/2,
	get_equation_members/3,
	get_atom_pre_rel_left_right/5,
%	instance_implies/4,
	inverse_atoms_if_possible/3,
	is_abbreviation/1,
	is_caseSplit/1,
	is_equational_atom/1,
	is_false_atom/1,
	is_true_literal/1,
	is_false_literal/1,
	is_propositional_literal/1,
	is_trivial_atom/1,
	lifted_inequation/2,
	match_atoms/2,
	match_atoms/5,
	match_atom_terms/5,
	orient_by_polarity/5,
	put_atom_preorientation/3,
	replacesubterm_ofatom_by_giving_/4,
	strongly_inverse_atom/2,
	weakly_inverse_atom/2,
	symmetric_polarity/3,
	symmetric_variants/4,
	true_atom/1,
	true_literal/1,
	var_on_predicate/2,
	variants_atoms/2
   ]).

:- use_module(classes,[
	and_cnf/3,
	antecedent/2,
	antecedent/3,
	is_atom/1,
	is_literal/1,
	bottom_cnf/1,
	change_rule/8,
	constraint/2,
	constraint/3,
	complexityBound/2,
	createClause/2,
	delete_rule/1,
	delete_rule/2,
	delete_rule/4,
	dont_simplify/1,
	empty/1,
	is_equation/1,
	history/2,
	history_and_number/3,
	implicants/2,
	indexEntries/2,
	is_old/1,
	is_trivial/1,
	kind/2,
	kind/3,
	literal_list_to_sequent/3,
	literals/2,
	literals/3,
	maximal_antecedent/2,
	maximal_succedent/2,
	negate/2,
	non_maximal_antecedent/2,
	non_maximal_succedent/2,
	number/2,
	or_cnf/3,
	pos_atom_as_literal/2,
	propositionalClause/1,
	reductive_for/3,
	redundancy_info/2,
	redundancy_info/3,
	simplifier/5,
	size/2,
	search_maximal_candidates/6,
	sequent_to_literals/3,
	succedent/2,
	succedent/3,
	theory_flag/2,
	to_clause/2
   ] ).

:- use_module(adt_declarations, [
	message/2
   ]).

:- use_module(adt_precedence, [
	greater_op/2,
	new_function_symbol_below/2,
	new_predicate_symbol_below/2,
	new_proposition_symbol/2
   ]).

:- use_module(adt_proof, [
	first_premise/2,
	second_premise/2,
	get_key_history/2

   ]).

:- use_module(clausal_rewriting, [
	redundant_by_clausal_rewriting/4
   ]).

:- use_module(equationalTheory, [
	match/2
   ]).

:- use_module(sequents, [
	sk/5,
	simplify_sequent/3,
	simple_sk/3
   ]).

:- use_module(adt_term, [
	get_term_vars/2,
	get_term_operators/2,
	instantiate/1,
	path_arg2/3,
	pi_constraints/4,
	pi_constraints_geq/4,
	ruleInstanceOf_in_/2,
	is_closed/1,
	term_size/2,
	variant/2
   ]).

:- use_module(myterms, [
	intersection_vars/3,
        identity_substitution/1,
	is_list_of_different_vars/1,
	skolemized/3
   ]).

:- use_module(auxiliary, [
	delete32/3,
	disjoint_vars/2,
	member212/2,
	member22/2,
	partition/3,
	subset_vars/2
   ]).

:- use_module(hgstuff, [
	setof0/3
   ]).


:- use_module(props, [
	composition_of_relations/3,
	dense_ordering_relation_symbol/1,
	equality_relation_symbol/1,
	reflexive_closure/2,
	irreflexive_relation_symbol/1,
	reflexive_relation_symbol/1,
	strict_part/2,
	symbol_admits_var_elim/1,
	symmetric_relation_symbol/1,
	total_ordering_relation_symbol/1,
	transitive_relation_symbol/1
   ]).

:- use_module(input, [
	permutativity_clause/5
   ]).

:- use_module(relational_theory,[
	subset_relation/6,
	union_relations/9
   ]).

:- use_module(match, [
	ac_equal/2,
	ac_match_pt/2
   ]).

:- use_module(orderConstraints, [
	and_eq/3,
	and_ge/3,
	and_gt/3,
	and_not_eq/3,
	false/1,
	implies/2,
	true/1
   ]).

:- use_module(termOrder, [
        greater/2,
        greaterEqual/2,
	greater_ground/2
   ]).

:- use_module(output, [
	print_terms/1
   ]).

:- use_module(myterms, [
	apply_substitution/3,
	list_of_vars/2
   ]).

:- use_module(adt_term, [
             free_vars/2
   ]).

:- use_module(library(arg), [
	genarg/3
   ]).

:- use_module(library(change_arg), [
	change_path_arg/4
   ]).

:- use_module(library(occurs), [
	contains_var/2
   ]).

:- use_module(library(maplist), [
	maplist/2,
	maplist/3,
	maplist/4
   ]).

:- use_module(global, [cont/2,store/2]).

:- use_module(thisProlog, [timeOut/3]).

:- use_module(timing, [
	push_timing/1,
	pop_timing/1
   ]).

:- use_module(index,[initial_index/2,initial_index_from/3,delete_terms/3]).

%end%

%_________________________________________________________________
%
%	DETECTION OF TAUTOLOGIES
%       new version for transitive relations
%
%_________________________________________________________________



tautology( C ):-
	is_trivial(C),
	!.
tautology( C ):-
	\+ is_equation(C), % this case is dealt with above
	is_tautology( C, _, yes ).


tautology( C, C1, A):-
	is_tautology(C,C1,A1),
	(A1==yes ->
	     message( taut, [C] )
	;
	     true
	),
	A1=A.


is_tautology( C, C, Ans ):-
	redundancy_info(C,(I,_)),
	(I==noTaut ->
	     Ans=no
	;I==tautology ->
	     Ans=yes),
	!.
is_tautology( C, C, Ans ):-
	is_trivial(C),
	!,
	Ans=yes.
is_tautology( C, C1, Ans ):-
	flag_on( rp, taut ),
	!,
	redundancy_info(C,(_,Implicants)),
	(is_tautology(C) ->
	     redundancy_info((tautology,Implicants),C,C1),
	     Ans=yes
	;
	     redundancy_info((noTaut,Implicants),C,C1),
	     Ans=no).
is_tautology(C,C,no).



% constraints in clauses are always fully simplified and nontrivial
%is_tautology( C ):-
%	constraint( C, Ct ),
%	simplify( Ct, Ct1 ),
%	false( Ct1 ).
is_tautology( C ):-
	antecedent( C, A ),
	succedent( C, S ),
	implicants(C,I),
	literal_list_to_sequent(I,A1,S1),
	append(S,S1,SI),
	append(A,A1,AI),
	is_tautology(AI,SI).

is_tautology(A,S):-
	is_simple_tautology(A,S),
	!.
is_tautology(A,S):-
        (member(P,A),\+boolean_atom(P) ->
%	     format("~ntaut? ~p -> ~p",[A,S]),
             is_tautology_special(A,S)
        ),
        !.

is_simple_tautology(_A,S):-
	member(P,S),
	is_trivial_atom(P),
	!.
is_simple_tautology(A,S):-
        member(P,A),
	(is_false_atom(P) ->
	     true
	;
	     member(Q,S),
	     equal_atoms(P,Q)
	),
	!.


is_tautology_special(A,S):-
        inverse_atoms_if_possible(S,SA,_SS),
        append(A,SA,AA),
        atoms_as_inequations( AA,I),
        ground_complete( I, [],[], R, Refl ),
        !,
        (       antecedentFalse(R,Refl)
        ;       succedentTrue(S,R,Refl)).




antecedentFalse(_,Refl):-
	member((RS,_),Refl),
	irreflexive_relation_symbol(RS),
	!.
antecedentFalse(R,Refl):-
	ground_convergent(R, Refl, (1,==,'$true$','$min')),
	!.

succedentTrue(S,R,Refl):-
	member( E, S ),
	get_atom_relation_symbol(E,RS),
	get_atom_members(E,U,V),
	ground_convergent(R, Refl, (1,RS,U,V)),
	!.




/*
ground_complete(+Eqs,+Rules,+Refl, ?RulesF, ?ReflF)
*/



ground_complete( [], R, Refl, R, Refl ):-
	!.
ground_complete( [(Pol,RS,U,V)|E], R, Refl, R1, Refl1 ):-
	ground_normal_form( R, RS, U, U1 ),
	ground_normal_form( R, RS, V, V1 ),

	(equal(U1,V1) ->
	      ground_complete(  E, R,[(RS,U1)| Refl], R1, Refl1 )
	;ground_convergent(R, Refl,(Pol,RS,U1,V1)) ->
	      ground_complete( E, R, Refl, R1, Refl1 )
	;     
	      orient_ground_eq_list_of_rules((Pol,RS,U1,V1),Rules),
	       % ground peaks with symmetric relations are simplifications
	      ground_simplify_with_eq_rules(Rules,R,Rules1,IrrR,RedR),

	       % ground peaks with non-symmetric relations are not
	      ground_peaks(Rules1, IrrR, E2),

	      append( E, RedR, E1),
	      append( E1, E2, E4 ),
	      append(Rules,IrrR,NewR),
	      ground_complete( E4, NewR, Refl, R1, Refl1 )
	),!.


ground_simplify_with_eq_rules([Ru1,Ru2],R,NonEqRules,IrrR,RedR):-
	ground_simplify_with_eq_rules([Ru1],R,Rules11,IrrR1,RedR1),
	ground_simplify_with_eq_rules([Ru2],IrrR1,Rules22,IrrR,RedR2),
	append(Rules11,Rules22,NonEqRules),
	append(RedR1,RedR2,RedR),
	!.
ground_simplify_with_eq_rules([(_,RS,S,T)],R,[],IrrR,RedR):-
	symmetric_relation_symbol(RS),
	ground_simplify_rules((1,RS,S,T),R,IrrR,RedR),
	!.
ground_simplify_with_eq_rules([Ru],R,[Ru],R,[]).




ground_simplify_rules( _, [], [], [] ):- !.
ground_simplify_rules(Ru , [(Pol,RS,I,D)|R], R1, [(Pol,RS,I1,D)|E1] ):-
	apply_ground_rule( Ru, RS, I, I1 ), !,
	ground_simplify_rules( Ru, R, R1, E1 ), !.
ground_simplify_rules(Ru , [(Pol,RS,I,D)|R], R1, [(Pol,RS,I,D1)|E1] ):-
	apply_ground_rule( Ru, RS, D, D1 ), !,
	ground_simplify_rules( Ru, R, R1, E1 ), !.
ground_simplify_rules(Ru , [Ru1|R], [Ru1|R1], E1 ):-
	ground_simplify_rules( Ru, R, R1, E1 ), !.



ground_convergent(_R, Refl, (_Pol,RS, U1, V1)):-
	equal(U1,V1),
	!,  % this cut is vital. if the following fails,
	    % there is no ground convergence
        (       reflexive_relation_symbol(RS)
	;	member212((RS,U1),Refl)),
	!.
ground_convergent([Rule|Rules], Refl, E):-
	orient_ground_eq( E,Ru),
	ground_narrow([Rule|Rules], Ru, E1),
	ground_convergent([Rule|Rules], Refl, E1).




ground_narrow(Rules, (Pol1,RS,U,V), (Pol1,CRS,U1,V) ):-
	member((Pol,RS1,S,T),Rules),
	\+ symmetric_variants(-Pol,RS1,Pol1,RS),
	ruleInstanceOf_in_((Sub,Pol,RS1,T),(U,Pol1,RS11,U1)),
	equal(Sub,S),
	(Pol1 = -1 ->
	      composition_of_relations( CRS, RS11, RS )
	;     
	      composition_of_relations( RS11, CRS, RS )
	),
	(congruence_relation_symbol(RS1) ->
	     !
	;
	     true
	).



ground_normal_form( R, RS, U, U2 ):-
	member22( Ru, R ),
	apply_ground_rule( Ru, RS, U, U1 ),
	ground_normal_form( R, RS, U1, U2 ),
	!.
ground_normal_form(_,_,T,T).

apply_ground_rule( Ru, RSU, U, NU ):-
	orient_ground_eq(Ru,(_,RS,I,D)),
	symmetric_relation_symbol(RS),
	ruleInstanceOf_in_((Sub,1,RS,D),(U,1,RS1,NU)),
	equal(Sub,I), 
	composition_of_relations(RS1,RSU,RSU),
	!.





orient_ground_eq( (Pol,RS,U,V), (Pol1,RS,V,U)):-
	greater_ground( V, U ), 
	!,
	Pol11 is -Pol,
	symmetric_polarity(Pol11,RS,Pol1).
orient_ground_eq( (Pol,RS,V,U), (Pol1,RS,V,U)):-
	greater_ground( V, U ), !,
	symmetric_polarity(Pol,RS,Pol1).
orient_ground_eq( (_Pol,_RS,U,V), _O ):-
%	trace,
%	orient_ground_eq( (_Pol,_RS,U,V), _O ),
	raise_exception(saturate_error("ground ordering not total: ~p  not orientable.",[(U,V)])).

orient_ground_eq_list_of_rules((Pol,RS,U,V),Rules):-
	(greater_ground( V, U ) ->
	     U1=V,
	     V1=U,
	     Pol1 is -Pol
	;
	     U1=U,
	     V1=V,
	     Pol1=Pol
	),
	(symmetric_relation_symbol(RS) ->
	     Rules = [(1,RS,U1,V1),(-1,RS,U1,V1)]
	;
	     Rules = [(Pol1,RS,U1,V1)]
	).



ground_peaks(Rules, R, E ):-
	setof0(Peak,Ru^(member(Ru,Rules),ground_peak_list(Ru,R,Peak)),E).


ground_peak_list(Ru, R, Peak ):-
	member(Ru1,R),
	(ground_peak(Ru,Ru1,Peak);
	ground_peak(Ru1,Ru,Peak)).


ground_peak((Pol1,RS1,L,R),(Pol2,RS2,I,D),(Pol2,RS3,R,D) ):-
	\+ symmetric_relation_symbol(RS1),
	\+ symmetric_relation_symbol(RS2),
	Pol1 is -Pol2,
	composition_of_relations(RS1,RS2,RS3),
	equal(I,L).



%____________________________________________________
%
%       remove repeated atoms in a clause
%       remove identities in the antecedent of a clause
%____________________________________________________

remove_repetitions_and_identities( C1, C2 ):-
        antecedent( C1, A1 ),
	lifted_inequations(A1,A11),
        succedent(  C1, S1 ),
	lifted_inequations(S1,S11),
	del_rep_and_id_eqs( A11, A2 ),
	del_rep_and_diff_eqs( S11, S2 ),
	(length(A1,NA),
	length(A2,NA),
	length(S1,NS),
	length(S2,NS) -> % no change
		  C2=C1
	;         antecedent( A2, C1, C3 ),
		  succedent(  S2, C3, C4 ), 
		  redundancy_info(_,C4,C2)
	),
	!.

del_rep_and_id_eqs( [], [] ).
del_rep_and_id_eqs( [E|R1], R2 ):-
	is_trivial_atom( E ),
	del_rep_and_id_eqs( R1, R2 ).
del_rep_and_id_eqs( [E|R1], [E|R2] ):-
	delete_all_eqs_equal_to( E, R1, R3 ),
	del_rep_and_id_eqs( R3, R2 ).

del_rep_and_diff_eqs( [], [] ).
del_rep_and_diff_eqs( [E|R1], R2 ):-
	is_false_atom( E ),
	del_rep_and_diff_eqs( R1, R2 ).
%del_rep_and_diff_eqs( [E|R1], R2 ):-
%	introduce_reflexive_closure(E,R1,R3),
%	del_rep_and_diff_eqs(R3,R2).
del_rep_and_diff_eqs( [E|R1], [E|R2] ):-
	delete_all_eqs_equal_to( E, R1, R3 ),
	del_rep_and_diff_eqs( R3, R2 ).



lifted_inequations([],[]).
lifted_inequations([E|Es],[L|Ls]):-
	lifted_inequation(E,L),
	lifted_inequations(Es,Ls).

irreducible(X):-
	var(X),
	!.
irreducible('$min').

delete_all_eqs_equal_to( _, [], [] ).
delete_all_eqs_equal_to( E1, [E2|R1], R2 ):-
	equal_atoms( E1, E2 ),
	delete_all_eqs_equal_to( E1, R1, R2 ).
delete_all_eqs_equal_to( E1, [E2|R1], [E2|R2] ):-
	delete_all_eqs_equal_to( E1, R1, R2 ).


%_________________________________________________________________
%
%	CONDENSEMENT
%_________________________________________________________________


condensement( C1, C, _Applicable ):-
	flag_on( rp, ccon ),
	literals(C1,[L|L1]),
	condensate( [L|L1], L2, CApplicable, What ),
	(var(CApplicable) ->
	    literals( L2, C1, C2 )
	;
	    C2=C1
	),
	antecedent(C2,A2),
	succedent(C2,S2),
	simplify_sequent((A2,S2),SApplicable,(A3,S3)),
	    
	(var(SApplicable) ->
	    antecedent(A3,C2,C3A),
	    succedent(S3,C3A,C3)
	;
	    var(CApplicable),
	    C3=C2
	),
	!,
	redundancy_info((noSimpleTaut,_),C3,C4),
	((nonvar(What);var(SApplicable)) ->
	     number( C1, N1 ),
	     history_and_number( [ccon, [N1],[C1]], C4, C ),
	     message( ccon, [C1,C] ),
	     recordProofOfClause(C)
	  	;
	     C=C4
	).
condensement( C, C, no ).


condensate( C1, C2, _Yes, What ):-
	one_condensate_step( C1, C3, What ),
	!,
	condensate( C3, C2, _, What ).
condensate( C, C, no, _nothing ).

/*

Condensement type I)
-----------------------

Suppose 

  C = (D or L or L')

such that

  a) |= (L\sigma -> L')
  b) (D or L')\sigma subset (D or L').

Then C is condensed (wrt. L) to C' = (D or L')\sigma.

Because of b) there exists n such that (D or L')\sigma^n subset (D or L')
and for any L'' in (D or L'), L''\sigma^{n+1}= L''\sigma^{n}

For a justification we prove that C |=| C'; C>C' is obvious.

b) implies that |= C' -> C, hence C' |= C

For the converse,
suppose C is true and there exists \tau such that
(D or L')\sigma^{n+1}\tau is false.

As C\sigma^{n+1}\tau is true, L\sigma^{n+1}\tau must be true.
Now a) implies that L'\sigma^n\tau is true.

As L'\sigma^{n+1} = L'\sigma^n,
L'\sigma^{n+1}\tau is true, hence C'\sigma^{n+1}\tau is true
which is a contradiction.

Therefore (D or L')\sigma^{n+1}\tau, and hence, (D or L')\sigma\tau
is true.


Condensement type II)
-----------------------

Condensement also applies orientable equations.


Condensement type III)
-----------------------

If a literal or a disjunction of two literals is equivalent
to true, the clause is reduced to true (-> exception)




*/


one_condensate_step( C1, C2, What ):-
	select( L, C1, C3 ),
	one_condensate_step_with(L,C3,C2,What).


% condensate (L or C1) to C2 with respect to L
one_condensate_step_with(L,_C1,_C2,_trivial):-
	is_true_literal(L),   
	raise_exception(reduced_to_true([])).
one_condensate_step_with(L,C1,C2,_trivial):-
	instance_implies_false(L,C1,C2).   % L'==false
one_condensate_step_with(L,C1,[L|C2],nontrivial):-
	reduceIfEq(L,C1,C2).
one_condensate_step_with(L,C1,C2,nontrivial):-
	reduceIfSubst(L,C1,C2).
one_condensate_step_with(L,C1,C2,nontrivial):-
	select( LPrime, C1, C3 ),
	matchLiteralOrConverse(L,LPrime,Sigma,Orientation),
	(
	     instance_implies(L,Sigma,Orientation,LPrime,C1,C2)
	;
	     identity_substitution(Sigma),
	     union_relations(L,LPrime,Orientation,L1),
	     C2=[L1|C3]
	),
	!.

reduceIfSubst([-|E],C,C):-
	atom_as_reductive_substitution(E).
%	atom_as_substitution(E). % seems to perform worse, is ``less complete''
                                 % anyway.



reduceIfEq([-|E],C,C1):-
	is_equational_atom(E),
	get_atom_members(E,L,R),

	(greaterEqual(L,R) ->
	     apply_substitution( [L=R], C,C1)
	;greater(R,L) ->
	     apply_substitution( [R=L], C,C1)
	),
	C\==C1.



instance_implies_false([+|E],C,C):-
	is_false_atom(E).
instance_implies_false([P|E],C,CSigma):-
	get_atom(E,RS,L,R),
	(P= - ->
	     reflexive_relation_symbol(RS)
	;
	     irreflexive_relation_symbol(RS)
	),
	mgu(L=R,Sigma),
	apply_substitution(Sigma,C,CSigma),
	is_contained_in(CSigma,C).

matchLiteralOrConverse([_|E],[_|EPrime],Sigma,Orientation):-
	match_atom_terms(E,EPrime,Orientation,[],Sigma),
	!.

	
instance_implies([P|E],Sigma,Orientation,[P1|E1],C,C1):-
	get_atom_relation_symbol(E,RS),
	get_atom_relation_symbol(E1,RS1),
	(
	    subset_relation(P,RS,+,P1,RS1,Orientation)
	;
	    subset_relation(P,RS,Orientation,P1,RS1,+)
	),
	!,
	apply_substitution( Sigma, C, C1 ),
	is_contained_in( C1, C ).



union_relations([P|E],[P1|E1],O,[P2|E2]):-
	get_atom(E,RS,L,R),
	get_atom_relation_symbol(E1,RS1),
	(  union_relations(P,RS,+,P1,RS1,O,P2,RS2,O2)
	;
	   union_relations(P,RS,O,P1,RS1,+,P2,RS2,O2)
	),
	!,
	(RS2='$all' ->
	     true_atom(E2)
	;
	     (O2= + ->
		  T=..[RS2,L,R]
	     ;
	          T=..[RS2,R,L]
	     ),
	     create_atom(T,E2)
	).



trans_match( [], [] ).
trans_match( [[V,T]|R1], [V=T|R2] ):-
	trans_match( R1, R2 ).

is_contained_in( [], _ ).
is_contained_in( [[P|E1]|R], S ):-
	member22( [P|E2], S ),
	equal_atoms( E1, E2 ),
	!,
	is_contained_in( R, S ).

%_________________________________________________________________
%
%	CLAUSE REDUCTION AND SUBSUMPTION
%_________________________________________________________________


% version for when subsumption is not integrated into rewriting

demodulate_clause( C1, C2,_Applicable, ST):-
	\+ rewriteStrength(max),
	literals(C1,LV),
	constraint(C1,Ct),
	skolemized((Ct,LV),(CtS,Lits),Subst),
	(flag_on( rp, cdem ) ->
	     demodulate_lits( Lits, LitsR,[],CtS,RN/RN1,ST)
	;
	     LitsR=Lits,
	     RN=RN1
	),

	
	(flag_on(rp,clsb), \+LitsR=[_] ->
	     normalform(clause,ct(LitsR,CtS),LitsR,LitsRR,RN1/[],ST)
	;
	     RN1=[],
	     LitsRR=LitsR
	),


	(LitsRR=[T],is_true_literal(T) ->
	     raise_exception(reduced_to_true(RN))
	;
	     RN\== [],   % one demodulation step performed?
	     skolemized(LRV,LitsRR,Subst),
	     literals(LRV,C1,C4),
	     redundancy_info((_,[]),C4,C5),   % old stuff
	     number( C1, N1 ),
	     history_and_number( [cdem, [N1,RN], [C1]], C5, C21 ),
	     message( cdem, [C1,RN,C21] ),
%	     /*
	     (RN1\== [] -> 
		  % subsumtion reduction may induce normal reducibility
		  demodulate_clause( C21, C2,Applicable, ST),
		  (nonvar(Applicable) ->
		       % no more reductions
		       C2=C21
		  ;
		       recordProofOfClause(C21)

		  )
	     ;
	          C21=C2

	     )
%	     */
%	     C21=C2
	),
	!.
%  version for when subsumption resolution is included in rewriting

demodulate_clause( C1, C2,_Applicable, ST):-
	rewriteStrength(max),
	literals(C1,LV),
	constraint(C1,Ct),
	skolemized((Ct,LV),(CtS,Lits),Subst),
	(flag_on( rp, cdem ) ->
	     demodulate_lits( Lits, LitsRR,[],CtS,RN/[],ST)
	;
	     LitsRR=Lits,
	     RN=[]
	),

	
	(LitsRR=[T],is_true_literal(T) ->
	     raise_exception(reduced_to_true(RN))
	;
	     RN\== [],   % one demodulation step performed?
	     skolemized(LRV,LitsRR,Subst),
	     literals(LRV,C1,C4),
	     redundancy_info((_,[]),C4,C5),   % old stuff
	     number( C1, N1 ),
	     history_and_number( [cdem, [N1,RN], [C1]], C5, C2 ),
	     message( cdem, [C1,RN,C2] ),
		       true
	),
	!.
demodulate_clause( C, C, no, _ST ).



/*
demodulate_lits(+Ls1, ?Res, +Irred, +Ct, ?N1/N2, +ST )
is true when the list of literals Res is the result of normalizing
the literals in Ls1 and in Irred, where the latter literals are irreducible.
Ct is the constraint of the clause from which the literals come from,
N1/N2 is the increment in rule numbers used for simplification;
ST is the Simplifier table.
*/

demodulate_lits([L1|Ls1], Res, Irred, Ct, N1/N2, ST ):-
	append(Ls1,Irred,Cl),
	(normalform1(lit, ct(Cl,Ct) , L1, L2, N1/N3, ST ) ->
	     (is_true_literal(L2) ->
	          N3=[],
		  raise_exception(reduced_to_true(N1))
	     ;    
	          (false_literal(L2) ->
		       Irred1=[]
		  ;	  
		       Irred1=[L2]
		  ),
	      % we have reduced something, so Irred may become reducible
	          append(Ls1,Irred,Reducible)
	     )
	;
	     Irred1=[L1|Irred],
	     Reducible=Ls1,
	     N1=N3
	),
	demodulate_lits( Reducible, Res, Irred1, Ct, N3/N2, ST ).
demodulate_lits([], Irred, Irred, _, N/N, _ ).



%________________________________________________________________
%
%       CLAUSE SIMPLIFICATION BY TOTALITY RESOLUTION
%
%________________________________________________________________

totality_resolution(C,C2,Applicable):-
	flag_on(rp, tore),
	totality_resolution0(C,C2,Applicable),
	!.
totality_resolution(C,C,no).

totality_resolution0(C,C1,_Yes):-
	one_step_totality_resolution(C,C2),
	totality_resolution0(C2,C11,_),
	number( C, N ),
	history_and_number( [tore, [N], [C]], C11, C1 ),
	message(tore, [C,C1]).
totality_resolution0(C,C,no).

one_step_totality_resolution(C,C1):-
	antecedent(C,A),
	append(A1,[E|A2],A),
	get_atom_relation_symbol(E,RS),
	(total_relation_symbol(RS) ->
	     get_atom_members(E,L,R),
	     \+equal(L,R),
	     succedent(C,S),
	     append(A1,A2,A3),
	     reflexive_closure(RS,RR),
	     P=..[RR,R,L]
	;total_relation_symbol(RSS),reflexive_closure(RSS,RS) ->
	     get_atom_members(E,L,R),
	     succedent(C,S),
	     append(A1,A2,A3),
	     P=..[RSS,R,L]
	),

	create_atom(P,E1),
	antecedent(A3,C,C2),
	succedent([E1|S],C2,C1),
	!.





%________________________________________________________________
%
%       CLAUSE SIMPLIFICATION BY EXTENSIONALITY RESOLUTION
%
%________________________________________________________________

extensionality_resolution(C1,C2):-
	(try_ext_res ->
	     (extensionality_resolution0(C1,Cs) -> 
		  (  Cs=[C2],
		     literals(C1,L1),
		     literals(C2,L2),
		     variant(L1,L2) ->
				 true   % in that case only metalogical
				        % information about the kind of C1
                                        % is changed
		  ;
	                         member(C,Cs),
				 antecedent(C,A),
				 succedent(C,S),
				 createClause([A,S],CC),
				 number( C1, N ),
				 history_and_number( [exre, [N], [C1]], CC, C2 ),
				 message(exre, [C1,C2])
		  )
	     ;
	          message(rws2, [C]),
		  fail
	     )
	          
	;
	     C1=C2
	).



try_ext_res:-
	flag_on(rp, exre),
	predicate_with_extension(_,_,_),
	!.

/*
   extensionality_resolution0(+C,?Cs)

is true when Cs is a nonempty
list of clauses which can be obtained by extensionality
resolution from C.
If it is false then C is a clause that comes from an extensionality
axiom of the form

  ~P(x) == exist y. Q(x,y)

in which the main predicate P occurs positively.
Such clauses are redundant once all negative occurrences
(outside extensionality axioms) of P  have been eliminated
by extensionality resolution. 
Failure of (the calling) extensionality_resolution leads to a 
deletion of C.

When C is a clause that comes from an extensionality axiom
with P occurring negatively, then Cs=[C1], with C1 a variant of
C in which the kind information is set to ``defClause(P/N)''.

*/

extensionality_resolution0(C,Cs):-
	findall(C2,'extensionality_resolution!'(C,C2),Cs),
	Cs\==[].

'extensionality_resolution!'(C,C):-
	kind(C,K),
	builtin_inferences_with(K,_),
	!.
'extensionality_resolution!'(C,C1):-
	predicate_with_extension(P/N,_SF,[_PArgs,Def,_NegatedDef]),
	nonvar(Def),
	literals(C,Ls),
        select_general_p_literal(P/N,-,Ls,Ls1),
	member(D,Def),
	subsumes_literalsV(D,Ls1),
	kind(defClause(P/N),C,C1),
	!.
'extensionality_resolution!'(C,_):-
	kind(C,K),
	extensionality_axiom(K,P/N),
	predicate_with_extension(P/N,_,_),
	!,
	fail.  
'extensionality_resolution!'(C,C1):-
	predicate_with_extension(P/N,_SF,[PArgs,_Def,NegatedDef]),
	antecedent(C,A),
	select(E,A,A1),
	atom_as_term(E,T),
	functor(T,P,N),
	!,
%	(P= (=) -> trace; true),
	T=..[P|PArgs],
	member(NDef,NegatedDef),
	literal_list_to_sequent(NDef,A2,S2),
	append(A1,A2,A3),
	succedent(C,S),
	append(S,S2,S3),
	antecedent(A3,C,C2),
	succedent(S3,C2,C3),
	'extensionality_resolution!'(C3,C1).
'extensionality_resolution!'(C,C).

extensionality_axiom(extensionality(P/N,_Args,_SFs),P/N).
extensionality_axiom(anti_symmetry(_R,P),P/2).

subsumes_literalsV(Ls,Ls1):-
        \+ \+ ( numbervars(Ls1,0,_),
                subsumes_literals(Ls,Ls1)
              ).


select_general_p_literal(P/N,Pol,Cl,Cl1):-
        select([Pol|A],Cl,Cl1),
        (   get_atom(A,==,E,'$true$'),
            E=..[P|Args],
            length(Args,N)
        ;
            N=2,
            get_atom(A,P,L,R),
            Args=[L,R]
        ),
        is_list_of_different_vars(Args).

select_p_literal(P/N,Pol,Args,Cl,Cl1):-
        select([Pol|A],Cl,Cl1),
        (   get_atom(A,==,E,'$true$'),
            functor(E,P,N),
            E=..[P|Args]
        ;
            N=2,
            get_atom(A,P,L,R),
            Args=[L,R]
        ).



%_________________________________________________________________
%
%	REDUNDANCY OF RULES
%_________________________________________________________________

delete_redundant_old_rules:-
	push_timing(rsim),
	delete_redundant_rules_by_case_an,
	pop_timing(rsim).



redundant_rule(R):-
	push_timing(rsim),
	number(R,N),
	(user:spy_clause(N)-> trace; true),
	(redundant_rule_by_case_analysis(R) ->
	     pop_timing(rsim)
	;
	     pop_timing(rsim),
	     fail
	).



%_________________________________________________________________
%
%	backward simplification
%_________________________________________________________________

backward_simplification(Ns, ST, STO):-
	store(new_rules,[]),
%	(member(67,Ns) -> trace; true),
	backward_simplification(Ns,_Any,ST,ST1),
	backward_simplification1(Ns,ST1,STO).


/* ----------------------------------------------------------------------------

  backward_simplification(+Ns,?Include,+ST,?STO)

is true when STO is the index resulting from ST by deleting
all rules in Include that can be reduced by the rules in Ns.
If Include is a variable, all current rules are addressed.
Reducible rules are deleted from the proofState as a side-effect.

---------------------------------------------------------------------------- */
/* treat units specifically
backward_simplification(Ns,Include,ST,STO):-
	select(N,Ns,),
	rule(N,R),
	literals(R,[P|A]),
	is_standard_atom(A),
	A=[_,==,At,'$true$'],
	!,
	
	
	
	
*/
backward_simplification(Ns,Include,ST,STO):-
	initial_index_from(Ns,STRu,Entries),
	(flag_on( rp, rdem ),containsReducer(Entries) ->
	     reduce_rules(Ns,Include,Ns,STRu,ST,ST1)
	;\+ rewriteStrength(min),
	 flag_on( rp, rusb ),
	 containsContextualReducer(Entries) ->
	     reduce_rules(Ns,Include,Ns,STRu,ST,ST1)
	;
	     ST1=ST
	),
	( \+ rewriteStrength(max),
	  flag_on( rp, rusb ),
	  containsSubsumer(Entries) ->
	     subsume_rules(Ns,Include,Ns,STRu,ST1,STO)
	;
	     ST1=STO
	).


backward_simplification1(Ns,ST,STO):-
	backward_simplification1([],Ns,ST,STO).

backward_simplification1(Ms,[N|Ns],ST,STO):-
	append(Ms,Ns,Others),
	backward_simplification([N],Others,ST,ST2),
	backward_simplification1([N|Ms],Ns,ST2,STO).
backward_simplification1(_,[],ST,ST).


/*
backward_simplification1([N|Ns],ST,STO):-
%	backward_simplification(Ns,[N],ST,ST1),
	backward_simplification([N],Ns,ST,ST2),
	backward_simplification1(Ns,ST2,STO).
backward_simplification1([],ST,ST).
*/



reduce_rules(Ns,Include,Exclude,STRu,ST,STO):-
	(var(Include) ->
	     true
	;
	     member(N,Include)
	),
	rule( N, R2 ),
	ord_nonmember(N,Exclude),
	\+ dont_simplify(R2),
	\+ (flag_on(rp,prop),propositionalClause(R2)),

	(larger_side_irreducible(R2,N,STRu,ST,ST1,Result) ->
	     (Result=nonMaximalReduced ->
	          ord_add_element(Exclude,N,Ex1),
		  !,
	          reduce_rules(Ns,Include,Ex1,STRu, ST1, STO)
	     ;Result=reduced_to_true->
	          message(rdem,[R2,Ns]),
	          delete_rule(N,R2,ST,ST1),
		  !,
		  reduce_rules(Ns,Include,Exclude,STRu, ST1, STO)
	     )
	;
	     to_clause( R2, C ),
	     recordClause(C),
	     (empty(C) -> raise_exception(proof_found); true), 
	     message(rdem,[R2,Ns]),
	     delete_rule(N,R2,ST,ST1),
	     !,
	     reduce_rules(Ns,Include,Exclude,STRu, ST1, STO)
	).
reduce_rules( _, _, _, _, ST, ST ).

larger_side_irreducible(Ru,N,STRu,ST,STO,Res):-
        on_exception(reduced_to_true(_),
		     larger_side_irreducible1(Ru,N,STRu,ST,STO,Res),
		     Res=reduced_to_true).



larger_side_irreducible1(Ru,N,STRu,ST,STO,Res):-
	reductive_for(Ru,[P|E],MinorLs),
	MinorLs=[],
%	true(Top),
	constraint(Ru,Ct),
	get_atom_relation_symbol(E,Rel),
	(congruence_relation_symbol(Rel) ->
	     atom_as_rule(E,ER,Pol),
	     get_atom_pre_rel_left_right(ER,l,_,L,R),
	     L1=L,
	     R1=R,
	     Ty=term
	;
	     fail,
	     L1=[P|E],
	     R1='$true$',
	     Ty=lit
	),
	!,
	(   numbervars(L1,0,_),
	    rewrite: reducible(Ty,ct([],Ct),L1,STRu)
	->
	    fail
	;
	    skolemized((L1,R1,MinorLs),(_L1S,R1S,MinorLsS),Subst),
	    (normalformH(R1S,Ty,MinorLsS,R11S,Minor1S,STRu,ST) ->
		  skolemized((R11,Minor1),(R11S,Minor1S),Subst),
		  (Ty=term ->
		       orient_by_polarity(Pol,L1,R11,L2,R2),
		       create_atom(L2,Rel,R2,E1),
		       put_atom_preorientation(l,E1,E2)
		  ;
		       E2=E
		  ),
%		  format('~NRule ~p: reduced ~p to ~p',
%			 [N,(R1S,MinorLsS),(R11S,Minor1S)]),
	          literal_list_to_sequent(Minor1,Ap,Sp),
		  (P= + ->
		       Am=[],Sm=[E2]
		  ;
		       Am=[E2],Sm=[]
		  ),
	          change_rule(N,Ru,Am,Ap,Sm,Sp,ST,STO),
		  Res=nonMaximalReduced
	     ;
	          ST=STO,
		  Res=irreducible
	     )
	),
	!.
larger_side_irreducible1(Ru,_,STRu,ST,ST,irreducible):-
%	true(Top),
	constraint(Ru,Ct), 
	literals(Ru,Ls),
	\+ (      numbervars(Ls,0,_),
		  select(L,Ls,LsL),
		  rewrite: reducible(lit,ct(LsL,Ct),L,STRu)
	   ).


normalformH(T,Ty,Ls,T1,Ls1,STRu,ST):-
	true(Top),
	normalform1(Ty,ct([],Top),T,TI, _,STRu),
	!,
	normalform( Ty,ct([],Top),TI,T1,_,ST),
	demodulate_lits(Ls,LsI,[],Top,X/Y,STRu),
	(X\==Y -> 
	     demodulate_lits(LsI,Ls1,[],Top,_,ST)
	;
	     Ls1=LsI
	).
normalformH(T,_,Ls,T,Ls1,STRu,ST):-
	true(Top),
	demodulate_lits(Ls,LsI,[],Top,X/Y,STRu),
	(X\==Y -> 
	     demodulate_lits(LsI,Ls1,[],Top,_,ST)
	).


subsume_rules(Ns,Include,Exclude,STRu,ST,STO):-
	(var(Include) ->
	     true
	;
	     member(RN,Include)
	),
	rule( RN, R2 ),
	ord_nonmember(RN,Exclude),
	\+ dont_simplify(R2),
	\+ (flag_on(rp,prop),propositionalClause(R2)),
	literals(R2,Ls),
	constraint(R2,Ct),
	\+ \+ (      numbervars((Ls,Ct),0,_),
		     rewrite: reducible(clause,ct(Ls,Ct),Ls,STRu)
	      ),
	to_clause( R2, C ),
	recordClause(C),
	message(rdem,[R2,Ns]),
	delete_rule(RN,R2,ST,ST1),
	subsume_rules(Ns,Include,Ns,STRu, ST1, STO).
subsume_rules( _, _, _, _, ST, ST ).




subsume_by_(RN,N):-
	rule(N,Ru),
	indexEntries(Ru,Entries),
	Entries\==[],
	initial_index(Entries,STRu),
	rule( RN, R2 ),
	\+ dont_simplify(R2),
	\+ (flag_on(rp,prop),propositionalClause(R2)),
	size(Ru,S),
	size(R2,S2),
	S2>=S,
	literals(R2,Ls),
	constraint(R2,Ct),
	\+ \+ (      numbervars((Ls,Ct),0,_),
		     rewrite: reducible(clause,ct(Ls,Ct),Ls,STRu)
	      ).



%_________________________________________________________________
%
%	CONSTRAINED REWRITING FOR ATOMAL CLAUSES
%_________________________________________________________________

redundant_clause_by_case_analysis( Cl ):-
	flag_on( rp, clca ),
	\+currentTheory(ac),
	is_atom( Cl ),
	succedent( Cl, [E] ),
	constraint(Cl, C),
	already_done( [E,C] ), !,
	message( clca, [Cl] ), !.
redundant_clause_by_case_analysis( Cl ):-
	flag_on( rp, clca ),
	\+currentTheory(ac),
	is_atom( Cl ),
	succedent( Cl, [E] ),
	constraint( Cl, C ), 
 	free_vars( E, LV ),
	get_equation_members( E, S, T ),
	\+permutativity_clause([],[S=T],_,_,_),
	case_analysis( LV, _, [[S=T,C]] ),
	recordProofByCaseAnalysis( E,C ),
	message( clca, [Cl] ).



already_done( [E1,C1] ):-
	proved_by_case_an( E2,C2 ),
	variants_atoms( E1, E2 ), 
	(     match(E1,E2),
	      implies(C2,C1)
	;     E1=(S=T),
	      match(T=S,E2),
	      implies(C2,C1)),
	!.

%_________________________________________________________________
%
% Case analysis is currently defined so that up to 3 vars it runs 
% Martin & Nipkow and constrained rewriting for more than 3 vars.
%_________________________________________________________________
case_analysis( LV, N, Eq ):-
	term_size(Eq,S),
	crw_size_bound(B),
	(S=<B ->
	     (flag_on(ms,clca),nonvar(N) ->
		  format('case analysis for rule ~p',N),
		  ttyflush
	     ;flag_on(ms,clca) ->
		  format('case analysis for clause ~p',Eq),
		  ttyflush
	     ;true),
	     (timeOut(case_analysis1(LV,N,Eq),30000,A) ->
		  (A=success ->
		       wr(' done.~n')
		  ;    wr(' time out.~n'),
		       fail
		  )
	     ;    wr(' no.~n'),
	          fail
	     )
	).

wr(S):-
	(flag_on(ms,clca) ->
	     format(S,[])
	;true).



case_analysis1( [], 	N, Eq ):-	!, martin_nipkow( [], 	   N, Eq ), !.
case_analysis1( [A], 	N, Eq ):-	!, martin_nipkow( [A],     N, Eq ), !.
case_analysis1( [A,B],	N, Eq ):-	!, martin_nipkow( [A,B],   N, Eq ), !.
case_analysis1( [A,B,C], N, Eq ):-	!, martin_nipkow( [A,B,C], N, Eq ), !.
%case_analysis1( [A,B,C,D], N, Eq ):-	!, martin_nipkow( [A,B,C,D], N, Eq ), !.
%case_analysis1( [A,B,C,D,E], N, Eq ):-	!, martin_nipkow( [A,B,C,D,E], N, Eq ), !.
case_analysis1(LV,N,Eq):-
	crw_depth(D),
	(flag_on(ms,clca),nonvar(N) ->
	    format('Constrained rewriting on ~p~n',[N])
	;true),
	constrained_rewriting(D,LV,N,Eq).
%	timeOut(constrained_rewriting(D,LV,N,Eq),60000,success).

%_________________________________________________________________





constrained_rewriting(  _, _, _,  [] ):- !.
constrained_rewriting(  Calls, LV, N, [[E,C]|Rest] ):-
	\+(\+( eq_taut( E, C ))), !, 
%	mess1_crw( LV,E,C ),
	constrained_rewriting( Calls, LV, N, Rest ), !.
constrained_rewriting( Calls, LV, N, [[E,C]|Rest] ):-
	one_step_selected_by_constraint(N,C,E,NE,_),
	!,
%	mess3_crw( LV,E,C,NE ),
	constrained_rewriting( Calls, LV, N, [ [NE,C]|Rest] ), !.
constrained_rewriting(  0, _, _,  _ ):- 
	!,
	fail.
constrained_rewriting( Calls, LV, N, [[E,C]|Rest] ):-
	one_step_crw(N,E,[C,[],[]],NE,[NC,CCs,ECs]),
	!, 
        Calls1 is Calls-1,
%	mess4_crw( LV,E,C,NE,NC ),
	constrained_rewriting( Calls1, LV, N, [ [NE,NC]|Rest ] ), !,
	% treating the complementary clauses
%	mess8_crw(LV,E,C,ECs),
	eq_complements(Calls,LV,N,[E,C],ECs),
%	mess9_crw(LV,E,C,CCs),
	ineq_complements(Calls,LV,N,[E,C],CCs),
	!.


eq_taut([+,_,=,T1,T2],C):-
	eq_taut(T1=T2,C).


eq_taut( T1=T2, _ ):- 
	ac_equal(T1,T2),!.
eq_taut( _, C ):-
	simplify( C, C1 ),
	false( C1 ),!.
eq_taut( T1=T2, _ ):- 
	\+ unify( T1, T2 ), !, fail.
eq_taut( T1=T2, C ):-
	mgu( [T1=T2], U ),
	constraint_implies_unifier( C, U ),!.


eq_complements(Calls,LV,N,[E,C],[L=R|EE]):-
	\+ \+ ( unify(L,R) ->
		    constrained_rewriting(Calls,LV,N,[[E,C]])
	      ;     true),
	!,
        eq_complements(Calls,LV,N,[E,C],EE).
eq_complements(_,_,_,_,[]).

ineq_complements(Calls,LV,N,[E,C],[[U,V]|EE]):-
	and_gt( C, [V,U], CC ),
        constrained_rewriting( Calls, LV, N, [ [E,CC] ] ),
	!,
        ineq_complements(Calls,LV,N,[E,C],EE).
ineq_complements(_,_,_,_,[]).





one_step_crw(N,E,Ci,NE,Co):-
	one_step_geq(N,E,Ci,NE,Co,_).
one_step_crw(N,E,Ci,NE,Co):-
	crw_rule(N,PreO,RS,L,R),
	ruleInstanceOf_in_((Sub,1,RS,R),(E,1,==,NE)),
	nonvar( Sub ),
	ac_match_pt( L, Sub),
	pi_constraints_geq([Sub],[L],Ci,C1),

	((PreO=l; greater(L,R)) ->
	  Co=C1
	; pi_constraints([L],[R],C1,Co)
	).
one_step_crw(N,E,Ci,NE,Co):-
	crw_rule(N,PreO,RS,L1,R1),
	ruleInstanceOf_in_((Sub,1,RS,R),(E,1,==,NE)),
	nonvar( Sub ),
	extended_rule(Sub,L1,R1,L,R),
	ac_match_pt( L, Sub),
	pi_constraints_geq([Sub],[L],Ci,C1),

	((PreO=l; greater(L,R)) ->
	  Co=C1
	; pi_constraints([L],[R],C1,Co)
	).

extended_rule(S,L,R,L1,R1):-
	S=..[F|A],
	nonvar(L),
	L=..[F|B],
	ac_del_assoc(F,A,AL),
	ac_del_assoc(F,B,BL),
	length(AL,N),
	length(BL,M),
	D is N-M,
	extension(F,D,D,L,R,L1,R1).


extension(F,I,D,L,R,L1,R1):-
	I=<D,
	length(E,I),
	append(E,[L],A),
	append(E,[R],B),
	rreduce(F,A,L1),
	rreduce(F,B,R1).
	



one_step_geq(N,E,Ci,NE,Co,M):-
        one_step_crw1(N,E,NE,_,[L,R],M),
	pi_constraints_geq([L],[R],Ci,Co),
	Co==Ci.  % in this case Ci implies L>=R


crw_rule(N,PreO,RS,L,R):-
	rule(N1,Ru),
	is_atom(Ru),
	N\==N1,
	\+ kind(Ru,commutativity(_,_)),
	\+ kind(Ru,associativity(_,_)),
	\+ kind(Ru,permutativity(_,_,_)),
	succedent(Ru,[E1]),
	atom_for_demodulation( E1,PreO, RS, L, R).

% we prefer selecting a rewrite step L->R that is ordered,
% that is, C implies L>R.
one_step_selected_by_constraint(N,C,E,NE,M):-
	\+ true(C),
	one_step_crw1(N,E,NE,Op,[L,R],M),
	( Op \== n
	; greater( L, R )
	; and_ge(C,[R,L],C1),
	  simplify(C1,CS),
	  false(CS)
	; true(T),
	  \+ \+ eq_taut(NE,T)
	).


one_step_crw1(N,E,NE,PreO,[L,R],N1):-
	rule(N1,Ru),
	is_atom(Ru),
	N\==N1,
%	\+ kind(Ru,commutativity(_,_)),
%	\+ kind(Ru,permutativity(_,_,_)),
	succedent(Ru,[E1]),
	atom_for_demodulation( E1,PreO, RS, L, R),
	ruleInstanceOf_in_((Sub,1,RS,R),(E,1,==,NE)),
	nonvar( Sub ),
	match( L, Sub ).






mess1_crw( LV,E,C ):-
	print_terms([LV,'
	**taut**:  ',E,'  ',C]),nl,nl, !.
mess2_crw( LV,E,C,E1 ):-
	print_terms([LV,'
	',E,'  ',C, '  **unif into**
	',E1,'  ',C]),nl,nl,!.
mess3_crw( LV,E,C,NE ):-
	print_terms([LV,'
	',E,'  ',C, '  **rewritten into**
	',NE]),nl,nl,!.
mess4_crw( LV,E,C,NE,NC,Calls ):-
	print_terms([Calls,' ',LV,'
	',E,'  ',C, '  **into**
	',NE,'  ',NC]),nl,nl,!.
mess4_crw( LV,E,C,NE,NC ):-
	print_terms([LV,'
	',E,'  ',C, '  **into**
	',NE,'  ',NC]),nl,nl,!.
mess5_crw( LV,E,C ):-
	print_terms([LV,' 
	don''t know how to get rid of ',E,'  ',C]),nl,nl,!.

mess6_crw(Calls,LV,E,C):-
	print_terms(['crw depth ',Calls,' ',LV,'
	',E,'  ',C ]),nl,nl,!.

mess7_crw(LV,C1):-
	print_terms(['forced crw with ',LV,' ',C1]),
	nl,nl,!.

mess8_crw( LV,E,C,ECs ):-
	print_terms(['Complementary case of ',LV,'
	',E,'  ',C, '
  with additional equality constraints
	',ECs]),nl,nl,!.

mess9_crw( LV,E,C,ECs ):-
	print_terms(['Complementary case of ',LV,'
	',E,'  ',C, '
  with additional ordering constraints
	',ECs]),nl,nl,!.




constraint_implies_unifier( _, [] ):- !.
constraint_implies_unifier( C, [X=T|U] ):-
	and_not_eq( C, [X,T], CC ),
	simplify( CC, CC1 ),
	false( CC1 ), !,
	constraint_implies_unifier( C, U ), !.


%___________________________________________________________________________
%
%	DELETE REDUNDANT RULES BY CONSTRAINED REWRITING
%
%  We consider only old rules. New rules are considered only
%  when they are selected for inferences. This avoids checking 
%  large rules repeatedly.
%___________________________________________________________________________


delete_redundant_rules_by_case_an:-
	flag_on( rp, ruca ),
	rule( N, R ),
	is_old(R),
	redundant_rule_by_case_analysis( R ),
	delete_rule( N, R  ), 
	message( ruca, [R] ),
	fail.
delete_redundant_rules_by_case_an.






redundant_rule_by_case_analysis(R):-
	flag_on( rp, ruca ),
	\+ dont_simplify(R),
	is_atom( R ),
	number( R, N ),
	(user:spy_clause(N) -> trace; true),
	(cont(new_small_rule,CheckedWithNewRule) -> 
	     (\+ member(N,CheckedWithNewRule) -> 
		  store(new_small_rule,[N|CheckedWithNewRule]),
		  maximal_succedent( R, [E] ),
		  true( C ), 
		  get_equation_members( E, S, T ),
		  free_vars( S=T, L ),
	          case_analysis( L, N, [[S=T,C]] )
	     )
	), 
	!.


%__________________________________________________________________________
%
%       CASE ANALYSIS ON VARIABLES (MARTIN AND NIPKOW)
%
%__________________________________________________________________________


% LV contains a subset of the variables in E 

martin_nipkow( LV, N, [[E,C]] ):-
	var_case( LV ),
	ground_rewrite_martin_nipkow( N, E, S=T ),
	\+equal(S,T), 
	length(LV,M),
	length(V,M),
	maplist(equate,LV,V,Subst),
	apply_substitution(Subst,(S,T,C),(L,R,C1)),

	var_case_as_constraint(LV,V,IEq,Eq),
	and_constraint_ineq_list(C1,IEq,C2),
	and_constraint_eq_list(C2,Eq,NC),

	crw_depth(D),
%	\+ timeOut(constrained_rewriting(D,V,N,[[L=R,NC]]),180000,success),
	\+ constrained_rewriting(D,V,N,[[L=R,NC]]),
	!, fail.
martin_nipkow( _, _, _ ).


and_constraint_ineq_list(C,[],C).
and_constraint_ineq_list(C,[I|Is],C1):-
	and_gt(C,I,C2),
	and_constraint_ineq_list(C2,Is,C1).



and_constraint_eq_list(C,[],C).
and_constraint_eq_list(C,[I|Is],C1):-
	and_eq(C,I,C2),
	and_constraint_eq_list(C2,Is,C1).


equate(X,Y,X=Y).



var_case_as_constraint(VC,V,IEq,Eq):-
	eq_constraints(VC,V,Eq),
	ineq_constraints(VC,V,IEq).


eq_constraints(VC,VL,IE):-
	VC1=..[v|VC],
	VL1=..[v|VL],
	setof0([X,Y],at_equal_positions(VC1,VL1,X,Y),IE).

at_equal_positions(VC,VL,X,Y):-
	genarg(I,VC,C),
	genarg(J,VC,C),
	I\==J,
	arg(I,VL,X),
	arg(J,VL,Y).


ineq_constraints(VC,VL,IE):-
	VC1=..[v|VC],
	VL1=..[v|VL],
	setof0([X,Y],at_different_positions(VC1,VL1,X,Y),IE).

at_different_positions(VC,VL,X,Y):-
	genarg(I,VC,CI),
	genarg(J,VC,CJ),
	I\==J,
	CI\==CJ,
	(greater_op(CI,CJ)->
	     arg(I,VL,X),
	     arg(J,VL,Y)
	;
	     arg(I,VL,Y),
	     arg(J,VL,X)).





% terms need not be ground really
ground_rewrite_martin_nipkow( N, T, NT ):-
	simplifier( N1, Op, RS, L, R ),
	ruleInstanceOf_in_((Sub,1,RS,R),(T,1,==,T1)),
	nonvar( Sub ),
	N1 \== N,
	match(L,Sub),
	( Op = l ; greater( Sub, R ) ),
	ground_rewrite_martin_nipkow( N, T1, NT ), !.
ground_rewrite_martin_nipkow( _, T, T ).


var_case( [] ).
var_case( ['a$'] ).

var_case( ['a$', 'a$'] ).
var_case( ['b$', 'a$'] ).
var_case( ['a$', 'b$'] ).

var_case( ['a$', 'a$', 'a$'] ).
var_case( ['a$', 'a$', 'b$'] ).
var_case( ['b$', 'b$', 'a$'] ).
var_case( ['a$', 'b$', 'b$'] ).
var_case( ['b$', 'a$', 'a$'] ).
var_case( ['a$', 'b$', 'a$'] ).
var_case( ['b$', 'a$', 'b$'] ).
var_case( ['a$', 'b$', 'c$'] ).
var_case( ['a$', 'c$', 'b$'] ).
var_case( ['b$', 'a$', 'c$'] ).
var_case( ['b$', 'c$', 'a$'] ).
var_case( ['c$', 'a$', 'b$'] ).
var_case( ['c$', 'b$', 'a$'] ).

var_case( ['d$', 'd$', 'd$', 'd$'] ).
var_case( ['d$', 'd$', 'd$', 'a$'] ).
var_case( ['d$', 'd$', 'a$', 'd$'] ).
var_case( ['d$', 'a$', 'd$', 'd$'] ).
var_case( ['a$', 'd$', 'd$', 'd$'] ).
var_case( ['d$', 'd$', A   , B   ] ):- var_case( [A,B] ).
var_case( ['d$', A   , 'd$', B   ] ):- var_case( [A,B] ).
var_case( ['d$', A   ,    B, 'd$'] ):- var_case( [A,B] ).
var_case( [A   , 'd$', 'd$', B   ] ):- var_case( [A,B] ).
var_case( [A   , 'd$', B   , 'd$'] ):- var_case( [A,B] ).
var_case( [A   , B   , 'd$', 'd$'] ):- var_case( [A,B] ).
var_case( ['d$', A   , B   , C   ] ):- var_case( [A,B,C] ).
var_case( [A   , 'd$', B   , C   ] ):- var_case( [A,B,C] ).
var_case( [A   , B   , 'd$', C   ] ):- var_case( [A,B,C] ).
var_case( [A   , B   , C   , 'd$'] ):- var_case( [A,B,C] ).






%_________________________________________________________________
%
%	SIMPLIFICATION BY CLAUSAL REWRITING
%
% Moved to redundant inference. Could be used here for initial clauses.
%_________________________________________________________________

redundant_clause_by_crw(C):-
	complexityBound(C,CB),
	nonvar(CB),
	constraint(C,Ct),
	redundant_by_clausal_rewriting(C,Ct,CB,[]),
	message(ccrw,[C]).

simplify_by_crw(C,C1):-
	complexityBound(C,CB),
	nonvar(CB),
	!,
	constraint(C,Ct),
	redundant_by_clausal_rewriting(C,Ct,CB,Ct1),
	(Ct1=[] ->
	     message(ccrw,[C]),
	     fail
	;
	     constraint(Ct1,C,C1)
	).
simplify_by_crw(C,C).


%_________________________________________________________________
%
%       Variable elimination and quantifier elimination 
%
%_________________________________________________________________


simplify_by_var_elim(C1,C2):-
	flag_on(rp,skex),
	sk(C1,C),
	!,
	number( C1, N1 ),
	history_and_number( [skex, [N1], [C1]], C, C2 ),
	message( skex, [C1,C2] ),
	recordProofOfClause(C2).
simplify_by_var_elim(C1,C2):-
	simplify_by_var_elim1(C1,C2).

simplify_by_var_elim1(C1,C2):-
	try_var_elim,
	variableElimination(C1,Cs),
	!,
	member(C,Cs),
	number( C1, N1 ),
	history_and_number( [vare, [N1], [C1]], C, CI ),
	message( vare, [C1,CI] ),
	recordProofOfClause(CI),
	simplify_by_var_elim1(CI,C2).
simplify_by_var_elim1(C,C).


try_var_elim:-
	symbol_admits_var_elim(_),
	flag_on(es, vare),
	!.



% ----------------------------------------------------------------
%
%         Ce is the list of clauses obtained from
%            variable elimination applied to clause Cl
%         don't-care non-deterministic selection of the
%         variable if more than one exists.
%
% ----------------------------------------------------------------


variableElimination(Cl,Ce):-
	veForTotalOrderings(Cl,Ce).
variableElimination(Cl,Ce):-
	veForPartialOrderings(Cl,Ce).
variableElimination(Cl,Ce):-
	\+ flag_on(rp,prop),   % otherwise incomplete selection strategy
	clausePartition(Cl,Ce),
%	format('~2n~p split into',[Cl]),
%	out(Ce),
	!.
variableElimination(Cl,Ce):-
	equation_splitting(Cl,Ce).




/*   causes incompleteness
variableElimination(Cl,Ce):-
	finite_domain(D),
	length(D,N),
	N=<2,
	ground_instances(Cl,D,Ce).
*/



variableElimination(Cl,[Cl]):-
	finite_domain([A,B]),
	(greater_op(A,B) ->
	     A1=A, B1=B
	;
	     A1=B,B1=A
	),
	constraint(Cl,[conj(Conj,[],[])]),
	member(X>Y,Conj),
        instantiate_lower(A1,B1,X,Y),
	!.
variableElimination(Cl,[]):-
	finite_domain([A,B]),
	(greater_op(A,B) ->
	     B1=B
	;
	     B1=A
	),
	constraint(Cl,[conj(Conj,[],[])]),
	member(X>_Y,Conj),
        below_min(B1,X).


/*

variableElimination(Cl,[Cl,C1]):-
	finite_domain([A,B]),
	succedent(Cl,S),
	member(E,S),
	is_equational_atom(E),
	get_atom_members(E,X,Y),
	var(X),var(Y),
	!,
	copy_term((Cl,X,Y),(C1,X1,Y1)),
	X=A,
	Y=B,
	X1=B,
	Y1=A.

*/




instantiate_lower(A,B,X,Y):-
	X==A,
	var(Y),
	!,
	Y=B.

below_min(B,X):-
	X==B.



out([]).
out([C|Ce]):-
	format('~n   ~p',[C]),
	out(Ce).

ground_instances(C,Domain,Cs):-
	antecedent(C,A),
	succedent(C,S),
	free_vars((A,S),V),
	length(V,N),
	N>0,
%	N=<2,
	findall([A,S],choose(V,Domain),Seqs),
	maplist(createClause,Seqs,Cs).


choose([X|Xs],D):-
	member(X,D),
	choose(Xs,D).
choose([],_).



sk(Cl,Ce):-
	firstOrder,
	antecedent(Cl,A),
	succedent(Cl,S),
	sequent_calculus_expansion((A,S),C),
	createClauseP(C,Ce).

/*

% testing
cl(([],[A])):-
%	E=(eq(X,Y)==all(member('$$0',X)==member('$$0',Y))),
%	E = (~all(all(sub(u('$$1','$$0'),u('$$0','$$1'))and sub(u('$$0','$$1'),u('$$1','$$0'))))),
%	E =(~all(all(all(in('$$0','$$1')or in('$$0','$$0')=>in('$$0','$$0')or in('$$0','$$1'))and all(in('$$0','$$0')or in('$$0','$$1')=>in('$$0','$$1')or in('$$0','$$0'))))),

	E=(~all((f('i$0')=>h('i$0'))and all(g('i$0')=>j('i$0')))==all(all(f('i$1')and g('i$0')=>h('i$1')and j('i$0')))),
	create_atom(E,A).






test:-
	cl(C),
	sequent_calculus_expansion(C,Seqs),
	maplist(createClauseP,Seqs,Ce),
	maplist(print_nl,Ce).

print_nl(C):-
        print(C),
	nl.



*/


clausePartition(Cl,Ce):-
	horn(off),
%	theory_flag(Cl,(goal)),   % this proved to not be appropriate
	literals(Cl,L),
	length(L,N),
	N=<8,    % otherwise we need a more clever implementation!
%	groundLiterals(L,LG,LNG),
	clauseSegments(L,L1,L2,Vs,Abbreviations),
	literal_list_to_sequent(L1,A1,S1),
	literal_list_to_sequent(L2,A2,S2),
	(Abbreviations=[P1] ->
	     % found P1 -> L1; L2 not found
	     % need to generate new abbreviation PS
	     % with PS -> L2 
	     % and generate PS -> L2 and P1 or PS
	     new_proposition_symbol(caseSplit,PS),
	     P=..[PS|Vs],
	     create_atom(P,==,'$true$',At),
	     A21=[At|A2],
	     S21=S2,
	     recordSplitCase(true(P),A2,S2),
	     createClause([A21,S21],C2),
	     literal_list_to_sequent([[+|At],P1],AA,SS),
	     createClause([AA,SS],C1),
%	     format('~n  ~p split into ~n    ~p and ~n    ~p',[Cl,C1,C2]),
	     Ce=[C1,C2]

	;Abbreviations=[P1,P2] ->
	     % found P1 -> L1 and P1 -> L2
	     % generate P1 or P2
	     literal_list_to_sequent([P1,P2],AA,SS),
	     createClause([AA,SS],C2),
%	     format('~n  ~p split into ~n    ~p and ~n    ~p',[Cl,P1,P2]),
	     Ce=[C2]
	;%true
	     new_proposition_symbol(caseSplit,PS),
%	     new_proposition_symbol(abbreviation,PS),
	     P=..[PS|Vs],
	     create_atom(P,==,'$true$',At),
	     A11=A1,
	     S11=[At|S1],
	     A21=[At|A2],
	     S21=S2,
	     createClause([A11,S11],C1),
	     createClause([A21,S21],C2),
	     recordSplitCase(false(P),A1,S1),
	     recordSplitCase(true(P),A2,S2),
%	     format('~n  ~p split into ~n    ~p and ~n    ~p',[Cl,C1,C2]),
	     Ce=[C1,C2]
	),
	!.


clauseSegments(L,L1,L2,Vs,Abbrev):-
	partition(L,L11,L12),
	proper_partition(L11,L12,Vs),
	\+ clauseSegments(L11,_,_),
	literal_list_to_sequent(L11,A1,S1),
	literal_list_to_sequent(L12,A2,S2),
	case_exists(A1,S1,Vs,P1),
	case_exists(A2,S2,Vs,P2),
	(P1\==[] ->
	     L1=L11,
	     L2=L12
	;
	     L1=L12,
	     L2=L11
	),
	append(P1,P2,Abbrev),
	!.

	
clauseSegments(L,L1,L2):-
	partition(L,L1,L2),
	proper_partition(L1,L2,_),
	\+ clauseSegments(L1,_,_).


combine_with_negation(A,S,LG,L,C):-
	negate(L,NL),
	sequent_to_literals(A,S,Ls),
	append(Ls,LG,LsLG),
	literal_list_to_sequent([NL|LsLG],A1,S1),
	createClause([A1,S1],C).


case_exists(A,S,Vs,[[-|PT]]):-
	sequent_to_literals(A,S,L),
	splitCase(false(P),A1,S1), % found ~P(x) == all y l1(x,y)
	sequent_to_literals(A1,S1,L1),
	mylists:permutation(Vs,Vsp),
	P=..[_|Vsp],
	\+ \+ (numbervars(Vs,0,K),   % now: ~P(c) == all y l1(c,y);
	       % check whether all y l1(c,y) can be alpha converted
	       % into all y' l(c,y') modulo AC of disjunction
	       \+ \+ (numbervars(L,K,_),subsumes(L1,L)),
	       % and symmetrically
	       (numbervars(L1,K,_),subsumes(L,L1))
	      ),
%	subsumes(L,L1),  % now we match the variables
	create_atom(P,PT),
	!.
case_exists(A,S,Vs,[[+|PT]]):-
	sequent_to_literals(A,S,L),
	splitCase(true(P),A1,S1),
	sequent_to_literals(A1,S1,L1),
	mylists:permutation(Vs,Vsp),
	P=..[_|Vsp],
	\+ \+ (numbervars(Vs,0,K),   % P(c) == all y l1(c,y)
	       \+ \+ (numbervars(L,K,_),subsumes(L1,L)),
	       (numbervars(L1,K,_),subsumes(L,L1))
	      ),
	create_atom(P,PT),
	!.
case_exists(_A,_S,_Vs,[]).

subsumes([],[]).
subsumes([[P|L]|Ls],Rs):-
	select([P|R],Rs,Rs1),
	match_atoms(L,R),
	!,
	subsumes(Ls,Rs1).







groundLiterals([],[],[]).
groundLiterals([L|Ls],[L|Ls1],Ls2):-
	is_closed(L),
	!,
	groundLiterals(Ls,Ls1,Ls2).
groundLiterals([L|Ls],Ls1,[L|Ls2]):-
	groundLiterals(Ls,Ls1,Ls2).



	

proper_partition(L1,L2,VI):-
	member([_|E],L1),
	\+is_caseSplit(E),
	member([_|F],L2),
	\+is_caseSplit(F),
	free_vars(L1,V1),
	free_vars(L2,V2),
	intersection_vars(V1,V2,VI),
	append(L1,L2,Ls),
	maplist(shieldedInLargeLiteral(Ls),VI),
	!,
	% we don't want splitting to interfere with 
	% quantifier elimination
%	(dense_ordering_relation_symbol(_) -> VI=[]; true),
	member(X,V1),
	\+ member212(X,VI),
	!,
	member(Y,V2),
	\+ member212(Y,VI),
	!.






/*
	
case_exists(A,S,Rem):-
	sequent_to_literals(A,S,L),
	( 
%	  clause(C)
%	;
	  rule(_,C)
	),
	literals(C,LC),
	partition(LC,LC1,Rem),
	maplist(is_propositional_literal,Rem),
	\+ \+ (numbervars(L,0,_),subsumes(LC1,L)),
	\+ \+ (numbervars(LC1,0,_),subsumes(L,LC1)),
	!.


*/




equation_splitting(Cl,[Cl1,Cl2]):-
	antecedent(Cl,A),
	succedent(Cl,S),
	select(E,S,S1),
	is_equational_atom(E),
	get_atom(E,RS,L,R),
	nonvar(L),
	nonvar(R),
	free_vars(L,VL),
	free_vars(R,VR),
	intersection_vars(VL,VR,VI),
	member(X,VL),
	\+ member212(X,VI),
	member(Y,VR),
	\+ member212(Y,VI),
	free_vars((A,S1),VC),
	subset_vars(VC,VI),
	length(VI,K),
	(RS= == ->
	     functor(L,F,N),
	     functor(R,G,M),
	     new_predicate_symbol_below([F/N,G/M],H/K)
	;
	     maximal_subterm_functor(VI,L,F), 
	     maximal_subterm_functor(VI,R,G), 
	     max_op(F,G,_F1),
%	     get_term_operators(L,OL),
%	     get_term_operators(R,OR),
%	     append(OL,OR,O),
%	     new_function_symbol_below(O,H/K)
%	     new_function_symbol_below([F,G],H/K)  % this should be it
	     new_function_symbol_below([_F1],H/K)  % this seems to work better
	),
	HT=..[H|VI],
	AL=..[RS,L,HT],
	AR=..[RS,R,HT],
	create_atom(AL,EL),
	create_atom(AR,ER),
	createClause([A,[EL|S1]],Cl1),
	createClause([A,[ER|S1]],Cl2),
	!.

max_op(F/N,G/_M,F/N):-
	greater_op(F,G),
	!.
max_op(_,G,G).


maximal_subterm_functor(_,T,_):-
	var(T),
	!,
	fail.
maximal_subterm_functor(V,T,F/N):-
	free_vars(T,FV),
	subset_vars(V,FV),
	functor(T,G,M),
	msf_subterms(T,0,M,V,G/M,F/N).


msf_subterms(_,M,M,_,G/N,G/N):-
	!.
msf_subterms(T,I,M1,V,G/M,H/K):-
	I1 is I+1,
	arg(I1,T,TI),
	(maximal_subterm_functor(V,TI,F/N) ->
	     (greater_op(F,G) ->
		  Max=F/N
	     ;
	          Max=G/M
	     )
	;
	     Max=G/M
	),
	msf_subterms(T,I1,M1,V,Max,H/K).

	


% ------------- sequent calculus expansion  -----------------
	


sequent_calculus_expansion(Cl,Ce):-
	'sequent_calculus_expansion!'(Effect,Cl,Ce),
	var(Effect).

'sequent_calculus_expansion!'(Effect,(A,S),Ce):-
	(simple_sk(A,S,C) ->
	     'sequent_calculus_expansion!'(_,C,Ce)
	;
	     Effect=notExpanded,
	     Ce=(A,S)
	).





createClauseP((A,S),C):-
	createClause([A,S],C).




% ------------- variable elimination for total orderings  -----------------

	

veForTotalOrderings(Cl,Ce):-
	antecedent(Cl,Ant),
	succedent(Cl,Suc),
	append(Ant,Suc,AS),
	!,
	unshieldedVariable(AS,X,RS),
	detachVariable(Ant,X,AntNoX,AntX),
	mkPositive(RS,AntX,Suc1X),
	detachVariable(Suc,X,SucNoX,Suc2X),
	append(Suc1X,Suc2X,SucX),
	bounds(X,SucX,L,R,E),
	varElim(RS,L,R,E,Variants),
	appendToClauses(AntNoX,SucNoX,Variants,Ce),
	!.


% varElim(RS,L,R,E,X)
% X is eliminated in clauses of the form
%        \/i L_i <1 X, \/j X <2 R_j, \/k X=E_k
% where < is RS and <i either < or =<, i=1,2.
% We generate clauses 
%        \/ij (L_i <12 R_j ,(L_i=R_j & \/k L_i=E_k)),
% with <12 is < if both <1 and <2 are <, and =<, otherwise.

% The list of these clauses (as pairs of antecedent and succedent)
% is returned in the last argument.
% If k=0 the empty disjunction \/k L_i=E_k denotes, as usual, false.

varElim(RS,_L,[],_E,C):-
	strict_part(RS,RSS),
	noendL_relation_symbol(RSS),
	bottom_cnf(C),
	!.
varElim(RS,[],_R,_E,C):-
	strict_part(RS,RSS),
	noendR_relation_symbol(RSS),
	bottom_cnf(C),
	!.
varElim(RS,L,R,[],Cs):-
	L\==[],
	R\==[],
	strict_part(RS,RSS),
	reflexive_closure(RSS,RSR),
	total_ordering_relation_symbol(RSS),
	(\+ (member((RSR,_),L), member((RSR,_),R)) ->
	     intervals(L,R,LltR),
	     disjunctionFromIneqs(LltR,[],[],Cs)
	;discrete_ordering_relation_symbol(RSS,Suc) ->
	     apply_suc(R,RSS,RSR,Suc,SucR),
	     intervals(L,SucR,LltR),
	     disjunctionFromIneqs(LltR,[],[],Cs)),
	!.
varElim(RS,L,R,E,Cs):-
	strict_part(RS,RSS),
	dense_ordering_relation_symbol(RSS),
	intervals(L,R,LltR),
	eqintervals(L,R,LeqR),
	disjunctionFromIneqs(LltR,LeqR,E,Cs),
	!.
varElim(RS,L,R,[(=,T)|Es],Cs):-
	strict_part(RS,RSS),
	reflexive_closure(RSS,RSR),
	discrete_ordering_relation_symbol(RSS,_),
	varElim(Rs,[(RSR,T)|L],R,Es,Cs1),
	varElim(Rs,L,[(RSR,T)|R],Es,Cs2),
	append(Cs1,Cs2,Cs).






veForPartialOrderings(Cl,[C]):-
	antecedent(Cl,Ant),
	succedent(Cl,Suc),
	append(Ant,Suc,AS),
	!,
	unshieldedVariable(AS,X,RS),
	transitive_relation_symbol(RS),
	reflexive_relation_symbol(RS),
	detachVariable(Suc,X,_,[]),
	detachVariable(Ant,X,AntNoX,AntX),
	bounds(X,AntX,L,R,E),
	vePO(RS,L,R,E,NewAnt),
	append(AntNoX,NewAnt,A),
	createClause([A,Suc],C),
	!.

/*
Elimination rules for nonstrict partial orderings >=
(x variable, not in a, b, t, C):

x>=t -> C     <=>  C
t>=x -> C     <=>  C

a>=x & /\ x>=bi  -> C    <=> /\ x>=bi -> C
/\ aj>=x & x=<b  -> C    <=> /\ aj>=b -> C
*/


vePO(RS,[(RS,_)],[],[],[]):-
	!.
vePO(RS,[],[(RS,_)],[],[]):-
	!.
vePO(_RS,L,R,[],C):-
        L=[_],
	intervals(L,R,LltR),
	disjunctionFromIneqs(LltR,[],[],[C]),
	!.
vePO(_RS,L,R,[],C):-
        R=[_],
	intervals(L,R,LltR),
	disjunctionFromIneqs(LltR,[],[],[C]),
	!.






apply_suc([],_,_,_,[]).
apply_suc([(RSR,R)|RR],RSS,RSR,Suc,[(RSS,SucR)|SucRR]):-
	SucR=..[Suc,R],
	apply_suc(RR,RSS,RSR,Suc,SucRR).
apply_suc([(RSS,R)|RR],RSS,RSR,Suc,[(RSS,R)|SucRR]):-
	apply_suc(RR,RSS,RSR,Suc,SucRR).


/*
in order for this to work well, we need to also eliminate
variables occurring under s and p only.
*/
%discrete_ordering_relation_symbol(<,s).  % for testing
discrete_ordering_relation_symbol(_,_):-     % needs to be extended
	fail.


intervals([],_,[]).
intervals(_,[],[]):-
	!.
intervals([(RS,L)|LL], RR, I):-
	intervals1(RS,L,RR,IL),
	intervals(LL,RR,ILL),
	append(IL,ILL,I).

intervals1(_,_,[],[]).
intervals1(RS,T,[(RS1,R)|RR],[A|I]):-
	(
	RS=RS1 ->
	     RS2=RS1;
	RS= = ->
	     RS2= = ;
	reflexive_closure(RS,RS2);
	RS2=RS
	),
	P=..[RS2,T,R],
	create_atom(P,A),
	intervals1(RS,T,RR,I).


eqintervals([],_,[]).
eqintervals(_,[],[]):-
	!.
eqintervals([(_RS,L)|LL], RR, I):-
	intervals1(=,L,RR,IL),
	eqintervals(LL,RR,ILL),
	append(IL,ILL,I).

%-------------------------------------------------














%  Generates
%        \/ij (L_i < R_j ,(L_i=R_j & \/k L_i=E_k))
%  from the lists L_i < R_j and L_i=R_j (in term format).
%
%  The second disjunct is empty if E is empty.
%
%------------------------------------------------

disjunctionFromIneqs(LltR,_,[],[LltR]):-
	!.
disjunctionFromIneqs([],_,_,C):-
	bottom_cnf(C),
	!.
disjunctionFromIneqs([IE|IEqs],[E|Es],ETs,D):-
	get_atom_members(IE,S,T),
	intervals1(=,S,ETs,DSE),
	(equal(S,T) ->
	     D1=[DSE],
	     D11=D1
	;    and_cnf([[E]],[DSE],D1),
	     or_cnf([[IE]],D1,D11)),
	disjunctionFromIneqs(IEqs,Es,ETs,D2),
	or_cnf(D11,D2,D).

	

appendToClauses(_A,_S,[],[]).
appendToClauses(A,S,[S1|V],[Cl|C] ):-
	append(S,S1,SS),
	sort(SS,S2),
	createClause([A,S2],Cl),
	appendToClauses(A,S,V,C).


detachVariable([],_X,[],[]).
detachVariable([E|Es],X,EsNoX,EsX):-
	get_term_vars(E,V),
	(\+member212(X,V) ->
		EsNoX=[E|No1],
		EsX=Yes1
	;	EsNoX=No1,
		EsX=[E|Yes1]),
	detachVariable(Es,X,No1,Yes1).
	

unshieldedVariable(Es,X,RS):-
	member(E,Es),
	get_atom_relation_symbol(E,RS),
	transitive_relation_symbol(RS),
	get_atom_members(E,S,T),
	member(X,[S,T]),
	var(X),
	\+shielded(Es,X,RS).
	

shielded(Es,X,RS):-
	member(E,Es),
	get_atom_relation_symbol(E,RS1),
	((RS=RS1;
	  RS1= =; 
	  reflexive_closure(RS,RS1); 
	  reflexive_closure(RS1,RS)) ->
		get_atom_members(E,S,T),
		((S==X,T==X); shieldedInTerm(S,X); shieldedInTerm(T,X))
	;	atom_as_term(E,T),
		shieldedInTerm(T,X)),
	!.


shieldedInTerm(T,X):-
	nonvar(T),
	genarg(_,T,T1),
	get_term_vars(T1,V1),
	member212(X,V1),
	!.

shieldedInLargeLiteral(Ls,X):-
	member([_Pol|E],Ls),
	\+ is_abbreviation(E),
	\+ is_caseSplit(E),
	get_atom_members(E,S,T),
	(shieldedInTerm(S,X); shieldedInTerm(T,X)),
	!.



mkPositive(_RS,[],[]).
mkPositive(RS,[E|Es],Fs):-
	strict_part(RS,RSS),
	reflexive_closure(RS,RR),
	get_atom_relation_symbol(E,RS1),
	get_atom_members(E,S,T),
	(RS1= = ->
		E1=..[RSS,S,T],
		create_atom(E1,A1),
		E2=..[RSS,T,S],
		create_atom(E2,A2),
	        Fs=[A1,A2|Fs1]
	;RS1=RSS ->
	        E1=..[RR,T,S],
		create_atom(E1,A1),
		Fs=[A1|Fs1]
	;% RS1=RR 
	        E1=..[RSS,T,S],
		create_atom(E1,A1),
		Fs=[A1|Fs1]
	),

	mkPositive(RS,Es,Fs1).
	

bounds(_,[],[],[],[]).
bounds(X,[Eq|Suc],L,R,E):-
	get_atom_members(Eq,S,T),
	get_atom_relation_symbol(Eq,RS),
	(RS= = ->
		(X==S ->
			E=[(=,T)|Es]
		;	E=[(=,S)|Es]),
		R=Rs,
		L=Ls
	;	(X==S ->
			L=Ls,
			R=[(RS,T)|Rs]
		;	L=[(RS,S)|Ls],
			R=Rs
		),
		E=Es),
	bounds(X,Suc,Ls,Rs,Es).
		
	









% eof _________________________________________________________________

