%declarations%

:- module(constraintSolverBasics, [ 
	implies/2,
	impliesA/2,
	simplify_constraint_vars/3,
	satisfiable_ineq/1,
	true_ineq/1
   ]).

:- use_module(termOrder, [
	greater/2,
	greaterEqual/2
   ]).


:- use_module(library(basics),[member/2]).

:- use_module(equationalTheory, [
	equal/2,
        mgu/2
     ]).


%end%


satisfiable_ineq( S>=T):-
	\+ greater( T, S ).
satisfiable_ineq( S>T ):-
	\+ greaterEqual( T, S ).

true_ineq( S>T):-
	greater( S, T ).
true_ineq( S>=T ):-
	greaterEqual( S, T ).


%------------------------------------------------------------------------------
% eliminates non used variables of equality constraints
%------------------------------------------------------------------------------

simplify_constraint_vars( LV, [conj(I,E,O)], [conj(I,U,O)] ):-
	mgu( E, U1 ),
	delete_non_used_vars( LV, U1, U ),!.
simplify_constraint_vars( _LV, C, C ).

delete_non_used_vars( _, [], [] ).
delete_non_used_vars( LV, [X=_|U], U1 ):-
	(member( X1, LV ), X1==X ->
	     delete_non_used_vars( LV, U, U1 )
	;    fail).
delete_non_used_vars( LV, [E|U], [E|U1] ):-
	delete_non_used_vars( LV, U, U1 ), !.




/*
implies(+A,+B)
is true when a conjunction of atomic ordering constraint A implies
an atomic constaint B (approximatively).
*/
implies(_,S>T):-
	greater(S,T),
	!.
implies(_,S>=T):-
	greaterEqual(S,T),
	!.
implies( [I|_],J ):-
	impliesA(I,J),
	!.
implies( [_|Is],J):-
	!,
	implies(Is,J).
implies(A,B):-
	impliesA(A,B).


/*
impliesA(+A,+B)
is true when an atomic ordering constraint A implies B
(approximatively).
*/
impliesA(A,B):-
	equal(A,B),
	!.
impliesA( S>T, S1>=T1 ):-
	equal(S>T,S1>T1), !.
impliesA( S>T, S1>T1 ):-
	equal(T,T1),
	greater( S1, S ), !.
impliesA( S>=T, S1>T1 ):-
	equal(T,T1),
	greater( S1, S ), !.
impliesA( S>T, S1>=T1 ):-
	equal(T,T1),
	greater( S1, S ), !.
impliesA( S>=T, S1>=T1 ):-
	equal(T,T1),
	greater( S1, S ), !.

impliesA( S>T, S1>T1 ):-
	equal(S,S1),
	greater( T, T1 ), !.
impliesA( S>=T, S1>T1 ):-
	equal(S,S1),
	greater( T, T1 ), !.
impliesA( S>T, S1>=T1 ):-
	equal(S,S1),
	greater( T, T1 ), !.
impliesA( S>=T, S1>=T1 ):-
	equal(S,S1),
	greater( T, T1 ), !.
