%declarations%

:- module(hgstuff, [ 
        endOfFile/1,
	setof0/3,
	apply/4,
	split/4,
	try/1,
	once/1
   ]).

:-meta_predicate setof0(+,:,-).
:-meta_predicate apply(:,+,+,-).
:-meta_predicate once(:).
:-meta_predicate try(:).
:-meta_predicate while(:,:).
:-meta_predicate split(:,+,-,-).
%:-meta_predicate memo(+,+,+,:).

:- use_module(proofState, [
	clause/1,
	rule/2,
	deleteClause/1
   ]).

:- use_module(classes, [
	number/2
   ]).

:- use_module(global, [
	cont/2,
	store/2
   ]).

:- use_module(clause_processing, [
	process_clause/3
   ]).

:- use_module(thisProlog,[prolog_version/1,
			  file_exists/1,
			  portray_clause2/2,
			  shell/1,
			  my_absolute_file_name/2,
			  savedOn/2]).

:- use_module(saturation,[add_inference/3]).

:- use_module(index, [initial_index/1]).

:- use_module(library(maplist),[maplist/2,maplist/3]).
:- use_module(library(basics),[member/2]).
:- use_module(library(moremaps),[exclude/3]).

:- use_module(auxiliary,[appendAtoms/3]).


%end%




%-------------------------------------------------------------
%
%      for manual selection of inferences: give rule/clause
%      number(s) as in-arguments.
%
% -----------------------------------------------------------

inferences(N):-
	turn_into_rule(N,R),
	initial_index(ST),
	on_exception(ruleSimplified,
		     add_inference(ST,R, [R]),
		     fail),
	fail.
inferences(_).


inferences(N,M):-
	turn_into_rule(N,RN),
	turn_into_rule(M,RM),
	initial_index(ST),
	on_exception(ruleSimplified,
		     (add_inference(ST,RN,[RN,RM]);add_inference(ST,RN,[RM,RN])),
		     fail),
	fail.
inferences(_,_):-
	process_clauses.


turn_into_rule(N,R):-
	rule(N,R),
	!.
turn_into_rule(N,R):-
	clause(Cl),
	number(Cl,N),
	!,
	deleteClause( Cl ),
	initial_index(ST),
	process_clause( Cl, ST, _ ),
	turn_into_rule(N,R).

process_clauses:-
	deleteClause(C),
	initial_index(ST),
	process_clause(C,ST,_),
	fail.
process_clauses.






% test sets


last_problem(LogFile,LP):-
	last_term(LogFile,TT),
	(arg(1,TT,file=LP) ->
	     true
;
	     LP='$start'
	).

last_term(F,'$start'):-
	\+ file_exists(F),
	!.
last_term(F,T):-
	open(F,read,S),
	last_term(S,'$start',T),
	close(S).

last_term(S,T1,T2):-
	read(S,T),
	(endOfFile(T) ->
	     T2=T1
	;
	     last_term(S,T,T2)
	).
	


satSamples(Samples,Log,MaxCls,TimeBound,Options):-
	samples(Samples,F),
	last_problem(Log,LP),
	sublist_after(LP,F,FS),
	open(Log,append,LogStr),
	testfiles((FS,0,0,MaxCls,TimeBound,LogStr,Options)),
	close(LogStr).



process_dir(D,OutDir):-
	file_members_of_directory(D,Fs1),
%	use_module(library(moremaps),[exclude/3]),
	exclude(hgstuff:improper,Fs1,Fs),
	maplist(process_file(OutDir),Fs).



process_file(OutDir,F):-
	name(F,FN),
	file_name_suffix(FN,FileN),
	name(File,FileN),
	appendAtoms([OutDir,File,'.in'],'',O),
	(file_exists(O) ->
	     format('~nFile ~p exists.~n',[O])
	;
	     user:insrc(F),
	     user:sarp([1,4,12,13]),
	     process_clauses,
	     user: out_as_spass(O)
	),
	!.

	

file_name_suffix(FN,FileN):-
	proper_suffix(FN,FileN),
	!.
file_name_suffix(F,F).

proper_suffix(FN,FileN):-
	append("/",FileN,SFileN),
	append(_Dir,SFileN,FN),
	\+ proper_suffix(FileN,_),
	!.


satdir(D):-
	appendAtoms([D,'logfile.log'],'',Log),
	satdir(D,Log,_,_,[]).



satdir(FPattern,Log,MaxCls,TimeBound,Options):-
	file_members_of_directory(FPattern,F1),
	exclude(hgstuff:improper,F1,F),
%	last_problem(Log,LP),
%	sublist_after(LP,F,FS),
	FS=F,
	open(Log,append,LogStr),
	testfiles((FS,0,0,MaxCls,TimeBound,LogStr,Options)),
	close(LogStr).
/*
satdir(D,Log,From,MaxCls,TimeBound,Options):-
	file_members_of_directory(D,F1),
%	use_module(library(moremaps),[exclude/3]),
	exclude(hgstuff:improper,F1,F),
	(From='$start';true),
	sublist_after(From,F,FS),
	open(Log,write,LogStr),
	testfiles((FS,0,0,MaxCls,TimeBound,LogStr,Options)),
	close(LogStr).

*/

/*

tptpdir(+P,?Version,?MaxCls,?TimeBound,?Options,?LogFile)

P is a filename pattern of the form DIR/name or just name.
(In the second case, DIR is taken to be the first 3 letters of name.)

DIR must be a 3 letter subdirectory of TPTP problems such as ``SET''.
If DIR is given, name can be empty in which case the ``/'' can
also be omitted.

Then sat is run over all problems DIR/name* .

All other parameters are optional.
The default logfile is /HG/local/provers/saturate/examples/test/DIR

*/

tptphome(H):-
	(user:os(windows) ->
	    H='c:/opt/TPTP'
	;
	    H='/HG/local/provers/TPTP'
	).

tptpdir(P,Version,MaxCls,TimeBound,Options,LogFile):-
	default(Version,'v2.2.1'),
	default(MaxCls,2000),
	default(TimeBound,400000),  % new SICStus: time in ms???
	  % get less time than what is measured by statistics!
	default(Options,[]),
	(Options=spass ->
	     Opt=spass(TmpFile),
	     thisProlog:gen_tmp_file(spass,TmpFile)
	;Opt=Options),
	tptphome(Home),
	appendAtoms([Home,Version],(-),TPTPDir),
	commands:cd(TPTPDir),
	tptpdir1(P,MaxCls,TimeBound,Opt,LogFile).


% obsolete; notice the swapping of the first 2 parameters
tptpdir(Version,D,MaxCls,TimeBound,Options):-
	tptpdir(D,Version,MaxCls,TimeBound,Options,_LogFile).

default(Version,V):-
	(var(Version) ->
	     Version=V
	;true).


tptpdir1(FPat,MaxCls,TimeBound,Options,Log):-
	format(user_output,
	 '~n- tptp ~p ~n  with options ~p -~2n',[FPat,Options]),
	pattern(FPat,D,F),
	appendAtoms(['Problems',D,F],'/',FPattern),
%	appendAtoms(['Logs',D],'/',LogFile),
	appendAtoms(['/HG/local/provers/saturate/examples/tests',D],'/',LogFile),
	default(Log,LogFile),
	satdir(FPattern,Log,MaxCls,TimeBound,Options).


runtptpproblems([P|Ps],Options):-
	runtptpproblem(P,_Version,_MaxCls,_TimeBound,Options,_LogFile),
	runtptpproblems(Ps,Options).
runtptpproblems([],_Options).


runtptpproblems([P|Ps],LogFile,Options):-
	runtptpproblem(P,_Version,_MaxCls,_TimeBound,Options,LogFile),
	runtptpproblems(Ps,LogFile,Options).
runtptpproblems([],_,_Options).

runtptpproblem(P,Options):-
	runtptpproblem(P,_Version,_MaxCls,_TimeBound,Options,_LogFile).

runtptpproblem(P,Version,MaxCls,TimeBound,Options,LogFile):-
	default(Version,'v2.5.0'),
	tptphome(Home),
	default(MaxCls,10000),
	default(TimeBound,600000),  % new SICStus: time in ms???
	  % get less time than what is measured by statistics!
	default(Options,[commands:simplify,sarp(+[13]),
			 option([
%				 iterative_saturation(on), % not useful if goals unknown
%				 skolem_functions(small)
				 skolem_functions(big)  % default
				])]),
	default(LogFile,'sat_tptp.log'),
	(Options=spass ->
	     Opt=spass(TmpFile),
	     thisProlog:gen_tmp_file(spass,TmpFile)
	;Opt=Options),
	appendAtoms([Home,Version],(-),TPTPDir),
%	appendAtoms(['d:/common/TPTP',Version],(-),TPTPDir),
	commands:cd(TPTPDir),
	pattern(P,Dir,F),
	appendAtoms([TPTPDir,'Problems',Dir,F],'/',P1),
	appendAtoms([P1,'p'],'.',P2),
	open(LogFile,append,LogStr),
	testfile(MaxCls,TimeBound,LogStr,P2,Result,Opt),
	close(LogStr).


pattern(FP,Dir,F):-
	( name(FP,[D,I,R,Slash|FN]), name('/',[Slash]) ->
	     name(Dir,[D,I,R]),
	     name(F,FN)
	;name(FP,[D,I,R,_|FN]) ->
	     name(Dir,[D,I,R]),
	     F=FP
	; % FP is directory
	     Dir=FP,
	     F=''
	).
	

analyse_dir(DF):-
	(DF=(D,F); D=DF, F=''),
	!,
	commands:cd('/HG/usr/hg/saturate/TPTP'),
	appendAtoms(['Problems',D,F],'/',Dir),
	appendAtoms(['/HG/usr/hg/saturate/TPTP/props',D],'/',LogFile),
	file_members_of_directory(Dir,F1),
%	use_module(library(moremaps),[exclude/3]),
	exclude(hgstuff:improper,F1,F2),
	open(LogFile,write,LogStr),
	anafiles(F2,LogStr),
	nl(LogStr),
	close(LogStr).
	

anafiles([],_).
anafiles([F|Fs],Log):-
	user:insrc(F),
	theory:summary(S),
	format(Log,'~nFile ~p: ~p',[F,S]),nl(Log),flush_output(Log),
	anafiles(Fs,Log).


sublist(_I,_J,[],[]):-
	!.
sublist(I,J,_,[]):-
	I>J,
	!.
sublist(1,J,[X|Xs],[X|Ys]):-
	!,
	J1 is J-1,
	sublist(1,J1,Xs,Ys).
sublist(I,J,[_X|Xs],Ys):-
	I1 is I-1,
	J1 is J-1,
	sublist(I1,J1,Xs,Ys).


sublist_after(_I,[],[]):-
	!.
sublist_after('$start',L,L):-
	!.
sublist_after(X,[X|Xs],Xs):-
	!.
sublist_after(X,[Y|Xs],[Y|Xs]):-
	name(X,NX),
	name(Y,NY),
	lexlt(NX,NY),
	!.
sublist_after(X,[_X|Xs],Ys):-
	sublist_after(X,Xs,Ys).


lexlt([],[_|_]).
lexlt([X|Xs],[X|Ys]):-
	lexlt(Xs,Ys).
lexlt([X|_],[Y|_]):-
	X<Y.

file_members_of_directory(D,M):-
	appendAtoms(['directoryAsList ',D,'* > /tmp/ls'],'',Command),
	shell(Command),
	open('/tmp/ls',read,Str),
	read(Str,M),
	close(Str).

samples(D,DM):-
	open(D,read,Str),
	read(Str,M),
	maplist(extract_dir,M,DM),
	close(Str).

extract_dir(F,DFP):-
	name(F,[D,I,R|_]),
	name(Dir,[D,I,R]),
	appendAtoms(['Problems',Dir,F],'/',DF),
	appendAtoms([DF,p],'.',DFP).


long_testsuite:-
	store(long,true),
	testsuite1.




testsuite:-
%	testsuite([sams([])]).
	testsuite([sarp(+[14]),sams([])]).

testsuite(O):-
	store(long,false),
	testsuite1(O).



testsuite1(O):-
%	tell('saturate.log'),
	user:cd(satsrc('../examples')),
	store(reductionTime,[0,0]),
	statistics( runtime, [T1,_] ),
	F1=[ac,aci,animals,asg,concept,ded,ex5,g,g1,g2,ag,hres15,
	    ind,intp,lion_uni1,lion_uni2,loveland,
	    nat1,oc15,or,overbeek1,overbeek3,pelletier45,pelletier46,
	    pigs,pl5,pl5hg,set,sr,tarski,to,wos22],
	testset(F1,O),
	nl,
	user:cd(bledsoe),
	testset([ivt,ivt1,ivtth,jar3],O),
	nl,
	user:cd('../LocalTheories'),
	testset([bool,cong,emb],O),
%	testset([bool,cong,emb,lk],O),
	nl,
	user:cd('KD45'),
	testset([kripke],O),
	nl,
	user:cd('../../hoare'),
	testset([qsort2iia,find8b,qs2],O),
	nl,
	user:cd('../modal-logic'),
	testset([kd45],O),
	nl,
	user:cd('../ChurchRosser'),
	testset([pr_lemmaF2,pr_lemmaF3],O),
	
	(cont(long,true) ->
	     user:cd('../happy'),
	     testset([andrews3,sr,halteproblem
		     ,verbandth1],O)
	;
	     true
	),

	(file_exists('../tests/LDA013-1') ->
	    nl,
	    user:cd('../tests'),
	    testset(['LDA013-1',cat015],O)
	;
	    true
	),
	!,
	statistics( runtime, [T2,_] ),
	T is T2-T1,
	TS is T // 1000,
%	told,
	savedOn(D,Ti),
	prolog_version(P),
	format('~nTotal execution time: ~d sec.',[TS]),
	cont(reductionTime,[TR,CR]),
	TRS is TR // 1000,
	CRS is CR // 1000,
	format('~nTotal rewriting time:  ~d sec.',[TRS]),
	format('~nTotal constraint time: ~d sec.',[CRS]),
	user:saturate_version(V),
	format('~n(Executed with version ~p (~p) of ~p at ~p)~n',[V,P,D,Ti]).




testset(F,O):-
	maplist(testfile(O),F).



p1(F-_,F).

improper(F):-
	name(F,N),
	(  append(_,"defaults",N)
	;  append(_,".log",N)
	;  append(_,".o",N)
	;  append(_,".ord",N)
	;  append(_,"*end*",N)
	;  append(_,"~",N)
	;  append(_,"\%",N)
	;  append(_,".rm",N)
	;  isDirectory(F)
	).


isDirectory(F):-
	(prolog_version(sicstus) ->
	     fail
	;prolog_version(sicstus3) ->
	     use_module(library(system),[file_exists/2]),
	     file_exists(F,search)
	;
	     % that directory_property returns an exception I consider a bug
	     use_module(library(directory),
		   [directory_property/2]),
	     on_exception(permission_error(_,_,_,_,_),directory_property(F,searchable),fail) 
	).

testfile(O,F):-
	format(user_output,
	 '~n----------------------------------------------~2n',[]),
	user:insrc(F),
	user:maplist(call,O),
%	user:sams([]),
%	user:satp([6]),
%	user:satp([4,6]),
%	user:satp(all),
%	user:sarp(+[15]),   % max rewriting
%	user:sarp(+[14]),   % min rewriting
%	user:saci([2]),
%	user:cnf,
	user:sat,
	cont(reductionTime,[TR,CR]),
	cont(rewr,T1),
	cont(cons,T2),
	TR1 is TR+T1,
	CR1 is CR+T2,
	store(reductionTime,[TR1,CR1]).

spassfile(F,TimeBound,(Success,SatFlag,TimeUsed,NInf,NClGen,NAllC)):-
	spassfile(F,TimeBound,(Success,SatFlag,TimeUsed,NInf,NClGen,NAllC),spass('/tmp/spassin')).

spassfile(F,TimeBound,(Success,SatFlag,TimeUsed,NInf,NClGen,NAllC),spass(SF)):-
	!,
	format(user_output,
	 '~n----------------------------------------------~2n',[]),
	user:insrc(F),
	user:sair([7]),
	user:sarp([1,2,3,4,6,7,12,13]),
	user:sams([]), user:saes([]),
	saturation:sat,
%	commands:out_as_dfg(SF,time(TimeBound)),
	commands:out_as_dfg(SF),
	spass(SF,TimeBound,SatFlag,TimeUsed,NInf,NClGen,NAllC),
	((SatFlag=timelimit; SatFlag=exception) ->
	       Success=fail
	;
	       Success=success
	),
	!.




testfile(_ClBound,TimeBound,LogStream,F,Success,spass(SF)):-
	spassfile(F,TimeBound,(Success,SatFlag,TimeUsed,NInf,NClGen,NAllC),spass(SF)),
	portray_clause2(LogStream,result(file=F,status=SatFlag,time=TimeUsed,'#inf'=NInf,'#kept'=NClGen,'#all'=NAllC)),
	flush_output(LogStream),
	!.
testfile(ClBound,TimeBound,LogStream,F,Success,Options):-
	format(user_output,
	 '~n----------------------------------------------~2n',[]),
	user:insrc(F),
%	user:sams([]),
%	user:option(ac(on)),
	user:maplist(call,Options),
	saturation: sat(ClBound,TimeBound,SatFlag,TimeUsed,NInf,NClGen,NAllC,SDepth),
	portray_clause2(LogStream,result(file=F,status=SatFlag,time=TimeUsed,'#inf'=NInf,'#kept'=NClGen,'#all'=NAllC, '#sdepth'=SDepth)),
	flush_output(LogStream),
	((SatFlag='?'; SatFlag=exception) ->
	       Success=fail
	;
	       Success=success
%	       saturation:terminate
	),
	!.


spass(F,TimeBound,SatFlag,TimeUsed,0,0):-
	Out='/tmp/spassresult',
	appendAtoms([callspass,TimeBound,F,Out],' ',Call),
	(user:unix(shell(Call)) ->
	     open(Out,read,O),
	     read(O,result(SatFlag)),
	     read(O,time(TimeUsed)),
	     close(O)
	;SatFlag=exception).




testfiles(([F|Fs],Successes,Fails,ClBound,TimeBound,LogStream,Options)):-
	testfile(ClBound,TimeBound,LogStream,F,Result,Options),
	(Result=success ->
	     Successes1 is Successes+1,
	     Fails=Fails1
	;
	     Successes1 = Successes,
	     Fails1 is Fails+1
	),
	      
	testfiles((Fs,Successes1,Fails1,ClBound,TimeBound,LogStream,Options)).
testfiles(([],Successes,Fails,_ClBound,_TimeBound,LogStream,_Options)):-
	(Successes+Fails > 0 ->
	     Percentage is integer(Successes/(Successes+Fails)*10000)
%	     format(LogStream,'~n======== Summary =========~n#Successes= ~p (= ~2d%), #Failures= ~p',
%		    [Successes,Percentage,Fails])
        ;
		   write('Empty problem set.'),nl
        ).


% higher-order


setof0(E,P,S):-
	setof(E,P,S),
	!.
setof0(_,_,[]).




apply(M:P,A,B,C):-
	!,
	Call=..[P,A,B,C],
	M:Call.
apply(P,A,B,C):-
	Call=..[P,A,B,C],
	Call.

once(P):-
	P,
	!.

try(P):-
	(P;true),
	!.

while(P,Q):-
	(\+ \+ P ->
	     Q,
	     while(P,Q)
	;
	     true
	).

split(_,[],[],[]).
split(M:P,[X|Xs],Ys,Ns):-
	split(M:P,Xs,Ys1,Ns1),
	P=..[Pred|Args],
	Call=..[Pred,X|Args],
	
	(M:Call ->
	     Ys=[X|Ys1],
	     Ns=Ns1
	;
	     Ys=Ys1,
	     Ns=[X|Ns1]
	).
	


% auxiliary


% to avoid a bug in maplist expansion at compile-time
endOfFile(F):-
	F==end_of_file.


	     

/* some experimental stuff


demodulate_eq( E, E1, RN, RN1):-
	get_atom_relation_symbol(E,RS),
	cancellative_relation_symbol(RS,F),
	neutral_element(F,Zero),
	get_atom_members(E,S,T),
	flatten(S,F,SF),
	flatten(T,F,TF),
	remove_terms_occuring_both_sides(Zero,SF,TF,SF1,TF1),
	!,
	unflatten(SF1,F,Zero,S1),
	unflatten(TF1,F,Zero,T1),
	A=..[RS,S1,T1],
	create_atom(A,E2),
	demodulate_eq(E2,E1,RN,RN1),
	!.

flatten(S,F,SF):-
	nonvar(S),
	S=..[F,S1,S2],
	!,
	flatten(S1,F,SF1),
	flatten(S2,F,SF2),
	append(SF1,SF2,SF).
flatten(S,_,[S]).


unflatten([],_F,Zero,Zero):-
	!.
unflatten([T],_F,_Z,T):-
	!.
unflatten([T|Ts],F,Zero,FT):-
	unflatten(Ts,F,Zero,FTs),
	FT=..[F,T,FTs].

remove_terms_occuring_both_sides(Zero,[Y],[X],_S1,_S2):-
	X==Zero,
	Y==Zero,
	!,
	fail.
remove_terms_occuring_both_sides(Zero,L1,L2,S1,S2):-
	member(T,L1),
	delete_one312(T,L2,L21),
	delete_one312(T,L1,L11),
	!,
	(      remove_terms_occuring_both_sides(Zero,L11,L21,S1,S2)
	;      S1=L11, S2=L21),
	!.


*/




time(Goal,T):-
        statistics(runtime,[T1,_]),
	Goal,
	statistics(runtime,[T2,_]),
	T is ((T2-T1)//100)/10.
