:- module(index,[
		empty_index/1,
		delete_terms/3,
		initial_index/1,
		initial_index_from/3,
		initial_index/2,
		index/3,
		index/5,
		insert_terms/3,
		insertSimplifierIntoIndex/4,
		retrieve_match/3
		]).




:-use_module(substitutiontree).
:-use_module(proofState,[rule/2,setRewriteFlag/1]).
:-use_module(equationalTheory,[currentTheory/1]).
:-use_module(classes,[indexEntries/2]).
:-use_module(library(basics),[member/2]).
:-use_module(library(change_arg),[change_arg/5]).
:-use_module(inference_rules,[do_unifications/1]).
:-use_module(achelp,[detach_nonAC_head/3]).
:-use_module(orderConstraints,[true/1]).
:-use_module(props,[firstOrder/0]).


empty_index(index(ST1,ST2,ST3,ST4,ST5,ST6)):-
	st_empty_tree(ST1),
	st_empty_tree(ST2),
	st_empty_tree(ST3),
	st_empty_tree(ST4),
	st_empty_tree(ST5),
	ST6=t(t,[]).                   % empty association list + clause list



%1
index([orientable,unconditional],I,ST):-
	arg(1,I,ST).
index(orientableUnits,I,ST):-
	arg(1,I,ST).

%2
index([nonorientable,unconditional],I,ST):-
	arg(2,I,ST).
index(nonorientableUnits,I,ST):-
	arg(2,I,ST).

%3
index([orientable,conditional],I,ST):-
	arg(3,I,ST).
index(reductiveReducers,I,ST):-
	arg(3,I,ST).

%4
index([nonorientable,conditional],I,ST):-
	arg(4,I,ST).
index(nonreductiveReducers,I,ST):-
	arg(4,I,ST).

%5
index(subsumption,I,ST):-
	arg(5,I,ST). 

%6
index(prop_assoc,I,ST):-
	arg(6,I,ST). 



% updating 



index([orientable,unconditional],I,ST,I1,ST1):-
	change_arg(1,I,ST,I1,ST1).


index([nonorientable,unconditional],I,ST,I1,ST1):-
	change_arg(2,I,ST,I1,ST1).

index([orientable,conditional],I,ST,I1,ST1):-
	change_arg(3,I,ST,I1,ST1).

index([nonorientable,conditional],I,ST,I1,ST1):-
	change_arg(4,I,ST,I1,ST1).

index(subsumption,I,ST,I1,ST1):-
	change_arg(5,I,ST,I1,ST1). 

index(prop_assoc,I,ST,I1,ST1):-
	change_arg(6,I,ST,I1,ST1). 






% intialization

initial_index(I):-
	findall(E,
		(rule(_,Ru),
		 indexEntries(Ru,Entries),
		 member(E,Entries)
		),
		RuleEntries),
	initial_index(RuleEntries,I).


initial_index_from(Rs,I,RuleEntries):-
	findall(E,
		(member(N,Rs),
		 rule(N,Ru),
		 indexEntries(Ru,Entries),
		 member(E,Entries)
		),
		RuleEntries),
	initial_index(RuleEntries,I).


initial_index(RuleEntries,I):-
	empty_index(I0),
	insert_terms(RuleEntries,I0,I1),
	insertStandardSimplifiers(I1,I).

/*
-------------------------------------------------------------------------------

insertSimplifierIntoIndex(+Ru,+ST,?STO,?Entries)

is true when Entries is the list of simplifier entries for the rule
Ru and when STO results from ST by inserting the Entries into ST.

-------------------------------------------------------------------------------
*/
insertSimplifierIntoIndex(Ru,ST,STO,Entries):-
	indexEntries(Ru,Entries),
	insert_terms(Entries,ST,STO).
insertSimplifierIntoIndex(_Ru,ST,ST,[]).


insertStandardSimplifiers(ST,ST2):-
	true(Top),
	% reflexivity of ``=''
	insert_term(t([orientable,unconditional],
		      [=,X,X],
		      t('L',[==,'$true$','$true$'],'$',nonvar,Top,_)),ST,ST1),
	logical_rules(Rs),
	st_insert_logical_simplifiers(Rs,ST1,ST2).

st_insert_logical_simplifiers(_,ST,ST):-
	\+ firstOrder,
	!.
st_insert_logical_simplifiers([],ST,ST):-
		  setRewriteFlag(term_equations_exist),
		  setRewriteFlag(equations_exist).
st_insert_logical_simplifiers([(L=R,RR)|Rs],ST1,ST3):-
	true(Top),
	insert_term(t([orientable,unconditional],
		      L,
		      t('L',R,RR,nonvar,Top,_)),ST1,ST2),
	st_insert_logical_simplifiers(Rs,ST2,ST3).


logical_rules([
	      % for beta reduction the irreducibility indicator is irrelevant
	('$apply'('$lambda'(_,B),E) = '$subst'(B,E), '$'),

	((X = X)           = '$true$','$'),  % for embedded equalities
	([==,X=Y,'$true$'] = [=,X,Y], [=,'$','$']),  % for embedded equalities

	([-,==,A,'$min'] = [+,==,A,'$true$'], [=,'$','$']),  
	([-,==,'$min',A] = [+,==,A,'$true$'], [=,'$','$']), 

	([+,==,A,'$min'] = [-,==,A,'$true$'], [=,'$','$']),  
	([+,==,'$min',A] = [-,==,A,'$true$'], [=,'$','$']),
	       
	([+,==,~A,'$true$'] = [-,==,A,'$true$'], [=,'$','$']),  
	([+,==,'$true$',~A] = [-,==,A,'$true$'], [=,'$','$']),  
	([-,==,~A,'$true$'] = [+,==,A,'$true$'], [=,'$','$']),  
	([-,==,'$true$',~A] = [+,==,A,'$true$'], [=,'$','$']),  

	((X and '$true$') = X,'$'),
	(('$true$' and X) = X,'$'),
	((X and X)        = X,'$'),
	((X and '$min')   = '$min','$'),
	(('$min' and X)   = '$min','$'),

        (('$all'(S,X => Y) and '$all'(_,Y => X))
	         = '$all'(S,X <=> Y),'$all'('$','$' <=> '$')),	% very very special
        (('$all'(S,'$all'(S1,X => Y)) and '$all'(_,'$all'(_,Y => X)))
	         = '$all'(S,'$all'(S1,X <=> Y)),'$all'('$','$' <=> '$')), % very special
	

%	((X == '$true$')  = X,'$'),
%	(('$true$' == X)  = X,'$'),
%	((X == '$min')    = ~(X),~('$')),
%	(('$min' == X)    = '$min','$'),
%	((X == X)         = '$true$','$'),

	((X = X)         = '$true$','$'),

	((X <=> '$true$')  = X,'$'),
	(('$true$' <=> X)  = X,'$'),
	((X <=> '$min')    = ~(X),~('$')),
	(('$min' <=> X)    = '$min','$'),
	((X <=> X)         = '$true$','$'),
	((X <=> ~X)        = '$min','$'),
	((~X <=> X)        = '$min','$'),

	(('$true$' or X)  = '$true$','$'),
	((X or '$true$')  = '$true$','$'),
	((X or '$min')    = X,'$'),
	(('$min' or X)    = X,'$'),
	((X or X)         = X,'$'),

	(('$true$' => X)  = X,'$'),
	((X => ~(X))      = ~(X),'$'),
	((X => '$true$')  = '$true$','$'),
	((X => '$min')    = ~(X),~('$')),
	(('$min' => X)    = '$true$','$'),
	((X => X)         = '$true$','$'),
	((~ ~X)           = X,'$'),
	((~X => X)        = X,'$'),
	((X => ~X)        = ~(X),~('$')),

	(~('$true$')      = '$min', '$'),
	(~('$min')        = '$true$', '$'),


	('$all'(_,'$true$')    = '$true$', '$'),
	('$all'(_,'$min')      = '$min', '$'),
	('$exists'(_,'$true$') = '$true$', '$'),
	('$exists'(_,'$min')   = '$min', '$'),
	('$exists'(_,'i$1'=_)  = '$true$', '$'),
	('$exists'(_,_='i$1')  = '$true$', '$'),
	       
	(~'$exists'(X,~P)      = '$all'(X,P), '$all'('$','$')),
	([P,==,'$exists'(_,'i$1'=_),E]
	                       = [P,==,'$true$', E],  [=,'$','$']),
	([P,==,'$exists'(_,_='i$1'),E]
	                       = [P,==,'$true$', E],  [=,'$','$']),
	([-,==,'$exists'(X,~P),'$true$']
	                       = [+,==,'$all'(X,P),'$true$'], [=,'$','$']),
	([+,==,'$exists'(X,~P),'$true$']
	                       = [-,==,'$all'(X,P),'$true$'],[=,'$','$']),
	       

	(~'$all'(X,~P)         = '$exists'(X,P), '$exists'('$','$')),
	([-,==,'$all'(X,~P),'$true$']
	                       = [+,==,'$exists'(X,P),'$true$'],[=,'$','$']),
	([+,==,'$all'(X,~P),'$true$']
	                       = [-,==,'$exists'(X,P),'$true$'], [=,'$','$'])
/*	      ,

	([P,=<,s(X),s(Y)]      = [P,=<,X,Y], [=,'$','$']),
	([P,=<,p(X),p(Y)]      = [P,=<,X,Y], [=,'$','$']),
	([P,<,s(X),s(Y)]      = [P,<,X,Y], [=,'$','$']),
	([P,<,p(X),p(Y)]      = [P,<,X,Y], [=,'$','$'])
	      */


	      ]).




% insertion

insert_term(t(Index,Key,Info),ST,STO):-
	index(Index,ST,STI,STO,STIO),
	(currentTheory(empty) ->
	     st_insert_term(Key,Info,STI,STIO)
	;
	     detach_nonAC_head(Key,NonAC,SubtermAssociation),
	     st_insert_term(NonAC,t(SubtermAssociation,Info),STI,STIO)
	).


insert_terms([T|Entries],ST,STO):-
	insert_term(T,ST,ST1),
	insert_terms(Entries,ST1,STO).
insert_terms([],ST,ST).
	

% deletion

delete_terms([t(Index,Key,Info)|Entries],ST,STO):-
	index(Index,ST,STI,ST1,STI1),
	(currentTheory(empty) ->
%	     (arg(1,Info,132) -> trace; true),
	     st_delete_term(Key,STI,STI1,rewrite:equalNumber(Info))
%	     ,(arg(1,Info,132) -> STI==STI1, write('*** equal'),nl; true)
	;
	     detach_nonAC_head(Key,NonAC,_),
	     st_delete_term(NonAC,STI,STI1,index:equalNum(Info))
	     
	),
	delete_terms(Entries,ST1,STO).
delete_terms([],ST,ST).


% retrieval
% this may instantiate variables in the index, hence should be used
% only in the context of a findall

retrieve_match(Pattern,Index,Info):-
	(currentTheory(empty) ->
	     st_us_unify(Pattern,Index,Info)
	;
	     st_us_unify(Pattern,Index,t(SubtermAssoc,Info)),
	     do_unifications(SubtermAssoc)
	).


retrieve_one_match(Pattern,Index,Info):-
	retrieve_match(Pattern,Index,Info),
	!.



% private

equalNum(Info,t(_,I)):-
	rewrite:equalNumber(Info,I).




