%declarations%

:- module(inference_computation, [ 
        builtin_inferences_with/2,
	explicit_inferences/2,
	simplifying_inference/5,
	treat_inference/7
   ]).

%:- use_module(user,_,[append/3]).

:- use_module(props, [
	ac_ist_C/1,
	flag_on/2,
	horn/0
   ]).

:- use_module(achelp, [
	ac_ist_AC/1
   ]).

:- use_module(equationalTheory, [
	currentTheory/1
   ]).

:- use_module(proofState, [recordClause/1]).


:- use_module(constraintSolver, [
	simplify/2,
	simplify_constraint_vars/3,
	simplify_light/2
   ]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(classes, [
	dont_simplify/1,
	empty/1,
	complexityBound/3,
	vars/2,
	constraint/2,
	constraint/3,
	is_trivial/1,
	redundancy_info/3,
	history_and_number/3,
	kind/2,
	number/2
   ]).

:- use_module(adt_declarations, [
	message/2
   ]).


:- use_module(inference_rules, [
	do_unifications/1
   ]).

:- use_module(props, [
	congruence_relation_symbol/1,
	dense_ordering_relation_symbol/1,
	dense_ordering_wo_end_relation_symbol/1,
	partial_congruence_relation_symbol/1,
	partial_equivalence_relation_symbol/1,
	total_ordering_relation_symbol/1,
	transitive_relation_symbol/1,
	reflexive_closure/2
   ]).

:- use_module(orderConstraints, [
	and/3,
	and_eq/3,
	false/1,
	true/1
   ]).

:- use_module(timing, [
	increase_inference_count/1,
	push_timing/1,
	pop_timing/1
   ]).



%end%


%						INFERENCE COMPUTATION


explicit_inferences(_Rs,_Key):-
	horn,
	!.
explicit_inferences(Rs,Key):-
	\+ builtin_inferences(Rs,Key).

builtin_inferences(Rs,Key):-
	member(R,Rs),
	kind(R,N),
	builtin_inferences_with(N,Key).
builtin_inferences([R1,R2],_Key):-
	number(R1,N),
	number(R2,N),
	kind(R1,monotonicity(_,_,RS,_,RS)),
	transitive_relation_symbol(RS).
builtin_inferences([R,R],chri):-
	kind(R,finite_domain(_,RS)),
	congruence_relation_symbol(RS).
	



builtin_inferences_with(composition(RS1,RS2,RS3),_Key):-
        transitive_relation_symbol(RS1),
	transitive_relation_symbol(RS2),
	transitive_relation_symbol(RS3).
builtin_inferences_with(monotonicity(_,_,RS,_,RS),_Key):-
	partial_congruence_relation_symbol(RS).
builtin_inferences_with(monotonicity(_,_,RS,_,==),_Key):-
	partial_equivalence_relation_symbol(RS).
builtin_inferences_with(symmetry(_RS),_Key).
builtin_inferences_with(anti_symmetry(RS),_Key):-
	transitive_relation_symbol(RS).	
builtin_inferences_with(density(RS,_,strict),_Key):-
	dense_ordering_relation_symbol(RS).
builtin_inferences_with(density(RS,_,nonstrict),_Key):-
	dense_ordering_relation_symbol(R),
	reflexive_closure(R,RS).
builtin_inferences_with(reflexive_closure1(RS,_),_Key):-
	total_ordering_relation_symbol(RS).
builtin_inferences_with(reflexive_closure2(RS,_),_Key):-
	total_ordering_relation_symbol(RS).
builtin_inferences_with(noendL(RS),_Key):-
	dense_ordering_wo_end_relation_symbol(RS).
builtin_inferences_with(noendR(RS),_Key):-
	dense_ordering_wo_end_relation_symbol(RS).
builtin_inferences_with(totality(RS,Eq),Key):-
	Key\==orre, 
	transitive_relation_symbol(RS),
	partial_congruence_relation_symbol(Eq).
builtin_inferences_with(irreflexivity(RS),sncr):-
	transitive_relation_symbol(RS).
builtin_inferences_with(reflexivity(RS),chri):-
	transitive_relation_symbol(RS).
builtin_inferences_with(reflexivity(RS),orre):-
	transitive_relation_symbol(RS).
builtin_inferences_with(commutativity(O,_R),_):-
	currentTheory(ac),
	ac_ist_C(O).
builtin_inferences_with(associativity(O,_Y),_):-
	currentTheory(ac),
	ac_ist_AC(O).








simplifying_inference(skex, [R], '$end', _, R).
simplifying_inference(orre, [R,R1], _, _, R1):-
	flag_on(rp,tore),   % only if totality resolution is turned on
	   % as a simplification we check as to whether the orre inference
	   % was in fact a resolution with totality
	kind(R,totality(RS,Eq)),
	transitive_relation_symbol(RS),
	partial_congruence_relation_symbol(Eq),
	kind(R1,N1),
	\+ dont_simplify(N1).
simplifying_inference(orre, [R,R1], _, [[X,Y]], R):-
	kind(R1,irreflexivity(_RS)),
	X==Y,
	kind(R,N),
	\+ dont_simplify(N).
simplifying_inference(rere, [R], _, [[X,Y]],R):-
	X==Y,
	!.
simplifying_inference(fact, [R], _, [[X,Y]],R):-
	fail,
	X==Y,
	!.
simplifying_inference(_,[R|_], C, _, R):-
	empty(C).


%---------------------------------
% special handling of pseudo conclusions of tableau inferences
%---------------------------------

treat_inference(_Key, _Premises, _Index, _Unif, '$end', _OrdConstraint, _CB ):-
	!.


%---------------------------------
% no constraint inheritance
%---------------------------------


treat_inference( Key, Premises, _Index, Unif, Conclusion, OrdConstraint, CB ):-
	\+flag_on( ci, orde ),
	\+flag_on( ci, equa ),
	do_unifications( Unif ),
	\+ is_trivial(Conclusion),
%	format('Constraint ~p~nfor ~p inference~nfrom ~p~2n',
%	       [OrdConstraint,Key,Premises]),
%	simplify_light( OrdConstraint, OrdConstraint1 ), 
	simplify( OrdConstraint, OrdConstraint1 ), 
%	OrdConstraint=OrdConstraint1,
	\+ false( OrdConstraint1 ),
	constraint(OrdConstraint1,Conclusion,Clause1),
	complexityBound(CB,Clause1,Clause),
%	format('simplified constraint: ~p~2n',[OrdConstraint1]),
	increase_inference_count(Key),
	assert_new_clause_and_construct_message( Key, Premises, Clause ).

%-----------------------------------
% ordering constraint inheritance
%-----------------------------------

treat_inference( Key, Premises, _Index, Unif, Conclusion, OrdConstraint, CB ):-
	flag_on( ci, orde ),
	\+ flag_on( ci, equa ),
	do_unifications( Unif ),
	\+ is_trivial(Conclusion),
	inherit_constraints( Premises, IHConstr ),
	and( IHConstr, OrdConstraint, Constraint ),
	\+ false( Constraint ),
	increase_inference_count(Key),
	constraint( Constraint, Conclusion, Clause1 ),
	complexityBound(CB,Clause1,Clause),
	assert_new_clause_and_construct_message( Key, Premises, Clause ).

%---------------------------------
% equality constraint inheritance
%---------------------------------

treat_inference( Key, Premises, _Index, Unif, Conclusion, OrdConstraint, CB ):-
	\+ flag_on( ci, orde ),
	flag_on( ci, equa ),
	inherit_constraints( Premises, IC ),
	add_unifications_to_constraint( IC, Unif, NewC ),	
	and( NewC, OrdConstraint, Check ),
	simplify_light( Check, Check1 ),
%	simplify( Check, Check1 ),
	\+ false( Check1 ),
	increase_inference_count(Key),
	vars( Conclusion, LV ),
	simplify_constraint_vars( LV, NewC, NewC1 ),
	constraint( NewC1, Conclusion, Clause1 ),
	\+ is_trivial(Clause1),
	complexityBound(CB,Clause1,Clause),
	assert_new_clause_and_construct_message( Key, Premises, Clause ).       

%---------------------------------------------
% ordering & equality constraint inheritance
%---------------------------------------------

treat_inference( Key, Premises, _Index, Unif, Conclusion, OrdConstraint, CB ):-
	flag_on( ci, orde ),
	flag_on( ci, equa ),
	inherit_constraints( Premises, IHConstr ),
	and( IHConstr, OrdConstraint, Constr ),
	add_unifications_to_constraint( Constr, Unif, Constraint ),
	simplify( Constraint, Constraint1 ),
%	Constraint=Constraint1,
	\+ false( Constraint1 ),
	increase_inference_count(key),
	constraint( Constraint1, Conclusion, Clause1 ),
	\+ is_trivial(Conclusion),
	complexityBound(CB,Clause1,Clause),
	assert_new_clause_and_construct_message( Key, Premises, Clause ).



add_unifications_to_constraint( Constr, [], Constr ):- !.
add_unifications_to_constraint( Constr, [[S,T]|Rest], Constraint ):-
	add_unifications_to_constraint( Constr, Rest, Constr1 ),
	and_eq( Constr1, [S,T], Constraint ), !.

 
%----------------------------------
% inference constraint inheritance
%----------------------------------

inherit_constraints( [], Top ):-
	true( Top ), !.
inherit_constraints( [R|Rest], Constraint ):-
	inherit_constraints( Rest, Constr1 ),
	constraint( R, Constr2 ),
	and( Constr1, Constr2, Constraint ), !.

%------------------------------------
% new clause and message construction
%------------------------------------

assert_new_clause_and_construct_message( Key, Premises, Conclusion ):-
	get_premises_numbers( Premises, NL ),
	history_and_number( [Key, NL, Premises], Conclusion, Cl ),
%	redundancy_info((noTaut,_),Cl, Clause),   % no! 
	Cl=Clause,
	recordClause( Clause ),
	append( Premises, [Clause], L ),
	message( Key, L ), !.

get_premises_numbers( [], [] ):- !.
get_premises_numbers( [P|R], [N|L] ):-
	number( P, N ),
	get_premises_numbers( R, L ), !.


	    

	
% eof ______________________________________________________________

