% -*- Module: inference_rules; -*-

%						INFERENCE_RULES
%declarations%

:- module(inference_rules, [ 
	constraint_greater_clause/3,
	constraint_grequal_clause/3,
	constraint_grequal_clause/4,
	constraint_grequal_literal/5,
	constraint_grequal_literal_in/4,
	constraint_greater_literal/3,
	constraint_greater_literal/5,
	constraint_greater_literal_in/4,
	do_unifications/1,
	inference/7
   ]).

:- use_module(props, [
	flag_on/2,
	full_cnf/1,
	horn/0
   ]).

:- use_module(equationalTheory, [
	unifiable/2,
	unify/2
   ]).

:- use_module(adt_term, [
	get_term_vars/2
   ]).


:- use_module(adt_atom, [
	atom_as_rule/3,
	atom_as_substitution/1,
	create_atom/2,
	equation_as_atom/2,
	equivalent_relation_symbols/2,
	get_atom_members/3,
	get_atom_preorientation/2,
	get_atom_relation_symbol/2,
	get_atom_pre_rel_left_right/5,
	get_binary_atom_pre_rel_left_right/5,
	get_equation_members/3,
	is_standard_atom/1,
	is_positive_standard_atom/1,
	is_equivalence/1,
	is_term_equation/1,
	is_proposition/1,
	greater_relation_symbol/2,
	orient_by_polarity/5,
	chain/8,
	factorization/7,
	negative_chain/7,
	resolution/3,
	residual/4
   ]).

:-use_module(sequents, [
	tableau_expansion_on/3
   ]).

:- use_module(classes, [
	add_antecedent/3,
	add_succedent/3,
	add_list_antecedent/3,
	add_list_succedent/3,
	antecedent/2,
	createSimpleClause/2,
	createClause/2,
	groundClause/1,
	literals/2,
	succedent/2,
	selectInAntecedent/3,
	selectInSuccedent/3,
	selectMaxInAntecedent/3,
	selectMaxInSuccedent/3,
	selectSelectedLiteral/3,
	kind/2,
	maximal_antecedent/2,
	maximal_succedent/2,
	non_maximal_succedent/2,
	number/2,
	hasSelectedLiteral/1,
	merge_two_rules_into_clause/3,
	maximal_succedent/3,
	non_maximal_succedent/3,
	propositionalClause/1,
	theory_flag/2,
	to_clause/2
   ]).


:- use_module(auxiliary, [
	delete32/3
   ]).

:- use_module(library(basics), [
	member/2
   ]).

%:- use_module(user, _, [
%	append/3
%   ]).

:- use_module(props, [
        congruence_relation_symbol/1,
	composition_of_relations/3,
	needs_trre_relation_symbol/1,
	partial_equivalence_relation_symbol/1,
	anti_symmetric_relation_symbol/1,
	reflexive_relation_symbol/1,
	special_relation/1,
	transitive_relation_symbol/1,
	strict_part/2,
	symmetric_relation_symbol/1,
	total_ordering_relation_symbol/1
   ]).

:- use_module(orderConstraints, [
	and/3,
	and_eq/3,
	and_ge/3,
	and_gt/3,
	and_list/2,
	false/1,
	not/2,
	or/3,
	or_eq/3,
	or_list/2,
	project/3,
	true/1
   ]).

:- use_module(constraintSolver, [
        simplify/2
   ]).

:- use_module(timing, [
	push_timing/1,
	pop_timing/1
   ]).

:- use_module(hgstuff, [
	apply/4
   ]).


%end%

%------------------------------------------------------------------------------
%
%       tableau expansion
%
%------------------------------------------------------------------------------


inference( skex, [R], _STree, [], Conclusion, Top, ComplexityBound ):-
	% expansion right
	number(R,N),
	(user:spy_clause(N) -> trace; true),
	(full_cnf(no)->
	    \+ hasSelectedLiteral( R ),
	    selectMaxInSuccedent( R, E, Ra )
	;
	    selectInSuccedent(R,E,Ra)
	    ),
	\+ is_positive_standard_atom(E),
	antecedent(Ra,A),
	succedent(Ra,S),
	tableau_expansion_on([+|E],(A,S),Ce),
	!,
	true(Top),

	createSimpleClause([[],[E]],ComplexityBound),
	(   member(C,Ce),
	    inference_conclusion( skex, C, Conclusion )

	;   Conclusion='$end'
	).
inference( skex, [R], _STree, [], Conclusion, Top, ComplexityBound ):-
	% expansion left
	number(R,N),
	(user:spy_clause(N) -> trace; true),
	(full_cnf(no)->
	    (hasSelectedLiteral( R ) ->
		selectSelectedLiteral( R, E, Ra ),
		ComplexityBound=R
	    ;    
		selectMaxInAntecedent( R, E, Ra ),
		createSimpleClause([[E],[]],ComplexityBound)
	    )
	;
	    selectInAntecedent(R,E,Ra),
	    (	is_equivalence(E)
	    ;
		\+ is_standard_atom(E)
	    ),
	    createSimpleClause([[E],[]],ComplexityBound)
	),
	antecedent(Ra,A),
	succedent(Ra,S),
	tableau_expansion_on([-|E],(A,S),Ce),
	!,
	true(Top),
	(   member(C,Ce),
	    inference_conclusion( skex, C, Conclusion )

	;   Conclusion='$end'
	).



%------------------------------------------------------------------------------
%
%       superposition + chaining right
%------------------------------------------------------------------------------

% Complexity bounds are in most cases just the maximal literal of
% the second premise. In some cases we take the entire premise.

inference( chri, [R1,R2], STree, Unif, Conclusion, ConstraintC, ComplexityBound ):-
	\+ hasSelectedLiteral( R1 ),
	\+ hasSelectedLiteral( R2 ),
        selectMaxInSuccedent( R1, E1, Ra ),
        selectMaxInSuccedent( R2, E2, Rb ), 
	number(R1,N1),
	number(R2,N2),

%	(N1=14,N2=20 -> trace; true),
	(N1=<N2 ->       % this may be incompatible with the check E1<E2
%	     trace,
	     chain( E1, E2, [Pol1|NE1], [Pol2|NE2], E, Unif, Pos, STree ),
	     finite_domain_pos(R1,E2,Pos)
	;    
	     chain( E1, E2, [Pol1|NE1], [Pol2|NE2], E, Unif, [_|_], STree)
	),
	inference_conclusion( chri, Ra, Rb, Unif, [Pol1|NE1], [Pol2|NE2], E, Conclusion ),
 	inference_ordering_constraint( chri, Pos, NE1, Ra, NE2, Rb, Constraint ),
	apply_unifier(Constraint,ConstraintC),
	createSimpleClause([[],[E2]],ComplexityBound).



%------------------------------------------------------------------------------
%
%       superposition left and negative chaining
%------------------------------------------------------------------------------

inference( sncr, [R1,R2], _, Unif, Conclusion, ConstraintC, ComplexityBound ):-
	(hasSelectedLiteral( R1 ) ->
	     horn1
	;
	     true
	),
	\+ (flag_on(rp,prop),propositionalClause(R1),propositionalClause(R2)),
	\+ (flag_on(rp,prop),groundClause(R1),groundClause(R2)), % incomplete even with positive strategy
	(horn1 ->
	     succedent(R2,[]),
%	     theory_flag(R2,goal),
%	     theory_flag(R1,theory),
	     selectInSuccedent(R1,E1,Ra)
	;
	     selectMaxInSuccedent( R1, E1, Ra )
	),
	(hasSelectedLiteral( R2 ) ->
	     selectSelectedLiteral( R2, E2, Rb ),
	     ComplexityBound=R2
	;    
	     selectMaxInAntecedent( R2, E2, Rb ),
	     createSimpleClause([[E2],[]],ComplexityBound)
	),

	(is_proposition(E1),
	 is_proposition(E2) 	->
	     % no self-resolution needed
	     number(R1,N1),
	     number(R2,N2),
	     N1\==N2
	;
	     true
	),

	negative_chain( E1, E2, NE1, NE2, E, Unif, Pos ),
	finite_domain_pos(R1,E2,Pos),
	inference_conclusion( sncr, Ra, Rb, Unif, E, Conclusion ),
	inference_ordering_constraint( sncr, NE1, Ra, NE2,Rb,Constraint ),
	apply_unifier(Constraint,ConstraintC).


%------------------------------------------------------------------------------
%
%       ordered resolution
%------------------------------------------------------------------------------
/*  unrestricted unit resolution
% doesn't work; needs resolution/3 to be less restrictive.

inference( orre, [R1,R2], _, Unif, Conclusion, Constraint, ComplexityBound ):-
	(literals(R1,[[+|E1]]);literals(R2,[[-|E2]])),
	selectInSuccedent(R1,E1,Ra),
	selectInAntecedent(R2,E2,Rb),
	ComplexityBound=R2,
        resolution( E1, E2, Unif ),
	merge_two_rules_into_clause( Ra, Rb, Conclusion ),
	write('&'),
	true(Constraint).
*/
inference( orre, [R1,R2], _, Unif, Conclusion, Constraint, ComplexityBound ):-
	(hasSelectedLiteral( R1 ) ->
	     horn1
	;
	     true
	),
	(horn1 ->
	     succedent(R2,[]),
	     selectInSuccedent(R1,E1,Ra)
	;
	     selectMaxInSuccedent( R1, E1, Ra )
	),
	
	(hasSelectedLiteral( R2 ) ->
	     selectSelectedLiteral( R2, E2, Rb ),
	     ComplexityBound=R2
	;    
	     selectMaxInAntecedent( R2, E2, Rb ),
	     createSimpleClause([[E2],[]],ComplexityBound)
	),

        resolution( E1, E2, Unif ),
	merge_two_rules_into_clause( Ra, Rb, Conclusion ),

	inference_ordering_constraint( orre, E1, Ra, E2,Rb,Constraint ).




%------------------------------------------------------------------------------
%
%      factoring
%------------------------------------------------------------------------------

inference(fact, [R], _,  Unif, Conclusion, ConstraintC, R) :-
	\+ hasSelectedLiteral( R ),
        selectMaxInSuccedent( R, E1, Ra ),
        selectInSuccedent( Ra, E2, Rb ),
	get_atom_relation_symbol(E1,RS),
	get_atom_relation_symbol(E2,RS),
	factorization( E1, E2, NE1, NE2, Ant, Suc, Unif ),
	inference_conclusion( fact,  Ant, Suc, Rb, Conclusion ),
	inference_ordering_constraint( fact, NE1, NE2, Ra, Constraint ),
	apply_unifier(Constraint,ConstraintC).



%------------------------------------------------------------------------------
%
%       reflexivity resolution
%------------------------------------------------------------------------------

inference( rere, [R], _, Unif, Conclusion, Constraint, ComplexityBound ):-
	\+ hasSelectedLiteral( R ),
	selectMaxInAntecedent( R, E, Ra ),
	reflexivity_resolution( E, Unif ),
	inference_conclusion( rere, Ra, Conclusion ),
	inference_ordering_constraint( rere, E, Ra, Constraint ),
	createSimpleClause([[E],[]],ComplexityBound).

inference( rere, [R], _, Unif, Conclusion, Constraint, R ):-
	hasSelectedLiteral( R ), !,
	selectSelectedLiteral( R, E, Ra ),
	reflexivity_resolution( E, Unif ),
	inference_conclusion( rere, Ra, Conclusion ),
	marked_inference_ordering_constraint( rere, E, Ra, Constraint ).

%------------------------------------------------------------------------------
%
%       transitivity resolution
%------------------------------------------------------------------------------
/* transitivity resolution is only needed is the following case
   in which we have two clauses 
   C,r1(s,u),r2(s,v)  <    D,r3(s,t)
   for some maximal term s, 

   r1(s,u) > r2(s,v) and  r3;r4 \subseteq r2, for some relation r2.

   For this to be possible, r1=r2=r3. This is a consequence of how
   we define the ordering on literals and the fact that we restrict ourselves
   to composition laws ri;rj \subseteq rmax(i,j) and ri>rj iff i>j.
   Moreover, residual(r3,r4,r2) must hold.
*/

inference( trre, [Ru1,Ru2], _, [[La,Lb],[La,L2]], Conclusion, Constraint, Ru1 ):-
	\+ hasSelectedLiteral( Ru1 ),
	\+ hasSelectedLiteral( Ru2 ),
	number(Ru1,N1),
	number(Ru2,N2),
	N1\==N2,   % the case N1=N2 is a special case of factorization
	selectMaxInSuccedent(Ru1, Ea, R11),
	atom_as_rule(Ea,[Pa,RS,La,Ra],Pol),
        needs_trre_relation_symbol(RS),
	selectMaxInSuccedent(R11, Eb, R12),
	selectMaxInSuccedent(Ru2, E2, R21),
	atom_as_rule(Eb,[Pb,RS,Lb,Rb],Pol),
	atom_as_rule(E2,[P2,RS,L2,R2],Pol),
        \+ \+ (unify(La,Lb), unify(La,L2)),
	orient_by_polarity(Pol,R2,Rb,L,R),
	orient_by_polarity(Pol,L2,Rb,U,V),
	residual(Pol,RS,RS1,RS),
	inference_ordering_constraint( trre, [Pa,RS,La,Ra], [Pb,RS,Lb,Rb], R12, [P2,RS,L2,R2], R21, Constraint),
        inference_conclusion( trre, [n,RS1,L,R], [n,RS,U,V], R21, Conclusion ).







%------------------------------------------------------------------------------
%       possible superpositions, factorizations, 
%	equality resolutions, merging paramodutations,
%	ordered factorizations
%------------------------------------------------------------------------------

	

reflexivity_resolution( E, [[S,T]] ):-
	get_atom_relation_symbol(E,RS),
	reflexive_relation_symbol(RS),
	get_atom_members( E, S, T ),
	\+ ( \+ unify( S, T ) ).




%------------------------------------------------------------------------------
%       inference conclusion computation
%------------------------------------------------------------------------------
/*

We optimize the form of the conclusion:
- we approximate maximal chaining
- if the critical pair is a substitution x=t, x not in t,
  we can safely apply it as even if t>x, t will be smaller than the peak

*/



%inference_conclusion( chri, Ra, Rb, [_,_,RS11|_],[_,_,RS22,Peak,_],E, Conclusion ):-
inference_conclusion( chri, Ra, Rb, Unif, [Pol1,_,RS11,L,R],[_,_,RS22,Peak,_],E, Conclusion ):-

       (flag_on( ci, equa ) ->
	    Rb=Rbr
       ;
	    do_unifications(Unif),
	    reduce_succedent_by_inequation(Rb,[Pol1,RS11,L,R],Rbr) % approximation
       ),
                  % of maximal chaining

       merge_two_rules_into_clause( Ra, Rbr, Clause ),
       get_atom_relation_symbol(E,RS),

       (
            add_succedent(E,Clause,Conclusion)
       ;
            (anti_symmetric_relation_symbol(RS),
	     RS11=RS, 
	     RS22=RS ->
	         get_atom_members(E,S,T),
		 \+ \+ unify(S,T),
		 create_atom(Peak=T,EP),
		 create_atom(S=T,E1),
		 (\+flag_on( ci, equa), atom_as_substitution(E1) ->
		      Cl=Clause
		 ;
		      add_antecedent(E1,Clause,Cl)
		 ),
		 add_succedent(EP,Cl,Conclusion)
	    )
       ).

inference_conclusion( sncr, Ra, Rb, Unif, E, Conclusion ):-
        merge_two_rules_into_clause( Ra, Rb, Clause ),
	(flag_on( ci, equa ) ->
	    add_antecedent( E, Clause, Conclusion )
	;
	    do_unifications(Unif),
	    (atom_as_substitution(E) ->
		 Conclusion=Clause
	    ;
	         add_antecedent( E, Clause, Conclusion )
	    )
	),
	!.



inference_conclusion( fact, Ant, Suc, R, Conclusion ):-
	!,
	to_clause(R,C),
	add_list_succedent( Suc, C, Clause1),
	add_list_antecedent( Ant, Clause1, Conclusion ),
	!.

inference_conclusion( mgpa, R1, R2, E, Conclusion ):-
	merge_two_rules_into_clause( R1, R2, Clause ),
        add_succedent( E, Clause, Conclusion ), !.

inference_conclusion( trre, E1, E2, R, Conclusion ):-
	to_clause(R,C),
	add_antecedent(E1,C,C1),
	add_succedent(E2,C1,Conclusion),
	!.


inference_conclusion( rere, R, Conclusion ):-
        to_clause( R, Conclusion ), !.


inference_conclusion( skex, (A,S), Conclusion ):-
	createClause([A,S],Conclusion),
	!.


inference_conclusion( ofact, R, E, Conclusion ):-
        to_clause( R, Clause ),
        add_succedent( E, Clause, Conclusion ), !.


% -------------------------------------------------------------
%  reduce_succedent_by_inequation(Rb,[Pol,RS,L,R],Rr)
% -------------------------------------------------------------
reduce_succedent_by_inequation(R,[Pol,RS|E],Rr):-
	total_ordering_relation_symbol(RS1),
	strict_part(RS,RS1),
	maximal_succedent(R,SM),
	reduce_ineqs(SM,[Pol,RS|E],SMr),
	maximal_succedent(SMr,R,R1),
	non_maximal_succedent(R,SN),
	reduce_ineqs(SN,[Pol,RS|E],SNr),
	non_maximal_succedent(SNr,R1,Rr),
	!.
reduce_succedent_by_inequation(R,_,R).


reduce_ineqs([E|S],[Pol1,RS1,S11,T11],[E1|Sr]):-
%	copy_term((S11,T11),(S1,T1)),
	atom_as_rule(E,[_,RS2,S2,T2],Pol2),
	Pol1 is -Pol2,
	composition_of_relations( RS1, RS2, CRS ),
	(   total_ordering_relation_symbol(CRS) % reduce in the strict case
	;   partial_equivalence_relation_symbol(RS1)),
	S11==S2, % match(S1,S2) ,
%	write('+'),print(S11),write('+'),
	orient_by_polarity(Pol2,T11,T2,L,R),
	P=..[CRS,L,R],
	create_atom(P,E1),
	reduce_ineqs(S,[Pol1,RS1,S11,T11],Sr),
	!.
reduce_ineqs([E|S],E1,[E|Sr]):-
	reduce_ineqs(S,E1,Sr).
reduce_ineqs([],_,[]).


apply_unifier(Constraint,ConstraintC):-
	(flag_on(ci,orde) ->
%	     (length(Constraint,K),K>1 -> format('~n~p',[K]);true),
	     member(conj(InEq,Eq,Ot),Constraint),
	     do_unifications(Eq),
	     ConstraintC=[conj(InEq,[],Ot)]
	;
	     ConstraintC=Constraint
	).

/*  % probably incomplete
finite_domain_pos(R1,E2,Pos):-
	(kind(R1,finite_domain(_,RS)), congruence_relation_symbol(RS) ->
	     (is_term_equation(E2) ->
		  Pos=[]
	     ;
	          Pos=[_]
	     )
	;
	     true
	).
*/

finite_domain_pos(_,_,_).



%------------------------------------------------------------------------------
%       unification treatment
%------------------------------------------------------------------------------

do_unifications( [] ).
do_unifications( [[S,T]|Rest] ):-
	unify( S, T ),
	do_unifications( Rest ).
do_unifications( [S=T|Rest] ):-
	unify( S, T ),
	do_unifications( Rest ).
do_unifications( [S-T|Rest] ):-
	unify( S, T ),
	do_unifications( Rest ).

%------------------------------------------------------------------------------
%       inference ordering restrictions constraint computation
%------------------------------------------------------------------------------

inference_ordering_constraint(_,_,_,_,_,_,T):-
	\+flag_on(ci,chco),
	\+flag_on(ci,orde),
	true(T),
	!.


inference_ordering_constraint( chri, _Pos, E1, Ra, E2, Rb, Constraint ):-
	(  constraint_oriented_eq( E1, C1 ),
	   constraint_oriented_eq_into( E2, C2 ),
	   and( C1, C2, C3 ),
%	\+ false(C3),
%     	constraint_greater_superposed_eqs( _Pos, E2, E1, C4 ),
%	and( C3, C4, C5 ),
	      % we have to exclude this check as 
              % symmetries are avoided anyway by simpler checks
	   C5=C3,
	   \+ false(C5),
	   constraint_strict_max_suc( E1, Ra, C6 ),
	   and( C5, C6, C7 ),
	   \+ false(C7),
	   constraint_strict_max_suc( E2, Rb, C8 ),
	   and( C7, C8, Constraint ) ->
		    true
	;
	            false(Constraint)
	),
	!.


inference_ordering_constraint( trre, Ea, Eb, Rab, E2, R2, Constraint ):-
	constraint_oriented_eq( Ea, C1 ),
	constraint_oriented_eq( Eb, C2 ),
	constraint_oriented_eq( E2, C3 ),
     	constraint_greater_superposed_eqs( [], Ea, Eb, C4 ),
     	constraint_greater_superposed_eqs( [], E2, Ea, C7 ),
	constraint_strict_max_suc( Ea, Rab, C5 ),
	constraint_strict_max_suc( E2, R2, C6 ),
	and_list([C1,C2,C3,C4,C5,C6,C7],Constraint),
	!.


inference_ordering_constraint(_,_,_,_,_,T):-
	\+flag_on(ci,chco),
	\+flag_on(ci,orde),
	true(T),
	!.
inference_ordering_constraint( sncr, _E1, _Ra, E2, Rb, Constraint ):-
	horn1,
	!,
	constraint_oriented_eq_into( E2, C5 ),
	(hasSelectedLiteral(Rb) ->
		true(C6)
	;	constraint_max_ant( E2, Rb, C6 )),
	and( C5, C6, Constraint ),
	!.
inference_ordering_constraint( sncr, E1, Ra, E2, Rb, Constraint ):-
	constraint_oriented_eq( E1, C1 ),
	constraint_oriented_eq_into( E2, C2 ),
	and( C1, C2, C3 ),
	constraint_strict_max_suc( E1, Ra, C4 ),
	and( C3, C4, C5 ),
	(hasSelectedLiteral(Rb) ->
		true(C6)
	;	constraint_max_ant( E2, Rb, C6 )),
	and( C5, C6, Constraint ),
	!.

inference_ordering_constraint(orre,E1,R1,E2,R2,Constraint):-
	constraint_strict_max_suc(E1,R1,C1),
	(hasSelectedLiteral(R2) ->
		true(C2)
	;	constraint_max_ant( E2, R2, C2 )),
	and(C1,C2,Constraint),
	!.
	



inference_ordering_constraint(_,_,_,_,T):-
	\+flag_on(ci,chco),
	\+flag_on(ci,orde),
	true(T),
	!.
inference_ordering_constraint( fact, E1, E2, Ra, Constraint ):-
	constraint_oriented_eq( E1, C1 ),
	constraint_oriented_eq( E2, C2 ),
	and( C1, C2, C3 ),
	constraint_max_suc( E1, Ra, C4 ),
	and( C3, C4, Constraint ),
	!.

inference_ordering_constraint(_,_,_,T):-
	\+flag_on(ci,chco),
	\+flag_on(ci,orde),
	true(T),
	!.

inference_ordering_constraint( rere, E, R, Constraint ):-
	constraint_max_ant( E, R, Constraint ),
	!.


% -------------------------------------------------------------------

project_constraint(Cl,Ct1,Ct):-
	antecedent(Cl,A),
	succedent(Cl,S),
	get_term_vars((A,S),V),
	project(V,Ct1,Ct).



%--------------------------------------------------------------------------------
%       marked case: inference ordering restrictions constraint computation
%--------------------------------------------------------------------------------



marked_inference_ordering_constraint(orre,E1,R1,_E2,_R2,C1):-
	constraint_strict_max_suc(E1,R1,C1),
	!.


marked_inference_ordering_constraint( rere, _, _, Constraint ):-
	true( Constraint ).

marked_inference_ordering_constraint_sncr( P, E1, Ra, E2, _Rb, Constraint ):-
	constraint_oriented_eq( P, E1, C1 ),
	constraint_oriented_eq( P, E2, C2 ),
	and( C1, C2, C3 ),
	equation_as_atom(E1,A1),
	constraint_strict_max_suc( A1, Ra, C4 ),
	and( C3, C4, Constraint ), !.

%------------------------------------------------------------------------------
%	constraint stating that an atom is oriented
%------------------------------------------------------------------------------

constraint_oriented_eq_into( E, Top ):-
	get_atom_preorientation( E, l ),
	true( Top ),!.
constraint_oriented_eq_into( E, C ):-
	true( Top ),
	get_atom_members( E, S1, T1 ),
	get_atom_relation_symbol(E,R),
	(partial_equivalence_relation_symbol(R),\+ reflexive_relation_symbol(R) ->
	     (S1==T1 -> 
		  C=Top
	     ;    and_ge( Top, [S1,T1], C ))
	;    and_gt( Top, [S1,T1], C )).

constraint_oriented_eq( E, Top ):-
	get_atom_preorientation( E, l ),
	true( Top ),!.
constraint_oriented_eq( E, C ):-
	true( Top ),
	get_atom_members( E, S1, T1 ),
	and_gt( Top, [S1,T1], C ).

%--------------------------------------------------------------------------------
%	constraint stating that an atom in a succedent is
%	strictly maximal in a clause
%--------------------------------------------------------------------------------

constraint_strict_max_suc( E, R, C ):-
	maximal_antecedent( R, Am ),
	constraint_greater_suc_ant_list( E, Am, Ca ),
        maximal_succedent( R, Sm ),
	constraint_greater_eq_list( E, Sm, Cs ),
	and( Ca, Cs, C ), !.

%--------------------------------------------------------------------------------
%	constraint stating that an atom in an succedent is
%	maximal in a clause
%--------------------------------------------------------------------------------

constraint_max_suc( E, R, C ):-
	maximal_antecedent( R, Am ),
        constraint_greater_suc_ant_list( E, Am, Ca ),
        maximal_succedent( R, Sm ),
        constraint_grequal_eq_list( E, Sm, Cs ),
        and( Ca, Cs, C ), !.

%------------------------------------------------------------------------------
%	constraint stating that an atom in an antecedent is
%	maximal in a clause
%------------------------------------------------------------------------------

constraint_max_ant( E, R, C ):-
	maximal_antecedent( R, Am ),
        constraint_grequal_eq_list( E, Am, Ca ),
        maximal_succedent( R, Sm ),
        constraint_greater_ant_suc_list( E, Sm, Cs ),
        and( Ca, Cs, C ), !.

%--------------------------------------------------------------------------------
%	constraint_greater_superposed_eqs( Pos, E1, E2, Ct ):
%	constraint stating that an atom E1 is greater than 
%	another one E2 when superposed lhs on lhs at position Pos
%	taking into account that they are supposed to be oriented.
%       E1== (L1 =< R1) is oriented from left to right, 
%       E2== (L2 =< R2) from right to left!. We need not compare
%       L1 and L2 as they are the same for root overlaps.
%--------------------------------------------------------------------------------

constraint_greater_superposed_eqs( [_|_], _, _, Ct ):-
	true( Ct ), !.
constraint_greater_superposed_eqs( [], E1, E2, Ct ):-
	get_atom_relation_symbol(E1,RS1),
	get_atom_relation_symbol(E2,RS2),
	true( C0 ),
	(RS1=RS2 ->
		(symmetric_relation_symbol(RS1) ->
			get_equation_members( E1, _L1, R1 ),
			get_equation_members( E2, _L2, R2 ),
		        and_gt(C0, [R1,R2], Ct)
		;	true(Ct))
	;	Ct=C0),
	!.        





%------------------------------------------------------------------------------
% constraint_greater_eq
%------------------------------------------------------------------------------

constraint_greater_eq(E1,E2,C):-
	constraint_greater_literal(E1,+,E2,+,C).
             
	     





%--------------------------------------------------------------------------------
%	constraint stating that an atom is equal to another one 
%--------------------------------------------------------------------------------

constraint_equal_eqs( E1, E2, CS ):-
	get_atom_relation_symbol(E1,RS1),
	get_atom_relation_symbol(E2,RS2),
	(RS1\==RS2 ->
	     false(C)
	;    get_atom_members( E1, S1, T1 ),
	     get_atom_members( E2, S2, T2 ),
	     true( CT ),
	     and_eq( CT,  [S1,S2], C11 ),
	     and_eq( C11, [T1,T2], C12 ),
	     (symmetric_relation_symbol(RS1) ->
		  and_eq( CT,  [S1,T2], C21 ),
		  and_eq( C21, [T1,S2], C22 ),
		  or( C12, C22, C )
	     ;    C12=C)),
	simplify(C,CS),
	!.


%--------------------------------------------------------------------------------
%	constraint stating that an atom (in the succedent of a clause)
%	is greater than another one (in the antecedent of the clause)
%	+ list
%--------------------------------------------------------------------------------
	
%--------------------------------------------------------------------------------
% constraint_greater_suc_ant
%--------------------------------------------------------------------------------

constraint_greater_suc_ant( E1, E2, C ):-
	constraint_greater_literal(E1,+,E2,-,C).

constraint_greater_suc_ant_list( _, [], C ):-
	true( C ), !.
constraint_greater_suc_ant_list( Eq, [E|R], C ):-
	constraint_greater_suc_ant( Eq, E, C1 ),
	constraint_greater_suc_ant_list( Eq, R, C2 ),
	and( C1, C2, C ), !.

%--------------------------------------------------------------------------------
%	constraint stating that an atom (in the antecedent of a clause)
%	is greater than another one (in the succedent of a clause)
%	+ list
%--------------------------------------------------------------------------------

%--------------------------------------------------------------------------------
% constraint_greater_ant_suc( S1=T1, S2=T2, 
%			(S1>=S2 or T1>=S2) and (S1>=T2 or T1>=T2)  ).
%--------------------------------------------------------------------------------

constraint_greater_ant_suc( E1, E2, C ):-
	constraint_greater_literal(E1,-,E2,+,C).

constraint_greater_ant_suc_list( _, [], C ):-
        true( C ), !.
constraint_greater_ant_suc_list( Eq, [E|R], C ):-
        constraint_greater_ant_suc( Eq, E, C1 ),
        constraint_greater_ant_suc_list( Eq, R, C2 ),
        and( C1, C2, C ), !.

%--------------------------------------------------------------------------------
% constraint_greater_clause
%--------------------------------------------------------------------------------

constraint_greater_clause( C1, C2, Ct ):-
	antecedent( C1, A10 ),
        succedent( C1, S10 ),
        antecedent( C2, A20 ),
        succedent( C2, S20 ),
        multiset_removal_equal_atoms( A10, A20, A1, A2 ),
        multiset_removal_equal_atoms( S10, S20, S1, S2 ),
        constraint_greater_clause1( A1, S1, A2, S2, Ct ), !.


multiset_removal_equal_atoms( A1, [], A1, [] ):- !.
multiset_removal_equal_atoms( A1, [X2|A2], B1, B2 ):-
	get_atom_members( X2, L2, R2),
	get_atom_relation_symbol(X2,RS),
        delete32( X1, A1, NA1 ),
	get_atom_relation_symbol(X1,RS),
	get_atom_members( X1, L1, R1 ),
	(symmetric_relation_symbol(RS) ->
	     ( ( L1==L2, R1==R2 ); ( R1==L2, L1==R2 ) )
	;    L1==L2,
	     R1==R2),
	multiset_removal_equal_atoms( NA1, A2, B1, B2 ), !.
multiset_removal_equal_atoms( A1, [X2|A2], B1, [X2|B2] ):-
        multiset_removal_equal_atoms( A1, A2, B1, B2 ), !.


constraint_greater_clause1( [], [], [], [], Ct ):-
	false( Ct ), !.
constraint_greater_clause1( A, S, [], [], Ct ):-
	( A \== [] ; S \== [] ),
	true( Ct ), !.
constraint_greater_clause1([A],[B],[A1],[B1],Ct):-
	constraint_gt_ms2(lit,[(A,-),(B,+)],[(A1,-),(B1,+)],Ct),
	!.
constraint_greater_clause1( A1, S1, [E|A2], S2, Ct ):-
	constraint_greater_list_eq( A1, E, Ct1 ),
	constraint_greater_list_suc_ant( S1, E, Ct2 ),
	constraint_grequal_clause( A1, S1, A2, S2, Ct3 ),
	and( Ct1, Ct3, Ct13 ),
	and( Ct2, Ct3, Ct32 ),
	or( Ct13, Ct32, CtG ),
	greater_constraint_equalize_ant( A1, [], S1, E, A2, S2, CtE ),
	or( CtG, CtE, Ct ), !.
constraint_greater_clause1( A1, S1, [], [E|S2], Ct ):-
	constraint_greater_list_ant_suc( A1, E, Ct1 ),
	constraint_greater_list_eq( S1, E, Ct2 ),
	constraint_grequal_clause( A1, S1, [], S2, Ct3 ),
	and( Ct1, Ct3, Ct13 ),
	and( Ct2, Ct3, Ct32 ),
	or( Ct13, Ct32, CtG ),
	greater_constraint_equalize_suc( A1, S1, [], E, [], S2, CtE ),
	or( CtG, CtE, Ct ), !.






constraint_grequal_clause( C1, C2, Ct ):-
	antecedent( C1, A10 ),
        succedent( C1, S10 ),
        antecedent( C2, A20 ),
        succedent( C2, S20 ),
        multiset_removal_equal_atoms( A10, A20, A1, A2 ),
        multiset_removal_equal_atoms( S10, S20, S1, S2 ),
        constraint_grequal_clause( A1, S1, A2, S2, Ct ), 
	!.

constraint_grequal_clause( _, _, [], [], Ct ):-
	true( Ct ), !.
constraint_grequal_clause( A1, S1, [E|A2], S2, Ct ):-
	constraint_greater_list_eq( A1, E, Ct1 ),
	constraint_greater_list_suc_ant( S1, E, Ct2 ),
	constraint_grequal_clause( A1, S1, A2, S2, Ct3 ),
	and( Ct1, Ct3, Ct13 ),
	and( Ct2, Ct3, Ct32 ),
	or( Ct13, Ct32, CtG ),
	grequal_constraint_equalize_ant( A1, [], S1, E, A2, S2, CtE ),
	or( CtG, CtE, Ct ), !.
constraint_grequal_clause( A1, S1, [], [E|S2], Ct ):-
	constraint_greater_list_ant_suc( A1, E, Ct1 ),
	constraint_greater_list_eq( S1, E, Ct2 ),
	constraint_grequal_clause( A1, S1, [], S2, Ct3 ),
	and( Ct1, Ct3, Ct13 ),
	and( Ct2, Ct3, Ct32 ),
	or( Ct13, Ct32, CtG ),
	grequal_constraint_equalize_suc( A1, S1, [], E, [], S2, CtE ),
	or( CtG, CtE, Ct ), !.


% the same thing for a conjunctive in-constraint CtIn

constraint_grequal_clause( C1, C2, CtIn, Ct ):-
	antecedent( C1, A10 ),
        succedent( C1, S10 ),
        antecedent( C2, A20 ),
        succedent( C2, S20 ),
        multiset_removal_equal_atoms( A10, A20, A1, A2 ),
        multiset_removal_equal_atoms( S10, S20, S1, S2 ),
        constraint_grequal_clause( A1, S1, A2, S2, CtIn, Ct ), 
	!.

constraint_grequal_clause( _, _, [], [], Ct, Ct ).
constraint_grequal_clause( A1, S1, [E|A2], S2, CtIn, Ct ):-
	constraint_greater_list_eq( A1, E, Ct1 ),
	constraint_greater_list_suc_ant( S1, E, Ct2 ),
	constraint_grequal_clause( A1, S1, A2, S2, Ct3 ),
	and(Ct1,CtIn, CtI1), and( CtI1, Ct3, Ct13 ),
	and(Ct2,CtIn, CtI2), and( CtI2, Ct3, Ct32 ),
	or( Ct13, Ct32, CtG ),
	grequal_constraint_equalize_ant( A1, [], S1, E, A2, S2, CtE ),
	and(CtIn,CtE,CtIE),
	or( CtG, CtIE, Ct ), !.
constraint_grequal_clause( A1, S1, [], [E|S2], CtIn, Ct ):-
	constraint_greater_list_ant_suc( A1, E, Ct1 ),
	constraint_greater_list_eq( S1, E, Ct2 ),
	constraint_grequal_clause( A1, S1, [], S2, Ct3 ),
	and(Ct1,CtIn, CtI1), and( CtI1, Ct3, Ct13 ),
	and(Ct2,CtIn, CtI2), and( CtI2, Ct3, Ct32 ),
	or( Ct13, Ct32, CtG ),
	grequal_constraint_equalize_suc( A1, S1, [], E, [], S2, CtE ),
	and(CtIn,CtE,CtIE),
	or( CtG, CtIE, Ct ), !.


%--------------------------------------------------------------------------------
% greater_constraint_equalize_ant( A1, [], S1, E, A2, S2, Ct ):
% computes Ct:  "A1-->S1  >  [E|A2]-->S2 with E=E1 for some E1 in A1"
%--------------------------------------------------------------------------------

greater_constraint_equalize_ant( [], _, _, _, _, _,  Ct ):-
	false( Ct ), !.
greater_constraint_equalize_ant( [EA|A1], DA, S1, E, A2, S2, Ct ):-
	constraint_equal_eqs( EA, E, Ceq ),
	append( A1, DA, A ),
	constraint_greater_clause1( A, S1, A2, S2, CR ),
	and( Ceq, CR, CtE ),
	greater_constraint_equalize_ant( A1, [EA|DA], S1, E, A2, S2, Ct1 ),
	or( CtE, Ct1, Ct ), !.


greater_constraint_equalize_suc( _, [], _, _, _, _,  Ct ):-
	false( Ct ), !.
greater_constraint_equalize_suc( A1, [ES|S1], DS, E, A2, S2, Ct ):-
	constraint_equal_eqs( ES, E, Ceq ),
	append( S1, DS, S ),
	constraint_greater_clause1( A1, S, A2, S2, CR ),
	and( Ceq, CR, CtE ),
	greater_constraint_equalize_suc( A1, S1, [ES|DS], E, A2, S2, Ct1 ),
	or( CtE, Ct1, Ct ), !.


grequal_constraint_equalize_ant( [], _, _, _, _, _,  Ct ):-
	false( Ct ), !.
grequal_constraint_equalize_ant( [EA|A1], DA, S1, E, A2, S2, Ct ):-
	constraint_equal_eqs( EA, E, Ceq ),
	append( A1, DA, A ),
	constraint_grequal_clause( A, S1, A2, S2, CR ),
	and( Ceq, CR, CtE ),
	grequal_constraint_equalize_ant( A1, [EA|DA], S1, E, A2, S2, Ct1 ),
	or( CtE, Ct1, Ct ), !.


grequal_constraint_equalize_suc( _, [], _, _, _, _,  Ct ):-
	false( Ct ), !.
grequal_constraint_equalize_suc( A1, [ES|S1], DS, E, A2, S2, Ct ):-
	constraint_equal_eqs( ES, E, Ceq ),
	append( S1, DS, S ),
	constraint_grequal_clause( A1, S, A2, S2, CR ),
	and( Ceq, CR, CtE ),
	grequal_constraint_equalize_suc( A1, S1, [ES|DS], E, A2, S2, Ct1 ),
	or( CtE, Ct1, Ct ), !.

%--------------------------------------------------------------------------------
%	constraint stating that there exists some atom 
%	in a list that is greater than the given atom
%--------------------------------------------------------------------------------

constraint_greater_list_eq( [], _, Ct ):-
	false( Ct ), !.
constraint_greater_list_eq( [E|L], X, Ct ):-
        constraint_greater_eq( E, X, Ct1 ),
        constraint_greater_list_eq( L, X, Ct2 ),
	or( Ct1, Ct2, Ct ), !.

%--------------------------------------------------------------------------------
%	constraint stating that there exists some atom 
%	in a list that is greater_suc_ant than the given atom
%--------------------------------------------------------------------------------

constraint_greater_list_suc_ant( [], _, Ct ):-
	false( Ct ), !.
constraint_greater_list_suc_ant( [E|L], X, Ct ):-
        constraint_greater_suc_ant( E, X, Ct1 ),
        constraint_greater_list_suc_ant( L, X, Ct2 ),
	or( Ct1, Ct2, Ct ), !.

%--------------------------------------------------------------------------------
%	constraint stating that there exists some atom 
%	in a list that is greater_ant_suc than the given atom
%--------------------------------------------------------------------------------

constraint_greater_list_ant_suc( [], _, Ct ):-
	false( Ct ), !. 
constraint_greater_list_ant_suc( [E|L], X, Ct ):-
	constraint_greater_ant_suc( E, X, Ct1 ),
        constraint_greater_list_ant_suc( L, X, Ct2 ),
	or( Ct1, Ct2, Ct ), !.


constraint_greater_eq_list( _, [], C ):-
	true( C ), !.
constraint_greater_eq_list( Eq, [E|R], C ):-
	constraint_greater_eq( Eq, E, C1 ),
	constraint_greater_eq_list( Eq, R, C2 ),
	and( C1, C2, C ), !.

%------------------------------------------------------------------------------
%	constraint stating that an atom is greater or equal to another one 
%------------------------------------------------------------------------------

constraint_grequal_eq( E1, E2, C ):-
	constraint_greater_eq( E1, E2, C0 ),
	constraint_equal_eqs(E1,E2,C1),
	or(C0,C1,C),
	!.



constraint_grequal_eq_list( _, [], C ):-
        true( C ), !.
constraint_grequal_eq_list( Eq, [E|R], C ):-
        constraint_grequal_eq( Eq, E, C1 ),
	simplify(C1,C1S),
        constraint_grequal_eq_list( Eq, R, C2 ),
%	simplify(C2,C2S),
        and( C1S, C2, C ), !.





%------------------------------------------------------------------------------
%  polarities of literals are denoted by '+' and '-' in the following predicate
%  
%------------------------------------------------------------------------------
constraint_greater_literal( [P1|E1], [P2|E2], C ):-
	constraint_greater_literal( E1, P1, E2, P2, C ).
	
	

constraint_grequal_literal( E1, P, E2, P, C ):-
	!,
	constraint_greater_literal( E1, P, E2, P, C0 ),
	constraint_equal_eqs(E1,E2,C1),
	or(C0,C1,C),
	!.
constraint_grequal_literal( E1, P1, E2, P2, C ):-
	constraint_greater_literal( E1, P1, E2, P2, C ).



constraint_greater_literal_in( [P1|E1], [P2|E2], CIn, C ):-
	constraint_greater_literal( E1, P1, E2, P2, CL ),
	and(CIn,CL,C).



constraint_grequal_literal_in( [P1|E1], [P2|E2], CIn, C ):-
	constraint_grequal_literal( E1, P1, E2, P2, CL ),
	and(CIn,CL,C).




constraint_greater_literal( E1, P1, E2, P2, C ):-
	setof(CL1,constrained_complexities(E1,P1,CL1),List1),
	setof(CL2,constrained_complexities(E2,P2,CL2),List2),
	true(CT),
	constraint_gt_complexity_lists(List1,List2,CT,C).


/* ---------------------------------------------------------------
With

         constrained_complexities(E,P,(C,[S,P,Side,RS,T]))

we compute the quintuples (S,P,Side,RS,T)
that represent the complexity of a literal (P,E) of the form 

           ([+/-], RS(L,R))  :

S is the singleton multiset with the maximal term among L and R
P is the polarity of the literal
Side=top, if S is the left term, Side=bot, otherwise
RS is the relation symbol
T is the singleton multiset with the minimal term among L and R.

For symmetric relations RS(L,R) which are not ``special'',
S is the multiset [L,R] and T is [], the empty multiset.

Since on the non-ground level maximal and minimal terms need not be unique,
more than one such complexity tuple can exist. 

constrained_complexities(E,P,(C,[S,P,Side,RS,T]))
computes them all, together with a constraint C such that 
for all substitutions that satisfy C, the corresponding
instance of (S,P,Side,RS,T) is the (unique) complexity measure
of the instance of (P,E).


----------------------------------------------------------------*/

/*

constrained_complexities(E,P,(C,[[S,T],P,bot,RS,[]])):-
%	fail,                        % experimental
	\+ props:case_split(_,_,_),  % for experimentation with equality axioms
	get_atom_relation_symbol(E,RS),
	symmetric_relation_symbol(RS),
	\+ special_relation(RS),
	!,
	get_atom_members(E,S,T),
	true(C).
*/
constrained_complexities(E,P,(C,[[S1],Pol1,Side,RS1,[T1]])):-
	% special treatment of (possibly non-special) binary atoms
	% experimental
	get_binary_atom_pre_rel_left_right(E,O,RS,S,T),
	(adt_precedence:dual_predicate_pair(RS,_RSNeg) ->
	     Pol1= +P,
	     RS1=RS
	;adt_precedence:dual_predicate_pair(RS1,RS) ->
	     Pol1= -P
	),
	!,
	true(CT),
	(    O\==r,
	     S1=S,
	     T1=T,
	     and_ge(CT,[S,T],C),
	     Side1=top
	;
	     O\==l,
             S1=T,
	     T1=S,
	     and_gt(CT,[T,S],C),
	     Side1=bot
	),
	(symmetric_relation_symbol(RS) -> Side=bot; Side=Side1).
constrained_complexities(E,P,(C,[[S1],P,Side,RS,[T1]])):-
	get_atom_pre_rel_left_right(E,O,RS,S,T),
	true(CT),
	(    O\==r,
	     S1=S,
	     T1=T,
	     and_ge(CT,[S,T],C),
	     Side1=top
	;
	     O\==l,
             S1=T,
	     T1=S,
	     and_gt(CT,[T,S],C),
	     Side1=bot
	),
	(symmetric_relation_symbol(RS) -> Side=bot; Side=Side1).



% for experimentation with specific literal orderings:




/*---------------------------------------------------------------

  constraint_gt_complexity_list(L1,L2,Cin,Cout)

computes the constraint Cout which is the conjunction
of all the constraints 

      C1 and C2 implies constraint_gt_complexity_tuple(Tuple1,Tuple2)

for (C1,Tuple1) in L1 and (C2,Tuple2) in L2,
with Cin.


-----------------------------------------------------------------*/

constraint_gt_complexity_lists([],_,C,C).
constraint_gt_complexity_lists([CComp|CC],CC1,Cin,Cout):-
	constraint_gt_complexity_tuple_list(CComp,CC1,Cin,Cint),
	constraint_gt_complexity_lists(CC,CC1,Cint,Cout).

constraint_gt_complexity_tuple_list(_,[],C,C).
constraint_gt_complexity_tuple_list((C,Co),[(C1,Co1)|CC],Cin,Cout):-
	constraint_gt_complexity_tuple(Co,Co1,Compare),
	and(C,C1,C11),
	not(C11,NotC11),
	or(NotC11,Compare,C3),
	and(Cin,C3,Cint),
	constraint_gt_complexity_tuple_list((C,Co),CC,Cint,Cout).


/* -----------------------------------------------------------------
constraint_gt_complexity_tuple
        ([Max,Pol,Side,RS,Min],[Max1,Pol1,Side1,RS1,Min1],C)

computes the constraint C which is satisfied for all substitutions
for which the first complexity tuple is greater than the second.
In this we assume that the Max terms are greater or equal than the
Min terms in each tuple.

The tuples are compared lexicographically using the
obvious orderings on the components
-------------------------------------------------------------------*/


constraint_gt_complexity_tuple([Max,Pol,Side,RS,Min],[Max1,Pol1,Side1,RS1,Min1],C):-
	constraint_gt_ms2(atom,Max,Max1,CStrictMax),
	constraint_ge_ms2(atom,Max,Max1,CMax),
	constraint_eq_ms2(atom,Max,Max1,CMaxEq),
	(gtPol(Pol,Pol1) ->
	      C=CMax
	;(Pol=Pol1, gtSide(Side,Side1)) ->
	      C=CMax
	;(Pol=Pol1, Side=Side1, greater_relation_symbol(RS,RS1)) ->
	      C=CMax
	;(Pol=Pol1,Side=Side1, equivalent_relation_symbols(RS,RS1)) ->
	      constraint_gt_ms2(atom,Min,Min1,CStrictMin),
	      and(CMaxEq,CStrictMin,CMin),
	      or(CStrictMax,CMin,C)
	;C=CStrictMax),
	!.



gtPol(P1,P2):-
	P1 @> P2.

gtSide(top,bot).	


as_ordering_constraint(atom,=,S,T,C):-
	true(CT),
	and_eq(CT,[S,T],C).
as_ordering_constraint(atom,>,S,T,C):-
	true(CT),
	and_gt(CT,[S,T],C).
as_ordering_constraint(atom,>=,S,T,C):-
	true(CT),
	and_ge(CT,[S,T],C).
as_ordering_constraint(lit,>,(A,P),(A1,P1),Ct):-
	constraint_greater_literal(A,P,A1,P1,Ct).
as_ordering_constraint(lit,>=,(A,P),(A1,P1),Ct):-
	constraint_grequal_literal(A,P,A1,P1,Ct).





constraint_gt_ms2(_,[],_,Ct):-
	false(Ct).
constraint_gt_ms2(_,[_|_],[],Ct):-
	true(Ct).
constraint_gt_ms2(Ty,[A],[C],Ct):-
	as_ordering_constraint(Ty,>,A,C,Ct).
constraint_gt_ms2(Ty,[A,B],[C],Ct):-
	as_ordering_constraint(Ty,>=,A,C,AgeC),
	as_ordering_constraint(Ty,>=,B,C,BgeC),
	or(AgeC,BgeC,Ct).
constraint_gt_ms2(Ty,[A],[C,D],Ct):-
	as_ordering_constraint(Ty,>,A,C,AgtC),
	as_ordering_constraint(Ty,>,A,D,AgtD),
	and(AgtC,AgtD,Ct).
constraint_gt_ms2(Ty,[A,B],[C,D],Ct):-
	(C==A ->
	     as_ordering_constraint(Ty,>,B,D,Ct)
	;C==B ->
	     as_ordering_constraint(Ty,>,A,D,Ct)
	;D==A ->
	     as_ordering_constraint(Ty,>,B,C,Ct)
	;D==B ->
	     as_ordering_constraint(Ty,>,A,C,Ct)

	;% no two equal terms
	     
	     as_ordering_constraint(Ty,>=,A,C,AgeC),as_ordering_constraint(Ty,>,B,D,BgtD),and(AgeC,BgtD,C1),
	     as_ordering_constraint(Ty,>,A,C,AgtC),as_ordering_constraint(Ty,>=,B,D,BgeD),and(AgtC,BgeD,C2),
	     as_ordering_constraint(Ty,>=,A,D,AgeD),as_ordering_constraint(Ty,>,B,C,BgtC),and(AgeD,BgtC,C3),
	     as_ordering_constraint(Ty,>,A,D,AgtD),as_ordering_constraint(Ty,>=,B,C,BgeC),and(AgtD,BgeC,C4),
	     as_ordering_constraint(Ty,>,A,C,AgtC),as_ordering_constraint(Ty,>,A,D,AgtD),and(AgtC,AgtD,C5),
	     as_ordering_constraint(Ty,>,B,C,BgtC),as_ordering_constraint(Ty,>,B,D,BgtD),and(BgtC,BgtD,C6),
	     or_list([C1,C2,C3,C4,C5,C6],Ct)
	).

constraint_ge_ms2(_,[],[],Ct):-
	true(Ct).
constraint_ge_ms2(_,[],[_|_],Ct):-
	false(Ct).
constraint_ge_ms2(_,[_|_],[],Ct):-
	true(Ct).
constraint_ge_ms2(Ty,[A],[C],Ct):-
	as_ordering_constraint(Ty,>=,A,C,Ct).
constraint_ge_ms2(Ty,[A,B],[C],Ct):-
	constraint_gt_ms2(Ty,[A,B],[C],Ct).
constraint_ge_ms2(Ty,[A],[C,D],Ct):-
	constraint_gt_ms2(Ty,[A],[C,D],Ct).
constraint_ge_ms2(Ty,[A,B],[C,D],Ct):-
	(C==A ->
	     as_ordering_constraint(Ty,>=,B,D,Ct)
	;C==B ->
	     as_ordering_constraint(Ty,>=,A,D,Ct)
	;D==A ->
	     as_ordering_constraint(Ty,>=,B,C,Ct)
	;D==B ->
	     as_ordering_constraint(Ty,>=,A,C,Ct)

	;% no two equal terms
	     
	     as_ordering_constraint(Ty,>=,A,C,AgeC),as_ordering_constraint(Ty,>=,B,D,BgeD),and(AgeC,BgeD,C1),
	     as_ordering_constraint(Ty,>=,A,D,AgeD),as_ordering_constraint(Ty,>=,B,C,BgeC),and(AgeD,BgeC,C2),
	     as_ordering_constraint(Ty,>,A,C,AgtC),as_ordering_constraint(Ty,>,A,D,AgtD),and(AgtC,AgtD,C5),
	     as_ordering_constraint(Ty,>,B,C,BgtC),as_ordering_constraint(Ty,>,B,D,BgtD),and(BgtC,BgtD,C6),
	     or_list([C1,C2,C5,C6],Ct)
	).


constraint_eq_ms2(_,[],[],Ct):-
	true(Ct).
constraint_eq_ms2(_,[],[_|_],Ct):-
	false(Ct).
constraint_eq_ms2(_,[_|_],[],Ct):-
	false(Ct).
constraint_eq_ms2(Ty,[A],[C],Ct):-
	as_ordering_constraint(Ty,=,A,C,Ct).
constraint_eq_ms2(_,[_A,_B],[_C],Ct):-
	false(Ct).
constraint_eq_ms2(_,[_A],[_C,_D],Ct):-
	false(Ct).
constraint_eq_ms2(Ty,[A,B],[C,D],Ct):-
	(C==A ->
	     as_ordering_constraint(Ty,=,B,D,Ct)
	;C==B ->
	     as_ordering_constraint(Ty,=,A,D,Ct)
	;D==A ->
	     as_ordering_constraint(Ty,=,B,C,Ct)
	;D==B ->
	     as_ordering_constraint(Ty,=,A,C,Ct)
	;% no two equal terms
	     
	     as_ordering_constraint(Ty,=,A,C,AeqC),as_ordering_constraint(Ty,=,B,D,BeqD),and(AeqC,BeqD,C1),
	     as_ordering_constraint(Ty,=,A,D,AeqD),as_ordering_constraint(Ty,=,B,C,BeqC),and(AeqD,BeqC,C2),
	     and(C1,C2,Ct)
	).


	     






:- dynamic sos/0.

%horn1:-horn.   % incomplete with simplification.
horn1:-sos.

% eof ---------------------------------------------------------------------------
