:-use_module(library(assoc)).
:-ensure_loaded('CLPPB/libclppb').


:-use_module(adt_atom,[atom_as_term/2]).
:-use_module(library(maplist),[maplist/2,maplist/3]).
:-use_module(library(basics),[member/2]).

:-use_module(index,[index/3,index/5]).

:-use_module(adt_proof,[parents/2]).

:-use_module(props,[flag_on/2]).

:-use_module(classes, [
        antecedent/2,
	createClause/2,
	groundClause/1,
	history_and_number/3,
	number/2,
	propositionalClause/1,
	succedent/2,
	vars/2
   ]).

:-use_module(proofState,[recordProofOfClause/1,operators/1]).


propLogicInit
	:- pb_init.

treatPropositionally(C,Domain):-
	flag_on(rp,prop),
	(groundClause(C),
	   \+groundParents(C) ->
	   Domain=[]
	;
	   operators(Domain),
	   length(Domain,M),
	   \+ (member(_F/N,Domain),N>0),
	   M>=1,
	   M=<3

	).
	   

tellPropClause(C,C1,ST,STO):-
	(treatPropositionally(C,Domain) -> 
	     ground_instances(C,Domain,Cs),
	     tellPropClauses(Cs,C1,ST,STO),
	     (var(C1) ->
		  C=C1
	     ;
	          true
	     )
	;
	     C1=C,
	     ST=STO
	).
	
tellPropClauses([C|Cs],C1,ST,STO):-
%	format('~Ntelling ~p',[C]),
	antecedent(C,A),
	succedent(C,S),
	number(C,N),
	maplist(atom_as_term,A,AP),
	maplist(atom_as_term,S,SP),
	index(prop_assoc,ST,t(Assoc,PCls),ST1,t(AssocO,[N|PCls])),
	toPVars(AP,PAP,Assoc,Assoc1),
	toPVars(SP,PSP,Assoc1,AssocO),
	toBExpr(PAP,-,BA),
	toBExpr(PSP,+,BS),
	(   sat('#>='(BA+BS,1))
	->  
	     tellPropClauses(Cs,C1,ST1,STO)
	;
	     ST1=STO,
	     createClause([[],[]],C2),
	     history_and_number([prop,[N,PCls],[C]],C2,C1),
	     recordProofOfClause(C1)
	).
tellPropClauses([],_,ST,ST).





ground_instances(C,Domain,Cs):-
	vars(C,V),
	findall(C,choose(V,Domain),Cs).

choose([X|Xs],D):-
	member(X/_,D),
	choose(Xs,D).
choose([],_).


groundParents(C):-
	parents(C,Cs),
%	(Cs\==[] -> maplist(groundClause,Cs)).
	(Cs\==[] -> maplist(propositionalClause,Cs)).



%
% associating ground atoms with
% propositional variables represented as
% Prolog variables
%
toPVars([],[],Assoc,Assoc).
toPVars([P|Ps],[PP|PPs],Assoc,AssocO):-
	(get_assoc(P,Assoc,PP) ->
	     Assoc1=Assoc
	;
	     put_assoc(P,Assoc,PP,Assoc1)
	),
	toPVars(Ps,PPs,Assoc1,AssocO).

	     
	
toBExpr([],_,0).
toBExpr([P],+,P).
toBExpr([P],-,'~'(P)).
toBExpr([P|Ps],Pol,BP+BPs):-
	toBExpr(Ps,Pol,BPs),
	(Pol= + ->
	     BP=P
	;
	     BP= '~'(P)
	).


sat(X):-
	pb_add(X).



var_assoc(V,ST,A):-
	index(prop_assoc,ST,Assoc),
	get_assoc(V,Assoc,A1),
	(A1==1 ->
	     A='$true$'
	;A1==0 ->
	     A='$min'
	).