/*
 *	file:		match.pl
 *	version:	1.5
 *	date:		October 10, 1989
 *	creation:	-
 *	author:		-
 *
 *	description:
 *	This file contains the function to do matching (AC-matching
 *	if necessary).
 *
 *	history:
 *	891010	js	Added this comment
 *  900412  huba  rewritten ac_match/2 to handle non-linear Patterns more efficently
 *                      removed ac_matchLinear/2 and ac_matchLinearC/4.
 *
 *	Copyright (C) 1989
 *		Hubert Bertling
 *		Harald Ganzinger
 *		Renate Schaefers
 *		University of Dortmund
 */






/*------------------------------------------------------------------*
 *								    *
 *	ac_match(GroundTerm,Pattern)						    *
 *								    *
 *------------------------------------------------------------------*/

%declarations%

:- module(match, [ 
	ac_equal/2,
	ac_match/2,
	ac_match/4,
	ac_match_pt/2,
	ac_matchvf/2,
	ac_subsumes_chk/2
   ]).

:- use_module(uniac, [
	ac_uniAC/2
   ]).

:- use_module(props, [
	ac_ist_C/1
   ]).

:- use_module(achelp, [
	ac_del_assoc/3,
	ac_ist_AC/1,
	ac_linear/3,
	ac_listACGleich/1,
	ac_mehrElemente/3,
	ac_removeAC/4,
	ac_removeACGleich/4,
	ac_removeACMatch/4,
	ac_split/4
   ]).

:- use_module(myterms, [
	apply_substitution/3,
	list_of_vars/2
   ]).

:- use_module(library(maplist), [
	maplist/3,
	maplist/4
   ]).

:- use_module(auxiliary, [
	lookup_ml/2
   ]).


%end%


ac_subsumes_chk(P,T):-
	\+ \+ (numbervars(T,1,_),ac_match(T,P)).
	




ac_equal(T1,T2):-
	T1==T2,
	!.
ac_equal(T1,T2):-
	ac_ist_C(_),
	!,
	\+ \+ (numbervars((T1,T2),1,_), ac_matchvf(T1,T2)).



ac_match_pt(P,T):-
	list_of_vars(T,V),   % need to also skolemize bound vars
	copy_term((T,V,P),(T1,V1,P1)),
	numbervars(T1,1,_),
	maplist(equate,V1,V,Su),
	ac_match(T1,P1),
	apply_substitution(Su,P1,PS),
	P=PS.


equate(X,Y,X=Y).


:- user:no_style_check(single_var).    % old code





ac_match(GTerm,Pattern):-
	\+ ac_ist_C(_),
	!,
	GTerm=Pattern.
ac_match(GTerm,Pattern) :- 
	var(Pattern),
	!,
	GTerm = Pattern.
ac_match(GTerm,_Pattern) :-
	var(GTerm),
	!,
	fail.
ac_match(GTerm,Pattern) :-
	Pattern =.. [Op|PatternArgs],
	ac_ist_AC(Op),
	!,
	GTerm =.. [Op|TermArgs],
	ac_del_assoc(Op,PatternArgs,PArgs),
	ac_del_assoc(Op,TermArgs,TArgs),
	ac_split(PArgs,VarSet,TermSet,GTermSet),
	ac_removeACGleich(GTermSet,TArgs,[],Erg),
	ac_removeACMatch(TermSet,Erg,[],Erg1),
	ac_split(VarSet,VarSetSplit,[],GTermSetSplit),
	ac_del_assoc(Op,GTermSetSplit,GTermSetSplit1),
	ac_removeACGleich(GTermSetSplit1,Erg1,[],Erg2),
	ac_linear(VarSetSplit,VarLinearSet,VarSubst),
	ac_mehrElemente(Op,Erg2,VarLinearSet),
	ac_listACGleich(VarSubst).
ac_match(GTerm,Pattern) :-
	Pattern =.. [Op,P1,P2],
	ac_ist_C(Op),
	!,
	GTerm =.. [Op,T1,T2],
	(	var(P1),
		P1 == P2,
		!,
		ac_match(T1,T2)
	;
		(	ac_match(T1,P1),
			ac_match(T2,P2)
		;
			ac_match(T1,P2),
			ac_match(T2,P1)
		)
	).
ac_match(GTerm,Pattern) :-
	ac_removeAC(Pattern,GTerm,ACPatterns,ACGTs),
	maplist(ac_match,ACGTs,ACPatterns).

/*------------------------------------------------------------------*
 *                                                                  *
 *        ac_matchvf(Term,Term1)				    *
 *                                                                  *
 *------------------------------------------------------------------*/

ac_matchvf(Term,Term1):-
	\+ ac_ist_C(_),
	!,
	Term=Term1.
ac_matchvf(Term,Term1) :-
	var(Term1),
	!,
	Term=Term1.
ac_matchvf(Term,Term1) :-
        Term =.. [F|Args],
        Term1 =.. [F|Args1],
        (	ac_ist_AC(F),
		!,
		ac_del_assoc(F,Args,Args10),
                ac_del_assoc(F,Args1,Args11),
                ac_removeACGleich(Args10,Args11,[],[])
	;
		ac_ist_C(F),
		!,
		ac_removeACGleich(Args,Args1,[],[])
	;
		maplist(ac_matchvf,Args,Args1)
	).



:- user:style_check(single_var).    % old code




%------------------------------------------------------------------------------
/*
match(+P,+T,+L1,?L2) 

is true when T = P(\sigma\cup\tau) (modulo AC),
where \sigma and \tau are the substitutions represented 
by L1 and L2, respectively.
The latter are equational problems in solved form.

*/

ac_match(P,T,I,O):-
	list_of_vars(P,PV),
	list_of_vars(T,TV),
	copy_term((T,TV,P,PV),(T1,TV1,P1,PV1)),
	numbervars(T1,1,_),
	maplist(equate,TV1,TV,Su),
	ac_match(T1,P1),
	apply_substitution(Su,PV1,PVS),
	composedSubstitution(PV,PVS,I,O).


composedSubstitution([X|PV],[T|PV1],I,O):-
	(lookup_ml(X=T1,I) ->
	     ac_equal(T1,T),
	     O=O1
	;
	     O=[X=T|O1]
	),
	composedSubstitution(PV,PV1,I,O1).
composedSubstitution([],[],I,I).
