% -*- Module: me; -*-

:- module(me,
	  [ prove_me/2,
	    prove_me/3 ]
	 ).



:- use_module(adt_term,[ get_term_vars/2,
			 term_info/4
		       ]).

:- use_module(termOrder,[ greater/2,
			  greaterEqual/2
		       ]).

:- use_module(mylists,[ reverse/2 ]).

:- use_module(auxiliary,[ member212/2 ]).

:- use_module(library(basics), [
	member/2
   ]).



:- use_module(adt_atom,[ atom_as_term/2 ]).

:- use_module(global,[ cont/2,
		       incr/1,
		       store/2
		     ]).

:- use_module(orderConstraints,[ and/3,
				 false/1,
				 true/1,
				 project/3
			       ]).

:- use_module(constraintSolver,[ simplify/2,
				 simplify_med/2 ]).

:- use_module(proofState,[ clause/1 ]).

:- use_module(classes,[ antecedent/2,
			succedent/2,
			constraint/2,
			theory_flag/2
		      ]).

:- use_module(equationalTheory,[ unify/2 ]).

:- use_module(library(maplist),[ maplist/3 ]).


:- dynamic proved/1, failed/2.


prove([],_,_,_,Delta,Delta):-
	!.
prove([Goal|Goals],D,Ancestors,Method,DeltaI,DeltaO):-
	D>0,
	\+ member212(Goal,Ancestors),
	resolvingClause(Goal,C),
	checkConstraint(Method,DeltaI,C,DeltaAndC),
%	format('~N~p:~p - ~p~n  if ~p',[D,Goal,Ancestors,DeltaAndC]),
	D1 is D-1,
	subgoals(C,Subgoals),
	projectConstraint(Goals+Subgoals,DeltaAndC,DeltaP),
	incr(steps),
	reverse(Subgoals,SubgoalsR),
	prove(SubgoalsR,D1,[Goal|Ancestors],Method,DeltaP,Delta),
	prove(Goals,D,Ancestors,Method,Delta,DeltaO).
%	memo_proved(Goal),


alreadyProved(G):-
	\+ \+ ( numbervars(G,0,_),
		proved(G)
	      ).

instanceProved(G):-
	proved(G1),
	unify(G,G1).

memo_proved(G):-
	(alreadyProved(G) ->
	     true
	;
	     assert(proved(G))
	).

memo_failed(G,D):-
	(\+ failedBefore(G,D) ->
	     assert(failed(G,D))
	),
	fail.

failedBefore(G,D):-
	failed(G1,D1),
	subsumes_chk((G,D),(G1,D1)).


resolvingClause(Goal,C):-
	clause(C),
	succedent(C,[A]),
	atom_as_term(A,Head),
	unify(Goal,Head).
	
checkConstraint(Method,Delta,C,DeltaAndC):-
	constraint(C,Gamma),
%	     format('~N~p',[Gamma]),
	orderConstraints:and1(Delta,Gamma,DandG1),
	simplify_constraint(Method,DandG1,DeltaAndC),
%	     format('~N*~p',[DeltaAndC]),
	\+ false(DeltaAndC),
	!.

subgoals(C,Goals):-
	antecedent(C,AnteC),
	maplist(atom_as_term,AnteC,Goals).




projectConstraint(_Goals,Delta,DeltaP):-
%	get_term_vars(_Goals,GVar),
%	project(GVar,Delta,DeltaP),
	DeltaP=Delta.
	



prove_me(D,Result):-
	true(Top),
	prove_me(D,Top,Result).

prove_me(D,Method,Result):-
	retractall(proved(_)),
	retractall(failed(_,_)),
	true(Top),
	goalClause(C),
	antecedent(C,AnteC),
%	reverse(AnteC,AnteR),
	maplist(atom_as_term,AnteC,Goals),
	store(steps,0),
	!,
	(prove(Goals,D,[],Method,Top,_) ->
	     cont(steps,S),
	     Result=proved(S)
	;
	     cont(steps,S),
	     Result=notProved(S)
	).
	


goalClause(C):-
	clause(C),
	theory_flag(C,goal).
	
atom_with_Depth_and_Ancestors(D,Anc,A,D+T+Anc):-
	atom_as_term(A,T).


simplify_constraint(nc,_C,[conj([],[],[])]).
simplify_constraint(med,C,C1):-
	simplify_med(C,C1).
simplify_constraint(full,C,C1):-
	simplify(C,C1).
simplify_constraint(light,C,C1):-
	simplify_constraint(C,C1).



simplify_constraint([conj([],E,O)],[conj([],E,O)]).
simplify_constraint([conj([S>=T|C],E,O)],C1):-
	(greaterEqual(S,T) ->
	     simplify_constraint([conj(C,E,O)],C1)
	;greater(T,S) ->
	     C1=[]
	;% uncomparable
	     simplify_constraint([conj(C,E,O)],[conj(C2,E,O)]),
	     C1=[conj([S>=T|C2],E,O)]
	).
simplify_constraint([conj([S>T|C],E,O)],C1):-
	(greater(S,T) ->
	     simplify_constraint([conj(C,E,O)],C1)
	;greaterEqual(T,S) ->
	     C1=[]
	;% uncomparable
	     simplify_constraint([conj(C,E,O)],[conj(C2,E,O)]),
	     C1=[conj([S>=T|C2],E,O)]
	).
