:- class(clause,simpleClause,
	 [history,
	  constraint,
	  redundancy_info,
	  kind,
	  complexityBound],
	 [
	 as_new_axiom/2,
	 as_new_goal/2,
	 as_dfg/2,
	 as_prolog_literals/2,
	 as_prolog_sequent/2,
	 as_spass/2,
	 as_tptp/2,
	 history/2,
	 history/3,
	 constraint/2,
	 complexityBound/2,
	 complexityBound/3,
	 constraint/3,
	 redundancy_info/2,
	 kind/2,
	 kind/3,
	 implicants/2,
	 history_and_number/3,
	 size_and_weight/3,
	 dont_simplify/1,
	 dont_demodulate_with/1,
	 operators/2,
	 proof_depth/2,
	 search_depth/2,
	 subterm/2,
	 subterms/2,
	 theory_flag/2
	 

]).

% -----------------------------------------------------------------------------
% -----------------------------------------------------------------------------
%
% clauses have the instance variables
%
%       history,constraint,redundancy_info,kind,compexityBound
% where
%
% redundancy_info ::- (noTaut|tautology|_),Implicants
%
% -----------------------------------------------------------------------------
% -----------------------------------------------------------------------------






% creation
% -----------------------------------------------------------------------------

createClause( [A,S],    myself(C,Cl)):-
	createSimpleClause([A,S],C),
	object_template(Cl),
	attr(constraint,Cl,Ct),
	attr(kind,Cl,regular),
	true( Ct ).
createClause( [A,S,Ct],    myself(C,Cl)):-
	createSimpleClause([A,S],C),
	object_template(Cl),
	attr(constraint,Cl,Ct),
	attr(kind,Cl,regular).






% accessing
% -----------------------------------------------------------------------------

history( self(C), F ):-
	attr(history,C,F).

constraint( self(C), Ct ):-
	attr(constraint,C,Ct).

redundancy_info(self(C),I):-
	attr(redundancy_info,C,I).

kind(self(C),K):-
	attr(kind,C,K).

complexityBound(self(C),CB):-
	attr(complexityBound,C,CB).

implicants(C,I):-
	redundancy_info(C,(_,IP)),
	(var(IP) ->
	     I=[]
	;    setof0(E,P^member((E,P),IP),I)).

proof_depth(C,D):-
	C=self(_),
	history(C,H),
	depth(H,D).

search_depth(C,D):-
	C=self(_),
	history(C,H),
	s_depth(H,D).

theory_flag(C,D):-
	C=self(_),
	history(C,H),
	theoryIndicator(H,D).



% modifying
% -----------------------------------------------------------------------------

history( F, self(_C,C), self(_C,C1) ):-
	change_attr(history,C,C1,F).

constraint( Ct, self(_C,C), self(_C,C1) ):-
	change_attr(constraint,C,C1,Ct).

redundancy_info(I, self(_C,C),self(_C,C1)):-
	change_attr(redundancy_info,C,C1,I).

kind(K, self(_C,C),self(_C,C1)):-
	change_attr(kind,C,C1,K).

complexityBound(CB, self(_C,C),self(_C,C1)):-
	asSimpleClause(CB,CB1),
	change_attr(complexityBound,C,C1,CB1).

history_and_number( H, C1, C ):-
        create_history( H, History ),
	newClauseNumber( N ),
        number( N, C1, C2 ),
        history( History, C2, C ), 
	!.
history_and_number( _, _, _ ):-
        raise_exception(saturate_error("~nerror in history_and_number",[])).

as_new_axiom(C1,C):-
	initial_history( theory, H ),
	newClauseNumber( N ),
	number( N, C1, C2 ),
	history( H, C2, C3 ),
	C=C3.

as_new_goal(C1,C):-
	initial_history( (goal), H ),
	newClauseNumber( N ),
	number( N, C1, C2 ),
	history( H, C2, C3 ),
	C=C3.


%	antecedent(C3,A),
%	succedent(C3,S),
%	analyse_clause(A,S,Kind),
%	kind(Kind,C3,C).



% converting
%------------------------------------------------------------------------------
	
as_tptp(Cl,input_clause(from_saturate,Flag,Ls)):-
	Cl=self(_),
	antecedent(Cl,A),
	succedent(Cl,S),
	elim_equivalences(A,S,A1,S1),
	theory_flag(Cl,TF),
	(TF=theory ->
	     Flag=axiom
	;
	     Flag=theorem
	),
	maplist(as_tptp_literal(-),A1,Negs),
	maplist(as_tptp_literal(+),S1,Poss),
	append(Negs,Poss,Ls),
	!.
	

as_spass(Cl,'$spass'(Constraint,Ante,ST)):-
	Cl=self(_),
	antecedent(Cl,A),
	succedent(Cl,S),
	elim_special_relations(A,S,A1,S1),
	maplist(atom_as_term,A1,AT),
	maplist(atom_as_term,S1,ST),
	split(sortAtom,AT,Constraint,Ante),
	!.
	

as_dfg(Mode,Cl,'$dfg'(C,TheoryFlag)):-
	theory_flag(Cl,TheoryFlag),
	antecedent(Cl,A),
	succedent(Cl,S),
	elim_special_relations(A,S,A1,S1),
	maplist(atom_as_term,A1,A2),
	maplist(deBruijn,A3,A2),
	maplist(as_dfg_formula(Mode),A3,A4),
	maplist(atom_as_term,S1,S2),
	maplist(deBruijn,S3,S2),
	maplist(as_dfg_formula(Mode),S3,S4),
	maplist(dfg_equal(Mode),A4,B),
	maplist(dfg_equal(Mode),S4,H),
	list_of_vars((H,B),Vs),
	maplist(apply(not),B,NB),
	append(NB,H,Literals),
	(Literals=[L] ->
	    Disj=L
	;
	    Disj=..[or|Literals])
	,
	(Vs=[] ->
	     C=clause(Disj)
	;
	     C=clause(forall(Vs,Disj))
	).


as_prolog_sequent(Cl,(B->H)):-
	antecedent(Cl,A),
	succedent(Cl,S),
	elim_equivalences(A,S,A1,S1),
	maplist(atom_as_term,A1,A2),
	maplist(atom_as_term,S1,S2),
	maplist(dfg_equal(dfg),A2,B),
	maplist(dfg_equal(dfg),S2,H).

as_prolog_literals(Cl,Lits):-
	antecedent(Cl,A),
	succedent(Cl,S),
	elim_equivalences(A,S,A1,S1),
	maplist(atom_as_term,A1,A2),
	maplist(atom_as_term,S1,S2),
	maplist(mk_negative,A2,Neg),
	append(Neg,S2,Lits).

dual_sequent(Cl,[S2,A2]):-
	antecedent(Cl,A),
	succedent(Cl,S),
	elim_equivalences(A,S,A1,S1),
	maplist(atom_as_term,A1,A2),
	maplist(atom_as_term,S1,S2).



mk_negative(A,-A).




dfg_equal(dfg,L=R,equal(L,R)):-
	!.
%dfg_equal(A,A1):-
%	A=..[P,S,T],
%	\+ special_relation(P),
%	A1=..[P,ttt(S,T)],
%	!.
dfg_equal(_,A,A).

:-use_module(adt_term,[binder/5,external_binder/4]).
:-use_module(adt_precedence,[special_symbol_conv/2]).

as_dfg_formula(_,X,X):-
	var(X),
	!.
as_dfg_formula(dfg,L=R,equal(L1,R1)):-
	!,
	as_dfg_formula(dfg,L,L1),
	as_dfg_formula(dfg,R,R1).
as_dfg_formula(dfg,S,T):-
	binder(S,_Skt,X,A,Op),
	!,
	as_dfg_formula(dfg,A,A1),
	external_binder(T,[X],A1,Op),
	!.
as_dfg_formula(Mode,L=>R,implies(L1,R1)):-
	!,
	as_dfg_formula(Mode,L,L1),
	as_dfg_formula(Mode,R,R1).
as_dfg_formula(Mode,L<=>R,equiv(L1,R1)):-
	!,
	as_dfg_formula(Mode,L,L1),
	as_dfg_formula(Mode,R,R1).
as_dfg_formula(Mode,L==R,equiv(L1,R1)):-
	!,
	as_dfg_formula(Mode,L,L1),
	as_dfg_formula(Mode,R,R1).
as_dfg_formula(Mode,and(L,R),and(L1,R1)):-
	!,
	as_dfg_formula(Mode,L,L1),
	as_dfg_formula(Mode,R,R1).
as_dfg_formula(Mode,or(L,R),or(L1,R1)):-
	!,
	as_dfg_formula(Mode,L,L1),
	as_dfg_formula(Mode,R,R1).
as_dfg_formula(Mode,(~L),not(L1)):-
	!,
	as_dfg_formula(Mode,L,L1).
as_dfg_formula(Mode,F,G):-
	(Mode=dfg,functor(F,O,N),special_symbol_conv(O/N,Odfg/N) ->
	    F=..[O|Args],
	    maplist(as_dfg_formula(Mode),Args,Argsdfg),
	    G=..[Odfg|Argsdfg]
	;
	    mapargs(as_dfg_formula(Mode),F,G)
	).



/*
elim_equivalences(A,[],A,[]).
elim_equivalences(A,[E|S],[E1|A1],[E2|S2]):-
	get_atom(E,==,L,R),
	R\=='$true$',
	!,
	create_atom(L=='$true$',E1),
	create_atom(R=='$true$',E2),
	elim_equivalences(A,S,A1,S2).
elim_equivalences(A,[E|S],A1,[E|S2]):-
	elim_equivalences(A,S,A1,S2).

*/

sortAtom(T):-
	T=..[_,X],
	var(X).

% clause size and weight
%------------------------------------------------------------------------------

size_and_weight( C,6,W):-
	currentTheory(ac),
	kind(C,commutativity(_,_)),
	segment_length(L),
	W is 30*L,
	!.
size_and_weight( C,10,W):-
	currentTheory(ac),
	kind(C,associativity(_,_)),
	segment_length(L),
	W is 30*L,
	!.
% the general case of size_and_weight is defined in the superclass




% simplification issues
%------------------------------------------------------------------------------

dont_simplify(C):-
	kind(C,K),
	dont_simplify_cl(K).

dont_demodulate_with(Cl):-
	kind(Cl,N),
	dont_demo_with(N).



% clause structure
% -----------------------------------------------------------------------------

operators( C, L3 ):-
	C=self(_),
	antecedent(C,A),
	succedent(C,S),
	constraint(C,Ct),
	get_list_of_atoms_operators( A, LA ),
	get_list_of_atoms_operators( S, LS ),
	constraint_operators( Ct, LCt ),
	append( LA, LS, L1 ),
	append( L1, LCt, L2 ),
	sort( L2, L3 ),
	!.        % essential

subterm(C,Sub):-
	C=self(_),
	(   antecedent(C,L)
	;   succedent(C,L)
	),
	member(A,L),
	get_atom_members(A,S,T),
	(    path_arg1(_S,S,Sub)
	;    path_arg1(_S,T,Sub)
	),
	(nonvar(Sub) ->
	     Sub\=='$true$',
	     Sub\=='$min',
	     functor(Sub,F,N),
	     \+ is_predicate(F/N)
	;    true).

subterms(C,Subs):-
	setof0(Sub,subterm(C,Sub),Subs).



% *** this should go somewhere else

%         conjunctive normal forms
%         sorting achieves the side-effect of factoring
%--------------------------------------------------------------------------


top_cnf([]).

bottom_cnf([[]]).

and_cnf(C1,C2,CS):-
	append(C1,C2,C3),
	sort(C3,CS).         % here one should check for subsumption

or_cnf([],_K,[]):-
	!.
or_cnf(_K,[],[]):-
	!.
or_cnf(K,[D|Ds],KS1):-
	sort(D,DS),
	or_conj_disj(K,DS,KD),
	or_cnf(K,Ds,KDs),
	append(KD,KDs,K1),
	sort(K1,KS1).
	
or_conj_disj([],_,[]).
or_conj_disj(K,[],K):-
	!.
or_conj_disj([D|Ds],D1,[DS1|D1s]):-
	append(D,D1,DD1),
	sort(DD1,DS1),
	or_conj_disj(Ds,D1,D1s).

and_cnf_atom(K,A,[[A]|K]).

or_cnf_atom(K,A,K1):-
	or_cnf(K,[A],K1).

and_cnf_literal(K,A,[[A]|K]).

or_cnf_literal(K,A,K1):-
	or_cnf(K,[A],K1).

not_cnf([],[[]]).
not_cnf([[]],[]).
not_cnf([A|B],NotAOrNotB):-
	not_disj(A,NotA),
	not_cnf(B,NotB),
	or_cnf(NotA,NotB,NotAOrNotB).


not_disj(A,NotA):-
	maplist(negateLitAsClause,A,NotA).

negateLitAsClause([+|A],[[-|A]]).
negateLitAsClause([-|A],[[+|A]]).




simplify_cnf([],[]).
simplify_cnf([C|Cs],CsS):-
	taut(C),
	!,
	simplify_cnf(Cs,CsS).
simplify_cnf([C|Cs],CsS):-
	member(D,Cs),
	subs(D,C),
	!,
	simplify_cnf(Cs,CsS).
simplify_cnf([C|Cs],CsS):-
	select(D,Cs,Cs1),
	subs(C,D),
	!,
	simplify_cnf([C|Cs1],CsS).
simplify_cnf([C|Cs],[C|CsS]):-
	simplify_cnf(Cs,CsS).


taut(C):-
	select([P|A],C,C1),
	member([P1|B],C1),
	A==B,
	dual(P,P1).

dual(-,+).
dual(+,-).


subs([],_C2).
subs([L|C],C2):-
	select(L1,C2,C21),
	equal(L,L1),
	subs(C,C2).

	



%	printing
%------------------------------------------------------------------------------
%:- mymultifile portray/1.

user:portray( X ):-
        is_object(X,self(R1,C)),
	history(X,H),
	nonvar(H),
	classes:printClause( self(R1,C) ).


% private
% -----------------------------------------------------------------------------



as_tptp_literal(-,A,'--'(T1)):-
	atom_as_term(A,T),
	(T=equal(U,V),\+ congruence_relation_symbol(equal) ->
	     T1=eq(U,V)
	;
	     T1=T
	).
as_tptp_literal(+,A,'++'(T1)):-
	atom_as_term(A,T),
	(T=equal(U,V),\+ congruence_relation_symbol(equal) ->
	     T1=eq(U,V)
	;
	     T1=T
	).



printClause( Cl ):-
	Cl=self(_,C),
	attr(constraint,C,Ct),
	\+ \+ printSimpleClause(Cl,Ct),
	!.

:-use_module(props,[special_relation/1]).

dont_simplify_cl(reflexivity(R)).
dont_simplify_cl(irreflexivity(_)).
dont_simplify_cl(symmetry(_)).
dont_simplify_cl(composition(R1,R2,R3)):-
	transitive_relation_symbol(R1), special_relation(R1),
	transitive_relation_symbol(R2), special_relation(R2),
	transitive_relation_symbol(R3), special_relation(R3).
dont_simplify_cl(monotonicity(_,_,R,_,R)):-
	partial_congruence_relation_symbol(R).
dont_simplify_cl(totality(_,_)).
dont_simplify_cl(density(_,_,_)).
dont_simplify_cl(reflexive_closure1(R,_S)):-
	total_ordering_relation_symbol(R).
dont_simplify_cl(reflexive_closure2(R,_S)):-
	total_ordering_relation_symbol(R).
dont_simplify_cl(noendL(_)).
dont_simplify_cl(noendR(_)).
dont_simplify_cl(commutativity(_,_)).
dont_simplify_cl(associativity(_,_)).
dont_simplify_cl(permutativity(_,_,_)):- \+ currentTheory(ac).
dont_simplify_cl(finite_domain(_,R)):-
	congruence_relation_symbol(R).



dont_demo_with(composition(_,_,_)).
dont_demo_with(totality(_,_)).
dont_demo_with(density(_,_,_)).
dont_demo_with(reflexive_closure1(R,_S)):-
	total_ordering_relation_symbol(R).
dont_demo_with(reflexive_closure2(R,_S)):-
	total_ordering_relation_symbol(R).
%dont_demo_with(noendL(_)).
%dont_demo_with(noendR(_)).
dont_demo_with(commutativity(_,_)):-currentTheory(ac).
dont_demo_with(associativity(_,_)):-currentTheory(ac).





