%   Package: errno
%   Author : whoever wrote /usr/include/errno.h
%   Updated: 1/21/91
%   Purpose: table of UNIX error codes for Prolog

%   Generated from licensed UNIX sources; prelude and postlude are
%   Copyright (C) 1987, Quintus Computer Systems, Inc.  All rights reserved.

%   The entries in this table are errno(Index, TAG, Gloss)
%   where Index is the value stored in the C variable errno by various
%   system calls, TAG is the word which appears in the UNIX manuals,
%   and Gloss is a short "English" description of the error.
%   We rely on /usr/include/sys/errno.h containing this information in
%   a standardised layout.

:- module(errno, [
	errno/2,
	errno/3,			% used by library(files)
	fileerrno/2			% used by library(files)
   ]).

sccs_id('"@(#)91/01/21 errno.ed    61.1"').


%   fileerrno(+Error, +Goal)
%   checks for a non-zero error code.  If the error code is non-zero,
%   the action depends on the setting of the 'fileerrors' flag.  When
%   the flag is on, the error code is looked up in the table of error
%   messages, a fuller text is printed out, and execution is aborted.
%   When the flag is off, fileerrno fails quietly.

fileerrno(0, _) :- !.		% ok
fileerrno(_, _) :-		% nofileerrors?
	prolog_flag(fileerrors, off),
	!,
	fail.
fileerrno(Error, Goal) :-
	errno(Error, TAG, Gloss),
	!,
	format(user_error, '~N! UNIX error ~w (~w)~n! Goal: ~q~n',
	    [TAG, Gloss, Goal]),
	abort.
fileerrno(Error, Goal) :-
	format(user_error, '~N! Unknown error ~w~n! Goal: ~q~n',
	    [Error, Goal]),
	abort.



%   errno(+Errno, +Goal)
%   checks for a non-zero error code.  If the error code is non-zero,
%   it is looked up in the table of UNIX error messages, and a fuller
%   text is printed out.  If it isn't a UNIX error either (or one of
%   the things we've added to the table), it just prints the code.

errno(0, _) :- !.		% ok
errno(Error, Goal) :-
	errno(Error, TAG, Gloss),
	!,
	format(user_error, '~N! UNIX error ~w (~w)~n! Goal: ~q~n',
	    [TAG, Gloss, Goal]),
	fail.			% should be more drastic
errno(Error, Goal) :-
	format(user_error, '~N! Unknown error ~w~n! Goal: ~q~n',
	    [Error, Goal]),
	fail.


%   The auxiliary table errno/3 is NOT intended for general use.
%   library(files) accesses it; that may be redesigned.

errno( 0,	'EOK',		'operation successful').
errno(1,	'EPERM',	'Not owner').
errno(2,	'ENOENT',	'No such file or directory').
errno(3,	'ESRCH',	'No such process').
errno(4,	'EINTR',	'Interrupted system call').
errno(5,	'EIO',	'I/O error').
errno(6,	'ENXIO',	'No such device or address').
errno(7,	'E2BIG',	'Arg list too long').
errno(8,	'ENOEXEC',	'Exec format error').
errno(9,	'EBADF',	'Bad file number').
errno(10,	'ECHILD',	'No children').
errno(11,	'EAGAIN',	'No more processes').
errno(12,	'ENOMEM',	'Not enough core').
errno(13,	'EACCES',	'Permission denied').
errno(14,	'EFAULT',	'Bad address').
errno(15,	'ENOTBLK',	'Block device required').
errno(16,	'EBUSY',	'Mount device busy').
errno(17,	'EEXIST',	'File exists').
errno(18,	'EXDEV',	'Cross-device link').
errno(19,	'ENODEV',	'No such device').
errno(20,	'ENOTDIR',	'Not a directory').
errno(21,	'EISDIR',	'Is a directory').
errno(22,	'EINVAL',	'Invalid argument').
errno(23,	'ENFILE',	'File table overflow').
errno(24,	'EMFILE',	'Too many open files').
errno(25,	'ENOTTY',	'Not a typewriter').
errno(26,	'ETXTBSY',	'Text file busy').
errno(27,	'EFBIG',	'File too large').
errno(28,	'ENOSPC',	'No space left on device').
errno(29,	'ESPIPE',	'Illegal seek').
errno(30,	'EROFS',	'Read-only file system').
errno(31,	'EMLINK',	'Too many links').
errno(32,	'EPIPE',	'Broken pipe').
errno(33,	'EDOM',	'Argument too large').
errno(34,	'ERANGE',	'Result too large').
errno(35,	'EWOULDBLOCK',	'Operation would block').
errno(36,	'EINPROGRESS',	'Operation now in progress').
errno(37,	'EALREADY',	'Operation already in progress').
errno(38,	'ENOTSOCK',	'Socket operation on non-socket').
errno(39,	'EDESTADDRREQ',	'Destination address required').
errno(40,	'EMSGSIZE',	'Message too long').
errno(41,	'EPROTOTYPE',	'Protocol wrong type for socket').
errno(42,	'ENOPROTOOPT',	'Protocol not available').
errno(43,	'EPROTONOSUPPORT',	'Protocol not supported').
errno(44,	'ESOCKTNOSUPPORT',	'Socket type not supported').
errno(45,	'EOPNOTSUPP',	'Operation not supported on socket').
errno(46,	'EPFNOSUPPORT',	'Protocol family not supported').
errno(47,	'EAFNOSUPPORT',	'Address family not supported by protocol family').
errno(48,	'EADDRINUSE',	'Address already in use').
errno(49,	'EADDRNOTAVAIL',	'Can''t assign requested address').
errno(50,	'ENETDOWN',	'Network is down').
errno(51,	'ENETUNREACH',	'Network is unreachable').
errno(52,	'ENETRESET',	'Network dropped connection on reset').
errno(53,	'ECONNABORTED',	'Software caused connection abort').
errno(54,	'ECONNRESET',	'Connection reset by peer').
errno(55,	'ENOBUFS',	'No buffer space available').
errno(56,	'EISCONN',	'Socket is already connected').
errno(57,	'ENOTCONN',	'Socket is not connected').
errno(58,	'ESHUTDOWN',	'Can''t send after socket shutdown').
errno(59,	'ETOOMANYREFS',	'Too many references: can''t splice').
errno(60,	'ETIMEDOUT',	'Connection timed out').
errno(61,	'ECONNREFUSED',	'Connection refused').
errno(62,	'ELOOP',	'Too many levels of symbolic links').
errno(63,	'ENAMETOOLONG',	'File name too long').
errno(64,	'EHOSTDOWN',	'Host is down').
errno(65,	'EHOSTUNREACH',	'No route to host').
errno(66,	'ENOTEMPTY',	'Directory not empty').
errno(67,	'EPROCLIM',	'Too many processes').
errno(68,	'EUSERS',	'Too many users').
errno(69,	'EDQUOT',	'Disc quota exceeded').
errno(70,	'ESTALE',	'Stale NFS file handle').
errno(71,	'EREMOTE',	'Too many levels of remote in path').
errno(72,	'ENOSTR',	'Device is not a stream').
errno(73,	'ETIME',	'Timer expired').
errno(74,	'ENOSR',	'Out of streams resources').
errno(75,	'ENOMSG',	'No message of desired type').
errno(76,	'EBADMSG',	'Trying to read unreadable message').
errno(77,	'EIDRM',	'Identifier removed').
errno(78,	'EDEADLK',	'Deadlock condition').
errno(79,	'ENOLCK',	'No record locks available').
errno(80,	'ENONET',	'Machine is not on the network').
errno(81,	'ERREMOTE',	'Object is remote').
errno(82,	'ENOLINK',	'the link has been severed').
errno(83,	'EADV',	'advertise error').
errno(84,	'ESRMNT',	'srmount error').
errno(85,	'ECOMM',	'Communication error on send').
errno(86,	'EPROTO',	'Protocol error').
errno(87,	'EMULTIHOP',	'multihop attempted').
errno(88,	'EDOTDOT',	'Cross mount point (not an error)').
errno(89,	'EREMCHG',	'Remote address changed').

% Errors specific to the Quintus Prolog library

errno(996,	'EUNIMP',	'Operation not implemented').
errno(997,	'ENOTARCH',	'Archive not in ar(5) format').
errno(998,	'ENOTMEMB',	'Member not present in archive').
errno(999,	'EUNKVAR',	'$variable not defined in environment').

% Error specific to VMS

errno(65535,	'EVMSERR',	'Untranslateable VMS error').
