:- module(polyC, [ 
	simplify/2,
	simplify_light/2,
	simplify_med/2,
	simplify_constraint_vars/3
   ]).

:- use_module(sto, [
	greater/2,
	greater_op_sto/2,
	greaterEqual/2
   ]).

:- use_module(lpoC,[]).

:- use_module(genericC,[]).


:- use_module(proofState, [
        s_con/2,
	record_s_con/2
   ]).


:- use_module(props, [
        memo_constraints/1
   ]).


:- use_module(constraintSolverBasics, [
	implies/2,
	satisfiable_ineq/1
   ]).

:- use_module(orderConstraints, [
	true/1,
	delete_implying_ands/2

   ]).

:- use_module(equationalTheory, [
	equal/2

   ]).

:- use_module(myterms, [
	apply_substitution/3,
	list_of_vars/2,
	skolemized/3
   ]).


:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(maplist), [
	maplist/2
   ]).


:- use_module(poly,[term_to_poly/2]).


%:- use_module(thisProlog, [timeOut/3]).

:- use_module(library(clpr)).



%------------------------------------------------------------------------------
% CONSTRAINT SIMPLIFICATION
%------------------------------------------------------------------------------

simplify_constraint_vars(A,B,C):-
	constraintSolverBasics:simplify_constraint_vars(A,B,C).

%------------------------------------------------------------------------------
% simplify_constraint_light( C1, C2 ): 
% 	efficient method that simplifies constraints, although not totally, i.e.
% 	if C1 is insatisfiable, then C2 will not always be syntactically bottom
%------------------------------------------------------------------------------

simplify_light( C,C1 ):-
	lpoC:simplify_light(C,C1).

%------------------------------------------------------------------------------
% simplify_constraint_med( C1, C2 ): 
% 	simplifies a constraint at medium level
% 	if C1 is insatisfiable, then C2 will not necessarily 
%	be syntactically bottom
%------------------------------------------------------------------------------

simplify_med( C, D ):- simplify_light( C, D ).



%------------------------------------------------------------------------------
% simplify( C1, C2 ): 
% 	simplifies a constraint totally
% 	e.g. if C1 is insatisfiable, then C2 will be syntactically bottom
%------------------------------------------------------------------------------

simplify( C, C ):- true( C ), !.
simplify( [], [] ).
simplify( C, [conj([],[],[])] ):-
	member( conj([],[],[]), C ), !.	
simplify( C, C2):-
	(memo_constraints(on) ->
	     lpoC:normalize_constraint(C,CN),  % to normalize a little before caching
	     skolemized(CN,CS,Subst),
	     hash_term(CS,Key),
	     (s_con(Key,(CS, CS1)) ->
%	          write('|'),ttyflush,
		  skolemized(C2,CS1,Subst)
	     ;
	          simplify_n(CN,C2),
		  record_s_con(Key,(CN,C2))
	     )
	;
	     simplify_n(C,C2)
	).




simplify_n(CN,C2):-
%	format('~Nworking on constraint ~p',[CN]),ttyflush,
%	write('['),ttyflush,
%	(timeOut(simplify1(CN,C1),5000,success) ->
%	     true
%	; 
%	     format('~Ntime-out on constraint ~p',[CN]),
%	     C1=CN
%	),

	simplify1(CN,C1),

%	write(']'),ttyflush,

	(member( conj([],[],[]), C1 ) ->
	     C2 = [conj([],[],[])]
	;	 
	     delete_implying_ands( C1, C2 )
	), 
	!.



simplify1([],[]).
simplify1( [A|C], [A|C1] ):-
	\+ \+ simplified_satisfiable_and( A ),
	simplify1( C, C1 ), 
	!. 
simplify1( [_|C], C1 ):-
	simplify1( C, C1 ).


simplified_satisfiable_and( conj(I,E,O) ):-
	genericC:partially_solved_form(2,I,I1),
	!,
	process(E),
	process(I1),
	!.

process([]).
process([S=T|Is]):-
	term_to_poly(S,SP),
	term_to_poly(T,TP),
	{SP=TP},
	process(Is).
process([S>=T|Is]):-
	term_to_poly(S,SP),
	term_to_poly(T,TP),
	{SP>=TP},
	process(Is).
process([S>T|Is]):-
	term_to_poly(S,SP),
	term_to_poly(T,TP),
	{SP>=TP},                 % > is quasi-ordering
	process(Is).

