:-module(proofState,

	 [ props/1,
	   clause/1,
	   tempClause/1,
	   rule/2,
	   predicates/1,
	   firstPredicatePrecedence/1,
	   precedence/1,
	   operators/1,
	   proved_by_case_an/2,
	   s_con/2,
	   axiom/1,
	   proof_file/2,
	   clause_counter/1,
	   module_counter/1,
	   inferenceComputed/2,
	   splitCase/3,
	   nf/4,
	   nf_by_current_rules/3, 
	   irreducible_by_current_rules/1,
	   theory_symbol/1,
	   goal_symbol/1,

	   abbreviationFor/3,
	   
	   rewriteFlag/1,
	   setRewriteFlag/1,

	   newClauseNumber/1,
	   newModuleNumber/1,

	   create_new_axiom/2,
	   create_new_axiom/3,

	   recordAbbreviation/3,
	   recordClause/1,
	   recordGoalOperators/1,
	   recordTempClause/1,
	   recordRule/2,
	   recordProofOfClause/1,
	   recordIrreducibility/1,
	   recordProofByCaseAnalysis/2,
	   recordPrecedence/1,
	   recordFirstPredicatePrecedence/1,
	   recordPredicate/1,
	   recordPredicate/2,
	   recordPredicates/1,
	   recordOperator/1,
	   recordOperator/2,
	   recordOperators/1,
	   recordProp/1,
	   recordProps/1,
	   record_s_con/2,
	   recordSplitCase/3,
	   recordInferenceComputed/2,

	   retractTempClause/1,

	   deleteAllFirstPredicatePrecedences/0,
	   deleteAllPrecedences/0,
	   deleteAllRules/0,
	   deleteAllTempClauses/0,

	   deleteIrreducibility/2,
	   deleteNormalform/2,

	   deleteRule/2,
	   deleteClause/1,
	   deleteClause/2,
	   deleteOperators/1,

	   addTransformedClauses/4,
	   transformClauses/4,

	   initialize/0



	 ]).


:- meta_predicate 
	deleteIrreducibility(+,:), 
	deleteNormalform(+,:),
	deleteClause(+,:),
	addTransformedClauses(:,:,+,-),
	transformClauses(:,:,+,-).
		 
:- use_module(classes,[
        as_new_axiom/2,
	createClause/2,
        number/2,
	kind/3
   ]).


:- use_module(saturation,[update_minimalWeight/1]).

:- use_module(propLogic,[propLogicInit/0]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(maplist), [
	maplist/2
   ]).

:- dynamic 
        abbreviationFor/3,
	props/1,
	clause/1,
	tempClause/1,
	rule/2,
	predicates/1,
	firstPredicatePrecedence/1,
	theory_symbol/1,
	goal_symbol/1,
	precedence/1,
	operators/1,
	proved_by_case_an/2,
	s_con/2,
	rewriteFlag/1,
	axiom/1,
	proof_file/2,
	clause_counter/1,
	module_counter/1,
	inferenceComputed/2,
	splitCase/3,
	nf/4,
	nf_by_current_rules/3, 
	irreducible_by_current_rules/1.


initialize:-
	retractall(abbreviationFor(_,_,_)),
	retractall(props(_)),
	retractall(clause(_)),
	retractall(tempClause(_)),
	retractall(rule(_,_)),
	retractall(predicates(_)),
	retractall(firstPredicatePrecedence(_)),
	retractall(theory_symbol(_)),
	retractall(goal_symbol(_)),
	retractall(precedence(_)),
	retractall(operators(_)),
	retractall(proved_by_case_an(_,_)),
	retractall(s_con(_,_)),
	retractall(rewriteFlag(_)),
	retractall(axiom(_)),
	retractall(proof_file(_,_)),
	retractall(clause_counter(_)),
	retractall(module_counter(_)),
	retractall(inferenceComputed(_,_)),
	retractall(splitCase(_,_,_)),
	retractall(nf(_,_,_,_)),
	retractall(nf_by_current_rules(_,_,_)),
	retractall(irreducible_by_current_rules(_)),

	assert(clause_counter( 1 )),
	assert(module_counter( 1 )),
	assert(props([])),
	assert(operators([])),
	assert(predicates([])),

	propLogicInit,
	dynamicCode:initialize.











newClauseNumber(N):-
        (retract( clause_counter(N) ) -> true),
        N1 is N + 1,
        asserta( clause_counter(N1) ).


newModuleNumber( N ):-
        retract( module_counter(N) ),
        N1 is N + 1,
        asserta( module_counter(N1) ), !.
	
	
create_new_axiom(A,S):-
	createClause([A,S],C),
	as_new_axiom(C,C1),
	assertz(clause(C1)).

create_new_axiom(A,S,Kind):-
	createClause([A,S],C),
	as_new_axiom(C,C1),
	kind(Kind,C1,C2),
	assertz(clause(C2)).


recordAbbreviation(L,Sign,P):-
	assertz(abbreviationFor(L,Sign,P)).	


recordClause(R):-
	assertz(clause(R)).	

recordTempClause(R):-
	assertz(tempClause(R)).	

retractTempClause(R):-
	retract(tempClause(R)).

recordRule(N,R):-
	assertz(rule(N,R)).


recordProofOfClause(C):-
	number(C,N),
	(proof_file(N,_) -> 
	     true
	;
	     assertz( proof_file(N,C))
	).

recordIrreducibility(S):-
	assert(irreducible_by_current_rules(S)).

recordPrecedence(P):-
	assert(precedence(P)).


recordOperators(P):-
	assert(operators(P)).

recordOperator(P,Th):-
	recordOperator(P),
	(Th=(theory) ->
	     (theory_symbol(P) ->
		  true
	     ;
	          assert(theory_symbol(P))
	     )
	;Th= (goal) ->
	     (goal_symbol(P) ->
		  true
	     ;
	          assert(goal_symbol(P))
	     )
	;
	     true
	).


recordPredicate(P,Th):-
	recordPredicate(P),
	(Th=(theory) ->
	     (theory_symbol(P) ->
		  true
	     ;
	          assert(theory_symbol(P))
	     )
	;Th= (goal) ->
	     (goal_symbol(P) ->
		  true
	     ;
	          assert(goal_symbol(P))
	     )

	;
	     true
	).

recordGoalOperators(Ops):-
	maplist(assert_goal_symbol,Ops).


assert_goal_symbol(O):-
	goal_symbol(O),
	!.
assert_goal_symbol(O):-
	assert(goal_symbol(O)).



recordOperator(P):-
	(operators(L) ->
	     true
	;
	     L=[]
	),
	(member(P,L) ->
	     true
	;
	     retractall(operators(_)),
	     assert(operators([P|L]))
	).



recordFirstPredicatePrecedence(P):-
	assert(firstPredicatePrecedence(P)).

recordPredicates(P):-
	assert(predicates(P)).

recordPredicate(P):-
	(predicates(L) ->
	     true
	;
	     L=[]
	),
	(member(P,L) ->
	     true
	;
		 
	     retractall(predicates(_)),
	     assert(predicates([P|L]))
	).


recordProofByCaseAnalysis(E,C):-
	assert( proved_by_case_an(E,C)).

recordProp(P):-
	retract(props(Ps)),
	merge_props(Ps,[P],Ps1),
	assert(props(Ps1)),
	!.

recordProps(P):-
	retract(props(Ps)),
	merge_props(Ps,P,Ps1),
	assert(props(Ps1)),
	!.



updated(op_complexity_(O,_),Qs):-
	member(op_complexity_(O,_),Qs),
	!.


merge_props([],Ps,Ps).
merge_props([P|Ps],Qs,Rs):-
	updated(P,Qs),
	!,
	merge_props(Ps,Qs,Rs).
merge_props([P|Ps],Qs,[P|Rs]):-
	merge_props(Ps,Qs,Rs).



record_s_con(A,B):-
	assert(s_con(A,B)).


recordInferenceComputed(R1,R2):-
	assert(inferenceComputed(R1,R2)).



recordSplitCase(P,A,S):-
	assert(splitCase(P,A,S)).



deleteAllRules:-
	retractall(rule(_,_)),
	retractall(clause(_)).

deleteAllFirstPredicatePrecedences:-
	retractall(firstPredicatePrecedence(_)).

deleteAllPrecedences:-
	retractall(precedence(_)).


deleteAllTempClauses:-
	retractall(tempClause(_)).

deleteOperators(O):-
	retract(operators(O)).


deleteRule(N,Ru):-
	(var(Ru) ->
	     retract(rule(N,Ru))
	;
	     retract(rule(N,_))
	).

deleteClause(O):-
	retract(clause(O)).



/* the following delete procedures delete conditionally
   the first parameter is the bound variable in the condition
   with is to be instantiated by the facts under consideration
*/

deleteClause(C,Condition):-
	clause(clause(C),_,Ref),
	(Condition -> erase(Ref)).

deleteIrreducibility(T,Condition):-
	clause(irreducible_by_current_rules(T),_,Ref),
	(Condition -> erase(Ref)).

deleteNormalform(T,Condition):-
	clause(nf_by_current_rules(_S,T,_),_,Ref),
	(Condition -> erase(Ref)).

transformClauses(Condition,Transformation,C,D):-
	deleteClause(C,Condition),
	Transformation,
	assertz(clause(D)),
	fail.
transformClauses(_,_,_,_).
	
	

addTransformedClauses(Condition,Transformation,C,D):-
	clause(C),
	Condition,
	Transformation,
	assertz(clause(D)),
	fail.
addTransformedClauses(_,_,_,_).


setRewriteFlag(F):-
	(rewriteFlag(F) ->
	     true
	;
	     assert(rewriteFlag(F))
	).
