:- module(res,
	  [resolve/0,
	   resolve/1
	  ]).


:- use_module(equationalTheory, [
        currentTheory/1,
	equal/2,
	extension/4,
	match/2,
	match/4,
	unify/2
   ]).

:- use_module(library(basics), [member/2]).

:- use_module(library(terms), [subsumes_chk/2]).

:- use_module(library(maplist), [
	maplist/3
   ]).

:-use_module(library(flatten),[flatten/2]).

:- use_module(state).

:- use_module(commands,[state_as_clauses_and_precedence1/3]).

:- use_module(classes,[as_prolog_literals/2]).



resolve:-
	state_as_clauses_and_precedence1(Cls,_,_),
	maplist(as_prolog_literals,Cls,ClsL),
	statistics(runtime,[T1,_]),
	resolve(ClsL),
	statistics( runtime, [T2,_] ),
	T is T2-T1,
%	TS is T // 1000,
	format('~nTotal execution time: ~d msec.',[T]).



resolve(Cls):-
	state_from_clause_list(Cls,S),
	resolve([],S).

resolve(Old,S):-
	get_min_clause(S,C,S1),
	(subsumes(Old,C) ->
	    resolve(Old,S1)
	;
	    !,
	    resolve_cls(Old,C,S1,S2),
	    resolve([C|Old],S2)
	),
	!.
resolve(Old,_):-
	% no new clauses
	length(Old,N),
	format('Saturation Terminated.~n~p~n~p clauses.',[final_clauses(Old),N]).





resolve_cls(Ds,C,S,S1):-
	positive_clause(C),
	!,
	literals(C,CL),
	resolve_pos(Ds,CL,S,S1).
resolve_cls(Ds,C,S,S1):-
	literals(C,CL),
	resolve_notpos(Ds,CL,S,S1).

resolve_pos([D|Ds],C,S,S1):-
	literals(D,DL),
	resolve_cls_cls(C,DL,S,S2),
	resolve_pos(Ds,C,S2,S1).
resolve_pos([],_C,S,S).

resolve_notpos([D|Ds],C,S,S1):-
	(positive_clause(D) ->
	    literals(D,DL),
	    resolve_cls_cls(DL,C,S,S2)
	;
	    S2=S
	),
	resolve_notpos(Ds,C,S2,S1).
resolve_notpos([],_C,S,S).


resolve_cls_cls(C,D,S,S1):-
	C=[A|Ls],
	append(J1,[-B|J2],D),
	\+ \+ unify(A,B),
	!,
	copy_term((B,J1,J2),(B1,Res11,Res12)),
	copy_term((A,Ls),(A1,Res2)),
	A1=B1,
	flatten([Res11,Res12,Res2],Resolvent),
	( Resolvent=[] ->
	    raise_exception('empty clause generated')
	;   
	    clause_from_literals(Resolvent,RC),
	    add_clause(S,RC,S1)).
resolve_cls_cls(_C,_D,S,S).





subsumes(Ds,C):-
	clause_size(C,S),
	literals(C,CL),
	subsumes(Ds,CL,S).

subsumes([D|_Ds],CL,SC):-
	clause_size(D,SD),
	SD=<SC,
	literals(D,DL),
	\+ \+subsumes_literals(DL,CL),
	!.
subsumes([_D|Ds],CL,SC):-
	subsumes(Ds,CL,SC).


subsumes_literals([L|Ls],C):-
	append(_C1,[M|C2],_C),
	match(L,M),
	subsumes_literals(Ls,C).
subsumes_literals([],_):-
	!.
