%declarations%
:- module(rpo, [ 
	greater/2,
	greater_ground/2,
	greaterEqual/2,
	greater_op/2
   ]).

:- use_module(adt_precedence, [
	greater_op/2,
	is_first_predicate/1
   ]).

:- use_module(library(arg), [genarg/3]).
:- use_module(library(basics), [member/2]).
:- use_module(auxiliary, [delete_one312/3]).
:- use_module(library(occurs), [contains_var/2]).

:- use_module(adt_term, [term_size/2]).
:- use_module(myterms, [
	apply_subst/3,
	occurs_in/2
   ]).
:- use_module(mylists, [
	permutation/2]).
:- use_module(library(sets), [
	select/3
   ]).





%end%

%------------------------------------------------------------------------------
% greater( S, T ):
%	S is greater than T with respect to the recursive path ordering
%------------------------------------------------------------------------------


% new implementation

greater( S, _ ):-
	var1(S),
	!,
	fail.
greater( S, T ):-
	\+var1(T),
	!,
	functor(S,F,N),
	functor(T,G,M),
	greater(F,G,N,M,S,T).

greater( S, T ):-
	functor(S,F,_),
	(is_first_predicate(F) -> 
	     true
	;    contains_var(T,S)),
	!.
	

greater(F,F,N,_,S,T):-
	multisetStatus(F/N),
	!,
	S=..[_|Ss],
	T=..[_|Ts],
	msext(greater,Ss,Ts),
	!.
greater(F,F,N,M,S,T):-
	\+multisetStatus(F/N),
	(is_first_predicate(F) ->
	     lexicographic1(0,S,N,T,M)
	;     lexicographic(0,S,N,T,M)),
	!.
greater(F,G,_N,M,S,T):-
	(greater_op(F,G)->
	     greater_than_list(S,0,T,M)
	; % greater_op(G,F) or F=G 
             genarg(_,S,S1),
             greaterEqual( S1, T)), 
	!.
	


msext(O,U,V):-
	difference(U,V,UV),
	difference(V,U,VU),
	msext1(O,UV,VU).


difference(A,[],A).
difference(A,[X|Xs],B):-
	(select(Y,A,A1),equal(X,Y) ->
	     difference(A1,Xs,B)
	;difference(A,Xs,B)).

msext1(_,[],[]):-
	!,
	fail.
msext1(O,U,V):-
	\+ (member(T,V), \+ (member(S,U), P=..[O,S,T], P)).





lexicographic1(N,_,N,_,_M):-
	!,
	fail.
lexicographic1(M,_,_N,_,M):-
	!,
	fail.
lexicographic1(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     lexicographic1(I1,S,N,T,M)
	;greater(SI,TI)).
	

lexicographic(N,_,N,_,_M):-
	!,
	fail.
lexicographic(M,_,_N,_,M):-
	!,
	fail.
lexicographic(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     lexicographic1(I1,S,N,T,M)
	;greater(SI,TI) ->
	     greater_than_list(S,I1,T,M)).

greater_than_list(_,N,_,N):-
	!.
greater_than_list(S,I,T,M):-
	I1 is I+1,
	arg(I1,T,TI),
	greater(S,TI),
	greater_than_list(S,I1,T,M).



/* 
The RPO is a quasi-ordering in ehich two terms are equivalent
if they are equal up to permutation of arguments of function
symbols with multiset status.

unifiable(+S,+T)
is true when there exists a substitution making
S and T equal up to permutativity.

For mgu we repeat code from equationalTheory.
*/

unifiable(S,T):-
	\+ \+mgu([S=T],_).



mgu( L, U ):-
	mgu( L, [], U ),
	!.

mgu( [], U, U ).				% the end
mgu( [T1=T2|R], S, U ):-			% delete
	T1==T2,
	!,
	mgu( R, S, U ).
mgu( [T1=T2|R], S, U ):-			
	(var(T1) ->
	     (\+ occurs_in( T1, T2 ) ->           % occur check
		  apply_subst( T1=T2, R, R1 ), 
		  apply_subst( T1=T2, S, S1 ), 
		  mgu(R1, [T1=T2|S1], U )
	     )
	;var(T2) ->
	     (\+ occurs_in( T2, T1 ) ->           % occur check
		  apply_subst( T2=T1, R, R1 ),
		  apply_subst( T2=T1, S, S1 ),
		  mgu(R1, [T2=T1|S1], U )
	     )
	; % decompose
	     functor(T1,F,N),
	     functor(T2,F,N),
	     (multisetStatus(F/N) ->
	         permutedSubterms(T1,T1P),
		 equalArgs( N, T1P, T2, L/R )
	     ;	 equalArgs( N, T1, T2, L/R )),
	     mgu( L, S, U )
	).
mgu(T1=T2,S,U):-
	mgu([T1=T2],S,U).


multisetStatus(_/2).       % experimental


permutedSubterms(T,TP):-
	T=..[F|Ts],
	permutation(Ts,TPs),
	TP=..[F|TPs].
		  


% equalArgs(+I,+N,+S,+T,?E) is true when E is the list of unification
% problems between corresponding subterms of S and T between 
% indexes N-I+1 and N. E is represented in form of a diffence list
% E1/E2 to allow easy appending with surrounding unification problems

equalArgs(0,_,_,L/L):-
	!.
equalArgs(I,T1,T2,[A1=A2|L]/R):-
	arg(I,T1,A1),
	arg(I,T2,A2),
	I1 is I-1,
	equalArgs(I1,T1,T2,L/R).



var1(X):-
	var(X),
	!.
var1('$VAR'(_)).   % for skolemized variables




%------------------------------------------------------------------------------
% greaterEqual( S, T ):
%  S is greater than T with respect rpo or S is equal to T up to permutativity
%------------------------------------------------------------------------------

greaterEqual(T1,T2):- equal(T1,T2),!.
greaterEqual(_T1,T2):- '$min' == T2,!.
greaterEqual(_T1,T2):- '$true$' == T2,!.
greaterEqual(T1,T2):- greater(T1,T2),!.

equal(T1,T2):-
	\+ \+ 
	(  numbervars((T1,T2),1,_),
	   unifiable(T1,T2) ).


%------------------------------------------------------------------------------
% maximal_term( S, T, M ):
%------------------------------------------------------------------------------

maximal_term( S, T, S ):- greaterEqual( S, T ), !.
maximal_term( S, T, T ):- greater( T, S ), !.

%------------------------------------------------------------------------------
%  for skolemized terms we use just lpo again
%
%------------------------------------------------------------------------------



greater_ground(S,T):-
	term_size(S,N),
	term_size(T,M),
	(	N>M
	;	M=N,
		gt_ground(S,T)
		),
	!.


gt_ground( S, T ):-
	ground_functor(S,F,N),
	ground_functor(T,G,M),
	gt_ground(F,G,N,M,S,T).

gt_ground(F,F,N,M,S,T):-
	% no cut here! the two cases of greater/6 are not exclusive
	(ground_is_first_predicate(F) ->
	     ground_lexicographic1(0,S,N,T,M)
	;     ground_lexicographic(0,S,N,T,M)),
	!.
gt_ground(F,G,_N,M,S,T):-
	(gt_ground_op(F,G)->
	     ground_greater_than_list(S,0,T,M)
	; % gt_ground_op(G,F) or F=G
             ground_genarg(_,S,S1),
             geq_ground( S1, T)), 
	!.
	

ground_genarg(I,T,S):-
	nonvar(T),
	genarg(I,T,S).
	

ground_functor(T,T,0):-
	var(T),
	!.
ground_functor(T,F,N):-
	functor(T,F,N).

ground_is_first_predicate(F):-
	nonvar(F),
	is_first_predicate(F).




ground_lexicographic1(N,_,N,_,_M):-
	!,
	fail.
ground_lexicographic1(M,_,_N,_,M):-
	!,
	fail.
ground_lexicographic1(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     ground_lexicographic1(I1,S,N,T,M)
	;gt_ground(SI,TI)).
	

ground_lexicographic(N,_,N,_,_M):-
	!,
	fail.
ground_lexicographic(M,_,_N,_,M):-
	!,
	fail.
ground_lexicographic(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     ground_lexicographic1(I1,S,N,T,M)
	;gt_ground(SI,TI) ->
	     ground_greater_than_list(S,I1,T,M)).

ground_greater_than_list(_,N,_,N):-
	!.
ground_greater_than_list(S,I,T,M):-
	I1 is I+1,
	arg(I1,T,TI),
	gt_ground(S,TI),
	ground_greater_than_list(S,I1,T,M).

geq_ground( T1, T2 ):- T1 == T2,!.
geq_ground( T1, T2 ):- gt_ground( T1, T2 ).

% ordering: first_predicates > variables > other symbols
gt_ground_op(F,G):-
	(var(F) ->
	     (var(G) -> F @> G; \+is_first_predicate(G))
	;    (var(G) -> is_first_predicate(F); greater_op(F,G))).

