:- use_module(library(lists)).
:- use_module(library(system)).
%:- load_files(library(detcheck),
%              [when(compile_time), if(changed)]).

% [(X,x), (Y,y), (X1,'x\''), (Y1, 'y\'')]
%c(p(s, X, Y), p(s, X1, Y1), [X>=0, Y>=0, X1=X+1, Y1=Y-2], 1).
%c(p(s, X, Y), p(s, X1, Y1), [X>=0, X1=X-1, Y1=Y+1], 2).


command2stream([], _, _) :- !.
command2stream([C], Var2Name, In) :-
	c2cvc(C, Var2Name, In), !.
command2stream([C|Cs], Var2Name, In) :-
	c2cvc(C, Var2Name, In), !, write(In, ' AND '), !,
	command2stream(Cs, Var2Name, In).

var_names2stream([], _) :- !.
var_names2stream([(_, Name)], Out) :-
	write(Out, Name), write(Out, ' : REAL;'),
	nl(Out),nl(Out).
var_names2stream([(_, Name)|Var2Names], Out) :-
	write(Out, Name), write(Out, ', '), !,
	var_names2stream(Var2Names, Out).

% Convert a (SICStus) constraint term into CVC expression and write it to stream.
c2cvc(Var, Var2Name, In) :-
	var(Var), !, look_up_name(Var, Var2Name, Name), !, 
	write(In, Name).
c2cvc(A, _, In) :-
	atomic(A), !, write(In, A).
c2cvc(+A, Var2Name, In) :-
	!, write(In, +), c2cvc(A, Var2Name, In).
c2cvc(-A, Var2Name, In) :-
	!, write(In, -), c2cvc(A, Var2Name, In).
c2cvc(OpAB, Var2Name, In) :-
	functor(OpAB, Op, 2), arg(1, OpAB, A), arg(2, OpAB, B),
	op2cvc(Op, CVC_Op), !, 
	c2cvc(A, Var2Name, In), write(In, CVC_Op), c2cvc(B, Var2Name, In).
c2cvc((A), Var2Name, In) :-
	write('('), c2cvc(A, Var2Name, In), !, write(')').


look_up_name(Var, [(V, Name)|_], Name) :-
	Var == V.
look_up_name(Var, [_|Var2Names], Name) :-
	look_up_name(Var, Var2Names, Name).
look_up_name(_, [], _) :-
	raise_exception(unknown_variable).
/*
C -->           Expr=:= Expr       equation 
              | Expr  =  Expr       equation 
              | Expr  <  Expr       strict inequation 
              | Expr  >  Expr       strict inequation 
              | Expr =<  Expr       nonstrict inequation 
              | Expr >=  Expr       nonstrict inequation 
              | Expr =\= Expr       disequation 

Expr -->     variable              Prolog variable 
            | number             floating point or integer 
            | + Expr             unary plus 
            | - Expr             unary minus 
            | Expr + Expr        addition 
            | Expr - Expr        subtraction 
            | Expr * Expr        multiplication 
            | Expr / Expr        division 
*/	    
op2cvc(Op, CVC_Op) :-
	look_up_cvc_op(Op, [(=:=,=), (=,=), (=\=,/=),
			    (<,<), (>,>), (=<,<=), (>=,>=),
			    (+,+), (-,-), (*,*), (/,/)], CVC_Op).
look_up_cvc_op(Op, [(Op, CVC_Op)|_], CVC_Op) :- !.
look_up_cvc_op(Op, [_|CVC_Ops], CVC_Op) :-
	look_up_cvc_op(Op, CVC_Ops, CVC_Op).
look_up_cvc_op(_, [], _) :-
	raise_exception(unknown_operation).


write_list([], _) :- !.
write_list([L|Ls], Out) :-
	write(Out, L), !, write_list(Ls, Out).

cvc_query(Out) :-
	read_line(Out, Line), length(Line, 11).


cvc_entails(_, [], _, _, _) :- !.
cvc_entails(As, Bs, Var2Names, In, Out) :-
	write(In, 'QUERY ('),
	command2stream(As, Var2Names, In),
	write(In, ') => ('),
	command2stream(Bs, Var2Names, In),
	write(In, ');'), nl(In),
	flush_output(In), !,
	cvc_query(Out).

cvc_start(In, Out, Err, Pid) :-
	cvc_start('~/dis/cvcl-0.1-alpha/bin/linux-i686/cvcl +int', In, Out, Err, Pid).
%	cvc_start('~/dis/cvclib/bin/linux-i686/cvc', In, Out, Err, Pid).
cvc_start(Cmd, In, Out, Err, Pid) :-
	exec(Cmd, [pipe(In), pipe(Out), pipe(Err)], Pid).

cvc_stop(In, Out, Err, Pid, Status) :-
	close(In), close(Out), close(Err), kill(Pid, 9), wait(Pid, Status).

cvc_init_vars(In, Var2Names) :-
	var_names2stream(Var2Names, In).


clp_benchmark(Max, Time, Spec) :-
	statistics(Spec, [Time1, _]),
	bb_put(count, 0),
	repeat,
	bb_get(count, C), C1 is C+1, bb_put(count, C1), 
	entails([X>=10, X>=40, Y=<0 ], [X>=1, Y<10]),
	C1 >= Max,
	statistics(Spec, [Time2, _]), Time is Time2-Time1.

cvc_benchmark(Max, Time) :-
	Var2Names = [(X,x), (Y,y), (X1,'x\''), (Y1, 'y\'')],
	statistics(runtime, [Time1, _]),
	bb_put(count, 0),
	cvc_start(In, Out, Err, Pid),
	cvc_init_vars(In, Var2Names), 
	repeat,
	bb_get(count, C), C1 is C+1, bb_put(count, C1), 
	cvc_entails([X>=10, X>=40, Y=<0 ], [X>=1, Y<10], Var2Names, In, Out), 
	C1 >= Max,
	cvc_stop(In, Out, Err, Pid, Status), 
	statistics(runtime, [Time2, _]), Time is Time2-Time1.
