:- module(state,
	  [add_clause/3,
	   clause_from_literals/2,
	   clause_list/2,
	   clause_size/2,
	   get_min_clause/3,
	   literals/2,
	   positive_clause/1,
	   state_from_clause_list/2,
	   init/1]
	 ).



:- use_module(adt_term, [term_size/2]).



:- use_module(library(heaps)).
:- use_module(library(basics), [member/2]).


init(S):-
	empty_heap(S).

add_clause(S,C,NewS):-
	clause_size(C,Size),
	add_to_heap(S,Size,C,NewS).

get_min_clause(S,C,NewS):-
	get_from_heap(S,_Size,C,NewS),
	!.


clause_list(S,Cls):-
	heap_to_list(S,Cls).


state_from_clause_list([CL|Cs],S):-
	state_from_clause_list(Cs,S1),
	clause_from_literals(CL,C),
	add_clause(S1,C,S).
state_from_clause_list([],S):-
	init(S).


clause_from_literals(Ls,(attrs(Type,Size),Ls)):-
	(\+member(-_,Ls) ->
	    Type=positive
	;
	    Type=non_positive
	),
	term_size(Ls,Size).

positive_clause((attrs(positive,_Size),_Ls)).


literals((_,Ls),Ls).

clause_size((attrs(_positive,Size),_Ls),Size).
