%declarations%
:- module(sto, [ 
	greater/2,
	greater_op_sto/2,
	greater_ground/2,
	greaterEqual/2
   ]).

:- use_module(adt_precedence, [
	greater_op/2,
	is_predicate/1
   ]).

:- use_module(library(arg), [genarg/3]).
:- use_module(library(occurs), [contains_var/2]).
:- use_module(sound_unify,[sound_unify/2]).

:- use_module(adt_term, [term_size/2, var1/1]).

:- use_module(lpo,[greater_ground/2]).

:- use_module(library(basics),[member/2]).

%end%

%------------------------------------------------------------------------------
% greater( S, T ):
%	S is greater than T with respect to the lpo
%       for the precedence greater_op_sto, cf. below.
%------------------------------------------------------------------------------

greater( S, _T ):-
	var1(S),
	!,
	fail.
greater('$true$',T2):-
	'$min'==T2,
	!.
greater('$true$',_):-
	!,
	fail.
greater( S, T ):-
	cfunctor(S,F,N),
	cfunctor(T,G,M),
	greater(F,G,N,M,S,T).







	

greaterEqual(T1,T2):- T1 == T2,!.
greaterEqual(_T1,T2):- '$min' == T2,!.
greaterEqual(T1,T2):- 
	'$true$' == T2,
	T1\=='$min'.
greaterEqual(T1,T2):- greater(T1,T2),!.


greater(F,G,N,M,S,T):-
	% note that F and or G can be arbirtary terms!
	F==G,
	lexicographic(0,S,N,T,M),
	!.
greater(F,G,N,M,S,T):-
	(greater_op_sto(F,G)->
	     greater_than_list(S,0,T,M)
	; % greater_op_sto(G,F) or F=G
             genarg(N,_,S,S1),
             greaterEqual( S1, T)), 
	!.







lexicographic(N,_,N,_,_M):-
	!,
	fail.
lexicographic(M,_,_N,_,M):-
	!,
	fail.
lexicographic(I,S,N,T,M):-
	I1 is I+1,
	arg(I1,S,SI),
	arg(I1,T,TI),
	(SI==TI ->
	     lexicographic(I1,S,N,T,M)
	;greater(SI,TI) ->
	     greater_than_list(S,I1,T,M)
	).





greater_than_list(_,N,_,N):-
	!.
greater_than_list(S,I,T,M):-
	I1 is I+1,
	arg(I1,T,TI),
	greater(S,TI),
	greater_than_list(S,I1,T,M).




% This is the precedence:
% All terms t are considered as constants ct such that
% ct>cs iff s is a proper subterm of t. There is a given
% (total) precedence on predicate symbols, and
% predicates are smaller than the constants ct.

greater_op_sto(F,G):-
	is_predicate(G/_),
	!,
	(is_predicate(F/_) ->
	     greater_op(F,G)
	;
	     true
	).
greater_op_sto(F,G):-
	\+ is_predicate(F/_),
	greater_sto(F,G).





greater_sto(S,T):-
	(S\==T, \+ var1(S), S=..[F|Ss] ->
	         (  \+var1(T),
		    T=..[G|Ts],
		    greater_op(F,G),
		    \+ (member(T1,Ts), \+ properSubterm(T1,S))
		 ;
		    member(S1,Ss),
		    greaterEqual_sto(S1,T)
		 )
	),
	!.

greaterEqual_sto(S,T):-
	S==T,
	!.
greaterEqual_sto(S,T):-
	greater_sto(S,T).


properSubterm(Sub,Term):-
	Sub\==Term,
	contains_var(Sub,Term).



% cfunctor is like functor, except that nonatom terms are considered
% as constants, and the returned functor may in fact be an arbitrary
% such term.
cfunctor(T,T,0):-
	var1(T),
	!.
cfunctor(T,P,N):-
	functor(T,P,N),
	is_predicate(P/N),
	!.
cfunctor(T,T,0).

genarg(N,I,T,TI):-
	(N>0 ->
	     genarg(I,T,TI)
	).


