% substitutiontree.pl
% Substitution trees; the interface module
% Author: pb

:- module(substitutiontree,[
	st_insert_term/4,
	st_insert_term/3,
	st_empty_tree/1,
	st_is_empty_tree/1,
	st_unify/3,
	st_unify/2,
	st_retrieve/4,
	st_retrieve/3,
	st_match/3,
	st_match/2,
	st_us_unify/3,
	st_us_unify/2,
	st_delete_term/3,
	st_delete_term/4
]).

:- use_module(st_insert).
:- use_module(st_retrieval).
:- use_module(st_delete).


% A substitution tree is of the form
% substitutiontree(U,ST,StarVars), where
%   U is the toplevel variable in the first substitution
%   ST is the tree itself
%   StarVars is a list of star variables occurring in ST
% ST is the substitutiontree itself. It is of the form
%   st(Subst,Nodes) ,
% where
%   Subst is a substitution and
%   Nodes is a list of ST's or
%            a list of leadnodes leaf(Info)
%            where Info is typically a list of substitutions
% An empty substitutiontree is substitutiontree(_,t,[]).
% A Substitution is defined in substitution.

/*
% Examples
ex_st1(substitutiontree(U,ST,[A,B])) :-
	ST = st([U-f(X1,X2)],[SN1,SN2]),
	SN1 = st([X1-A],[SN11,SN12]),
	SN11 = st([X2-A],leaf([[A-X],[A-Z]])),
	SN12 = st([X2-B],leaf([[A-X,B-_Y]])),
	SN2 = st([X2-g(A)],[SN21,SN22]),
	SN21 = st([X1-a],leaf([[A-Z]])),
	SN22 = st([X1-g(X3)],[SN221,SN222]),
	SN221 = st([X3-A],leaf([[A-X]])),
	SN222 = st([X3-d],leaf([[A-X]])).

*/




