% routines for term management
% pb

:- module(term,[
	copyterm/2,
	groundcopy/2,
	subsumes/2,
	vars_of_term/2,
	vars_of_term/3
]).

:- use_module(vars).

% T1 is a copy of T0 where all variables are replaced by new ones
copyterm(T0,T1) :-
	copyterm(T0,T1,[],_).

copyterm(T0,T1,V0,V1) :-
	var(T0),!,
	varassoc(V0,V1,T0,T1).
copyterm(T0,T1,V0,V1) :-
	atomic(T0),!,
	V0 = V1,
	T0 = T1.
copyterm(T0,T1,V0,V1) :-
	functor(T0,F,A),
	functor(T1,F,A),
	copyterm(A,T0,T1,V0,V1).

copyterm(0,_,_,V,V) :- !.
copyterm(N,T0,T1,V0,V2) :-
	arg(N,T0,ST0),
	arg(N,T1,ST1),
	copyterm(ST0,ST1,V0,V1),
	M is N-1,
	copyterm(M,T0,T1,V1,V2).

groundcopy(T0,T1) :-
	copyterm(T0,T1,[],V),
	bindnewvars(V,0).

bindnewvars([],_).
bindnewvars([_-V0|V],N) :- 
	genconstant(V0,N),
	N1 is N+1,
	bindnewvars(V,N1).

genconstant(V0,N) :-
	name(N,LN),
	name(V0,[0's|LN]).


% subsumes(General,Instance)
% suceeds if there exists a substitution such that
% General is an instance of Instance.
% Does not bind variables.
subsumes(General,Instance) :-
	subsumes(General,Instance,[],_).

subsumes(General,Instance,V0,V2) :-
	var(General),!,
	varassoc_subsumes(V0,V1,General,T,New),
	(var(T)->
	    (New == true ->
		T = Instance % just match
	    ;
		T == Instance
	    ),
	    V1 = V2
	;
	    subsumes(T,Instance,V0,V2)
	).
subsumes(General,Instance,V0,V1) :-
	atomic(General),!,
	General == Instance,
	V0 = V1.
subsumes(_General,Instance,_V0,_V1) :-
	var(Instance),!,
	fail.
subsumes(General,Instance,V0,V1) :-
	functor(General,F,A),
	functor(Instance,F,A),
	subsumes(A,General,Instance,V0,V1).

subsumes(0,_,_,V,V) :- !.
subsumes(N,General,Instance,V0,V2) :-
	arg(N,General,SG),
	arg(N,Instance,SI),
	subsumes(SG,SI,V0,V1),
	N1 is N-1,
	subsumes(N1,General,Instance,V1,V2).


	
% varassoc_subsumes(V0,V1,T0,T1,New).
% takes a list of V-W pairs V0 and searches for an
% entry where V = T0. If yes, then W = T1 and V0 = V1.
% Else a new entry T0-T1 is append on V0 yielding V1.
% New is true iff a new entry is appended.
varassoc_subsumes([],[T0-T1],T0,T1,true).
varassoc_subsumes([V-W|V0],[V-W|V1],T0,T1,New) :-
	(V == T0 ->
	    V0 = V1,
	    T1 = W,
	    New = false
	;
	    varassoc_subsumes(V0,V1,T0,T1,New)
	).
	    
vars_of_term(Term,Vars) :-
	vars_of_term(Term,[],Vars).

vars_of_term(Term,Vars0,Vars1) :-
	var(Term),!,
	(var_memberchk(Term,Vars0) ->
	    Vars0 = Vars1
	;
	    [Term|Vars0] = Vars1
	).
vars_of_term(Term,Vars0,Vars1) :-
	functor(Term,_,A),
	vars_of_term(A,Term,Vars0,Vars1).

vars_of_term(0,_Term,Vars,Vars) :- !.
vars_of_term(A,Term,Vars0,Vars2) :-
	arg(A,Term,SubTerm),
	vars_of_term(SubTerm,Vars0,Vars1),
	A1 is A-1,
	vars_of_term(A1,Term,Vars1,Vars2).
	
	



