%declarations%

:- module(timing, [ 
	benchmark/3,
	increase_inference_count/1,
	recordNewRule/0,
	runtime/1,
	start_timing/0,
	stime/0,
	push_timing/1,
	pop_timing/1,
	timing/2
   ]).

:-meta_predicate timing(+,:).


:- use_module(props, [
	flag_on/2
   ]).

:- use_module(global, [
	cont/2,
	store/2
   ]).

:- use_module(adt_declarations, [
	all_keys/2,
	all_texts/2
   ]).

:- use_module(proofState, [
	clause_counter/1
   ]).			  


%end%

%						TIMING


%______________________________________________________________________
%
%	COMPUTING TIME
%______________________________________________________________________

:-dynamic total_time/1.

start_timing:-
        retractall( total_time(_) ),
        statistics( runtime, [T,_] ),
        asserta( total_time( T ) ),
	all_keys( tp, L ),
        assert_time_count( L ), 
	store(timing_stack,[tc_other]),
	store(nmb_inferences_skex,0),
	store(nmb_inferences_forward,0),
	store(nmb_rules,0),
	!.

increase_inference_count(skex):-
	cont(nmb_inferences_skex,N),
	M is N+1,
	store(nmb_inferences_skex,M),
	!.
increase_inference_count(_):-
	cont(nmb_inferences_forward,N),
	M is N+1,
	store(nmb_inferences_forward,M),
       !.


assert_time_count( [] ):-
	store(tc_other,0),
	!.
assert_time_count( [Key|L] ):-
        store(Key, 0 ),
        assert_time_count( L ), !.

begin_time( _Type ):-
	statistics( runtime, _ ).


end_time( Type ):-
	statistics( runtime, [_,T] ),
	cont( Type, S ),
	S1 is S + T,
	store( Type, S1 ).



recordNewRule:-
	cont(nmb_rules,NN),
	NN1 is NN+1,
	(0 is NN1 mod 100,flag_on(ms,nwru) ->
	     format('~nKeeping ~p clauses currently.~n',[NN1])
	;
	     true
	),
	store(nmb_rules,NN1).


benchmark(N,P,T):-
	statistics( runtime, [_,BT] ),
	time0(N,P),
	statistics( runtime, [_,ET] ),
	T is ET-BT.

time0(N,P):-
	(N=<0 ->
	     true
	;    \+ \+P,
	     N1 is N-1,
	     time0(N1,P)).
	

push_timing(T):-
	(flag_on( tp, T ) ->
	     cont(timing_stack,[T1|Ts]),
%	     format("~Npush ~p to ~p",[T,[T1|Ts]]),
             end_time(T1),
	     store(timing_stack,[T,T1|Ts])
	;
	     true
	).


pop_timing(T):-
	(flag_on( tp, T ) ->
%	     format("~Npop ~p",[T]),

             (cont(timing_stack,[T|Ts]) ->
%	           format(" from ~p",[[T|Ts]]),
                   end_time(T),
	           store(timing_stack,Ts)
	     ;
	           cont(timing_stack,TSt),
		   raise_exception(saturate_error("~ntiming stack ~p unbalanced upon pop ~p",[TSt,T]))
	     )
	;
             true
	).


% this works only for deterministic goals
timing(Key,Goal):-
	push_timing(Key),
	on_exception(E,Goal,(pop_timing(Key),
			     format('~nPassing exception ~p',[E]),ttyflush,
			     raise_exception(E))),
	pop_timing(Key),
	!.
timing(Key,_Goal):-
	pop_timing(Key),
	fail.


stime:-
	statistics( runtime, [T1,_] ),
	total_time( T2 ),
	Total is T1 - T2,
	nl,nl,write('Total time: '), 
	write( Total ), write(' milliseconds' ),
	(flag_on( tp, _ ) ->
	     nl, nl, write( 'of which: '), nl, nl,
	     all_keys( tp, KL ),
	     all_texts( tp, TL ),
	     write_times( KL, TL, 0, _S ),
	     cont(tc_other,Rest),
	     write( ' ' ), write( Rest ), 
	     write('		milliseconds for the rest'),nl
	;
	     cont(nmb_inferences_forward,C),
	     format('~2nNumber of forward inferences: ~d',[C]),nl,
	     cont(nmb_inferences_skex,C1),
	     format('Number of tableau inferences: ~d',[C1]),nl,
	     clause_counter(NAllC),
	     NA is NAllC-1,
	     format('Number of generated clauses: ~d',[NA]),nl,
	     cont(nmb_rules,R),
	     format('Number of kept clauses: ~d',[R])
	     
	),
	nl.


runtime(Total):-
	statistics( runtime, [T1,_] ),
	total_time( T2 ),
	Total is T1 - T2.



write_times( [], [], S, S ):- !.
write_times( [K|KL], [T|TL], I, F ):-
	flag_on( tp, K ), !, 
	cont( K , SK ),
	write( ' ' ), write( SK ),
	write('		milliseconds of ' ), 
	write( T ), nl,
	I1 is I+SK,
	write_times( KL, TL, I1, F ), !.
write_times( [_|KL], [_|TL], I, F ):-
	write_times( KL, TL, I, F ), !.















%eof_____________________________________________________________________






