
/* for eclipse
:- set_flag(occur_check, on).
sound_unify(X,X).
:- set_flag(occur_check, off).
*/

sound_unify(X, Y) :-
	unify2(X,Y).
/* if many unification problems are structural not unifiable
	(X == Y ->
	    true
	;
	    (?=(X,Y) ->
		fail
	    ;
		unify2(X, Y)
	    )
	).
*/

unify2(X, Y) :-
        var(X),var(Y),!,
        X = Y.        
unify2(X, Y) :-
        var(X),!,
        occurs_check(Y, X),  
        X = Y.
unify2(X, Y) :-
        var(Y),!,
        occurs_check(X, Y),
        X = Y.
unify2(X, Y) :-
        atomic(X),!,
        X = Y.
unify2(X, Y) :-
        functor(X, F, N),
        functor(Y, F, N),
        unify2(N, X, Y).
        
unify2(0, _, _) :- !.
unify2(N, X, Y) :-
        arg(N, X, Xn),arg(N, Y, Yn),
        unify2(Xn, Yn),
        M is N-1,
        !,
        unify2(M, X, Y).

occurs_check(Term, Var) :-
        var(Term),!,
        Term \== Var.
occurs_check(Term, Var) :-
        functor(Term, _, Arity),
        occurs_check(Arity, Term, Var).

occurs_check(0, _, _) :- !.
occurs_check(N, Term, Var) :-
        arg(N, Term, Arg),
        occurs_check(Arg, Var),
        M is N-1,
        occurs_check(M, Term, Var).


% sound_unifiable/2
% just checks whether two terms are unifiable
% with occurcheck

sound_unifiable(T0,T1) :-
	unifiable(T0,T1,[],_).

unifiable(T0,T1,V0,V1) :-
	var(T0),!,
	var_unifiable(V0,V1,T0,T1).
unifiable(T0,T1,V0,V1) :-
	var(T1),!,
	var_unifiable(V0,V1,T1,T0).
unifiable(T0,T1,V0,V0) :-
	(atomic(T0);atomic(T1)),!,
	 T0 == T1.
unifiable(T0,T1,V0,V1) :-
	functor(T0,F,A),
	functor(T1,F,A),
	unifiable(A,T0,T1,V0,V1).

unifiable(0,_,_,V,V) :- !.
unifiable(N,T0,T1,V0,V2) :-
	arg(N,T0,ST0),
	arg(N,T0,ST1),
	unifiable(ST0,ST1,V0,V1),
	N1 is N-1,
	unifiable(N1,T0,T1,V1,V2).

var_unifiable(V0,V1,T0,T2) :-
	apply_subst(V0,T0,T1),
	(var(T1) -> 
	    occurs_check(T2,T1),
	    V1 = [T1-T2|V0]
	;
	    unifiable(T1,T2,V0,V1)
	).

apply_subst([],T,T).
apply_subst([V-T|V0],T0,T1) :-
	(V == T0 ->
	    T = T1
	;
	    apply_subst(V0,T0,T1)
	).
