#include "/HG/usr/barth/G++/PBIneq/PBCS.h"

extern "C" {
#include "intlist.h"
#include "stdio.h"
#include "stdlib.h"

  int pbcs_initialized = 0;

  int c_backup();
  int c_restore();
  int c_pbcs_init();
  int c_initialize();

  int c_addlpbi();

  int c_fixing();
  int c_fixingcheck();
  int c_eqcheck();
  int c_coeffred();
  int c_coeffredm();
  int c_deleteimplied();
  int c_resetenumstatistics();
  int c_resetcoverstatistics();
  int c_getnocovers();
  int c_getnonodes();
  int c_getnofixed();
  int c_addcover(long* len);
  int c_coversolve();
  int c_solve();
  int c_maximize();
  int c_minimize();
  
  int c_getfixed();
  int c_getsol();
  // getting the complete constraint store
  int c_first();
  int c_last();
  int c_next();

  int c_setheuristic(int i);
  int c_setmaxcutnodes(int i);
  int c_setdellitheu(int i);
  int c_setverbosity(int i);
  int c_setenumverbosity(int i);

  int c_writefile(char* s);
}

#define MAXBACKUPSTACK 2048
PBCS* constraint_store;
LPBIneqSet* backupstack[MAXBACKUPSTACK];
OrdInt* backupstackfixed[MAXBACKUPSTACK];
int bs_index = 0;

int c_backup()
{
//  printf("B");fflush(stdout);
  LPBIneqSet& Copy = *(backupstack[bs_index]);
  LPBIneqSet& Orig = constraint_store->S;
  
  Copy.clear();
  Orig.first();
  while(!Orig.last())
    {
      Copy.add(Orig.next());
    }
  OrdInt& Fixed = *(backupstackfixed[bs_index]);
  Fixed.clear();
  OrdInt FixedOrig = constraint_store->get_fixed();
  FixedOrig.first();
  while(!FixedOrig.last())
    {
      Fixed += FixedOrig.next();
    }
//  printf("backed up on %d\n",bs_index-1);fflush(stdout);
  bs_index++;
  return 0;
}

int c_restore()
{
//  printf("R[%d]",bs_index);fflush(stdout);
  if (bs_index == 0)
    return -1;
  constraint_store->clear();
  bs_index -= 1;
  LPBIneqSet& S = *(backupstack[bs_index]);
  S.first();
  while(!S.last())
    {
      constraint_store->add(S.next());
    }
  OrdInt& Fixed = *(backupstackfixed[bs_index]);
  Fixed.first();
  while(!Fixed.last())
    {
      constraint_store->fix(Fixed.next());
    }
//  printf("restored %d\n",bs_index);fflush(stdout);
  return 0;
}

int c_pbcsinit()
{
  if (!pbcs_initialized)
    {
      int i;

      constraint_store = new PBCS;
      for (i=0;i < MAXBACKUPSTACK;i++)
	{
	  backupstack[i] = new LPBIneqSet;
	  backupstackfixed[i] = new OrdInt;
	}
      pbcs_initialized = 1;
    }
  return 0;
}

int c_initialize()
{
  c_pbcsinit();
  constraint_store->clear();
  bs_index = 0;
//  printf("pB constraint solver ready and cleared \n");
  return 0;
}



int c_addlpbi()
{
  c_pbcsinit();
/*  printf("in do_add_lpbi\n"); */
  LPBIneq lpbi;
  int i = 0;
  while(i < intlistlength-1)
    {
//      printf(" %d*%d +",intlist[i],intlist[i+1]);
      lpbi.add(intlist[i],intlist[i+1]);
      i += 2;
    }
  lpbi.set_rhs(lpbi.get_rhs() + intlist[intlistlength-1]);
//  printf("0 >= %d\n",intlist[intlistlength-1]);
/*
  lpbi.first();
  while(!lpbi.last())
    {
      Product& p = lpbi.next();
      printf(" %d*%d +",p.coefficient,p.literal);      
    }
  printf("0 >= %d\n",lpbi.get_rhs());
*/
  constraint_store->add(lpbi);
  return 0;
}


int c_fixing()
{
  OrdInt fixed = constraint_store->fixing();  
  return fixed.length();
}

int c_fixingcheck()
{
  OrdInt fixed = constraint_store->fixingcheck();  
  fixed.first();
  while(!fixed.last())
    {
      constraint_store->fix(fixed.next());
    }
 return fixed.length();
}

int c_eqcheck()
{
  return constraint_store->eq_check();  
}

int c_coeffred()
{
  return constraint_store->coefficient_reduction();  
}

int c_coeffredm()
{
  return constraint_store->coefficient_reduction_bigM();  
}

int c_deleteimplied()
{
  return constraint_store->delete_implied();  
}

int c_resetenumstatistics()
{
  constraint_store->enum_stat_reset();
  return 0;
}

int c_resetcoverstatistics()
{
  return constraint_store->reset_cover_statistics();
}

int c_getnocovers()
{
  return constraint_store->get_no_covers();
}

int c_getnonodes()
{
  return constraint_store->Enum.stat_nodes();
}

int c_getnofixed()
{
  return constraint_store->Enum.stat_fixedlits();
}

int c_addcover(long* len)
{
  int ilen;
  int val = constraint_store->addcover(ilen);
  *len = (long) ilen;
  // 1 if sat; -1 if unsat; 0 if open
  return val;
}

int c_coversolve()
{
//  printf("maxcutnodes is %d \n",constraint_store->max_cut_nodes);
  int ret = constraint_store->coversolve();
//  printf("maxcutnodes is %d \n",constraint_store->max_cut_nodes);
//  printf ("coversolve : %d ",ret);
  return ( ret == 1) ? 1 : -1;
}

int c_solve()
{
  // lift that to a solve in class PBCS
  constraint_store->Enum.init_enumeration();
  int ret = constraint_store->Enum.solve();
  return ( ret == 1) ? 1 : -1;  
}

int c_maximize()
{
  Products objf;
  int consistent;
  OrdInt sol;

  if (intlistlength % 2 == 1)
    {
      fprintf(stderr,"Error : maximize: invalid integer list");
      return 0;
    }
  for(int i=0;i<intlistlength;i = i+2)
    {
      objf.add(intlist[i],intlist[i+1]);
    }
  int val = constraint_store->optimize(objf, consistent, sol, 1);
  return val;
}

int c_minimize()
{
  Products objf;
  int consistent;
  OrdInt sol;

  if (intlistlength % 2 == 1)
    {
      fprintf(stderr,"Error : minimize: invalid integer list");
      return 0;
    }
  for(int i=0;i<intlistlength;i = i+2)
    {
      objf.add(intlist[i],intlist[i+1]);
    }
  int val = constraint_store->optimize(objf, consistent, sol, 0);
  return val;
}



int c_getfixed()
{
  OrdInt fixed_lits = constraint_store->get_fixed();

  int i=0;

  fixed_lits.first();
  while(!fixed_lits.last())
    {
      intlist[i++] = fixed_lits.next();
    }
  return intlistlength=i;
}

int c_getsol()
{
  OrdInt sol;
  sol = constraint_store->get_last_solution();

  int i=0;
//  printf("getsol : ");
  sol.first();
  while (!sol.last())
    {
      int lit = sol.next();
//      printf("%d ",lit);
      intlist[i++] = lit;
    }
  intlistlength = i;
//  printf(" : %d elements",intlistlength);
  return intlistlength;
}

int c_first()
{
  constraint_store->S.first();
  return 0;
}
int c_last()
{
  if (constraint_store->S.last())
    return 1;
  else
    return 0;
}

int c_next()
{
  LPBIneq& lpbi = constraint_store->S.next();
  int i=0;

  lpbi.first();
  while(!lpbi.last())
    {
      Product& p = lpbi.next();
      intlist[i++] = p.coefficient;
      intlist[i++] = p.literal;
    }
  intlist[i++] = lpbi.get_rhs();
  intlistlength = i;
  return 0;
}




int c_setmaxcutnodes(int i)
{
//  printf("setting maxcutnodes to %d \n",i);
  return constraint_store->set_max_cut_nodes(i);
}

int c_setheuristic(int i)
{
  return constraint_store->set_enum_heuristic(i);
}

int c_setdellitheu(int i)
{
  return constraint_store->set_dellitheu(i);
}

int c_setverbosity(int i)
{
  return constraint_store->set_verbosity(i);
}

int c_setenumverbosity(int i)
{
  return constraint_store->set_enum_verbosity(i);
}

int c_writefile(char* s)
{
  FILE* out = fopen(s,"w");

  if (out == NULL)
    return -1;
  fprintf(out,"max: 1*A + 1*B\n");
  LPBIneqSet& S = constraint_store->S;
  S.first();
  while(!S.last())
    {
      LPBIneq& lpbi = S.next();
      lpbi.first();
      while(!lpbi.last())
        {
          Product& p = lpbi.next();
          if (p.literal > 0)
            fprintf(out,"%d*x%d ",p.coefficient,p.literal);
          else
            fprintf(out,"%d*~x%d ",p.coefficient,-p.literal);       
        }
      fprintf(out," >= %d\n",lpbi.get_rhs());
    }  
  fclose(out);
  return 0;
}
