% Magic Series as pB problem.
% For an n+1-element list, we have (n+1)^2 Variables Xij for i,j \in \{0,..,n\}.
% Xij = true iff ith element in list has value j
% Constraints:
% We know that 
% forall i : \sum_{j} Xij = 1  % one value!
% If Xij = 1 then l[i] = j and thus \sum_{k} Xki = j
% Otherwise stated.
% When summing up a row \sum_{i} Xij for some j then this equals
% the weighted sum over the columns \sum_{k} k*Xjk.
% There is also a surrogate constraint, stating that the sum over all
% values equals n+1.

magic(N,T) :-
	make_matrix(N,M),
	transpose(M,T),
	fwrite('Matrix generated'),
	one_value(T),
	fwrite('one_value_assured'),
%	surrogate(N,M),    % to big after transformation!
%	fwrite('surrogated'),
	(1 = 0 ->
	    M1 = M,
	    T1 = T
	;
	    reverse(M,M1),
	    reverse(T,T1)
	),
	magic_condition(M1,T1),
	fwrite('done').

get_sol(M,L) :-
	matrix_enum(M),
	assign_list(M,L).

matrix_enum([]).
matrix_enum([L|Ls]) :-
	list_enum(L),
	matrix_enum(Ls).

list_enum([]).
list_enum([Var|Vars]) :- Var #<= 0,list_enum(Vars).
list_enum([Var|Vars]) :- Var #>= 1,list_enum(Vars).




% for transposed matrix
assign_list([],[]).
assign_list([L|Ls],[Val|Vals]) :-
	assign_val(L,0,Val),
	assign_list(Ls,Vals).

assign_val([1|_],N,N) :- !.
assign_val([_|Rest],N,Val) :- 
	N1 is N + 1,
	assign_val(Rest,N1,Val).


% constraints

magic_condition([],[]).
magic_condition([Row|Rows],[Column|Columns]) :-
	weighted_sum(Column,0,Term),
	sum_list_equal(Row,0,Term),
%	fwrite('one magic condition'),
	magic_condition(Rows,Columns).

weighted_sum([],_,0).
weighted_sum([Var|Vars],Weight,Term + Weight*Var) :-
	Weight1 is Weight+1,
	weighted_sum(Vars,Weight1,Term).


% one value per list position.
one_value([]).
one_value([Column|Rest]) :-
	sum_list_equal(Column,0,1),
	one_value(Rest).



% surrogate; sum vals = n+1

surrogate(N,M) :- surrogate(M,0,T),N1 is N+1,T #= N1.
surrogate([],_,0).
surrogate([Row|Rows],Weight,T + WT) :-
	make_wsum(Row,Weight,WT),
	Weight1 is Weight + 1,
	surrogate(Rows,Weight1,T).

make_wsum([],_,0).
make_wsum([Var|Vars],Weight,T + Weight*Var) :-
	make_wsum(Vars,Weight,T).
	



% stuff

sum_list_equal([],T,R) :-
	T #= R.
sum_list_equal([L|Ls],T,R) :-
	sum_list_equal(Ls,L + T,R).

	
make_matrix(N,M) :-
	make_matrix(N,N,M).
	
make_matrix(-1,_,[]) :- !.
make_matrix(N,RN,[Row|M]) :-
	N1 is N-1,
	make_row(RN,Row),
	make_matrix(N1,RN,M).

make_row(-1,[]) :- !.
make_row(N,[_Xij|Row]) :-
	N1 is N-1,
	make_row(N1,Row).

transpose([[]|_],[]) :- !.
transpose(M0,[Firsts|T]) :-
	del_firsts(M0,Firsts,M1),
	transpose(M1,T).

del_firsts([],[],[]) :- !.
del_firsts([[E|ERest]|Rest],[E|Firsts],[ERest|M]) :-
	del_firsts(Rest,Firsts,M).



fwrite(X) :- write(X),nl,ttyflush.

