/* Property 3 of the Sheffer's 3-basis (for Boolean algebra).
   f is the Sheffer stroke. The property is derived from Veroff's 2-basis of f.

   ``Basis'' means that the set of axioms axiomatize Boolean algebra.
   

   The automatic introduction of new functions from splitting
   equations seems essential in solving the problem.
   Saturate invents f$0 (which is verum) and f$1 (negation).

   This example was communicated to Thomas Hillenbrand by Dana Scott,
   also see the JAR 29,1 (2002) paper by McCune et al.

*/

% Veriff's 2-basis:

:- use_module(dp(formulas)).

theory([A1,A2,A3,A4,G]):-
	X='$VAR'(0),
	Y='$VAR'(1),
	Z='$VAR'(2),

	mk_equation(f(X,Y), f(Y,X), o, axiom(1),A1),
	mk_equation(f(f(X,Y),f(X,f(Y,Z))), X, o, axiom(2), A2),
	mk_equation(n(X), f(X,X), o, axiom(3), A3),
	mk_equation(f(X,f(X,X)), t, o, axiom(4), A4),

	% goal
	mk_equation('$neq'(f(f(a,f(b,c)),f(a,f(b,c))),
			   f(f(f(b,b),a),f(f(c,c),a))),'$true',prop,goal(1),G).

goal:-
	theory([A1,A2,A3,A4,G]),
%	global_vars:initialize_vars,
	hgstuff:time(saturation:sat(equational(sig([
						  t:([]->o),
						  n:([o]->o),
						  f:([o,o]->o),
						  a:([]->o),
						  b:([]->o),
						  c:([]->o)
						 ],
					     cons([])),
					     _),

				  ord(lpo,prec([],[f,n,t,a,b,c])),[],
%				  [A2,G],
				 [A1,A2,A3,A4,G],
				 _A,_P),
		    T),
	format('~2nTime: ~p~2n',[T]).


/*
20021106 600 Mhz PIII by Saturate


Proof:
======

   2 : f(f(A,B),f(A,f(B,C)))=A                                          [input]
   3 + !A,B,C.(f(f(C,f(B,A)),f(C,f(B,A)))=f(f(f(B,B),C),f(f(A,A),C))) ->  [input]
   4 + !A,B.(f(f(B,f(A,s$3)),f(B,f(A,s$3)))=f(f(f(A,A),B),f(f(s$3,s$3),B))) ->  [expansion from 3]
   5 + !A.(f(f(A,f(s$2,s$3)),f(A,f(s$2,s$3)))=f(f(f(s$2,s$2),A),f(f(s$3,s$3),A))) ->  [expansion from 4]
   6 + f(f(s$1,f(s$2,s$3)),f(s$1,f(s$2,s$3)))=f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1)) ->  [expansion from 5]
   7 : f(f(f(A,B),A),A)=f(A,B)                           [chaining of 2 from 2]
  15 : f(f(f(A,B),f(B,C)),A)=f(A,B)                      [chaining of 2 from 2]
  27 : f(f(f(A,f(B,C)),B),A)=f(A,f(B,C))                 [chaining of 2 from 2]
  73 : f(f(A,A),f(A,B))=A                                [chaining of 7 from 2]
 105 : f(f(f(A,A),B),A)=f(A,A)                          [chaining of 73 from 2]
 800 : f(A,A)=f(A,f(f(A,B),B))                          [chaining of 2 from 27]
 990 : f(f(f(A,B),f(A,C)),f(f(A,B),f(A,C)))=f(f(f(A,B),f(A,C)),A) [chaining of 27 from 15]
1174 : f(f(A,B),f(A,f(A,A)))=f(f(A,B),f(A,B))         [chaining of 73 from 800]
1214 : f(A,f(f(A,B),f(f(A,B),f(B,C))))=f(A,A)         [chaining of 15 from 800]
1223 : f(f(A,A),B)=f(B,f(A,B))                        [chaining of 800 from 27]
1264 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f(f(f(s$2,s$3),s$1),f(f(f(s$3,f(s$3,s$3)),f(s$3,f(s$3,s$3))),f(s$2,s$3))) ->  [reduction of 6 by [1174,1174,1223]]
1300 : f(f(A,f(B,C)),A)=f(f(f(A,B),f(A,B)),f(A,f(B,C))) [chaining of 2 from 1223]
1310 : f(f(f(A,A),B),B)=f(B,A)                        [chaining of 1223 from 7]
1358 : f(f(A,A),A)=f(f(f(A,B),f(A,B)),f(A,A))        [chaining of 73 from 1223]
1426 : f(f(f(A,B),f(B,A)),f(A,B))=f(f(A,A),f(f(A,B),f(B,A))) [chaining of 15 from 1223]
1485 : f(f(f(A,B),f(B,A)),f(A,B))=f(f(A,A),A)     [reduction of 1426 by [1358]]
1980 : f(f(f(A,A),B),f(f(f(A,A),B),f(B,C)))=f(f(f(f(A,A),B),f(B,C)),A) [chaining of 15 from 1310]
2201 : f(f(f(A,f(A,A)),B),f(f(A,B),f(A,B)))=f(A,f(A,A)) [chaining of 1174 from 2]
2258 : f(A,f(f(A,f(B,B)),f(f(A,f(B,B)),f(A,f(B,B)))))=f(f(f(A,f(B,B)),f(A,B)),f(f(A,f(B,B)),f(A,B))) [chaining of 2 from 1174]
2548 : f(f(A,B),f(f(A,B),f(A,B)))=f(f(f(A,f(A,A)),f(A,f(A,A))),f(A,B)) [chaining of 1174 from 1223]
2592 : f(f(A,f(B,f(B,B))),A)=f(B,f(B,B))          [reduction of 2201 by [1300]]
2593 : f(A,f(B,f(B,B)))=f(A,A)       [reduction of 2258 by [1980,1223,105,2,2]]
2597 : f(f(A,A),A)=f(f(f(A,f(A,A)),f(A,f(A,A))),f(A,B)) [reduction of 2548 by [1485]]
2598 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f(f(f(s$2,s$3),s$1),f(f(s$3,s$3),s$3)) ->  [reduction of 1264 by [2597]]
2795 : f(f(A,f(B,f(B,B))),f(A,C))=A                  [chaining of 2593 from 73]
3003 : f(f(A,f(B,f(B,B))),C)=f(C,f(A,C))           [chaining of 2593 from 1223]
3007 : f(f(A,f(A,A)),f(B,f(A,f(A,A))))=f(f(B,B),f(B,B)) [chaining of 1223 from 2593]
3171 : f(A,f(B,f(B,B)))=f(A,f(C,f(C,C)))           [chaining of 2593 from 2593]
3179 : f(f(A,f(A,A)),f(B,f(A,f(A,A))))=B            [reduction of 3007 by [73]]
3182 : f(A,f(B,f(B,B)))=f$1(A)                 [variable elimination from 3171]
3183 : f(A,f(B,f(B,B)))=f$1(A)                 [variable elimination from 3171]
3187 : f(f$1(A),A)=f(B,f(B,B))                    [reduction of 2592 by [3182]]
3189 : f(A,f(A,A))=f$2                         [variable elimination from 3187]
3201 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f(f(f(s$2,s$3),s$1),f$2) ->  [reduction of 2598 by [3189]]
3202 : f(A,f$2)=f(A,A)                            [reduction of 2593 by [3189]]
3207 : f(f$2,f(A,f$2))=A                     [reduction of 3179 by [3189,3189]]
3209 : f(f(A,f$2),B)=f(B,f(A,B))                  [reduction of 3003 by [3189]]
3213 : f(f(A,f$2),f(A,B))=A                       [reduction of 2795 by [3189]]
3215 : f(A,f$2)=f$1(A)                            [reduction of 3183 by [3189]]
3218 + f(f(f(s$2,s$2),s$1),f(f(s$3,s$3),s$1))=f$1(f(f(s$2,s$3),s$1)) ->  [reduction of 3201 by [3215]]
3233 : f(f$1(A),f(A,B))=A                         [reduction of 3213 by [3215]]
3237 : f(f$1(A),B)=f(B,f(A,B))                    [reduction of 3209 by [3215]]
3239 : f$1(f$1(A))=A                         [reduction of 3207 by [3215,3215]]
3241 : f$1(A)=f(A,A)                              [reduction of 3202 by [3215]]
3261 : f(A,f(f$1(f(B,C)),f(A,B)))=f$1(A)     [reduction of 1214 by [3237,3241]]
3267 + f(f(f$1(s$2),s$1),f(f$1(s$3),s$1))=f$1(f(f(s$2,s$3),s$1)) ->  [reduction of 3218 by [3241,3241]]
3268 : f(f$1(f(A,B)),f(A,f(B,C)))=f(f$1(f(B,C)),A) [reduction of 1300 by [3241]]
3269 : f$1(f(f(A,B),f(A,C)))=f(f(f(A,B),f(A,C)),A) [reduction of 990 by [3241]]
6293 : f(f$1(A),f$1(f(B,C)))=f(f$1(f(B,C)),f(A,B))   [chaining of 3261 from 27]
6719 : f(f$1(f(A,B)),f$1(f(C,A)))=f(f$1(f(A,B)),C) [reduction of 3268 by [6293]]
6960 : f(f$1(f(A,B)),f(f$1(C),f$1(f(A,B))))=f(f$1(f(C,A)),f$1(f(A,B))) [chaining of 6293 from 3237]
7242 : f(A,f$1(f(B,C)))=f(f$1(f(B,A)),C) [reduction of 6960 by [3237,3239,6719]]
7447 : f(A,f(f$1(f(B,A)),C))=f(f$1(f$1(f(B,C))),A) [chaining of 7242 from 3237]
7816 : f(A,f(f$1(f(B,A)),C))=f(f(B,C),A)          [reduction of 7447 by [3239]]
9976 : f(A,f(f$1(B),f$1(f(C,A))))=f(f(C,f(B,f$1(f(C,A)))),A) [chaining of 3237 from 7816]
10461 : f(f(A,f$1(B)),C)=f(f(f(C,B),A),C)    [reduction of 9976 by [7816,7816]]
10482 : f(f(f$1(A),f$1(B)),C)=f$1(f(f(C,A),f(C,B))) [reduction of 3269 by [10461,10461]]
10559 : f$1(f(f(f(A,B),C),A))=f(f(f$1(C),f$1(f$1(A))),f(A,B)) [chaining of 3233 from 10482]
10993 : f$1(f(f(A,f$1(B)),C))=f(f(f$1(A),C),f(C,B)) [reduction of 10559 by [10461,3239]]
11020 + false                             [reduction of 3267 by [10993,3239,L]]

Length = 65, Depth = 25, Search Depth = 0



Total time: 324754 milliseconds

Number of forward inferences: 10719
Number of tableau inferences: 3
Number of generated clauses: 11020
Number of kept clauses: 501


*/



/*  Satellite on mpino2202 20030823

Proofs 
23651:inference(reflexivity,[23650])
23650:inference(reduction,inference(reduction,eq(7),[22804]),[47,47,5])
22804:inference(cp,[5,22598])
22705:inference(cp,[5,22598])
22598:inference(reduction,inference(cp,[22201,33]),[152])
22201:inference(reduction,inference(cp,[51,3815]),[47,5,7464,5,7448,47])
3815:inference(reduction,inference(cp,[3687,53]),[47])
4112:inference(reduction,inference(cp,[823,3645]),[5])
7456:inference(reduction,inference(cp,[217,6749]),[6583])
21186:inference(reduction,inference(cp,[20792,7448]),[52])
20956:inference(reduction,inference(cp,[20670,7448]),[52])
20792:inference(cp,[5,20670])
20670:inference(reduction,inference(cp,[962,7455]),[1392,5])
7455:inference(reduction,inference(cp,[218,6749]),[6749])
7452:inference(reduction,inference(cp,[224,6749]),[6749])
7451:inference(reduction,inference(cp,[253,6749]),[6583])
15524:inference(cp,[6326,6552])
15523:inference(cp,[6058,6552])
15140:inference(cp,[6058,6388])
15139:inference(cp,[6058,6388])
15137:inference(cp,[5873,6388])
11851:inference(reduction,inference(cp,[1321,11581]),[6326,5,1321])
6806:inference(reduction,inference(cp,[6326,5873]),[6326,5])
6653:inference(reduction,inference(cp,[6058,6058]),[5])
6552:inference(reduction,inference(cp,[6058,6058]),[5])
6388:inference(reduction,inference(cp,[5873,6058]),[6058,5])
7517:inference(reduction,inference(cp,[6749,6749]),[6749])
7465:inference(reduction,inference(cp,[6583,6749]),[6749])
1303:inference(reduction,inference(cp,[1282,218]),[5,152,1282])
2426:inference(cp,[823,824])
2289:inference(cp,[822,824])
8988:inference(reduction,inference(cp,[195,8953]),[152,47,5,6326,5,5,55])
14316:inference(reduction,inference(cp,[14277,6470]),[5,53,5])
14277:inference(reduction,inference(reduction,inference(cp,[169,801]),[47,5]),[6326,5])
9275:inference(reduction,inference(cp,[180,6137]),[5,5,5])
9258:inference(reduction,inference(cp,[12,6137]),[5,5,5])
7462:inference(reduction,inference(cp,[3553,6749]),[6749])
13858:inference(reduction,inference(reduction,inference(cp,[1254,822]),[47,5,1254,47]),[11581])
6736:inference(reduction,inference(cp,[6326,85]),[5])
6723:inference(reduction,inference(cp,[6326,103]),[5])
6570:inference(reduction,inference(cp,[6058,85]),[5])
6557:inference(reduction,inference(cp,[6058,103]),[5])
6470:inference(reduction,inference(cp,[6058,85]),[5])
6457:inference(reduction,inference(cp,[6058,103]),[5])
6146:inference(reduction,inference(cp,[5873,85]),[5])
6133:inference(reduction,inference(cp,[5873,103]),[5])
2423:inference(cp,[823,9])
2422:inference(cp,[823,184])
2286:inference(cp,[822,9])
11581:inference(reduction,inference(cp,[248,2285]),[7448,3586,47])
2285:inference(cp,[822,184])
1507:inference(reduction,inference(cp,[1321,248]),[5])
1506:inference(cp,[1321,13])
962:inference(reduction,inference(cp,[832,248]),[5])
944:inference(cp,[832,13])
929:inference(reduction,inference(cp,[828,248]),[5])
911:inference(cp,[828,13])
881:inference(reduction,inference(cp,[824,13]),[5])
853:inference(reduction,inference(cp,[801,13]),[5])
7727:inference(reduction,inference(cp,[7464,6749]),[6749])
7634:inference(reduction,inference(cp,[7448,6749]),[6749])
1991:inference(cp,[1519,53])
1291:inference(reduction,inference(cp,[1254,832]),[47,1254])
9899:inference(reduction,inference(reduction,inference(cp,[195,217]),[152,5,195]),[6326,5,6326,5,6326,6326])
9407:inference(cp,[5,6461])
9283:inference(cp,[5,6137])
9105:inference(cp,[5,1509])
6727:inference(reduction,inference(cp,[6326,53]),[47])
6561:inference(reduction,inference(cp,[6058,53]),[47])
6461:inference(reduction,inference(cp,[6058,53]),[47])
6137:inference(reduction,inference(cp,[5873,53]),[47])
4180:inference(reduction,inference(cp,[1254,3568]),[47,1254,47,5])
1509:inference(reduction,inference(cp,[1321,832]),[47,1321])
8953:inference(reduction,inference(reduction,inference(cp,[9,1392]),[5,1389]),[6326])
8837:inference(reduction,inference(reduction,inference(cp,[4,1392]),[5,1392]),[6326])
8704:inference(reduction,inference(reduction,inference(cp,[13,1392]),[5,1392]),[6326])
8568:inference(reduction,inference(reduction,inference(cp,[248,1389]),[5,1389]),[6326])
7466:inference(reduction,inference(cp,[3587,6749]),[5,47,6749,3587])
6582:inference(reduction,inference(cp,[6058,1389]),[3587,47,5,6058,5])
7579:inference(reduction,inference(cp,[4,7448]),[3,6326,5])
7578:inference(reduction,inference(cp,[9,7448]),[3,6326,5])
7971:inference(reduction,inference(reduction,eq(4034),[5873]),[6326,5])
7732:inference(reduction,inference(reduction,eq(4032),[5873]),[6326,5])
7464:inference(reduction,inference(cp,[1389,6749]),[6749])
7448:inference(reduction,inference(cp,[1392,6749]),[6749])
6749:inference(reduction,inference(cp,[6326,1392]),[47])
6583:inference(reduction,inference(cp,[6058,1392]),[47])
6482:inference(reduction,inference(cp,[6058,1392]),[47])
6158:inference(reduction,inference(cp,[5873,1392]),[47])
6326:inference(cp,[5,6058])
6058:inference(cp,[5,5873])
5873:inference(reduction,inference(cp,[55,3747]),[3747,3747])
3747:inference(reduction,inference(cp,[9,3554]),[5])
3746:inference(reduction,inference(cp,[180,3554]),[5])
3901:inference(reduction,inference(cp,[3554,3582]),[47,47])
4875:inference(reduction,inference(cp,[1321,4032]),[47,5,1321,47])
3557:inference(reduction,inference(cp,[1518,3548]),[47,47,5])
4035:inference(reduction,inference(cp,[4,3584]),[5])
4034:inference(reduction,inference(cp,[13,3584]),[5])
4033:inference(reduction,inference(cp,[9,3584]),[5])
4032:inference(reduction,inference(cp,[180,3584]),[5])
3587:inference(reduction,inference(cp,[247,3548]),[5])
3586:inference(reduction,inference(cp,[258,3548]),[5])
3572:inference(reduction,inference(cp,[12,3548]),[5])
3568:inference(reduction,inference(cp,[180,3548]),[5])
3645:inference(reduction,inference(cp,[3548,3548]),[47,47])
3584:inference(reduction,inference(cp,[248,3548]),[5])
3582:inference(reduction,inference(cp,[184,3548]),[5])
3687:inference(reduction,inference(cp,[3553,1392]),[1392,47])
3554:inference(cp,[5,3548])
3553:inference(reduction,inference(cp,[47,3548]),[5,47])
3548:inference(cp,[3078,285])
3078:inference(reduction,inference(cp,[2681,1392]),[39,47])
3168:inference(cp,[5,3069])
3140:inference(cp,[5,3069])
3096:inference(cp,[5,3069])
3069:inference(reduction,inference(cp,[905,2681]),[47])
2681:inference(reduction,inference(cp,[184,2598]),[5,5])
2678:inference(reduction,inference(cp,[12,2598]),[5])
2667:inference(reduction,inference(cp,[180,2598]),[5])
2653:inference(cp,[5,2598])
2642:inference(reduction,inference(cp,[47,2598]),[5])
2598:inference(reduction,inference(cp,[1404,1504]),[5])
1404:inference(reduction,inference(cp,[184,1389]),[5,1389])
823:inference(reduction,inference(cp,[787,247]),[5,61])
2303:inference(cp,[5,2274])
2274:inference(reduction,inference(cp,[169,822]),[47,5,169,47])
822:inference(reduction,inference(cp,[787,258]),[5,61])
1579:inference(reduction,inference(cp,[1518,336]),[47])
1505:inference(reduction,inference(cp,[1321,333]),[47])
1504:inference(reduction,inference(cp,[1321,647]),[47])
1503:inference(reduction,inference(cp,[1321,373]),[47])
1946:inference(cp,[5,179])
1887:inference(cp,[5,175])
1576:inference(reduction,inference(cp,[1518,905]),[47,1518])
1519:inference(reduction,inference(cp,[1321,801]),[47])
179:inference(reduction,inference(cp,[53,9]),[5])
175:inference(reduction,inference(cp,[51,9]),[5])
933:inference(reduction,inference(cp,[169,832]),[47,169])
1687:inference(cp,[5,924])
1405:inference(reduction,inference(cp,[103,1389]),[5,1389])
965:inference(reduction,inference(cp,[47,905]),[5,5])
924:inference(cp,[828,53])
849:inference(reduction,inference(cp,[52,801]),[47])
321:inference(reduction,inference(cp,[248,4]),[5,5])
1518:inference(reduction,inference(cp,[1321,85]),[1389])
1514:inference(reduction,inference(cp,[1321,1392]),[6])
1502:inference(reduction,inference(cp,[1321,147]),[47])
1321:inference(reduction,inference(cp,[253,285]),[3])
1392:inference(cp,[5,1389])
1389:inference(cp,[1319,285])
1319:inference(reduction,inference(cp,[13,285]),[3,5])
285:inference(reduction,inference(cp,[184,4]),[5,5])
1282:inference(reduction,inference(cp,[1254,147]),[47,5])
1254:inference(cp,[832,56])
56:inference(reduction,inference(cp,[51,4]),[5])
647:inference(reduction,inference(cp,[146,253]),[152,146])
376:inference(reduction,inference(cp,[146,218]),[5,152,146])
374:inference(reduction,inference(cp,[135,218]),[5,152,135])
373:inference(reduction,inference(cp,[147,218]),[5,152,147])
336:inference(reduction,inference(cp,[146,217]),[152,146])
334:inference(reduction,inference(cp,[135,217]),[5,152,135])
333:inference(reduction,inference(cp,[147,217]),[5,152,147])
938:inference(cp,[5,832])
905:inference(cp,[5,828])
832:inference(cp,[248,801])
828:inference(cp,[51,801])
824:inference(cp,[5,801])
801:inference(reduction,inference(cp,[47,787]),[47])
787:inference(reduction,inference(cp,[164,55]),[3])
164:inference(reduction,inference(cp,[146,4]),[5,152])
258:inference(reduction,inference(cp,[180,180]),[5])
253:inference(reduction,inference(cp,[180,13]),[5])
247:inference(reduction,inference(cp,[9,180]),[5])
246:inference(reduction,inference(cp,[4,180]),[5])
245:inference(reduction,inference(cp,[13,180]),[5])
224:inference(reduction,inference(cp,[13,13]),[5])
218:inference(reduction,inference(cp,[4,13]),[5])
217:inference(reduction,inference(cp,[9,13]),[5])
248:inference(cp,[5,180])
184:inference(cp,[5,9])
180:inference(cp,[5,9])
13:inference(cp,[5,4])
195:inference(reduction,inference(cp,[169,147]),[47,5])
169:inference(reduction,inference(cp,[33,9]),[5,39])
9:inference(cp,[5,4])
157:inference(cp,[152,51])
146:inference(reduction,inference(cp,[47,135]),[5])
147:inference(cp,[5,135])
155:inference(reduction,inference(cp,[152,135]),[47])
152:inference(reduction,inference(cp,[135,85]),[47,5,55])
135:inference(reduction,inference(cp,[51,122]),[5,39,5])
122:inference(cp,[47,35])
35:inference(reduction,inference(cp,[33,4]),[5])
103:inference(cp,[5,85])
85:inference(cp,[5,12])
12:inference(cp,[3,4])
73:inference(cp,[5,18])
18:inference(reduction,inference(cp,[4,4]),[5,5])
61:inference(reduction,inference(cp,[53,53]),[5])
55:inference(reduction,inference(cp,[4,51]),[5])
52:inference(reduction,inference(cp,[47,51]),[5])
53:inference(cp,[5,51])
51:inference(reduction,inference(reduction,eq(15),[39]),[5])
47:inference(reduction,eq(25),[39])
39:inference(reduction,inference(cp,[25,19]),[5,22,3])
19:inference(cp,[3,15])
22:inference(cp,[6,15])
34:inference(reduction,inference(cp,[33,15]),[3])
33:inference(reduction,inference(cp,[30,4]),[3,5])
30:inference(reduction,inference(cp,[25,6]),[5])
29:inference(cp,[3,25])
25:inference(cp,[3,15])
15:inference(cp,[6,4])
4:axiom(2)
7:inference(reduction,(goal 1),[3,5,3,5,5,5,5,3])
6:inference(reduction,axiom(4),[3])
5:axiom(1)
3:axiom(3)



Time: 204.968


*/
