/* -----------------------------------------------------------------------

Installation steps:
0. Create a directory <install_dir> and untar the tar file into it.

1. Edit ``saturate_directory'' below to make it return the <install_dir>.

2. cd to <install_dir>/sources/DP

3. Call Prolog.

4. :- compile(main).  (this file)

5. At the end of the compilation you'll be asked for the name of a file
to which the system image is to be saved into.
Write file names that are no Prolog atoms with quotes
such as in the example below and put a dot ``.'' at the end.

6. Start using Satellite .


-------------------------------------------------------------------------*/

:- ensure_loaded('../satinit').  % some initialization



os(windows):-
				% modify by commenting out inappropriate
				% clauses if this does not work for you:
	prolog_flag(version,V),
	name(V,VN),
	append1("SICStus",_,VN), % can't have Quintus under windows
	use_module(library(system), [file_exists/1]),
	file_exists('c:/').
os(linux):-
	prolog_flag(version,V),
	name(V,VN),
	append1("SICStus",_,VN), % don't have Quintus under Linux
	use_module(library(system),[shell/2]),
	shell('uname -a | grep Linux > /dev/null',0),
	!.
os(darwin):-
	prolog_flag(version,V),
	name(V,VN),
	append1("SICStus",_,VN), % can't have Quintus under Darwin
	use_module(library(system),[shell/2]),
	shell('uname -a | grep Darwin > /dev/null',0).
% the following clause may be called only after compilation
% since unix/1 is not defined for SICStus and will be defined
% only later by compiling thisProlog.pl
os(solaris):-	% it better not be SunOS4
	unix(system('uname -a | grep SunOS > /dev/null')),
	!.

/* ==============================================================================

   Below you need to state the directory to which the files
   have been unloaded.

============================================================================== */

%%%%% Installation directory:
:- dynamic saturate_directory/1.

% put your <install_dir> here
:- assert(saturate_directory( 'c:/home/hg/saturate' )).


%%%%% comment out this call as it only works in my environment
/*
:- 	(os(windows) ->
	    use_module(library(system),[shell/2]),
	    shell(testhost,H),
				%	format('~nMachine ~p~n',[H]),
	    ( H=100 -> M=mpino2100
	    ;	H=2 -> M=mpino2202),
	    
	    ( M=mpino2100 ->
		assert(saturate_directory( 'e:/home/hg/saturate' )),
		format('~nCompiling Saturate on Machine ~p~n',['MPINO2100'])
	    ;M=mpino2202 ->
		assert(saturate_directory( 'c:/home/hg/saturate' )),
		format('~n\%*Compiling Saturate on Machine ~p~n',['MPINO2202'])
	    )
	;os(darwin) ->
	    assert(saturate_directory( '/Users/hg/saturate' ))
	;% linux or solaris; notice we don't call os(solaris)
	    assert(saturate_directory( '/HG/local/provers/saturate' ))
	).

*/

%==============================================================================

            % you're not supposed to change anything from here on

%==============================================================================


satellite_version('0.1').




:- op(1150,fx,mymultifile).



:- op(1020,xfx,if).
:- op(940,yfx,[and,'&']).
:- op(930,yfx,[or,'|']).
:- op(950,xfx,[<=>]).
:- op(950,xfy,[=>]).
%:- op(900,fy,[~,dia,box]).
:- op(900,fy,[dia,box]).
:- op(700,yfx,@).    % apply operator
:- op(500,fy,[~]).   % should be smaller than :
:- op(600,yfx,[::]).
:- op(600,xfx,[^^]).
:- op(100,fx,[!,?]).
:-op(1050,xfx,<-).
:-op(1150,fx,goal).





:- multifile  
	portray/1,
	file_search_path/2, 
	query_hook/6, 
	message_hook/3, 
	portray_message/2.

:- dynamic 
	file_search_path/2, 
	query_hook/6, 
	message_hook/3, 
	portray_message/2,
	spy_clause/1,
	user_portray/1,
	no_portray/0.





add_search_path(Name,P):-
	file_search_path(Name,P),
	!.
add_search_path(Name,P):-
	asserta(file_search_path(Name,P)).

add_search_path_at_end(Name,P):-
	file_search_path(Name,P),
	!.
add_search_path_at_end(Name,P):-
	assertz(file_search_path(Name,P)).




message_hook(_,help,_).


%unknown_predicate_handler(_,dynamicCode,fail).


 

:-
	add_search_path(library,satsrc(plib)),
	saturate_directory( P ),
	add_search_path(sathome,P),
	name(P,PN),
	append1(PN,"/sources/DP",PSDPN),
	name(PSDP,PSDPN),
	add_search_path(dp,PSDP),
	append1(PN,"/sources",PSN),
	name(PS,PSN),
	add_search_path(satsrc,PS),
	append1(PN,"/sources/DP/examples",PEN),
	name(PE,PEN),
	add_search_path(examples,PE).



:- use_module(satsrc(mylists),[is_list/1]).


portray(X):-
	var(X),
	!,
	fail.
portray(date(Y,M,D)):-
	format('~d ~p ~p',[D,month(M),Y]).
portray(time(H,M,S)):-
	format('~d:~p:~p',[H,dig2(M),dig2(S)]).
portray(dig2(Z)):-
	Z1 is Z mod 100,
	(Z1<10 -> write(0)
	;true),
	write(Z1).
portray('$list'(X)):-
	print('$list'(X,',')).
portray('$listnl'(X)):-
	print('$list'(X,'
')).
portray('$list'([],_)).
portray('$list'([X],_)):-
       !,
       print(X).
portray('$list'([X|Xs],Sep)):-
	print(X),
	print(Sep),
	print('$list'(Xs,Sep)).
portray('$neq'(L,R)):-
	format('(~p \\= ~p)',[L,R]).
portray(month(0)):-write('January').
portray(month(1)):-write('February').
portray(month(2)):-write('March').
portray(month(3)):-write('April').
portray(month(4)):-write('May').
portray(month(5)):-write('June').
portray(month(6)):-write('July').
portray(month(7)):-write('August').
portray(month(8)):-write('September').
portray(month(9)):-write('October').
portray(month(10)):-write('November').
portray(month(11)):-write('December').
portray(X):-
	user_portray(X).


%==============================================================================
%  start : welcome message, install defaults and initialize
%==============================================================================



start:- 
	savedOn(D,T),
	prolog_version(P),
	satellite_version(V),
	os(S),
	format('~2nSatellite Version ~p (~p-~p) of ~p at ~p.~2n',[V,S,P,D,T]).





:-
	use_module(satsrc(os)),
	use_module(satsrc(thisProlog)           ,[prolog_version/1,
						  savedOn/2,
						  save_state/0,
						  save_state/1
						  ]),
	use_module(lpo,[]),   % need to load all ordering modules
	use_module(lpoc,[]),
	
	use_module(equational,[]),    % load all theory modules
	use_module(cong_closure,[]),
	
	ensure_loaded(error_handling),
	use_module(saturation),
%	use_module(global_vars),  % avoids segmentation faults
	save_state,
	start.




/*

to be done/fixed

extra variables on right side: equation might be orientable
  because of top predicates; should be checked


reconsider proof computation: indirect orphans not detected;
  reconsider adt_proof

handling of trivial goals


*/