/* -----------------------------------------------------------------------

Installation steps:

1. Edit ``saturate_directory'' below to where you have unloaded the system.
2. cd to the sources subdirectory.
3. Call Prolog.
4. Generate Saturate by saying

?- compile(sat), genSystem.

5. At the end you'll be asked for the name of a file
on which the system is to be saved.
Write file names that are no Prolog atoms with quotes
such as in the example below and put a dot ``.'' at the end.

6. Start using Saturate right away or exit by typing ^D.
Subsequently you call Saturate by executing the saved file.


-------------------------------------------------------------------------*/


:- use_module(library(system),[shell/2,file_exists/1]).


% change these definitions appropriatetely

os(windows):-
	file_exists('c:/').
os(linux):-
	shell('uname -a | grep Linux > /dev/null',0),
	!.
os(solaris):-	% it better not be SunOS4
	shell('uname -a | grep SunOS > /dev/null',0),
	!.
os(darwin):-
	shell('uname -a | grep Darwin > /dev/null',0).




				% Satellite home directory:
:- dynamic install_directory/1.

% put your directory here and uncomment
% :- assert(install_directory( '/HG/local/provers/saturate' ))).


% comment this call as it only works in my environment
:- 	(os(windows) ->
	    shell(testhost,H),
				%	format('~nMachine ~p~n',[H]),
	    ( H=100 -> M=mpino2100
	    ;	H=2 -> M=mpino2202),
	    
	    ( M=mpino2100 ->
		assert(install_directory( 'e:/home/hg/saturate' )),
		format('~nCompiling Saturate on Machine ~p~n',['MPINO2100'])
	    ;M=mpino2202 ->
		assert(install_directory( 'c:/home/hg/saturate' )),
		format('~n\%*Compiling Saturate on Machine ~p~n',['MPINO2202'])
	    )
	;os(darwin) ->
	    assert(install_directory( '/Users/hg/saturate' ))
	;% linux or solaris
	    assert(install_directory( '/HG/local/provers/saturate' ))
	).


%==============================================================================

            % you're not supposed to change anything from here on

%==============================================================================


satellite_version('0.1').




:- op(1150,fx,mymultifile).



:- op(1020,xfx,if).
:- op(940,yfx,[and,'&']).
:- op(930,yfx,[or,'|']).
:- op(950,xfx,[<=>]).
:- op(950,xfy,[=>]).
%:- op(900,fy,[~,dia,box]).
:- op(900,fy,[dia,box]).
:- op(700,yfx,@).    % apply operator
:- op(500,fy,[~]).   % should be smaller than :
:- op(600,yfx,[::]).
:- op(600,xfx,[^^]).
:- op(100,fx,[!,?]).
:-op(1050,xfx,<-).
:-op(1150,fx,goal).





:- multifile  
	portray/1,
	file_search_path/2, 
	query_hook/6, 
	message_hook/3, 
	portray_message/2.

:- dynamic 
	file_search_path/2, 
	query_hook/6, 
	message_hook/3, 
	portray_message/2,
	spy_clause/1,
	user_portray/1,
	no_portray/0.


append([], List, List).
append([Head|Tail], List, [Head|Rest]) :- 
        append(Tail, List, Rest).



add_search_path(Name,P):-
	file_search_path(Name,P),
	!.
add_search_path(Name,P):-
	asserta(file_search_path(Name,P)).

add_search_path_at_end(Name,P):-
	file_search_path(Name,P),
	!.
add_search_path_at_end(Name,P):-
	assertz(file_search_path(Name,P)).




:-
	install_directory( P ),
	name(P,PN),
	add_search_path(dp,P),
	add_search_path(satsrc,PS),
	append(PN,"/examples",PEN),
	name(PE,PEN),
	add_search_path(examples,PE).



:- use_module(hgstuff,[is_list/1]).


portray(X):-
	var(X),
	!,
	fail.
portray(date(Y,M,D)):-
	format('~d ~p ~p',[D,month(M),Y]).
portray(time(H,M,S)):-
	format('~d:~p:~p',[H,dig2(M),dig2(S)]).
portray(dig2(Z)):-
	Z1 is Z mod 100,
	(Z1<10 -> write(0)
	;true),
	write(Z1).
portray('$list'(X)):-
	print('$list'(X,',')).
portray('$listnl'(X)):-
	print('$list'(X,'
')).
portray('$list'([],_)).
portray('$list'([X],_)):-
       !,
       print(X).
portray('$list'([X|Xs],Sep)):-
	print(X),
	print(Sep),
	print('$list'(Xs,Sep)).
portray('$neq'(L,R)):-
	format('(~p \\= ~p)',[L,R]).
portray(month(0)):-write('January').
portray(month(1)):-write('February').
portray(month(2)):-write('March').
portray(month(3)):-write('April').
portray(month(4)):-write('May').
portray(month(5)):-write('June').
portray(month(6)):-write('July').
portray(month(7)):-write('August').
portray(month(8)):-write('September').
portray(month(9)):-write('October').
portray(month(10)):-write('November').
portray(month(11)):-write('December').
portray(X):-
	user_portray(X).


%==============================================================================
%  start : welcome message, install defaults and initialize
%==============================================================================



start:- 
	savedOn(D,T),
	prolog_version(P),
	satellite_version(V),
	os(S),
	format('~2nSatellite Version ~p (~p-~p) of ~p at ~p.~2n',[V,S,P,D,T]).





:-
	use_module(thisProlog,[
			       prolog_version/1,
			       savedOn/2,
			       save_state/0,
			       save_state/1
			      ]),
	use_module(lpo,[]),   % need to load all ordering modules
	use_module(lpoc,[]),
	
	use_module(equational,[]),    % load all theory modules
	use_module(cong_closure,[]),
	
	ensure_loaded(error_handling),
	use_module(saturation),
%	use_module(global_vars),
	save_state,
	start.




/*

to be done/fixed

extra variables on right side: equation might be orientable
  because of top predicates; should be checked


reconsider proof computation: indirect orphans not detected;
  reconsider adt_proof

handling of trivial goals


*/