%declarations%

:- module(adt_flag, [ 
	active_flag_numbers/2,
	compile_flags/2,
	create_initial_default_file/0,
	default_flag_numbers/2,
	save_defaults/2,
	compile_defaults/0,
	set_default_flags_as_on_flags/0
   ]).

:- use_module(props, [
        compileProps/1,
	default_flags/2,
	flag_on/2
   ]).

:- use_module(adt_declarations, [
	all_flag_types/1
   ]).

:- use_module(auxiliary, [
	delete_one312/3
   ]).

:- use_module(global, [
	cont/2
   ]).

:- use_module(os, [
	command_name/2
   ]).

:- use_module(thisProlog,[portray_clause2/2]).

%end%

%							ADT_FLAG

%------------------------------------------------------------------------------
% ci,ir,ma,ms,rp,tp: -> flag_type	% see module adt_declarations
% all_keys: flag_type -> [key]
% all_numbers: flag_type -> [nat]
% all_texts: flag_type -> [text]
% flag_on: flag_type -> [key]
% default_flags: flag_type -> [[nat],[key]]
%------------------------------------------------------------------------------


%------------------------------------------------------------------------------
% default_flags( Flagtype, [NumberList,KeyList] ).
%------------------------------------------------------------------------------

active_flags( Flagtype, [NL,KL] ):-
	( bagof( Key, flag_on( Flagtype, Key ), KL ) ; KL = [] ), !,
	relate_nums_with_keys( Flagtype, NL, KL ), !.

active_flag_numbers( Flagtype, NL ):-
	( bagof( Key, flag_on( Flagtype, Key ), KL ) ; KL = [] ), !,
	relate_nums_with_keys( Flagtype, NL, KL ), !.

default_flag_numbers( Flagtype, NL ):-
	default_flags( Flagtype, [NL,_] ), !.

%------------------------------------------------------------------------------
% creates the key word list for the given nat list
%------------------------------------------------------------------------------

relate_nums_with_keys( _, [], [] ):- !.
relate_nums_with_keys( Flagtype, [N|NL], [K|KL] ):-
	Call =..[Flagtype,N,K,_],
	adt_declarations:call( Call ),        % sic
        relate_nums_with_keys( Flagtype, NL, KL ), !.

%--------------------------------------------------------------------------------
% sets the default flag options as the current active flag options
%--------------------------------------------------------------------------------

set_default_flags_as_on_flags:-
	all_flag_types( FL ),
	cont(int_flags_tmp_file,T),
        open(T,write,StrT),
	portray_clause2(StrT,(flag_on( _, _ ) :- fail) ),
	write_default_flags_as_on_flags( StrT, FL ),
        close( StrT ),
        compileProps( T ), !.


write_default_flags_as_on_flags( _, [] ):- !.
write_default_flags_as_on_flags( Str, [Flagtype|FTL] ):-
	default_flags( Flagtype, [_,KL] ),
	write_flag_on( Str,Flagtype, KL ),
	write_default_flags_as_on_flags( Str,FTL ), !.

write_flag_on( _, _, [] ):- !.
write_flag_on( Str, Flagtype, [Key|Rest] ):-
	portray_clause2(Str, flag_on( Flagtype, Key ) ),
	write_flag_on( Str, Flagtype, Rest ), !.

%------------------------------------------------------------------------------
% saves the current active flag options as the default flag options
%------------------------------------------------------------------------------

save_defaults( Flagtype , Others):-
        active_flags( Flagtype, L ),
	T='defaults',
        tell( T ),
	write( default_flags( Flagtype, L ) ), write('.'), nl,
        all_flag_types( FTL ),
        delete_one312( Flagtype, FTL, FTL1 ),
        (Others=defaults ->
	     write_default_flags_as_default_flags( FTL1 )
	;
	     write_default_flags_as_default_flags( FTL1 )
	),
        close( T ),
	copy_defaults_tmp(F),
        compileProps( F ), !.


write_default_flags_as_default_flags( [] ):- !.
write_default_flags_as_default_flags( [Flagtype|FTL] ):-
	default_flags( Flagtype, F ),
	write( default_flags( Flagtype, F ) ), write('.'), nl,
	write_default_flags_as_default_flags( FTL ), !.

write_active_flags_as_default_flags( [] ):- !.
write_active_flags_as_default_flags( [Flagtype|FTL] ):-
	active_flags( Flagtype, F ),
	write( default_flags( Flagtype, F ) ), write('.'), nl,
	write_active_flags_as_default_flags( FTL ), !.

%------------------------------------------------------------------------------
% creates an initial default file (all flags are off)
% used when there is no user initial default file
%------------------------------------------------------------------------------

create_initial_default_file:-
	user:saturate_directory(D),
	name(D,DN),
	append(DN,"/sources/defaults.pl",DN1),
	command_name(cp,CP),
	append(CP,DN1,Co1),
	append(Co1," ./defaults",Co2),
	name(Command,Co2),
	unix(system(Command)).


write_empty_default_flags( [] ):- !.
write_empty_default_flags( [Flagtype|FTL] ):-
	write( default_flags( Flagtype, [[],[]] ) ), write('.'), nl,
	write_empty_default_flags( FTL ), !.


%------------------------------------------------------------------------------
% compiles a new user flag setting as the current active flag option
%------------------------------------------------------------------------------

compile_flags( Flagtype, NL ):-
	relate_nums_with_keys( Flagtype, NL, KL ), !,
	cont(int_flags_tmp_file,T),
        open( T, write, StrT ),
	write( StrT, 'flag_on( _, _ ) :- fail.' ), nl(StrT),
	write_flag_on( StrT, Flagtype, KL ),
	write_other_flags( StrT, Flagtype ),
        close( StrT ),
        compileProps( T ),
	!.

% writes the active flags with type different from Flagtype
write_other_flags( Str, Flagtype ):-
	all_flag_types( FL ),
	delete_one312( Flagtype, FL, FL1 ),
	write_active_flags_as_on_flags( Str, FL1 ), !.

write_active_flags_as_on_flags( _, [] ):- !.
write_active_flags_as_on_flags( Str, [Flagtype|FTL] ):-
	active_flags( Flagtype, [_,KL] ),
	write_flag_on( Str, Flagtype, KL ),
	write_active_flags_as_on_flags( Str, FTL ), !.




compile_defaults:-
	nofileerrors,
	copy_defaults_tmp(T),
	compileProps( T ),
	fileerrors, !.
compile_defaults:-
	write('* creating defaults file'),
	nl,
	create_initial_default_file,
	copy_defaults_tmp(T),
	compileProps( T ),
	fileerrors, !.


copy_defaults_tmp(T):-
	cont(defaults_tmp_file,T),
	name(T,TN),
	command_name(cp,CP),
	user:append(CP,"defaults ",CPD),
	user:append(CPD,TN,ComN),
	name(Com,ComN),
	unix(system(Com)).









% eof----------------------------------------------------------------------------








