
%declarations%

:- module(adt_precedence, [ 
        abbreviation_for_literal/4,
        abbreviating_proposition/1,
        adaptPrecedence/2,
	compile_precedence/0,
	data_complexity/2,
	define_precedence/2,
	definition/2,
	generated_proposition/1,
	greater_op/2,
	has_external_sort/1,
	is_first_predicate/1,
	is_magic_goal_predicate/1,
	is_operator/1,
	is_predicate/1,
	is_propositional_variable/1,
	is_skolem_function/1,
	is_total/2,
	minimal_constant/1,
	max_skolem_function/1,
	new_function_symbol_below/2,
        new_minimal_function_symbols/1,
	new_predicate_symbol_below/2,
	new_proposition_symbol/1,
	new_proposition_symbol/2,
	new_skolem_function/2,
	oriented_definition/3,
	precedence_as_facts/1,
	predicate/1,
	record_generated_proposition/1,
	record_skolem_function/1,
	strip_arity/2,
	special_symbol_conv/2,
	split_proposition/1		    
			  
   ]).


:- use_module(library(maplist),[maplist/3]).

:- use_module(more_maps,library(moremaps),[include/3]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(sets), [
	subtract/3
   ]).

:- use_module(mylists, [
	delete/3,
	nth0/4
   ]).

:- use_module(myterms,[
        general_function_call/1,
	linear/1,
	subset_vars/2
   ]).

:- use_module(proofState, [
        abbreviationFor/3,
	clause/1,
	firstPredicatePrecedence/1,
	precedence/1,
	deleteAllFirstPredicatePrecedences/0,
	deleteAllPrecedences/0,
	recordAbbreviation/3,
	recordFirstPredicatePrecedence/1,
	recordOperator/1,
	recordPrecedence/1,
	recordPredicate/1,
	recordProps/1,
	operators/1,
	predicates/1
   ]).

:- use_module(classes, [
        antecedent/2,
	kind/2,
	negate/2,
	succedent/2,
	theory_flag/2
   ]).

:- use_module(equationalTheory, [match/2]).

:- use_module(adt_atom, [
	 atom_as_term/2,
	 create_atom/2,
         get_atom/4,
	 get_atom_members/3,
	 is_equational_atom/1,
	 is_equivalence/1,
	 is_term_equation/1,
	 is_standard_atom/1,
	 lifted_equation/2,
	 predicate_in_atom/2
   ]).

:- use_module(adt_term, [
         deBruijn_index/1,
	 free_vars/2,
	 instantiate/1,
         logical_operator/1,
         logical_symbol_complexity/2,
	 path_arg1/3
   ]).

:- use_module(input, [analyse_clause/3,sat_meta_predicate/1]).

:- use_module(output, [store_props/0]).

:-use_module(library(flatten),[flatten/2]).

:-use_module(props,[
         expand_definition_for/1,
         firstOrder/0,
	 max_cases/1,
	 max_deBruijn/1,
	 max_skolem/1,
	 op_complexity_/2,
	 partial_equivalence_relation_symbol/1,
	 predicate_with_extension/3,
	 skolem_functions/1,
	 special_relation/1
   ]).

:-use_module(global,[cont/2,incr/1,store/2]).

:-use_module(auxiliary,[appendAtoms/3,interval/3]).

:-use_module(hgstuff,[setof0/3]).

:-use_module(thisProlog,[prolog_version/1]).

:-use_module(library(sets),[select/3]).

:- use_module(library(samsort),[samsort/3]).

:- use_module(softtypes,[original_predicate/1]).


:- ((prolog_version(sicstus);prolog_version(sicstus3)) ->
	use_module(library(ugraphs))
   ;
	use_module(slibrary(ugraphs))
   ).
	


%end%

%						ADT_PRECEDENCE

%______________________________________________________________________________
%
% precedence( L1 ), L1 = [f1,f2, ..., fr] where f1>f2> ... >fr
% predicates( L2 ), L2 = [p1,p2, ..., pr] where p1>p2> ... >pr
% firstPredicatePrecedence( L3 ), L3 = [q1,q2, ..., qk]
%				
% it is supposed that L3 is a subset of L2 and
% they are considered to be the bigger operators
%
% it is also supposed that L1 and L3 are disjoint	
%______________________________________________________________________________

 

is_predicate( P/_ ):-
	var(P),
	!,
	fail.
is_predicate( P/_ ):-
	\+atomic(P),
	!,
	fail.
is_predicate( PN ):-
	% tests for predicateship 
	% P can be a variable, and then it should fail.
	predicates( L ),
	member( PN, L ).
is_predicate( P/N ):-
	generated_proposition(P/N).



is_propositional_variable( P ):-
	nonvar(P),
	(  predicates( L ),
	   member( P/0, L )
	;
	   name(P,[112,36|_])  %p 
	;
	   name(P,[113,36|_])  %q
	),
	!.

is_magic_goal_predicate(P):-
	name(P,N),
	name('goal$',G),
	append(G,_,N),
	!.


predicate( P/N ):-
	predicates( L ),
	member( P/N, L ).

is_predicate_no_arity(P):-
	is_predicate(P/_).


is_operator( P ):-
	operators( L ),
	member( P, L ), 
	!.

operator( P ):-
	operators( L ),
	member( P, L ).


is_operator_no_arity(P):-
	is_operator(P/_).

is_skolem_function(F/_):-
	name(F,[115,36|_]),  % s$...
	!.

special_symbol_conv(F/N,O/N):-
	var(O),
	(F='$true$',N=0 ->
	    O=true
	;F='$min',N=0 ->
	    O=false
	;
	    name(F,[N1,36|N2]),
	    name(O,[115,115,N1|N2])
	).
special_symbol_conv(F/N,O/N):-
	var(F),
	(O=true,N=0 ->
	    F='$true$'
	;O=false,N=0 ->
	    F='$min'
	;
	    name(O,[115,115,N1|N2]),
	    name(F,[N1,36|N2])
	).

max_skolem_function(S):-
	max_skolem(M),
	name(M,MN),
	name(S,[115,36|MN]).

is_first_predicate( P ):-
	firstPredicatePrecedence( L ),
	member( P, L ), !.

has_external_sort(F):-
	% actually I don't remember for what kind of experiments
	% this is good for
	name(F,[36,36|_]).  % F='$$...'

minimal_constant(M):-
	operators(O),
	precedence(L),
	member(M/0,O),
	append(_,[M|Smaller],L),
	\+ (member(M1/0,O), member(M1,Smaller)),
	!.

new_proposition_symbol(P):-
	new_proposition_symbol(abbreviation,P).

new_proposition_symbol(caseSplit,P):-
	incr(propCtr),
	cont(propCtr,M),
	max_cases(Max),
	M=<Max,
	store(propCtr,M),
	M1 is Max-M,
	number_chars(M1,M1C),
	atom_chars(M1A,M1C),
	appendAtoms([p,M1A],'$',P).
new_proposition_symbol(abbreviation,P):-
	incr(propCtr),
	cont(propCtr,M),
	max_cases(Max),
	M=<Max,
	store(propCtr,M),
	M1 is Max-M,
	number_chars(M1,M1C),
	atom_chars(M1A,M1C),
%	appendAtoms([p,M1],'$',P).
	appendAtoms([q,M1A],'$',P).
new_proposition_symbol(_,_):-	
	raise_exception(saturate_error('max number aux propositions exceeded',[])).



abbreviation_for_literal([Sign|L],_LA,[Sign|L],old):-
	is_standard_atom(L),
	!.
abbreviation_for_literal([Sign|L],_LA,[Sign|PA],old):-
	atom_as_term(L,T),
	abbreviationFor(L1,Sign,P),
	match(L1,T),
	!,
%	format('~nfound ~p ~p == ~p ~p',[Sign,P,Sign,T]),
	create_atom(P,PA).
/*
abbreviation_for_literal([Sign|L],[Sign1|PA],new):-
	atom_as_term(L,T),
	negate([Sign],[Sign1]),
	abbreviationFor(L1,Sign1,P),
	subsumes_chk(L1,T),
	!,
	recordAbbreviation(L1,Sign,P),
	match(L1,T),    
	create_atom(P,PA),
	format('~nfound ~p ~p == ~p ~p',[Sign1,P,Sign,T]).
*/
abbreviation_for_literal([Sign|_],LA,[Sign|PA],new):-
	new_proposition_symbol(PSymbol),
	atom_as_term(LA,T),
	free_vars(T,Vs),
	P=..[PSymbol|Vs],
	recordAbbreviation(T,Sign,P),
%	format('~nexpansion ~p ~p == ~p ~p',[Sign,P,Sign,T]),
	create_atom(P,PA),
	!.
abbreviation_for_literal(_L,_PA,_Status):-
	raise_exception(saturate_error('too many propositional variables',[])).



new_minimal_function_symbols(Fs):-
	precedence(Prec),
	deleteAllPrecedences,
	maplist(recordOperator,Fs),
	maplist(strip_arity,Fs,FsNoArity),
	sort(FsNoArity,FsNoArityS),
	append(Prec,FsNoArityS,Prec1),
	recordPrecedence(Prec1),
	compile_precedence.


new_function_symbol_below(Os,F/K):-
	cont(funcCtr,N),
	M is N+1,
	store(funcCtr,M),
	!,
	number_chars(N,NC),
	atom_chars(NA,NC),
	appendAtoms([f,NA],'$',F),
	precedence(Prec),
	samsort(gtt(Os,F),[F|Prec],Prec1),
	deleteAllPrecedences,
	recordPrecedence(Prec1),
	recordOperator(F/K),
	compile_precedence,
	!.

new_skolem_function(_Arity,F):-
	incr(skolemCtr),
	cont(skolemCtr,M),
	max_skolem(Max),
	M=<Max,
	!,
%	(_Arity==0 -> M1 is Max-M; M1=M),
	  % larger numbers are smaller in the ordering
	  % we make constants smaller, but generate them increasingly
	M1=M,
	number_chars(M1,M1C),
	atom_chars(M1A,M1C),
	appendAtoms([s,M1A],'$',F).
new_skolem_function(_Arity,_F):-	
	raise_exception(saturate_error('max number Skolem constants exceeded',[])).


record_skolem_function(F):-
	name(F,[115,36|NN]),
	name(N,NN),
	cont(skolemCtr,N1),
	M is max(N,N1),
	store(skolemCtr,M),
	max_skolem(Max),
	M=<Max,
	!.
record_skolem_function(_F):-	
	raise_exception(saturate_error('max number Skolem constants exceeded',[])).

record_generated_proposition(P):-
	(name(P,[113,36|NN]) % q$...
	;   
	name(P,[112,36|NN])),  % p$...
	name(N,NN),
	max_cases(Max),
	cont(propCtr,N1),
	N2 is Max-N,
	M is max(N2,N1),
	store(propCtr,M),
	M=<Max,
	!.
record_generated_proposition(_F):-	
	raise_exception(saturate_error('max number aux propositions exceeded',[])).



new_predicate_symbol_below(Os,F/K):-
	(cont(precCtr,N);N=0),
	M is N+1,
	store(precCtr,M),
	!,
	number_chars(N,NC),
	atom_chars(NA,NC),
	appendAtoms([q,NA],'$',F),
	precedence(Prec),
	samsort(gtt(Os,F),[F|Prec],Prec1),
	deleteAllPrecedences,
	recordPrecedence(Prec1),
	recordPredicate(F/K),
	compile_precedence,
	!.


compile_precedence:-
	precedence_as_facts(G),
	recordProps(G), % should not append new op_complexity_ to old ones
	store_props,
	!.


gtt(Greater,New,O,P):-
	(O==New ->
	     P\==New,
	     (member(P/_,Greater) ->
		  false
	     ;member(P1/_,Greater), greater_op(P,P1) ->
		  false
	     ;
	          member(P1/_,Greater), greater_op(P1,P)
	     )
	;P==New ->
	     O\==New,
	     (member(O/_,Greater) ->
		  true
	     ;member(P1/_,Greater),  greater_op(O,P1) ->
		  true
	     ;member(P1/_,Greater),  greater_op(P1,O)
	     )
	;
	     greater_op(O,P)
	).
	   

	          
	          




%------------------------------------------------------------------------------
% greater_op( Op1, Op2):
%	checks whether Op1 is greater thar Op2 w.r.t the operator precedence
%
% there are reserved constant names, used in redundancy proofs by 
% Martin and Nipkow's techniques
%------------------------------------------------------------------------------


greater_op( X, Y ):- ( var(X);var(Y) ),
	raise_exception(saturate_error('~nerror in greater_op.',[])).
greater_op( X, X ):- !, fail.
greater_op(A,B):-
	op_complexity_(A,C),
	op_complexity_(B,D),
	!,
	C>D.




/* complexities of operators:

   $min                  = -1
   $true                 = 0
   prop variable     p$j = 1000-j (for case splits)
   logical symbols       in [1800,2000]

   1. bei Standard Klauseln

      user symbols          in [10001,18000] according to their precedence

   2. im first-order Fall:

      deBjuin-Index   i$j   = 2000+j
      skolem constants s$j  = 3000+j  
      predicates      q$j (abbreviations for subformulas)
                           in [4001,5000]
      user functions       in [10001,15000] according to their precedence
      skolem functions s$j  = 8000-j  
      user predicates      in [16001,19999] according to their precedence


      ordered variables     in [40000,40100]
      first predicates      in [51001,60000]
      $c                     = 100000


   The op_complexity relation will be compiled into op_complexity_ facts
   in module props.

*/

op_complexity('$min',-1).
op_complexity('$true$',0).
op_complexity(P,C):-
	max_cases(M),
	interval(1,M,Int),
	member(I,Int),
	C is 1000-I+1,
	name(I,IN),
	name(P,[112,36|IN]).
%op_complexity(P,C):-		% shouldn't be smaller than abbreviations; incomplete
%	precedence(Prec),
%	predicate(P/0),
%	length(Prec,N),
%	nth0(I,Prec,P,_),
%	C is 1000+N-I.
op_complexity(P,C):-
	logical_symbol_complexity(P,C).
op_complexity(P,C):-
	max_deBruijn(M),
	interval(1,M,Int),
	member(I,Int),
	C is 2000+I,
	name(I,IN),
	name(P,[105,36|IN]). % i$...
op_complexity(P,C):-
	precedence(Prec),
	operator(P/_),
	length(Prec,N),
	nth0(I,Prec,P,_),
	C is 10000+N-I.
op_complexity(P,C):-
	% skolem function; new ones have to be smaller than old ones
	% smaller than user functions
	skolem_functions(small),
	max_skolem(M),
	interval(1,M,Int),
	member(I,Int),
	C is 8000-I,         % small skolem functions
	name(I,IN),
	name(P,[115,36|IN]). % s$...
op_complexity(P,C):-
	% skolem functions; new ones are smaller than old ones???
	% they are bigger than user functions
	skolem_functions(big),
	max_skolem(M),
	interval(1,M,Int),
	member(I,Int),
	C is 16000-I,        % bigger than user-defined functions
%	I2 is 2*I+1,         % odd numberef
	name(I,IN),
	name(P,[115,36|IN]). % s$...
op_complexity(P,C):-
	precedence(Prec),
	predicate(P/_M),
%	_M>0,
	length(Prec,N),
	nth0(I,Prec,P,_),
%	(firstOrder ->       % this test is not compatible with module structure
	                     % and hence may lead to cyclic precedences
%	     C is 16000+N-I   
%	;
	     C is 10000+N-I
%	)
.
op_complexity(P,C):-
	max_cases(M),
	interval(1,M,Int),
	member(I,Int),
	name(I,IN),
	name(P,[113,36|IN]),   % q
	name(I,IN),
%	C is 15000+I. 
	C is 4000+I.   % newer ones smaller ( minus in new_proposition...)
%	C is 5000-I.
op_complexity('a$',40004).
op_complexity('b$',40003).
op_complexity('c$',40002).
op_complexity('d$',40001).
op_complexity(P,C):-
	firstPredicatePrecedence(Prec),
	length(Prec,N),
	nth0(I,Prec,P,_),
	C is 51000+N-I.
op_complexity('$c',100000).



/*


% from 11/2001

op_complexity('$min',-1).
op_complexity('$true$',0).
op_complexity(P,C):-
	max_cases(M),
	interval(1,M,Int),
	member(I,Int),
%	C is 1000-I+1,
	C is 1000-M+I,
	name(I,IN),
	name(P,[112,36|IN]).
op_complexity(P,C):-
	
	precedence(Prec),
	predicate(P/0),
	length(Prec,N),
	nth0(I,Prec,P,_),
	C is 1000+N-I.
op_complexity(P,C):-
	logical_symbol_complexity(P,C).
op_complexity(P,C):-
	max_deBruijn(M),
	interval(1,M,Int),
	member(I,Int),
	C is 2000+I,
	name(I,IN),
	name(P,[105,36|IN]). % i$...
op_complexity(P,C):-
	precedence(Prec),
	operator(P/_),
	length(Prec,N),
	nth0(I,Prec,P,_),
	C is 10000+N-I.
%op_complexity(P,C):-
	% skolem constants (arity 0); new ones have to be bigger than old ones
	% smaller than user functions
%	max_skolem(M),
%	interval(1,M,Int),
%	member(I,Int),
%	C is 16000-I,         % small skolem functions
%	I2 is I*2,           % skolem constants are even-numbered
%	name(I2,IN),
%	name(P,[115,36|IN]). % s$...
op_complexity(P,C):-
	% skolem functions; new ones are smaller than old ones???
	% they are bigger than user functions
	max_skolem(M),
	interval(1,M,Int),
	member(I,Int),
	C is 16000-I,        % bigger than user-defined functions
%	I2 is 2*I+1,         % odd numberef
	name(I,IN),
	name(P,[115,36|IN]). % s$...
op_complexity(P,C):-
	precedence(Prec),
	predicate(P/M),
	M>0,
	length(Prec,N),
	nth0(I,Prec,P,_),
	(firstOrder ->    
	     C is 16000+N-I   
	;
	     C is 10000+N-I
	).
op_complexity(P,C):-
	max_cases(M),
	interval(1,M,Int),
	member(I,Int),
	name(I,IN),
	name(P,[113,36|IN]),   % q
	name(I,IN),
%	C is 20000+I.
	C is 4000+I.
op_complexity('a$',40004).
op_complexity('b$',40003).
op_complexity('c$',40002).
op_complexity('d$',40001).
op_complexity(P,C):-
	firstPredicatePrecedence(Prec),
	length(Prec,N),
	nth0(I,Prec,P,_),
	C is 51000+N-I.
op_complexity('$c',100000).

*/

data_complexity(O,C1):-
	((operator(O/_);predicate(O/_)) ->
	    precedence(Prec),
	    length(Prec,N),
	    op_complexity_(O,C),
	    C1 is (C-10000)/N
	;   
	    C1 is 1
	).


logical_op('a$').
logical_op('b$').
logical_op('c$').
logical_op('d$').
logical_op(O):-
	logical_operator(O).
logical_op(O):-
	generated_proposition(O/_).
logical_op(O):-
	is_skolem_function(O/_).



generated_proposition(P/_):-
	name(P,[112,36|_]).  % p$...
generated_proposition(P/_):-
	name(P,[113,36|_]).  % q$...

abbreviating_proposition(P/_):-
	name(P,[113,36|_]).  % q$...

split_proposition(P/_):-
	name(P,[112,36|_]).  % p$...
	




%--------------------------------------------------------------------------------
% is_total( L1, L2 ):
%	L1 is a list of precedences (which are lists of operators )
% 	checks whether the precedence represented by L1 is total
%	returns a simplified single precedence list L2 
%--------------------------------------------------------------------------------

is_total( L, [] ):-       
        list_of_empty_lists( L ), !.
is_total( L, [X|R] ):-
        unique_maximal( X, L ),
        delete_of_precedence( L, X, L1 ),
        is_total( L1, R ), !.

unique_maximal( _, L ):-
        \+ maximal(_,L),
        nl, nl, write(' Circularity: incompatible precedences!'), nl, !, fail.
unique_maximal( _, L ):-
        maximal( X, L ),
        maximal( Y, L ),
        Y \== X,
        nl, nl, write(' The precedence is not total: '),
        write(X), write(' and '), write(Y),
        write(' are not comparable!'),
        nl, nl,
        write( ' Please use command "prec" to complete its definition' ),
        nl, nl, !, fail.

unique_maximal( X, L ):-
        maximal( X, L ), !.

maximal(X,L):-                          % X is a maximal element in L
        member( [X|_], L ),             % X es max en alguna precedencia de L
        \+ gr(L, _,X ).                 % X es maximal en todo L (tampoco X>X).

gr( L, X, Y ):-                         % X > Y en alguna precedencia de L
        member( P, L ),
        user:append( _, [X|R], P ),
        member( Y, R ).

delete_of_precedence( [], _, [] ):-!.
delete_of_precedence( [P|R], X, [P1|R1] ):-
        delete(P,X,P1), delete_of_precedence(R,X,R1),!.


list_of_empty_lists( L ):- \+ member([_|_],L), !.





all_operators_related( TL, NotDecl ):-
	predicates(Ps),
	operators( Os ),
	append(Ps,Os,Ops),
	setof(X,(exists_one_not_related( Ops, TL, X ),
		 \+ builtInOp(X)),NotDecl),
	!.
all_operators_related( _TL, [] ).






	
exists_one_not_related( Ops, TL, X ):-
	member22( X/_, Ops ),
	\+ member212( X, TL ).

disjoint_preds( L1, L2 ):-
	member22( X1, L1 ),
	member22( X2, L2 ),
	X1 == X2,
	nl, nl, write( '*** The predicate '), write( X1 ),
	write( ' is declared twice in the precedence,' ),
	nl, !, fail.
disjoint_preds( _, _ ).




% precedence adaptation for changing between lpo and sto
% going from lpo to sto we have to delete first_predicate_precedences,
% we do that by appending these to the precedence of the other symbols,
% making the former minimal

adaptPrecedence(lpo,sto):-
	firstPredicatePrecedence(FP),
	precedence(P),
	store(lpofp,(FP,P)),
	deleteAllFirstPredicatePrecedences,
	deleteAllPrecedences,
	recordFirstPredicatePrecedence([]),
	append(P,FP,PFP),
	recordPrecedence(PFP).
adaptPrecedence(sto,lpo):-
	cont(lpofp,(FP,P)),
	deleteAllFirstPredicatePrecedences,
	deleteAllPrecedences,
	recordPrecedence(P),
	recordFirstPredicatePrecedence(FP).
adaptPrecedence(_,_).

% for experimenting with renaming

dual_predicate_pair('+eq','-eq').
dual_predicate_pair('+p','-p').
dual_predicate_pair('+q','-q').
dual_predicate_pair('+r','-r').

dual_predicate(P):-
	dual_predicate_pair(P,_).
dual_predicate(P):-
	dual_predicate_pair(_,P).

% ------------------------ precedence finding --------------------------------

define_precedence_from_edges(EsF,Es,FPPrec,Prec):-
	vertices_edges_to_ugraph([],EsF,GF),
	vertices(GF,FirstPreds),
	vertices_edges_to_ugraph([],Es,G),
	vertices(G,Others),
	(member(O,Others),member(O,FirstPreds) ->
	     format('~2nCircular precedence: ~p occurs both as first predicate and otherwise.',[O]),
	     fail
	 ;
	     true
 	),
	(cycle(GF,Path) ->
	     format('~2nCircular precedence among first predicates: ~p.',[Path]),
	     fail
	 ;
	     true
 	),

	(cycle(G,Path) ->
	     format('~2nCircular precedence: ~p.',[Path]),
	     fail
	 ;
	     true
 	),
	
	complete_precedence(GF,G,FPPrec,Prec),
	!.




precedence_as_facts(G):-
	setof0(op_complexity_(A,B),op_complexity(A,B),G).




precedes(A,B,L,_):-
	append(_,[A|Right],L),
	member(B,Right),
	!.
precedes(A,B,_,L):-
	append(_,[A|Right],L),
	member(B,Right),
	!.
precedes(A,B,F,L):-
	member(A,F),
	member(B,L),
	!.




complete_precedence(GF,G,FP1,Prec):-
	top_sort(GF,_),
	top_sort(G,_),
	findPrecedence(FPAuto,PrecAuto),
	vertices(GF,FPs),
	vertices(G,Ps),
	(FPs=[] -> FPs1=FPAuto; FPs=FPs1),
	(FPs=[],Ps=[] ->
	     FP1=FPAuto,
	     Prec=PrecAuto
	;
	     precedenceAsEdges(FPAuto,EFPAuto),
	     precedenceAsEdges(PrecAuto,EPrecAuto),
	     append(EFPAuto,EPrecAuto,EAll),
	     append(FPs,FPAuto,FPsA),
	     merge_edges_into(EAll,FPsA,GF,GFC),
	     merge_edges_into(EAll,-FPsA,G,GC),
	     top_sort(GFC,FP1),
	     top_sort(GC,Prec1),
%	     (firstOrder->
%		  include(is_predicate_no_arity,Prec1,PrecPs),
%		  include(is_operator_no_arity,Prec1,PrecOs),
%		  append(PrecPs,PrecOs,Prec)
%	     ;
	          Prec1=Prec
%	     )
		  
	).



merge_right([],Ys,Ys).
merge_right([X|Xs],Ys,Zs):-
	append(Ys1,[X|Ys2],Ys),
	!,
	subtract(Xs,Ys1,Xs1),
	merge_right(Xs1,Ys2,Zs2),
	append(Ys1,[X|Zs2],Zs).
merge_right([X|Xs],Ys,[X|Zs]):-
	% X not in Ys
	merge_right(Xs,Ys,Zs).



merge_left_graphs(G1,G2,G3):-
	edges(G2,Es),
	merge_edges_into(Es,G1,G3).

merge_edges_into([],G,G).
merge_edges_into([P-Q|Es],G,G1):-
	(pathFromTo(Q,P,G) ->
	     merge_edges_into(Es,G,G1)
	;
	     add_edges(G,[P-Q],G2),
	     merge_edges_into(Es,G2,G1)
	).

merge_edges_into([],_,G,G).
merge_edges_into([P-Q|Es],-Vs,G,G1):-
	!,
	(\+member(P,Vs),\+member(Q,Vs),\+pathFromTo(Q,P,G) ->
	     add_edges(G,[P-Q],G2),
	     merge_edges_into(Es,-Vs,G2,G1)
	;member(Q,Vs),Es=[Q-P1|Es1] -> 
	     merge_edges_into([P-P1|Es1],-Vs,G,G1)
	;
	     merge_edges_into(Es,-Vs,G,G1)
	).
merge_edges_into([P-Q|Es],Vs,G,G1):-
	(member(P,Vs),member(Q,Vs),\+pathFromTo(Q,P,G) ->
	     add_edges(G,[P-Q],G2),
	     merge_edges_into(Es,Vs,G2,G1)
	;\+member(Q,Vs),Es=[Q-P1|Es1] -> 
	     merge_edges_into([P-P1|Es1],Vs,G,G1)
	;
	     merge_edges_into(Es,Vs,G,G1)
	).

cyclic_with(G,P-Q):-
	add_edges(G,[P-Q],G1),
	\+top_sort(G1,_).

cycle(G,G):-
	\+ top_sort(G,_).


define_precedence(FPPrec,Prec):-
	collect_first_predicate_precedence( LL1 ),
	precedencesAsEdges(LL1,EsF),
	collect_precedence( LL2 ),
	precedencesAsEdges(LL2,Es),
	(define_precedence_from_edges(EsF,Es,FPPrec1,Prec1) ->
	    true
	;
	    format('~nPrecedences from submodules cannot be combined:~n top predicates: ~p ~n other symbols: ~p',[LL1,LL2]),
	    fail
	),
	(select('$min',Prec1,Prec2) ->
	     true
	;
	     Prec1=Prec2
	),
	(flatten(LL1,[FirstP]), 
	 (select(FirstP,Prec2,Prec); Prec3=Prec2) ->
	     FPPrec=[FirstP|FPPrec1]
	     
	;
	     Prec2=Prec3,
	     FPPrec=FPPrec1
	),
	(flatten(LL2,[SomeSy]) ->
	  (member(SomeSy,Prec3) ->
	       Prec=Prec3
	  ;Prec=[SomeSy|Prec3]
	  )
	;Prec=Prec3
	).


precedencesAsEdges([],[]).
precedencesAsEdges([P|Ps],Es):-
	precedenceAsEdges(P,Es1),
	precedencesAsEdges(Ps,Es2),
	append(Es1,Es2,Es).

precedenceAsEdges([],[]).
precedenceAsEdges([_],[]).
precedenceAsEdges([P,Q|Ps],[P-Q|Es]):-
	precedenceAsEdges([Q|Ps],Es).


collect_first_predicate_precedence( LL ):-
	bagof( L, firstPredicatePrecedence( L ), LL ), !.
collect_first_predicate_precedence( [] ):- !.

collect_precedence( LL ):-
	bagof( L, precedence( L ), LL ), !.
collect_precedence( [] ):- !.

pathFromTo(A,E,G):-
	reachable(A,G,R),
	member(E,R),
	!.




%findPrecedence(OsOut,PsOut):-
findPrecedence(FPPrec,Out):-
	predicate_strata(G),
	top_sort(G,GS),
	maplist(samsort(gtGoalOrArity),GS,GSS),
	flatten(GSS,Ps11),
	splitFPs(Ps11,[],FPPrec1,Ps1),
	maplist(strip_arity,FPPrec1,FPPrec),

	functionPrecedence(Fs,Cs),
	maplist(strip_arity,Fs,FsOut),
	maplist(strip_arity,Cs,CsOut),
	maplist(strip_arity,Ps1,PsOut),
	subtract(PsOut,FsOut,PsOut1),
	subtract(PsOut1,CsOut,PsOut2),
	append(FsOut,PsOut2,Out1),
	append(Out1,CsOut,Out).


splitFPs([],Greater,[],Greater).
splitFPs([P/N|Ps],Greater,[P/N|FPs],Ps1):-
	predicate_with_extension(P/N,_SF,_),
	\+ (N=2, special_relation(P)),
	!,
	splitFPs(Ps,Greater,FPs,Ps1).
splitFPs([P/N|Ps],Greater,[P/N|FPs],Ps1):-
	has_definition(P/N),
	!,
	splitFPs(Ps,Greater,FPs,Ps1).
splitFPs([P/N|Ps],Greater,FPs,Ps1):-
	append(Greater,[P/N],Greater1),
	splitFPs(Ps,Greater1,FPs,Ps1).



functionPrecedence(Fs1,[]):-
	operators(Os1),
	samsort(greaterArity1,Os1,Os),
	setof0(Q-P,func_priority(Q,P),Es),
	vertices_edges_to_ugraph(Os,Es,G1),
	setof0(Q-P,(member(P,Os),member(Q,Os),greaterArity1(Q,P)),Es2),
	vertices_edges_to_ugraph(Os,Es2,G2),
	merge_left_graphs(G1,G2,G3),
	reduce(G3,G),
	top_sort(G,GS),
	maplist(samsort(greaterArity1),GS,GSS),
	flatten(GSS,Fs1).








split_constants(Os1,Fs,Cs):-
	(append(Fs1,[F/N,C/0|Cs1],Os1),N>=1 ->
	     append(Fs1,[F/N],Fs),
	     Cs=[C/0|Cs1]
	;Os1=[C/0|_] ->
	     Fs=[],
	     Cs=Os1
	;
	     Fs=Os1,
	     Cs=[]
	).

predicate_strata(G):-
	(cont(better_precedence,no) ->
	    predicates(Ps),
	    findall(Q-P,pred_priority(Q,P),Es),
	    acyclic_ugraph(Ps,Es,G1),
	    reduce(G1,G)
	;
	    predicate_strata_with_merge(G)).


predicate_strata_with_merge(G):-
	predicates(Ps),
	findall(Q-P,pred_priority(Q,P),Es),
	acyclic_ugraph(Ps,Es,G1),
	findall(Q-P,(member(P,Ps),member(Q,Ps),greaterArity(Q,P)),Es2),
	vertices_edges_to_ugraph(Ps,Es2,G2),
	merge_left_graphs(G1,G2,G3),
	reduce(G3,G).


acyclic_ugraph(Ps,Es,G):-
	vertices_edges_to_ugraph(Ps,[],G1),
	acyclic_ugraphE(Es,G1,G).


acyclic_ugraphE([P-Q|Es],G1,G):-
	add_edges(G1,[P-Q],G2),
	(top_sort(G2,_) ->
	    acyclic_ugraphE(Es,G2,G)
	;
	    acyclic_ugraphE(Es,G1,G)
	).
acyclic_ugraphE([],G,G).



% P is defined in terms of Q

has_definition(P/N):-
	positive_equivalence(E),
	is_equivalence(E),
	get_atom_members(E,L,R),
	(definition(L,R),
	    proper_predicate_in_term(L,P/N)
	;definition(R,L),
	    proper_predicate_in_term(R,P/N)),
	!.




pred_priority(P/N,Q/M):-
	expand_definition_for(P/N),
	positive_equivalence(E),
	is_equivalence(E),
	get_atom_members(E,L,R),
	(definition(L,R) ->
	     proper_predicate_in_term(L,P/N),
	     proper_predicate_in_term(R,Q/M)
	;definition(R,L) ->
	     proper_predicate_in_term(R,P/N),
	     proper_predicate_in_term(L,Q/M)
	).
pred_priority(P/N,Q/M):-
	clause(C),
	antecedent(C,A),
	succedent(C,[H]),
	member(B,A),
	proper_predicate_in_atom(H,P/N),
	proper_predicate_in_atom(B,Q/M),
	\+special_relation(P),
	\+special_relation(Q).


oriented_definition(E,L,R):-
	get_atom_members(E,L,R),
	definition(L,R).
%	proper_predicate_in_term(L,P/_N),
%	is_first_predicate(P),
%	\+ (   proper_predicate_in_term(R,Q/_M),
%	       (Q=P;greater_op(Q,P))).

	
	

% generate (not test!) priorities between function symbols
func_priority(F/N,G/M):-
	original_predicate(G/M),
	operator(F/N),
	\+ original_predicate(F/N).
func_priority(F/N,G/M):-
	positive_equation(E),
	(inverse_law(E,F,G) ->
	     N=1,
	     M=2
	;
	     get_atom_members(E,L,R),
	     pre_orient(L,R,L1,R1),
	     nonvar(L1),
	     functor(L1,F,N),
	     \+ logical_op(F),
	     path_arg1(_,R1,Sub),
	     nonvar(Sub),
	     functor(Sub,G,M),
	     \+ logical_op(G)
	).
func_priority(F/N,G/M):-
	constructor(G/M),
	operator(F/N),
	\+ constructor(F/N).
/*
func_priority(F/N,G/0):-
	operator(F/N),
	operator(G/0),
	N>0.
*/


constructor(F/N):-
	positive_equation(E),
	(inverse_law(E,F,_G) ->
	     N=1
	;
	     get_atom_members(E,L,R),
	     pre_orient(L,R,L1,_),
	     nonvar(L1),
	     functor(L1,F,N)
	),
	!,
	fail.
constructor(_).	

positive_equation(E1):-
	clause(C),
	succedent(C,S),
	member(E,S),
	lifted_equation(E,E1),
	is_term_equation(E1),
	is_equational_atom(E1).

	

positive_equivalence(E):-
	clause(C),
	succedent(C,S),
	member(E,S),
	is_equivalence(E).


inverse_law(E,F,G):-
	atom_as_term(E,T),
	analyse_clause([],[T],inverse(F,G,_,_)).



pre_orient(L,R,R,L):-
	subset_vars(L,R),
	\+ subset_vars(R,L),
	!.
pre_orient(L,R,R,L):-
	lpo:greater(R,L),
	!.
%/*   
%     causes unfolding of definitions
%     producing *very* large term in some SET examples
pre_orient(L,R,L,R):-
	subset_vars(R,L),
	\+ subset_vars(L,R),
	!.
pre_orient(L,R,R,L):-  
        definition(R,L),  % unfolding
%        definition(L,R),  % folding
	!.
pre_orient(L,R,L,R):-
        definition(L,R),  % unfolding
%        definition(R,L),  % folding
	!.
pre_orient(L,R,R,L):-
	subset_vars(L,R),
	linear(R),
	\+ linear(L),
	!.
pre_orient(L,R,L,R):-
	subset_vars(R,L),
	linear(L),
	\+ linear(R),
	!.
%*/
pre_orient(L,R,L,R).


definition(L,R):-
	general_function_call(L),
	\+general_function_call(R),
	!.



gtGoalOrArity(P,Q):-
	isGoalPredicate(P),
	\+ isGoalPredicate(Q),
	!.
gtGoalOrArity(P,Q):-
	isGoalPredicate(Q),
	\+ isGoalPredicate(P),
	!,
	fail.
gtGoalOrArity(P,Q):-
	greaterArity(P,Q).



proper_predicate_in_atom(H,P/N):-
	predicate_in_atom(H,P/N),
	\+ sat_meta_predicate(P/N),
	\+ logical_op(P).


proper_predicate_in_term(T,P/N):-
	path_arg1(_,T,Sub),
	nonvar(Sub),
	functor(Sub,P,N),
	is_predicate(P/N),
	\+ sat_meta_predicate(P/N),
	\+ logical_op(P).

isGoalPredicate(P):-
	clause(C),
	theory_flag(C,(goal)),
	(  antecedent(C,A),
	   member(E,A),
	   predicate_in_atom(E,P)
	;
	   succedent(C,A),
	   member(E,A),
	   predicate_in_atom(E,P)
	),
	!.

	



strip_arity(P/_,P).


greaterArity(P/N,Q/M):-
	t(N,P)@>t(M,Q).

greaterArity1(_/1,_/2):-
	!.
greaterArity1(_/2,_/1):-
	!,
	fail.
greaterArity1(P/N,Q/M):-
	greaterArity(P/N,Q/M).



% eof -------------------------------------------------------------------------
