:-module(classes, [
        active_term/2,
        add_antecedent/3,
	add_list_antecedent/3,
	add_list_succedent/3,
	add_succedent/3,
	and_cnf/3,
	antecedent/2,
        antecedent/3,
	as_dfg/3,
	as_new_axiom/2,
	as_new_goal/2,
        as_prolog_literals/2,
	as_prolog_sequent/2,
	as_tptp/2,
	as_spass/2,
        asSequent/2,
	blocked/1,
	bottom_cnf/1,
	change_rule/8,
	complexityBound/2,
	complexityBound/3,
	constraint/2,
	constraint/3,
	createClause/2,
	createSimpleClause/2,
	createSimpleClauseFromLiterals/2,
	createRule/4,
	createRuleWithSelection/5,
	delete_rule/1,
	delete_rule/2,
	delete_rule/3,
	delete_rule/4,
	delete_rules_no_dependents/3,
	dual_sequent/2,
	dont_demodulate_with/1,
	dont_simplify/1,
	empty/1,
	groundClause/1,
	hasSelectedLiteral/1,
	history/2,
	history/3,
	history_and_number/3,
	implicants/2,
	incr_weight_bound/2,
	indexEntries/2,
	indexEntries/3,
	is_atom/1,
	is_brand_new/1,
	is_equation/1,
	is_literal/1,
	is_new/1,
	is_object/2,
	is_old/1,
	is_trivial/1,
	kind/2,
	kind/3,
	literal_list_to_sequent/3,
	literals/2,
	literals/3,
	makeNew/2,
	makeOld/2,
	maximal_antecedent/2,
	maximal_antecedent/3,
	maximal_succedent/2,
	maximal_succedent/3,
	merge_two_rules_into_clause/3,
	moduleNumber/2,
	moduleNumber/3,
	monadicClause/1,
	neg_atom_as_literal/2,
	negate/2,
	non_maximal_antecedent/2,
	non_maximal_antecedent/3,
	non_maximal_succedent/2,
	non_maximal_succedent/3,
	not_cnf/2,
	number/2,
	number/3,
	operators/2,
	or_cnf/3,
	printSimpleClause/2,
	proof_depth/2,
	propositionalClause/1,
	pos_atom_as_literal/2,
	reductive/1,
	reductive_for/3,
	redundancy_info/2,
	redundancy_info/3,
	search_depth/2,
	search_maximal_candidates/6,
	selectSelectedLiteral/3,
	selectInAntecedent/3,
	selectInSuccedent/3,
	selectMaxInAntecedent/3,
	selectMaxInSuccedent/3,
	selectedLiteral/2,
	sequent_to_literals/3,
	simplifier/5,
	simplifier/6,
	simplify_cnf/2,
	size/2,
	size_and_weight/3,
        size_weight_vars/4,
	small_rule/1,
	subterm/2,
	subterms/2,
	succedent/2,
	succedent/3,
	terms/2,
	theory_flag/2,
	to_clause/2,
	vars/2,
	weight/2,
	weight/3,
	weightBound/2,
	weight_with_bound/2
  ]).


:- use_module(library(maplist),[maplist/2,maplist/3]).
:- use_module(library(sets), [
	select/3
   ]).
:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(flatten),[flatten/2]).

:- use_module(mylists, [
	is_list/1
   ]).

:- use_module(myterms, [
        apply/3,
	mapargs/3,
	list_of_vars/2
   ]).

:- use_module(library(samsort),[samsort/3]).


%:- use_module(user,_,[append/3]).


:- use_module(input, [
	analyse_clause/3,
	associativity_clause/4,
	commutativity_clause/4,
	permutativity_clause/5
    ]).

:- use_module(props, [
	equality_relation_symbol/1,
	congruence_relation_symbol/1,
	partial_equivalence_relation_symbol/1,
	partial_congruence_relation_symbol/1,
	reflexive_closure/2,
	reflexive_relation_symbol/1,
	total_ordering_relation_symbol/1,
	transitive_relation_symbol/1,
	special_relation/1,

	flag_on/2,
        newCOperator/1,
	newAOperator/1,
	newReflRelation/1,
	ac/1,
        ac_ist_C/1,
	horn/0,
	segment_length/1,
	scaled_clause_weight/7,
	term_weight/2
   ]).

:- use_module(equationalTheory, [
        equal/2,
	currentTheory/1
   ]).

:- use_module(thisProlog,[prolog_version/1]).


:- use_module(auxiliary, [ 
	delete_one312/3,
	list_as_conjunction/2,
	member212/2,
	appendAtoms/3
   ]).


:- use_module(adt_atom, [
	active_term_in_atom/2,
	atom_admits_variable_elimination/1,
	atom_as_rule/3,
	atom_as_term/2,
	atom_for_demodulation/5,
	create_atom/2,
	create_atom/4,
	get_atom/4,
	get_atom_members/3,
	get_atom_preorientation/2,
	get_atom_relation_symbol/2,
	get_atom_terms/2,
	get_atom_vars/2,
	get_list_atoms_vars/2,
	get_list_of_atoms_operators/2,
	greater_literal/4,
	is_false_atom/1,
	is_marked/1,
	is_trivial_atom/1,
	is_true_atom/1,
	is_proposition/1,
	mark_antecedent/2,
	orient_atom/2,
	print_list_atoms/1,
	selection/10
   ]).

:- use_module(adt_proof, [
	create_history/2,
	depth/2,
	get_key_history/2,
	get_first_history/2,
	get_second_history/2,
	initial_history/2,
	s_depth/2,
	theoryIndicator/2
   ]).

:- use_module(adt_precedence, [
	is_predicate/1
   ]).

:- use_module(proofState, [
	clause/1,
	rule/2,
	deleteRule/2,
	deleteClause/1,
	recordGoalOperators/1,
	recordRule/2,
	setRewriteFlag/1
   ]).

:- use_module(orderConstraints, [
	constraint_operators/2,
	print_constraint/1,
	project/3,
	true/1
   ]).


:- use_module(timing,[recordNewRule/0]).


:- use_module(proofState,[newClauseNumber/1]).
        



:- use_module(hgstuff, [
	setof0/3,
	split/4
   ]).




:- use_module(adt_term, [
	find_value_of_extra_vars/3,
	free_vars/2,
	get_term_vars/2,
	path_arg1/3,
	term_info/4,
	term_size/2
   ]).




:- use_module(index).



:- dynamic currentClass/1, classDefinition/4.




:- 
	user:no_style_check(discontiguous),
	ensure_loaded(classesCode),
	user:style_check(discontiguous).


