(:-true).
(:-use_module(mylists,[is_list/1])).
(:-use_module(props,[firstOrder/0,search_depth/1,scaled_clause_weight/5])).
(:-use_module(library(maplist),[scanlist/4])).
(:-use_module(dynamicCode,[recordPrologClause/2,known_proposition/1])).
(:-use_module(adt_atom,[is_equational_atom/1,orient_atom_fresh/2])).
(:-use_module(rewrite,[simplifier/4,injectivitySimplifiers/4])).
createRule(A, B, C, D) :-
        createRule(A, E),
        generateSimplifiers(E, B, C, D).
createRule(A, [object(B),simpleClause(C),clause(D),rule([E,F,0,0,[G,H],[I,J],_,K,_])]) :-
        antecedent(A, L),
        succedent(A, M),
        kind(A, N),
        maplist(atom_as_term, L, O),
        maplist(atom_as_term, M, P),
        analyse_rule(O, P, Q),
        select_kind(N, Q, R),
        kind(R, A, S),
        size_and_weight(S, E, F),
        maplist(orient_atom_fresh, L, T),
        maplist(orient_atom_fresh, M, U),
        search_maximal_candidates(T, U, G, H, I, J),
        sequent_to_literal_list([T,U], K),
        (   flag_on(ci, orde),
            \+flag_on(ci,equa) ->
            get_term_vars((L,M), V),
            constraint(S, W),
            project(V, W, X)
        ;   true(X)
        ),
        (   theory_flag(S, goal) ->
            operators(S, Y),
            recordGoalOperators(Y)
        ;   true
        ),
        constraint(X, S, Z),
        recordNewRule,
        number(Z, _),
        antecedent([], Z, A1),
        succedent([], A1, [object(B),simpleClause(C),clause(D)]), !.
generateSimplifiers(A, B, C, D) :-
        insertSimplifierIntoIndex(A, C, D, E),
        indexEntries(E, A, B), !.
createRuleWithSelection(poss, A, B, C, D) :-
        createRule(A, [object(E),simpleClause(F),clause(G),rule([H,I,0,0,[J,K],[L,M],N,O,P])|Q]),
        R=[object(E),simpleClause(F),clause(G),rule([H,I,0,1,[S,T],[U,V],N,O,P])|Q],
        kind(R, W),
        selection(W, J, K, L, M, S, T, U, V, 1),
        generateSimplifiers(R, B, C, D).
createRuleWithSelection(maxs, A, B, C, D) :-
        createRule(A, E),
        E=[object(F),simpleClause(G),clause(H),rule([I,J,0,0,[K,L],[M,N],O,P,Q])|R],
        S=[object(F),simpleClause(G),clause(H),rule([I,J,0,T,[U,V],[W,X],O,P,Q])|R],
        kind(S, Y),
        selection(Y, K, L, M, N, U, V, W, X, T),
        generateSimplifiers(S, B, C, D).
createRuleWithSelection(A, B, C, D, E) :-
        F=[object(G),simpleClause(H),clause(I),rule([J,K,0,1,[[],L],[[],M],_,N,[]])],
        number(A),
        antecedent(B, O),
        put_marked_eq_in_first_place(A, O, L),
        succedent(B, M),
        sequent_to_literal_list([L,M], N),
        size_and_weight(B, J, K),
        antecedent([], B, P),
        succedent([], P, [object(G),simpleClause(H),clause(I)]),
        generateSimplifiers(F, C, D, E), !.
delete_rules_no_dependents([], A, A).
delete_rules_no_dependents([A|B], C, D) :-
        number(A, E),
        deleteRule(E, A),
        indexEntries(A, F),
        delete_terms(F, C, G),
        delete_rules_no_dependents(B, G, D).
delete_rule(A) :-
        number(A, B),
        delete_rule(B, A).
delete_rule(A, B) :-
        deleteRule(A, B),
        delete_dependents(A).
delete_rule(A, B, C, D) :-
        deleteRule(A, B),
        indexEntries(B, E),
        delete_terms(E, C, D),
        delete_dependents(A).
delete_rule(A, B, C) :-
        number(A, D),
        delete_rule(D, A, B, C).
change_rule(A, B, C, D, E, F, G, H) :-
        deleteRule(A, _),
        append(C, D, I),
        append(E, F, J),
        antecedent(I, B, K),
        succedent(J, K, L),
        maximal_succedent(E, L, M),
        maximal_antecedent(C, M, N),
        literals(N, O),
        literals(O, N, P),
        size_and_weight(P, Q, R),
        size(Q, P, S),
        weight(R, S, T),
        indexEntries(B, U),
        delete_terms(U, G, V),
        indexEntries(_, T, W),
        generateSimplifiers(W, X, V, H),
        recordRule(A, X).
size([object(_),simpleClause(_),clause(_),rule([A,_,_,_,_,_,_,_,_])|_], A) :- !.
weight([object(_),simpleClause(_),clause(_),rule([_,A,_,_,_,_,_,_,_])|_], A).
weightBound([object(_),simpleClause(_),clause(_),rule([_,_,A,_,_,_,_,_,_])|_], A).
weight_with_bound(A, B) :-
        A=[object(_),simpleClause(_),clause(_),rule([_,C,_,_,_,_,_,_,_])|_],
        search_depth(A, D),
        theory_flag(A, E),
        (   orientableEquivalence(A) ->
            F=orientableEquivalence
        ;   kind(A, F)
        ),
        scaled_clause_weight(C, D, E, F, B).
antecedent([object(_),simpleClause(_),clause(_),rule([_,_,_,_,[A,B],_,_,_,_])|_], C) :- !,
        append(A, B, C).
maximal_antecedent([object(_),simpleClause(_),clause(_),rule([_,_,_,_,[A,_],_,_,_,_])|_], A).
non_maximal_antecedent([object(_),simpleClause(_),clause(_),rule([_,_,_,_,[_,A],_,_,_,_])|_], A).
succedent([object(_),simpleClause(_),clause(_),rule([_,_,_,_,_,[A,B],_,_,_])|_], C) :- !,
        append(A, B, C).
maximal_succedent([object(_),simpleClause(_),clause(_),rule([_,_,_,_,_,[A,_],_,_,_])|_], A).
non_maximal_succedent([object(_),simpleClause(_),clause(_),rule([_,_,_,_,_,[_,A],_,_,_])|_], A).
max_a_min_a_max_s_min_s([object(_),simpleClause(_),clause(_),rule([_,_,_,_,[A,B],[C,D],_,_,_])|_], A, B, C, D).
moduleNumber([object(_),simpleClause(_),clause(_),rule([_,_,_,_,_,_,A,_,_])|_], A).
selectedLiteral([object(_),simpleClause(_),clause(_),rule([_,_,_,1,[[],[A|_]],_,_,_,_])|_], A).
indexEntries(A, B) :-
        A=[object(_),simpleClause(_),clause(_),rule([_,_,_,_,_,_,_,_,B])|_],
        (   var(B) ->
            simplifierEntries(A, B),
            (   hasTermSimplifier(B) ->
                setRewriteFlag(term_equations_exist),
                setRewriteFlag(equations_exist)
            ;   true
            ),
            (   hasBoolSimplifier(B) ->
                setRewriteFlag(equations_exist),
                (   firstOrder ->
                    setRewriteFlag(term_equations_exist)
                ;   true
                ),
                setRewriteFlag(bool_equations_exist)
            ;   true
            ),
            printEntries(B)
        ;   true
        ), !.
selectInAntecedent(A, B, C) :-
        maximal_antecedent(A, D),
        select(B, D, E),
        maximal_antecedent(E, A, C).
selectInAntecedent(A, B, C) :-
        non_maximal_antecedent(A, D),
        select(B, D, E),
        non_maximal_antecedent(E, A, C).
selectMaxInAntecedent(A, B, C) :-
        maximal_antecedent(A, D),
        select(B, D, E),
        maximal_antecedent(E, A, C).
selectInSuccedent(A, B, C) :-
        maximal_succedent(A, D),
        select(B, D, E),
        maximal_succedent(E, A, C).
selectInSuccedent(A, B, C) :-
        non_maximal_succedent(A, D),
        select(B, D, E),
        non_maximal_succedent(E, A, C).
selectMaxInSuccedent(A, B, C) :-
        maximal_succedent(A, D),
        select(B, D, E),
        maximal_succedent(E, A, C).
active_term(A, B) :-
        maximal_succedent(A, C),
        member(D, C),
        active_term_in_atom(D, B).
active_term(A, B) :-
        maximal_antecedent(A, C),
        member(D, C),
        active_term_in_atom(D, B).
active_term(A, B) :-
        selectedLiteral(A, C),
        active_term_in_atom(C, B).
makeOld([object(A),simpleClause(B),clause(C),rule([D,E,_,F,G,H,I,J,K])|L], [object(A),simpleClause(B),clause(C),rule([D,E,old,F,G,H,I,J,K])|L]).
makeNew([object(A),simpleClause(B),clause(C),rule([D,E,_,F,G,H,I,J,K])|L], [object(A),simpleClause(B),clause(C),rule([D,E,1,F,G,H,I,J,K])|L]).
weight(A, [object(B),simpleClause(C),clause(D),rule([E,_,F,G,H,I,J,K,L])|M], [object(B),simpleClause(C),clause(D),rule([E,A,F,G,H,I,J,K,L])|M]).
size(A, [object(B),simpleClause(C),clause(D),rule([_,E,F,G,H,I,J,K,L])|M], [object(B),simpleClause(C),clause(D),rule([A,E,F,G,H,I,J,K,L])|M]).
incr_weight_bound([object(A),simpleClause(B),clause(C),rule([D,E,F,G,H,I,J,K,L])|M], [object(A),simpleClause(B),clause(C),rule([D,E,N,G,H,I,J,K,L])|M]) :-
        number(F),
        segment_length(O),
        N is F+O, !.
incr_weight_bound(A, B) :-
        makeNew(A, B).
maximal_antecedent(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,[_,I],J,K,L,M])|N], [object(B),simpleClause(C),clause(D),rule([E,F,G,H,[A,I],J,K,L,M])|N]).
non_maximal_antecedent(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,[I,_],J,K,L,M])|N], [object(B),simpleClause(C),clause(D),rule([E,F,G,H,[I,A],J,K,L,M])|N]).
maximal_succedent(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,[_,J],K,L,M])|N], [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,[A,J],K,L,M])|N]).
non_maximal_succedent(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,[J,_],K,L,M])|N], [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,[J,A],K,L,M])|N]).
moduleNumber(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,_,K,L])|M], [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,A,K,L])|M]).
literals(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,K,_,L])|M], [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,K,A,L])|M]).
indexEntries(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,K,L,_])|M], [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,K,L,A])|M]).
selectSelectedLiteral([object(A),simpleClause(B),clause(C),rule([D,E,F,1,[[],[G|H]],I,J,K,L])|M], G, [object(A),simpleClause(B),clause(C),rule([D,E,F,1,[[],H],I,J,K,L])|M]).
is_new(A) :-
        \+is_old(A).
is_brand_new([object(_),simpleClause(_),clause(_),rule([_,_,0,_,_,_,_,_,_])|_]).
is_old([object(_),simpleClause(_),clause(_),rule([_,_,old,_,_,_,_,_,_])|_]).
hasSelectedLiteral([object(_),simpleClause(_),clause(_),rule([_,_,_,1,_,_,_,_,_])|_]).
reductive(A) :-
        \+hasSelectedLiteral(A),
        (   maximal_succedent(A, []),
            maximal_antecedent(A, [B])
        ;   maximal_succedent(A, [B]),
            maximal_antecedent(A, [])
        ),
        get_atom_preorientation(B, C),
        (   C=l
        ;   C=r
        ), !.
small_rule(A) :-
        is_atom(A),
        size(A, B),
        small_rule_size(C),
        B=<C.
productivity_analysis(A) :-
        \+hasSelectedLiteral(A),
        \+flag_on(rp,prop),
        maximal_antecedent(A, []),
        maximal_succedent(A, [B]),
        non_maximal_succedent(A, C),
        maplist(is_proposition, [B|C]),
        as_Prolog_clause(A, D, E),
        recordPrologClause(D, E), !.
productivity_analysis(_).
blocked(_) :-
        flag_on(rp, prop), !,
        fail.
blocked(_) :-
        search_depth(A),
        A<10, !,
        fail.
blocked(A) :-
        antecedent(A, B),
        member(C, B),
        is_proposition(C),
        atom_as_term(C, D),
        known_proposition(D),
        \+dynamicCode:D, !.
blocked(A) :-
        non_maximal_succedent(A, B),
        member(C, B),
        atom_as_term(C, D),
        \+predicate_property(D,built_in),
        \+ \+ (numbervars(D,0,_),dynamicCode:D), !.
to_clause([object(A),simpleClause(B),clause(C),rule([D,E,F,G,H,I,J,K,L])|M], N) :-
        antecedent([object(A),simpleClause(B),clause(C),rule([D,E,F,G,H,I,J,K,L])|M], O),
        succedent([object(A),simpleClause(B),clause(C),rule([D,E,F,G,H,I,J,K,L])|M], P),
        history([object(A),simpleClause(B),clause(C),rule([D,E,F,G,H,I,J,K,L])|M], Q),
        constraint([object(A),simpleClause(B),clause(C),rule([D,E,F,G,H,I,J,K,L])|M], R),
        number([object(A),simpleClause(B),clause(C),rule([D,E,F,G,H,I,J,K,L])|M], S),
        createClause([O,P,R], T),
        history(Q, T, U),
        number(S, U, N), !.
as_Prolog_clause(A, (B:-C), [D|E]) :-
        A=[object(_),simpleClause(_),clause(_),rule([_,_,_,_,_,_,_,_,_])|_],
        maximal_succedent(A, [F]),
        atom_as_term(F, B),
        \+predicate_property(B,built_in),
        non_maximal_succedent(A, G),
        maximal_antecedent(A, []),
        non_maximal_antecedent(A, H),
        maplist(atom_as_term, G, I),
        maplist(prolog_negation, I, J),
        maplist(atom_as_term, H, K),
        append(K, J, L),
        append(I, K, E),
        functor(B, D, _),
        scanlist(prolog_and, L, true, C).
prolog_negation(A, \+A).
prolog_and(A, B, (A,B)).
simplifierEntries(A, B) :-
        kind(A, injectivity(_,_,_)), !,
        antecedent(A, [C]),
        succedent(A, [D]),
        number(A, E),
        injectivitySimplifiers(E, C, D, B),
        productivity_analysis(A).
simplifierEntries(A, B) :-
        (   dont_demodulate_with(A) ->
            B=[]
        ;   number(A, C),
            constraint(A, D),
            maximal_antecedent(A, E),
            non_maximal_antecedent(A, F),
            maximal_succedent(A, G),
            non_maximal_succedent(A, H),
            (   hasSelectedLiteral(A) ->
                I=hasSelectedLit
            ;   I=nothingSelected
            ),
            setof(J, simplifier(t(E,F,G,H,I),C,D,J), B),
            productivity_analysis(A)
        ).
hasTermSimplifier(A) :-
        member(t(_,B,_), A),
        \+B=[_,_,_],
        (   var(B)
        ;   functor(B, C, D),
            \+is_predicate(C/D)
        ), !.
hasBoolSimplifier(A) :-
        member(t(_,B,_), A),
        \+B=[_,_,_],
        nonvar(B),
        functor(B, C, D),
        is_predicate(C/D), !.
simplifier(A, B, =, C, D) :-
        rule(A, E),
        indexEntries(E, F),
        member([G,C,[_,A,_,D]], F),
        (   G=orientable_units_index ->
            B=l
        ;   B=n
        ).
simplifier(A, B, C, D, E, [[],[]]) :-
        simplifier(A, B, C, D, E).
simplifier(A, B, C, D, E, [F,G]) :-
        rule(A, H),
        H=[object(_),simpleClause(_),clause(_),rule([_,_,_,_,[[],F],[[I],G],_,_,_])|_],
        (   F\==[]
        ;   G\==[]
        ),
        atom_for_demodulation(I, B, C, D, E).
merge_two_rules_into_clause(A, B, C) :-
        antecedent(A, D),
        antecedent(B, E),
        (   hyper(B) ->
            mark_antecedent(E, F),
            append(F, D, G)
        ;   append(D, E, G)
        ),
        succedent(A, H),
        succedent(B, I),
        append(H, I, J),
        createClause([G,J], C).
hyper(_) :-
        horn.
hyper(A) :-
        hasSelectedLiteral(A),
        kind(A, B),
        hykind(B).
hykind(anti_symmetry(_,A)) :-
        congruence_relation_symbol(A).
hykind(composition(A,A,A)).
select_kind(regular, A, A) :- !.
select_kind(_, A, A) :-
        A\==regular, !.
select_kind(A, _, A).
:-mymultifile user:portray/1.
user:portray(A) :-
        is_object(A, [object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,K,L,M])|N]),
        [E,F,G,H,I,J,K,L,M]=[_,_,_,_,[O,P],[Q,R]|_],
        maplist(nonvar, [O,P,Q,R]),
        classes:printRule([object(B),simpleClause(C),clause(D),rule([E,F,G,H,I,J,K,L,M])|N]).
printRule(A) :-
        A=[object(_),simpleClause(_),clause(_),rule([_,_,_,B,_,_,_,_,_])|_],
        (   blocked(A) ->
            write(-)
        ;   B==1 ->
            write(*)
        ;   true
        ),
        to_clause(A, C),
        print(C).
analyse_rule(A, B, commutativity(C,D)) :-
        commutativity_clause(A, B, C, D),
        congruence_relation_symbol(D),
        (   ac(off) ->
            newCOperator(C)
        ;   true
        ).
analyse_rule(A, B, associativity(C,D)) :-
        associativity_clause(A, B, C, D),
        congruence_relation_symbol(D),
        (   ac(off) ->
            newAOperator(C)
        ;   true
        ).
analyse_rule(A, B, permutativity(C,D,E)) :-
        permutativity_clause(A, B, C, D, E).
analyse_rule(A, B, composition(C,C,C)) :-
        analyse_clause(A, B, composition(C,C,C)).
analyse_rule(A, B, reflexivity(C)) :-
        analyse_clause(A, B, reflexivity(C)),
        newReflRelation(C).
analyse_rule(A, B, finite_domain(C,D)) :-
        analyse_clause(A, B, finite_domain(C,D)),
        congruence_relation_symbol(D).
analyse_rule(_, _, regular).
small_rule_size(15).
reductive_for(A, B, C) :-
        \+hasSelectedLiteral(A),
        max_a_min_a_max_s_min_s(A, D, E, F, G),
        (   D=[H],
            F=[] ->
            B=[-|H],
            sequent_to_literal_list([E,G], C)
        ;   D=[],
            F=[H] ->
            B=[+|H],
            sequent_to_literal_list([E,G], C)
        ).
put_marked_eq_in_first_place(A, B, C) :-
        find_eq(A, B, D),
        delete_one312(D, B, E),
        search_greater_terms([D], F),
        append(F, E, C), !.
find_eq(1, [A|_], A) :- !.
find_eq(A, [_|B], C) :-
        D is A-1,
        find_eq(D, B, C), !.
search_greater_terms([], []) :- !.
search_greater_terms([A|B], [C|D]) :-
        orient_atom(A, C),
        search_greater_terms(B, D).
delete_dependents(A) :-
        dependent(A, B),
        deleteClause(B),
        fail.
delete_dependents(_).
dependent(A, B) :-
        clause(B),
        history(B, C),
        get_first_history(C, D),
        (   D==A ->
            true
        ;   get_second_history(C, E),
            E==A
        ).
sequent_to_literal_list([A,B], C) :-
        maplist(neg_atom_as_literal, A, D),
        maplist(pos_atom_as_literal, B, E),
        append(D, E, C), !.
printEntries(_).
(:-true).
(:-true).
createClause([A,B], [object(C),simpleClause(D),clause([_,E,_,regular,_])]) :-
        createSimpleClause([A,B], [object(C),simpleClause(D)]),
        true(E).
createClause([A,B,C], [object(D),simpleClause(E),clause([_,C,_,regular,_])]) :-
        createSimpleClause([A,B], [object(D),simpleClause(E)]).
history([object(_),simpleClause(_),clause([A,_,_,_,_])|_], A).
constraint([object(_),simpleClause(_),clause([_,A,_,_,_])|_], A).
redundancy_info([object(_),simpleClause(_),clause([_,_,A,_,_])|_], A).
kind([object(_),simpleClause(_),clause([_,_,_,A,_])|_], A).
complexityBound([object(_),simpleClause(_),clause([_,_,_,_,A])|_], A).
implicants(A, B) :-
        redundancy_info(A, (_,C)),
        (   var(C) ->
            B=[]
        ;   setof0(D, E^member((D,E),C), B)
        ).
proof_depth(A, B) :-
        A=[object(_),simpleClause(_),clause([_,_,_,_,_])|_],
        history(A, C),
        depth(C, B).
search_depth(A, B) :-
        A=[object(_),simpleClause(_),clause([_,_,_,_,_])|_],
        history(A, C),
        s_depth(C, B).
theory_flag(A, B) :-
        A=[object(_),simpleClause(_),clause([_,_,_,_,_])|_],
        history(A, C),
        theoryIndicator(C, B).
history(A, [object(B),simpleClause(C),clause([_,D,E,F,G])|H], [object(B),simpleClause(C),clause([A,D,E,F,G])|H]).
constraint(A, [object(B),simpleClause(C),clause([D,_,E,F,G])|H], [object(B),simpleClause(C),clause([D,A,E,F,G])|H]).
redundancy_info(A, [object(B),simpleClause(C),clause([D,E,_,F,G])|H], [object(B),simpleClause(C),clause([D,E,A,F,G])|H]).
kind(A, [object(B),simpleClause(C),clause([D,E,F,_,G])|H], [object(B),simpleClause(C),clause([D,E,F,A,G])|H]).
complexityBound(A, [object(B),simpleClause(C),clause([D,E,F,G,_])|H], [object(B),simpleClause(C),clause([D,E,F,G,I])|H]) :-
        asSimpleClause(A, I).
history_and_number(A, B, C) :-
        create_history(A, D),
        newClauseNumber(E),
        number(E, B, F),
        history(D, F, C), !.
history_and_number(_, _, _) :-
        raise_exception(saturate_error([126,110,101,114,114,111,114,32,105,110,32,104,105,115,116,111,114,121,95,97,110,100,95,110,117,109,98,101,114],[])).
as_new_axiom(A, B) :-
        initial_history(theory, C),
        newClauseNumber(D),
        number(D, A, E),
        history(C, E, F),
        B=F.
as_new_goal(A, B) :-
        initial_history(goal, C),
        newClauseNumber(D),
        number(D, A, E),
        history(C, E, F),
        B=F.
as_tptp(A, input_clause(from_saturate,B,C)) :-
        A=[object(_),simpleClause(_),clause([_,_,_,_,_])|_],
        antecedent(A, D),
        succedent(A, E),
        elim_equivalences(D, E, F, G),
        theory_flag(A, H),
        (   H=theory ->
            B=axiom
        ;   B=theorem
        ),
        maplist(as_tptp_literal(-), F, I),
        maplist(as_tptp_literal(+), G, J),
        append(I, J, C), !.
as_spass(A, '$spass'(B,C,D)) :-
        A=[object(_),simpleClause(_),clause([_,_,_,_,_])|_],
        antecedent(A, E),
        succedent(A, F),
        elim_special_relations(E, F, G, H),
        maplist(atom_as_term, G, I),
        maplist(atom_as_term, H, D),
        split(sortAtom, I, B, C), !.
as_dfg(A, B, '$dfg'(C,D)) :-
        theory_flag(B, D),
        antecedent(B, E),
        succedent(B, F),
        elim_special_relations(E, F, G, H),
        maplist(atom_as_term, G, I),
        maplist(deBruijn, J, I),
        maplist(as_dfg_formula(A), J, K),
        maplist(atom_as_term, H, L),
        maplist(deBruijn, M, L),
        maplist(as_dfg_formula(A), M, N),
        maplist(dfg_equal(A), K, O),
        maplist(dfg_equal(A), N, P),
        list_of_vars((P,O), Q),
        maplist(apply(not), O, R),
        append(R, P, S),
        (   S=[T] ->
            U=T
        ;   U=..[or|S]
        ),
        (   Q=[] ->
            C=clause(U)
        ;   C=clause(forall(Q,U))
        ).
as_prolog_sequent(A, (B->C)) :-
        antecedent(A, D),
        succedent(A, E),
        elim_equivalences(D, E, F, G),
        maplist(atom_as_term, F, H),
        maplist(atom_as_term, G, I),
        maplist(dfg_equal(dfg), H, B),
        maplist(dfg_equal(dfg), I, C).
as_prolog_literals(A, B) :-
        antecedent(A, C),
        succedent(A, D),
        elim_equivalences(C, D, E, F),
        maplist(atom_as_term, E, G),
        maplist(atom_as_term, F, H),
        maplist(mk_negative, G, I),
        append(I, H, B).
dual_sequent(A, [B,C]) :-
        antecedent(A, D),
        succedent(A, E),
        elim_equivalences(D, E, F, G),
        maplist(atom_as_term, F, C),
        maplist(atom_as_term, G, B).
mk_negative(A, -(A)).
dfg_equal(dfg, A=B, equal(A,B)) :- !.
dfg_equal(_, A, A).
(:-use_module(adt_term,[binder/5,external_binder/4])).
(:-use_module(adt_precedence,[special_symbol_conv/2])).
as_dfg_formula(_, A, A) :-
        var(A), !.
as_dfg_formula(dfg, A=B, equal(C,D)) :- !,
        as_dfg_formula(dfg, A, C),
        as_dfg_formula(dfg, B, D).
as_dfg_formula(dfg, A, B) :-
        binder(A, _, C, D, E), !,
        as_dfg_formula(dfg, D, F),
        external_binder(B, [C], F, E), !.
as_dfg_formula(A, B=>C, implies(D,E)) :- !,
        as_dfg_formula(A, B, D),
        as_dfg_formula(A, C, E).
as_dfg_formula(A, B<=>C, equiv(D,E)) :- !,
        as_dfg_formula(A, B, D),
        as_dfg_formula(A, C, E).
as_dfg_formula(A, B==C, equiv(D,E)) :- !,
        as_dfg_formula(A, B, D),
        as_dfg_formula(A, C, E).
as_dfg_formula(A, B and C, D and E) :- !,
        as_dfg_formula(A, B, D),
        as_dfg_formula(A, C, E).
as_dfg_formula(A, B or C, D or E) :- !,
        as_dfg_formula(A, B, D),
        as_dfg_formula(A, C, E).
as_dfg_formula(A, ~B, not(C)) :- !,
        as_dfg_formula(A, B, C).
as_dfg_formula(A, B, C) :-
        (   A=dfg,
            functor(B, D, E),
            special_symbol_conv(D/E, F/E) ->
            B=..[D|G],
            maplist(as_dfg_formula(A), G, H),
            C=..[F|H]
        ;   mapargs(as_dfg_formula(A), B, C)
        ).
sortAtom(A) :-
        A=..[_,B],
        var(B).
size_and_weight(A, 6, B) :-
        currentTheory(ac),
        kind(A, commutativity(_,_)),
        segment_length(C),
        B is 30*C, !.
size_and_weight(A, 10, B) :-
        currentTheory(ac),
        kind(A, associativity(_,_)),
        segment_length(C),
        B is 30*C, !.
dont_simplify(A) :-
        kind(A, B),
        dont_simplify_cl(B).
dont_demodulate_with(A) :-
        kind(A, B),
        dont_demo_with(B).
operators(A, B) :-
        A=[object(_),simpleClause(_),clause([_,_,_,_,_])|_],
        antecedent(A, C),
        succedent(A, D),
        constraint(A, E),
        get_list_of_atoms_operators(C, F),
        get_list_of_atoms_operators(D, G),
        constraint_operators(E, H),
        append(F, G, I),
        append(I, H, J),
        sort(J, B), !.
subterm(A, B) :-
        A=[object(_),simpleClause(_),clause([_,_,_,_,_])|_],
        (   antecedent(A, C)
        ;   succedent(A, C)
        ),
        member(D, C),
        get_atom_members(D, E, F),
        (   path_arg1(G, E, B)
        ;   path_arg1(G, F, B)
        ),
        (   nonvar(B) ->
            B\=='$true$',
            B\=='$min',
            functor(B, H, I),
            \+is_predicate(H/I)
        ;   true
        ).
subterms(A, B) :-
        setof0(C, subterm(A,C), B).
top_cnf([]).
bottom_cnf([[]]).
and_cnf(A, B, C) :-
        append(A, B, D),
        sort(D, C).
or_cnf([], _, []) :- !.
or_cnf(_, [], []) :- !.
or_cnf(A, [B|C], D) :-
        sort(B, E),
        or_conj_disj(A, E, F),
        or_cnf(A, C, G),
        append(F, G, H),
        sort(H, D).
or_conj_disj([], _, []).
or_conj_disj(A, [], A) :- !.
or_conj_disj([A|B], C, [D|E]) :-
        append(A, C, F),
        sort(F, D),
        or_conj_disj(B, C, E).
and_cnf_atom(A, B, [[B]|A]).
or_cnf_atom(A, B, C) :-
        or_cnf(A, [B], C).
and_cnf_literal(A, B, [[B]|A]).
or_cnf_literal(A, B, C) :-
        or_cnf(A, [B], C).
not_cnf([], [[]]).
not_cnf([[]], []).
not_cnf([A|B], C) :-
        not_disj(A, D),
        not_cnf(B, E),
        or_cnf(D, E, C).
not_disj(A, B) :-
        maplist(negateLitAsClause, A, B).
negateLitAsClause([+|A], [[-|A]]).
negateLitAsClause([-|A], [[+|A]]).
simplify_cnf([], []).
simplify_cnf([A|B], C) :-
        taut(A), !,
        simplify_cnf(B, C).
simplify_cnf([A|B], C) :-
        member(D, B),
        subs(D, A), !,
        simplify_cnf(B, C).
simplify_cnf([A|B], C) :-
        select(D, B, E),
        subs(A, D), !,
        simplify_cnf([A|E], C).
simplify_cnf([A|B], [A|C]) :-
        simplify_cnf(B, C).
taut(A) :-
        select([B|C], A, D),
        member([E|F], D),
        C==F,
        dual(B, E).
dual(-, +).
dual(+, -).
subs([], _).
subs([A|B], C) :-
        select(D, C, _),
        equal(A, D),
        subs(B, C).
user:portray(A) :-
        is_object(A, [object(B),simpleClause(C),clause([D,E,F,G,H])|I]),
        nonvar(D),
        classes:printClause([object(B),simpleClause(C),clause([D,E,F,G,H])|I]).
as_tptp_literal(-, A, --(B)) :-
        atom_as_term(A, C),
        (   C=equal(D,E),
            \+congruence_relation_symbol(equal) ->
            B=eq(D,E)
        ;   B=C
        ).
as_tptp_literal(+, A, ++(B)) :-
        atom_as_term(A, C),
        (   C=equal(D,E),
            \+congruence_relation_symbol(equal) ->
            B=eq(D,E)
        ;   B=C
        ).
printClause(A) :-
        A=[object(_),simpleClause(_),clause([_,B,_,_,_])|_],
        \+ \+printSimpleClause(A,B), !.
(:-use_module(props,[special_relation/1])).
dont_simplify_cl(reflexivity(_)).
dont_simplify_cl(irreflexivity(_)).
dont_simplify_cl(symmetry(_)).
dont_simplify_cl(composition(A,B,C)) :-
        transitive_relation_symbol(A),
        special_relation(A),
        transitive_relation_symbol(B),
        special_relation(B),
        transitive_relation_symbol(C),
        special_relation(C).
dont_simplify_cl(monotonicity(_,_,A,_,A)) :-
        partial_congruence_relation_symbol(A).
dont_simplify_cl(totality(_,_)).
dont_simplify_cl(density(_,_,_)).
dont_simplify_cl(reflexive_closure1(A,_)) :-
        total_ordering_relation_symbol(A).
dont_simplify_cl(reflexive_closure2(A,_)) :-
        total_ordering_relation_symbol(A).
dont_simplify_cl(noendL(_)).
dont_simplify_cl(noendR(_)).
dont_simplify_cl(commutativity(_,_)).
dont_simplify_cl(associativity(_,_)).
dont_simplify_cl(permutativity(_,_,_)) :-
        \+currentTheory(ac).
dont_simplify_cl(finite_domain(_,A)) :-
        congruence_relation_symbol(A).
dont_demo_with(composition(_,_,_)).
dont_demo_with(totality(_,_)).
dont_demo_with(density(_,_,_)).
dont_demo_with(reflexive_closure1(A,_)) :-
        total_ordering_relation_symbol(A).
dont_demo_with(reflexive_closure2(A,_)) :-
        total_ordering_relation_symbol(A).
dont_demo_with(commutativity(_,_)) :-
        currentTheory(ac).
dont_demo_with(associativity(_,_)) :-
        currentTheory(ac).
(:-true).
(:-true).
createSimpleClause([A,B], [object(_),simpleClause([_,A,B])]).
createSimpleClauseFromLiterals(A, B) :-
        literal_list_to_sequent(A, C, D),
        elim_equivalences(C, D, E, F),
        createSimpleClause([E,F], B).
antecedent([object(_),simpleClause([_,A,_])|_], A).
succedent([object(_),simpleClause([_,_,A])|_], A).
number([object(_),simpleClause([A,_,_])|_], A).
literals(A, B) :-
        A=[object(_),simpleClause([_,_,_])|_],
        antecedent(A, C),
        succedent(A, D),
        sequent_to_literals(C, D, B), !.
antecedent(A, [object(B),simpleClause([C,_,D])|E], [object(B),simpleClause([C,A,D])|E]).
succedent(A, [object(B),simpleClause([C,D,_])|E], [object(B),simpleClause([C,D,A])|E]).
number(A, [object(B),simpleClause([_,C,D])|E], [object(B),simpleClause([A,C,D])|E]).
literals(A, [object(B),simpleClause([C,_,_])|D], [object(B),simpleClause([C,E,F])|D]) :-
        literal_list_to_sequent(A, E, F).
asSimpleClause(A, B) :-
        antecedent(A, C),
        succedent(A, D),
        createSimpleClause([C,D], B).
asSequent(A, [B,C]) :-
        antecedent(A, D),
        succedent(A, E),
        maplist(atom_as_term, D, B),
        maplist(atom_as_term, E, C).
to_clause(A, A) :-
        A=[object(_),simpleClause([_,_,_])|_].
is_atom(A) :-
        antecedent(A, []),
        succedent(A, [_]).
is_equation(A) :-
        antecedent(A, []),
        succedent(A, [B]),
        get_atom_relation_symbol(B, C),
        partial_equivalence_relation_symbol(C),
        (   C= == ->
            get_atom_members(B, D, C),
            C\=='$true$',
            D\=='$true$'
        ;   true
        ).
is_literal(A) :-
        antecedent(A, [_]),
        succedent(A, []), !.
is_literal(A) :-
        is_atom(A).
is_trivial(A) :-
        succedent(A, B),
        (   succedentTrivial(B)
        ;   antecedent(A, C),
            (   antecedentTrivial(C)
            ;   member(D, C),
                member212(D, B)
            )
        ), !.
empty(A) :-
        antecedent(A, []),
        succedent(A, []).
orientableEquivalence(A) :-
        antecedent(A, []),
        succedent(A, [B]),
        adt_atom:equivalence(B).
vars(A, B) :-
        succedent(A, C),
        antecedent(A, D),
        append(D, C, E),
        get_list_atoms_vars(E, B).
terms(A, B) :-
        antecedent(A, C),
        succedent(A, D),
        maplist(get_atom_terms, C, E),
        maplist(get_atom_terms, D, F),
        append(E, F, G),
        flatten(G, B).
operators(A, B) :-
        succedent(A, C),
        antecedent(A, D),
        get_list_of_atoms_operators(D, E),
        get_list_of_atoms_operators(C, F),
        append(E, F, G),
        sort(G, B), !.
groundClause(A) :-
        antecedent(A, B),
        free_vars(B, []),
        succedent(A, C),
        free_vars(C, []).
monadicClause(A) :-
        antecedent(A, B),
        succedent(A, C),
        free_vars((B,C), D),
        (   D=[]
        ;   D=[_]
        ), !.
propositionalClause(A) :-
        succedent(A, B),
        antecedent(A, C),
        append(C, B, D),
        maplist(is_proposition, D).
size(A, B) :-
        succedent(A, C),
        antecedent(A, D),
        term_size(D, E),
        term_size(C, F),
        B is E+F, !.
size_weight_vars(A, B, C, D) :-
        succedent(A, E),
        antecedent(A, F),
        term_info([E,F], B, G, D),
        length(D, H),
        term_weight((B,G,H), C).
size_and_weight(A, B, C) :-
        size_weight_vars(A, B, C, _).
add_antecedent(A, B, C) :-
        antecedent(B, D),
        antecedent([A|D], B, C).
add_succedent(A, B, C) :-
        succedent(B, D),
        succedent([A|D], B, C).
add_list_antecedent(A, B, C) :-
        antecedent(B, D),
        append(A, D, E),
        antecedent(E, B, C).
add_list_succedent(A, B, C) :-
        succedent(B, D),
        append(A, D, E),
        succedent(E, B, C).
selectInAntecedent(A, B, C) :-
        antecedent(A, D),
        select(B, D, E),
        antecedent(E, A, C).
selectInSuccedent(A, B, C) :-
        succedent(A, D),
        select(B, D, E),
        succedent(E, A, C).
elim_equivalences([], [], [], []).
elim_equivalences(A, [B|C], [D|E], [F|G]) :-
        get_atom(B, ==, H, I),
        I\=='$true$', !,
        create_atom(H=='$true$', D),
        create_atom(I=='$true$', F),
        elim_equivalences(A, C, E, G).
elim_equivalences(A, B, C, D) :-
        elim_special_relations(A, B, C, D).
elim_special_relations([], [], [], []).
elim_special_relations(A, [B|C], D, [E|F]) :-
        get_atom(B, G, H, I),
        G\== ==,
        partial_equivalence_relation_symbol(G),
        reflexive_relation_symbol(G),
        \+congruence_relation_symbol(G),
        congruence_relation_symbol(J), !,
        K=..[G,H],
        L=..[G,I],
        create_atom(K, J, L, E),
        elim_special_relations(A, C, D, F).
elim_special_relations([A|B], C, [D|E], F) :-
        get_atom(A, G, H, I),
        G\== ==,
        partial_equivalence_relation_symbol(G),
        reflexive_relation_symbol(G),
        \+congruence_relation_symbol(G),
        congruence_relation_symbol(J), !,
        K=..[G,H],
        L=..[G,I],
        create_atom(K, J, L, D),
        elim_special_relations(B, C, E, F).
elim_special_relations(A, [B|C], D, [B|E]) :-
        elim_special_relations(A, C, D, E).
elim_special_relations([A|B], C, [A|D], E) :-
        elim_special_relations(B, C, D, E).
indexEntries(A, B) :-
        A=[object(_),simpleClause([_,C,D])|_],
        E=9999999,
        true(F),
        setof(G, simplifier(t([],C,D,[],nothingSelected),E,F,G), B).
search_maximal_candidates(A, B, C, D, E, F) :-
        ant([], A, B, G, H, I, J),
        suc(G, [], I, C, K, E, L),
        append(H, K, D),
        append(J, L, F), !.
ant(A, [B|C], D, E, F, G, H) :-
        orient_atom(B, I),
        smallerones(I, A, J, K),
        smallerones(I, C, L, M),
        smallerones_ant_suc(I, D, N, O),
        append(J, [I], P),
        append(K, M, Q),
        ant(P, L, N, E, R, G, S),
        append(Q, R, F),
        append(O, S, H), !.
ant(A, [], B, A, [], B, []).
suc(A, B, [C|D], E, F, G, H) :-
        orient_atom(C, I),
        smallerones(I, B, J, K),
        smallerones(I, D, L, M),
        smallerones_suc_ant(I, A, N, O),
        append(J, [I], P),
        append(K, M, Q),
        suc(N, P, L, E, R, G, S),
        append(O, R, F),
        append(Q, S, H), !.
suc(A, B, [], A, [], B, []).
smallerones(A, [B|C], D, [B|E]) :-
        greater_literal(+, A, +, B), !,
        smallerones(A, C, D, E).
smallerones(A, [B|C], [B|D], E) :-
        smallerones(A, C, D, E), !.
smallerones(_, [], [], []).
smallerones_ant_suc(A, [B|C], D, [B|E]) :-
        greater_literal(-, A, +, B), !,
        smallerones_ant_suc(A, C, D, E).
smallerones_ant_suc(A, [B|C], [B|D], E) :-
        smallerones_ant_suc(A, C, D, E).
smallerones_ant_suc(_, [], [], []).
smallerones_suc_ant(A, [B|C], D, [B|E]) :-
        greater_literal(+, A, -, B), !,
        smallerones_suc_ant(A, C, D, E).
smallerones_suc_ant(A, [B|C], [B|D], E) :-
        smallerones_suc_ant(A, C, D, E).
smallerones_suc_ant(_, [], [], []).
user:portray(A) :-
        is_object(A, [object(B),simpleClause([C,D,E])|F]),
        nonvar(D), !,
        true(G),
        \+ \+classes:printSimpleClause([object(B),simpleClause([C,D,E])|F],G).
user:portray(proofClause(A)) :-
        \+ \+classes:printProofClause(A).
(:-use_module(adt_term,[deBruijn/2])).
printSimpleClause(A, B) :-
        A=[object(_),simpleClause([C,_,_])|_],
        antecedent(A, D),
        succedent(A, E),
        maplist(deBruijn, F, D),
        maplist(deBruijn, G, E),
        antecedent(F, A, H),
        succedent(G, H, I),
        numbervars((I,B), 0, _),
        proof_depth(I, J),
        theory_flag(I, K),
        (   K=theory ->
            L= :
        ;   L=(+)
        ),
        (   number(C) ->
            format('~t~d(~d) ~p~10| ', [C,J,L])
        ;   true
        ),
        split(is_true_atom, F, _, M),
        split(is_false_atom, G, _, N),
        print_list_atoms(M),
        (   M\==[] ->
            write(' -> ')
        ;   N=[] ->
            write(false)
        ;   true
        ),
        print_list_atoms(N),
        print_constraint(B).
printProofClause(A) :-
        number(A, B),
        antecedent(A, C),
        succedent(A, D),
        maplist(deBruijn, E, C),
        maplist(deBruijn, F, D),
        antecedent(E, A, G),
        succedent(F, G, H),
        numbervars(H, -1, _),
        proof_depth(H, _),
        theory_flag(H, I),
        (   I=theory ->
            J= :
        ;   J=(+)
        ),
        (   number(B) ->
            format('~n~t~d ~p~6| ', [B,J])
        ;   true
        ),
        split(is_true_atom, E, _, K),
        split(is_false_atom, F, _, L),
        print_list_atoms(K),
        (   K\==[] ->
            write(' -> ')
        ;   L=[] ->
            write(false)
        ;   true
        ),
        print_list_atoms(L),
        (   flag_on(ci, orde) ->
            constraint(H, M),
            print_constraint(M)
        ;   true
        ).
succedentTrivial(A) :-
        member(B, A),
        is_trivial_atom(B), !.
antecedentTrivial(A) :-
        member(B, A),
        is_false_atom(B), !.
literal_list_to_sequent([], [], []).
literal_list_to_sequent([[-|A]|B], [A|C], D) :-
        literal_list_to_sequent(B, C, D).
literal_list_to_sequent([[+|A]|B], C, [A|D]) :-
        literal_list_to_sequent(B, C, D).
sequent_to_literals(A, B, C) :-
        maplist(neg_atom_as_literal, A, D),
        maplist(pos_atom_as_literal, B, E),
        append(D, E, C), !.
neg_atom_as_literal(A, [-|A]).
pos_atom_as_literal(A, [+|A]).
negate([+|A], [-|A]).
negate([-|A], [+|A]).
(:-true).
(:-true).
is_object(A, B) :-
        subsumes_chk([object(_)|_], A),
        A=B.
(:-true).
