%declarations%

:- module(clausal_rewriting, [ 
	redundant_by_clausal_rewriting/4
   ]).

:- use_module(constraintSolver, [
	simplify_med/2,
	simplified_satisfiable_conjunct/2
   ]).

:- use_module(orderConstraints, [
	and_gt/3,
	false/1,
	implies/2,
	nonstrict_inequalities_as_disjunction/2,
	true/1
   ]).


:- use_module(global, [
	store/2,
	cont/2
   ]).

:- use_module(rewrite, [
        subsumes_literals/2
   ]).

:- use_module(props, [
	flag_on/2,
	case_split/3,
	crw_depth/1
   ]).

:- use_module(adt_atom, [
	create_atom/2,
	subsumes_literals/4,
	match_atoms/2,
	match_atoms/4,
	neg_atom_as_literal/2,
	symmetric_version_of_atom/2,
	get_atom_members/3,
	get_atom_preorientation/2,
	get_atom_relation_symbol/2
   ]).

:- use_module(classes, [
	add_antecedent/3,
	add_succedent/3,
	literals/2,
	literals/3,
	constraint/2,
	createSimpleClause/2,
	vars/2,
	terms/2,
	selectInAntecedent/3,
	selectInSuccedent/3,
	constraint/3,
	number/3,
	selectMaxInAntecedent/3,
	selectMaxInSuccedent/3,
	antecedent/2,
	maximal_antecedent/3,
	non_maximal_antecedent/3,
	number/2,
	kind/2,
	literals/2,
	subterms/2,
	succedent/2,
	to_clause/2
   ]).

:- use_module(adt_declarations, [
	message/2
   ]).


:- use_module(adt_term, [
	extraVars/2,
	find_value_of_extra_vars/3,
	free_vars/2,
	instantiate/1
   ]).

:- use_module(myterms, [
        skolemizedR/3,
        skolemized/3,
	list_of_vars/2,
	apply_substitution/3
   ]).


:- use_module(auxiliary, [
	delete32/3,
	member212/2
   ]).

:- use_module(library(basics), [member/2]).

:- use_module(library(maplist), [maplist/3]).

:- use_module(mylists,[nth1/3]).

:- use_module(clause_redundancy_prover, [
	is_tautology/3
   ]).

:- use_module(inference_rules, [
        constraint_greater_literal/3,			       
        constraint_grequal_literal_in/4,			       
	constraint_grequal_clause/4
   ]).

:- use_module(proofState, [
	rule/2
   ]).

:- use_module(thisProlog, [
	timeOut/3
   ]).


:- use_module(library(sets), [
	select/3
   ]).

%:- use_module(user,_,[append/3]).

:- use_module(commands, [l/0]).

%end%


/*

CtNonRed describes the ground insatnces of Conc
that were not found to be redundant.

*/



redundant_by_clausal_rewriting( Conc, Constraint, CBound, CtNonRed):-
	flag_on( rp, ccrw ),
	!,
	subterms( Conc, LV ),
%	terms( Conc, LV ),
	number(Conc,N),
	number( _, Conc, C ),
	crw_depth(D),

	(flag_on( ms, ccrw ) ->
	     \+ \+ (numbervars((LV,Conc,CBound),1,_),
		    format('~2nClausal rewriting for~n ~p~n wrt. bound ~p~n...',[Conc,CBound]),
		    ttyflush
		   )
	;
	     true
	),
	
	store(betterMatching,false),
%	(N=23 -> trace; true),
	case_split(C,Cases),
	orderConstraints:and1(Constraint,Cases,Ct),
	nonstrict_inequalities_as_disjunction(Ct,Ct1),
	(flag_on( ms, ccrw ) ->
	     constraint(Ct1,Conc,ConcCt1),
	     \+ \+ (numbervars((LV,ConcCt1,CBound),1,_),
		    format('~2nClausal rewriting for~n ~p~n wrt. bound ~p~n...',[ConcCt1,CBound]),
		    ttyflush
		   )
	;
	     true
	),
	whichDisjunctsRedundantByClr(N,Ct1,LV,CBound,C,D,CtNonRed),
	!.
redundant_by_clausal_rewriting( _Conc, Constraint, _CBound, Constraint).

case_split(C,Ct):-
	case_split(AT,ST,CtS),
        maplist(create_atom,AT,A),
        maplist(create_atom,ST,S),
        createSimpleClause([A,S],Cl1),
        literals(Cl1,Ls1),
        skolemized(C,CS,Subst),
        literals(CS,Ls),
        subsumes_literals(Ls1,Ls),
        skolemized(Ct,CtS,Subst),
        !.
case_split(_,Top):-
	true(Top).



whichDisjunctsRedundantByClr(_,[],_,_CBound,_C,_,[]).
whichDisjunctsRedundantByClr(N,[Conj|Ct],LV,CBound,C,D,[Co|CtNonRed]):-
	simplified_satisfiable_conjunct(Conj,Co),
	\+ disjunctRedundantByClr(N,Co,LV,CBound,C,D),
	!,
	writewnl(' Instance cannot be eliminated.'),
	whichDisjunctsRedundantByClr(N,Ct,LV,CBound,C,D,CtNonRed).
whichDisjunctsRedundantByClr(N,[_|Ct],LV,CBound,C,D,CtNonRed):-
	whichDisjunctsRedundantByClr(N,Ct,LV,CBound,C,D,CtNonRed).

	
disjunctRedundantByClr(N,conj(I,E,O),LV,CBound,C,D):-
	\+ \+ ( numbervars((LV,conj(I,E,O),C,CBound),1,_),
		constraint([conj(I,E,O)],C,CCase),
		  ( flag_on( ms, ccrw ) ->
		      format('~2nWorking on case~n ~p at depth ~p~n wrt. bound ~p~n...',[CCase,D,CBound]),
		      ttyflush
		  ;   
		      true
		  )
	      ),
	approximatedCBound(CBound,ACB),
	apply_substitution(E,(ACB,LV,C),(CBound1,LV1,C1)),
	free_vars(LV1,VV),
	union(VV,LV1,V),
	constraint([conj(I,[],O)],C1,C2),
	retractall(cl_rew(_,_)),
	skolemizedR(C2,_,Subst),
	redundant_by_cl_rewrT(N,Subst,CBound1,V,C2,D,Proof/[],10000000),
%	redundant_by_cl_rewrT(N,Subst,CBound1,LV1,C2,D,Proof/[],10000000),
	(flag_on( ms, ccrw ) ->
	     format('~nProof of redundancy for instance~n ~p:',[C2]),
	     printProof(Proof)
	;
	     true
	),
	!.

:-dynamic interactive/1,cl_rew/2.



redundant_by_cl_rewrT(N,Subst,CBound,LV,C,D,Proof,T):-
	(interactive(N) ->
	     store(interactive,true)
	;
	     store(interactive,false)
	),
	(timeOut(redundant_by_cl_rewr(Subst,CBound,LV,C,D,Proof),T,S) ->
	     (S==success ->
		  true
	     ;    \+ \+ (  interactive(_),
			   numbervars(C,1,_),
                           format('~nCan''t eliminate~n ~p~n by clausal rewriting: time out after ~p msec.~n Shall I try again? Enter time [msec], or 0 for not trying again: ',[C,T])
			),
	          read(T1),
		  (number(T1),T1>0 ->
		       redundant_by_cl_rewrT(N,Subst,CBound,LV,C,D,Proof,T1)
		  )
	     )
	;
	     \+ \+ (  interactive(_),
		      numbervars(C,1,_),
		      format('~nCan''t eliminate~n ~p~n by clausal rewriting: current proof depth = ~p.~n Shall I try again? Enter new depth, or ~n  -1 for going into interactive mode~n  0 for not trying again: ',[C,D])
		   ),
	     read(D1),
	     integer(D1),
	     (D1>0 ->
		  redundant_by_cl_rewrT(N,Subst,CBound,LV,C,D1,Proof,T)
	     ;D1<0 ->
		  store(interactive,true),
		  (redundant_by_cl_rewrT(N,Subst,CBound,LV,C,D1,Proof,T) ->
		       store(interactive,false)
		  ;
		       store(interactive,false),
		       fail
		  )
		  
	     )
	).





approximatedCBound(CBound,ACB):-
	% to avoid large constraints
	literals(CBound,Ls),
	member(ACB,Ls).



writewnl(X):-
	(flag_on(ms,ccrw) ->
	     write(X),
	     nl
	;    true
	).

writew(X):-
	(flag_on(ms,ccrw) ->
	     write(X),
	     ttyflush
	;    true
	).


iterativeDeepening(CurrentDepth,D,CurrentDepth,Pred):-
	CurrentDepth=<D,
	Pred,
	!.
iterativeDeepening(CurrentDepth,D,DPar,Pred):-
	CurrentDepth<D,
	D1 is max(CurrentDepth+2,D),
	iterativeDeepening(D1,D,DPar,Pred),
	!.



printProof([]).
printProof([(Step,Info)|P]):-
	format('~n',[]),
	mess_cr( Step, Info ),
	printProof(P).

tauto(C):-
	antecedent(C,A),
	succedent(C,S),
	member(E,A),
	symmetric_version_of_atom(E,E1),
	member(E2,S),
	E1==E2.






/*
redundant_by_cl_rewr(Subst,Max,Subterms,Clause,Depth,Proof):-
	redundant_by_cl_rewr1(Subst,Max,Subterms,Clause,Depth,Proof).
*/

proofok([(taut,_)|P]/[]):-
	proofok(P/[]).
proofok([(subs, [_LV,C,R,Max])|P]/[]):-
	constraint( C, Ct ),
	constraint_implies_greater_rule( Max, R, Ct ),
	!,
	proofok(P/[]).
proofok([(rewr,[_LV,C,R,Max,_S])|P]/[]):-
	constraint( C, Ct ),
	constraint_implies_greater_rule( Max, R, Ct ),
	!,
	proofok(P/[]).

proofok([]/[]).




redundant_by_cl_rewr(Subst,Max,Subterms,[[C,D]],Proof):-
	% the common case
	!,
	redundant_by_cl_rewr(Subst,Max,Subterms,C,D,Proof).
redundant_by_cl_rewr(Subst,_Max,_Subterms,Clauses,_Proof):-
	member([Clause,_Depth],Clauses),
	copy_term((Clause,Subst),(CS,CSubst)),
	instantiate(CSubst),
	numbervars(CS,1,_),
	hash_term(CS,Key),
	(var(Key) ->
	     raise_exception(saturate_error("~nredundant_by_cl_rewr: Key should not be a variable.~n",[])),
	     trace
	     
	;
	     true
	),
	cl_rew(Key,(CS,no,_)),
	writew('-'),
	!,
	fail.
redundant_by_cl_rewr(Subst,Max,Subterms,Clauses,Proof):-
	redundant_by_cl_rewr1(Subst,Max,Subterms,Clauses,Proof).



redundant_by_cl_rewr1(Subst,Max, LV, [[C,D]|R], P1/P2):-
	redundant_by_cl_rewr(Subst,Max, LV, C, D, P1/P3),
	redundant_by_cl_rewr1(Subst,Max, LV, R, P3/P2),
	!.
redundant_by_cl_rewr1( _, _, _, [], Proof/Proof).

redundant_by_cl_rewr(Subst,Max,Subterms,Clause,Depth,Proof):-
	copy_term((Clause,Subst),(CS,CSubst)),
	instantiate(CSubst),
	hash_term(CS,Key),
	(var(Key) ->
	     raise_exception(saturate_error("~Nredundant_by_cl_rewr/6: Key should not be a variable.~n",[]))
	;
	     true
	),
	(cl_rew(Key,(CS,Answer,Proof)) ->
	     (Answer=yes ->
	          writew('+')
	     ;
	          writew('-'),
	          fail
	     )
	;writew('.'),redundant_by_cl_rewr1(Subst,Max,Subterms,Clause,Depth,Proof) ->
	     assert(cl_rew(Key,(CS,yes,Proof)))
	;
	     assert(cl_rew(Key,(CS,no,_))),
	     fail
	).



redundant_by_cl_rewr1( _, _Max, LV, C, _, [(taut,[LV,C])|P]/P):-
	tauto(C),
	mess_cr1( taut, [LV,C] ).
redundant_by_cl_rewr1( _, _Max, LV, C, _, [(taut,[LV,C])|P]/P):-
	is_tautology( C, _, yes ), 
	mess_cr1( taut, [LV,C] ).
redundant_by_cl_rewr1( _, Max, LV, C, _, [P|P1]/P1 ):-
	is_subsumed_by_rule_smaller_than_max( LV, Max, C,P ).
redundant_by_cl_rewr1( Subst,Max, LV, C, D, [P|P1]/P2 ):-
	cont(interactive,true),
%	D >= 0, D1 is D - 1,
	D=D1,
	askForRule(LV,C,N),
	rule(N,Ru),
	P=(rewr,[LV,C,Ru,Max,Comps]),
	one_step_of_clausal_rewriting(Max, LV, C, D1,Comps, Checked,Ru ), 
	\+ \+ askForOk(LV,P),
	P=(rewr,[LV,C,Ru,Max,Comps]),
	redundant_by_cl_rewr( Subst,Max, LV, Comps, P1/P2 ),
	(Checked=notChecked ->
	     constraint(C,Ct),
	     constraint_implies_greater_rule( Max, Ru, Ct )
	;    true),
	!.
redundant_by_cl_rewr1( Subst,Max, LV, C, D, [P|P1]/P2 ):-
	D >= 0, D1 is D - 1,
	one_step_of_clausal_rewriting(Max, LV, C, D1, S, Checked,Ru ), 
	P=(rewr,[LV,C,Ru,Max,S]),
	redundant_by_cl_rewr( Subst,Max, LV, S, P1/P2 ),
	(Checked=notChecked ->
	     constraint(C,Ct),
	     constraint_implies_greater_rule( Max, Ru, Ct )
	;    true),
	!.
% extremely prolific or buggy
%redundant_by_cl_rewr1( Subst, Max, LV, C, D, P/P ):-
%	flag_on( rp, clca ),
%	D >= 0, D1 is D - 1,
%	\+ ( \+ all_var_cases_redundant_by_cl_rewr( Subst,Max, LV, LV, C, D1 ) ).
redundant_by_cl_rewr1( _, _Max, LV, C, _, _ ):-
	mess_cr1(rid,[LV,C]),
	fail.



askForRule(LV,C,N):-
	prompt((LV,C),'Working on clause~n~p.~nTry which rule (l for listing)? ',[C],N1),
	(N1=l ->
	     l,
	     askForRule(LV,C,N)
	;
	     N=N1
	;
	     askForRule(LV,C,N)
	).



askForOk(V,P):-
	numbervars((V,P),1,_),
	P=(L,R),
	mess_cr(L,R),
	format('~nOK (y/n)? ',[]),
	read(T),
	T=y.


% positive predicate on positive predicate
one_step_of_clausal_rewriting(Max,LV,C,D,Comps,Checked,R):-
	applyRule(LV,C,R,A,S,Max,Checked),
	\+ \+ showInstance(LV,C,R,Max),
	complementary_clauses( C, D, A, S, Comps ).


	
showInstance(LV,C,R,Max):-
	(cont(interactive,true) ->
	     numbervars((LV,C,R,Max),1,_),
	     format('~nApplying rule~n ~p~n to ~p~n with bound ~p,',[R,C,Max])
	;    true).




applyRule(V,C,R,[],SR,Max,checked):-
	% hyperresolution version; we check complexity together with matching
	% doing this before trying the top-down direction speeds some
	% examples up by a factor of 10!  (emb2 from 28 Ksec to 2.6 Ksec )
	(cont(interactive,true) ->
	     prompt(_,'Apply rule bottom up (y/n)',[],y)
	;
	     true
	),
	constraint(C,Ct),
	antecedent(C,A),
	bottomUpRuleForCRW(R,AR),
	matchAtomList(AR,A,_Matched,_NotMatched), % don't need to check here
	succedent(R,SR),
	vars(R,VR),
	(extraVars(VR,V) ->
	     instantiate_extra_vars( V, VR, R )
	;  
	     true
	),
	makingProgress(SR,A),
	maplist(neg_atom_as_literal,SR,SLR),
            % we treat the succedent in R negatively
	    % so as to assure that it is smaller than Max even
	    % when the sign is changed upon construction of
            % a complementary clause.
	literalsSmallerThanIn(SLR,Max,Ct,V),
	constraint(R,CtR),
	implies(Ct,CtR).

applyRule(V,C,R,AR,SR,Max,checked):-
%	fail,
	succedent(C,S),
	selectAtom(V, C, S, EC ),
	constraint(C,Ct),
	topDownRuleForCRW(V,C,R,AR,SR,ER),
	
	match_atoms( ER,EC ),
	vars(R,VR),
	(extraVars(VR,V) ->
	     instantiate_extra_vars( V, VR, R)
	;
	     true
	),
	makingProgress(AR,S),
	maplist(neg_atom_as_literal,AR,ALR),
	maplist(neg_atom_as_literal,SR,SLR),  % SR does not contain the matches
                    % atom which we don't need to check.
	literalsSmallerThanIn(ALR,Max,Ct,V),
	literalsSmallerThanIn(SLR,Max,Ct,V),
	constraint(R,CtR),
	implies(Ct,CtR).




makingProgress(SR,A):-
	\+ (member(E,SR),symmetric_version_of_atom(E,ES),member212(ES,A)).

bottomUpRuleForCRW(R,A):-
	rule(_,R),
	\+kind(R,symmetry(_)),
	antecedent(R,A),
	A\==[].

topDownRuleForCRW(V,C,Ru,A,S,E):-
	rule(_,Ru),
	\+kind(Ru,symmetry(_)),
	(cont(interactive,true) ->
	     succedent(Ru,Succ),
	     selectAtomR(V,C,Ru,Succ,ER)
	;
	     true
	),
	selectMaxInSuccedent( Ru, ER, Ra ),
	get_atom_members(ER,L,R),
	get_atom_relation_symbol(ER,RS),
	antecedent(Ru,A1),
	succedent(Ra,S),
	((R=='$true$';RS\== ==) ->
	      A1=A,
	      E=ER
	;% the case of a nontrivial equivalence L==R
	      create_atom(L,AL),
	      create_atom(R,AR),
%	      get_atom_preorientation(ER,P),
	      P=u,
	      (P=l ->
		   E=AL,
		   A=[AR|A1]
	      ;P=r ->
		   E=AR,
		   A=[AL|A1]
	      ;   
	           E=AL,
		   A=[AR|A1]
	      ;
		   E=AR,
		   A=[AL|A1]
	      )
	).
		   
	            

matchAtomList(Ps,As,Matched,NonMatched):-
	matchAtomList(Ps,As,Matched,NonMatched,[],I),
	instantiate(I).




matchAtomList([E|Es],A,[EC|M],NM,I,O):-
	select(EC,A,A1),
	match_atoms(E,EC,I,O1),
	matchAtomList(Es,A1,M,NM,O1,O).
matchAtomList([],A,[],A,I,I).


/*
matchAtomList(Max,Ct,Ps,As,Matched,NonMatched):-
	matchAtomList(Max,Ct,Ps,As,Matched,NonMatched,[],I),
	instantiate(I).
*/




matchAtomList(Max,Ct,[E|Es],A,[EC|M],NM,I,O):-
	select(EC,A,A1),
	match_atoms(E,EC,I,O1),
	literalSmallerThanIn([-|EC],Max,Ct),
	matchAtomList(Max,Ct,Es,A1,M,NM,O1,O).
matchAtomList(_,_,[],A,[],A,I,I).

matchLiteralList(Max,Ct,[[P|E]|Es],A,[[P|EC]|M],NM,I,O):-
	select([P|EC],A,A1),
	match_atoms(E,EC,I,O1),
	literalSmallerThanIn([P|EC],Max,Ct,_),
	matchLiteralList(Max,Ct,Es,A1,M,NM,O1,O).
matchLiteralList(_,_,[],A,[],A,I,I).



literalsSmallerThanIn([L],M,C,V):-
	literalSmallerThanIn(L,M,C,V).
literalsSmallerThanIn([L|Ls],M,C,V):-
	literalSmallerThanIn(L,M,C,V),
	literalsSmallerThanIn(Ls,M,C,V).
literalsSmallerThanIn([],_M,_C,_V).



literalSmallerThanIn(L,M,Ct,V):-
	constraint_grequal_literal_in(L,M,Ct,C),
	!,
	(false(C) ->
	     true
	;cont(interactive,true) ->
	     numbervars((V,Ct),1,_),
	     format('~nLiteral ~p not smaller than ~p ~n  in context ~p.',[L,M,Ct]),
	     fail
	).
	     


constraint_implies_greater_rule( M, R, Ct ):-
	createSimpleClause( [[],[]], C ),
	literals([M],C,CM),
	to_clause( R, CR ),
	constraint_implies_greater_clause( CM, CR, Ct ), !.

constraint_implies_greater_clause( C1, C2, Ct ):-
	constraint_grequal_clause( C2, C1, Ct, C ),
	!,
	false(C).



complementary_clauses( _, _, [], [], [] ):-!.
complementary_clauses( C, D, [E|A], S, [[C1,D]|Comps] ):-
	\+ (( selectInSuccedent( C, EC, _ ), E == EC )),
	add_succedent(  E, C, C1 ),
	complementary_clauses( C, D, A, S, Comps ), !.
complementary_clauses( C, D, [], [E|S], [[C1,D]|Comps] ):-
	\+ (( selectInAntecedent( C, EC, _ ), E == EC )),
	add_antecedent( E, C, C1 ),
	complementary_clauses( C, D, [], S, Comps ), !.


is_subsumed_by_rule_smaller_than_max( LV, _Max, C, (subs, [LV,C,R,Subst]) ):-
	% we don't need to check the complexity bound as all literals
	% in all clauses that we construct by clausal rewriting
	% are below that bound
	literals(C,L),
        rule(_, R ),
	literals(R,LR),
	subsumes_literals(LR,L,[],Subst),
	constraint(C,Ct),
	constraint(R,CtR),
	implies(Ct,CtR),
	mess_cr1( subs, [LV,C,R,Subst] ),
	!. 


% ----------------------------------------------------------------------
% Combination of clausal rewriting with case analysis on variables.
% ----------------------------------------------------------------------

all_var_cases_redundant_by_cl_rewr( Subst, M, LV, _, C, D ):-
	redundant_by_cl_rewr( Subst, M, LV, C, D , _), !.
all_var_cases_redundant_by_cl_rewr( _, _, _, LVM, _, _ ):-
	\+ var_case( LVM, _, _, _ ), !, fail.
all_var_cases_redundant_by_cl_rewr( Subst, M, LV, LVM, C, D ):-
	constraint( C, Ct ),
	var_case( LVM, Ct, LVM1, Ct1 ), 
	constraint( Ct1, C, C1 ),
	\+ all_var_cases_redundant_by_cl_rewr( Subst, M, LV, LVM1, C1, D ), !, fail.
all_var_cases_redundant_by_cl_rewr( _, _, _, _, _, _ ):- !.


var_case( LV, Ct, LV1, Ct ):-
	append( L1, [X|L2], LV ),
	member( Y, L2 ),
	X = Y,
	append( L1, L2, LV1 ).
var_case( LV, Ct, LV1, Ct1 ):-
	delete32( X, LV, LV1 ),
	LV1 \== [],
	maximality_constraint( X, LV1, Ct, Ct2 ),
	simplify_med( Ct2, Ct1 ).

maximality_constraint( _, [], Ct, Ct ).
maximality_constraint( X, [Y|LV], Ct1, Ct2 ):-
	and_gt( Ct1, [X,Y], Ct3 ),
	maximality_constraint( X, LV, Ct3, Ct2 ).




% interactive proof support

selectAtom(V,C,S,E):-
	(cont(interactive,true) ->
	     prompt(V,
       'Working on ~p.~nSelect literal for back-chaining from ~p: ',[C,S],I),
	     nth1(I,S,E)
	;
	     nth1(_,S,E)
	).

selectAtomR(V,C,R,S,E):-
	(cont(interactive,true) ->
	     prompt((V,R),
       'Working on ~p~nusing ~p.~nSelect literal for back-chaining from ~p: ',[C,R,S],I),
	     nth1(I,S,E)
	;
	     nth1(_,S,E)
	).





instantiate_extra_vars(From,For1,InRule):-
	difference(For1,From,For),
	!,
	(cont(interactive,true) ->
	     prompt((From,InRule),'Instantiation of extra variables ~p from ~p in ~n~p: ',
		    [For,From,InRule],
		    
		    IndexesInFrom
		   ),
	     mapnth(For,From,IndexesInFrom)
	 
	;
	     find_value_of_extra_vars(From,For,For)
	).

mapnth([],_,[]).
mapnth([E|Es],L,[I|Is]):-
	nth1(I,L,E),
	mapnth(Es,L,Is).


union([],S,S).
union([X|Xs],S,T):-
	member212(X,S),
	!,
	union(Xs,S,T).
union([X|Xs],S,[X|T]):-
	union(Xs,S,T).

difference([],_,[]).
difference([A|As],Bs,As1):-
	(member212(A,Bs) ->
	     difference(As,Bs,As1)
	;
	     difference(As,Bs,As2),
	     As1=[A|As2]
	).


% -------------------------------------------------------------------------

prompt(V,S,A,R):-
	nl,
	\+ \+ (numbervars(V,1,_),format(S,A)),
	read(R),
	(R=a ->
	     store(interactive,false),
	     fail
	;
	     true
	).


mess_cr1(_S,_I):- !.
mess_cr1(S,I):-
	mess_cr(S,I).

mess_cr( old, [LV,C] ):-
	numbervars( LV, 1, _ ),
	nl,write(' '),
	print( C ), nl,
	write(' is redundant; proof as above.'),
	nl,
	fail.
mess_cr( rewr, [LV,C,R,Max,Comps] ):-
	numbervars( LV, 1, _ ),
	nl,write(' '),
	print( C ), nl,
	write(' is rewritten by '), nl,
	print( R ), nl,
	write(' which is smaller than '), nl,
	print( Max ), nl,
	(Comps\==[] ->
	     write(' complementary clauses: '), nl,
	     printCC(Comps), nl
	;    true),
	fail.
mess_cr( subs, [LV,C,R,Subst] ):-
	instantiate(Subst),
	numbervars( LV, 1, _ ),
	nl,write(' '),
	print( C ), nl,
	write(' is subsumed by '), nl,
	print( R ), nl,
%	write(' which is smaller than '), nl,
%	print( Max ), nl,
	fail.
mess_cr( taut, [LV,C] ):-
	numbervars( LV, 1, _ ),
	nl,write(' '),
	print( C ), nl,
	write(' is a tautology.'), nl,fail. 
mess_cr( rid, [LV,C] ):-
	numbervars( LV, 1, _ ),
	nl,write(' '),
	write('*** can''t prove '),nl,
	print( C ), nl,
	fail. 
mess_cr( _, _ ):- !.



printCC([]).
printCC([C|Cs]):-
	format('~n     ~p',[C]),
	printCC(Cs).


% eof -------------------------------------------------------------------------

