
%							COMMANDS

%------------------------------------------------------------------------------

%declarations%

:- module(commands, [ 
        % user predicates

	cl/0,                           % clear database
        cnf/0,
        cnf/1,
	command_error/2,
	user_help/0,        % help
	in/0,in/1,insrc/1,inc/1,add/1,cd/1,     % input
	l/0,
	loadTheory/0,
	loadTheory/1,
	lprec/0,

	horn_as_eqth/0,
	magic/0,

	out/1,
	out_as_dfg/2,
	out_as_spass/1,
	prec/0,
	proof/1,

	prove/1,
	prove/2,
	prove/3,
	prove/4,

	rp/0,
	rp/1,
	rpf/0,
	saci/0, saci/1,
	sair/0, sair/1,
	saes/0, saes/1,
	sama/0, sama/1,
	sams/0, sams/1,
	sarp/0, sarp/1,
	sass/0, sass/1,
	satp/0, satp/1,
	sv/0, sv/1,
	svird/0,
	svcid/0,
	svrpd/0,
	svmsd/0,
	svmad/0,
	svtpd/0,

	sat/0,
	sat/1,
	sat/2,
	simplify/0,
	simplify_and_reanalyse/0,
	state_as_clauses_and_precedence/3,
	state_as_clauses_and_precedence1/3,
	storeTheory/0,
	storeTheory/1

   ]).

:- use_module(props, [
	age_selection_modulus/1,
	flag_on/2,
	firstOrder/0,
	option/1,
	special_relation/1

   ]).

:- use_module(theory, [
	compile_theory_properties/0

   ]).

:- use_module(proofState, [
	clause/1,
	firstPredicatePrecedence/1,
	recordFirstPredicatePrecedence/1,
	deleteAllFirstPredicatePrecedences/0,
	newModuleNumber/1,
	operators/1,
	precedence/1,
	recordPrecedence/1,
	deleteAllPrecedences/0,
	predicates/1,
	recordPredicates/1,
	recordProps/1,
	rule/2
	
   ]).


:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(sets), [
	subtract/3
   ]).

:- use_module(library(maplist), [
	maplist/3
   ]).

:- use_module(library(flatten), [
	flatten/2
   ]).


:- use_module(mylists, [
	delete/3,
	is_list/1
   ]).

:- use_module(myterms, [
	mapargs/3
   ]).


:- use_module(classes, [
	antecedent/2,
	createClause/2,
	succedent/2,
        as_dfg/3,
	as_prolog_sequent/2,
        as_tptp/2,
        as_spass/2,
        asSequent/2,
	dual_sequent/2,
        empty/1,
	kind/2,
	literal_list_to_sequent/3,
	indexEntries/3,
	number/3
   ]).

:- use_module(adt_declarations, [
	all_flag_types/1,
	all_numbers/2
   ]).

:- use_module(adt_flag, [
	active_flag_numbers/2,
	compile_flags/2,
	create_initial_default_file/0,
	default_flag_numbers/2,
	save_defaults/2,
	compile_defaults/0,
	set_default_flags_as_on_flags/0
   ]).

:- use_module(adt_precedence, [
	define_precedence/2,
	precedence_as_facts/1,
	is_predicate/1,
	is_skolem_function/1,
	abbreviating_proposition/1,
	split_proposition/1,
	is_total/2,
	new_minimal_function_symbols/1,
        special_symbol_conv/2,
	strip_arity/2
   ]).

:- use_module(adt_proof, [
	recover_proof/0,
	recover_proof_for/1
   ]).


:- use_module(adt_term, [
	builtInOp/1
   ]).

:- use_module(adt_atom, [
	get_atom/4,
	create_atom/4
   ]).

:- use_module(auxiliary, [
        appendAtoms/3,
	interval/3,
	member212/2,
	member22/2
   ]).

:- use_module(global, [
	cont/2,
	initialize_vars/0,
	initialize_theory_vars/0,
	store/2
   ]).

:- use_module(hgstuff, [
        endOfFile/1,
	setof0/3,
	split/4,
	try/1
   ]).

:- use_module(input, [
	parse/2,
	read_file_name/1,
	read_file/3
   ]).

:- use_module(output, [
	displayall/0,
	store_props/0
   ]).

:- use_module(thisProlog, [
	command_exists/1,
	file_exists/1,
	gen_tmp_files/0,
	gen_tmp_file/2,
	my_absolute_file_name/2,
	save_state/0,
	savedOn/2,
	portray_clause2/2,
	prolog_version/1,
	compile_nqa/2
   ]).

:- use_module(timing, [
	start_timing/0
   ]).

:-use_module(saturation, [
	make_rules_new/0,
        sat/0,
	sat/1,
	sat/2
   ]).

:- use_module(os, [
	command_name/2
   ]).


:-use_module(softtypes, [
	magicS/2
   ]).

/*
:- op(200,yfx,[and]).
:- op(300,yfx,[or]).
:- op(400,yfx,[=>, <=>]).
:- op(170,fy,[~,dia,box]).
:- op(600,yfx,[::]).
:- op(600,xfx,[^^]).
*/
:- use_module(translate).


%end%



%==============================================================================
%  clean : CLEAN the database
%==============================================================================

cl:- 
	clean_database.

	

clean_database:-
	compile_initial,
	initialize_vars,
	initialize_theory_vars,
	start_timing.
%	statistics(heap,[Size,_]),
%	format('Heap size = ~p~2n',[Size]).





compile_initial:-
	compile_nqa(user,satsrc(initialize)).



%==============================================================================
%  hlp, hlp(command_name) : display the helps
%==============================================================================

% this is the user help hook to be called upon ``help''

user_help:- 
	try(start_www),
	!.

/* outdated

user_help:-
	hlp.

hlp:-
	display_help_file( hlp      ), nl.



hlp(X) :- 
	name(X,N), 
	user:append("hlp",N,HN), 
	name(H,HN), 
	nofileerrors,
	nl, display_help_file(H), nl,
	fileerrors,
	!.
hlp(X) :- 
	nl, write('No help information available for '),
	write(X),
	write('.'),
	nl,
	fileerrors.

*/



start_www :-
	user:os(windows),
	!,
	user:saturate_directory(D),
	appendAtoms([D,'doc/Saturate/index.html'],'/',Manual_URL),
	user:html_browser(startup(Manual_URL),Startup_CmdL),
	appendAtoms(Startup_CmdL,'',Startup_Cmd),
	format('[Attempting to spawn Web browser with Saturate manual...]',[]),
	unix(system(Startup_Cmd)).
start_www :-
	user:saturate_directory(D),
	appendAtoms([D,'doc/Saturate/index.html'],'/',Manual_URL),
	user:html_browser(startup(Manual_URL),Startup_CmdL),
	appendAtoms(Startup_CmdL,'',Startup_Cmd),
	user:html_browser(already_running,Test_CmdL),
	appendAtoms(Test_CmdL,'',Test_Cmd),
	user:html_browser(inform(Manual_URL),Inform_CmdL),
	appendAtoms(Inform_CmdL,'',Inform_Cmd),

	(unix(system(Test_Cmd)) ->
	     % netscape is already running
	     format('[Displaying manual with running instantiation of Web browser...~n Check your Web browser window.]',[]),
	     Cmd=Inform_Cmd
	;
	     format('[Attempting to spawn Web browser with Saturate manual...]',[]),
	     Cmd=Startup_Cmd
	),
	     unix(system(Cmd)).



%------------------------------------------------------------------------------
% display_help_file
%------------------------------------------------------------------------------

display_help_file( F ):-
	command_name(cat,More),
	user:saturate_directory( D1 ),
	name( D1, D2 ),
	user:append( More, D2, D3 ),
	name( '/helps/', Slash ),
	user:append( D3, Slash, D4 ),
	name( F, F1 ),
	user:append( D4, F1, F2 ),
	user:append(F2,".html",F4),
	name( F3, F4 ),
	unix( system( F3 ) ),!.

%==============================================================================
%  in	: read in a file
%		  This file can be user written or created by the 
%		  command "out", and can contain axioms, clauss
%		  rules and operator precedences. + predicate declarations?
%==============================================================================



in:- 	
	read_file_name(F),
	in(F).




% -------------------------------------------------
% for reading in a file from the current examples
% directory; this can be set by the cd command
% below; cleans the database
% -------------------------------------------------



in(F):-
	cl,
	set_default_flags_as_on_flags,
	add(F).
/*
	(props:(horn,special_user_relation(_),special_relations(on)) ->
	     format("~2n[Re-reading file with option(special_relations(off))~n",[]),
	     cl,
	     set_default_flags_as_on_flags,
	     props:option(special_relations(off)),
	     add(F)
	;
	     true
	).
*/

insrc(F):-
	cl,
	set_default_flags_as_on_flags,
	add(src,F).

inc(F):-
	cl,
	add(F).


% this adds the file to the current state
addsrc(F):-
	add(src,F).

add(F):-
	add(saved,F).

add(Saved,F):-

	newModuleNumber( M ),
	(F=[FO,F1] ->
	     (atom(FO)->
		  (atom(F1) ->
		       true
		  ;
	               command_error("~p should be an atom.",[F1])
		  )
	     ;
	          command_error("~p should be an atom.",[FO]) 
	     )
	           
	;
	     (atom(F) ->
		  F=F1
	     ;
	          command_error("~p should be an atom.",[F])
	     )
	     
	),

	read_file(Saved,F,[module(F1,M)]),

	declare_predicates,
	compile_theory_properties,

	!,
	control_precedence,
	displayall,
	display_precedence,
	store_props.

cd(D1):-
	my_absolute_file_name(D1,D),
	unix(cd(D)),
	nl,
	command_name(ls,LSN),
	name(LS,LSN),
	unix(system(LS)),
	nl,
	nl,
	ttyflush,
	user:no_style_check(multiple),
	compile_defaults,
	set_default_flags_as_on_flags,
	!.
cd(D):-
	nl,
	user:style_check(multiple),
	format('~nCannot cd to ~p.~n',[D]).

	



declare_predicates:-
	(predicates(P) ->
	     format('~n~nParsed as predicates: ~p',[P])
	;
	     recordPredicates([])
	).


%==============================================================================
%  l 	: list all the current axioms in the database
%==============================================================================

l:- 	displayall,
	(rule( _N, Ru ),empty( Ru ) ->
	     true
	;
	     control_precedence
	).

%==============================================================================
%  lprec : list the current precedence
%==============================================================================


lprec:-	nl, control_precedence, !.

%==============================================================================
%  out	: save current database in a file
%==============================================================================

out(F):- 
	open( F, write, Str),
	write_clauses(Str),
	write_rules(Str),
	write_precedence(Str),
	write_first_predicate_precedence(Str),
%	write_predicates,
	close( Str ).


write_rules(Str):-
	rule(_, Ru ),
	indexEntries(_,Ru,Ru1), % invalidate them
	numbervars( Ru1, 0, _ ),
	portray_clause2( Str, rule( Ru1 )),
	fail.
write_rules(_):- !.

write_clauses(Str):-
	clause( C ),
	numbervars( C, 0, _ ),
%	write_term( clause( C ), [quoted(true),numbervars(true)] ), 
	portray_clause2(Str, clause( C )),
	fail.
write_clauses(_):- !.

write_precedence(Str):-
	precedence( P ),
	portray_clause2( Str, precedence( P ) ), 
	fail.
write_precedence(_):- !.

write_first_predicate_precedence(Str):-
	firstPredicatePrecedence( P ),
	portray_clause2( Str, first_predicate_precedence( P ) ),
	fail.
write_first_predicate_precedence(_):- !.

write_predicates:-
	predicates( P ),
	write( predicates( P ) ), write('.' ), nl, fail.
write_predicates:- !.

%==============================================================================
%  precedence
%==============================================================================

%------------------------------------------------------------------------------
% well definedness control of the precedence definition, that is, if it is total
% and all the operators appearing in the initial axioms are related
%
% it is supposed that L1 and L2 are disjoint,
% where first_predicate_precedende(L1)and precedence(L2)
%------------------------------------------------------------------------------

control_precedence:-
	define_precedence(FP,Prec),
	deleteAllFirstPredicatePrecedences,
	recordFirstPredicatePrecedence( FP ),
	deleteAllPrecedences,
	recordPrecedence( Prec ),
	precedence_as_facts(G),
	recordProps(G),
	!.

display_precedence:-
	nl, nl, write( 'Current precedence:' ),nl, nl,
	write( '   First predicates: ' ), 
	display_first_predicate_precedence,
	write( '   Other symbols: ' ), 
	display_rest_precedence.

display_first_predicate_precedence:-
	firstPredicatePrecedence( L ),
	write( '     ' ), write( L ), nl, fail.
display_first_predicate_precedence:- !.

display_rest_precedence:-
	precedence( L ),
	write( '     ' ), write( L ), nl, fail.
display_rest_precedence:- !.


%==============================================================================
%  prec	: reads a new PRECedence from a file
%==============================================================================

prec:-	write('File: '),
	read(F), see( F ),
	deleteAllFirstPredicatePrecedences,
	deleteAllPrecedences,
	repeat, read( Input ),
	(endOfFile(Input) ->  
	     control_precedence
	;Input=precedence(P) -> 
	     recordPrecedence(P),
	     fail
	;Input=first_predicate_precedence(P) -> 
	     recordFirstPredicatePrecedence(P),
	     fail
	;Input=top_predicates_precedence(P) -> 
	     recordFirstPredicatePrecedence(P),
	     fail
	),
	seen,
	store_props.



%==============================================================================
%  prect : is Prececedence Total
%==============================================================================

prect:-	control_precedence.

%==============================================================================
%  rp [proof]	: Recover the Proof [for a particular clause]
%  rpf      	: Recover the Proof For a given clause number
%==============================================================================

% this is in module: proof_recovering

rp:-	recover_proof.

rp(N):-
	recover_proof_for(N).

proof(N):-
	rp(N).

rpf:-	write(' Number of the clause for which you want the proof: '),
	read( N ),
	recover_proof_for( N ),!.

%==============================================================================
%  saci	: Show/Set Constraint Inheritance
%==============================================================================

saci:-	print_menu( ci ),
	active_flag_numbers( ci, Act ),
	write( 'The active constraint treatment strategy is now: ' ),
	write( Act ), nl, nl,
	default_flag_numbers( ci, Def ),
	write( 'Your default file option for constraint treatment is: ' ),
	write( Def ), nl, nl,
        write( 'Type: ' ), nl, nl,
        write( '  n		no changes' ), nl, nl, 
        write( '  d		to activate the constraint inheritance' ), nl,
	write( '		as specified in your default file' ), nl,nl, 
	write( '  [n1,...,ns]	to activate n1,...,ns, s >= 0' ), nl,
	write( '		intervals n-m may be used' ),
	nl, nl,
        read( L ),
	saci(L).


saci(L):-
	treat_answer( ci, L ).




%==============================================================================
%  sair	: Show/Set Active Inference Rules
%==============================================================================

sair:-	print_menu( ir ),
	active_flag_numbers( ir, Act ),
	write( 'The active inference rules are now: ' ),
	write( Act ),nl,
	( flag_on( ma, _ ) ->
	     write( '                       with selection of negative literals' ); true ),
	nl, nl,
	default_flag_numbers( ir, Def ),
	write( 'Your default file option for inference rules is: ' ),
	write( Def ), nl, nl,
	write( 'Type: ' ), nl, nl,
        write( '  n             no changes' ), nl, nl,
        write( '  d             to activate the inference rules as specified'),
	nl,
	write( '		in your default file' ), nl, nl,
	write('  [n1, ..., ns]	to activate inference rules n1,..., ns, s >= 0'),
	nl,
	write('  		intervals n-m may be used'), nl, nl,
	read( L ),
	treat_answer( ir, L ), !.
sair(L):-
	treat_answer( ir, L ), !.
	

%==============================================================================
%  saes	: Show/Set Active Eager Simplifications
%==============================================================================

saes:-	print_menu( es ),
	active_flag_numbers( es, Act ),
	write( 'The active eager simplifications are: ' ),
	write( Act ),nl,
	nl, nl,
	default_flag_numbers( es, Def ),
	write( 'Your default file option for eager simplifications is: ' ),
	write( Def ), nl, nl,
	write( 'Type: ' ), nl, nl,
        write( '  n             no changes' ), nl, nl,
        write( '  d             to activate eager simplifications as specified'),
	nl,
	write( '		in your default file' ), nl, nl,
	write('  [n1, ..., ns]	to activate eager simplifications n1,..., ns, s >= 0'),
	nl,
	write('  		intervals n-m may be used'), nl, nl,
	read( L ),
	treat_answer( es, L ), !.
saes(L):-
	treat_answer( es, L ), !.

%==============================================================================
%  sama	: Show/Set Active Marking
%==============================================================================

sama:-  print_menu( ma ),
        active_flag_numbers( ma, Act ),
        write( 'The active selection strategy is: ' ),
        write( Act ), nl, nl,
        default_flag_numbers( ma, Def ),
        write( 'Your defaults file option for selection is: ' ),
        write( Def ), nl, nl,
        write( 'Type: ' ), nl, nl,
        write( '  n             no changes' ), nl, nl,
        write( '  d             to activate selection as specified' ), nl,
        write( '                in the defaults file' ), nl, nl,
        write( '  [m]           to activate selection strategy m' ),
        nl,nl,
        read( L ),
        treat_answer( ma, L ),!.
sama(L):-
        treat_answer( ma, L ),!.



%==============================================================================
%  sass: new name for sama
%==============================================================================

sass:-
	sama.

sass(L):-
	sama(L).



%==============================================================================
%  sams	: Show/Set Active Messages
%==============================================================================

sams:-	print_menu( ms ),
	active_flag_numbers( ms, Act ),
	write( 'The active messages are now: ' ),
	write( Act ), nl, nl,
	default_flag_numbers( ms, Def ),
	write( 'Your default file option for messages is: ' ),
	write( Def ), nl, nl,
        write( 'Type: ' ), nl, nl,
        write( '  n		no changes' ), nl, nl,
        write( '  d		to activate the messages as specified' ), nl, 
	write( '		in your default file' ), nl, nl,
        write( '  all		to activate all messages' ), nl, nl,
        write( '  [n1, ..., ns]	to activate messages n1, ..., ns, s >= 0' ),
	nl,
	write('  		intervals n-m may be used'), nl, nl,
        read( L ),
	treat_answer( ms, L ),!.
sams(L):-
	treat_answer( ms, L ),!.

%==============================================================================
%  sarp	: Show/Set Active Redundancy Proofs
%==============================================================================

sarp:-  print_menu( rp ),
	active_flag_numbers( rp, Act ),
	write( 'The active redundancy proofs are now: ' ),
	write( Act ), nl, nl,
	default_flag_numbers( rp, Def ),
        write( 'Your default file option for active redundancy proofs is: ' ),
	write( Def ), nl, nl,
        write( 'Type: ' ), nl, nl,
        write( '  n             no changes' ), nl, nl,
        write( '  d             to activate the redundancy proofs as specified'),
	nl,
        write( '		in your default file' ), nl, nl,
        write( '  all           to activate all redundancy proofs' ), nl, nl,
        write('  [n1, ..., ns]	to activate redundancy proofs n1,...,ns, s >=0'),
	nl,
	write('  		intervals n-m may be used'), nl, nl,
        read( L ),
	treat_answer( rp, L ),!.
sarp(L):-
	treat_answer( rp, L ),!.




%==============================================================================
%  satp	: Show/Set Active Timing Points
%==============================================================================

satp:-  print_menu( tp ),
	active_flag_numbers( tp, Act ),
	write( 'The active timing points are now: ' ),
	write( Act ), nl, nl,
	default_flag_numbers( tp, Def ),
        write( 'Your default file option for active timing points is: ' ),
	write( Def ), nl, nl,
        write( 'Type: ' ), nl, nl,
        write( '  d             to activate timing points as specified'),
	nl,
        write( '		in your default file' ), nl, nl,
        write( '  all           to activate all timing points' ), nl, nl,
        write('  [n1, ..., ns] to activate timing points n1,...,ns (s >=0)'),
	nl, nl,
        write( '  n             no changes' ), nl, nl,
        read( L ),
	treat_answer( tp, L ),!.
satp(L):-
	treat_answer( tp, L ),!.

%------------------------------------------------------------------------------
% print menu
%------------------------------------------------------------------------------



print_menu( Tabletype ):-
       	nl,nl,
        Term =.. [Tabletype,N,_,TextL],
        adt_declarations:call( Term ),
       	write_line( N, TextL ), fail.
print_menu( _ ):- nl,nl,!.

write_line( 0, [Text|_] ):-
       	write( Text ), nl, !.
write_line( N, [Text|_] ):-
       	( 10 > N -> write( ' ' ); true ),
	write( N ), write(':    '),
       	write( Text ), nl,!.

%------------------------------------------------------------------------------
% treat_answer
%------------------------------------------------------------------------------

treat_answer( _, n ):-!.
treat_answer( Flagtype, d ):-
	default_flag_numbers( Flagtype, NL ),
	treat_answer( Flagtype, NL ), !.
treat_answer( Flagtype, all ):-
	all_numbers( Flagtype, NL ),
	treat_answer( Flagtype, NL ), !.
treat_answer( Flagtype, +L):-
	active_flag_numbers( Flagtype, Act ),
	append(Act,L,L1),
	sort(L1,L2),
	treat_answer( Flagtype, L2 ), !.
treat_answer( Flagtype, -L):-
	active_flag_numbers( Flagtype, Act ),
	subtract(Act,L,L1),
	sort(L1,L2),
	treat_answer( Flagtype, L2 ), !.
treat_answer( Flagtype, L ):-
	construct_integer_list( L, L1 ),
	(compile_flags( Flagtype, L1 ) ->
	     true
	;
	     format("~n* I'm having problems with setting flags.~n* Maybe the defaults file in this directory is OUTDATED or otherwise corrupted.~n* Remove or correct the file and cd again into here.",[])
	),!.
treat_answer( Flagtype, _ ):-
	write(' Incorrect option. Please try again.'), nl, 
	read( L ), treat_answer( Flagtype, L ), !.

construct_integer_list( [], [] ):- !.
construct_integer_list( [X|L], [X|L1] ):-
	integer( X ),
	construct_integer_list( L, L1 ), !.
construct_integer_list( [X|L], L3 ):-
	X =..[-,N,M],
	interval( N, M, L1 ),
	construct_integer_list( L, L2 ),
	user:append( L1, L2, L3 ), !.


	

%==============================================================================
%  sv**	: Save defaults of flag types.
%==============================================================================

sv(K):- all_flag_types(T), member(K,T), save_defaults( K, active ).

sv:-
	sv(_),
	fail.
sv.

svird:- save_defaults( ir, defaults ), !.
svcid:- save_defaults( ci, defaults ), !.
svrpd:- save_defaults( rp, defaults ), !.
svmsd:- save_defaults( ms, defaults ), !.
svmad:- save_defaults( ma, defaults ), !.
svtpd:- save_defaults( tp, defaults ), !.


%------------------------------------------------------------------------------
% Theorem proving with translation
%------------------------------------------------------------------------------

:- dynamic problem_file/1.


prove(Thm):-
	prove(Thm,[]).


prove(Thy=>Thm,Options):-
	!,
	cl,
	loadTheory(Thy),
	prove(Thm,classic,_,Options).


prove(Thm,Lo,TMode):-
	prove(Thm,Lo,TMode,[]).


prove(Thm,classic,_TMode,Options):-
	!,
	as_goal_axiom([[Thm],[]],G),
	clauses_as_state([G]),
	maplist(call,Options),
	store_props,
	sat.
prove(Thm,Lo,TMode,Options):-
	cl,
	translate(Thm,Lo,TMode,_Skolems,Theory,Goal),
	maplist(as_th_axiom,Theory,Th),
	maplist(as_goal_axiom,Goal,Gs),
	append(Th,Gs,Ax),
	maplist(deskolemize,Ax,Ax1),
	clauses_as_state(Ax1),
	maplist(call,Options),
	store_props,
	sat.

clauses_as_state(Ax):-
%	cl,
%	set_default_flags_as_on_flags,
	newModuleNumber( M ),
	% now let's parse everything to deal properly with
	% theory properties, skolem functions etc
	parse(Ax,[module(user,M)]),  
	option(horn(off)),
	compile_theory_properties,
	declare_predicates,
	displayall,
	!,
	control_precedence,  % required for compiling greater_op
	display_precedence,
	store_props.


as_th_axiom(A,axiom(A)).

as_goal_axiom(A,goal(axiom(A))).


theory_exists(Th):-
	(atom(Th) ->
	     (cont(theory(Th),_);file_exists(Th))
	;% Th is explicit representation
	     true
	).

storeTheory:-
	storeTheory(ct).
storeTheory(Th):-
	atom(Th),
	state_as_clauses_and_precedence(C,P,FP),
	appendAtoms([theory,Th],'$',TTh),
	store(TTh,[precedence(P),top_predicates_precedence(FP)|C]).

	

loadTheory:-
	loadTheory(ct).

loadTheory(Th):-
	% by modularity considerations we don't compute
	% any inferences between the clauses of any theory
	active_flag_numbers(ir,I),
	active_flag_numbers(rp,RP),
	active_flag_numbers(ma,MA),
	active_flag_numbers(ci,CI),
	age_selection_modulus(Age),
	appendAtoms([theory,Th],'$',TTh),
	cont(TTh,Cl),
	deleteAllFirstPredicatePrecedences,
	deleteAllPrecedences,
	cl,
	clauses_as_state(Cl),
	sama(MA),
	saci(CI),
	sair(I),
	sarp(RP),
	option(age_selection_modulus(Age)).




state_as_clauses_and_precedence(Ks,P,FP):-
	findall(R,rule(_,R),Rs),
	maplist(indexEntries(_),Rs,Rs1),  % invalidate them
	maplist(mk_rule,Rs1,Rs2),
	findall(R,clause(R),Cs),
	maplist(mk_clause,Cs,Cs1),
	append(Rs2,Cs1,Ks),
	precedence(P),
	firstPredicatePrecedence(FP).
	

state_as_clauses_and_precedence1(Ks,P,FP):-
	findall(R,rule(_,R),Rs),
	maplist(indexEntries(_),Rs,Rs1),  % invalidate them
	findall(R,clause(R),Cs),
	append(Rs1,Cs,Ks),
	firstPredicatePrecedence(FP),
	precedence(P).

mk_rule(X,rule(X)).

mk_clause(X,clause(X)).

mk_axiom((A->S),axiom([A,S])).
mk_axiom([A,S],axiom([A,S])).

magic:-
	state_as_clauses_and_precedence1(Ks,P,FP),
	maplist(as_prolog_sequent,Ks,Ps),
	maplist(magicS,Ps,Mss),
	flatten(Mss,Ms),
	maplist(mk_axiom,Ms,Ax),
	cl,
	set_default_flags_as_on_flags,
	clauses_as_state([precedence(P),top_predicates_precedence(FP)|Ax]).

reset:-
	active_flag_numbers(ir,I),
	active_flag_numbers(rp,RP),
	active_flag_numbers(ma,MA),
	active_flag_numbers(ci,CI),
	age_selection_modulus(Age),
	state_as_clauses_and_precedence1(Ks,_P,_FP),
	deleteAllFirstPredicatePrecedences,
	deleteAllPrecedences,
	maplist(asSequent,Ks,Cs),
	maplist(mk_axiom,Cs,Ax),
	cl,
%	clauses_as_state(Ax).
	clauses_as_state([precedence(_P),top_predicates_precedence(_FP)|Ax]),
	sama(MA),
	saci(CI),
	sair(I),
	sarp(RP),
	option(age_selection_modulus(Age)).



simplify_and_reanalyse:-
	active_flag_numbers(ir,I),
	active_flag_numbers(rp,RP),
	active_flag_numbers(ma,MA),
	active_flag_numbers(ci,CI),
	sair([]),
	sarp(+[3,15]),
	sat,
	out_as_dfg('/tmp/sat.dfg',dfg),
	in('/tmp/sat.dfg'),
	sama(MA),
	saci(CI),
	sair(I),
	sarp(RP).

simplify:-
	active_flag_numbers(ir,I),
	active_flag_numbers(rp,RP),
	active_flag_numbers(ma,MA),
	active_flag_numbers(ci,CI),
	sair([]),
	sarp(+[3,15]),
	sat,
	make_rules_new,
	sama(MA),
	saci(CI),
	sair(I),
	sarp(RP).





dual:-
	state_as_clauses_and_precedence1(Ks,P,FP),
	maplist(dual_sequent,Ks,DSs),
	maplist(mk_axiom,DSs,Ax),
	cl,
	set_default_flags_as_on_flags,
	clauses_as_state([precedence(P),top_predicates_precedence(FP)|Ax]).


cnf:-
	(firstOrder ->
	    active_flag_numbers(ir,I),
	    active_flag_numbers(rp,RP),
	    active_flag_numbers(ma,MA),
	    active_flag_numbers(ci,CI),
	    sair([7]),
%	    sarp(+[5]), % turns blocking off
	    sama([]),
	    sat,
	    format('~2nResetting proof state.~n',[]),
%	    reset,
	    out_as_dfg('/tmp/sat.dfg',sat),
	    in('/tmp/sat.dfg'),
	    sair(I),
	    sarp(RP),
	    sama(MA),
	    saci(CI)
	;
	    true).




cnf(full):-
	option(full_cnf(yes)),
	cnf.
cnf(partial):-
	option(full_cnf(no)),
	cnf.


	




%------------------------------------------------------------------------------
% auxiliary commands
%------------------------------------------------------------------------------

ns:- nospyall.

nd:- nodebug.

da:- displayall.

ls:- listing.

un:-
	write('Unix command: '), read( Com ),
	unix( system( Com ) ), !.




out_as_spass(F):-
	findall(R,rule(_,R),Rs),
	findall(C,clause(C),Cs),
	maplist(as_spass,Rs,Ts),
	maplist(as_spass,Cs,Us),
%	format('~2nWriting this Spass Output to ~p:`2n',[F]),
%	maplist(printSpass(user_output),Ts),
%	maplist(printSpass(user_output),Us),
	open( F, write, Str),
	maplist(printSpass(Str),Ts),
	maplist(printSpass(Str),Us),
	close(Str).


printSpass(Str,'$spass'(C,A,S)):-
	\+ \+ (numbervars((C,A,S),1,_),
	       write((C,A,S)),nl,
	       format(Str,"~p || ~p -> ~p.~n",['$spassTerms'(C),'$spassTerms'(A),'$spassTerms'(S)])
	      ).

out_as_dfg(F,Mode):-
	out_as_dfg(F,Mode,_TimeBound).

out_as_dfg(F,Mode,TimeBound):-
	findall(R,rule(_,R),Rs),
	findall(C,clause(C),Cs),
	maplist(as_dfg(Mode),Rs,Ts),
	maplist(as_dfg(Mode),Cs,Us),
	format('~2nWriting theory to ~p in DFG notation ...',[F]),
	ttyflush,
	predicates(Ps),
	delete(Ps,equal/2,Ps1),
	operators(Os),
	spass_precedence(Prec),
	append(Ts,Us,AllCls),
	collect_symbols(is_skolem_function,AllCls,SkfsL),
	collect_symbols(abbreviating_proposition,AllCls,AbbrevL),
	collect_symbols(split_proposition,AllCls,CasesL),
	sort(SkfsL,SkfsSpass),
	sort(AbbrevL,AbbrevsSpass),
	sort(CasesL,CasesSpass),
	maplist(strip_arity,SkfsSpass,SkfsSpassNA),
	maplist(strip_arity,AbbrevsSpass,AbbrevsSpassNA),
	maplist(strip_arity,CasesSpass,CasesSpassNA),
	flatten([Prec,SkfsSpassNA,AbbrevsSpassNA,CasesSpassNA],PrecS),
	flatten([Os,SkfsSpass],OsS),
	flatten([Ps1,AbbrevsSpass,CasesSpass],PsS),
	write_dfg_code(F,(PsS,OsS,PrecS,TimeBound,AllCls)),
	!.


% special symbols are already converted to DFG-notation!
collect_symbols(_,X,[]):-
	var(X),
	!.
collect_symbols(P,C,[(F/N)|Fs]):-
	functor(C,F,N),
	special_symbol_conv(FD/N,F/N),  % de-convert
	Call=..[P,FD/N],
	Call,
	C=..[F|Args],
	!,
	maplist(collect_symbols(P),Args,FsL),
	flatten(FsL,Fs).
collect_symbols(P,C,Fs):-
	C=..[_F|Args],
	maplist(collect_symbols(P),Args,FsL),
	flatten(FsL,Fs).




:- dynamic condop_for_arity/1, cond_number/1.


horn_as_eqth:-
	findall(R,rule(_,R),Rs),
	findall(C,clause(C),Cs),
	retractall(condop_for_arity(_)),
	retractall(cond_number(_)),
	assert(cond_number(c_0)),
	maplist(horn_clause_as_eq,Rs,Es1),
	maplist(horn_clause_as_eq,Cs,Es2),
	append(Es1,Es2,Es),
	declare_condops,
%	create_atom(if_then_(X,X,T),'=',T,IFTrue),
	IFTrue=(if_then_(X,X,T,_)=T),
	maplist(clauseFromUnitLiteral,[[+|IFTrue]|Es],Cls),
	precedence(P),
	cl,
	clauses_as_state([precedence(P)|Cls]).


clauseFromUnitLiteral([+|E],C):-
	mk_axiom([[],[E]],C).
clauseFromUnitLiteral([-|E],C):-
	mk_axiom([[E],[]],C).


horn_clause_as_eq(R,[+|E]):-
	antecedent(R,[]),
	!,
	succedent(R,[Eq]),
	get_atom(Eq,Rel,S,T),
	eqRight(Rel,T,T1),
	eqLeft(Rel,S,T,S1),
	E=(S1=T1).
%	create_atom(S1,'=',T1,E).
horn_clause_as_eq(R,[+|E]):-
	antecedent(R,A),
	length(A,N),
	succedent(R,[Eq]),
	!,
	get_atom(Eq,Rel,S,T),
	eqRight(Rel,T,T1),
	eqLeft(Rel,S,T,S1),
	auxiliary_ops(N,TupleN),
	E=(S1=if_then_(CondL,CondR,T1,S1)),
	functor(CondL,TupleN,N),
	functor(CondR,TupleN,N),
	condition_term(A,1,N,CondL,CondR).
horn_clause_as_eq(R,[-|E]):-
	antecedent(R,A),
	succedent(R,[]),
	!,
	length(A,N),
	auxiliary_ops(N,TupleN),
	E=(if_then_(CondL,CondR,tt_,tt_)=tt_),
	functor(CondL,TupleN,N),
	functor(CondR,TupleN,N),
	condition_term(A,1,N,CondL,CondR).


declare_condops:-
	findall(O,condop_for_arity(O),COps),
	findall(C/0,(cond_number(C),C\==c_0),DOps),
	append(COps,DOps,Ops),
	new_minimal_function_symbols([tt_/0|Ops]).


/*
auxiliary_ops(N,TupleN,CondN):-
	appendAtoms(['tuple',N],'_',TupleN),
	(condop_for_arity(TupleN/N) ->
	    true
	;   assert(condop_for_arity(TupleN/N))
	),
	(cond_number(Last) ->
	    name(Last,LastName),
	    append("c_",LName,LastName),
	    name(L,LName),
	    Next is L+1,
	    appendAtoms([c,Next],'_',CondN),
	    asserta(cond_number(CondN))
	),
	!.
*/

auxiliary_ops(N,TupleN):-
	appendAtoms(['tuple',N],'_',TupleN),
	(condop_for_arity(TupleN/N) ->
	    true
	;   assert(condop_for_arity(TupleN/N))
	),
	!.



eqRight(Rel,T,T1):-
	( (T=='$true$'; (Rel\== '=', Rel\== '==')) ->
	    T1=tt_
	;   T1=T
	).

eqLeft('=',S,_T,S):-
	!.
eqLeft('==',S,_T,S):-
	!.
eqLeft(R,S,T,RST):-
	RST=..[R,S,T].

condition_term(_,I,N,_,_):-
	I>N,
	!.
condition_term([E|Es],I,N,CS,CT):-
	get_atom(E,Rel,S,T),
	eqRight(Rel,T,T1),
	eqLeft(Rel,S,T,S1),
	arg(I,CS,S1),
	arg(I,CT,T1),
	I1 is I+1,
	condition_term(Es,I1,N,CS,CT).



spass_precedence(Prec):-
	precedence(Prec1),
	split(special_relation,Prec1,Prec1S,Prec1N),
	firstPredicatePrecedence(Prec2),
	append(Prec1S,Prec1N,Prec3),
	append(Prec2,Prec3,Prec4),
	delete(Prec4,equal,Prec).


ftop(F/N,(F,N)).

write_dfg_code(F,(Ps,Os,Prec,time(TimeBound),P)):-
	maplist(ftop,Ps,Ps1),
	maplist(ftop,Os,Os1),
	open(F,write,S),
	format(S,"begin_problem(Unknown).~2n",[]),
	format(S,"list_of_descriptions.~n",[]),
	format(S,"name({*some theory*}).~n",[]),
	format(S,"author({*Saturate*}).~n",[]),
	format(S,"status(satisfiable).~n",[]),
	format(S,"description({*type theory of some logic program*}).~n",[]),
	format(S,"end_of_list.~2n",[]),

	format(S,"list_of_symbols.~n",[]),
	(Os1\==[] -> format(S,"functions~p.~n",[Os1]);true),
	(Ps1\==[] -> format(S,"predicates~p.~n",[Ps1]);true),
	format(S,"end_of_list.~2n",[]),
	split(dfg_theory_clause,P,PT,PG),
	maplist(negate_dfg_clause,PG,PGN),
	format(S,"list_of_formulae(axioms).~n",[]),
%	P=PT,
	maplist(print(S),PT),
	format(S,"end_of_list.~2n",[]),
	format(S,"list_of_formulae(conjectures).~n",[]),
	maplist(print(S),PGN),
	format(S,"end_of_list.~2n",[]),
	

	format(S,"list_of_settings(SPASS).~n{*~nset_flag(Ordering,1).~nset_precedence(~p).",['$list'(Prec)]),  % RPO mit Status (0 fuer KBO)
%	format(S,"list_of_settings(SPASS).~n{*~nset_flag(Auto,0).~nset_flag(Ordering,1).~nset_precedence(~p).",['$list'(Prec)]),  % RPO mit Status (0 fuer KBO)
	(number(TimeBound) ->
	     format(S," ~p ",[TimeBound])
	;true),

	format(S,"~n*}~nend_of_list.~nend_problem.~n",[]),
        close(S).



dfg_theory_clause('$dfg'(_C,theory)).

negate_dfg_clause('$dfg'(clause(forall(V,not(C))),TF),
		  '$dfg'(clause(exists(V,C)),TF)):- !.
negate_dfg_clause('$dfg'(clause(not(C)),TF),'$dfg'(clause(C),TF)):- !.
negate_dfg_clause('$dfg'(clause(C),TF),'$dfg'(clause(not(C)),TF)).


command_error(S,Args):-
	append("~n!! ",S,S1),
	format(S1,Args),
	nl,
	!,
	fail.

	
:- mymultifile user:portray/1.

user:portray('$spassTerms'([])).
user:portray('$spassTerms'([A])):-
	print('$spassTerm'(A)).
user:portray('$spassTerms'([A,B|As])):-
	print('$spassTerm'(A)),
	write(' '),
	print('$spassTerms'([B|As])).
user:portray('$spassTerm'('$VAR'(N))):-
	!,
	format("x~p",[N]).
user:portray('$spassTerm'(T)):-
	T=..[F|Ts],
	(F=equal ->
	     (props: (\+ congruence_relation_symbol(equal)) ->
		  F1= F
	     ;
	          F1= =
	     )
	;
	     F=F1
	),
%	format("(~p ~p)",[F1,'$spassTerms'(Ts)]).  % too slow in SICStus
	write('('),
	print(F1),
	write(' '),
	print('$spassTerms'(Ts)),
	write(')').
user:portray('$dfg'(clause(forall(V,C)),_)):-
	format("formula(forall(~p,~p)).~n",[V,'$dfg_inner'(C)]).
user:portray('$dfg'(clause(or),_)):-
	!,
	format("formula(or(false)).~n",[]).
user:portray('$dfg'(clause(C),_)):-
	format("formula(~p).~n",['$dfg_inner'(C)]).
user:portray('$dfg_inner'(X)):-
	var(X),
	!,
	write(X).
user:portray('$dfg_inner'(C)):-
	is_list(C),
	!,
	format('[~p]','$dfg_args'(C)).
user:portray('$dfg_inner'(C)):-
	C=..[O|Args],
	!,
	(Args==[] ->
	    format('~k',[O])
	;
	    format('~k(~p)',[O,'$dfg_args'(Args)])
	).
user:portray('$dfg_args'([Arg|Args])):-
	(Args==[] ->
	    format('~p','$dfg_inner'(Arg))
	;
	    format('~p,~p',['$dfg_inner'(Arg),'$dfg_args'(Args)])
	).

user:portray(saturate_error(A,B)):-
	( atom(A)-> A1=A
	;   is_list(A) ->name(A1,A)),
	format(A,B).


	      
dfg_predefined([
    '3TAP'/_, 'KIV'/_, 'LEM'/_, 'PROTEIN'/_, 'SATURATE'/_, 'SETHEO'/_,
    'SPASS'/_, 'and'/_, 'author'/_, 'axioms'/_, 'begin_problem'/_,
    'clause'/_, 'cnf'/_, 'conjectures'/_, 'date'/_, 'description'/_,
    'dnf'/_, 'end_of_list'/_, 'end_problem'/_, 'equiv'/_, 'exists'/_,
    'false'/_, 'forall'/_, 'formula'/_, 'functions'/_, 'hypothesis'/_,
    'implied'/_, 'implies'/_, 'list_of_clauses'/_, 'list_of_declarations'/_,
    'list_of_descriptions'/_, 'list_of_formulae'/_,
    'list_of_general_settings'/_, 'list_of_proof'/_, 'list_of_settings'/_,
%----Need to () not because op(900,fy,(not)) and op(400,yfx,(/)).
    'list_of_symbols'/_, 'logic'/_, 'name'/_, ('not')/_, 'operators'/_,
    'or'/_, 'predicate'/_, 'predicates'/_, 'quantifiers'/_, 'resolution'/_,
    'sorts'/_, 'status'/_, 'subsort'/_, 'true'/_, 'version'/_]).




% eof -------------------------------------------------------------------------
:-write(end).

