:-module(expand,[expand/3,st_prove/1]).



/* 

For experimentation with Abrial's set theory calculus.
The automatic typing is not perfect.
In some examples one might have to introduce additional equality
predicates for different types


*/


:-use_module(library(basics),[member/2]).
:-use_module(global,[cont/2,incr/1,store/2]).
:-use_module(auxiliary,[appendAtoms/3]).




expand(X,X):-
	var(X),
	!.
expand(T,S):-
	T=..[F|Args],
	maplist(expand,Args,Args1),
	T1=..[F|Args1],
	(def(T1=T2) ->
	     expand(T2,S)
	;
	     S=T1
	).


norm(X,X):-
	var(X),
	!.
norm(app(F,X),FX):-
	atom(F),
	!,
	norm(X,X1),
	FX=..[F,X1].
norm(subst(P,X,X),P1):-
	norm(P,P1).
norm(all(X,P),P1):-
       subsumes_chk((X1,X2),X),
       !,
       X=(X1,X2),
       norm(all(X1,all(X2,P)),P1).
norm(exists(X,P),P1):-
       subsumes_chk((X1,X2),X),
       !,
       X=(X1,X2),
       norm(exists(X1,exists(X2,P)),P1).
norm(T,S):-
	T=..[F|Args],
	maplist(norm,Args,Args1),
	S=..[F|Args1].





exp(S,T):-
	store(varCtr,0),
	expand(S,S1),
	norm(S1,S2),
	myterms:skolemized(S2,SS,Subst),
	maplist(newvar,Subst),
	myterms:skolemized(T,SS,Subst).



new(X,X1):-
	incr(varCtr),
	cont(varCtr,N),
	appendAtoms([X,N],'$',X1).


newvar(_V-X):-
	new(x,X).



st_prove(T):-
	exp(T,Th),
	format('~2n~p expanded to ~n   ~p.~n',[T,Th]),
	user:prove(all(x,all(y,all(u,all(v,(x,u)=(y,v) => x=y and u=v)))) => Th).

/*
		   classic,[],[option(
(select(E,[A,_S]):-
        (E=(U=(X,Y));E=((X,Y)=U)),
	member(E,A),
	var(U)))]).

*/


/* -----------------------------------------------------------------------


           Expansion rules


--------------------------------------------------------------------------*/




def(eq1(S,T) = all(X,member(X,S) <=> member(X,T))).
def(eq(S,T) = all((X,Y),member((X,Y),S) <=> member((X,Y),T))).

def(member(((A,B),(C,D)),par(S,T)) = (member((A,C),S) and member((B,D),T))).
def(member((X,Y),(S;T)) = exists(Z,member((X,Z),S) and member((Z,Y),T))).
def(member(X,union(S,T)) = (member(X,S) or member(X,T))).
def(member(X,intersection(S,T)) = (member(X,S) and member(X,T))).
def(member((A,B),product(S,T)) = (member(A,S) and member(B,T))).
def(member((X,Y),id(T)) = (X=Y and member(X,T))).
def(member(Y,image(F)) = exists(X,member((X,Y),F))).
def(member(Y,image(F,S)) = exists(X,member(X,S) and member((X,Y),F))).
def(member(X,set(Y,P)) = subst(P,Y,X)).


def(rel(R,D,C)  = all((X,Y),(member((X,Y),R) => member(X,D) and member(Y,C)))).
def(total(F,D)  = all(X,member(X,D) => exists(Y,member((X,Y),F)))).
def(unique(F)   = all((X,Y,Z),member((X,Y),F) and member((X,Z),F) => X=Z)).
def(pfunc(F,D,C) =  (rel(F,D,C) and unique(F))).
def(dom(F)       = set(X,exists(Y,member((X,Y),F)))).
def(injective(F) = all((X,Y,Z),member((X,Z),F) and member((Y,Z),F) => X=Y)).
def(func(F,D,C)  = (pfunc(F,D,C) and total(F,D))).
def(singleton(X) = set(Y,X=Y)).




/*  Example theorems

st_prove(

rel(p,s,t) and rel(q,s,t) 
and func(f,t,b) and injective(f) =>
                eq((p;f),(q;f)) => eq(p,q)
	).


st_prove(pfunc(f,s,t) and pfunc(g,s,t) and all(X,member(X,intersection(dom(f),dom(g))) => all(Y,all(Z, member((X,Y),f) and member((X,Z),g) => Y=Z))) =>
       pfunc(union(f,g),s,t)).


st_prove(rel(p,s,s) and rel(r,t,s) =>
     eq((r;p),image(par(id(t),p),r))).

st_prove(func(f,s,t) => all(X,member(X,s) => exists(Y,member((X,Y),f) and eq1(image(f,singleton(X)),singleton(Y))))).

*/
