% -*- Module: input; -*-

%							INPUT
%declarations%

:- module(input, [ 
	analyse_clause/3,
	associativity_clause/4,
	collect_operators/2,
	commutativity_clause/4,
	create_clauses/0,
	create_clause_from_temp/1,
	parse/2,
	permutativity_clause/5,
	read_file/3,
	read_file_name/1,
	sat_meta_predicate/1,
	theory_presentation/2
   ]).





:- use_module(dfg_parser, [ % <DFG>
        read_file_as_list/2,
        parse_dfg_tail/2,
        saturate_info/8
   ]).




:- use_module(myterms, [
	is_list_of_different_vars/1,
	is_var_set/1,
	list_of_vars/2,
	occurs_in/2
   ]).

:- use_module(mylists, [
        is_list/1,
	nth1/4,
	permutation/2
   ]).

:- use_module(global,[cont/2]).

:- use_module(hgstuff,[split/4]).


:- use_module(props, [
	option/1,
	ordering/1,
	ordering/2,
	congruence_relation_symbol/1,
	function_as_predicate/5,
%	predicate_with_extension/2,
	special_relation/1,
	special_relations/1,
	symmetric_relation_symbol/1
   ]).




:- use_module(library(change_arg), [
	change_arg/5
   ]).

:- use_module(library(sets), [
	select/3
   ]).

:- use_module(library(maplist), [
	maplist/2,
	maplist/3
   ]).

:- use_module(library(basics), [
	member/2
   ]).

:- use_module(library(flatten), [
	flatten/2
   ]).

:- use_module(proofState, [
	clause/1,
	newClauseNumber/1,
	newModuleNumber/1,
	clause_counter/1,
	operators/1,
	predicates/1,
	props/1,
	recordClause/1,
	recordPredicates/1,
	recordPredicate/2,
	recordProofOfClause/1,
	recordFirstPredicatePrecedence/1,
	recordRule/2,
	recordOperator/2,
	recordOperators/1,
	recordPrecedence/1,
	recordProp/1,
	rule/2,
	tempClause/1,
	recordTempClause/1,
	deleteAllTempClauses/0,
	deleteOperators/1
   ]).

:- use_module(output, [
	store_props/0
   ]).

:- use_module(adt_term, [
	binder/4,
	binder/5,
	external_binder/4,
	deBruijn/2,
	deBruijn_index/1,
	logical_symbol/1
   ]).

:- use_module(adt_atom, [
	create_atom/2,
	get_atom/4,
	get_atom_members/3,
	get_atom_relation_symbol/2,
	is_atom/1
   ]).

:- use_module(clause_redundancy_prover, [
	sequent_calculus_expansion/2
   ]).


:- use_module(classes, [
        antecedent/2,
	succedent/2,
	indexEntries/2,
	indexEntries/3,
	is_new/1,
	is_old/1,
	literal_list_to_sequent/3,
	moduleNumber/3,
	makeNew/2,
	createClause/2,
	kind/2,
	operators/2,
	history/3,
	kind/3,
	number/3,
	theory_flag/2
   ]).

:- use_module(adt_precedence, [
        generated_proposition/1,
	is_operator/1,
	is_predicate/1,
	is_skolem_function/1,
	record_generated_proposition/1,
	record_skolem_function/1
   ]).

:- use_module(adt_proof, [
	initial_history/2
   ]).


:- use_module(equationalTheory, [
	match/2,
	equal/2
   ]).


:- use_module(auxiliary, [
        appendAtoms/3,
	delete_all/3,
	get_prolog_data_operators/2,
	less_or_equal/2,
	member212/2
   ]).

:- use_module(orderConstraints, [
	and_eq/3,
	and_ge/3,
	and_gt/3,
	false/1,
	or/3,
	true/1
   ]).

:- use_module(library(arg), [
	genarg/3
   ]).


:- use_module(thisProlog, [
	file_exists/1
   ]).

:- use_module(commands, [command_error/2]).

%end%


%______________________________________________________________________________
%
% input clause ::-
%	[antecedent,succedent] | [antecedent,succedent,input_constraint]
%
% antecedent, succedent = [input_atom]
% input_constraint ::- [[[input_inatom],[input_atom]]]
%				representing a disjunctive normal form
% input_atom ::- term=term | term
% input_atom_ ::- term=term
% input_inatom ::-| term>term | term>=term
%
%______________________________________________________________________________




read_file_name(F):-
	seen,
	write('File: '),
	read(F).



read_file(_Saved,[OF,F],Context):-
	!,
	(file_exists(OF) ->
	     parse(OF,Context),
	     parse(F,Context)
	;
	     command_error("file does not exist: ~p",[OF])
	).
read_file(Saved,Th,Context):-
	theory_presentation(Th,F),
	(name(F,FN),append(FBN,".dfg",FN),name(FB,FBN) ->
	    FBase=FB
	;
	    FBase=F
	),
	appendAtoms([FBase,'.o'],'',FO),
	(Saved=saved,file_exists(FO) ->
	     F1=FO
	;
	     appendAtoms([FBase,'.ord'],'',OF),
	     (Saved\==only_clauses,file_exists(OF) ->
		  parse(OF,Context)
	     ;
	          true
	     ),
	     F1=Th
	),

	(file_exists(F1) ->
	     parse(F1,Context)
	;    
	     command_error("theory does not exist: ~p",[F1])
	).



theory_presentation(Th,_):-
	var(Th),
	!,
	fail.
theory_presentation(Th,Th):-
	atom(Th).
theory_presentation(th(Name,Clauses),Name):-
	atom(Name),
	is_list(Clauses).
theory_presentation(th(Name),Name):-
	atom(Name),
	cont(theory(Name),_).


parse(F,Context):-
	atom(F),
	open(F,read,S),
	length(Context,N),
	M is (N-1)*2,
	format("~n~*c[reading ~p",[M,32,F]),ttyflush,
	read_theory(S, Input, S1 ),
	parse(S1, F, Context, Input ),
	write(']'),ttyflush,
	close(S).
parse(S,Context):-
	is_list(S),
	read_theory(S, Input, S1 ),
	parse(S1, user, Context, Input ).



read_theory([C|Cs],C1,Cs):-
	!,
	copy_term(C,C1).
read_theory([],end_of_file,[]):-
	!.
read_theory(Str,First,Str):-
	read(Str,First).

	
parse([],_F, _M, 'end_of_file'):-
	!.
parse(_Str, _F, _M, 'end_of_file' ):-
	!.
parse(Str, F, _M, begin_problem(_ProblemLabel) ):-                 % <DFG: parse dfg-files>
	!,
        read_file_as_list(Str,FileTailList),                        % read file
	(parse_dfg_tail(FileTailList,ParsedFile) ->                   % parse content with DCG's
	    th_line_count(Str,Line),
	    saturate_info(ParsedFile,_Pred,_Func,Ax,Hyp,Goals,ThF,GoF), % saturate relevant information
	     ((ThF = [], GoF = []) ->                                    % if no formulae
			on_exception(input_error(S,P),
			      treat_input_terms(Ax,theory),
			      input_error(Ax,F,Line,S,P)
			     ),
			on_exception(input_error(S,P),
			      treat_input_terms(Hyp,theory),
			      input_error(Ax,F,Line,S,P)
			     ),
			on_exception(input_error(S,P),
			      treat_input_terms(Goals,(goal)),
			      input_error(Ax,F,Line,S,P)
			     )
	     ;
		 on_exception(input_error(S,P),
			      treat_input_terms(ThF, theory), % but prefer formulae if existen
			      input_error(ThF,F,Line,S,P)
			     ),
		 on_exception(input_error(S,P),
			      treat_input_terms(GoF, (goal)),
			      input_error(ThF,F,Line,S,P)
			     )
	     )
	;   
	th_line_count(Str,Line),
	input_error(begin_problem(_ProblemLabel),F,Line,[],[])
	).
parse(Str, F, Context, Input ):-
	th_line_count(Str,Line),
	on_exception(input_error(S,P),
		     treat_input( Input, Context, theory),
		     input_error(Input,F,Line,S,P)
		    ),
	read_theory( Str, Input1, Str1 ),
	parse( Str1, F, Context, Input1 ).


treat_input_terms([Term|Tail],Type) :- % <DFG: treat a list of clauses>
	treat_input(Term, _, Type),
	treat_input_terms(Tail,Type).
treat_input_terms([],_).




th_line_count(Str,L1):-
	is_list(Str),
	!,
	length(Str,L),
	L1 is -L-1.
th_line_count(Str,L):-
	line_count(Str,L).


%------------------------------------------------------------------------------
%	treat input of axioms, clauses and rules
%	precedences, predicates, ...
%------------------------------------------------------------------------------




treat_input( P, _, T ):-
	var(P),
	!,
	treat_input( axiom([[],[P]]), _, T).

treat_input(all(_X,B),M,T):-
	treat_input(B,M,T).

treat_input((A and B =>S),M,T):-
	!,
	treat_input(((A,B)->S),M,T).

treat_input((A=>S),M,T):-
	treat_input((A->S),M,T).



treat_input(Disj,M,T):-
	is_disjunction(Disj,Disjuncts),
	!,
	split(is_negation,Disjuncts,NegNeg,Pos),
	maplist(complement,NegNeg,Neg),
	treat_input(axiom([Neg,Pos]),M,T).


treat_input('$select'(A,B),M,T):-
	treat_input(option(select(A,B)),M,T).

treat_input(('$select'(A,B):-Bo),M,T):-
	treat_input(option((select(A,B):-Bo)),M,T).



treat_input(T,_, _):-
	T=..[':-',Goal],     % to avoid possible term expansion problems
	(user:Goal ->
	     true
	;
	     raise_exception(input_error('Execution of goal ~p failed.',[Goal]))
	),
	!.
treat_input( option(O),_, _ ):-
	option(O).

treat_input(use(F),Context,_):-
	(member(module(F,_),Context) ->
	     true
	;
	     newModuleNumber( M ),
	     read_file(saved,F,[module(F,M)|Context])
	).
	     
treat_input(clause( C ), _, _ ):-
	clause_counter( N ),
	antecedent(C,A),
	succedent(C,S),
	append(A,S,AS),
	theory_flag(C,Th),
	type_check_clause_literals(Th,AS),
	number( N, C, C1 ),
	initial_history( theory, H ),
	history( H, C1, C2 ),
	recordProofOfClause(C2),
	recordClause( C1 ),
	newClauseNumber(N).
treat_input( rule( R ), M, _ ):-
	antecedent(R,A),
	succedent(R,S),
	append(A,S,AS),
	theory_flag(R,Th),
	type_check_clause_literals(Th,AS),
	preprocess_rule(M,R).

treat_input( precedence( L ), _, _ ):-
	recordPrecedence( L ).

treat_input( first_predicate_precedence( L ), _, _ ):-
	recordFirstPredicatePrecedence( L ).

treat_input( top_predicates_precedence( L ), _, _ ):-  % new terminology
	recordFirstPredicatePrecedence( L ).

treat_input( predicates( _L ), _, _ ).
%	recordPredicates( _L ).  % obsolete

treat_input(ordering(O), _, _):-
	ordering(O).

treat_input( ordering(O,S), _, _):-
	ordering(O,S).

treat_input(reflexive_closure(A,B), M, T ):-
	(atom(A),atom(B),A\==B ->
	     At=..[A,X,Y],
	     Bt=..[B,X,Y],
	     treat_input((At->Bt),M,T),
	     treat_input((Bt->At,X=Y),M,T)
	;
	     raise_exception(input_error('Reflexive closure declaration ~p:~n    ~p and ~p must be different atoms~2n',[reflexive_closure(A,B),A,B]))
	).
treat_input(partial_congruence(R),_,T):-
	create_clause_from_input_axiom( T,partial_congruence(R) , Cl ),
        recordTempClause( Cl ).

treat_input(goal(Cl), M, _ ):-
	treat_input(Cl, M,(goal)).

treat_input( axiom( InAx ), _, T  ):-
	create_clause_from_input_axiom( T, InAx, Cl ),
        recordTempClause( Cl ).

treat_input( L=R, M, T ) :-
	treat_input( axiom([[],[L=R]]), M, T ).

treat_input( L==R, M, T ) :-
	treat_input( axiom([[],[L==R]]), M, T ).

treat_input( (A ->[]), M, T ):-
	!,
	pairs_to_list(A,AL),
	treat_input( axiom([AL,[]]), M, T ).

treat_input( (A ->S), M, T ):-
	pairs_to_list(A,AL),
	pairs_to_list(S,SL),
	treat_input( axiom([AL,SL]), M, T ).

treat_input( ((A -> []) if Gamma), M, T ):-
	!,
	pairs_to_list(A,AL),
	treat_input( axiom([AL,[],[[Gamma,[]]]]), M, T ).

treat_input( ((A -> S) if Gamma), M, T ):-
	!,
	pairs_to_list(A,AL),
	pairs_to_list(S,SL),
	treat_input( axiom([AL,SL,[[Gamma,[]]]]), M, T ).

treat_input( ([] <- A if Gamma), M, T ):-
	!,
	pairs_to_list(A,AL),
	treat_input( axiom([AL,[],[[Gamma,[]]]]), M, T ).

treat_input(([] <- A), M, T ):-
	!,
	pairs_to_list(A,AL),
	treat_input( axiom([AL,[]]), M, T).

treat_input( (S <- A if Gamma), M, T ):-
	!,
	pairs_to_list(A,AL),
	pairs_to_list(S,SL),
	treat_input( axiom([AL,SL,[[Gamma,[]]]]), M, T ).

treat_input( (S <- A), M, T ):-
	pairs_to_list(A,AL),
	pairs_to_list(S,SL),
	treat_input( axiom([AL,SL]), M, T ).

treat_input( (P,Q), M, T ) :-
	pairs_to_list((P,Q),SL),
	!,
	treat_input( axiom( [[], SL] ), M, T ).

% Otter-like notation
treat_input((P;Q), M, T) :-
	disj_to_sequent((P;Q),A,S),
	treat_input(  axiom([A,S]), M, T ).

treat_input(-P, M, T) :-
	treat_input( axiom([[P],[]]), M, T ).

% TPTP notation

treat_input(input_clause(_Name,Kind,Cl),M,_T):-
	(Kind=axiom ->
	     T1=theory
	;
	     T1=(goal)
	),
	maplist(tptp_literal_to_literal,Cl,Lits),
	literal_list_to_sequent(Lits,A,S),
	treat_input(  axiom([A,S]), M, T1 ).

treat_input(input_formula(_Name,Kind,F),M,_T):-
	(Kind=axiom ->
	     T1=theory,
	     F1=F
	;Kind=conjecture ->
	     T1=(goal),
	     F1=(~F)
	;
	     T1=(goal),
	     F1=F
	),
	tptp_formula_conv(F1,FC),
	treat_input(  FC , M, T1 ).

treat_input(include(F),Context,_):-
	treat_input(use(F),Context,_).

/*
treat_input(include(F),Context,_):-
	(name(F,S),
	 append(_,".eq",S) ->
		    true
	;
	            treat_input(use(F),Context,_)
	).
*/
	

treat_input( (P if Gamma), M, T ):-
	!,
	treat_input( axiom([[],[P],[[Gamma,[]]]]), M, T ).

treat_input( P, M, T ):-
	treat_input( axiom([[],[P]]), M, T ).

treat_input( L, _, _):-
	raise_exception(input_error('Error in input file: ~p',[L])).



is_disjunction(D,Disjuncts):-
	D=..[or|Or],
	maplist(is_disjunction_or_formula,Or,LDisjuncts),
	flatten(LDisjuncts,Disjuncts).
is_disjunction(~A,[~A]).



is_disjunction_or_formula(A,AD):-
	!,
	(is_disjunction(A,AD) ->
	     true
	;
	AD=[A]).

	


is_negation(~_).

complement(~A,A):-
	!.
complement(A,A).


:-op(200,fx,++).
:-op(200,fx,--).


tptp_literal_to_literal(++(equal(A,B)),[+|A=B]):- 
	cont(tptpspecials,true),
	!.
tptp_literal_to_literal(--(equal(A,B)),[-|A=B]):- 
	cont(tptpspecials,true),
	!.
tptp_literal_to_literal(++(P),[+|P]).
tptp_literal_to_literal(--(P),[-|P]).




tptp_formula_conv( (! _Vs : (L<=>R)), L1==R1):-
	% special case to catch definitions
	tptp_formula_conversion(L,L1),
	tptp_formula_conversion(R,R1),
	!.
tptp_formula_conv( F, F1):-
	tptp_formula_conversion(F,F1).



tptp_formula_conversion(X,X):-
	var(X),
	!.
tptp_formula_conversion(X&Y,X1 and Y1):-
	tptp_formula_conversion(X,X1),
	tptp_formula_conversion(Y,Y1),
	!.
tptp_formula_conversion((X|Y),X1 or Y1):-
	tptp_formula_conversion(X,X1),
	tptp_formula_conversion(Y,Y1),
	!.
tptp_formula_conversion((X<=>Y),X1==Y1):-
	tptp_formula_conversion(X,X1),
	tptp_formula_conversion(Y,Y1),
	!.
tptp_formula_conversion((X=>Y),X1=>Y1):-
	tptp_formula_conversion(X,X1),
	tptp_formula_conversion(Y,Y1),
	!.
tptp_formula_conversion(equal(X,Y),X1=Y1):-
	tptp_formula_conversion(X,X1),
	tptp_formula_conversion(Y,Y1),
	!.
tptp_formula_conversion(~X,~X1):-
	tptp_formula_conversion(X,X1),
	!.
tptp_formula_conversion( (! [V] : X), all(V,X1)):-
	tptp_formula_conversion(X,X1),
	!.
tptp_formula_conversion( (! [V|Vs] : X), all(V,X1)):-
	tptp_formula_conversion((! Vs : X), X1),
	!.
tptp_formula_conversion( (? [V] : X), exists(V,X1)):-
	tptp_formula_conversion(X,X1),
	!.
tptp_formula_conversion( (? [V|Vs] : X), exists(V,X1)):-
	tptp_formula_conversion((? Vs : X), X1),
	!.
tptp_formula_conversion(X,X).





preprocess_rule(M,R):-
	clause_counter( N ),
	number( N, R, R1 ),
	initial_history( theory, H ),
	history( H, R1, R2 ),
	recordProofOfClause(R2),
%	( ( is_new( R ), moduleNumber( _, R2, R4 ) )
%	;
%	  ( is_old( R ),
	    M=[module(_,ModN)|_],
	    moduleNumber( ModN, R2, R3 ),
	    makeNew( R3, R4 )
%	  )
%	)
	,
	indexEntries(_,R4,R5),  % invalidation; will be recomputed upon
	                        % next call to sat.
	                        % Should not be recomputed earlier since
	                        % it relies on the theory analysis
%	indexEntries(R5,_E),     % recomputes them
	recordRule( N,R5 ),
	newClauseNumber(N).



predicate_or_meta_predicate(R):-
	is_predicate(R),
	!.
predicate_or_meta_predicate(R):-
	sat_meta_predicate(R).
	
sat_meta_predicate(= /2).
sat_meta_predicate(== /2).
sat_meta_predicate('$true$'/0).
sat_meta_predicate('$min'/0).
%sat_meta_predicate('$all'/2).
%sat_meta_predicate('$exists'/2).
%sat_meta_predicate(and/2).
%sat_meta_predicate(or/2).
%sat_meta_predicate(=> /2).
%sat_meta_predicate(<=> /2).






create_clauses:-
	store_props,
	tempClause(C),
	create_clause_from_temp(C),
	fail.
create_clauses:-
	deleteAllTempClauses.



pairs_to_list((A,B),L):-
	!,
	pairs_to_list(A,AL),
	pairs_to_list(B,BL),
	user:append(AL,BL,L).
pairs_to_list(X,[X]).

disj_to_sequent((P;Q),A,S):-
	nonvar(P),
	nonvar(Q),
	disj_to_sequent1((P;Q),A,S).


disj_to_sequent1((-P;Q),[P|A],S):-
	!,
	disj_to_sequent1(Q,A,S).
disj_to_sequent1((P;Q),A,[P|S]):-
	!,
	disj_to_sequent1(Q,A,S).
disj_to_sequent1(-P,[P],[]):-
	!.
disj_to_sequent1(P,[],[P]).




%------------------------------------------------------------------------------
% control operator definition
%------------------------------------------------------------------------------
declare_operators( Ops ):-
	(deleteOperators( Ops1 ) ->
	     user:append( Ops, Ops1, Ops2 )
	;    Ops2=Ops),
	sort( Ops2, Ops3 ),
	recordOperators( Ops3 ).	




create_clause_from_input_axiom( T, [A,S|R], C3 ):-
	!,
	user:append(A,S,AS),
%	type_check_clause_literals(T,AS),
	(R=[] -> 
	     analyse_clause(A,S,Kind),
	     C3 = ( [T,A1,S1], Kind )
	; 
	     R=[Ct],
	     create_constraint( Ct, Ct1 ),
	     Kind=regular,
	     C3 = ( [T,A1,S1,Ct1], Kind )
	),
	(Kind=monotonicity(_,_,_,_,_) ->
	     type_check_clause_literals(logical,AS)
	;
	     type_check_clause_literals(T,AS)
	),
	maplist(deBruijn,A,A1),
	maplist(deBruijn,S,S1).

	
create_clause_from_input_axiom(_T,Nonstandard,(Nonstandard,Nonstandard)).

type_check_clause_literals(_,[]).
type_check_clause_literals(Th,[L|Ls]):-
	nonvar(L),
	(is_atom(L) ->
	     get_atom(L,RS,P,Q)
	;
	     create_atom(L,A),
	     get_atom(A,RS,P,Q)
	),
	(RS= == ->
	     type_check_atom(Th,P),
	     type_check_atom(Th,Q)
	;
	     predicate_symbol(RS/2,Th),
	     type_check_term(Th,P),
	     type_check_term(Th,Q)
	),
	!,
	type_check_clause_literals(Th,Ls).



type_check_term(_,X):-
	var(X),
	!.
type_check_term(Th,(P@T)):-
	!,
	type_check_term(Th,P),
	type_check_term(Th,T).
type_check_term(Th,T):-
	(   binder(T,_Var,Body,B)
	;   binder(T,_,Var,Body,B)
	;   external_binder(T,Var,Body,B)
	),
	!,
	(nonvar(Var) ->
	     raise_exception(input_error('~p should be a variable in binding construct ~p.',[Var,T]))
	;
	     true
	),
	(B\==lambda ->
	     raise_exception(input_error('Binder ~p not allowed in term ~p.',[B,T]))
	;
	     true
	),
	type_check_term(Th,Body).
type_check_term(Th,T):-
	functor(T,F,N),
	T=..[F|Ts],
	function_symbol(F/N,Th),
	maplist(type_check_term(Th),Ts).

%type_check_atom(_,_):-
%	!.
type_check_atom(_,X):-
	var(X),
	!.
%	raise_exception(input_error('Atomic expression ~p should not be a variable.',[X])).
type_check_atom(Th,(P@T)):-
	type_check_term(Th,P),
	type_check_term(Th,T).
type_check_atom(Th,T):-
	functor(T,P,N),
	T=..[P|Ts],
	logical_symbol(P/N),
	\+ P/N= (=)/2,
	!,
	maplist(type_check_atom(Th),Ts).
type_check_atom(Th,T):-
	(binder(T,_,Var,Body,_);external_binder(T,Var,Body,_)),
	!,
	(nonvar(Var) ->
	     raise_exception(input_error('~p should be a variable in binding construct ~p.',[Var,T]))
	;
	     true
	),
	type_check_atom(Th,Body),
	( binder(T,SkT,_Var,_Body,_B) ->
	    type_check_term(Th,SkT)
	;
	    true
	).
type_check_atom(Th,T):-
	binder(T,_Var,Body,_),
	!,
	type_check_atom(Th,Body).
type_check_atom(Th,T):-
	functor(T,P,N),
	T=..[P|Ts],
	predicate_symbol(P/N,Th),
	maplist(type_check_term(Th),Ts).

function_symbol('$min'/0,_):-
	!.
function_symbol(F/N,Th):-
	(predicate_or_meta_predicate(F/_N) ->
	     raise_exception(input_error('~p was parsed as predicate symbol and now occurs as function symbol.',[F]))
	;
	     (is_skolem_function(F/N) ->
		 record_skolem_function(F)  % don't reuse them
	     ;deBruijn_index(F) ->
		  true
	     ;
	          recordOperator(F/N,Th)
	     )
	).

predicate_symbol(= /2,_):- !.
predicate_symbol('$true$'/0,_):- !.
predicate_symbol('$min'/0,_):- !.
predicate_symbol(F/N,Th):-
	(is_operator(F/_N) ->
	     raise_exception(input_error('~p was parsed as function symbol and now occurs as predicate symbol.',[F]))
	;
	     (generated_proposition(F/N) ->
		  record_generated_proposition(F)  % don't reuse them
	     ;
	          recordPredicate(F/N,Th)
	     )
	).



create_clause_from_temp(([T,A,S|R],Kind)):-
	normalize_clause(A,S,Kind,A1,S1),
	createClause([A1,S1|R],C1),
	newClauseNumber( N ),
	number( N, C1, C2 ),
	initial_history( T, H ),
	history( H, C2, C3 ),
	kind(Kind,C3,C4),
	recordClause(C4),
	!.


normalize_clause(A,S,Kind,A2,S2):-
	maplist(literal_as_atom(-,Kind), A, A1 ),
	maplist(literal_as_atom(+,Kind), S, S1 ),
	normalize_monotonicity_clause(A1,S1,Kind,A2,S2).


normalize_monotonicity_clause(A1,S1,Kind,A2,S2):-
	(special_relations(on),
	 Kind=monotonicity(_,_,R,1,==),
	 symmetric_relation_symbol(R) ->
	     select(E1,A1,[E2]),
	     get_atom_relation_symbol(E1,R),
	     S1=[E3],
	     get_atom_members(E3,L3,R3),R3=='$true$',
	     get_atom_members(E2,L2,R2),R2=='$true$',
	     create_atom(L2==L3,E4),
	     A2=[E1],
	     S2=[E4]
	;
	     A2=A1,
	     S2=S1
	),
	!.
normalize_monotonicity_clause(A,S,_Kind,A,S).
	     

	     

literal_as_atom(Pol,_,T,A):-
	T=..[P|As],
	function_as_predicate(P,N,I,F,R),
	length(As,N),
	!,
	nth1(I,As,X,As1),
	FT=..[F|As1],
	(var(X), Pol= -, \+ occurs_in(X,FT) ->
	     X=FT,
	     create_atom('$true$',A)
	;
	     E=..[R,X,FT],
	     create_atom(E,A)
	).

literal_as_atom(_,_,E,A):-
	create_atom(E,A).

	


%------------------------------------------------------------------------------
% create_constraint( DNF, Ct ):
% 	creates a constraint Ct from an input_constraint DNF wich is 
%	a disjunctive normal form of type [[input_atom_disatom]]
%	where input_atom_disatom ::- term=term | term>term | term>=term
%------------------------------------------------------------------------------

create_constraint( [], Ct ):-
	false( Ct ), !.
create_constraint( [[AI,AE]|DNF], Ct2 ):-
	user:append( AI, AE, A ),
	create_and( A, And ),
	create_constraint( DNF, Ct1 ),
	or( And, Ct1, Ct2 ), !.

create_and( [], Ct ):-
	true( Ct ), !.
create_and( [S>T|R], Ct2 ):-
	create_and( R, Ct1 ),
	and_gt( Ct1, [S,T], Ct2 ), !.
create_and( [S>=T|R], Ct2 ):-
	create_and( R, Ct1 ),
	and_ge( Ct1, [S,T], Ct2 ), !.
create_and( [S=T|R], Ct2 ):-
	create_and( R, Ct1 ),
	and_eq( Ct1, [S,T], Ct2 ), !.
	
%---------------------------------------------------
%	collect operators of axioms, clauses and rules
%---------------------------------------------------
	
collect_operators( clause( C ), L ):-
	operators( C, L ), !.
collect_operators( rule( R ), L  ):-
	operators( R, L ), !.
collect_operators( axiom( Ax ), L  ):-
	get_input_axiom_operators( Ax, L ), !.
collect_operators( _, [] ).

%---------------------------------------------------
%	axiom operators
%---------------------------------------------------

get_input_axiom_operators( [A,S|_Ct], L ):-
	get_prolog_data_operators( [A,S], L ).
%	get_prolog_data_operators( Ax, L0 ),
%	delete_all( == /2, L0, L1 ),
%	delete_all( = /2, L1, L ).
% this used to be valid for constraints in input clauses
%	delete_all( >, L2, L3 ),
%	delete_all( >=, L3, L ), !.

%---------------------------------------------------
%	clause analysis
%       transitivity, reflexivity, irreflexivity
%       are recognized
%---------------------------------------------------


analyse_clause(A,S,anti_symmetry(R,ER)):-
	anti_symmetry_clause(A,S,R,ER),
	!.
analyse_clause(A,S,irreflexivity(R)):-
	irreflexivity_clause(A,S,R),
	!.
analyse_clause(A,S,reflexivity(R)):-
	reflexivity_clause(A,S,R),
	!.
analyse_clause(A,S,composition(R1,R2,R3)):-
	composition_clause(A,S,R1,R2,R3),
	!.
analyse_clause(A,S,symmetry(R)):-
	symmetry_clause(A,S,R),
	!.
analyse_clause(A,S,symmetry(R,N,I,J)):-
	symmetry_clause(A,S,R,N,I,J),
	!.
analyse_clause(A,S,nonstrict_totality(R)):-
	nonstrict_totality_clause(A,S,R),
	!.
analyse_clause(A,S,totality(R,ER)):-
	totality_clause(A,S,R,ER),
	!.
analyse_clause(A,S,reflexive_closure1(P,Q)):-
	binary_subset_clause(A,S,P,Q).
analyse_clause(A,S,reflexive_closure2(P,Q)):-
	binary_subset_wo_diag_clause(A,S,P,Q).
analyse_clause(A,S,D):-
	density_clause(A,S,D),
	!.
analyse_clause(A,S,extensionality(P,Args,SF)):-
	extensionality_clause(A,S,P,Args,SF),
	!.
analyse_clause(A,S,noendL(R)):-
	noendL_clause(A,S,R),
	!.
analyse_clause(A,S,noendR(R)):-
	noendR_clause(A,S,R),
	!.
analyse_clause(A,S,monotonicity(F,I,R,P,R1)):-
	monotonicity_clause(A,S,F,R,I,P,R1),
	!.
analyse_clause(A,S,commutativity(O,R)):-
	commutativity_clause(A,S,O,R),
	!.
analyse_clause(A,S,associativity(O,R)):-
	associativity_clause(A,S,O,R),
	!.
analyse_clause(A,S,permutativity(O,X,Y)):-
	permutativity_clause(A,S,O,X,Y),
	!.
analyse_clause(A,S,injectivity(O,N,I)):-
	injectivity_clause(A,S,O,N,I),
	!.

analyse_clause(A,S,total_predicate(P,N,I,F)):-
	total_predicate_clause(A,S,P,N,I,F),
	!.

analyse_clause(A,S,det_predicate(P,N,I,R)):-
	det_predicate_clause(A,S,P,N,I,R),
	!.

analyse_clause(A,S,finite_domain(D,R)):-
	finite_domain_clause(A,S,D,R),
	!.

analyse_clause(A,S,inverse(I,M,U,R)):-
	inverse_clause(A,S,I,M,U,R),
	!.


analyse_clause(_,_,regular).





%------------------------------------------------------------------------------
%	clause analysing
%------------------------------------------------------------------------------




irreflexivity_clause([A],[],R):-
	A=..[R,X,Y],
	var(X),
	X==Y,
	!.

	
reflexivity_clause(A,S,R):-
	irreflexivity_clause(S,A,R).





composition_clause(Gamma,[C],R1,R2,R3):-
	\+ special_relations(off),
	permutation(Gamma,[A,B]),
	A=..[R1,X,Y],
	B=..[R2,Y1,Z],
	C=..[R3,X1,Z1],
	(R3=R1;R3=R2),
	var(X), var(Y), var(Z),
	Y==Y1, Z==Z1, X==X1,
	!.

totality_clause([],S,R,ER):-
	permutation(S,[A,B,C]),
	A=..[R,X,Y],
	var(X), var(Y),
	X\==Y,
	B=..[R,Y1,X1],
	Y==Y1, X==X1, 
	(C=..[ER,X2,Y2];C=..[ER,Y2,X2]),
	Y==Y2, X==X2,
	!.

nonstrict_totality_clause([],S,R):-
	permutation(S,[A,B]),
	A=..[R,X,Y],
	var(X), var(Y),
	X\==Y,
	B=..[R,Y1,X1],
	Y==Y1, X==X1, 
	!.

binary_subset_clause([A],[B],R,S):-
	A=..[R,X,Y],
	var(X), var(Y),
	X\==Y,
	B=..[S,Y1,X1],
	X==Y1, Y==X1, 
	!.

binary_subset_wo_diag_clause([A],Su,R,S):-
	select(B,Su,[C]),
	A=..[S,X,Y],
	var(X), var(Y),
	X\==Y,
	B=..[R,Y1,X1],
	Y==X1, X==Y1,
	C=(X2=Y2),
	(X2==X, Y2==Y; X2==Y, Y2==X),
	!.



symmetry_clause([A],[B],R):-
	A=..[R,X,Y],
	var(X), var(Y),
	X\==Y,
	B=..[R,Y1,X1],
	Y==Y1, X==X1, 
	!.

symmetry_clause([A],[B],R,N,I,J1):-
	% R(...XI...XJ...) -> R(...XJ...XI...)
	A=..[R|As],
	B=..[R|Bs],
	length(As,N),
	length(Bs,N),
	nth1(I,As,XI,As1),
	var(XI),
	nth1(J,As1,XJ,As2),
	var(XJ),
	nth1(I,Bs,YJ,Bs1),
	YJ==XJ,
	nth1(J,Bs1,YI,Bs2),
	YI==XI,
	As2==Bs2,
	list_of_vars(As2,Vs),
	As2==Vs,
	J1 is J+1,
	!.



anti_symmetry_clause(Gamma,[C],R,ER):-
	permutation(Gamma,[A,B]),
	A=..[R,X,Y],
	var(X), var(Y),
	X\==Y,
	B=..[R,Y1,X1],
	Y==Y1, X==X1,
	C=..[ER,X2,Y2],
	(X2==X, Y2==Y; X2==Y, Y2==X),
	!.

monotonicity_clause([E],[E1],F/N,RS,[I],Pol,RS1):-
	E=..[RS,X,Y],
	var(X), var(Y), X\==Y,
	E1=..[RS1,FX,FY],
	nonvar(FX),
	FX=..[F|Xs],
	length(Xs,N),
	is_var_set(Xs),
	(   
	    genarg(I,FX,X1),
	    X1==X,
	    change_arg(I,FX,X,FX1,Y),
	    FY==FX1,
	    Pol = 1
	;
	    genarg(I,FX,Y1),
	    Y1==Y,
	    change_arg(I,FX,Y,FX1,X),
	    FY==FX1,
	    Pol= -1
	),
	!.
monotonicity_clause(A,[E1],F/2,RS,[1,2],1,RS1):-
	% RS(X,Y),RS(U,V) -> RS1(F(X,U),F(Y,V))
	permutation(A,[A1,A2]),
	A1=..[RS,X,Y],
	A2=..[RS,U,V],
	is_list_of_different_vars([X,Y,U,V]),
	E1=..[RS1,FX,FY],
	nonvar(FX),
	FX=..[F,X1,U1],
	nonvar(FY),
	FY=..[F,Y1,V1],
	X1==X,Y1==Y,U1==U,V1==V,
	!.
monotonicity_clause(A,[E2],RS1/N,RS,[I],1,==):-
	permutation(A,[E,E1]),
	E=..[RS,X,Y],
	var(X), var(Y), X\==Y,

	functor(E1,RS1,N),RS1\== =,RS1\== ==,
	E1=..[RS1|LV1],
	is_list_of_different_vars(LV1),

	E2=..[RS1|LV2],
	is_list_of_different_vars(LV2),
	length(LV2,N),

	genarg(I,E1,X1),
	X1==X,
	change_arg(I,E1,X,E12,Y),
	E2==E12.


density_clause([A],[B],density(R,left,strict)):-
	A=..[R,X,Y],
	B=..[R,X1,F],
	var(X), var(Y),
	nonvar(F),
	F=..[_,X2,Y1],
	X==X1,Y==Y1,X1==X2,
	!.
density_clause([B],[A],density(R,left,nonstrict)):-
	A=..[R,X,Y],
	B=..[R,X1,F],
	var(X), var(Y),
	nonvar(F),
	F=..[_,Y1,X2],
	X==X1,Y==Y1,X1==X2,
	!.
density_clause([A],[B],density(R,right,strict)):-
	A=..[R,X,Y],
	B=..[R,F,Y1],
	var(X), var(Y),
	Y==Y1,
	nonvar(F),
	F=..[_,X1,Y2],
	Y1==Y2,
	X==X1,
	!.
density_clause([B],[A],density(R,right,nonstrict)):-
	A=..[R,X,Y],
	B=..[R,F,Y1],
	var(X), var(Y),
	Y==Y1,
	nonvar(F),
	F=..[_,Y2,X1],
	Y1==Y2,
	X==X1,
	!.

extensionality_clause(A,S,P/N,Args,SFs):-
	select(D,S,S1),
	D=..[P|Args],
	is_list_of_different_vars(Args),
	append(A,S1,Ls),
	length(Args,N),
	maplist(extension_call(Args,P/N,SF),Ls),
	(var(SF) ->
	     SFs=[]
	;
	     SFs=[SF]
	).



noendR_clause([],[A],R):-
	A=..[R,X,F],
	var(X),
	nonvar(F),
	F=..[_,X1],
	X==X1,
	!.

noendL_clause([],[A],R):-
	A=..[R,F,X],
	var(X),
	nonvar(F),
	F=..[_,X1],
	X==X1,
	!.




commutativity_clause([],[E],O,RS):-
	E=..[RS,L,R],
	commutativity_clause1([],[L=R],O).

commutativity_clause1([],[L=R],O):-
	nonvar(L),
	nonvar(R),
	L=..[O,A,B],
	R=..[O,B1,A1],
	A==A1,
	B==B1,
	var(A),
	var(B),
	!.

associativity_clause([],[E],O,RS):-
 	E=..[RS,L,R],
	associativity_clause1([],[L=R],O).

associativity_clause1([],[L=R],O):-
	nonvar(L),
	functor(L,O,2),
	L1=..[O,A,BC], BC=..[O,B,C],
	R1=..[O,AB,C], AB=..[O,A,B],
	(match((L,R),(L1,R1)),
	match((L1,R1),(L,R))
	;match((R,L),(L1,R1)),
	match((L1,R1),(R,L))),
	!.
	
permutativity_clause([],[L=R],O,A,B):-
	nonvar(L),
	functor(L,O,2),
	L1=..[O,A,BC], BC=..[O,B,C],
	R1=..[O,B,AC], AC=..[O,A,C],
	(match((L,R),(L1,R1)),
	match((L1,R1),(L,R))
	;match((R,L),(L1,R1)),
	match((L1,R1),(R,L))),
	!.


injectivity_clause([E],[F],O,N,I):-
	% E=(O(X1,...,X(i-1),A,X(i+1),...,Xn)=O(X1,...,X(i-1),B,X(i+1),...,Xn))
	functor(E,=,2),
	arg(1,E,L),
	arg(2,E,R),
	nonvar(L),
	nonvar(R),
	functor(L,O,N),
	functor(R,O,N),
	genarg(I,L,X),
	arg(I,R,Y),
	var(X),var(Y),X\==Y,

	same_var_args(L,R,I,N),
	I1 is I-1,
	same_var_args(L,R,0,I1),
	

	% F=(X=Y) or F=(Y=X)
	(   F1=(X=Y),
	    F==F1
	;   F1=(Y=X),
	    F==F1),
	!.


% is true when the clause states that P, a predicate of arity N,
% is total with respect to (result) argument position I, and
% F is the skolem function to be used for P

total_predicate_clause([],[A],P,N,I,F):-
	% A == P(X1,...,X(I-1),F(X1,...,X(I-1),X(I+1),...,Xn),X(I+1),...,Xn)
	A=..[P|Args],
	length(Args,N),
	append(Left,[FF|Right],Args),
	nonvar(FF),
	FF=..[F|FArgs],
	append(Left,Right,ProperPArgs),
	maplist(var,ProperPArgs),
	list_of_vars(ProperPArgs,Vs),  % |Vs|=|ProperPArgs| => no duplicates
	length(Vs,M),
	M is N-1,
	ProperPArgs==FArgs,
	length(Left,I1),
	I is I1+1.

det_predicate_clause([E1,E2],[E],P,N,I,R):-
	% P(...XI...),P(...YI...) -> R(XI,YI) ; usually: R is equality
	E1=..[P|A1],
	E2=..[P|A2],
	E=..[R,XI,YI],
	length(A1,N),
	append(L1,[UI|R1],A1),
	append(L2,[VI|R2],A2),
	var(UI),var(VI),
	(UI==XI,VI==YI;VI==XI,UI==YI),
	append(L1,R1,A11),
	append(L2,R2,A22),
	A11==A22,
	!,
	length(L1,I1),
	I is I1+1.
	


finite_domain_clause([],[E1,E2],[D1,D2],R):-
	% we treat only a special case right now
	(  E1=..[R,X1,D1]
	;  E1=..[R,D1,X1]
	),
	(  E2=..[R,X2,D2]
	;  E2=..[R,D2,X2]
	),
	var(X1),
	X1==X2,
	atomic(D1),
	atomic(D2),
	!.

inverse_clause([],[E],I,M,T,R):-
	(E=..[R,S,T];E=..[R,T,S]),
	nonvar(S),
	nonvar(T),
	atom(T),
	functor(S,M,2),
	(arg(1,S,S1),arg(2,S,S2); arg(1,S,S2),arg(2,S,S1)),
	var(S1),
	nonvar(S2),
	S2=..[I,S21],
	S21==S1,
	!.

/*

definition_clause(_A,S,P/N,I):-
	member(F,S),
	positive_subformula(F,Sub),
	functor(Sub,==,2),
	arg(1,Sub,Left),
	arg(2,Sub,Right),

*/

extension_call(Args,P/N,F,L):-
	\+ functor(L,P,N),
	L=..[_|PArgs],
	maplist(extension_call_arg(Args,F),PArgs),
	nonvar(F).


extension_call_arg(Args,_,T):-
	member212(T,Args),
	!.
extension_call_arg(Args,F,T):-
	nonvar(T),
	T=..[F|Args1],
	Args==Args1.




	

same_var_args(_L,_R,N,N):- !.
same_var_args(L,R,I,N):-
	I1 is I+1,
	arg(I1,L,LI1),
	arg(I1,R,RI1),
	var(LI1),
	LI1==RI1,
	same_var_args(L,R,I1,N).


is_predicate_symbol(=).
is_predicate_symbol(==).
is_predicate_symbol(P):-
	is_predicate(P/_).





input_error(Input,File,Line,S,P):-
	(Line<0 ->
	     LC=clause, Near=in
	;
	     LC=line, Near=near
	),
	appendAtoms(['~n! Error ~w ~w ~p, file ~p:~n! expression ~p:~n!',S],' ',SS),
	format(SS,[Near,LC,Line,File,Input|P]), nl, nl.






% eof -------------------------------------------------------------------------
