% -*- Module: magic; -*-

:-module(magic,[qa/3
	       ]).


:-use_module(myterms,[list_of_vars/2,
		      skolemized/3]).


:-use_module(commands,[out_as_dfg/1
		       ]).


:-use_module(thisProlog,[portray_clause2/2]).

:-use_module(auxiliary,[appendAtoms/3,
		  list_as_conjunction/2,
		  lookup_ml/2,
		  member212/2]).



:-use_module(library(maplist),[maplist/3]).
:-use_module(library(basics),[member/2]).
:-use_module(library(flatten),[flatten/2]).

% Query-answer transformation: basic version
% JPG: 27.3.92
%
% qa(F,F1,Q): qa(+,+,+)
%
%	F = source file
%	F1 = file for writing transformed program
%	Q = (atomic) query to generate the transformation
%
% e.g. ?- qa('rev.pl', transf_rev, rev(_,_)).
%
% Note: if source program contains quoted atoms, they will be
% unquoted and you'll have to edit them back in to the transformed program.
%
% Note:  operator definitions are not handled.


qa(F,F1,Q) :-
	see(F),
	tell(F1),
	read(C),
	ans_clauses(C),
	seen,
	see(F),
	read(C1),
	query_clauses(C1),
	flat_name(query(Q),QQ),
	canonical( QQ ),
	displayqClauses([(QQ :- true)]),
	nl,
	seen,
	told.

ans_clauses(X) :- 
	end_of_file(X),
	!.
ans_clauses((A :- B)) :- 
	!,
	answerlits(B,Bs),
	flat_name(ans(A),AA),
	flat_name(query(A),QA),
	flat_names(Bs,FBs),
	canonical((AA :- QA,FBs)),
	conjoin(QA, FBs, Body),
	displayqClauses([(AA :- Body)]),
	read(C1),
	!,
	ans_clauses(C1).
ans_clauses(A) :- 
	flat_name(ans(A),AA),
        flat_name(query(A),QA),
        canonical((AA :- QA)),
        displayqClauses([(AA :- QA)]),
	read(C1),
	!,
	ans_clauses(C1).

answerlits((B,Bs),(B,Bs1)) :-
	sp_builtin(B),
	!,
	answerlits(Bs,Bs1).
answerlits((B,Bs),(ans(B),Bs1)) :-
	!,
	answerlits(Bs,Bs1).
answerlits(B,B) :-
	sp_builtin(B).
answerlits(B,ans(B)).

flat_name(A,A) :-
	sp_builtin(A),
	!.
flat_name(A,A1) :-
	A =.. [F,As],
	As =.. [G|Args],
	name(F,Fn),
	name(G,Gn),
	concat(Gn,[95|Fn],An),
	name(H,An),
	A1 =.. [H|Args].

flat_names((B,Bs),(B1,Bs1)) :-
	!,
	flat_name(B,B1),
	flat_names(Bs,Bs1).
flat_names(B,B1) :-
	flat_name(B,B1).


query_clauses(Z) :- 
	end_of_file(Z),
	!.
query_clauses((A :- B)) :- 
	!,
	bodylitqueries(A,B),
	read(C1),
	!,
	query_clauses(C1).
query_clauses(_) :- 
	read(C1),
	!,
	query_clauses(C1).

bodylitqueries(_,true) :-
	!.
bodylitqueries(A,B) :-
	removelast(B,Bj,Bs1),
	sp_builtin(Bj),
	!,
	bodylitqueries(A,Bs1).
bodylitqueries(A,B) :-
	removelast(B,Bj,Bs1),
	!,
	answerlits(Bs1,Bs2),
	flat_names(Bs2,QBs2),
	flat_name(query(Bj),QBj),
	flat_name(query(A),QA),
	canonical((QBj :- QA,QBs2)),
	conjoin(QA, QBs2, Body),
	displayqClauses([(QBj :- Body)]),
	bodylitqueries(A,Bs1).

removelast((B,C,Bs),Bn,(B,Bs1)) :-
	!,
	removelast((C,Bs),Bn,Bs1).
removelast((B,C),C,B) :-
	!.
removelast(B,B,true).
	
concat([],Xs,Xs).
concat([X|Xs],Ys,[X|Zs]) :-
	concat(Xs,Ys,Zs).

end_of_file(end_of_file).	



canonical(T) :-
        name('X',[A]),
        mynumbervars(T,A,1,_).

mynumbervars(X,Ch,N,N1) :-
        var(X),
        !,
        N1 is N+1,
        name(N,Nlist),
        name(X,[Ch|Nlist]).
mynumbervars(X,_,N,N) :-
        atomic(X),
        !.
mynumbervars(T,Ch,N,N1) :-
        nonvar(T),
        !,
        T =.. [_|Args],
        mynumbervarlist(Args,Ch,N,N1).

mynumbervarlist([],_,N,N).
mynumbervarlist([T|Ts],Ch,N,N2) :-
        mynumbervars(T,Ch,N,N1),
        mynumbervarlist(Ts,Ch,N1,N2).


displayqClauses([]).
displayqClauses([(H :- B)|Cls]) :-
%        name('X',[A]),
%        mynumbervars((H :- B),A,0,_),
        write(H),
        write(' :-'),nl,
        writeqBody(B),
        displayqClauses(Cls).

writeqBody((G,Gs)) :-
        !,
        write('      '),
        /*writeq(G), */
        write(G),
        write(','),nl,
        writeqBody(Gs).
writeqBody(G) :-
        write('      '),
        /*writeq(G), */
        write(G),
        write('.'),nl.



conjoin(true,B,B) :-
	!.
conjoin(B,true,B) :-
	!.
conjoin((B,Bs),Cs,(B,Ds)) :-
	!,
	conjoin(Bs,Cs,Ds).
conjoin(B,Cs,(B,Cs)).

% the following to be used with Sicstus

sp_builtin(G) :-
        predicate_property(G,built_in).

