%declarations%

:- module(output, [ 
	display_rules/0,
	displayall/0,
	print_number/1,
	print_terms/1,
	skolem/3,
	store_props/0
   ]).


:- use_module(library(basics), [
	member/2
   ]).

:- use_module(global, [
	cont/2
   ]).

:- use_module(theory, [
	display_theory/0
   ]).

:- use_module(proofState, [
	clause/1,
	props/1,
	rule/2
   ]).

:- use_module(classes, [number/2]).

:- use_module(thisProlog, [compile_nqa/2]).



%end%


%						OUTPUT

%_________________________________________________________________________
%
%	OUTPUT
%_________________________________________________________________________

displayall:-
	display_clauses,
	display_rules,
	display_theory,
	!.

store_props:-
	props(Ps),
	sort(Ps,P),
%	Ps=P,
	cont(props_tmp_file,F),
	tell(F),
	declare_dynamic_preds(P),
	assert_props(P),
	told,
	compile_nqa(props,F),
	!.

declare_dynamic_preds(P):-
	member(ac_ist_C(_),P),
	write(':- dynamic ac_ist_C/1.'),
	nl,
	fail.
declare_dynamic_preds(P):-
	member(ac_ist_A(_),P),
	write(':- dynamic ac_ist_A/1.'),
	nl.
declare_dynamic_preds(_).


declare_dynamic_preds:-
	write(':- dynamic ac_ist_C/1, ac_ist_A/1.'),
	nl.
	  
assert_props([]).
assert_props([P|Ps]):-
	writeq(P),
	write('.'),
	nl,
	assert_props(Ps).

display_clauses:-
	\+ clause( _ ), !.
display_clauses:-
	nl, nl, write('Fresh Clauses :'), nl, nl,
	clause( Cl ),
	print( Cl ), nl, fail.
display_clauses.


display_rules:-
	(setof((N,R),rule(N,R),Rules) ->
             % this gives a list which is sorted according to rule numbers
	     display_rules(Rules)
	;    true).


display_rules(RS):-
	nl, % nl, 
	write('Persisting Clauses:'), nl, nl,
	member((_,Ru),RS),
	print( Ru ), nl, fail.
display_rules(_).


print_number( N ):-  
	N<10, print_terms( [' ', N, '    : '] ), !.
print_number( N ):-
	N<100, print_terms( [' ', N, '   : '] ), !.
print_number( N ):-
	N<1000, print_terms( [' ', N, '  : '] ), !.
print_number( N ):-
	N<10000, print_terms( [' ', N, ' : '] ), !.
print_number( N ):-
	print_terms( [' ', N, ' : '] ), !.


print_terms( L ):- \+(\+( print_skolemized( L, 1, _ ) )), !.


print_skolemized( [], _, _ ):-!.
print_skolemized( [T|R], N1, N2 ):-
		skolem( T, N1, N3 ), 
		write( T ),
		print_skolemized( R, N3, N2 ), !.


% skolemizes variables in any Prolog data structure, by
% x1,  x2, ... in order of appearance.    

skolem( PVar, N, N1 ):-
                var( PVar ),  
		ascii( N, [], L ),
		name( PVar, [120|L] ), 
		N1 is N + 1.
skolem( [], N, N ):-!.
skolem( [Term|RTerm], N, N2 ):-
                skolem( Term, N, N1 ), 
                skolem( RTerm, N1, N2 ), !.
skolem( Term, N, N1 ):- 
                Term =.. [_|LTerm], 
                skolem( LTerm, N, N1 ).


ascii( 0, L, L ):-!.
ascii( N, L1, L2 ):-
	K is N mod 10,
	A is K+48,
	N1 is N // 10,
	ascii( N1, [A|L1], L2 ), !.




%eof ____________________________________________________________


