%declarations%
:- module(poly, [ 
	default_interpretation/1,
	greater/2,
	greater_ground/2,
	greaterEqual/2,
	term_to_poly/2
   ]).


:- use_module(library(basics), [member/2]).

:- use_module(library(maplist), [maplist/3]).

:- use_module(lpo,[]).

:- use_module(equationalTheory,[equal/2]).

:- use_module(proofState,[operators/1,theory_symbol/1]).

:- use_module(achelp,[ac_ist_AC/1]).

:- use_module(props,[interpretation/1]).

:- use_module(library(clpr)).

:- use_module(adt_term, [var1/1]).


%end%






term_to_poly(X,X):-
	var(X),
	{X>=2},
	!.
term_to_poly('$min',2):-
	!.
term_to_poly('$true$',2):-
	!.
term_to_poly('$VAR'(I),I):-
	!.
term_to_poly1(V,2):-
	% the case of deBruijn indexes
	var1(V),
	!.
term_to_poly(T,T):-         
    % this case may occur as a side-effect during constraint solving
	float(T),
	!.
term_to_poly(T,P):-
	T=..[F|Ts],
	maplist(term_to_poly,Ts,Ps),
	PF=..[F|Ps],
	(interpretation(I),member(PF-P,I) ->
	     true
	;
	     length(Ps,N),
	     defaultInterpretationFor(F/N,PF-P)
	),
	!.


term_to_poly1(X,2):-
	var(X),
	!.
term_to_poly1('$min',2):-
	!.
term_to_poly1('$true$',2):-
	!.
term_to_poly1('$VAR'(I),I):-
	!.
term_to_poly1(V,2):-
	var1(V),
	!.
term_to_poly1(T,P):-
	T=..[F|Ts],
	maplist(term_to_poly1,Ts,Ps),
	PF=..[F|Ps],
	(interpretation(I),member(PF-P,I) ->
	     true
	;
	     length(Ps,N),
	     defaultInterpretationFor(F/N,PF-P)
	),
	!.


defaultInterpretationFor(F/N,FT-Poly):-
	length(ArgPolys,N),
	FT=..[F|ArgPolys],
	(N=2,ac_ist_AC(F) ->
	     Poly=..[+|ArgPolys]
	;N=0 ->
	     Poly=2
	;N=1 ->
	     ArgPolys=[P],
	     Poly=2*P+2
	;
	     sum([2|ArgPolys],1,Poly)

	),
	!.

	


sum([],_,0).
sum([X,Y],F,F*X+F*Y).
sum([X,Y|Xs],F,F*X+S):-
	sum([Y|Xs],F,S).








default_interpretation(I):-
	operators(Ops),
	maplist(defaultInterpretationFor,Ops,I).


%------------------------------------------------------------------------------
% greater( S, T ):
%	S is greater than T with respect to the poly interpretation
%  
%------------------------------------------------------------------------------


greater(S,T):-
	\+ \+ gt(S,T).

gt( S, _T ):-
	var1(S),
	!,
	fail.
gt( S, T ):-
	term_to_poly(S,PS),
	term_to_poly(T,PT),
	\+ {PT>=PS}.

greaterEqual(T1,T2):- equal(T1,T2),!.
greaterEqual(_T1,T2):- '$min' == T2,!.
greaterEqual(T1,T2):- 
	'$true$' == T2,
	T1\=='$min'.
greaterEqual(T1,T2):- greater(T1,T2),!.


greater_ground(S,T):-
	      term_to_poly1(S,P),
	      term_to_poly1(T,Q),
	      (P>Q ->
		   true
	      ;P=Q ->
		   lpo:greater_ground(S,T)
	      ).



