:-module(props,[
	       compileProps/1,
	       ordering/1,
	       ordering/2,
	       option/1,
	       theory/1,

	       user_selection/1,
	       user_selection/2,
	       theory_clause/1,
	       blocked_clause/1,

	       ac/1,
	       ac_ist_A/1,
	       ac_ist_C/1,
	       age_selection_modulus/1,
	       anti_symmetric_relation_symbol/1,
	       composition_of_relations/3,
	       congruence_relation_symbol/1,
	       dense_ordering_relation_symbol/1,
	       dense_ordering_wo_end_relation_symbol/1,
	       equality_relation_symbol/1,
	       expand_definition/1,
	       expand_definition_for/1,
	       explicit_nonstrict_total_ordering_relation_symbol/1,
	       finite_domain/1,
	       full_cnf/1,
	       function_as_predicate/5,
	       irreflexive_relation_symbol/1,
	       needs_trre_relation_symbol/1,
	       nonstrict_total_ordering_relation_symbol/1,
	       noendL_relation_symbol/1,
	       noendR_relation_symbol/1,
	       partial_congruence_relation_symbol/1,
	       partial_equivalence_relation_symbol/1,
	       predicate_with_extension/3,
	       reflexive_closure/2,
	       reflexive_relation_symbol/1,
	       residual/3,
	       strict_part/2,
	       special_relation/1,
	       special_user_relation/1,
	       symbol_admits_var_elim/1,
	       symmetric_relation_symbol/1,
	       total_ordering_relation_symbol/1,
	       total_relation_symbol/1,
	       transitive_relation_symbol/1,

	       agoal/0,
	       flat_proofs/1,
	       flat_proofs/4,
	       iterative_saturation/1,
	       horn/1,
	       horn/0,
	       firstOrder/0,
	       search_depth/1,
	       skolem_functions/1,
	       structural_expansion/1,
	       tactic/1,

	       newCOperator/1,
	       newAOperator/1,
	       newReflRelation/1,

	       flag_on/2,
	       default_flags/2,

	       interpretation/1,
	       opOccursInBuiltInClauses/1,
	       op_complexity_/2,
	       term_weight/2,
	       scaled_clause_weight/5,
	       scaled_clause_weight/7,
	       scaled_inference_weight/11,
	       scaled_clause_weight_D/7,
	       scaled_inference_weight_D/11,
	       segment_length/1,
	       case_split/3,

	       max_cases/1,
	       max_skolem/1,
	       max_deBruijn/1,

	       crw_depth/1,
	       crw_size_bound/1,
	       initial_var_depth/1,
	       special_relations/1,
	       memo_constraints/1,
	       var_overlaps/1

	       ]).

:-use_module(adt_term,[
		      term_size_weight/2,
		      term_var_weight/2,
		      term_depth_weight/2,
		      term_var_weight_gt_small/2,
		      normal_term_weight/2
		      ]).
:-use_module(adt_precedence,[adaptPrecedence/2,
			     compile_precedence/0]).

:- use_module(equationalTheory, [
	currentTheory/1,
	match/2,
	match/4
   ]).

:-use_module(global,[cont/2,store/2]).
:-use_module(proofState, [recordProp/1,clause/1,rule/2]).
:-use_module(thisProlog, [compile_nqa/2,prolog_version/1,portray_clause2/2]).
:-use_module(termOrder,[currentOrdering/1]).
:-use_module(constraintSolver,[currentSolver/1]).
:-use_module(classes,[kind/2,operators/2]).

:-use_module(library(basics),[member/2]).
:-use_module(library(maplist),[maplist/2]).
:-use_module(mylists,[is_list/1]).
:-use_module(auxiliary,[appendAtoms/3,interval/3]).





:-ensure_loaded(initialprops).
:-ensure_loaded(defaults).
:-ensure_loaded('int_flags').



compileProps(F):-
	compile_nqa(props,F),
	!.
	



ordering(Ord):-
	ordering(Ord,Ord).

ordering(poly,poly):-
	ordering(poly([]),poly).
ordering(poly(I),S):-
	prolog_version(V),
	V\==sicstus3,
	!,
	raise_exception(not_implemented(ordering(poly(I),S),'polynomial interpretations not implemented for this Prolog',V)),
	fail.
ordering(Ord2,Solver):-
	(Ord2=poly(I1) ->
	     Ord1=poly,
	     (I1=[] ->
		  use_module(satsrc(poly),[default_interpretation/1]),
		  default_interpretation(I),
		  format('~nUsing interpretation ~p',[I])
	     ;
	          I1=I
	     )
	;
	     Ord1=Ord2
	),
	member((Ord1,Solver),
	       [(elpo,lpo),(elpo,generic),
		(lpo,lpo),(lpo,generic),(rpo,generic),
		(sto,sto),(sto,generic),(sto,generic),
		(poly,poly), (poly,generic),
		(aclpo,generic), (rpo,generic)
               ,(apo,generic)
		]),
	!,
	(Ord1=elpo ->   % lpo with empty precedence
	     Ord=lpo
	;
	     Ord=Ord1
	),
	currentOrdering(OldOrd),
	(Ord=OldOrd -> 
	     true
	;    name(Ord,ON),
	     (Ord1\==elpo ->  % lpo with precedence adapted from OldOrd
		  adaptPrecedence(OldOrd,Ord)
	     ;
	          true
	     ),
	     user:append(ON,"Interface",ONI),
	     name(OI,ONI),
%	     termOrder:compile(satsrc(OI))
	     compile_nqa(termOrder,satsrc(OI))
	     
	),
	(currentSolver(Solver) -> 
	     true
	;
	     name(Solver,SN),
	     user:append(SN,"ConstraintSolver",FN),
	     name(F,FN),
	     compile_nqa(constraintSolver,satsrc(F))
	),
	(Ord2=poly(I) ->
	     setPolyInterpretation(I)
	;
	     true
	).
ordering(Ord,Solver):-
	format('~n*** unknown ordering ~p or constraint solver ~p~n',[Ord,Solver]).



theory(Th):-
	(currentTheory(Th) ->
	     true
	;
	     appendAtoms([Th,'Theory'],'',TT),
	     compile_nqa(equationalTheory,satsrc(TT))
	).


ac_ist_C(O):-
	commutative(O).

ac_ist_A(O):-
	associative(O).



option(Os):-
	is_list(Os),
	!,
	cont(int_flags_tmp_file,F),
	open(F,write,FStream),
	maplist(write_code(FStream),Os),
	!,
	close(FStream),
	cont(int_flags_tmp_file,F),
	compileProps(F),
	(member(skolem_functions(_),Os) ->
	    compile_precedence
	;   true
	).
option(O):-
	option([O]).

setPolyInterpretation(I):-
	option(interpretation(I)).


newCOperator(O):-
	(ac_ist_C(O) -> true; 	assert(commutative(O))).

newAOperator(O):-
	(ac_ist_A(O) -> true; 	assert(associative(O))).

newReflRelation(O):-
        (reflexive_relation_symbol(O) -> 
	     true
	;   
	     assert(reflexive_relation_symbol(O))
	).


% residual(?R1,+R2,+R3)  or residual(+R1,?R2,+R3)
residual(=,R,R):-
	partial_equivalence_relation_symbol(R),
	R\== ==,
	!.
residual(R,=,R):-
	partial_equivalence_relation_symbol(R),
	R\== ==,
	!.
residual(R,RR,R):-
	reflexive_closure(R,RR),
	!.
residual(RR,R,R):-
	reflexive_closure(R,RR),
	!.
residual(R1,R2,R3):-
	composition_of_relations(R1,R2,R3).


needs_trre_relation_symbol(RS):-
        \+ symmetric_relation_symbol(RS),
	(strict_part(RS,RSS) ->
	     \+ total_ordering_relation_symbol(RSS)
	;    true).




opOccursInBuiltInClauses(O):-
	dense_ordering_relation_symbol(R),
	(clause(C);rule(_,C)),
	kind(C,density(R,_,strict)),
	operators(C,[O/_]).
opOccursInBuiltInClauses(O):-
	dense_ordering_relation_symbol(R),
	reflexive_closure(R,RR),
	clause(C),
	kind(C,density(RR,_,nonstrict)),
	operators(C,[O/_]).
opOccursInBuiltInClauses(O):-
	dense_ordering_wo_end_relation_symbol(R),
	clause(C),
	kind(C,noendL(R)),
	operators(C,[O/_]).
opOccursInBuiltInClauses(O):-
	dense_ordering_wo_end_relation_symbol(R),
	rule(_,C),
	kind(C,noendL(R)),
	operators(C,[O/_]).
opOccursInBuiltInClauses(O):-
	dense_ordering_wo_end_relation_symbol(R),
	clause(C),
	kind(C,noendR(R)),
	operators(C,[O/_]).
opOccursInBuiltInClauses(O):-
	dense_ordering_wo_end_relation_symbol(R),
	rule(_,C),
	kind(C,noendR(R)),
	operators(C,[O/_]).
opOccursInBuiltInClauses(R):-
	explicit_total_ordering_relation_symbol(P),
	reflexive_closure(P,R),
	P\==R.
opOccursInBuiltInClauses(P):-
	explicit_nonstrict_total_ordering_relation_symbol(P),
	reflexive_closure(P,R),
	P\==R.

expand_definition_for(PN):-
	expand_definition(Ps),
	(Ps=all -> true; member(PN,Ps)).



%%%%%%%%%%%%%% PRIVATE %%%%%%%%%%%%%%%%%%%




% the cuts in the output clauses allow for multiple and contradictionary
% options for the same flag

write_code(_,X):-
	var(X),
	!,
	raise_exception(input_error('Option should not be a variable.',[])).

write_code(_,ordering(Ord,Solver)):-
	ordering(Ord,Solver).

write_code(F,(select(P,Cond):-Body)):-
	!,
	portray_clause2(F,(user_selection(P,Cond):-Body)),
	recordProp(user_selection(on)).

write_code(F,(H:-B)):-
	portray_clause2(F,(H:- B)).

write_code(F,term_var_weight):-
	portray_clause2(F,(term_weight(T,S):- term_var_weight(T,S),!)),
	portray_clause2(F,(segment_length(L):-large_increment(L),!)).

write_code(F,term_depth_weight):-
	portray_clause2(F,(term_weight(T,S):- term_depth_weight(T,S),!)),
	portray_clause2(F,(segment_length(L):-large_increment(L),!)).

write_code(F,term_var_weight_gt_small):-
	portray_clause2(F,(term_weight(T,S):-term_var_weight_gt_small(T,S),!)),
	portray_clause2(F,(segment_length(L):-large_increment(L),!)).

write_code(F,normal_term_weight):-
	portray_clause2(F,(term_weight(T,S):-normal_term_weight(T,S),!)),
	portray_clause2(F,(segment_length(L):-normal_increment(L), !)).

write_code(F,term_size_weight):-
	portray_clause2(F,(term_weight(T,S):-term_size_weight(T,S),!)),
	portray_clause2(F,(segment_length(L):-normal_increment(L), !)).

write_code(F,increment(I)):-
	portray_clause2(F,(segment_length(I):- !)).

write_code(F,crw_depth(I)):-
	portray_clause2(F,(crw_depth(I):- !)).

write_code(F,search_depth(I)):-
	(number(I),I>= -1 ->
	    portray_clause2(F,(search_depth(I):- !))
	;raise_exception(input_error('Search depth should be a number >= -1.',[]))).

write_code(F,age_selection_modulus(I)):-
	(number(I),I>0 ->
	     portray_clause2(F,(age_selection_modulus(I):- !))
	;raise_exception(input_error('Age selection modulus should be a positive number.',[]))).

write_code(F,var_depth(I)):-
	portray_clause2(F,(initial_var_depth(I):- !)).

write_code(F,crw_size_bound(I)):-
	portray_clause2(F,(crw_size_bound(I):- !)).

write_code(F,special_relations(off)):-
	portray_clause2(F,(special_relations(off):- !)).
write_code(F,special_relations(on)):-
	portray_clause2(F,(special_relations(on):- !)).

write_code(F,skolem_functions(big)):-
	portray_clause2(F,(skolem_functions(big):- !)).
write_code(F,skolem_functions(small)):-
	portray_clause2(F,(skolem_functions(small):- !)).


write_code(F,ac(off)):-
	portray_clause2(F,(ac(off):- !)).
write_code(F,ac(on)):-
	portray_clause2(F,(ac(on):- !)).

write_code(F,select(P,Cond)):-
	portray_clause2(F,user_selection(P,Cond)),
	recordProp(user_selection(on)).

write_code(F,memo_constraints(off)):-
	portray_clause2(F,(memo_constraints(off):- !)).

write_code(F,memo_constraints(on)):-
	portray_clause2(F,(memo_constraints(on):- !)).

write_code(F,var_overlaps(off)):-
	portray_clause2(F,(var_overlaps(off):- !)).

write_code(F,var_overlaps(on)):-
	portray_clause2(F,(var_overlaps(on):- !)).

write_code(F,horn(off)):-
	portray_clause2(F,(horn(off))).

write_code(F,horn(on)):-
	portray_clause2(F,(horn(on))).

write_code(F,structural_expansion(S)):-
	(S=on;S=off),
	portray_clause2(F,structural_expansion(S)).

write_code(F,full_cnf(S)):-
	(S=yes;S=no),
	portray_clause2(F,full_cnf(S)).

write_code(F,cnf(full)):-
	portray_clause2(F,full_cnf(yes)).

write_code(F,cnf(partial)):-
	portray_clause2(F,full_cnf(no)).


write_code(F,iterative_saturation(S)):-
	(S=on;S=off;is_list(S)),
	portray_clause2(F,iterative_saturation(S)).

write_code(F,expand_definition(D)):-
	(D=all;is_list(D)),
	portray_clause2(F,expand_definition(D)).

write_code(F,horn(on)):-
	portray_clause2(F,(horn(on))).

write_code(F,case_split((A->S),C)):-
	portray_clause2(F,case_split(A,S,C)).

% for compatibility with earlier version
write_code(F,term_var_size):-
	write_code(F,term_var_weight).
write_code(F,normal_term_size):-
	write_code(F,normal_term_weight).
write_code(_F,reflexive_closure(P,Q)):-
	recordProp(refl_clo(P,Q)).

write_code(F,interpretation(I)):-
	portray_clause2(F,interpretation(I)).

write_code(F,flat_proofs(on)):-
	write_code(F,flat_proofs(500,1000000,5,5)),
	portray_clause2(F,flat_proofs(on)).

write_code(F,flat_proofs(off)):-
	write_code(F,flat_proofs(0,0,1,5)),
	portray_clause2(F,flat_proofs(off)).


write_code(F,flat_proofs(LB,UB,Freq,Depth)):-
	portray_clause2(F,flat_proofs(LB,UB,Freq,Depth)),
	weight_clause1(C1,Depth),
	weight_clause2(C2,Depth),
	portray_clause2(F,C1),
	portray_clause2(F,C2).

write_code(F,tactic(T)):-
	portray_clause2(F,tactic(T)).

write_code(_F,O):-
	raise_exception(input_error('Unknown option ~p.',[O])).



weight_clause1((scaled_clause_weight_D(W,D,B,_T,Red,K,S):-
	                    (Red=reductive ->
				 Delta=0
			    ;
			         Delta= 0
			    ),
		            (D<Depth ->
			         S1 is W/2 
			    ;D<4*Depth -> 
				 S1 is W/2 + D/2 +B/5 
			    ;
			         S1 is W + B/5 + D/2
			    ),
		            S2 is S1-Delta,
			    (K=extensionality(_,_,_) ->
				 S is S2/2
			    ;
			         S=S2
			    )
		       ),
	       Depth).

weight_clause2((scaled_inference_weight_D(W1,W2,D1,D2,T1,T2,Red1,Red2,_K1,K2,W):-
	                    (Red1=reductive,Red2=reductive ->
				 Delta=0
			    ;
			         Delta= 0
			    ),
		            ((T1=theory,T2=theory ; T1=(goal),T2=(goal))->
				 (D1+D2 <8*Depth ->
				      W3=W2  
				 ;
				      W3=W2+D2/2-5
				 )
			    ;T1=(goal),T2=theory,D2<Depth ->
				 W3=W2/3
			    ;
			         W3=(W1+W2+D2)/4   
				 
			    ),
		            W4 is W3-Delta,
			    (K2=extensionality(_,_,_) ->
				 W is W4/2
			    ;
			         W=W4
			    )
		       ),
	       Depth).
	
set_weights:-
	(flat_proofs(on), agoal, \+horn  ->
	     weight_clause1(C1),
	     weight_clause2(C2),
	     option([C1,C2])
	;
	     true
	).

