
:- module(relational_theory,[

         subset_relation/6,
	 union_relations/9

   ]).



:- use_module(props, [
         reflexive_relation_symbol/1,
         symmetric_relation_symbol/1,
	 transitive_relation_symbol/1,
	 anti_symmetric_relation_symbol/1,
         partial_congruence_relation_symbol/1,
         total_ordering_relation_symbol/1,
	 reflexive_closure/2
   ]).









/*
subset_relation(+P,+S,+C,+Q,+T,+D)

is true when 
   converse(C,S)^P is known to be a subset of converse(D,T)^Q

*/



subset_relation(P,S,C,Q,T,D):-
	normalized_relation(P,S,C,P1,S1,+),
	normalized_relation(Q,T,D,Q1,T1,+),
	(P1=Q1 ->
	     (P1= + ->
		  subset_relation(S1,T1)
	     ;
	          subset_relation(T1,S1)
	     )
	).


subset_relation(R,R).
subset_relation(=,R):-
	R\== ==,
	reflexive_relation_symbol(R).
subset_relation(S,T):-
	reflexive_closure(S,T).

union_relations(P,R,C,Q,S,D,X,Y,Z):-
	normalized_relation(P,R,C,P1,R1,C1),
	normalized_relation(Q,S,D,Q1,S1,D1),
	(partial_congruence_relation_symbol(R1) ->
	     union_relations1(Q1,S1,D1,P1,R1,C1,X,Y,Z)
	;
	     union_relations1(P1,R1,C1,Q1,S1,D1,X,Y,Z)
	).


union_relations1(+,R,C,+,S,_,+,T,C):-
	partial_congruence_relation_symbol(S),
	reflexive_closure(R,T).
union_relations1(+,R,C,+,R,C1,-,=,+):-
	minus(C,C1),
	total_ordering_relation_symbol(R).
union_relations1(-,R,C,-,R,C1,-,=,-):-
	minus(C,C1),
	nonstrict_partial_ordering_relation_symbol(R).
union_relations1(P,R,C,Q,S,D,+,'$all',+):-
	minus(P,P1),
	subset_relation(P1,R,C,Q,S,D).




nonstrict_partial_ordering_relation_symbol(R):-
	transitive_relation_symbol(R),
	reflexive_relation_symbol(R),
	anti_symmetric_relation_symbol(R).

normalized_relation(P,R,C,P,R,D):-
	minus(C,D),
	symmetric_relation_symbol(R).
normalized_relation(P,R,-,P1,R1,+):-
	total_ordering_relation_symbol(R),
	reflexive_closure(R,R1),
	minus(P,P1).
normalized_relation(P,R,-,P1,R1,+):-
	reflexive_closure(R1,R),
	total_ordering_relation_symbol(R1),
	minus(P,P1).
normalized_relation(P,R,C,P,R,C).



minus(-,+).
minus(+,-).
