:- module(rewrite,[
        containsReducer/1,
        containsContextualReducer/1,
        containsSubsumer/1,
	normalform/6,
	normalformBy/3,
	normalformBy/4,
	reducible/4,
	rewriteStrength/1,
	normalform1/6,
	injectivitySimplifiers/4,
	simplifier/4,
	subsumes_literals/2
   ]).


:- use_module(adt_atom, [
        atom_as_term/2,
        create_atom/2,
	equivalent_relation_symbols/2,
	false_literal/1,
	get_atom_pre_rel_left_right/5,
	get_atom_vars/2,
	greater_literal/4,
	induced_equivalence/2,
	normalize_atom/2,
	orient/6,
	orient_atom/2,
	true_atom/1,
	true_literal/1,
	match_atoms/2,
	var_on_literal/2,
	var_on_predicate/2
   ]).

:- use_module(adt_term, [
	containsAllVarsOf/2,
	find_value_of_extra_vars/3,
	free_vars/2,
	get_term_vars/2,
	insert_shifts/2,
	beta_reduction/4,
	standard_atom/1,
	term_size/2,
	var1/1
   ]).

:- use_module(adt_precedence, [is_predicate/1,is_operator/1]).

:- use_module(library(maplist),[maplist/2,maplist/3]).

:- use_module(library(samsort),[samsort/3]).

:- use_module(classes, [
	createSimpleClause/2,
        constraint/2,
	dont_demodulate_with/1,
	hasSelectedLiteral/1,
	literal_list_to_sequent/3,
	literals/2,
	negate/2,
	neg_atom_as_literal/2,
	maximal_antecedent/2,
	maximal_succedent/2,
	non_maximal_antecedent/2,
	non_maximal_succedent/2,
	number/2,
	pos_atom_as_literal/2,
	propositionalClause/1
]).

:- use_module(clause_redundancy_prover, [is_tautology/2]).

:- use_module(props, [
	composition_of_relations/3,
	equality_relation_symbol/1,
	firstOrder/0,
	partial_congruence_relation_symbol/1,
	partial_equivalence_relation_symbol/1,
	symmetric_relation_symbol/1,
	reflexive_relation_symbol/1,
	reflexive_closure/2,
	strict_part/2,
	special_relation/1,
	total_ordering_relation_symbol/1
   ]).

% :- use_module(propLogic,[var_assoc/3]).

:- use_module(props,[
	flag_on/2,
	horn/0,
	total_relation_symbol/1
   ]).

:- use_module(equationalTheory, [
	equal/2,
	extension/4,
	match/2
   ]).

:- use_module(termOrder, [
        greater/2,
        greaterEqual/2
   ]).

:- use_module(global, [
        cont/2,
	incr/1,
	store/2
   ]).


:- use_module(inference_rules, [
        constraint_greater_literal/3
   ]).

:- use_module(orderConstraints, [
	and/3,
	and_gt/3,
	and_list/2,
	false/1,
	implies_s/2,
	true/1,
	valid_s/1
   ]).

:- use_module(proofState, [
	nf/4,
	irreducible_by_current_rules/1,
	rewriteFlag/1,
	recordIrreducibility/1,
	rule/2,
	setRewriteFlag/1
   ]).

:- use_module(auxiliary, [
	delete32/3,
	delete_all/3,
	member212/2
   ]).

:- use_module(library(basics),[member/2]).

:- use_module(library(sets),[select/3]).

:- use_module(library(arg),[genarg/3]).

:- use_module(myterms,[
	mapargs/3,
	subset_vars/2
   ]).


:- use_module(index,[
        empty_index/1,
	index/3,
	insertSimplifierIntoIndex/4,
	insert_terms/3,
	retrieve_match/3
   ]).


:- use_module(timing,[
	timing/2
   ]).




/*
normalform1(Type,Context,S,T,Used,ST) ist wahr, wenn S reduzibel ist und T die
Normalform von S ist.

normalform(Type,Context,S,T,Used,ST) ist wahr, wenn T die Normalform von S ist.

Used ist in beiden Faellen
die Liste der angewendeten Regelnummern als Differenzliste.

Die Verwendung beider Praedikate vermeidet das Kopieren
irreduzibler Terme. Erst wenn eine Regel irgenwo in einem
Subterm angewendet wurde (siehe *), werden neue Oberteile generiert.

ST is der aktuelle substitution tree, der die Simplifikationsregeln
enthaelt.

Type = [clause|lit|term] gibt an, welches Objekt in S vorliegt.
Context = ct(Cl,Ct) gibt den Kontext der Simplifikation an:
Cl ist die Klausel (ohne das Literal, aus dem S stammt, wenn Type=lit
oder Type=term).
Ct ist der contraint der Klausel, aus der S stammt.

*/


normalform(Type,Context,S,T,U1/U2,ST):-
	(\+rewriteStrength(min) ->
	     retractall(nf(_,_,_,_))
	;
	     true
	),
%	format('~N---',[]),
	timing(rewr,normalform(Type,Context,S,S,T,U1/U2,ST)).
	


normalform1(Ty,Context,S,T,Us,ST):-
	(\+rewriteStrength(min) ->
	     retractall(nf(_,_,_,_))
	;
	     true
	),
%	format('~N---',[]),
	timing(rewr,normalform1(Ty,Context,S,S,T,Us,ST)).



reducible(Ty,Context,S,ST):-
	(\+rewriteStrength(min) ->
	     retractall(nf(_,_,_,_))
	;
	     true
	),
%	format('~N---',[]),
	timing(rewr,reducible1(Ty,Context,S,ST)).



assertNf(A,B,C,D):-
	(\+rewriteStrength(min) ->
	     assert(nf(A,B,C,D))
	;
	     true
	).

assertLit(A,B,D):-
	(\+rewriteStrength(min) ->
	     assert(nf(A,B,literal,D))
	;
	     true
	).



/* ----------------------------------------------------------------------------
   normalform[1](Type,Context,S,S1,T,Used,ST) 

ist wie

   normalform[1](Type,Context,S,T,Used,ST), 

wobei S1 die Positionen in S mit ``$'' markiert, die
irreduzibel sind.

---------------------------------------------------------------------------- */

normalform(Type,Context,S,SR,T,U1/U2,ST):-
%	invariant((get_term_vars(S,L),L\==[]),['~n*** variable in ~p',[S]]),
%	(get_term_vars(S,L),L\==[] -> trace; true),
	(normalform1(Type,Context,S,SR,T,U1/U2,ST) -> 
	     true
	;    
	     U1=U2,
	     T=S
	).


normalform1(Type,_Context,S,'$',_T,_Us,_ST):-
	% S is irreducible, indicated by the ``$'' in the S1-argument
	assertNf(S,'$',Type,_),
	!,
	fail.
normalform1(Type,_Context,S,_,T,Used,_ST):-
	nf(S,T,Type,Used),
	!,
%	format('~Ngot ~p ->! ~p',[S,T]),
        T\=='$'.
normalform1(term,Context,S,SR,T,Us,ST):-
	\+ var1(S),
%	invariant((get_term_vars(S,L),L\==[]),['~n*** variable in ~p',[S]]),
	!,
	(functor(S,F,N) ->
	     (normalformArgs1(Context,N,S,SR,S1,F,N,Us1,ST) ->
		  (one_step_at_root(Context,S1,T1,T1R,U1/U2,ST) ->
		       normalform(term,Context,T1,T1R,T,Us2,ST),
		       append_dl(Us1,U1/U2,Us3),
		       append_dl(Us3,Us2,Us)
		  ;    T=S1,
		       Us=Us1
		  )
	     ; % Subterme von S sind irreduzibel

		  (one_step_at_root(Context,S,T1,T1R,U1/U2,ST) ->
		       normalform(term,Context,T1,T1R,T,Us1,ST),
		       append_dl(U1/U2,Us1,Us)
		  ; % S ist irreduzibel, also fail
		   assertNf(S,'$',term,_),
		   fail
		   )
	     )
	),
%	format('~N~p ~p ->! ~p',[term,S,T]),
	assertNf(S,T,term,Us).
normalform1(lit,Context,[Pol,_|A],[_,_|AR],Lit,Us,ST):-
	A=[RS,L,R],
	AR=[RS,LR,RR],
	(equations_exist(RS),
	 normalform1(term,Context,L,LR,L1,Us1,ST)->
	     normalform(term,Context,R,RR,R1,Us2,ST),
	     A1=[n,RS,L1,R1],
	     normalize_atom(A1,A1N),
	     A1R=[n,RS,'$','$'],
	     normalform(lit,Context,[Pol|A1N],[Pol|A1R],Lit,Us3,ST),
	     append_dl(Us1,Us2,Us4),
	     append_dl(Us4,Us3,Us)
/*	; % we might have an equivalence for a special relation symbol
	 RS\== ==,
	 equations_exist(==),
	 atom_as_term([_|A],AT),
	 normalform1(term,Context,AT,ATR,AT1,Us1,ST)->
	     A1=[l,==,AT1,'$true$'],
	     A1R=[n,==,ATR,'$'],
	     normalform(lit,Context,[Pol|A1],[Pol|A1R],Lit,Us2,ST),
	     append_dl(Us1,Us2,Us)
	*/
	;equations_exist(RS),
	 normalform1(term,Context,R,RR,R1,Us1,ST)->
	     A1=[n,RS,L,R1],
	     normalize_atom(A1,A1N),
  	     A1R=[n,RS,'$','$'],
	     normalform(lit,Context,[Pol|A1N],[Pol|A1R],Lit,Us2,ST),
	     append_dl(Us1,Us2,Us)
	;one_step_at_root(Context,[Pol|A],[Pol1|A1],A1R,U1/U2,ST) ->
	     normalize_atom([_|A1],[_|A1N]),
	     normalform(lit,Context,[Pol1,n|A1N],[Pol1,n|A1R],Lit,Us1,ST),
	     append_dl(U1/U2,Us1,Us)
	;one_step_at_root(Context,A,A1,A1R,U1/U2,ST) ->
	     normalize_atom([_|A1],[_|A1N]),
	     normalform(lit,Context,[Pol,n|A1N],[Pol1,n|A1R],Lit,Us1,ST),
	     append_dl(U1/U2,Us1,Us)
	; % irreduzibel, also fail
	     assertNf([Pol,_|A],'$',lit,_),
	     fail
	),
%	format('~N~p ~p ->! ~p',[lit,[Pol,_|A],[Pol|B]]),
	assertNf([Pol,_|A],Lit,lit,Us).
normalform1(clause,ct(C,Ct),_,_,D,U1/U3,ST):-
	(one_step_of_subsumption_reduction(ct(C,Ct),C1,U1/U2,ST) ->
	     normalform(clause,ct(C1,Ct),C1,C1,D,U2/U3,ST)
	).


equations_exist(==) :-
	firstOrder,
	!.
equations_exist(==) :-
	!,
	rewriteFlag(equations_exist).
equations_exist(_) :-
	rewriteFlag(term_equations_exist).




reducible1(termByUnits,Context,S,ST):-
	rewriteFlag(equations_exist),
	\+ var1(S),
	(one_step_at_root_by_units(Context,S,_,_,_,ST) ->
	     true
	;
	     rewriteFlag(term_equations_exist),
%	     rewriteFlag(equations_exist),
	     genarg(_,S,SI),
	     reducible1(termByUnits,Context,SI,ST)
	).
reducible1(term,Context,S,ST):-
	rewriteFlag(equations_exist),
	\+ var1(S),
	(one_step_at_root(Context,S,_,_,_,ST) ->
	     true
	;
	     rewriteFlag(term_equations_exist),
%	     rewriteFlag(equations_exist),
	     genarg(_,S,SI),
	     reducible1(term,Context,SI,ST)
	).
reducible1(lit,Context,[_Pol,_PreO|A],ST):-
	(one_step_at_root(Context,A,_,_,_,ST) ->
	     true
	;
	     A=[RS,L,R],
	     equations_exist(RS),
	     (   reducible1(term,Context,L,ST)
	     ;   reducible1(term,Context,R,ST)
	     )
	),
	!.
reducible1(clause,Context,_,ST):-
	one_step_of_subsumption_reduction(Context,_C1,_Used,ST).


/* ----------------------------------------------------------------------------
   normalformArgs1(+Context,+I,+S,+S1,+T,+F,+N,+ST) ist wahr, wenn die Subterme
   T[j] von T, 1=<j<=I, die Normalformen der Subterme S[j] von T
   sind. Zudem mu\3 mindestens einer der Terme S[j] reduzibel sein,
   und es mu\3 F das Topsymbol von S und T sein.
   Context provides constraint and clause context for the reduction.
   S1 marks irreducible positions in S.
   ST is the set of potential simplifiers as substitution tree.

---------------------------------------------------------------------------- */


normalformArgs1(Context,I,S,SR,T,F,N,U1/U2,ST):-
	I>0,
	rewriteFlag(term_equations_exist),
%	rewriteFlag(equations_exist),
	arg(I,S,SI),
	arg(I,SR,SRI),
	I1 is I-1,
	(normalform1(term,Context,SI,SRI,TI,U1/U3,ST) ->     % irgendwo in SI wurde eine
	                                 % Regel angewendet
	     functor(T,F,N),             % neues F-Oberteil
	     arg(I,T,TI),                % die Normalform TI von SI einhaengen
	     sameSubterms(I,N,S,T),      % die alteren Brueder von SI in S 
                                         % in T einhaengen
	     normalformArgs(Context,I1,S,SR,T,U3/U2,ST)
	;    
	     normalformArgs1(Context,I1,S,SR,T,F,N,U1/U2,ST)).



sameSubterms(N,N,_,_).
sameSubterms(I,N,S,T):-
	I<N,
	I1 is I+1,
	arg(I1,S,Sub),
	arg(I1,T,Sub),
	sameSubterms(I1,N,S,T).


normalformArgs(_,0,_,_,_,U/U,_).
normalformArgs(Context,I,S,SR,T,U1/U3,ST):-
	I>0,
	I1 is I-1,
	arg(I,S,SI),
	arg(I,SR,SRI),
	arg(I,T,TI),
	normalform(term,Context,SI,SRI,TI,U1/U2,ST),
	normalformArgs(Context,I1,S,SR,T,U2/U3,ST).
% -----------------------------------------------------------------------------
%       normalization by unit equations


% -----------------------------------------------------------------------------
%           Rules is a list of equations L=R
%           an index is generated from the list of rules
% -----------------------------------------------------------------------------

normalformBy(Type,S,T,Rules):-
	maplist(create_atom,Rules,Es1),
	maplist(orient_atom,Es1,Es),
	indexOfSimplifiers(Es,ST),
	true(Top),
	normalform(Type,ct([],Top),S,T,_,ST).




indexOfSimplifiers([],ST):-
	empty_index(ST).
indexOfSimplifiers([E|Es],ST):-
	indexOfSimplifiers(Es,ST1),
	createSimpleClause([[],[E]],C),
	insertSimplifierIntoIndex(C,ST1,ST,_).



% -----------------------------------------------------------------------------
%           Rules is a list of equations L=R
%           S is any term
%           no rule extensions are applied, no index is generated
%           the rules are assumed to be terminating
% -----------------------------------------------------------------------------


normalformBy(S,T,Rs):-
	(normalformBy1(S,T,Rs) -> 
	     true
	;    T=S).


normalformBy1(S,T,Rs):-
	(functor(S,F,N) ->
	     (normalformByArgs1(N,S,S1,F,N,Rs) ->
		  (member(S11=T1,Rs), match(S1,S11)->
		       normalformBy(T1,T,Rs)
		  ;    T=S1)
	     ; % Subterme von S sind irreduzibel

		  (member(S11=T1,Rs), match(S,S11) ->
		       normalformBy(T1,T,Rs))
		  % S ist irreduzibel, also fail

	     )
	;    raise_exception(illegal_term(S))
	).




normalformByArgs1(I,S,T,F,N,Rs):-
	I>0,
	arg(I,S,SI),
	I1 is I-1,
	(normalformBy1(SI,TI,Rs) ->           % irgendwo in SI wurde eine
	                                 %    Regel angewendet
	     functor(T,F,N),     % neues F-Oberteil
	     arg(I,T,TI),           % die Normalform TI von SI einh"angen
	     sameSubterms(I,N,S,T),     % die alteren Br"uder von SI in S 
                                         %    in T einh"angen
	     normalformByArgs(I1,S,T,N,Rs)
	;    
	     normalformByArgs1(I1,S,T,F,N,Rs)).



normalformByArgs(0,_,_,_,_).
normalformByArgs(I,S,T,N,Rs):-
	I>0,
	I1 is I-1,
	arg(I,S,SI),
	arg(I,T,TI),
	normalformBy(SI,TI,Rs),
	normalformByArgs(I1,S,T,N,Rs).



% ------------- single step rewriting ---------------------------------

one_step_at_root(Context,Sub,R,RR,N,ST) :-
	incr(rewriteCtr),
	one_step_at_root_by_units(Context,Sub,R,RR,N,ST).
one_step_at_root(Context,Sub,R,RR,N,ST) :-
	\+ rewriteStrength(min),
	rewrite_term_contextually(Context,Sub,R,RR,N,ST).


one_step_at_root_by_units(Context,Sub,R,RR,N,ST):-
%	rewrite_term(Context,Sub,R,RR,N,ST).

	\+ \+ (rewrite_term(Context,Sub,R,RR,N,ST) ->
                   store('$found',t(R,RR,N))
	      ),
	cont('$found',t(R,RR,N)),
	!.

%	findall((R,RR,N),rewrite_term(Context,ST,Sub,R,RR,N),[(R,RR,N)]).
%	on_exception(rewritten(R,RR,N),'rewrite_term!'(Context,ST,Sub,R,RR,N),true). % best choice for Quintus


one_step_of_subsumption_reduction(C,D,[N|Ns]/Ns,ST):-
	\+ \+ (rewrite_clause(C,D,N,ST) ->
                   store('$found',t(D,N))
	      ),
	cont('$found',t(D,N)).
%	findall((D,N),rewrite_clause(C,D,N,ST),[(D,N)]).
%	on_exception(rewritten(D,N),'rewrite_clause!'(C,D,N,ST),true).


'rewrite_term!'(Context,ST,Sub,R,RR,N):-
	 (rewrite_term(Context,ST,Sub,R,RR,N) -> 
	      raise_exception(rewritten(R,RR,N))
	 ).

'rewrite_clause!'(C,D,N,ST):-
	 (rewrite_clause(C,D,N,ST) -> 
	      raise_exception(rewritten(D,N))
	 ).


retrieve(Key,Index,Info):-
	\+ \+ (retrieve_match(Key,Index,Info) ->
		   store('$found',Info)
	      ),
	cont('$found',Info).


rewrite_term(_,Sub,R,RR,[N|Ns]/Ns,ST):-
	% we use preoriented unit rules first
	index(orientableUnits,ST,STL),
%	retrieve(Sub,STL,t(N,R1,RR1,V,_,_)),
	retrieve_match(Sub,STL,t(N,R1,RR1,V,_,_)),
%	(rule(N,_);N='L'),  % this should be an invariant, but apparently is not
        \+ var_on_predicate( V, Sub ),%	format('~n vp: rewrite ~p with a variable',[Sub]),
        \+ var_on_literal( V, Sub ), % format('~nvl: rewrite ~p with a variable',[Sub]),
	beta_reduction(R1,RR1,R,RR),
	!.
rewrite_term(ct(_Cl,_Ct),Sub,R,RR,[N|Ns]/Ns,ST):-
	% next we try unoriented units
	index(nonorientableUnits,ST,STU),
	retrieve_match(Sub,STU,t(N,R1,RR1,V,Ct1,_)),
%	(rule(N,_);N='L'),  % this should be an invariant,
				% but apparently is not
%	retrieve(Sub,STU,t(N,R1,RR1,V,Ct1,_)),
        \+ var_on_predicate( V, Sub ),	
        \+ var_on_literal( V, Sub ),
%	valid_s(Ct1),
	implies_s(_Ct,Ct1),
	beta_reduction(R1,RR1,R,RR),
	!.

rewrite_term_contextually(ct(Cl,Ct),Sub,R,RR,[N|RN]/RN1,ST):-
	% next we try reductive simplifiers
	index(reductiveReducers,ST,STU),
	   % we should not try to verify the condition with an
           % instantiated ST, hence this findall, followed by member
	findall(Reducer,reducer(Sub,STU,Reducer),Reducers),  % expensive
	member(t(N,R1,RR1,_V,_,Cond),Reducers),

		  %	copy_term(STU,STUCopy),  % much worse
		  %	reducer(Sub,STUCopy,t(N,R1,RR1,_V,_,Cond)),

		  %	reducer(Sub,STU,t(N,R1,RR1,_V,_,Cond)),   % worse, too

%	spy_rewrite(N,Sub,Cond,Cl),
%	(rule(N,_);N='L'),  % this should be an invariant,
				% but apparently is not
	true(Top),
	trueInContext(Cond,Top,ct(Cl,Ct),N,RN/RN1,ST),
	beta_reduction(R1,RR1,R,RR),
%	format('~N~p -> ~p by ~p successful',[Sub,R,N]),
	!.
rewrite_term_contextually(ct(Cl,Ct),Sub,R,RR,[N|RN]/RN1,ST):-
	% next we try nonreductive simplifiers
	index(nonreductiveReducers,ST,STU),
	   % we should not try to verify the condition with an
           % instantiated ST, hence this findall, followed by member
	findall(Reducer,reducer(Sub,STU,Reducer),Reducers),
	member(t(N,R1,RR1,_V,Ct1,Cond),Reducers),
			  %	copy_term(STU,STUCopy),
			  %	reducer(Sub,STUCopy,t(N,R1,RR1,_V,Ct1,Cond)),
			  %	reducer(Sub,STU,t(N,R1,RR1,_V,Ct1,Cond)),

%	spy_rewrite(N,Sub,Cond,Cl),
%	(rule(N,_);N='L'),  % this should be an invariant,
  			% but apparently is not
	trueInContext(Cond,Ct1,ct(Cl,Ct),N,RN/RN1,ST),
	beta_reduction(R1,RR1,R,RR),
%	format('~N~p -> ~p by ~p successful',[Sub,R,N]),
	!.




reducer(Sub,STU,t(N,R,RR,V,Ct,Cond)):-
	retrieve_match(Sub,STU,t(N,R,RR,V,Ct,Cond)),
%	retrieve(Sub,STU,t(N,R,RR,V,Ct,Cond)),
        \+ var_on_predicate( V, Sub ),	
        \+ var_on_literal( V, Sub ).



spy_rewrite(N,Sub,Cond,Cl):-
%	incr(rewrite),
	(spy_rewrite(M) -> 
	     proofState:rule(N,Ru),
	     ((N==M;var(M)) ->
		 (nonvar(M) -> trace; true),
		 format('~ntrying ~p on ~p ',[Ru,Sub]),
		 format('~n[cond ~p in context~n  ~p',[Cond,Cl])
	     ;
	          true
	     )
	;
	     true
	).

%spy_rewrite(N,Sub,Cond,Cl).


/*
-------------------------------------------------------------------------------



                  checking the condition of a simplifier



-------------------------------------------------------------------------------
*/

:-dynamic spy_rewrite/1.


trueInContext(t(RedConditions,[],ExtraVarConditions),Ct1,ct(Cl,Ct),N,Used,ST):-
	% ``[]'' means that as of now we don't want to
	% check at simplification time as to whether some
	% condition of a simplifier is sufficiently small.
	
	subsumes_literals(ExtraVarConditions,Cl),
	% now all variables are instantiated and we may check
	% the constraint

	implies_s(Ct,Ct1),
	samsort(reverseSizeOrder,RedConditions,NISorted),
	(spy_rewrite(N) -> 
	     global:incr(condDepth),
	     global:cont(condDepth,D),
	     format('~nchecking red=~p evc=~p of ~p',[RedConditions,ExtraVarConditions,N]),
	     format('~n[~p',[D])
        ; 
	     true
	),
	impliedNormalforms(NISorted,ct(Cl,Ct),Used,ST),  

	(spy_rewrite(N) -> 
	     format('~n]~pt',[D]),
	     global:decr(condDepth)
	;true),
	!.
trueInContext(_,_,_,N,_,_):-
	(spy_rewrite(N) -> 
	     global:cont(condDepth,D),
	     format('~n]~pf',[D]),
	     global:decr(condDepth)
	),
	fail.



impliedNormalforms([],_Context,RN/RN,_).
impliedNormalforms([L|Ls],Context,RN1/RN2,ST):-
	impliedNormalformMemo(L,Context,RN1/RN3,ST),
	!,
	impliedNormalforms(Ls,Context,RN3/RN2,ST).

impliedNormalformMemo(L,_Context,Used,_ST):-
	nf(L,Value,literal,Used),
%	format('~ngot ~p ->! ~p',[L,Value]),
	!,
	Value.
impliedNormalformMemo(L,Context,N1/N2,ST):-
	(contextImplies(L,Context) -> 
	     N1=N2,
	     assertLit(L,true,N1/N2)
	;
	     (  % format('~nreducing ~p',[L]),
%	      fail,
	      normalform1(lit,Context,L,L,LN,N1/N2,ST) ->
		  contextImplies(LN,Context),
		  assertLit(L,true,N1/N2)
	     ;
	          
	          assertLit(L,fail,_),
		  fail

	     )		  
	).
	


contextImplies([],_Ct,[]).
contextImplies([L|Ls],Context,Not1):-
	(contextImplies(L,Context) ->
	     Not1=Not
	;
	     Not1=[L|Not]
	),
	contextImplies(Ls,Context,Not).

contextImplies(L,ct(Cl,_)):-
	negate(L,NotL),
	literal_list_to_sequent([NotL|Cl],A,S),
	is_tautology(A,S).




/*
-------------------------------------------------------------------------------



                  Subsumption resolution

rewrite_clause(+ct(C,Ct),?D,?N,+ST)

is true when clause C (as list of literals) with constraint Ct
rewrites to clause D (as list of literals) by subsumption resolution
with N in state ST.


-------------------------------------------------------------------------------
*/


rewrite_clause(ct(C,Ct),D,N,ST):-
	index(subsumption,ST,STS),
	select([P,_|A],C,C1),
	retrieve_match(A,STS,t(N,A2,_,_V,Ct1,t([],[],CR))),

	(P= +  ->
	     find_inclusion(C,CR,C1,Res),
	     (Res=subsumed ->
		  detectTrivial([P,n|A2],C1,D)
	     ;% Res is a reduced clause and [P|A] is one of the side literals
	          true_lit([P,n|A2]),
		  D=Res
	     )
	;% opposite signs
	     subsumes_literals(CR,C1),
	     detectTrivial([P,n|A2],C1,D)
	),
	implies_s(Ct,Ct1),  % ok as always reductive

%	invariant((free_vars(A2,L),L\==[]),
%	          ['~n*** reduction with ~p introduces new variables',[N]]),
	!.

detectTrivial(A,_C,[A]):-
	true_lit(A),
	!.
detectTrivial(A,C,[A|C]).


true_lit([+,_,RS,L,L]):-
	reflexive_relation_symbol(RS),
	!.
true_lit(A):-
	true_literal(A).


find_inclusion(_,[],_,subsumed).
find_inclusion(WholeCL,[[Pol|A]|RL],CL,CL2):-
	delete([Pol|A],_,CL,CL1),
	find_inclusion(WholeCL,RL,CL1,CL2),
	!.
find_inclusion(WholeCL,[[Pol,_,RS,L,R]|RL],CL,CL2):-
	symmetric_relation_symbol(RS),
	delete([Pol,_,RS,R,L],_,CL,CL1),
	find_inclusion(WholeCL,RL,CL1,CL2),
	!.
find_inclusion(WholeCL,[[+,_,RS,L,R]|RL],CL,WCL1):-
	total_ordering_relation_symbol(RS),
	(RS1=RS;reflexive_closure(RS,RS1)),
	delete([+,_,RS1,R,L],A1,CL,CL1),
	subsumes_literals(RL,CL1),
	delete_all(A1,WholeCL,WCL1),
	!.
find_inclusion(WholeCL,[[+,_,RS,L,R]|RL],CL,WCL1):-
	reflexive_closure(RS1,RS),
	delete([+,_,RS1,R,L],A1,CL,CL1),
	subsumes_literals(RL,CL1),
	delete_all(A1,WholeCL,WCL1),
	!.
find_inclusion(WholeCL,[[+,O1,RS,L,R]|RL],CL,[[+,O1,=,L,R]|WCL1]):-
	reflexive_closure(_,RS),
	!,
	delete([+,_,RS,R,L],A1,CL,CL1),
	subsumes_literals(RL,CL1),
	delete_all(A1,WholeCL,WCL1),
	!.
find_inclusion(WholeCL,[[Sign|E]|RL],CL,WCL1):-
	inverse_sign(Sign,SI),
	delete([SI|E],E1,CL,CL1),
	subsumes_literals(RL,CL1),
	delete32(E1,WholeCL,WCL1),
	!.
find_inclusion(WholeCL,[[Sign,_,==,L,'$true$']|RL],CL,[[+,n,==,L,R]|WCL2]):-
	fail,
	\+ horn,
	inverse_sign(Sign,SI),
	delete([SI,_,==,L,'$true$'],A1,CL,CL1),
	select([SI,_,==,R,'$true$'],RL,RL1),
	delete([Sign,_,==,R,'$true$'],A2,CL1,CL2),
	subsumes_literals(RL1,CL2),
	delete32(A1,WholeCL,WCL1),
	delete32(A2,WCL1,WCL2),
	!.
find_inclusion(WholeCL,[[Sign,_,RS,L,R]|RL],CL,WCL1):-
	symmetric_relation_symbol(RS),
	inverse_sign(Sign,SI),
	delete([SI,_,RS,R,L],E1,CL,CL1),
	subsumes_literals(RL,CL1),
	delete32(E1,WholeCL,WCL1),
	!.



subsumes_literals(C1,C2):-
	subsumes_literals(C1,C2,_),
	!.

subsumes_literals([],CR,CR).
subsumes_literals([[Sign,_|E]|RL],CL,CR):-
	delete([Sign,_|E],_,CL,CL1),
	subsumes_literals(RL,CL1,CR).
subsumes_literals([[Sign,_,RS,R,L]|RL],CL,CR):-
	symmetric_relation_symbol(RS),
	delete([Sign,_,RS,L,R],_,CL,CL1),
	subsumes_literals(RL,CL1,CR).

	     
/*
delete(+L,?L1,+C,?C1) 
is true when C = C1,L1 and L matches literal L1 in C.
*/
delete([P|A],[P|A1],M,M1):-
	delete32([P|A1],M,M1),
	match_atoms(A,A1).



inverse_sign(-,+).
inverse_sign(+,-).




memoize_ir(S):-
	\+ \+ (numbervars(S,1,_),
	       recordIrreducibility(S)).





append_dl(U1/U3,U3/U2,U1/U2).






/*
-------------------------------------------------------------------------------

simplifier(+Cl,+Ct,+N,?Simplifier)

is true when ?Simplifier is a simplifier aspect of the clause represented by
Cl.

Cl is a tuple of the form t(Am,A,Sm,S,Sel)
where
Am and A are the maximal and nonmaximal atoms of the antecedent,
Sm and S are the maximal and nonmaximal atoms of the succedent,
and Sel= [hasSelectedLit|nothingSelected]
indicates as to whether some negative literal has been selected in Cl.

Ct is the constraint and N the number of the clause.

?Simplifier is of the form

    t(Index,Lhs,t(N,To,To',V,OrdC,Cond))

where
Index is the name of the index structure into which the simplifier is to be stored,

Lhs   is the term or literal that is reduced by the simplifier,

To   is the result of simplification

To'  is as To with variables replaced by '$' os as to indicate irreducibility
     when innermost reduction can be assumed as stretegy
     (this should also be extended to conditions of reductive simplifiers)

V    is a variable if the left side of a simplfier is a variable 
     that has to be type checked, and V=nonvar, otherwise

OrdC is an ordering constraint which, when satisfied,
     ensures that Lhs > To, and also entails the ordering constraint
     of the simplifier

Cond is the list of literals, the negations of which
     have to be logically implied by the
     context in which the simplifier is applied.

Cond is of the form t(Red,NRed,Extra), where Red are the reductive
conditions, Inst are potentially nonreductive conditions that will
be fully instantiated after matching the left side of the simplifier,
and Extra are the remaining conditions.

We basically distinguish between simplification by reductive rewriting
(first clause of ``simplifier'')
and subsumption [reduction] (second clause).


!! all Simplifiers need to be enumerated, don't put cuts nowhere !!


*/

simplifier(Cl,N,Ct,Simplifier1):-
	(spy_rewrite(N) -> trace; true),
	simplifier1(Cl,N,Ct,Simplifier1),
	(spy_rewrite(N) ->
	    Simplifier1=t(Index,Le1,t(N,Ri1,_RR,_V,Ct2,CondA)),
	    format('~ni=~p l=~p r=~p ~n  ct=~p ~n    cond=~p',
		   [Index,Le1,Ri1,Ct2,CondA])
	;
	    true
	).


simplifier1(Cl,N,Ct,Simplifier1):-
	rewriteStrength(max),
	% reduction and subsumption are integrated 
	contextAndSimplifier(red,Cl,Cond1,L,RS),
	literalAsSimplifier(L,t(Le,Ri,V,CondL,OrdCL,Ty)),

	Ty\== lifted,  % Ty=lifted would only be needed if reduction by
	               % non-orientable / non-reductive clauses
	               % is dealt with in restricted form like
	               % as unit reduction + subsumption
	               % Right now we get only all or nothing
	               % with reduction.

	append(Cond1,CondL,Cond),
	hasCondition(Cond,HCo),

	contrapositiveOrSymm(Le,Ri,Le1,Ri1),

	(Ty=term ->
	     Lit=[RS,Le1,Le1]
	;Ty=bool ->
	     Lit=[RS,Le1,'$true$']
	;% Ty in [lit|lifted]
	     Lit=Le1
	),
	reductivityAnalysis(Ty,Cond,Lit,CondA),
	combineOC(OrdCL,Ct,Ct2),
	\+ false(Ct2),
	hasConstraint(Ct2,CondA,HCt),
	Index=[HCt,HCo],
	irredIndicators(Ri1,RR),
	Simplifier=t(Index,Le1,t(N,Ri1,RR,V,Ct2,CondA)),
	instantiateExtraVars(Simplifier,Simplifier1).
simplifier1(Cl,N,Ct,Simplifier1):-
	rewriteStrength(min),
	%  reduction only with unit literals
	contextAndSimplifier(red,Cl,[],L,RS),
	literalAsSimplifier(L,t(Le,Ri,V,[],OrdCL,Ty)),

	hasCondition([],HCo),

	contrapositiveOrSymm(Le,Ri,Le1,Ri1),
	(Ty=term ->
	     Lit=[RS,Le1,Le1]
	;Ty=bool ->
	     Lit=[RS,Le1,'$true$']
	;% Ty in [lit|lifted]
	     Lit=Le1
	),
	reductivityAnalysis(Ty,[],Lit,CondA),
	combineOC(OrdCL,Ct,Ct2),
	\+ false(Ct2),
	hasConstraint(Ct2,CondA,HCt),
	Index=[HCt,HCo],
%	contrapositiveOrSymm(Le,Ri,Le1,Ri1),
	irredIndicators(Ri1,RR),
	Simplifier=t(Index,Le1,t(N,Ri1,RR,V,Ct2,CondA)),
	instantiateExtraVars(Simplifier,Simplifier1).
simplifier1(Cl,N,Ct,Simplifier1):-
	rewriteStrength(medium),
%	((N=82;N=154;N=314) -> trace;true),
	%  reduction only with uniformly reductive clauses
	contextAndSimplifier(red,Cl,Cond1,L,RS),
	literalAsSimplifier(L,t(Le,Ri,V,CondL,OrdCL,Ty)),

	append(Cond1,CondL,Cond),
	hasCondition(Cond,HCo),
	contrapositiveOrSymm(Le,Ri,Le1,Ri1),

	(Ty=term ->
	     Lit=[RS,Le1,Le1]
	;Ty=bool ->
	     Lit=[RS,Le1,'$true$']
	;% Ty in [lit|lifted]
	     Lit=Le1
	),
	reductivityAnalysis(Ty,Cond,Lit,CondA),
	combineOC(OrdCL,Ct,Ct2),
	\+ false(Ct2),
	hasConstraint(Ct2,CondA,HCt), 
	(Cond=[] ->
	     true
	;
	     HCt=orientable % uniformly reductive
%	     ,!  % in this case subsumption checks will not involve this clause
                % hence no subsumption index entry needs to be generated
	),

	Index=[HCt,HCo],
%	contrapositiveOrSymm(Le,Ri,Le1,Ri1),
	irredIndicators(Ri1,RR),
	Simplifier=t(Index,Le1,t(N,Ri1,RR,V,Ct2,CondA)),
	instantiateExtraVars(Simplifier,Simplifier1).
simplifier1(Cl,N,Ct,Simplifier1):-
	\+ rewriteStrength(max),
	% reduction is complemented by subsumption resolution
	contextAndSimplifier(subs,Cl,Cond,L,_),
	(Ty=lit;Ty=lifted),     
	literalAsSimplifier(L,t(Le,Ri,VTag,[],OrdCL,Ty)),
	CondA=t([],[],Cond), % simplification with non-selected literals 
                             % only via subsumption resolution
	combineOC(OrdCL,Ct,Ct2),
	\+ false(Ct2),
	Index=subsumption,  
	contrapositiveOrSymm(Le,Ri,Le1,Ri1),
	irredIndicators(Ri1,RR),
	Simplifier=t(Index,Le1,t(N,Ri1,RR,VTag,Ct2,CondA)),
	instantiateExtraVars(Simplifier,Simplifier1).


hasSelection(t(_,_,_,_,1)).


rewriteStrength(min):-
	% flag rws2 overrules flag rws0
	flag_on(rp,rws0),
	\+flag_on(rp,rws2),
	!.
rewriteStrength(medium):-
	% default flag setting
	\+flag_on(rp,rws0),
	\+flag_on(rp,rws2),
	!.
rewriteStrength(max):-
	% flag rws2 overrules flag rws0
	flag_on(rp,rws2).



injectivitySimplifiers(N,[_|EF],[_|E],
           [t([orientable,unconditional],EF,
            t(N,E,ER,nonvar,Top,t([],[],[])))]):-
	irredIndicators(E,ER),
	true(Top).

	





/*
-------------------------------------------------------------------------------

contextAndSimplifier(+Kind,+Cl,?Cl1,?Equation)

is true when the clause Cl is of the form 

    Cl = Cl1 vee Equation

such that it makes sense to attempt simplification of other clauses
by rewriting with Equation, provided that the context of the
simplification entails the negation of Cl1.

Kind in [red|subs]
indicates as to whether one is interested in recursive reductive 
rewriting with Cl or, alternatively, in subsumption (reduction).
In the first case the clause needs to have at least one potentially maximal
literal. This is not the case when one has something negative selected
in Cl.


-------------------------------------------------------------------------------
*/

% the following choices of turning clauses into contextual simplifiers
% for reductive simplification or subsumption reduction
% is still somewhat arbitray and subject to further experimentation



contextAndSimplifier(RedKind,t(Am,[],Sm,[],_),[],t(Pol,P,R,U,V),R):-
	% reduction for unit clauses
	(
	     Am=[],
	     Sm=[E],
	     Pol= +
	;
	     Am=[E],
	     Sm=[],
	     Pol= -
	),
	!,
	RedKind=red,
	get_atom_pre_rel_left_right(E,P,R,U,V).
contextAndSimplifier(red,t([],[E|A],MaxS,S,hasSelectedLit),Cond,t(-,P,R,U,V),R):-
	% simplification with the selected literal E
	append(S,MaxS,S1),
	maplist(pos_atom_as_literal,S1,SL),
	maplist(neg_atom_as_literal,A,AL),
	append(AL,SL,Cond),
	get_atom_pre_rel_left_right(E,P,R,U,V).
contextAndSimplifier(red,t(Am,Ap,Sm,Sp,nothingSelected),Cond,t(+,P,R,U,V),R):-
	append(Sm,Sp,S),
	select(E,S,S1),
	get_atom_pre_rel_left_right(E,P,R,U,V),
	append(Am,Ap,A),
	maplist(pos_atom_as_literal,S1,SL),
	maplist(neg_atom_as_literal,A,AL),
	append(AL,SL,Cond).
contextAndSimplifier(red,t(Am,Ap,Sm,Sp,nothingSelected),Cond,t(-,P,R,U,V),R):-
	append(Ap,Am,A),
	select(E,A,A1),
	get_atom_pre_rel_left_right(E,P,R,U,V),
	append(Sm,Sp,S),
	maplist(pos_atom_as_literal,S,SL),
	maplist(neg_atom_as_literal,A1,AL),
	append(AL,SL,Cond).

%%% subsumption resolution  
contextAndSimplifier(subs,t([],[E1|A],MaxS,S,hasSelectedLit),[[-|E1]|Cond],t(Pol,P,R,U,V),R):-
	append(S,MaxS,S1),
	maplist(pos_atom_as_literal,S1,SL),
	maplist(neg_atom_as_literal,A,AL),
	append(AL,SL,Lits),
	samsort(sizeOrder,Lits,[[Pol|Max]|Cond]),
	get_atom_pre_rel_left_right(Max,P,R,U,V).
contextAndSimplifier(subs,t(Am,Ap,MaxS,S,nothingSelected),Cond,t(Pol,P,R,U,V),R):-
	% not needed when subsumption is in rewriting
	append(S,MaxS,S1),
	maplist(pos_atom_as_literal,S1,SL),
	append(Am,Ap,A),
	maplist(neg_atom_as_literal,A,AL),
	append(AL,SL,Lits),
	Lits=[_|_],     % not a unit clause
	samsort(sizeOrder,Lits,[[Pol|Max]|Cond]),
	get_atom_pre_rel_left_right(Max,P,R,U,V).



sizeOrder(A,B):-
	term_size(A,N),
	term_size(B,M),
	M<N.

reverseSizeOrder(A,B):-
	sizeOrder(B,A).





/*
literalAsSimplifier(+ t(Pol,P,RS,L,R), ?SInfo)

is true when SInfo specifies a simplifier derived from the
given literal with polarity Pol, preorientation P, relation symbol RS
left and right sides L and R, respectively.

SInfo = [Le,Ri,V,Cond,OrdC,+Ty]
where 
  V is a variable if Le is a variable that has to be type checked, 
    and V=nonvar, otherwise, and
  Cond is a logical condition for the simplification step.
  Ord is a list (conjunction) of inequalities which represent
    an ordering constraint for the application of the rewrite rule Le -> Re,
  Ty=[term|bool|lit|lifted] 
    specifies the type of object that the simplifier reduces.

Simplifications preserve logical equivalence if the condition
is implied by the context of its application.

Note: Ty should be instantiated in calls where Ty=bool is not 
      wanted.

Note: At this point we cannot reduce with rules where left
      sides contain binders. In that case shifting of
      free variables in the right sides may be required
      if they are instantiated with a bound variable at reduction time.
      At this time we can only shift up not down after
      reduction (when right sides have free variables in the scope
      of a binder).
*/




literalAsSimplifier(t(+,_,==,[=,L,R],[=,L1,R1]),SInfo):-
	SInfo=t(L=R,L1=R1,nonvar,[],[],lit).
literalAsSimplifier(t(+,P,==,L,R),SInfo):-
%	(L=step2 -> trace; true),
	orient(P,L,R,P1,L1,R1),
	extension(L1,R1,L2,R2),
	standard_atom(L2),   % cf. Note 
	preOrientationAsConstraint(P1,L1,R1,Ct),
	SInfo=t(L2,R2,nonvar,[],Ct,bool).
literalAsSimplifier(t(+,P,RS,L,R),SInfo):-
	partial_congruence_relation_symbol( RS ),
	orient(P,L,R,P1,L1,R1),
	extension(L1,R1,L2,R2),
	preOrientationAsConstraint(P1,L1,R1,Ct),
	typeCheckVariable(L1,V),
	SInfo=t(L2,R2,V,[],Ct,term).
literalAsSimplifier(t(+,P,RS,L,R),SInfo):-
	induced_equivalence([P,RS,L,R],[P,==,L1,R1]),
	literalAsSimplifier(t(+,P,==,L1,R1),SInfo).
literalAsSimplifier(t(+,P,RS,L,R),SInfo):-
	partial_equivalence_relation_symbol( RS ),
	!,     % from here on RS is not a partial equivalence
	       % if the polarity is +
	(P=n ->
	     (    Ty=lit,
		  L1=[RS,R,L],
		  R1=[==,'$true$','$true$'],
		  Ct=[]
	     ;
	          Ty=lifted,
		  nonvar(R),
	          L1=[RS,L,X], 
		  R1=[RS,R,X],
		  Ct=[L>R]
	     ;
	          Ty=lifted,
		  nonvar(L),
	          L1=[RS,R,X],
		  nonvar(X),
		  R1=[RS,L,X],
		  Ct=[R>L]
	     )
 
	;
	     (P=l ->
		  (nonvar(R) ->
		       L1=[RS1,L,X], 
		       R1=[RS1,R,X]
		  ;
		       L1=[RS,R,L],
		       R1=[==,'$true$','$true$']
		  )

	     ;
	          (nonvar(L) ->
		       L1=[RS1,R,X],
		       R1=[RS1,L,X]
		  ;
		       L1=[RS,R,L],
		       R1=[==,'$true$','$true$']
		  )
	     ),
	     ((partial_congruence_relation_symbol(RS);RS= ==) ->
	          Ty=lifted,
		  RS1=RS
             ;
	          Ty=lit,
		  special_relation(RS1),
		  RS1\== =,
		  RS1\== ==
	     ),
	     Ct=[]
	),
	SInfo=t(L1,R1,nonvar,[],Ct,Ty).
literalAsSimplifier(t(-,_P,==,L,'$true$'),SInfo):-
	SInfo=t(L,'$min',nonvar,[],[],bool),
	!.
literalAsSimplifier(t(Pol,_,RS,L,R),SInfo):-
	equivalent_relation_symbols(RS,RS1),
	L1=[RS1,L,R],
	polarityToTruthvalue(Pol,T),
	SInfo=t(L1,[==,'$true$',T],nonvar,[],[],lit).
literalAsSimplifier(t(-,_,RS,L,R),SInfo):-
	% assuming totality resolution, 
	% this can only come from an irreflexivity axiom
	total_relation_symbol(RS),
	reflexive_closure(RS,RR),
	L1=[RR,L,R],
	SInfo=t(L1,[==,'$true$','$true$'],nonvar,[],[],lit).
literalAsSimplifier(t(Pol,P,RS,L,R),SInfo):-

	% simplification with inequations RS(L,R) is implemented
	% as conditional simplification with 

	%     RS1(R,X) ->  [Pol] RS2(L,X) == [Pol] true 
	% 
	% Ordering constraints apply.

	Pol= +,  % because unsound this way for Pol= -
	         % as contrapositive form is different from what is
                 % suggested above; need to be extended correctly
	(    P\== r,
	     L1=L,
	     R1=R,
	     composition_of_relations(RS,RS1,RS2),
	     (    Left1=[RS2,L,X]
	     ;
	          Left1=[RS2R,L,X],
		  reflexive_closure(RS2,RS2R),
		  strict_part(RS2R,RS2)
	     ),
	     AddlCond=[-,n,RS1,R,X]
	;
	     P\==l,
	     L1=R,
	     R1=L,
	     composition_of_relations(RS1,RS,RS2),
	     (    Left1=[RS2,X,R]
	     ;
	          Left1=[RS2R,X,R],
		  reflexive_closure(RS2,RS2R),
		  strict_part(RS2R,RS2)
	     ),
	     AddlCond=[-,n,RS1,X,L]
	),
	\+partial_congruence_relation_symbol( RS1 ),
	\+partial_congruence_relation_symbol( RS2 ),
	preOrientationAsConstraint(P,L1,R1,Ct),
	polarityToTruthvalue(Pol,T),
	SInfo=t(Left1,[==,'$true$',T],nonvar,[AddlCond],Ct,lit).



% warum hier nicht auch einen Fall Ri=[==,'$true$','$min']  ?

contrapositiveOrSymm(Le,Ri,Le,Ri).
contrapositiveOrSymm([RR,L,R],[==,'$true$','$true$'],L1,R1):-
	L\==R,
	reflexive_closure(RS,RR),
	total_relation_symbol(RS),
	(    L1=[RS,R,L],
	     R1=[==,'$true$','$min']
	;
	     L1= [RR,R,L],
	     R1= [=,R,L]
	).
contrapositiveOrSymm([RS,L,R],[==,'$true$','$true$'],L1,[==,'$true$','$min']):-
	total_relation_symbol(RS),
	reflexive_closure(RS,RR),
	(  L1=[RR,R,L]
	;
	   L1=[RS,R,L]
	).
contrapositiveOrSymm([RS,L,R],Ri,[RS,R,L],Ri):-
	symmetric_relation_symbol(RS),
	L\==R.


/* too weak since nonreductive rules might be useful in step n>1,
   and this is not covered by subsumption
reductivityAnalysis(Cond,Le,t(RedConds,NRedConds,ExtraVarConds)):-
	reductivityAnalysis1(Cond,Le,t(RedConds,NRedConds,ExtraVarConds)),
	!,

	(RedConds=[] ->
	     Cond=[]
	;
	     true
	).
*/

greaterSubgoal(Ty,E1,E) :-
        (Ty=lit;Ty=lifted),
	!,
	greater_literal(+,E1,+,E),  % it suffices to make the literal smaller
	containsAllVarsOf(E1,E).
greaterSubgoal(_Ty,E1,E) :-
	 greater_literal(+,E1,-,E),  
	 containsAllVarsOf(E1,E).


reductivityAnalysis(_,[],_,t([],[],[])).
reductivityAnalysis(Ty,[[Pol|E]|Cond1],Le,
                       t(RedConds,NRedConds,ExtraVarConds)):-
	(greaterSubgoal(Ty,[n|Le],E) ->
	     RedConds=[[Pol|E]|RedConds1],
	     ExtraVarConds=ExtraVarConds1,
	     NRedConds=NRedConds1
%   that case would require reductivity checking at simplification time
%   essentially the Inst component is always empty right now
%	;subset_vars(E,Le) ->
%	     RedConds=RedConds1,
%	     NRedConds=[[Pol|E]|NRedConds1],
%	     ExtraVarConds=ExtraVarConds1
	;
	     RedConds=RedConds1,
	     NRedConds=NRedConds1,
	     ExtraVarConds=[[Pol|E]|ExtraVarConds1]
	),
	reductivityAnalysis(Ty,Cond1,Le,t(RedConds1,NRedConds1,ExtraVarConds1)).


	  



preOrientationAsConstraint(n,L,R,[L>R]):-
	!,
	containsAllVarsOf(L,R).
preOrientationAsConstraint(_,_,_,[]).




containsReducer(E):-
	member(t([_,_],_,t(_,_,_,_,_,t([],[],[]))),E),
	!.

containsContextualReducer(E):-
	member(t([_,_],_,_),E),
	!.

containsSubsumer(E):-
	member(t(subsumption,_,_),E).


hasCondition([],unconditional).
hasCondition([_|_],conditional).	



hasConstraint(Top,t(_,[],[]),orientable):-
	true(Top),
	!.
hasConstraint(_,_,nonorientable).



typeCheckVariable(X,_):-
	var(X),
	!.
typeCheckVariable(_X,nonvar).


polarityToTruthvalue(+,'$true$').
polarityToTruthvalue(-,'$min').




instantiateExtraVars(t(I,L,t(N,R,RR,V,OrdC,C)),t(I,L1,t(N,R2,RR,V,OrdC1,C1))):-
	Def=(C,L),
	Inst=R,
	free_vars(Def,VL),
	free_vars(Inst,VR),
	(setof(X,(member(X,VR),\+member212(X,VL)),Extra)->
	     copy_term((L,R,C,OrdC,Extra),(L1,R1,C1,OrdC1,Extra1)),
%	     (is_operator(K/0) -> true; K='$min'),
	     K='$min',
	     find_value_of_extra_vars([K],Extra1,[])
	;    
	     L1=L,
	     R1=R,
	     C1=C,
	     OrdC1=OrdC
	),
	insert_shifts(R1,R2).




equalNumber(T1,T2):-
	arg(1,T1,N),
	arg(1,T2,N).



combineOC([],C,C):-
	!.  
combineOC([S>T|_A],_B,F):-
	greaterEqual(T,S),
	!,
	false(F).
combineOC([S>T|A],B,C):-
	and_gt(B,[S,T],B1),
	combineOC(A,B1,C).

irredIndicators(T,TR):-
	copy_term(T,T1),
	irredInd(T1,TR).

irredInd('$','$'):-    % irreducibility at or below variables 
	!.
irredInd(T,TR):-
	mapargs(irredInd,T,TR).
