
:- class(rule,clause,
	 [size,
	  weight,
	  status,
	  mark,
	  antecedentOrdered,
	  succecentOrdered,
	  moduleNumber,
	  literals,
	  indexEntries],

	 [
	 active_term/2,
	 antecedent/2,
	 succedent/2,
	 size/2,
	 size/3,
	 weight/2,
	 weight/3,
	 weightBound/2,
	 weight_with_bound/2,
	 maximal_antecedent/2,
	 non_maximal_antecedent/2,
	 maximal_succedent/2,
	 non_maximal_succedent/2,
	 moduleNumber/2,
	 selectedLiteral/2,
	 makeOld/2,
	 makeNew/2,
	 incr_weight_bound/2,
	 maximal_antecedent/3,
	 non_maximal_antecedent/3,
	 maximal_succedent/3,
	 non_maximal_succedent/3,
	 moduleNumber/3,
	 literals/3,
	 selectSelectedLiteral/3,
	 indexEntries/2,
	 indexEntries/3,
	 is_new/1,
	 is_brand_new/1,
	 is_old/1,
	 hasSelectedLiteral/1,
	 small_rule/1,
	 to_clause/2,
	 simplifier/5,
	 simplifier/6,
	 merge_two_rules_into_clause/3,
	 selectInAntecedent/3,
	 selectInSuccedent/3,
	 selectMaxInAntecedent/3,
	 selectMaxInSuccedent/3
	 ]

). 




:- use_module(mylists,[is_list/1]).
:- use_module(props,[firstOrder/0,search_depth/1,scaled_clause_weight/5]).
:- use_module(library(maplist),[scanlist/4]).
:- use_module(dynamicCode,[recordPrologClause/2,known_proposition/1]).
:- use_module(adt_atom,[is_equational_atom/1,orient_atom_fresh/2]).
:- use_module(rewrite, [simplifier/4,injectivitySimplifiers/4]).




% -----------------------------------------------------------------------------
% -----------------------------------------------------------------------------
% 
%
% rules have the instance variables
%
%  size, weight, status, mark, antecedentO, succedentO, module_number, literals
%

% where
%
% size ::- nat
% weight ::- nat
% status ::- {old} \cup N
%   The status indicates what inferences have been computed with a rule.
%   If it is ``old'', all inferences have been computed.
%   If it is a number n, then all inferences with rules up to size n
%   have been computed
% mark ::- {0,1} ( 0: not marked, 1: marked )
% antecedentO, succedentO ::- [maximals,non_maximals]
% maximals,non_maximals ::- [atom]
% module_number ::- nat
% literals ::- [literal]
%   literals is only defined correctly for rules that are stored
%   in the database under predicate rule/1 if they have been recorded
%   using recordRule.  (???)
% -----------------------------------------------------------------------------
% -----------------------------------------------------------------------------




% creation
% -----------------------------------------------------------------------------

%------------------------------------------------------------------------------
% create a rule from a clause:
% 	- maximal term in every atom
%       - determine (possible) maximal atoms in the antecedent
%       - determine (possible) maximal atoms in the succedent
%	- the rule becomes new
%------------------------------------------------------------------------------
createRule(Cl, Ru, STI, STO):-
	createRule(Cl, Ru1),
	generateSimplifiers(Ru1,Ru,STI,STO).


createRule(Cl, myself(Cl3,[Z,W,0,0,[Am1,Ap1],[Sm1,Sp1],_,L,Entries])):-
	% no selection at first
	antecedent( Cl, A ),
	succedent( Cl, S ),

	kind( Cl, K ),
	maplist(atom_as_term,A,AT),
	maplist(atom_as_term,S,ST),
	analyse_rule(AT,ST,K1),
	select_kind(K,K1,K2),
	kind(K2,Cl,Cl1),

	size_and_weight( Cl1, Z, W ),
	
	maplist(orient_atom_fresh,A,A1),
	maplist(orient_atom_fresh,S,S1),
        search_maximal_candidates( A1, S1, Am1, Ap1, Sm1, Sp1 ),

	sequent_to_literal_list([A1,S1],L),

	(flag_on( ci, orde ),
	 \+ flag_on( ci, equa )
%	      , \+ is_literal(Cl1)   % this produces unsoundness if constraints
                                     % have a logical meaning
	          ->
			   get_term_vars((A,S),V),
			   constraint(Cl1,Ct1),
			   project(V,Ct1,Ct)
			   
	;
	                   true(Ct)
	),
	(theory_flag(Cl1,(goal)) ->
	     operators(Cl1,Ops),
	     recordGoalOperators(Ops)
	;
	     true
	),
	constraint(Ct,Cl1,Cl2),

	recordNewRule,
	number(Cl2,N),
	% delete literals in super-object
	antecedent([],Cl2,Cl4),
	succedent([],Cl4,Cl3), 
	!.



generateSimplifiers(Ru,Ru1,STI,STO):-
	insertSimplifierIntoIndex(Ru,STI,STO,Entries),
	indexEntries(Entries,Ru,Ru1),
	!.






% create a rule from a clause and select some negative literal in it
% 	in a rule with selection there is no computation of maximal candidates
% 	and all the atoms are considered as non maximal.
% 	the selected atom is the first one
% The first parameter determines the selection strategy.
% maxs: select a maximal negative literal, if there is no unifomly  maximal
%       positive literal
% poss: positive strategy
% a number indicates the literal to be selected
%------------------------------------------------------------------------------

createRuleWithSelection( poss, Cl, Ru2, STI, STO ):-
	createRule( Cl, self(_C,[Z,W,0,0,[Am1,Ap1],[Sm1,Sp1]|C])),
	Ru=self(_C,[Z,W,0,1,[Am,Ap],[Sm,Sp]|C]),
	kind(Ru,K),
	selection(K,Am1,Ap1,Sm1,Sp1,Am,Ap,Sm,Sp,1),
	generateSimplifiers(Ru,Ru2,STI,STO).
createRuleWithSelection(maxs,Cl,Ru2,STI,STO):- 
	createRule( Cl, Ru1 ),
	Ru1=self(_C,[Z,W,0,0,[Am1,Ap1],[Sm1,Sp1]|C]),
	Ru=self(_C,[Z,W,0,M,[Am,Ap],[Sm,Sp]|C]),
	kind(Ru,K),
	selection(K,Am1,Ap1,Sm1,Sp1,Am,Ap,Sm,Sp,M),
	generateSimplifiers(Ru,Ru2,STI,STO). 
createRuleWithSelection(MK,Cl,Ru2,STI,STO ):- 
	Ru=myself(Cl2,[Z,W,0,1,[[],A1],[[],S],_,L,[]]),
	number(MK),
	antecedent( Cl, A ),
	put_marked_eq_in_first_place( MK, A, A1 ),
	succedent( Cl, S ),
	sequent_to_literal_list([A1,S],L),
	size_and_weight( Cl, Z, W ),
	antecedent([],Cl,Cl4),
	succedent([],Cl4,Cl2),
	generateSimplifiers(Ru,Ru2,STI,STO),
	!.


delete_rules_no_dependents([],ST,ST).
delete_rules_no_dependents([R|Rs],ST,STO):-
	number(R,N),
	deleteRule(N,R),
	indexEntries(R,Entries),
	delete_terms(Entries,ST,ST1),
	delete_rules_no_dependents(Rs,ST1,STO).

	


delete_rule(R):-
	number(R,N),
	delete_rule(N,R).



% deletion with deletion of dependent *clauses*
% that's ok as long as we don't use clauses for redundancy proofs
% right now we're using rules only
% -----------------------------------------------------------------------------
delete_rule(N,R):-
	deleteRule(N,R),
	delete_dependents(N).

delete_rule(N,R,ST,STO):-
	deleteRule(N,R),
	indexEntries(R,Entries),
	delete_terms(Entries,ST,STO),
	delete_dependents(N).

delete_rule(R,ST,STO):-
	number(R,N),
	delete_rule(N,R,ST,STO).


change_rule(N,R,Am,Ap,Sm,Sp,ST,STO):-
	deleteRule(N,_R),    % R == _R up to renaming, but we need the R copy
	append(Am,Ap,A),
	append(Sm,Sp,S),
	antecedent(A,R,R1),
	succedent(S,R1,R2),
	maximal_succedent(Sm,R2,R3),
	maximal_antecedent(Am,R3,R31),
	literals(R31,Ls),
	literals(Ls,R31,R32),
	size_and_weight(R32,Z,W),
	size(Z,R32,R33),
	weight(W,R33,R34),
	indexEntries(R,Entries),
	delete_terms(Entries,ST,ST1),
	indexEntries(_,R34,R4),
	generateSimplifiers(R4,R5,ST1,STO),
	recordRule(N,R5).

	


% accessing
% -----------------------------------------------------------------------------

% The cut is essential to avoid calling the more general definition
% of size in super class $simpleClause upon backtracking.
size(self(_R,[Z|_]),Z):-
	!.

weight(self([_,W|_]),W).
	

weightBound( self([_,_,B|_]), B ).


% not used right now; new definition in saturation.pl!
weight_with_bound( R, WB ):-
	R =  self([_,W,B|_]),
%	proof_depth(R,D),
	search_depth(R,D),
	theory_flag(R,T),
%	(reductive(R) ->
%	     RedFlag=reductive
%	;
%	     RedFlag=nonReductive
%	),
	(orientableEquivalence(R) ->
	    K=orientableEquivalence
	;
	    kind(R,K)
	),
	scaled_clause_weight(W,D,T,K,WB).
%	scaled_clause_weight(W,D,B,T,RedFlag,K,WB).




antecedent( self([_,_,_,_,[Am,Ap]|_]), A ):-
	!,
	append(Am,Ap,A).

maximal_antecedent( self([_,_,_,_,[Am,_A]|_]), Am ).

non_maximal_antecedent( self([_,_,_,_,[_Am,A]|_]), A ).

succedent( self([_,_,_,_,_,[Sm,Sp]|_]), S ):-
	!,
	append(Sm,Sp,S).

maximal_succedent( self([_,_,_,_,_,[Sm,_S]|_]), Sm ).

non_maximal_succedent( self([_,_,_,_,_,[_Sm,S]|_]), S).


max_a_min_a_max_s_min_s(self([_,_,_,_,[Am,Ap],[Sm,Sp]|_]),Am,Ap,Sm,Sp).


moduleNumber( self([_,_,_,_,_,_,M|_]), M ).

selectedLiteral( self([_,_,_,1,[[],[E|_]]|_]), E ).


indexEntries(Ru,Entries):-
	Ru=self([_Z,_W,_ON,_MK,_A,_S,_M,_L,Entries|_]),
	(var(Entries) ->
	     simplifierEntries(Ru,Entries),
	     (hasTermSimplifier(Entries) ->
		  setRewriteFlag(term_equations_exist),
		  setRewriteFlag(equations_exist)
	     ;
	          true
	     ),
	     (hasBoolSimplifier(Entries) ->
		  setRewriteFlag(equations_exist),
		  (firstOrder ->
		       setRewriteFlag(term_equations_exist)
		  ;
		       true
		  ),
		  setRewriteFlag(bool_equations_exist)
	     ;
	          true
	     ),
	     printEntries(Entries)
	;
	     true
	),
	!.

	




% literal access
% -----------------------------------------------------------------------------

%------------------------------------------------------------------------------
% gives one atom in the antecedent
%------------------------------------------------------------------------------

selectInAntecedent( R, E, R1 ):-
        maximal_antecedent( R, Am ),
        select( E, Am, Am1 ),
        maximal_antecedent( Am1, R, R1 ).

selectInAntecedent( R, E, R1 ):-
        non_maximal_antecedent( R, Am ),
        select( E, Am, Am1 ),
        non_maximal_antecedent( Am1, R, R1 ).

%------------------------------------------------------------------------------
% gives one maximal atom in the antecedent
%------------------------------------------------------------------------------

selectMaxInAntecedent( R, E, R1 ):-
        maximal_antecedent( R, Am ),
        select( E, Am, Am1 ),
        maximal_antecedent( Am1, R, R1 ).

%------------------------------------------------------------------------------
% gives one atom in the succedent
%------------------------------------------------------------------------------

selectInSuccedent( R, E, R1 ):-
        maximal_succedent( R, Sm ),
        select( E, Sm, Sm1 ),
        maximal_succedent( Sm1, R, R1 ).
selectInSuccedent( R, E, R1 ):-
        non_maximal_succedent( R, S ),
        select( E, S, S1 ),
        non_maximal_succedent( S1, R, R1 ).

%------------------------------------------------------------------------------
% gives one maximal atom in the succedent
%------------------------------------------------------------------------------


selectMaxInSuccedent( R, E, R1 ):-
        maximal_succedent( R, Sm ),
        select( E, Sm, Sm1 ),
        maximal_succedent( Sm1, R, R1 ).




active_term(R,T):-
	maximal_succedent(R,S),
	member(E,S),
	active_term_in_atom(E,T).
active_term(R,T):-
	maximal_antecedent(R,S),
	member(E,S),
	active_term_in_atom(E,T).
active_term(R,T):-
	selectedLiteral(R,E),
	active_term_in_atom(E,T).




% modifying
% -----------------------------------------------------------------------------


makeOld( self(_R,[Z,W,_|R]), self(_R,[Z,W,old|R]) ).

%makeNew( self(_R,[Z,W,_|R]), self(_R,[Z,W,0|R]) ).
makeNew( self(_R,[Z,W,_|R]), self(_R,[Z,W,1|R]) ).

weight(W,self(_R,[Z,_W|R]),self(_R,[Z,W|R])).
	
size(Z,self(_R,[_Z|R]),self(_R,[Z|R])).
	
incr_weight_bound( self(_R,[Z,W,B|R]), self(_R,[Z,W,B1|R]) ):-
	number(B),
	segment_length(L),
	B1 is B+L,
	!.
incr_weight_bound( R, R1 ):-
	makeNew(R,R1).

maximal_antecedent( 
	Am1, self(_R,[Z,W,ON,MK,[_,A]|R]), self(_R,[Z,W,ON,MK,[Am1,A]|R])).

non_maximal_antecedent(
	A1, self(_R,[Z,W,ON,MK,[Am,_]|R]), self(_R,[Z,W,ON,MK,[Am,A1]|R])).

maximal_succedent(
	Sm1, self(_R,[Z,W,ON,MK,A,[_,S]|R]), self(_R,[Z,W,ON,MK,A,[Sm1,S]|R]) ).

non_maximal_succedent(
	S1, self(_R,[Z,W,ON,MK,A,[Sm,_]|R]), self(_R,[Z,W,ON,MK,A,[Sm,S1]|R]) ).

moduleNumber( M, self(_R,[Z,W,ON,MK,A,S,_|R]), self(_R,[Z,W,ON,MK,A,S,M|R]) ).

literals( L, self(_R,[Z,W,ON,MK,A,S,M,_|R]), self(_R,[Z,W,ON,MK,A,S,M,L|R]) ).

indexEntries(E,self(_R,[_Z,_W,_ON,_MK,_A,_S,_M,_L,_|Ru]),self(_R,[_Z,_W,_ON,_MK,_A,_S,_M,_L,E|Ru])).


selectSelectedLiteral(
	self(_R,[Z,W,ON,1,[[],[E|R]]|Ru]), E, self(_R,[Z,W,ON,1,[[],R]|Ru]) ).


% testing
% -----------------------------------------------------------------------------


is_new( R):-
	\+ is_old(R).

is_brand_new( self([_,_,0|_]) ).


is_old( self([_,_,old|_]) ).



hasSelectedLiteral( self([_,_,_,1|_]) ).


reductive(R):-
	\+ hasSelectedLiteral(R),
	(  maximal_succedent(R,[]),
	   maximal_antecedent(R,[A])
	;
	   maximal_succedent(R,[A]),
	   maximal_antecedent(R,[])
	),
	get_atom_preorientation(A,P),
	(P= l; P= r),
	!.





small_rule(R):-
	is_atom(R),
	size(R,S),
	small_rule_size(Sm),
	S=<Sm.

/* ----------------------------------------------------------------------------

We attempt to predict produced atoms.
At present only for purely propositional atoms.

---------------------------------------------------------------------------- */

productivity_analysis(Cl):-
	\+ hasSelectedLiteral(Cl),
	\+ flag_on(rp,prop),
	maximal_antecedent(Cl,[]),
	maximal_succedent(Cl,[E]),
	non_maximal_succedent(Cl,S),
	maplist(is_proposition,[E|S]),
	as_Prolog_clause(Cl,CP,Props),
	recordPrologClause(CP,Props),
	!.
productivity_analysis(_Cl).



blocked(_Ru):-
	flag_on(rp,prop),   % need to do propositional resolution
	% need to think mor about this
	% for blocking
	!,
	fail.
blocked(_Ru):-
	search_depth(D),
	D<10,
	!,  % otherwise too incomplete
	fail.

/* 
blocked(_Ru):-
	flag_on(es,vare),   % counterproductive with clause splitting
	  % as we want to merge split cases
	!,
	fail.
*/
blocked(Ru):-
	antecedent(Ru,A),
	member(E,A),
	is_proposition(E),
	atom_as_term(E,T),
	known_proposition(T),
	\+ dynamicCode:T,
	!.
blocked(Ru):-
	non_maximal_succedent(Ru,A),
	member(E,A),
%	\+ is_equational_atom(E),
%	is_proposition(E),
	atom_as_term(E,T),
	\+ predicate_property(T,built_in),
	\+ \+ (numbervars(T,0,_),dynamicCode:T),
	!.



% converting
% -----------------------------------------------------------------------------




to_clause( self(Cl,Ru),Cl3):-
	antecedent(self(Cl,Ru),A),
	succedent(self(Cl,Ru),S),
	history(self(Cl,Ru),H),
	constraint(self(Cl,Ru),Ct),
	number(self(Cl,Ru),N),
	createClause([A,S,Ct],Cl1),
	history(H,Cl1,Cl2),
	number(N,Cl2,Cl3),
	!.

	

as_Prolog_clause(C,(Head:-Body),[HP|Props]):-
	C=self(_),
	maximal_succedent(C,[E]),
%	is_proposition(E),
%	\+ is_equational_atom(E),
	atom_as_term(E,Head),
	\+ predicate_property(Head,built_in),
	non_maximal_succedent(C,S),
	maximal_antecedent(C,[]),
	non_maximal_antecedent(C,A),
	maplist(atom_as_term,S,ST),
	maplist(prolog_negation,ST,B2),
	maplist(atom_as_term,A,B1),
	append(B1,B2,B3),
	append(ST,B1,Props),
	functor(Head,HP,_),
	scanlist(prolog_and,B3,true,Body).



prolog_negation(A,(\+ A)).

prolog_and(A,B,(A,B)).
	


% reduction and narrowing
% -----------------------------------------------------------------------------


/*
-------------------------------------------------------------------------------

simplifierEntries(+Ru,?Entries)

is true when Entries is the list of simplifier entries for the rule
Ru.

The main working horse for this predicate is

rewrite:simplifier/4


-------------------------------------------------------------------------------
*/


simplifierEntries(Ru,Entries):-
	kind(Ru,injectivity(_F,_N,_I)),
	!,
	antecedent(Ru,[EF]),
	succedent(Ru,[E]),
	number(Ru,N),
	injectivitySimplifiers(N,EF,E,Entries),
	productivity_analysis(Ru).
simplifierEntries(Ru,Entries):-
	(dont_demodulate_with(Ru) ->
	     Entries=[]
	;
	     number(Ru,N),
%	     (N=87 -> trace; true),
	     constraint(Ru,Ct),
	     maximal_antecedent(Ru,Am),
	     non_maximal_antecedent(Ru,Ap),
	     maximal_succedent(Ru,Sm),
	     non_maximal_succedent(Ru,Sp),
	     (hasSelectedLiteral(Ru) ->
		  Sel=hasSelectedLit
	     ;
	          Sel=nothingSelected
	     ),
	     setof(Entry,simplifier(t(Am,Ap,Sm,Sp,Sel),N,Ct,Entry),Entries),
	     productivity_analysis(Ru)
	).





hasTermSimplifier(Entries) :-
	member(t(_,LS,_),Entries),
	\+ LS=[_,_,_],   % otherwise atom simplifier
	(    var(LS)
	;
	     functor(LS,F,N),
	     \+ is_predicate(F/N)
	),
	!.

hasBoolSimplifier(Entries) :-
	member(t(_,LS,_),Entries),
	\+ LS=[_,_,_],   % otherwise atom simplifier
	nonvar(LS),
	functor(LS,F,N),
	is_predicate(F/N),
	!.



% should be eliminated
simplifier( N, Op, =, L, R ):-
	rule(N,Ru),
	indexEntries(Ru,Entries),
	member([Index,L,[_,N,_,R]],Entries),
	(Index=orientable_units_index ->
	      Op=l
	;
	      Op=n
	).
	

simplifier( N, Op, RS, L, R, [[],[]] ):-
	simplifier( N, Op, RS, L, R ).
simplifier( N, Op, RS, L, R, [Ap,Sp] ):-
	rule(N,Ru),
	Ru=self([_,_,_,_,[[],Ap],[[E],Sp]|_]),
	(Ap\==[];Sp\==[]),
	atom_for_demodulation( E, Op, RS, L, R ).




% binary operations
% -----------------------------------------------------------------------------

merge_two_rules_into_clause( R1, R2, Cl ):-
	antecedent( R1, A1 ),
	antecedent( R2, A2 ),
	(hyper(R2) ->
	     mark_antecedent(A2,A2M),
	     append( A2M, A1, A )   % essential for goal-oriented resolution
	                        % for horn clauses; A2 are the remaining goals
	                        % which should be tried first.
	;
	     append( A1, A2, A )
	),
	
	succedent( R1, S1 ),
	succedent( R2, S2 ),
	append( S1, S2, S ),
	createClause( [A,S], Cl ).


hyper(_R):-
	horn.
hyper(R):-
	hasSelectedLiteral(R),
	kind(R,K),
	hykind(K).

hykind(anti_symmetry(_,ER)):-
	congruence_relation_symbol(ER).
hykind(composition(R,R,R)).


select_kind(regular,K,K):-
	!.
select_kind(_,K,K):-
	K\==regular,
	!.
select_kind(K,_,K).



%	printing
%------------------------------------------------------------------------------

:- mymultifile user:portray/1.

user:portray( X ):-
	is_object(X,self(R1,R2)),
	R2=[_,_,_,_,[Am,Ap],[Sm,Sp]|_],
	maplist(nonvar,[Am,Ap,Sm,Sp]),  % sufficiently instantiated
	classes:printRule( self(R1,R2) ).
/*

user:portray('$simplifierEntries'([])).
user:portray('$simplifierEntries'([E|Es])):-
	user:portray('$simplifierEntry'(E)),
	user:portray('$simplifierEntries'(Es)).
user:portray('$simplifierEntry'(t(Index,L,t(N,R,_RR,_V,Ct,CondA)))):-
	    format('~n  rule ~p : ~p -> ~p if ~p ~n          [I=~p, Ct=~p]', [N,L,R,CondA,Index,Ct]).


*/



printRule( Ru ):-
	Ru=self([_,_,_,M|_] ),
	(blocked(Ru) ->
	     write('-')
	;M==1 -> 
	     write('*')
	;true),
	to_clause(Ru,Cl),
	print(Cl).




% private
% -----------------------------------------------------------------------------


% this should be unified with analyse_clause

analyse_rule(A,S,commutativity(O,R)):-
	commutativity_clause(A,S,O,R),
	congruence_relation_symbol(R),
	(ac(off) ->
	     newCOperator(O)
	;
	     true % need to revert everything for this,otherwise
	).
analyse_rule(A,S,associativity(O,R)):-
	associativity_clause(A,S,O,R),
	congruence_relation_symbol(R),
	(ac(off) ->
	     newAOperator(O)
	;
	     true % need to revert everything for this,otherwise
	).
analyse_rule(A,S,permutativity(O,X,Y)):-
	permutativity_clause(A,S,O,X,Y).
analyse_rule(A,S,composition(R,R,R)):-
	analyse_clause(A,S,composition(R,R,R)).
analyse_rule(A,S,reflexivity(R)):-
	analyse_clause(A,S,reflexivity(R)),
	newReflRelation(R).
analyse_rule(A,S,finite_domain(D,R)):-
	analyse_clause(A,S,finite_domain(D,R)),
	congruence_relation_symbol(R).
analyse_rule(_,_,regular).



small_rule_size(15).


reductive_for(Ru,Max,Minors):-
	\+ hasSelectedLiteral(Ru),
	max_a_min_a_max_s_min_s(Ru,Am,Ap,Sm,Sp),
	(Am=[E],Sm=[] ->
	     Max=[-|E],
	     sequent_to_literal_list([Ap,Sp],Minors)
	;Am=[],Sm=[E] ->
	     Max=[+|E],
	     sequent_to_literal_list([Ap,Sp],Minors)
	).


put_marked_eq_in_first_place( MK, A, A2 ):-
	find_eq( MK, A, E ),
	delete_one312( E, A, A1 ),
	search_greater_terms( [E], L ),
	append( L, A1, A2 ), !.

find_eq( 1, [E|_], E ):- !.
find_eq( MK, [_|R], E ):-
	MK1 is MK-1,
	find_eq( MK1, R, E ), !.

search_greater_terms( [], [] ):- !.
search_greater_terms( [E1|L1], [E2|L2] ):-
	orient_atom( E1, E2 ),
        search_greater_terms( L1, L2 ).





delete_dependents(N):-
	dependent(N,Cl),
	deleteClause(Cl),
	fail.
delete_dependents(_).

dependent(N,Cl):-
        clause(Cl),
	history(Cl,H),
	get_first_history(H,M),
	(M==N -> 
	     true
	;get_second_history(H,K),
	 K==N
	).


sequent_to_literal_list([A,S],CL):-
	maplist(neg_atom_as_literal,A,AL),
	maplist(pos_atom_as_literal,S,SL),
	append(AL,SL,CL),
	!.



printEntries(_E):-
%	nl,nl,
%	print('$simplifierEntries'(_E)),
%	nl,nl,
	true.



