:- class(simpleClause,object,
	 % instance variables
	 [number,
	  antecedent,
	  succedent],

	 % (re-)defined methods

	 [
	 asSequent/2,
	 asSimpleClause/2,
	 number/2,
	 number/3,
	 antecedent/2,
	 antecedent/3,
	 succedent/2,
	 succedent/3,
	 indexEntries/2,
	 is_atom/1,
	 is_equation/1,
	 is_literal/1,
	 is_trivial/1,
	 empty/1,
	 vars/2,
	 operators/2,
	 size/2,
	 size_and_weight/3,
	 terms/2,
	 to_clause/2,
	 terms/2,
	 literals/2,
	 literals/3,
	 add_antecedent/3,
	 add_succedent/3,
	 add_list_antecedent/3,
	 add_list_succedent/3,
	 search_maximal_candidates/6,
	 selectInAntecedent/3,
	 selectInSuccedent/3,
	 size_weight_vars/4
	 ]).




% -----------------------------------------------------------------------------
% -----------------------------------------------------------------------------
%
% simple clauses are pairs of lists of atoms
%
% -----------------------------------------------------------------------------
% -----------------------------------------------------------------------------



% creation
% -----------------------------------------------------------------------------

createSimpleClause( [A,S], myself(_,[_,A,S])).


createSimpleClauseFromLiterals(Ls,C):-
	literal_list_to_sequent(Ls,A,S),
	elim_equivalences(A,S,A1,S1),
%	maplist(create_atom,A1,AA),
%	maplist(create_atom,S1,SS),
	createSimpleClause([A1,S1],C).


% accessing
% -----------------------------------------------------------------------------
antecedent(self([_,A,_]), A ).

succedent( self([_,_,S]), S ).

number( self([N|_]), N ).


literals(C,L):-
	C=self(_),
	antecedent(C,A),
	succedent(C,S),
	sequent_to_literals(A,S,L),
	!.






% modifying
% -----------------------------------------------------------------------------

antecedent(A, self(C,[N,_,S]), self(C,[N,A,S])).

succedent( S, self(C,[N,A,_]), self(C,[N,A,S])).

number( N, self(_C,[_|C]), self(_C,[N|C]) ).

literals(CL,self(_C,[N,_A,_S]),self(_C,[N,A,S])):-
	literal_list_to_sequent(CL,A,S).



% converting
% -----------------------------------------------------------------------------


asSimpleClause(C,D):-
	antecedent(C,A),
	succedent(C,S),
	createSimpleClause([A,S],D).
	
asSequent(C,[A2,S2]):-
	antecedent(C,A1),
	succedent(C,S1),
	maplist(atom_as_term,A1,A2),
	maplist(atom_as_term,S1,S2).

	



to_clause( C,C ):-
	C=self(_).



% testing
% -----------------------------------------------------------------------------

is_atom(C):-
	antecedent(C,[]),
	succedent(C,[_]).

is_equation( C ):-
	antecedent(C,[]),
	succedent(C,[E]),
	get_atom_relation_symbol(E,R),
	partial_equivalence_relation_symbol(R),
	(R= == ->
	     get_atom_members(E,L,R),
	     R\== '$true$',
	     L\== '$true$'
	;
	     true
	).



is_literal(C):-
	antecedent(C,[_]),
	succedent(C,[]),
	!.
is_literal(C):-
	is_atom(C).

is_trivial(C):-
	succedent( C, S ),
	(  succedentTrivial(S)
	;
	   antecedent( C, A ),
	   (   antecedentTrivial(A)
	   ;
	       member(E,A),
	       member212(E,S)
	   )
	),
	!.

	
empty(C):-
	antecedent(C,[]),
	succedent(C,[]).

orientableEquivalence(C):-
	antecedent(C,[]),
	succedent(C,[E]),
	adt_atom:equivalence(E). % at this point we don't orient


% clause structure
% -----------------------------------------------------------------------------


vars( C, L ):-       
	succedent( C, S ),
	antecedent( C, A ),
	append( A, S, T ),
	get_list_atoms_vars( T, L ). 

terms( C, Ts ):-
       antecedent(C,A),
       succedent(C,S),
       maplist(get_atom_terms,A,TA),
       maplist(get_atom_terms,S,TS),
       append(TA,TS,TTs),
       flatten(TTs,Ts).



operators( C, L2 ):-
	succedent( C, S ),
	antecedent( C, A ),
	get_list_of_atoms_operators( A, LA ),
	get_list_of_atoms_operators( S, LS ),
	append( LA, LS, L1 ),
	sort( L1, L2 ), 
	!.

groundClause(C):-
	antecedent(C,A),
	free_vars(A,[]),
	succedent(C,S),
	free_vars(S,[]).

monadicClause(C):-
	antecedent(C,A),
	succedent(C,S),
	free_vars((A,S),V),
	(V=[];V=[_]),
	!.

propositionalClause(C):-
	succedent( C, S ),
	antecedent( C, A ),
	append( A, S, T ),
	maplist(is_proposition,T).





% clause size and weight
%------------------------------------------------------------------------------
size( C, Z):-
	succedent( C, S ),
	antecedent( C, A ),
	term_size( A, Z1 ),
	term_size( S, Z2 ),
	Z is Z1+Z2,
	!.
/*

size_and_weight( C, Z, W ):-
	succedent( C, S ),
	antecedent( C, A ),
	term_info( A, Z1, D1, V1 ),
	length(V1,N1),
	term_weight( (Z1,D1,N1), W1 ),
	length(S,SN),
	length(A,AN),
	CN is AN+SN,
	term_info( S, Z2, D2 ,V2 ),
	length(V2,N2),
	D2SN is D2+CN-1,
	term_weight( (Z2,D2SN,N2), W2 ),
	Z is Z1+Z2,
	W3 is W1+W2,
	((orientableEquivalence(C); proof_depth(C,1)) ->
	      % equivalences and axioms are preferred candidates
	      % for inference computation
	      W is W3/2
	;
	      W=W3
	).

*/

size_weight_vars( C, Z, W , V):-
	succedent( C, S ),
	antecedent( C, A ),
	term_info( [S,A], Z, D, V ),
	length(V,N),
	term_weight( (Z,D,N), W ).

size_and_weight( C, Z, W ):-
	size_weight_vars( C, Z, W , _V).


%	length(S,SN),
%	length(A,AN),



/*
	(reduced_but_large(C,W4,W) ->
	     true
	;
	     W=W4
	),
	!.

reduced_but_large(C,W,WC):-
	history(C,H),
	get_key_history(H,cdem),
	get_first_history(H,C1),
	size_and_weight(C1,_,W1),
	(W>=W1 ->
	     WC is W1 - 4
	;
	     WC=W
	).
*/

% logical operations
%------------------------------------------------------------------------------


add_antecedent( E, C1, C2 ):-
	antecedent( C1, S1 ),
	antecedent( [E|S1], C1, C2 ).

add_succedent( E, C1, C2 ):-
	succedent( C1, S1 ),
	succedent( [E|S1], C1, C2 ).


add_list_antecedent( A, C1, C2 ):-
	antecedent( C1, S1 ),
	append(A,S1,S2),
	antecedent( S2, C1, C2 ).

add_list_succedent( S, C1, C2 ):-
	succedent( C1, S1 ),
	append(S,S1,S2),
	succedent( S2, C1, C2 ).



selectInAntecedent( C, E, C1 ):-
        antecedent( C, Am ),
        select( E, Am, Am1 ),
        antecedent( Am1, C, C1 ).

selectInSuccedent( C, E, C1 ):-
        succedent( C, Sm ),
        select( E, Sm, Sm1 ),
        succedent( Sm1, C, C1 ).
	


elim_equivalences([],[],[],[]).
elim_equivalences(A,[E|S],[E1|A1],[E2|S2]):-
	get_atom(E,==,L,R),
	R\=='$true$',
	!,
	create_atom(L=='$true$',E1),
	create_atom(R=='$true$',E2),
	elim_equivalences(A,S,A1,S2).
elim_equivalences(A,S,A1,S1):-
	elim_special_relations(A,S,A1,S1).

elim_special_relations([],[],[],[]).
elim_special_relations(A,[E|S],A1,[E1|S2]):-
	get_atom(E,RS,L,R),
	RS\== ==,
	partial_equivalence_relation_symbol(RS),
	reflexive_relation_symbol(RS),
	\+ congruence_relation_symbol(RS),
	congruence_relation_symbol(RS1),
	!,
	L1=..[RS,L],
	R1=..[RS,R],
	create_atom(L1,RS1,R1,E1),
	elim_special_relations(A,S,A1,S2).
elim_special_relations([E|A],S,[E1|A1],S2):-
	get_atom(E,RS,L,R),
	RS\== ==,
	partial_equivalence_relation_symbol(RS),
	reflexive_relation_symbol(RS),
	\+ congruence_relation_symbol(RS),
	congruence_relation_symbol(RS1),
	!,
	L1=..[RS,L],
	R1=..[RS,R],
	create_atom(L1,RS1,R1,E1),
	elim_special_relations(A,S,A1,S2).
elim_special_relations(A,[E|S],A1,[E|S2]):-
	elim_special_relations(A,S,A1,S2).
elim_special_relations([E|A],S,[E|A1],S2):-
	elim_special_relations(A,S,A1,S2).


	
%	rewriting
%------------------------------------------------------------------------------




indexEntries(C,Entries):-
	C=self([_N,A,S]),
	N=9999999,
	true(Ct),
	setof(Entry,simplifier(t([],A,S,[],nothingSelected),N,Ct,Entry),Entries).

	

%	maximal literals
%------------------------------------------------------------------------------

%search_maximal_candidates( A, S, AO, [], SO, [] ):-
%	free_vars(A,[]),
%	free_vars(S,[]),
%	maplist:map_list(orient_atom,A,AO),
%	maplist:map_list(orient_atom,S,SO),
%	!.
search_maximal_candidates( A, S, Am2, Ap, Sm2, Sp ):-
        ant( [], A, S, Am1, Ap1, Sm1, Sp1 ),
        suc( Am1, [], Sm1, Am2, Ap2, Sm2, Sp2 ),
        append( Ap1, Ap2, Ap ),
        append( Sp1, Sp2, Sp ), !.

ant( AI, [E|AD], S, Am1, Ap, Sm1, Sp ):-
	orient_atom(E,EO),
        smallerones( EO, AI, AIm, AIp ),
        smallerones( EO, AD, ADm, ADp ),
        smallerones_ant_suc( EO, S, NSm, NSp ),
        append( AIm, [EO], NAI ),
        append( AIp, ADp, NAp ),
        ant( NAI, ADm, NSm, Am1, Ap1, Sm1, Sp1 ),
        append( NAp, Ap1, Ap ),
        append( NSp, Sp1, Sp ), !.
ant( AI, [], S, AI, [], S, [] ).

suc( A, SI, [E|SD], Am1, Ap, Sm1, Sp ):-
	orient_atom(E,EO),
        smallerones( EO, SI, SIm, SIp ),
        smallerones( EO, SD, SDm, SDp ),
        smallerones_suc_ant( EO, A, NAm, NAp ),
        append( SIm, [EO], NSI ),
        append( SIp, SDp, NSp ),
        suc( NAm, NSI, SDm, Am1, Ap1, Sm1, Sp1 ),
        append( NAp, Ap1, Ap ),
        append( NSp, Sp1, Sp ), !.
suc( A, SI, [], A, [], SI, [] ).

smallerones( E, [E1|L1], Lm1, [E1|Lp1] ):-
        greater_literal(+, E, +, E1 ), !,
        smallerones( E, L1, Lm1, Lp1 ).
smallerones( E, [E1|L1], [E1|Lm1], Lp1 ):-
        smallerones( E, L1, Lm1, Lp1 ), !.
smallerones( _, [], [], [] ).

smallerones_ant_suc( E, [E1|L1], Lm1, [E1|Lp1] ):-
        greater_literal( -, E, +, E1 ), !,
        smallerones_ant_suc( E, L1, Lm1, Lp1 ).
smallerones_ant_suc( E, [E1|L1], [E1|Lm1], Lp1 ):-
        smallerones_ant_suc( E, L1, Lm1, Lp1 ).
smallerones_ant_suc( _, [], [], [] ).

smallerones_suc_ant( E, [E1|L1], Lm1, [E1|Lp1] ):-
        greater_literal( +, E, -, E1 ), !,
        smallerones_suc_ant( E, L1, Lm1, Lp1 ).
smallerones_suc_ant( E, [E1|L1], [E1|Lm1], Lp1 ):-
        smallerones_suc_ant( E, L1, Lm1, Lp1 ).
smallerones_suc_ant( _, [], [], [] ).


%	printing
%------------------------------------------------------------------------------
%:- mymultifile portray/1.

user:portray( X ):-
        is_object(X,self(R1,[N,A,S])),
	nonvar(A),
	!,
	true(Ct),
	\+ \+ classes:printSimpleClause( self(R1,[N,A,S]), Ct ).
user:portray( proofClause(Cl) ):-
	\+ \+ classes:printProofClause(Cl).



% private
% -----------------------------------------------------------------------------


:- use_module(adt_term,[deBruijn/2]).

printSimpleClause( ClI , Ct):-
	ClI=self([N|_S]),
	antecedent(ClI,A2),
	succedent(ClI,S2),
	maplist(deBruijn,A,A2),
	maplist(deBruijn,S,S2),
	antecedent(A,ClI,ClI1),
	succedent(S,ClI1,Cl),

	numbervars((Cl,Ct),0,_),
	proof_depth(Cl,D),
	theory_flag(Cl,Th),
	(Th=theory -> 
	     Dot=':'
	; 
	     Dot='+'
	),
	(number(N) -> 
	     format('~t~d(~d) ~p~10| ',[N,D,Dot])
	;
	     true
	),
	split(is_true_atom,A,_,A1),
	split(is_false_atom,S,_,S1),
	print_list_atoms(A1),
	(A1\==[] -> 
	     write( ' -> ' )
	;S1=[] ->
	     write(false)
	;true),
	print_list_atoms(S1),
	print_constraint(Ct).


printProofClause( Cl1 ):-
	number(Cl1,N),
	antecedent(Cl1,A2),
	succedent(Cl1,S2),
	maplist(deBruijn,A,A2),
	maplist(deBruijn,S,S2),
	antecedent(A,Cl1,Cl3),
	succedent(S,Cl3,Cl),
	numbervars(Cl,-1,_),
	proof_depth(Cl,D),
	theory_flag(Cl,Th),
	(Th=theory -> 
	     Dot=':'
	; 
	     Dot='+'
	),
	(number(N) -> 
	     format('~n~t~d ~p~6| ',[N,Dot])
	;
	     true
	),
	split(is_true_atom,A,_,A1),
	split(is_false_atom,S,_,S1),
	print_list_atoms(A1),
	(A1\==[] -> 
	     write( ' -> ' )
	;S1=[] ->
	     write(false)
	;true),
	print_list_atoms(S1),
	(flag_on(ci,orde) ->
	     constraint(Cl,Ct),
	     print_constraint(Ct)
	;
	     true
	).

succedentTrivial(S):-
	member(E,S),
	is_trivial_atom(E),
	!.

antecedentTrivial(S):-
	member(E,S),
	is_false_atom(E),
	!.


literal_list_to_sequent([],[],[]).
literal_list_to_sequent([[-|E]|L],[E|A],S):-
	literal_list_to_sequent(L,A,S).
literal_list_to_sequent([[+|E]|L],A,[E|S]):-
	literal_list_to_sequent(L,A,S).



sequent_to_literals(A,S,L):-
	maplist(neg_atom_as_literal,A,AL),
	maplist(pos_atom_as_literal,S,SL),
	append(AL,SL,L),
	!.




neg_atom_as_literal(A,[-|A]).
pos_atom_as_literal(A,[+|A]).


negate([+|A],[-|A]).
negate([-|A],[+|A]).
