% ADT substitution tree.   Syntax: 
%
% <stree> :==	[ [V1,...V10], U, <list of trees> ]
% <tree>  :== 	[ X = T, <list of trees> ]
%		|
%		[<list of info>]

allrules( [Y1,Y2,Y3,Y4,Y5,Y6,Y7,Y8,Y9,Y10],
[U,[[ U=f(X1,X2), 
	[[X1=Y1,
		[[X2=Y1,[[g(Y1)]]],
		[X2=Y2,[[g(Y2)]]]]
	],
	[X2=g(X3),
		[[X3=Y1,
			[[X1=a,[[b]]],
			[X1=g(X4),
				[[X4=Y1,[[Y1]]],
				[X4=d,[[d]]]]
			]]
		]]
	]]
]]]).


%	f(Y1,Y1) 	--> g(Y1)
%	f(Y1,Y2) 	--> g(Y2)
%	f(a,g(Y1)) 	--> b
%	f(g(Y1),g(Y1)) 	--> Y1
%	f(g(d),g(Y1)) 	--> d

%
% operations currently available:
% 
% 	empty_rules(Rules)
% 	insert_rule( Rules, Rules1, L, R )
% 	find_unifying_rule( Rules, Query, RHS )
% 	find_matching_rule( Rules, Query, RHS )
%	display_rules( Rules )

empty_rules( [_, []] ).


insert( [Vars,U,Trees], [Vars,U,Trees1], [U=L], Info ):-
	normalize( Vars, [[U=L],Info] ),
	insert1(Trees,Trees1,[U=L],Info).


insert1([[X=T,[Infos]]|Trees],[[X=T,[Info|Infos]]|Trees],[Y=U|Subst],Info):-
	==T.
insert1([[X=T,STrees]|Trees],[[X=T,[Info|Infos]]|Trees],L,Info):-
	nonvar(L),
	nonvar(T),
	functor(L,F,N),
	functor(T,F,N),
	



normalize( Vars, T ):-
	list_of_vars( T, VT ),
	append( VT, _, Vars ).

find_unifying_rule( [_,U,R], Query, RHS ):-
	U=Query,
	member( R1, R ),
	find_unifying_rule1( R1, RHS ).

find_unifying_rule1( [Infos], Info ):-
	member(Info,Infos).
find_unifying_rule1( [X=T,R], RHS ):-
	unify( T, X ),
	member( R1, R ),
	find_unifying_rule1( R1, RHS ).



find_matching_rule( [_,U,R], Query, RHS ):-
	U = Query,
	member( R1, R ),
	find_matching_rule1( R1, RHS ), !.

find_matching_rule1( [Infos], Info ):-
	member(Info,Infos).
find_matching_rule1( [X=T,R], RHS ):-
	subsumes_chk( T, X ), T = X,
	member( R1, R ),
	find_matching_rule1( R1, RHS ).



% ********** end ADT rules *******************

% Testing area:
%

query(S):- S = f(g(a),g(a)).
query(S):- S = f(g(d),g(a)).
query(S):- S = f(a,g(a)).
query(S):- S = f(X,a).
query(S):- S = f(g(d),g(X)).
query(S):- S = f(g(d),b).

q:-	rules(R),
	query(S),
	normal_form(R,S,T),
	print_esp([S,' se reescribe en ',T]), nl, fail.

q1:-	rules(R),
	query(S),
	find_unifying_rule(R,S,RHS),
	print_esp([S,'	unifica. Parte derecha correspondiente: ',RHS]), 
	nl, fail.




		

member(X,[X|_]).
member(X,[_|L]):- member(X,L).


test:-	q(1000)

normal_form( S, T ):-
		allrules(Rules),
		subterm( Sub, S, NS, RHS ),
		find_matching_rule( Rules, Sub, RHS ),!,
		normal_form( NS, T ),!.
normal_form( T, T ).


% ====== eof ============

