%tptp2sat
%for sicstus only
:- write('%%%%%%%%%%%%% IMPORTANT: tptp2sat for sicstus only %%%%%%%%%%%%%%'),nl.

% This file is a collection of three sources:
% 1. tptpread
% 2. tptp2X
% 3. an extension to this to convert from tptp to saturate
% not as given, the first two files are treated as toolboxes for
% conversion output routines (#3).




%%%
%%%  tptpread.pl
%%%

%======================================================================
%----This module holds the clauses for reading input sets in TPTP 
%----format. This code has to be executed in the parent directory of 
%----the Problems and Axioms directories, due to the references to 
%----axiom files in the problem files.

%----Written by Geoff Sutcliffe, July, 1992.
%======================================================================
%----This specifies the parent directory of the TPTP Problems and Axioms
%----directories. To work from the current directory, use '.' here.
tptp_directory('/tmp_mnt/HG/local/provers/tptp-library/TPTP-v1.0.0').
%---------------------------------------------------------------------
%----Make a completed path name from the TPTP directory and the
%----supplied file name
%----If the pathname is user, then do nothing
tptp_path_name(user,user):-
    !.

%----If there is a leading /,~ or ., then do nothing.
tptp_path_name(FileName,FileName):-
    name(FileName,[FirstAscii|_]),
%----Would be better to use member here, but it's the only place
    (FirstAscii == 46 ;           %----  .
     FirstAscii == 47 ;           %----  /
     FirstAscii == 126),          %----  ~
    !.

%----If no leading /, then put the TPTP directory on front
tptp_path_name(FileName,PathName):-
    name(FileName,FileNameList),
    tptp_directory(Directory),
    name(Directory,DirectoryList),
    appendx(DirectoryList,[47|FileNameList],PathNameList),
    name(PathName,PathNameList).
%---------------------------------------------------------------------
%----Determine what to do with the term read.
%----Input clauses are returned in a list. Note the dictionary is 
%----embedded in another list to align with the list of lists that can
%----return from an include.
filter_for_clauses(input_clause(Name,Status,Literals),Dictionary,[input_clause(Name,Status,Literals)],[Dictionary]):-
    !.
    
%----Include directive cause recursive call to the top level to read
%----the included file.
filter_for_clauses(include(FileName),_,InputClauses,Dictionary):-
    !,
    tptp_path_name(FileName,PathName), %%%%%%%%%%%%%%% additional, see %-quotes in read_input_clauses_from_file
    read_input_clauses_from_file(PathName,InputClauses,_,Dictionary).
    
filter_for_clauses(Term,_,[],[]):-
    write('Input term '),
    write(Term),
    write(' ignored'),
    nl.
%---------------------------------------------------------------------
%----This pulls apart the dictionary and instantiates the variables
%----with their names.
instantiate_variables_in_clause([]):-
    !.

%----Otherwise instantiate this node and do sub-dictionaries
instantiate_variables_in_clause([Name=Name|RestOfList]):-
    instantiate_variables_in_clause(RestOfList).
%----------------------------------------------------------------------
%----This pulls apart the dictionary and instantiates the variables
%----with their names.
instantiate_variables_from_dictionary([]):-
    !.

%----Special case of numbervars, simply call it
instantiate_variables_from_dictionary([[numbervars=Term]|RestOfList]):-
    !,
    numbervars(Term,0,_),
    instantiate_variables_from_dictionary(RestOfList).

%----If it's a list, then it's the list for a clause. Do recursively
instantiate_variables_from_dictionary([ClauseList|RestOfList]):-
    !,
    instantiate_variables_in_clause(ClauseList),
    instantiate_variables_from_dictionary(RestOfList).
%----------------------------------------------------------------------
%----Convert a SICStus dictionary into a list of variable-name pairs
make_dictionary_from_SICStusDictionary(SICStusDictionary,[]):-
    var(SICStusDictionary),
    !.

make_dictionary_from_SICStusDictionary(dic(NameList,[Variable|_],
SmallerSICStusDictionary,LargerSICStusDictionary),
[Name=Variable|RestOfDictionary]):-
    name(Name,NameList),
    make_dictionary_from_SICStusDictionary(SmallerSICStusDictionary,
SmallerDictionary),
    make_dictionary_from_SICStusDictionary(LargerSICStusDictionary,
LargerDictionary),
    appendx(SmallerDictionary,LargerDictionary,RestOfDictionary).
%----------------------------------------------------------------------
%----Convert Sepia's dictionary to standard dictionary. Supplied by
%----Max Moser of TUM.
make_dictionary_from_Sepia_dictionary([],[]).

make_dictionary_from_Sepia_dictionary([[Name|Variable]|Rest],
[Name=Variable|RestConverted]):-
    make_dictionary_from_Sepia_dictionary(Rest,RestConverted).
%----------------------------------------------------------------------
%----This is the entry point for reading
%----Generic version (does not preserve variable names)
%real_read(Term,Dictionary):-
%    read(Term),
%----Set the dictionary to a special flag for later instantiating.
%    Dictionary = [numbervars=Term].

%----System specific reading to preserve variable names
%----If you use one of these ...
%
%    DON'T FORGET TO COMMENT OUT THE GENERIC VERSION ABOVE
%    

%----Sepia version (supplied by Max Moser of TUM)
%Sepia real_read(Term,Dictionary):-
%Sepia     readvar(input,Term,SepiaDictionary),
%Sepia     (Term == end_of_file -> 
%Sepia         SepiaDictionary=[]; 
%Sepia         true),
%Sepia     make_dictionary_from_Sepia_dictionary(SepiaDictionary,Dictionary).

%----Quintus version
%Quintus real_read(Term,Dictionary):-
%Quintus     read_term([variable_names(Dictionary)],Term).

%----SISCtus 0.7 version
%SISCtus 0.7 real_read(Term,Dictionary):-
%SISCtus 0.7    'SYSCALL'(read_top_level(Term,SICStusDictionary)),
%SISCtus 0.7    make_dictionary_from_SICStusDictionary(SICStusDictionary,
%SISCtus 0.7 Dictionary).

%----SICStus 2.1 version
real_read(Term,Dictionary):-
  current_input(InputStream),
  prolog:read_top_level(InputStream,Term,SICStusDictionary),
  make_dictionary_from_SICStusDictionary(SICStusDictionary,Dictionary).
%----------------------------------------------------------------------
%----Read in cluases from current input device, until eof
read_input_clauses(InputClauses,Dictionary):-
%----Need to deal with skolem_functor facts here, if and when I add
%----the equality stuff again
    real_read(PrologTerm,TermDictionary),
    \+ (PrologTerm == end_of_file),
    !,
%----Check if an include, and return all clauses resulting
    filter_for_clauses(PrologTerm,TermDictionary,InputClauseList,
ClauseDictionary),
    read_input_clauses(RestOfClauses,RestOfDictionary),
    appendx(InputClauseList,RestOfClauses,InputClauses),
    appendx(ClauseDictionary,RestOfDictionary,Dictionary).

read_input_clauses([],[]).
%----------------------------------------------------------------------
%----Read input clauses from a file, doing includes too
read_input_clauses_from_file(FileName,InputClauses,ProblemName,Dictionary):-
    seeing(Current_input),
    %tptp_path_name(FileName,PathName), %%%%%%% second line instead of (original) this
    FileName=PathName,                  %%%%%%% to force user to use cd
                                        %%%%%%% (cd puts default variables into current directory)
    see(PathName),
    read_input_clauses(InputClauses,Dictionary),
%----Cut, so that later errors do not try to read more input clauses
    !,
    seen,
    see(Current_input),
%----Make problem name from file name
    basename(FileName,'.p',ProblemName).
%--------------------------------------------------------------------










%%%
%%%  tptp2X.pl
%%%
%====================================================================
%----The tptp2X  utility is used  to make transformations on the 
%----clauses in TPTP problem, and to output the clauses in formats 
%----accepted  by existing ATP systems. See the associated ReadMe
%----file for details.
%----
%----Written by Geoff Sutcliffe, July 1992.
%----Updated by Geoff Sutcliffe, November 1992.
%----Updated by Geoff Sutcliffe, March 1993.
%----Revised by Geoff Sutcliffe, January, 1994.
%----Updated by Geoff Sutcliffe, April, 1994.
%----
%----Written for SICStus Prolog. Now Quintus and Sepia can use it too.
%====================================================================
%--------------------------------------------------------------------
%----These lines should be uncommented for use with Sepia Prolog/ECRC
%----Eclipse. Supplied by Max Moser from TUM.
%Sepia :-use_module(library(quintus)).
%Sepia :-local(select/3).
%Sepia :-local(halt/0).
%Sepia halt:-
%Sepia     exit(0).
%--------------------------------------------------------------------
%----These are used in the TPTP. They are also declared at runtime
%----in tptp2X/4
:-op(100,fx,++).
:-op(100,fx,--).
%----Use the standard TPTP read code
%:-compile(tptpread).
%----Load the format files
%:-compile('format.setheo').
%:-compile('format.otter').
%:-compile('format.mgtp').
%:-compile('format.pttp').
%:-compile('format.sprfn').
%:-compile('format.tptp'). 
%----Load the tranformation files
%:-compile('transform.arrange').
%:-compile('transform.shorten').
%:-compile('transform.equality').
%:-compile('transform.magic').
%--------------------------------------------------------------------
%----Simple append, required in tptpread
appendx([Head|Tail],List,[Head|TailList]):-
    !,
    appendx(Tail,List,TailList).

appendx([],List,List):-
    !.
%--------------------------------------------------------------------
%----Split (not append Tsang!)
%----This must make the front list short as possible first
split([],L,L).

split([H|T],L,[H|T1]):-
    split(T,L,T1).
%----------------------------------------------------------------------
%----Simple select
selectx(Element,[Element|Tail],Tail).

selectx(Element,[Head|Tail],[Head|SelectedTail]):-
    selectx(Element,Tail,SelectedTail).
%--------------------------------------------------------------------
%----Simple member
memberx(Element,[Element|_]).

memberx(Element,[_|Tail]):-
    memberx(Element,Tail).
%--------------------------------------------------------------------
%----Remove all elements upto the last delimiter
list_tail([],_,[],not_found):-
    !.
    
%----Check if the delimiter is further down the list
list_tail([_|ListTail],Delimiter,TailTail,found):-
    list_tail(ListTail,Delimiter,TailTail,found),
    !.

%----Check if this is the delimiter
list_tail([Delimiter|ListTail],Delimiter,ListTail,found):-
    !.

%----Otherwise pass back this list
list_tail(List,_,List,not_found).
%----------------------------------------------------------------------
%----Clauses to the same as the UNIX basename command
basename(FullName,Suffix,Basename):-
    name(FullName,FullNameList),
    name(Suffix,SuffixList),
%----Delete upto last / (ASCII 47)
    list_tail(FullNameList,47,FullNameTailList,_),
%----Try to delete suffix from list
    (split(BasenameList,SuffixList,FullNameTailList) ->
        name(Basename,BasenameList);
%----If not possible, then ignore the suffix
        name(Basename,FullNameTailList)).
%----------------------------------------------------------------------
%----Clauses the same as the UNIX dirname command
dirname(FullName,Dirname):-
    name(FullName,FullNameList),
%----Remove the basename
    basename(FullName,'',Basename),
    name(Basename,BasenameList),
%----Find what is before the basename, if anything
    (split(DirnameList,[47|BasenameList],FullNameList) ->
        name(Dirname,DirnameList);
%----If no directory then it's the current one
        Dirname = '.').
%----------------------------------------------------------------------
%-----Convert from an integer to a list of characters
integer_name(Integer,FinalString):-
    integer_name(Integer,[],FinalString).

integer_name(Integer,StringSoFar,[Character|StringSoFar]):-
    Integer < 10,
    !,
    Character is Integer + 48.

integer_name(Integer,StringSoFar,FinalString):-
    TopHalf is Integer // 10,
    BottomHalf is Integer mod 10,
    Character is BottomHalf + 48,
    integer_name(TopHalf,[Character|StringSoFar],FinalString).
%----------------------------------------------------------------------
%----Do name for atoms and integers
any_name(Atom,AtomList):-
    atom(Atom),
    !,
    name(Atom,AtomList).

any_name(Integer,IntegerList):-
    integer(Integer),
    integer_name(Integer,IntegerList).
%----------------------------------------------------------------------
%----Concatenate a list of atoms into a list of the ASCII
concatenate_atoms_to_list([],[]):-
    !.

concatenate_atoms_to_list([FirstTerm|RestOfTerms],TermsList):-
    any_name(FirstTerm,FirstList),
    concatenate_atoms_to_list(RestOfTerms,RestList),
    appendx(FirstList,RestList,TermsList).
%----------------------------------------------------------------------
%----Concatenate a list of terms into a big atom
concatenate_atoms(Atoms,Atom):-
    concatenate_atoms_to_list(Atoms,AtomsList),
    name(Atom,AtomsList).
%----------------------------------------------------------------------
%-----Fast reverse a list
fast_reverse_list([],ReverseList,ReverseList):-
    !.

fast_reverse_list([Head|Tail],ReverseSoFar,ReverseList):-
    fast_reverse_list(Tail,[Head|ReverseSoFar],ReverseList).
%----------------------------------------------------------------------
read_line_as_list(10,[]):-
    !.

read_line_as_list(LastCharacter,[LastCharacter|RestOfLine]):-
    get0(NextCharacter),
    read_line_as_list(NextCharacter,RestOfLine).

read_line(Line):-
    get0(FirstCharacter),
    read_line_as_list(FirstCharacter,LineAsList),
    atom_chars(Line,LineAsList).
%----------------------------------------------------------------------
%----Check if this is a comment line
comment_line(Line):-
    atom_chars(Line,[37,45,45,45,45,45,45,45,45|_]).
%----------------------------------------------------------------------
%----Keep reading until
read_until_comment_line(HeaderLine,[]):-
    comment_line(HeaderLine),
    !.

read_until_comment_line(TPTPLine,[Line|RestOfHeaderLines]):-
%----Remove TPTP comment character, if any
    (atom_chars(TPTPLine,[37|RestOfAscii]) ->
        atom_chars(Line,RestOfAscii);
        Line = TPTPLine),
%----Read next line
    read_line(NextLine),
    read_until_comment_line(NextLine,RestOfHeaderLines).
%----------------------------------------------------------------------
%----Read lines between two comment lines
read_header_lines(FileHeader):-
%----Get the first line and check it is a comment line
    read_line(FirstHeaderLine),
    comment_line(FirstHeaderLine),
    !,
    read_line(Line),
    read_until_comment_line(Line,FileHeader).
    
%----If the first line is not a comment then return nothing
read_header_lines([]).
%----------------------------------------------------------------------
%----Read the header from a TPTP file, one line at a time
read_input_file_header(TPTPFileName,FileHeader):-
    %tptp_path_name(TPTPFileName,TPTPPathName), %%%%%%% second line instead of (original) this
    TPTPFileName=TPTPPathName,                  %%%%%%% to force user to use cd
                                                %%%%%%% (cd puts default variables into current directory)
%----Redirect input
    seeing(CurrentInput),
    see(TPTPPathName),
    read_header_lines(FileHeader),
%----Reset input to old input
    seen,
    see(CurrentInput).
%---------------------------------------------------------------------
%----Write a comment line
output_comment_line(CommentAtom):-
    write(CommentAtom),
    write('-------------------------------------'),
    write('-------------------------------------'),
    nl.
%----------------------------------------------------------------------
%----Output the file header using the given comment atom
output_file_header([],_):-
    !.

output_file_header([FirstHeaderLine|RestOfHeaderLines],CommentAtom):-
    write(CommentAtom),
    write(FirstHeaderLine),
    nl,
    output_file_header(RestOfHeaderLines,CommentAtom).
%----------------------------------------------------------------------
%----Make the output file name
%----If the output directory is user, then output to stdout
make_output_file_name(user,_,_,user):-
    !.

make_output_file_name(OutputDirectory,OutputName,FileNameExtension,
OutputFileName):-
    concatenate_atoms([OutputDirectory,'/',OutputName,FileNameExtension],
OutputFileName).
%----------------------------------------------------------------------
%----Format the clauses and write to file
output_clauses_in_format(FileHeader,OutputClauses,OutputFormat,
OutputName,OutputDirectory):-
%----Extract the main name of the output format
    OutputFormat =.. [FormatName|_],
%----Get the file name extension (don't like this yet)
    concatenate_atoms([FormatName,'_format_information'],
ExtensionQueryPredicate),
    ExtensionQuery =.. [ExtensionQueryPredicate,FileNameExtension,
CommentAtom],
    ExtensionQuery,
%----Make the output file name
    make_output_file_name(OutputDirectory,OutputName,FileNameExtension,
OutputFileName),
%----Make the output query
    Query =.. [FormatName,OutputClauses,OutputFormat],
%----Redirect the output
    telling(CurrentOutput),
    tell(OutputFileName),
%----Output the file header
    output_comment_line(CommentAtom),
    output_file_header(FileHeader,CommentAtom),
    output_comment_line(CommentAtom),
%----Output the clauses in the required format
    Query,
%----Put a comment line
    output_comment_line(CommentAtom),
%----Restore output direction
    told,
    tell(CurrentOutput).
%---------------------------------------------------------------------
%----Control of a single transformation, including output.
do_transformation(InputClauses,InputName,Dictionary,Transformation,OutputClauses,OutputName,NewDictionary):-
%----Extract the main name of the transformation
    Transformation =.. [TransformationName|_],
    Query =.. [TransformationName,InputClauses,Dictionary,
Transformation,OutputClauses,NewDictionary,TransformationSuffix],
    Query,
    !,
    concatenate_atoms([InputName,TransformationSuffix],OutputName).

do_transformation(_,_,_,Transformation,_,_,_):-
    write('The tranformation '),
    write(Transformation),
    write(' failed'),
    nl,
    fail.
%---------------------------------------------------------------------
%----Do one transformation that is a sequence in the transformations
%----list.
%----When no more transformation to do, then copy to output
transform_clauses(Clauses,Name,Dictionary,[],Clauses,Name,Dictionary):-
    !.
    
%----If a list of transformations, then do each one. Use member due
%----to backtracking approach
transform_clauses(InputClauses,InputName,Dictionary,[[OneTransformation|OtherTransformations]|RestOfTransformations],OutputClauses,OutputName,NewDictionary):-
    !,
    memberx(Transformation,[OneTransformation|OtherTransformations]),
    do_transformation(InputClauses,InputName,Dictionary,Transformation,IntermediateClauses,IntermediateName,IntermediateDictionary),
    transform_clauses(IntermediateClauses,IntermediateName,IntermediateDictionary,RestOfTransformations,OutputClauses,OutputName,NewDictionary).

%----If a definite transformation, then do it
transform_clauses(InputClauses,InputName,Dictionary,[Transformation|RestOfTransformations],OutputClauses,OutputName,NewDictionary):-
    do_transformation(InputClauses,InputName,Dictionary,Transformation,IntermediateClauses,IntermediateName,IntermediateDictionary),
    transform_clauses(IntermediateClauses,IntermediateName,IntermediateDictionary,RestOfTransformations,OutputClauses,OutputName,NewDictionary).
%---------------------------------------------------------------------
%----Do a transformation by combination from the transformation list.
%----Done by backtracking over the transformations.
do_all_transformations_and_output(InputClauses,BaseInputName,Transformations,FileHeader,OutputFormat,OutputDirectory,Dictionary):-
    transform_clauses(InputClauses,BaseInputName,Dictionary,Transformations,OutputClauses,BaseOutputName,NewDictionary),	
    instantiate_variables_from_dictionary(NewDictionary),	
    output_clauses_in_format(FileHeader,OutputClauses,OutputFormat,BaseOutputName,OutputDirectory),
    fail.

%----Succeed when all transformation have been backtracked over
do_all_transformations_and_output(_,_,_,_,_,_,_).
%---------------------------------------------------------------------
%----The entry point to convert a TPTP file to other formats
tptp2X(TPTPFileName,Transformations,OutputFormat,OutputDirectory):-
%----These operators are used in the TPTP. They must be declared here
%----so that they are declared at run, rather than compile time. If 
%----they are only done as :-op(100,fx,++)., then compiled versions of 
%----this code will not work
    op(100,fx,++),
    op(100,fx,--),
%----Read in the file header
    read_input_file_header(TPTPFileName,FileHeader),
%----Read in the clauses
    read_input_clauses_from_file(TPTPFileName,InputClauses,_,
Dictionary),
%----Extract the base file name
    basename(TPTPFileName,'.p',BaseInputName),
%----Do each transformation and output the clauses in the required
%----format. 
    do_all_transformations_and_output(InputClauses,BaseInputName,Transformations,FileHeader,OutputFormat,OutputDirectory,Dictionary).


tptp2X:-
    write('Enter TPTP file name         : '),
    read(TPTPFileName),
    write('Available transformations are lr, cr, clr, shorten, magic,
    equality(<keep> (<keep> = reflexivity or none)'),
    nl,
    write('Enter transformations list   : '),
    read(Transformations),
    write('Available output formats are mgtp, pttp, setheo, sprfn, tptp, and
    otter(<sos>,<otter options>), where <sos> is one of theorem, hypothesis, 
    positive, negative, unit, none and <otter options> is a list of Otter 
    option directives.'),
    nl,
    write('Enter output format          : '),
    read(OutputFormat),
    write('Enter output directory       : '),
    read(OutputDirectory),
    tptp2X(TPTPFileName,Transformations,OutputFormat,OutputDirectory).
%---------------------------------------------------------------------
%----Get the Transformations and output format from tptp2Xrc file
%----First look locally
get_transformation_and_output_directory(Transformations,OutputFormat):-
    see(tptp2Xrc),
    read(Transformations),
    read(OutputFormat),
    seen,
    !.
    
%----Second look in tptp2X directory
get_transformation_and_output_directory(Transformations,OutputFormat):-
    tptp_path_name('TPTP2X/tptp2Xrc',PathName),
    see(PathName),
    read(Transformations),
    read(OutputFormat),
    seen,
    !.

%----Default is TPTP format with no transformation
get_transformation_and_output_directory([],tptp).
%---------------------------------------------------------------------
%----This executes tptp2X as a filter, with the tptp2Xrc file
execute_filter:-
    get_transformation_and_output_directory(Transformations,
OutputFormat),
    tptp2X(user,Transformations,OutputFormat,user),
    halt.
%---------------------------------------------------------------------
%----Make a saved state filter.
%----There is no generic version, but the SICStus version is common

%----SICStus version.
make_tptp2X_filter:-
  nofileerrors,
  save(tptp2X_filter),
  execute_filter.

%----Sepia version
%Sepia make_tptp2X_filter:-
%Sepia     nofileerrors,
%Sepia     save(tptp2X_filter),
%Sepia     execute_filter.

%----Quintus version
%Quintus make_tptp2X_filter:-
%Quintus     save_program(tptp2X_filter,execute_filter).
%---------------------------------------------------------------------












%%%
%%%  tptp2sat.pl  (main extension)
%%%

%:- compile(tptp2X). %special versions of tptp2X and tptpread, extended versions of files distributed with tptp

output_literal(L):-
	L = equal(X1,X2), !,
	write(X1=X2).
output_literal(L):-
	\+ L = equal(_,_), !,
	write(L).

output_neg_literals([--L])    :- output_literal(L).
output_neg_literals([--L|Ls]) :- \+ member(--_,Ls), output_literal(L).
output_neg_literals([--L|Ls]) :- member(--_,Ls), output_literal(L), write(', '), output_neg_literals(Ls).
output_neg_literals([++_]).
output_neg_literals([++_|Ls]) :- \+ Ls==[], output_neg_literals(Ls).

output_pos_literals([++L])    :- output_literal(L).
output_pos_literals([++L|Ls]) :- \+ member(++_,Ls), output_literal(L).
output_pos_literals([++L|Ls]) :- member(++_,Ls), output_literal(L), write(', '), output_pos_literals(Ls).
output_pos_literals([--_]).
output_pos_literals([--_|Ls]) :- \+ Ls==[], output_pos_literals(Ls).

output_clause(C):-
	\+ member(--_,C), !,
	output_pos_literals(C), write('.').
output_clause(C):-
	\+ member(++_,C), !,
	output_neg_literals(C), write(' -> [].').
output_clause(C):-
	member(++_,C), member(--_,C), !,
	output_neg_literals(C), write(' -> '), output_pos_literals(C), write('.').

output_axiom(C):-
	write('axiom( [ ['),
	output_neg_literals(C), write('],'), put(9), write('['), output_pos_literals(C),
	write('] ] ).').

output_clauses(_,[]).
output_clauses(inf,[input_clause(Cname,Ckind,Cclause)|Clauses]):-
	write('%'), write(Ckind), write(': '), write(Cname), nl,
        output_clause(Cclause), nl, nl,
	output_clauses(inf,Clauses).
output_clauses(ax,[input_clause(Cname,Ckind,Cclause)|Clauses]):-
        output_axiom(Cclause), 	put(9), write('%'), write(Ckind), write(': '), write(Cname), nl,
	output_clauses(ax,Clauses).

output_prop(String,List):-
	write(String), write('('), write(List), write(').'), nl.

term_arglist(0,_,[]).
term_arglist(1,T,[X]):-arg(1,T,X).
term_arglist(N,T,XXX):-
	N>1,
	N1 is N-1,
	term_arglist(N1,T,XX),
	arg(N,T,X),
	appendx(XX,[X],XXX).

term_prop(Term,[]):-
	var(Term).
term_prop(Term,[o:Name/Arity|Rest]):-
	nonvar(Term),
	functor(Term,Name,Arity),
	term_arglist(Arity,Term,Args),
	terms_prop(Args,Rest).

terms_prop([H|T],P):-
	term_prop(H,HP),
	terms_prop(T,TP),
	appendx(HP,TP,P).
terms_prop([],[]).

clause_to_proplist([],[]).
clause_to_proplist([++H|T],X):-
	clause_to_proplist([--H|T],X).
clause_to_proplist([--H|C],[p:N/A|Props]):-
	H=..[N|Arg],
	length(Arg,A),
	terms_prop(Arg,TermProp),
	clause_to_proplist(C,ClauseProp),
	appendx(TermProp,ClauseProp,Props).

clauses_to_prop([],[]).
clauses_to_prop([input_clause(_,_,C)|Clauses],Props):-
	clause_to_proplist(C,P),
	clauses_to_prop(Clauses,Prop),
	appendx(P,Prop,Props).

neat_and_clean([],[]).
neat_and_clean([H|T1],T2):-
	member(H,T1),
	neat_and_clean(T1,T2).
neat_and_clean([H|T1],[H|T2]):-
	\+member(H,T1),
	neat_and_clean(T1,T2).

properties_of_clauses(Clauses,Props):-
	clauses_to_prop(Clauses,P),
	neat_and_clean(P,Props).


xofprop(_,[],[]).
xofprop(X,[X:N/A|T],[[A,N]|TT]):-
	xofprop(X,T,TT).
xofprop(X,[Y:_/_|T],TT):-
	\+ X=Y,
	xofprop(X,T,TT).

reverse_unkey([],[]).
reverse_unkey([[_,H]|T],X):-
	reverse_unkey(T,X1),
	appendx(X1,[H],X).

precedence_of_props(Props,Prec):-
	xofprop(o,Props,O),
	sort(O,OS),
	reverse_unkey(OS,Prec).

first_predicate_precedence_of_props(Props,Prec):-
	xofprop(p,Props,P),
	sort(P,PS),
	reverse_unkey(PS,Prec).

predicates_of_props([],[]).
predicates_of_props([p:N/_|T],[N|TT]):-
	predicates_of_props(T,TT).
predicates_of_props([o:_/_|T],TT):-
	predicates_of_props(T,TT).

output_transform_sat(BaseInputName,FileHeader,InputClauses,Dictionary,Rep):-
	properties_of_clauses(InputClauses,Props), !,
        precedence_of_props(Props,Precedence),
        predicates_of_props(Props,Predicates),
        first_predicate_precedence_of_props(Props,PredPrec),
	seeing(OldFile), see(user),
	write('%'), write(BaseInputName), write('; converted by tptp2sat'), nl,
        output_comment_line('%'),
        output_file_header(FileHeader,'%'),
        output_comment_line('%'),
	output_prop('predicates',Predicates), nl,
        output_prop('precedence',Precedence), nl,
	output_prop('first_predicate_precedence',PredPrec), nl,
	nl,
        instantiate_variables_from_dictionary(Dictionary),
        output_clauses(Rep,InputClauses),
	seen, seeing(OldFile).

tptp2sat(InFileName,Rep,OutFileName):-
	%These operators are used in the TPTP.
	%They must be declared here so that they are declared at run, rather than compile time. If 
	%they are only done as :-op(100,fx,++)., then compiled versions of this code will not work
	op(100,fx,++),
	op(100,fx,--),
	%Read in the file header
	read_input_file_header(InFileName,FileHeader),
	%Read in the clauses
	read_input_clauses_from_file(InFileName,InputClauses,_,Dictionary),
	%Extract the base file name
	basename(InFileName,'.p',BaseInputName),
	%Do each transformation and output the clauses in the required format.
	telling(OldOutput), tell(OutFileName),
	output_transform_sat(BaseInputName,FileHeader,InputClauses,Dictionary,Rep),
	told, tell(OldOutput).

intptp([N1,N2,N3,N4,N5,N6,-|NT]):-
	tptp_directory(X),
	appendx(X,[/,N1,N2,N3],Path),
	cd(Path),
	intptp([N1,N2,N3,N4,N5,N6,-|NT]).

intptp(F):-
	tptp2sat(F,inf,'~/tptp.tmp'), !,
	%Show the file header
	read_input_file_header(F,FileHeader), output_file_header(FileHeader,'%'),
	in('~/tptp.tmp').

addtptp(F):-
	tptp2sat(F,inf,'~/tptp.tmp'), !,
	add('~/tptp.tmp').

cdtptp:-
	tptp_directory(X),
	cd(X).
	




