%declarations%

:- module_interface('lpo-protein').

:- export
	greater/2,
	greaterEqual/2,
	simplify_constraint/3.



/* 

:- use_module(library(occurs), [contains_var/2]).

:- use_module(adt_term, [term_size/2]).


*/

%end%


:- begin_module('lpo-protein').

:- import lpo_precedence/2 from protein_proof.
:- import is_first_predicate/1 from protein_proof.

%------------------------------------------------------------------------------
% greater( S, T ):
%	S is greater than T with respect to the lexicographic path ordering
%------------------------------------------------------------------------------


% new implementation

greater( S, _):-
	var(S),
	!,
	fail.
greater( S, T ):-
	\+var(T),
	!,
	functor(S,F,N),
	functor(T,G,M),
	greater(F,G,N,M,S,T).
greater( S, T ):-
	% T is a variable
	functor(S,F,_),
	(is_first_predicate(F) -> 
	     true
	;    contains_var(T,S)),
	!.
	



greater(F,F,N,M,S,T):-
	!,
        (is_first_predicate(F) ->
             lexicographic1(0,S,N,T,M)
        ;
             lexicographic(0,S,N,T,M)
	).
greater(F,G,N,M,S,T):-
        (greater_op(F,G)->
             greater_than_list(S,0,T,M)

        ; % greater_op(G,F)
             some_subterm_greater_equal(0,S,N,T)
        ),
        !.



some_subterm_greater_equal(N,_S,N,_T):-
	!,
	fail.
some_subterm_greater_equal(I,S,N,T):-
	I1 is I+1,
	arg(I1,S,SI1),
	(greaterEqual( SI1, T) ->
	     true
	;
	     some_subterm_greater_equal(I1,S,N,T)
	).

	




lexicographic1(N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic1(I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg(I1,S,SI),   % otherwise: N=<M
	(arg(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic1(I1,S,N,T,M)
	     ;
	          greater(SI,TI)
	     )
	;% M<N
	     true
	).
	

lexicographic(N,_,N,_,_M):-
	!,  % here : S[j]==T[j], for 1=<j=<N and N=<M
	fail.
lexicographic(I,S,N,T,M):-
	% here : S[j]==T[j], for 1=<j=<I,  and I=<M and I<N
	I1 is I+1,
	arg(I1,S,SI),   % otherwise: N=<M
	(arg(I1,T,TI) ->
	     (SI==TI ->
		  lexicographic(I1,S,N,T,M)
	     ;greater(SI,TI) ->
		  greater_than_list(S,I1,T,M)
	     ;
                  some_subterm_greater_equal(I1,S,N,T)
	     )
	;% M<N     
	     true
	).

greater_than_list(_,N,_,N):-
	!.
greater_than_list(S,I,T,M):-
	I1 is I+1,
	arg(I1,T,TI),
	greater(S,TI),
	greater_than_list(S,I1,T,M).








%------------------------------------------------------------------------------
% greaterEqual( S, T ):
%	S is greater than T with respect lpo or S is equal to T
%------------------------------------------------------------------------------

greaterEqual(T1,T2):- T1 == T2,!.
greaterEqual(T1,T2):- greater(T1,T2),!.







% auxiliary



greater_op(A,B):-
	lpo_precedence(A,P),
	lpo_precedence(B,Q),
	P>Q.





%   contains_var(+Variable, +Term)
%   is true when the given Term contains at least one sub-term which
%   is identical to the given Variable.  We use '=='/2 to check for
%   the variable (contains_term/2 uses '=') so it can be used to check
%   for arbitrary terms, not just variables.

contains_var(Variable, Term) :-
	\+ free_of_var(Variable, Term).


%   free_of_var(+Variable, +Term)
%   is true when the given Term contains no sub-term identical to the
%   given Variable (which may actually be any term, not just a var).
%   For variables, this is precisely the "occurs check" which is
%   needed for sound unification.

free_of_var(Variable, Term) :-
	Term == Variable,
	!,
	fail.
free_of_var(Variable, Term) :-
	compound(Term),
	!,
	functor(Term, _, Arity),
	free_of_var(Arity, Term, Variable).
free_of_var(_, _).

free_of_var(1, Term, Variable) :- !,
	arg(1, Term, Argument),
	free_of_var(Variable, Argument).
free_of_var(N, Term, Variable) :-
	arg(N, Term, Argument),
	free_of_var(Variable, Argument),
	M is N-1, !,
	free_of_var(M, Term, Variable).




simplify_constraint(nc,_C,[conj([],[],[])]).
simplify_constraint(med,C,C1):-
	simplify_med(C,C1).
simplify_constraint(full,C,C1):-
	simplify(C,C1).
simplify_constraint(light,C,C1):-
	simplify_constraint(C,C1).



simplify_constraint([conj([],E,O)],[conj([],E,O)]).
simplify_constraint([conj([S>=T|C],E,O)],C1):-
	(greaterEqual(S,T) ->
	     simplify_constraint([conj(C,E,O)],C1)
	;greater(T,S) ->
	     C1=[]
	;% uncomparable
	     simplify_constraint([conj(C,E,O)],[conj(C2,E,O)]),
	     C1=[conj([S>=T|C2],E,O)]
	).
simplify_constraint([conj([S>T|C],E,O)],C1):-
	(greater(S,T) ->
	     simplify_constraint([conj(C,E,O)],C1)
	;greaterEqual(T,S) ->
	     C1=[]
	;% uncomparable
	     simplify_constraint([conj(C,E,O)],[conj(C2,E,O)]),
	     C1=[conj([S>=T|C2],E,O)]
	).
